/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <jni.h>
#include <vector>
#include <map>
#include "STAFServiceInterface.h"
#include "STAFJavaService.h"
#include "STAFLocalConnection.h"
#include "STAFUtil.h"
#include "STAFThread.h"
#include "STAFEventSem.h"
#include "STAFMutexSem.h"
#include "STAFProcess.h"
#include "STAFFileSystem.h"
#include "STAF_fstream.h"

struct JVMData
{
    STAFString fName;
    STAFString fExec;
    STAFString fOptions;
    STAFString fIPCName;
    STAFEventSemPtr fJVMExitedSem;
    STAFProcessID_t fJVM_PID;
    unsigned int fNumServices;
};

typedef STAFRefPtr<JVMData> JVMDataPtr;
typedef std::map<STAFString, JVMDataPtr> JVMDataMap;

struct JVMStartThreadData
{
    JVMStartThreadData (STAFString_t startString, STAFEventSemPtr &exitedSem) :
        fStartString(startString), fJVMExitedSem(exitedSem)
    { /* Do nothing */ }

    STAFString_t fStartString;
    STAFEventSemPtr fJVMExitedSem;
};

struct STAFProcJavaServiceData
{
    STAFString fName;
    STAFString fExec;
    JVMDataPtr fJVM;
};

static JVMDataMap sJVMDataMap;
static STAFMutexSem sJVMDataSem;

STAFRC_t STAFServiceGetLevelBounds(unsigned int levelID,
                                   unsigned int *minimum,
                                   unsigned int *maximum)
{
    switch (levelID)
    {
        case kServiceInfo:
        {
            *minimum = 2;
            *maximum = 2;
            break;
        }
        case kServiceInit:
        {
            *minimum = 1;
            *maximum = 1;
            break;
        }
        case kServiceAcceptRequest:
        {
            *minimum = 2;
            *maximum = 2;
            break;
        }
        case kServiceTerm:
        case kServiceDestruct:
        {
            *minimum = 0;
            *maximum = 0;
            break;
        }
        default:
        {
            return kSTAFInvalidAPILevel;
        }
    }

    return kSTAFOk;
}


unsigned int STAFDoShutdownJVM(STAFString &jvmIPCName)
{
    try
    {
        STAFConnAllocInfo allocInfo =
                          STAFLocalConnection::create(jvmIPCName);
        STAFConnPtr connPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                            STAFConnPtr::CUSTOM, allocInfo.pFreeFunc);

        connPtr->writeUInt(JAVA_SERVICE_JVMEXIT);

        STAFRC_t jvmRC = connPtr->readUInt();
        STAFString jvmResultString = connPtr->readString();

        // Note, this last connect is required depending on the thread timing
        // in STAFJavaServiceHelper.cpp.   In that file, if the request
        // thread gets started before the creating thread gets control back,
        // then this call is unnecessary.  Otherwise, we need to send this
        // final request so that the main thread will be able to check the
        // exit flag.

        STAFConnAllocInfo allocInfo2 =
                          STAFLocalConnection::create(jvmIPCName);
        STAFConnPtr conn2Ptr(allocInfo2.pConnection, STAFConnPtr::INIT,
                             STAFConnPtr::CUSTOM, allocInfo2.pFreeFunc);

        conn2Ptr->writeUInt(JAVA_SERVICE_JVMFIN);
    }
    catch (STAFConnectionException)
    {
        // These execptions are anticipated in cases where the JVM has
        // already shutdown before we send the shutdown commands
    }

    return 0;
}


// Note: You should already own sJVMDataSem before calling STAFShutdownJMV

unsigned int STAFShutdownJVM(STAFString &jvmName)
{
    try
    {
        JVMDataPtr jvm = sJVMDataMap[jvmName];

        sJVMDataMap.erase(jvmName);

        STAFDoShutdownJVM(jvm->fIPCName);
    }
    catch (STAFException &e)
    {
        e.write("JSTAF.STAFShutdownJVM()");
    }
    catch (...)
    {
        cout << "Caught unknown exception in JSTAF.STAFShutdownJVM()" << endl;
    }

    return 0;
}


STAFRC_t STAFServiceConstruct(STAFServiceHandle_t *pServiceHandle,
                              void *pServiceInfo, unsigned int infoLevel,
                              STAFString_t *pErrorBuffer)
{
    try
    {
        if (infoLevel != 2) return kSTAFInvalidAPILevel;

        STAFProcJavaServiceData data;
        STAFServiceInfoLevel2 *pInfo =
            static_cast<STAFServiceInfoLevel2 *>(pServiceInfo);
        STAFString jvmName = "STAFJVM1";
        STAFString jvmExec = "java";
        STAFString jvmOptions = " ";
        bool jvmSpecified = false;
        bool j2Specified = false;
        unsigned int maxLogs = 5;    // Defaults to keeping 5 JVM logs
        long maxLogSize = 1048576;   // Default size of each log is 1M

        data.fName = pInfo->name;
        data.fExec = pInfo->exec;

        // Walk through and verify the config options

        for (unsigned int i = 0; i < pInfo->numOptions; ++i)
        {
            STAFString upperOptionName =
                       STAFString(pInfo->pOptionName[i]).upperCase();
            STAFString optionValue(pInfo->pOptionValue[i]);

            if (upperOptionName == "JVMNAME")
            {
                jvmName = optionValue;
            }
            else if (upperOptionName == "JVM")
            {
                jvmExec = optionValue;
                jvmSpecified = true;
            }
            else if (upperOptionName == "J2")
            {
                jvmOptions += optionValue + " ";
                j2Specified = true;
            }
            else if (upperOptionName == "MAXLOGS")
            {
                // Check to make sure it is an integer value > 0
                
                STAFString maxLogsStr = optionValue;

                if (maxLogsStr.isDigits() && (maxLogsStr.asUInt() > 0))
                {
                    maxLogs = maxLogsStr.asUInt();
                }
                else
                {
                    STAFString errmsg("Error JVM using JVMName: " + 
                        jvmName + ", JVM: " + jvmExec + ", JVMOptions: " +
                        jvmOptions + ", RC: " + STAFString(kSTAFInvalidValue) +
                        ", MAXLOGS must be an positive integer");
                    *pErrorBuffer = errmsg.adoptImpl();
                    return kSTAFServiceConfigurationError;
                }
            }
            else if (upperOptionName == "MAXLOGSIZE")
            {
                // Check to make sure it is an integer value > 0
                
                STAFString maxLogSizeStr = optionValue;

                if (maxLogSizeStr.isDigits() && (maxLogSizeStr.asUInt() > 0))
                {
                    maxLogSize = (long)maxLogSizeStr.asUInt();
                }
                else
                {
                    STAFString errmsg("Error JVM using JVMName: " + 
                        jvmName + ", JVM: " + jvmExec + ", JVMOptions: " +
                        jvmOptions + ", RC: " + STAFString(kSTAFInvalidValue) +
                        ", MAXLOGSIZE must be an positive integer");
                    *pErrorBuffer = errmsg.adoptImpl();
                    return kSTAFServiceConfigurationError;
                }
            }
            else
            {
                STAFString optionError(pInfo->pOptionName[i]);
                *pErrorBuffer = optionError.adoptImpl();
                return kSTAFServiceConfigurationError;
            }
        }

        // Check to see if the specified JVM already exists

        STAFMutexSemLock lock(sJVMDataSem);

        if (sJVMDataMap.find(jvmName) == sJVMDataMap.end())
        {
            // The JVM doesn't exist so we need to start it

            JVMData jvmData;

            jvmData.fName = jvmName;
            jvmData.fExec = jvmExec;
            jvmData.fOptions = jvmOptions;
            jvmData.fIPCName = "JSTAF_" + jvmName;
            jvmData.fJVMExitedSem = STAFEventSemPtr(new STAFEventSem,
                                                    STAFEventSemPtr::INIT);
            jvmData.fNumServices = 0;

            STAFString jvmStartString = jvmOptions +
                       "com.ibm.staf.service.STAFServiceHelper " + jvmName;

            // We need to shutdown any JVM that might happen to be leftover
            // from a "bad" exit from a previous STAFProc

            STAFDoShutdownJVM(jvmData.fIPCName);
            
            // We need to capture stdout/stderr for diagnostic purposes
            // Create directory for JVM log file if doesn't already exist

            STAFFSPath logPath;
            logPath.setRoot(pInfo->writeLocation);
            logPath.addDir("JSTAF");
            logPath.addDir(jvmName);

            if (!logPath.exists())
            {
                try
                {
                    // Don't want exceptions here
                    STAFFSEntryPtr dir = 
                        logPath.createDirectory(kSTAFFSCreatePath);
                }
                catch (...)
                { /* Do Nothing */ }
            }

            // Name the current JVM log file - JVMLog.1

            logPath.setName("JVMLog");
            logPath.setExtension("1");
            STAFString logName = logPath.asString();
            
            // Instead of replacing the JVM log file each time a JVM is created,
            // use the following JVM OPTIONs to determine when to create a new
            // JVM log file and how many JVM log files to save:
            // - MaxLogs   : Maximum number of JVM log files to keep.
            // - MaxLogSize: Maximum size of a JVM log file in bytes.
            
            // Open the JVM log file
                
            fstream outfile(logName.toCurrentCodePage()->buffer(),
                            ios::out | ios::app);

            if (!outfile)
            {
                STAFString errmsg("Error JVM using JVMName: " + 
                    jvmName + ", JVM: " + jvmExec + ", JVMOptions: " +
                    jvmOptions + ", RC: " + STAFString(kSTAFFileOpenError) +
                    ", Error opening file " + logName);
                *pErrorBuffer = errmsg.adoptImpl();
                return kSTAFServiceConfigurationError;
            }

            // Figure out how big the JVM log file is

            outfile.seekp(0, ios::end);
            unsigned int fileLength = (unsigned int)outfile.tellp();

            if (fileLength > maxLogSize)
            {
                // Roll any existing log files (e.g. Rename JVMLog.2.out to
                // JVMLog.3, JVMLog.1 to JVMLog.2, etc) and create a new
                // JVMLog.1 file.  If the # of existing logs > MAXLOGS, don't
                // save the oldest log.

                outfile.close();

                STAFFSPath fromLogPath(logPath);
                
                for (int i = maxLogs; i > 0; --i)
                {
                    fromLogPath.setExtension(STAFString(i));

                    if (fromLogPath.exists() && i < maxLogs)
                    {
                        // Rename JVMLog.<i> to JVMLog.<i+1>

                        STAFFSPath toLogPath(fromLogPath);
                        toLogPath.setExtension(STAFString(i + 1));

                        fromLogPath.getEntry()->move(toLogPath.asString());
                    }
                }
                
                // Open a new empty current log file

                outfile.open(logName.toCurrentCodePage()->buffer(),
                             ios::out | ios::trunc);

                if (!outfile)
                {
                    STAFString errmsg("Error JVM using JVMName: " + 
                        jvmName + ", JVM: " + jvmExec + ", JVMOptions: " +
                        jvmOptions + ", RC: " + STAFString(kSTAFFileOpenError) +
                        ", Error opening file " + logName);
                    *pErrorBuffer = errmsg.adoptImpl();
                    return kSTAFServiceConfigurationError;
                }
            }

            // Write the JVM start information to the JVM log file

            STAFString separatorLine("***************************************"
                                     "***************************************");
            STAFString line1("*** " + STAFTimestamp().asString() +
                                  " - Start of Log for JVMName: " + jvmName);
            STAFString line2("*** JVM Executable: " + jvmExec);
            STAFString line3("*** JVM Options   :");
            if (jvmOptions != STAFString(" "))
                line3 += jvmOptions;
            else
                line3 += " none";

            outfile << separatorLine.toCurrentCodePage()->buffer() << endl 
                    << line1.toCurrentCodePage()->buffer() << endl
                    << line2.toCurrentCodePage()->buffer() << endl
                    << line3.toCurrentCodePage()->buffer() << endl
                    << separatorLine.toCurrentCodePage()->buffer() << endl;

            outfile.close();

            // Start a process for the JVM

            STAFProcessStartInfoLevel1 startInfo = { 0 };

            startInfo.command     = jvmExec.getImpl();
            startInfo.parms       = jvmStartString.getImpl();
            startInfo.consoleMode = kSTAFProcessSameConsole;

            startInfo.stdoutMode  = kSTAFProcessIOAppendFile;
            startInfo.stderrMode  = kSTAFProcessIOStdout;

            startInfo.stdoutRedirect = logName.getImpl();
            
            unsigned int osRC = 0;
            STAFRC_t rc = STAFProcessStart(0, 0, &startInfo, 1, &osRC);

            if (rc != kSTAFOk)
            {
                STAFString startError("Error starting JVM using JVMName: " +
                                      jvmName + ", JVM: " + jvmExec +
                                      ", JVMOptions: " + jvmOptions +
                                      ", RC: " + STAFString(rc) +
                                      ", OSRC: " + STAFString(osRC));

                // Add more details to the error msg when the Java executable
                // cannot be found.

                if (rc == 10)
                {
                    if (!jvmSpecified)
                    {
                        startError = startError + ".  Make sure the java " +
                            "executable is in the PATH.";
                    }
                    else
                    {
                        startError = startError + ".  Make sure " + jvmExec +
                            " exists.";
                    }
                }

                *pErrorBuffer = startError.adoptImpl();
                return kSTAFServiceConfigurationError;
            }

            // Now we need to wait for it to start

            bool jvmReady = false;

            for (int i = 0; (i < 30) & !jvmReady; ++i)
            {
                // XXX: Need to check to see if the JVM is actually still running

                try
                {
                    // First connect to the JVM

                    STAFConnAllocInfo allocInfo =
                        STAFLocalConnection::create(jvmData.fIPCName);
                    STAFConnPtr connPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                                        STAFConnPtr::CUSTOM,
                                        allocInfo.pFreeFunc);

                    // Now see if it's alive

                    connPtr->writeUInt(JAVA_SERVICE_JVMPING);

                    STAFRC_t jvmRC = connPtr->readUInt();
                    STAFString jvmResultString = connPtr->readString();

                    if (jvmRC != kSTAFOk)
                    {
                        jvmResultString = "Error starting JVM: " +
                                          jvmResultString;
                        *pErrorBuffer = jvmResultString.adoptImpl();
                        return kSTAFServiceConfigurationError;
                    }

                    jvmReady = true;
                }
                catch (STAFException e)
                {
                    // XXX: This should really be more specific
                }

                if (!jvmReady) STAFThreadSleepCurrentThread(1000, 0);
            }

            // If we didn't time out waiting for the JVM then we add the
            // JVM to the map of JVMs

            if (!jvmReady)
            {
                *pErrorBuffer =
                    STAFString("Unable to connect to JVM").adoptImpl();
                return kSTAFServiceConfigurationError;
            }

            sJVMDataMap[jvmName] = JVMDataPtr(new JVMData(jvmData),
                                              JVMDataPtr::INIT);
        }
        else if (jvmSpecified || j2Specified)
        {
            *pErrorBuffer = STAFString("You may not specify the JVM or J2 "
                                       "options without specifying a new "
                                       "JVMNAME").adoptImpl();
            return kSTAFServiceConfigurationError;
        }

        // Ok. We have a JVM.  Now let's try to load the service.

        data.fJVM = sJVMDataMap[jvmName];

        STAFConnAllocInfo allocInfo =
                          STAFLocalConnection::create(data.fJVM->fIPCName);

        STAFConnPtr connPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                            STAFConnPtr::CUSTOM, allocInfo.pFreeFunc);

        connPtr->writeUInt(JAVA_SERVICE_LOAD);
        connPtr->writeString(data.fName);
        connPtr->writeString(data.fExec);
        connPtr->writeString(pInfo->writeLocation);

        unsigned int constructRC = connPtr->readUInt();
        STAFString constructResult = connPtr->readString();

        if (constructRC != kSTAFOk)
        {
            if (data.fJVM->fNumServices == 0) STAFShutdownJVM(jvmName);

            *pErrorBuffer = constructResult.adoptImpl();
            return constructRC;
        }

        // The service is now loaded

        STAFProcJavaServiceData *pService = new STAFProcJavaServiceData(data);

        *pServiceHandle = pService;
        ++pService->fJVM->fNumServices;

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.write("JSTAF.STAFServiceConstruct()");
    }
    catch (...)
    {
        cout << "Caught unknown exception in JSTAF.STAFServiceConstruct()"
             << endl;
    }

    return kSTAFUnknownError;
}

STAFRC_t STAFServiceDestruct(STAFServiceHandle_t *serviceHandle,
                             void *pDestructInfo,
                             unsigned int destructLevel,
                             STAFString_t *pErrorBuffer)
{
    try
    {
        if (destructLevel != 0) return kSTAFInvalidAPILevel;

        STAFProcJavaServiceData *pData =
            static_cast<STAFProcJavaServiceData *>(*serviceHandle);

        STAFMutexSemLock lock(sJVMDataSem);

        // Debug: cout << "Destructing service: " << pData->fName << endl;
        // Debug: cout << "Service pData->fJVM: " << pData->fJVM << endl;
        // Debug: cout << "Service JVMName: " << pData->fJVM->fName << endl;

        if (--pData->fJVM->fNumServices == 0)
            STAFShutdownJVM(pData->fJVM->fName);

        delete pData;
        *serviceHandle = 0;

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.write("JSTAF.STAFServiceDestruct()");
    }
    catch (...)
    {
        cout << "Caught unknown exception in JSTAF.STAFServiceDestruct()"
             << endl;
    }

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceInit(STAFServiceHandle_t serviceHandle,
                         void *pInitInfo, unsigned int initLevel,
                         STAFString_t *pErrorBuffer)
{
    try
    {
        if (initLevel != 1) return kSTAFInvalidAPILevel;

        STAFProcJavaServiceData *pData =
            static_cast<STAFProcJavaServiceData *>(serviceHandle);
        STAFServiceInitLevel1 *pInfo =
            static_cast<STAFServiceInitLevel1 *>(pInitInfo);

        STAFConnAllocInfo allocInfo =
                          STAFLocalConnection::create(pData->fJVM->fIPCName);

        STAFConnPtr connPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                            STAFConnPtr::CUSTOM, allocInfo.pFreeFunc);

        connPtr->writeUInt(JAVA_SERVICE_INIT);
        connPtr->writeString(pData->fName);
        connPtr->writeString(pInfo->parms);

        unsigned int initRC = connPtr->readUInt();
        STAFString initResult = connPtr->readString();

        if (initRC != kSTAFOk)
        {
            *pErrorBuffer = initResult.adoptImpl();
            return initRC;
        }

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.write("JSTAF.STAFServiceInit()");
    }
    catch (...)
    {
        cout << "Caught unknown exception in JSTAF.STAFServiceInit()" << endl;
    }

    return kSTAFUnknownError;
}

STAFRC_t STAFServiceTerm(STAFServiceHandle_t serviceHandle,
                         void *pTermInfo, unsigned int termLevel,
                         STAFString_t *pErrorBuffer)
{
    try
    {
        if (termLevel != 0) return kSTAFInvalidAPILevel;

        STAFProcJavaServiceData *pData =
            static_cast<STAFProcJavaServiceData *>(serviceHandle);

        STAFConnAllocInfo allocInfo =
                          STAFLocalConnection::create(pData->fJVM->fIPCName);

        STAFConnPtr connPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                            STAFConnPtr::CUSTOM, allocInfo.pFreeFunc);

        connPtr->writeUInt(JAVA_SERVICE_TERM);
        connPtr->writeString(pData->fName);

        unsigned int termRC = connPtr->readUInt();
        STAFString termResult = connPtr->readString();

        if (termRC != kSTAFOk)
        {
            *pErrorBuffer = termResult.adoptImpl();
            return termRC;
        }

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        e.write("JSTAF.STAFServiceTerm()");
    }
    catch (...)
    {
        cout << "Caught unknown exception in JSTAF.STAFServiceTerm()" << endl;
    }

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceAcceptRequest(STAFServiceHandle_t serviceHandle,
                                  void *pRequestInfo, unsigned int reqLevel,
                                  STAFString_t *pResultBuffer)
{
    try
    {
        if (reqLevel != 2) return kSTAFInvalidAPILevel;

        STAFProcJavaServiceData *pData =
            static_cast<STAFProcJavaServiceData *>(serviceHandle);
        STAFServiceRequestLevel2 *pInfo =
            static_cast<STAFServiceRequestLevel2 *>(pRequestInfo);

        STAFConnAllocInfo allocInfo =
                          STAFLocalConnection::create(pData->fJVM->fIPCName);

        STAFConnPtr connPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                            STAFConnPtr::CUSTOM, allocInfo.pFreeFunc);

        unsigned int  localMachineLength     = 0;
        const char   *localMachineBuffer     = 0;
        unsigned int  machineLength          = 0;
        const char   *machineBuffer          = 0;
        unsigned int  effectiveMachineLength = 0;
        const char   *effectiveMachineBuffer = 0;
        unsigned int  processNameLength      = 0;
        const char   *processNameBuffer      = 0;
        unsigned int  requestLength          = 0;
        const char   *requestBuffer          = 0;

        STAFStringGetBuffer(pInfo->localMachine, &localMachineBuffer,
                            &localMachineLength, 0);
        STAFStringGetBuffer(pInfo->machine, &machineBuffer,
                            &machineLength, 0);
        STAFStringGetBuffer(pInfo->effectiveMachine, &effectiveMachineBuffer,
                            &effectiveMachineLength, 0);
        STAFStringGetBuffer(pInfo->processName, &processNameBuffer,
                            &processNameLength, 0);
        STAFStringGetBuffer(pInfo->request, &requestBuffer,
                            &requestLength, 0);

        // IMPORTANT:  Increase the numFields value if add a field to the
        // ServiceRequest class for a new STAFServiceInterfaceLevel.
        unsigned int numFields = 11;

        unsigned int bufferLength = (numFields * sizeof(unsigned int)) +
                                    pData->fName.length() + localMachineLength +
                                    machineLength + effectiveMachineLength +
                                    processNameLength + requestLength;
        STAFBuffer<char> buffer(new char[bufferLength], STAFBuffer<char>::INIT,
                                STAFBuffer<char>::ARRAY);
        unsigned int *uintBuffer = 
                     reinterpret_cast<unsigned int *>((char *)buffer);

        uintBuffer[0] = 
            STAFUtilConvertNativeUIntToLE(JAVA_SERVICE_ACCEPT_REQUEST);
        uintBuffer[1] = 
            STAFUtilConvertNativeUIntToLE(bufferLength - 
                (2 * sizeof(unsigned int)));
        uintBuffer[2] = pData->fName.length();
        uintBuffer[3] = pInfo->handle;
        uintBuffer[4] = pInfo->trustLevel;
        uintBuffer[5] = localMachineLength;
        uintBuffer[6] = machineLength;
        uintBuffer[7] = effectiveMachineLength;
        uintBuffer[8] = processNameLength;
        uintBuffer[9] = requestLength;
        uintBuffer[10] = pInfo->diagEnabled;
        
        char *currBuffer = buffer + (numFields * sizeof(unsigned int));

        memcpy(currBuffer, pData->fName.buffer(), pData->fName.length());
        currBuffer += pData->fName.length();
        memcpy(currBuffer, localMachineBuffer, localMachineLength);
        currBuffer += localMachineLength;
        memcpy(currBuffer, machineBuffer, machineLength);
        currBuffer += machineLength;
        memcpy(currBuffer, effectiveMachineBuffer, effectiveMachineLength);
        currBuffer += effectiveMachineLength;
        memcpy(currBuffer, processNameBuffer, processNameLength);
        currBuffer += processNameLength;
        memcpy(currBuffer, requestBuffer, requestLength);

        /* Debug:
        cout << "Buffer length: " << bufferLength << endl;
        cout << endl << "Buffer: " << endl;

        for (unsigned int i = 0; i < (bufferLength - 8); ++i)
        {
            if ((i != 0) && (i % 24 == 0)) cout << endl;
            unsigned int currChar = buffer[i + 8];
            if (currChar < 16) cout << "0";
            cout << hex << currChar << dec << " ";
        }

        cout << endl;
        */

        connPtr->write(buffer, bufferLength);

        unsigned int reqRC = connPtr->readUInt();
        STAFString reqResult = connPtr->readString();

        *pResultBuffer = reqResult.adoptImpl();
        return reqRC;
    }
    catch (STAFException &e)
    {
        e.write("JSTAF.STAFServiceAcceptRequest");
    }
    catch (...)
    {
        cout << "Caught unknown exception in JSTAF.STAFServiceAcceptRequest()"
             << endl;
    }

    *pResultBuffer = STAFString().adoptImpl();

    return kSTAFUnknownError;
}
