/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <jni.h>
#include <vector>
#include <map>
#include "STAFJavaService.h"
#include "com_ibm_staf_service_STAFServiceHelper.h"
#include "STAFLocalConnectionFactory.h"
#include "STAFThreadManager.h"

struct JavaServiceData
{
    STAFString fName;
    jmethodID fInitMethodID;
    jmethodID fAcceptRequestMethodID;
    jmethodID fTermMethodID;
    jobject fServiceObject;
    unsigned int fInterfaceLevel;
};

// Function prototypes

unsigned int HandleRequest(void *allocInfoPtr);
void HandleServiceConstruct(STAFConnPtr &connection);
void HandleServiceInit(STAFConnPtr &connection);
void HandleServiceRequest(STAFConnPtr &connection);
void HandleServiceTerm(STAFConnPtr &connection);

#if defined(STAF_OS_NAME_HPUX) && !defined(__ia64)
extern "C"
{
    void _main();
}
#endif

// Global variables

jobject gHelperObj;
JavaVM *gJVM;
bool gExitJVM = false;

// These variables reference strings which will be passed into the JVM.  See
// additional comments in the initJVMStrings() function below.

static const char *sJavaStringFieldType             = 0;
static const char *sJavaIntFieldType                = 0;
static const char *sJavaConstructorMethod           = 0;
static const char *sSTAFResultClass                 = 0;
static const char *sSTAFResultRCField               = 0;
static const char *sSTAFResultResultField           = 0;
static const char *sServiceInitClass                = 0;
static const char *sServiceInitConstructorMethodSig = 0;
static const char *sServiceRequestClass             = 0;
static const char *sServiceRequestConstructorSig    = 0;
static const char *sJVMNameField                    = 0;
static const char *sLoadServiceMethod               = 0;
static const char *sLoadServiceMethodSig            = 0;
static const char *sInitServiceMethod               = 0;
static const char *sInitServiceMethodSig            = 0;
static const char *sCallServiceMethod               = 0;
static const char *sCallServiceMethodSig            = 0;
static const char *sTermServiceMethod               = 0;
static const char *sTermServiceMethodSig            = 0;

static const STAFString *sSTAFResultClassStringPtr     = 0;
static const STAFString *sServiceInitClassStringPtr    = 0;
static const STAFString *sServiceRequestClassStringPtr = 0;

static void initJVMStrings()
{
    // JVM string constants
    //
    // Note: All strings that are passed into the JVM via the JNI interfaces
    //       must be sent in as "ASCII".  The JNI will not accept strings in the
    //       current encoding.  They must use the UTF-8 encodings from 0x01 to
    //       0x7F.  That is why we have all of the constants below.
    //
    // Note: This data must reside in this function.  If it is placed at file
    //       scope, then there are problems with the AIX runtime initializing
    //       things in a non-desirable order.

    // JVM Field types

    static STAFString sJavaStringFieldTypeString =
        STAFString("Ljava/lang/String;") + kUTF8_NULL;
    static STAFString sJavaIntFieldTypeString = STAFString("I") + kUTF8_NULL;

    sJavaStringFieldType = sJavaStringFieldTypeString.buffer();
    sJavaIntFieldType    = sJavaIntFieldTypeString.buffer();

    // Default Java Methods

    static STAFString sJavaConstructorMethodString =
        STAFString("<init>") + kUTF8_NULL;

    sJavaConstructorMethod = sJavaConstructorMethodString.buffer();

    // STAFResult class definitions

    static STAFString sSTAFResultClassString =
        STAFString("com/ibm/staf/STAFResult") + kUTF8_NULL;
    static STAFString sSTAFResultRCFieldString =
        STAFString("rc") + kUTF8_NULL;
    static STAFString sSTAFResultResultFieldString =
        STAFString("result") + kUTF8_NULL;

    sSTAFResultClass       = sSTAFResultClassString.buffer();
    sSTAFResultRCField     = sSTAFResultRCFieldString.buffer();
    sSTAFResultResultField = sSTAFResultResultFieldString.buffer();

    sSTAFResultClassStringPtr = &sSTAFResultClassString;

    // STAFServiceHelper.ServiceInit class definitions

    static STAFString sServiceInitClassString =
        STAFString("com/ibm/staf/service/STAFServiceHelper$ServiceInit") +
        kUTF8_NULL;
    static STAFString sServiceInitConstructorMethodSigString =
        STAFString("(Ljava/lang/String;Ljava/lang/String;)V") + kUTF8_NULL;

    sServiceInitClass = sServiceInitClassString.buffer();
    sServiceInitConstructorMethodSig =
        sServiceInitConstructorMethodSigString.buffer();

    sServiceInitClassStringPtr = &sServiceInitClassString;

    // STAFServiceHelper.ServiceRequest class definitions

    static STAFString sServiceRequestClassString   =
        STAFString("com/ibm/staf/service/STAFServiceHelper$ServiceRequest") +
        kUTF8_NULL;
    static STAFString sServiceRequestConstructorSigString =
        STAFString("(Ljava/lang/String;Ljava/lang/String;"
                   "Ljava/lang/String;Ljava/lang/String;III"
                   "Ljava/lang/String;)V") + kUTF8_NULL;

    sServiceRequestClass = sServiceRequestClassString.buffer();
    sServiceRequestConstructorSig =
        sServiceRequestConstructorSigString.buffer();

    sServiceRequestClassStringPtr = &sServiceRequestClassString;

    // STAFServiceHelper fields

    static STAFString sJVMNameFieldString = STAFString("fJVMName") + kUTF8_NULL;

    sJVMNameField = sJVMNameFieldString.buffer();

    // STAFServiceHelper methods

    static STAFString sLoadServiceMethodString =
        STAFString("loadService") + kUTF8_NULL;
    static STAFString sLoadServiceMethodSigString =
        STAFString("(Ljava/lang/String;Ljava/lang/String;"
                   "Ljava/lang/String;)Lcom/ibm/staf/STAFResult;") + kUTF8_NULL;
    static STAFString sInitServiceMethodString =
        STAFString("initService") + kUTF8_NULL;
    static STAFString sInitServiceMethodSigString =
        STAFString("(Ljava/lang/String;"
                   "Lcom/ibm/staf/service/STAFServiceHelper$ServiceInit;)"
                   "Lcom/ibm/staf/STAFResult;") + kUTF8_NULL;
    static STAFString sCallServiceMethodString =
        STAFString("callService") + kUTF8_NULL;
    static STAFString sCallServiceMethodSigString =
        STAFString("(Ljava/lang/String;"
                   "Lcom/ibm/staf/service/STAFServiceHelper$"
                   "ServiceRequest;)Lcom/ibm/staf/STAFResult;") + kUTF8_NULL;
    static STAFString sTermServiceMethodString =
        STAFString("termService") + kUTF8_NULL;
    static STAFString sTermServiceMethodSigString =
        STAFString("(Ljava/lang/String;)Lcom/ibm/staf/STAFResult;") +
        kUTF8_NULL;

    sLoadServiceMethod    = sLoadServiceMethodString.buffer();
    sLoadServiceMethodSig = sLoadServiceMethodSigString.buffer();
    sInitServiceMethod    = sInitServiceMethodString.buffer();
    sInitServiceMethodSig = sInitServiceMethodSigString.buffer();
    sCallServiceMethod    = sCallServiceMethodString.buffer();
    sCallServiceMethodSig = sCallServiceMethodSigString.buffer();
    sTermServiceMethod    = sTermServiceMethodString.buffer();
    sTermServiceMethodSig = sTermServiceMethodSigString.buffer();
}
    

/*
 * Class:     com_ibm_staf_service_STAFServiceHelper
 * Method:    initialize
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_staf_service_STAFServiceHelper_initialize
    (JNIEnv *env, jclass)
{
    #if defined(STAF_OS_NAME_HPUX) && !defined(__ia64)
        _main();
    #endif

    initJVMStrings();
}


/*
 * Class:     com_ibm_staf_service_STAFServiceHelper
 * Method:    listen
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_ibm_staf_service_STAFServiceHelper_listen
    (JNIEnv *env, jobject theHelperObject)
{
    gHelperObj = env->NewGlobalRef(theHelperObject);
    env->GetJavaVM(&gJVM);

    // Get the class for STAFServiceHelper and the fJVMName field ID

    jclass helperClass = env->GetObjectClass(theHelperObject);
    jfieldID jvmNameFieldID = env->GetFieldID(helperClass, sJVMNameField,
                                              sJavaStringFieldType);
    if (jvmNameFieldID == 0)
    {
        // We could not get the Field ID of STAFServiceHelper.fJVMName
        // The Java VM should already have thrown an exception, so we
        // just return

        return;
    }

    jstring jvmNameObj = reinterpret_cast<jstring>(
                         env->GetObjectField(theHelperObject, jvmNameFieldID));

    const char *jvmNameUTF = env->GetStringUTFChars(jvmNameObj, 0);

    if (jvmNameUTF == 0)
    {
      // Pass back a string with the error
    }

    try
    {
        STAFString jvmName(jvmNameUTF, strlen(jvmNameUTF), STAFString::kUTF8);
        STAFString jvmIPCName = "JSTAF_"+ jvmName;
        STAFLocalConnectionFactory factory(jvmIPCName);
        STAFThreadManager threadManager(1);

        while (!gExitJVM)
        {
            STAFConnAllocInfo allocInfo = factory.acceptConnection();

            if (gExitJVM) return;

            threadManager.dispatch(HandleRequest,
                                   new STAFConnAllocInfo(allocInfo));
        }
    }
    catch (STAFException &se)
    { se.write("STAFServiceHelper.listen()"); }
    catch (...)
    { cout << "Caught unknown exception in STAFServiceHelper.listen()" << endl; }
}

unsigned int HandleRequest(void *allocInfoPtr)
{
    unsigned int doShutdown = 0;

    try
    {
        STAFConnAllocInfo *pAllocInfo =
                          static_cast<STAFConnAllocInfo *>(allocInfoPtr);
        STAFConnPtr connection(pAllocInfo->pConnection, STAFConnPtr::INIT,
                               STAFConnPtr::CUSTOM, pAllocInfo->pFreeFunc);
        delete pAllocInfo;

        unsigned int reqType = connection->readUInt();

        switch (reqType)
        {
            case JAVA_SERVICE_JVMPING:
            {
                connection->writeUInt(kSTAFOk);
                connection->writeString(STAFString());
                break;
            }
            case JAVA_SERVICE_LOAD:
            {
                HandleServiceConstruct(connection);
                break;
            }
            case JAVA_SERVICE_INIT:
            {
                HandleServiceInit(connection);
                break;
            }
            case JAVA_SERVICE_ACCEPT_REQUEST:
            {
                HandleServiceRequest(connection);
                break;
            }
            case JAVA_SERVICE_TERM:
            {
                HandleServiceTerm(connection);
                break;
            }
            case JAVA_SERVICE_JVMEXIT:
            {
                gExitJVM = true;

                connection->writeUInt(kSTAFOk);
                connection->writeString(STAFString());

                break;
            }
            default:
            {
                connection->writeUInt(kSTAFInvalidValue);
                connection->writeString(STAFString("Invalid Java service ") +
                                        "request type: " + STAFString(reqType));
                break;
            }
        }
    }
    catch (STAFException &se)
    { se.write("JSTAFSH.HandleRequest()"); }
    catch (...)
    { cout << "Caught unknown exception in JSTAFSH.HandleRequest()" << endl; }

    return 0;
}


void HandleServiceConstruct(STAFConnPtr &connection)
{
    STAFString serviceName   = connection->readString();
    STAFString execInfo      = connection->readString();
    STAFString writeLocation = connection->readString();

    // Attach the current thread

    JNIEnv *env = 0;

    jint attachRC = gJVM->AttachCurrentThread(reinterpret_cast<void **>(&env),
                                              NULL);

    if (attachRC != 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error attaching Java VM thread, RC: " << attachRC << endl;
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error attaching Java VM thread, RC: " +
                                STAFString(attachRC) + " in constructService "
                                "for " + serviceName);
        return;
    }

    // Call JVM to load the class

    jclass helperClass = env->GetObjectClass(gHelperObj);

    jmethodID loadService = env->GetMethodID(helperClass, sLoadServiceMethod,
                                             sLoadServiceMethodSig);
    if (loadService == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error getting STAFServiceHelper.loadService() method ID"
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting "
                                "STAFServiceHelper.loadService() method ID "
                                "in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Convert the execInfo string to UTF-8 for Java

    execInfo      += kUTF8_NULL;
    serviceName   += kUTF8_NULL;
    writeLocation += kUTF8_NULL;

    jstring javaExecInfo      = env->NewStringUTF(execInfo.buffer());

    if (javaExecInfo == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error converting execInfo string to UTF-8" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error converting execInfo string to UTF-8 "
                                "in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jstring javaServiceName   = env->NewStringUTF(serviceName.buffer());

    if (javaServiceName == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error converting service name string to UTF-8" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error converting service name to UTF-8 "
                                "in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jstring javaWriteLocation = env->NewStringUTF(writeLocation.buffer());

    if (javaWriteLocation == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error converting writeLocation string to UTF-8" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error converting writeLocation to UTF-8 "
                                "in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Call the loadService() method

    jobject javaResult = env->CallObjectMethod(gHelperObj, loadService,
                                               javaServiceName, javaExecInfo,
                                               javaWriteLocation);
    if (javaResult == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error calling the loadService method" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error calling the loadService method in "
                                "constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the field IDs of the STAFResult class

    jclass resultClass = env->FindClass(sSTAFResultClass);

    if (resultClass == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error finding Java class " << *sSTAFResultClassStringPtr
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error finding Java class " +
                                *sSTAFResultClassStringPtr +
                                " in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultRCID = env->GetFieldID(resultClass, sSTAFResultRCField,
                                          sJavaIntFieldType);

    if (resultRCID == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error getting rc field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting rc field of Java class " +
                                *sSTAFResultClassStringPtr +
                                " in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultStringID = env->GetFieldID(resultClass,
                                              sSTAFResultResultField,
                                              sJavaStringFieldType);
    if (resultStringID == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error getting result field of Java class " 
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting result field of Java class " +
                                *sSTAFResultClassStringPtr +
                                " in constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the return code and result string from the STAFResult object

    jint rc = env->GetIntField(javaResult, resultRCID);
    jstring javaResultString = (jstring)env->GetObjectField(javaResult,
                                                            resultStringID);
    const char *utfResultString =
               env->GetStringUTFChars(javaResultString, 0);

    if (utfResultString == 0)
    {
        cout << "Constructing service: " << serviceName << endl;
        cout << "Error getting UTF-8 result string" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting UTF-8 result string in "
                                "constructService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    STAFString resultString(utfResultString,
                            env->GetStringUTFLength(javaResultString),
                            STAFString::kUTF8);

    env->ReleaseStringUTFChars(javaResultString, utfResultString);

    connection->writeUInt(rc);
    connection->writeString(resultString);

    // Detach the current thread

    gJVM->DetachCurrentThread();
}


void HandleServiceInit(STAFConnPtr &connection)
{
    STAFString serviceName = connection->readString();
    STAFString parms = connection->readString();

    // Attach the current thread

    JNIEnv *env = 0;

    jint attachRC = gJVM->AttachCurrentThread(reinterpret_cast<void **>(&env),
                                              NULL);

    if (attachRC != 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error attaching Java VM thread, RC: " << attachRC << endl;
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error attaching Java VM thread, RC: " +
                                STAFString(attachRC) + " in initService "
                                "for " + serviceName);
        return;
    }

    // Call JVM to load the class

    jclass helperClass = env->GetObjectClass(gHelperObj);

    jmethodID initService = env->GetMethodID(helperClass, sInitServiceMethod,
                                             sInitServiceMethodSig);
    if (initService == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error getting STAFServiceHelper.initService() method ID"
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting "
                                "STAFServiceHelper.initService() method ID");
        gJVM->DetachCurrentThread();
        return;
    }

    // Convert the service name and parms strings to UTF-8 for Java

    serviceName += kUTF8_NULL;
    parms = parms.replace(kUTF8_NULL, kUTF8_NULL2);
    parms += kUTF8_NULL;

    jstring javaServiceName = env->NewStringUTF(serviceName.buffer());

    if (javaServiceName == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error converting service name to UTF-8" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error converting service name " +
                                serviceName + " to UTF-8 in initService");
        gJVM->DetachCurrentThread();
        return;
    }

    jstring javaParms = env->NewStringUTF(parms.buffer());

    if (javaParms == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error converting Java parameters to UTF-8" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Eror converting Java parameters for " +
                                serviceName + " to UTF-8 in initService");
        gJVM->DetachCurrentThread();
        return;
    }

    // Load and construct the ServiceInit class

    jclass serviceInitClass = env->FindClass(sServiceInitClass);

    if (serviceInitClass == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error finding Java class "
             << *sServiceInitClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error finding Java class: " +
                                *sServiceInitClassStringPtr + " for "
                                + serviceName + " in initService");
        gJVM->DetachCurrentThread();
        return;
    }

    jmethodID constructor = env->GetMethodID(serviceInitClass,
                                             sJavaConstructorMethod,
                                             sServiceInitConstructorMethodSig);
    if (constructor == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error loading constructor for Java class "
             << *sServiceInitClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error loading constructor for Java class " +
                                *sServiceInitClassStringPtr + " for " +
                                serviceName);
        gJVM->DetachCurrentThread();
        return;
    }
        
    jobject serviceInitObject = env->NewObject(serviceInitClass, constructor,
                                               javaServiceName, javaParms);

    if (serviceInitObject == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error creating Java object of type "
             << *sServiceInitClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error creating Java object of type " +
                                *sServiceInitClassStringPtr + " for " +
                                serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Call the initService() method

    jobject javaResult = env->CallObjectMethod(gHelperObj, initService,
                                               javaServiceName,
                                               serviceInitObject);
    if (javaResult == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error calling the initService method" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error calling the initService method for " +
                                serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the field IDs of the STAFResult class

    jclass resultClass = env->FindClass(sSTAFResultClass);

    if (resultClass == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error finding Java class" << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error finding Java class " +
                                *sSTAFResultClassStringPtr +
                                " in initService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultRCID = env->GetFieldID(resultClass, sSTAFResultRCField,
                                          sJavaIntFieldType);

    if (resultRCID == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error getting rc field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting rc field of Java class " +
                                *sSTAFResultClassStringPtr +
                                " in initService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultStringID = env->GetFieldID(resultClass,
                                              sSTAFResultResultField,
                                              sJavaStringFieldType);
    if (resultStringID == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error getting result field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting result field of Java class " +
                                *sSTAFResultClassStringPtr +
                                " in initService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the return code and result string from the STAFResult object

    jint rc = env->GetIntField(javaResult, resultRCID);
    jstring javaResultString = (jstring)env->GetObjectField(javaResult,
                                                            resultStringID);
    const char *utfResultString =
               env->GetStringUTFChars(javaResultString, 0);

    if (utfResultString == 0)
    {
        cout << "Initializing service: " << serviceName << endl;
        cout << "Error getting UTF-8 result string" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting UTF-8 result string in "
                                "initService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    STAFString resultString(utfResultString,
                            env->GetStringUTFLength(javaResultString),
                            STAFString::kUTF8);

    env->ReleaseStringUTFChars(javaResultString, utfResultString);

    connection->writeUInt(rc);
    connection->writeString(resultString);

    // Detach the current thread

    gJVM->DetachCurrentThread();
}


void HandleServiceRequest(STAFConnPtr &connection)
{
    // Note: The request type has already been read
    unsigned int totalLength = connection->readUInt();
    STAFBuffer<char> buffer(new char[totalLength], STAFBuffer<char>::INIT,
                            STAFBuffer<char>::ARRAY);
    unsigned int *uintBuffer = reinterpret_cast<unsigned int *>((char *)buffer);

    // IMPORTANT:  Increase the numFields value if add a field to the
    //             ServiceRequest class for a new STAFServiceInterfaceLevel.
    //             Note:  This value is 2 less than the numFields value in
    //                    STAFJavaService.cpp.
    unsigned int numFields = 9;  // # of fields in buffer

    // uintBuffer[0] = Service name length
    // uintBuffer[1] = Handle
    // uintBuffer[2] = Trust level
    // uintBuffer[3] = Local machine length
    // uintBuffer[4] = Machine length
    // uintBuffer[5] = Effective machine length
    // uintBuffer[6] = Handle name length
    // uintBuffer[7] = Request length
    // uintBuffer[8] = Diagnostics Enabled Flag

    connection->read(buffer, totalLength);

    
    char *serviceNameBuffer = buffer + (numFields * sizeof(unsigned int));
    char *localMachineBuffer = serviceNameBuffer + uintBuffer[0];
    char *machineBuffer = localMachineBuffer + uintBuffer[3];
    char *effectiveMachineBuffer = machineBuffer +  uintBuffer[4];
    char *handleNameBuffer = effectiveMachineBuffer + uintBuffer[5];
    char *requestBuffer = handleNameBuffer + uintBuffer[6];

    if (serviceNameBuffer == 0 || localMachineBuffer == 0 ||
        machineBuffer == 0 || effectiveMachineBuffer == 0 ||
        handleNameBuffer == 0 || requestBuffer == 0)
    {
        cout << "Memory allocation failure in "
             << "JSTAFSH.HandleServiceRequest()" << endl;
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Memory allocation failure in "
                                "JSTAFSH.HandleServiceRequest");
        return;
    }

    STAFString serviceName(serviceNameBuffer, uintBuffer[0], STAFString::kUTF8);
    STAFString localMachine(localMachineBuffer, uintBuffer[3],
                            STAFString::kUTF8);
    STAFString machine(machineBuffer, uintBuffer[4], STAFString::kUTF8);
    STAFString effectiveMachine(effectiveMachineBuffer, uintBuffer[5],
                                STAFString::kUTF8);
    STAFString handleName(handleNameBuffer, uintBuffer[6], STAFString::kUTF8);
    unsigned int handle = uintBuffer[1];
    unsigned int trustLevel = uintBuffer[2];
    STAFString request(requestBuffer, uintBuffer[7], STAFString::kUTF8);
    unsigned int diagEnabled = uintBuffer[8];
    
    /* Debug:
    cout << "Total length       : " << totalLength << endl
         << "Service name       : " << serviceName << endl
         << "Local machine      : " << localMachine << endl
         << "Machine            : " << machine << endl
         << "Effective machine  : " << effectiveMachine << endl
         << "Handle name length : " << uintBuffer[6] << endl
         << "Handle name        : " << handleName << endl
         << "Handle             : " << handle << endl
         << "Trust level        : " << trustLevel << endl
         << "Request length     : " << uintBuffer[7] << endl
         << "Request            : " << request << endl
         << "Diag enabled       : " << diagEnabled << endl;

    cout << endl << "Buffer: " << endl;
    
    for (unsigned int i = 0; i < totalLength; ++i)
    {
        if ((i != 0) && (i % 8 == 0)) cout << endl;
        unsigned int currChar = buffer[i];
        if (currChar < 16) cout << "0";
        cout << hex << currChar << dec << ":" << (char)currChar << " ";
    }

    cout << endl;
    */

    // Attach the current thread

    JNIEnv *env = 0;

    jint attachRC = gJVM->AttachCurrentThread(reinterpret_cast<void **>(&env),
                                              NULL);

    if (attachRC != 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error attaching Java VM thread, RC: " << attachRC << endl;
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error attaching Java VM thread, RC: " +
                                STAFString(attachRC));
        return;
    }

    // Call JVM to load the class

    jclass helperClass = env->GetObjectClass(gHelperObj);

    jmethodID callService = env->GetMethodID(helperClass, sCallServiceMethod,
                                             sCallServiceMethodSig);
    if (callService == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error getting STAFServiceHelper.callService() method ID"
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting "
                                "STAFServiceHelper.callService() method ID");
        gJVM->DetachCurrentThread();
        return;
    }

    // Convert the various strings to UTF-8 for Java

    serviceName += kUTF8_NULL;
    localMachine += kUTF8_NULL;
    machine += kUTF8_NULL;
    effectiveMachine += kUTF8_NULL;
    handleName = handleName.replace(kUTF8_NULL, kUTF8_NULL2);
    handleName += kUTF8_NULL;
    request = request.replace(kUTF8_NULL, kUTF8_NULL2);
    request += kUTF8_NULL;

    jstring javaServiceName = env->NewStringUTF(serviceName.buffer());
    jstring javaLocalMachine = env->NewStringUTF(localMachine.buffer());
    jstring javaMachine = env->NewStringUTF(machine.buffer());
    jstring javaEffectiveMachine = env->NewStringUTF(effectiveMachine.buffer());
    jstring javaHandleName = env->NewStringUTF(handleName.buffer());
    jstring javaRequest = env->NewStringUTF(request.buffer());

    // Load and construct the ServiceInit class

    jclass serviceRequestClass = env->FindClass(sServiceRequestClass);

    if (serviceRequestClass == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error finding Java class: " << *sServiceRequestClassStringPtr
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error finding Java class: " +
                                *sServiceRequestClassStringPtr);
        gJVM->DetachCurrentThread();
        return;
    }

    jmethodID constructor = env->GetMethodID(serviceRequestClass,
                                             sJavaConstructorMethod,
                                             sServiceRequestConstructorSig);
    if (constructor == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error loading constructor for Java class "
             << *sServiceRequestClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error loading constructor for Java class " +
                                *sServiceRequestClassStringPtr);
        gJVM->DetachCurrentThread();
        return;
    }

    jobject serviceRequestObject =
        env->NewObject(serviceRequestClass, constructor, javaLocalMachine,
                       javaMachine, javaEffectiveMachine, javaHandleName,
                       (jint)handle, (jint)trustLevel, (jint)diagEnabled,
                       javaRequest);

    if (serviceRequestObject == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error creating Java object of type "
             << *sServiceRequestClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error creating Java object of type " +
                                *sServiceRequestClassStringPtr);
        gJVM->DetachCurrentThread();
        return;
    }

    // Call the callService() method

    jobject javaResult = env->CallObjectMethod(gHelperObj, callService,
                                               javaServiceName,
                                               serviceRequestObject);
    if (javaResult == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error calling callService method for ServiceRequest object"
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error calling callService method "
                                "for ServiceRequest object");
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the field IDs of the STAFResult class

    jclass resultClass = env->FindClass(sSTAFResultClass);

    if (resultClass == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error finding Java class " << *sSTAFResultClassStringPtr
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error finding Java class " +
                                *sSTAFResultClassStringPtr);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultRCID = env->GetFieldID(resultClass, sSTAFResultRCField,
                                          sJavaIntFieldType);

    if (resultRCID == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error getting rc field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting rc field of Java class " +
                                *sSTAFResultClassStringPtr);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultStringID = env->GetFieldID(resultClass,
                                              sSTAFResultResultField,
                                              sJavaStringFieldType);
    if (resultStringID == 0)
    {
        cout << "Error getting result field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        cout << "Service: " << serviceName << " Request: " << request << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting result field of Java class " +
                                *sSTAFResultClassStringPtr);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the return code and result string from the STAFResult object

    int rc = env->GetIntField(javaResult, resultRCID);
    jstring javaResultString = (jstring)env->GetObjectField(javaResult,
                                                            resultStringID);
    const char *utfResultString =
               env->GetStringUTFChars(javaResultString, 0);

    if (utfResultString == 0)
    {
        cout << "Service: " << serviceName << " Request: " << request << endl;
        cout << "Error getting result string from the STAFResult object"
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting result string from the "
                                "STAFResult object");
        gJVM->DetachCurrentThread();
        return;
    }

    STAFString resultString(utfResultString,
                            env->GetStringUTFLength(javaResultString),
                            STAFString::kUTF8);

    env->ReleaseStringUTFChars(javaResultString, utfResultString);

    connection->writeUInt(rc);
    connection->writeString(resultString);

    // Detach the current thread

    env->DeleteLocalRef(javaLocalMachine);
    env->DeleteLocalRef(javaServiceName);
    env->DeleteLocalRef(javaMachine);
    env->DeleteLocalRef(javaEffectiveMachine);
    env->DeleteLocalRef(javaHandleName);
    env->DeleteLocalRef(javaRequest);
    env->DeleteLocalRef(serviceRequestObject);

    gJVM->DetachCurrentThread();
}


void HandleServiceTerm(STAFConnPtr &connection)
{
    STAFString serviceName = connection->readString();

    // Attach the current thread

    JNIEnv *env = 0;

    jint attachRC = gJVM->AttachCurrentThread(reinterpret_cast<void **>(&env),
                                              NULL);

    if (attachRC != 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error attaching Java VM thread, RC: " << attachRC << endl;
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error attaching Java VM thread, RC: " +
                                STAFString(attachRC) + " during termService "
                                " for " + serviceName);
        return;
    }

    jclass helperClass = env->GetObjectClass(gHelperObj);

    jmethodID termService = env->GetMethodID(helperClass, sTermServiceMethod,
                                             sTermServiceMethodSig);
    if (termService == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error getting STAFServiceHelper.termService() method ID"
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting "
                                "STAFServiceHelper.termService() method ID "
                                "for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Convert the service name to UTF-8 for Java

    serviceName += kUTF8_NULL;

    jstring javaServiceName = env->NewStringUTF(serviceName.buffer());

    if (javaServiceName == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error converting service name to UTF-8" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error converting service name to UTF-8 "
                                "during termService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Call the termService() method

    jobject javaResult = env->CallObjectMethod(gHelperObj, termService,
                                               javaServiceName);
    if (javaResult == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error calling termService method" << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error calling termService method for " +
                                serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the field IDs of the STAFResult class

    jclass resultClass = env->FindClass(sSTAFResultClass);

    if (resultClass == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error finding Java class " << *sSTAFResultClassStringPtr
             << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error finding load Java class " +
                                *sSTAFResultClassStringPtr +
                                " during termService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultRCID = env->GetFieldID(resultClass, sSTAFResultRCField,
                                          sJavaIntFieldType);

    if (resultRCID == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error getting rc field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting rc field of Java class " +
                                *sSTAFResultClassStringPtr +
                                " during termService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    jfieldID resultStringID = env->GetFieldID(resultClass,
                                              sSTAFResultResultField,
                                              sJavaStringFieldType);
    if (resultStringID == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error getting result field of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting result field of Java class " +
                                *sSTAFResultClassStringPtr +
                                " during termService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    // Get the return code and result string from the STAFResult object

    jint rc = env->GetIntField(javaResult, resultRCID);
    jstring javaResultString = (jstring)env->GetObjectField(javaResult,
                                                            resultStringID);
    const char *utfResultString =
               env->GetStringUTFChars(javaResultString, 0);

    if (utfResultString == 0)
    {
        cout << "Terminating service: " << serviceName << endl;
        cout << "Error getting result string of Java class "
             << *sSTAFResultClassStringPtr << endl;
        env->ExceptionDescribe();
        env->ExceptionClear();
        connection->writeUInt(kSTAFJavaError);
        connection->writeString("Error getting result string of Java class " +
                                *sSTAFResultClassStringPtr +
                                " during termService for " + serviceName);
        gJVM->DetachCurrentThread();
        return;
    }

    STAFString resultString(utfResultString,
                            env->GetStringUTFLength(javaResultString),
                            STAFString::kUTF8);

    env->ReleaseStringUTFChars(javaResultString, utfResultString);

    connection->writeUInt(rc);
    connection->writeString(resultString);

    // Detach the current thread

    gJVM->DetachCurrentThread();
}
