/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service;

import com.ibm.staf.*;
import java.util.jar.*;
import java.util.Map;
import java.util.HashMap;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.File;
import java.util.Collections;

// STAFServiceHelper - This class is used to help load and call Java services
//                     from jar files, as well as from classes in the CLASSPATH.
//
// Most of its methods are called from JNI code.  Essentially the static main
// method is invoked by STAFProc (via a standard java invocation).  The main
// method creates an instance of the STAFServiceHelper.  The service helper
// then calls into native code to listen on an IPC interface.  This native
// method doesn't return until it is told to by STAFProc.  The native method
// creates threads for each service request.  These threads call back into the
// helper object in order load/initialize/call/terminate the service.

public class STAFServiceHelper
{
    static final String STAF_ENTRY = new String("staf/service/info");
    static final String STAF_SERVICE_CLASS = new String("Service-Class");
    static final String STAF_SERVICE_JARS = new String("Packaged-Jars");

    static public class ServiceInit
    {
        public ServiceInit(String name, String parms)
        {
            this.name = name;
            this.parms = parms;
        }

        public String name;
        public String parms;
    }

    static public class ServiceRequest
    {
        public ServiceRequest(String localMachine, String machine,
                              String effectiveMachine, String handleName,
                              int handle, int trustLevel, int diagEnabled,
                              String request)
        {
            this.localMachine = localMachine;
            this.machine = machine;
            this.effectiveMachine = effectiveMachine;
            this.handleName = handleName;
            this.handle = handle;
            this.trustLevel = trustLevel;
            this.diagEnabled = diagEnabled;
            this.request = request;
        }

        public String localMachine;
        public String machine;
        public String effectiveMachine;
        public String handleName;
        public int handle;
        public int trustLevel;
        public int diagEnabled;
        public String request;
    }

    static public class ServiceData
    {
        public Object service;
        public JarFile serviceJar;
        public ClassLoader serviceClassLoader;

        public ServiceData(Object serviceObject, JarFile serviceJar, ClassLoader serviceClassLoader)
        {
            this.service = serviceObject;
            this.serviceJar = serviceJar;
            this.serviceClassLoader = serviceClassLoader;
        }

    }

    public static void main(String [] argv)
    {
        if (argv.length != 1)
        {
            System.out.println(
                "Usage: java com.ibm.staf.service.STAFServiceHelper <JVM Name>");
            System.exit(1);
        }

        STAFServiceHelper helper = new STAFServiceHelper(argv[0]);

        // We will only return from this call once there are no loaded services
        // left.

        helper.listen();

        System.exit(0);
    }

    public STAFServiceHelper(String jvmName)
    {
        fJVMName = jvmName;
    }

    // loadService - This method is reponsible for loading and constructing the
    //               implementation class for a given service.  It supports
    //               raw class files, as well as jar-based services.

    private STAFResult loadService(String service, String serviceImpl,
                                   String stafTemp)
    {
        try
        {
            String serviceClassName = serviceImpl;
            ClassLoader loader = this.getClass().getClassLoader();
            Class serviceClass = null;
            Object serviceObj = null;
            JarFile jarFile = null;

            if (serviceImpl.toLowerCase().endsWith(".jar"))
            {
                try
                {
                    jarFile = new JarFile(serviceImpl);
                }
                catch (IOException e)
                {
                    return new STAFResult(STAFResult.JavaError,
                                          "Unable to open jar file: " +
                                          serviceImpl);
                }

                Manifest manifest = null;

                try
                {
                    manifest = jarFile.getManifest();
                }
                catch (IOException e)
                {
                    return new STAFResult(STAFResult.JavaError,
                                          "Unable to load jar file manifest: " +
                                          serviceImpl);
                }

                if (!manifest.getEntries().containsKey(STAF_ENTRY))
                {
                    return new STAFResult(STAFResult.JavaError,
                                          "Invalid manifest for STAF service: " +
                                          "No " + STAF_ENTRY + " entry");
                }

                Attributes attrs = manifest.getAttributes(STAF_ENTRY);

                if (!attrs.containsKey(new Attributes.Name(STAF_SERVICE_CLASS)))
                {
                    return new STAFResult(STAFResult.JavaError,
                                          "Invalid manifest for STAF service: " +
                                          "No " + STAF_SERVICE_CLASS +
                                          " attribute");
                }

                serviceClassName = attrs.getValue(STAF_SERVICE_CLASS);

                String serviceDir = stafTemp + "/staf/jstaf/service/" + service;
                File serviceDirFile = new File(serviceDir);

                if (!serviceDirFile.exists() && !serviceDirFile.mkdirs())
                {
                    return new STAFResult(STAFResult.JavaError,
                                          "Error creating directory: " +
                                          serviceDir);
                }

                loader = new STAFServiceJarClassLoader(jarFile, serviceDir);
            }

            try
            {
                serviceClass = loader.loadClass(serviceClassName);

                // set context class loader

                Thread.currentThread().setContextClassLoader(loader);

                serviceObj = serviceClass.newInstance();
            }
            catch (ClassNotFoundException e)
            {
                return new STAFResult(STAFResult.JavaError,
                                      "ClassNotFound: " + serviceClassName);
            }
            catch (InstantiationException e)
            {
                return new STAFResult(STAFResult.JavaError,
                                      "Could not instantiate class: " +
                                      serviceClassName);
            }
            catch (IllegalAccessException e)
            {
                return new STAFResult(STAFResult.JavaError,
                                      "Illegal access to class: " +
                                      serviceClassName);
            }

            if (!((serviceObj instanceof
                   com.ibm.staf.service.STAFServiceInterfaceLevel1) ||
                  (serviceObj instanceof
                   com.ibm.staf.service.STAFServiceInterfaceLevel2) ||
                  (serviceObj instanceof
                   com.ibm.staf.service.STAFServiceInterfaceLevel3) ||
                  (serviceObj instanceof
                   com.ibm.staf.service.STAFServiceInterfaceLevel4)))
            {
                return new STAFResult(STAFResult.JavaError,
                                      "Not a valid interface level");
            }

            fServiceMap.put(service, new ServiceData(serviceObj, jarFile, loader));
        }
        catch (Throwable t)
        {
            // Go ahead and print the stack trace, since STAFProc doesn't really
            // expect a string back from Service construction

            t.printStackTrace();

            // Now return the stack trace as well

            StringWriter stringWriter = new StringWriter();
            t.printStackTrace(new PrintWriter(stringWriter));
            return new STAFResult(STAFResult.JavaError, stringWriter.toString());
        }

        return new STAFResult(STAFResult.Ok);
    }

    private STAFResult initService(String service, ServiceInit initInfo)
    {
        try
        {
            Object serviceDataObj = fServiceMap.get(service);
            int rc = STAFResult.UnknownError;

            if (serviceDataObj == null)
            {
                return new STAFResult(STAFResult.JavaError,
                                      "No such service implementation: " +
                                      service);
            }

            ServiceData serviceData = (ServiceData)serviceDataObj;

            // set context class loader

            Thread.currentThread().setContextClassLoader(serviceData.serviceClassLoader);

            if (serviceData.service instanceof STAFServiceInterfaceLevel1)
            {
                STAFServiceInterfaceLevel1 serviceL1 =
                    (STAFServiceInterfaceLevel1)serviceData.service;

                rc = serviceL1.init(service, initInfo.parms);
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel2)
            {
                STAFServiceInterfaceLevel2 serviceL2 =
                    (STAFServiceInterfaceLevel2)serviceData.service;

                rc = serviceL2.init(
                     new STAFServiceInterfaceLevel2.InitInfo(service,
                                                             initInfo.parms));
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel3)
            {
                STAFServiceInterfaceLevel3 serviceL3 =
                    (STAFServiceInterfaceLevel3)serviceData.service;

                rc = serviceL3.init(
                     new STAFServiceInterfaceLevel3.InitInfo(service,
                         initInfo.parms, serviceData.serviceJar));
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel4)
            {
                STAFServiceInterfaceLevel4 serviceL4 =
                    (STAFServiceInterfaceLevel4)serviceData.service;

                rc = serviceL4.init(
                     new STAFServiceInterfaceLevel4.InitInfo(service,
                         initInfo.parms, serviceData.serviceJar));
            }
            else
            {
                return new STAFResult(STAFResult.JavaError,
                                      "Unknown service interface type");
            }

            return new STAFResult(rc);
        }
        catch (Throwable t)
        {
            StringWriter stringWriter = new StringWriter();
            t.printStackTrace(new PrintWriter(stringWriter));
            return new STAFResult(STAFResult.JavaError, stringWriter.toString());
        }
    }

    private STAFResult callService(String service, ServiceRequest reqInfo)
    {
        try
        {
            Object serviceDataObj = fServiceMap.get(service);
            STAFResult result = new STAFResult(STAFResult.UnknownError);

            if (serviceDataObj == null)
            {
                return new STAFResult(STAFResult.JavaError,
                                      "No such service implementation: " +
                                      service);
            }

            ServiceData serviceData = (ServiceData)serviceDataObj;

            // set context class loader

            Thread.currentThread().setContextClassLoader(serviceData.serviceClassLoader);

            if (serviceData.service instanceof STAFServiceInterfaceLevel1)
            {
                STAFServiceInterfaceLevel1 serviceL1 =
                    (STAFServiceInterfaceLevel1)serviceData.service;

                result = serviceL1.acceptRequest(reqInfo.machine,
                                                 reqInfo.handleName,
                                                 reqInfo.handle,
                                                 reqInfo.request);
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel2)
            {
                STAFServiceInterfaceLevel2 serviceL2 =
                    (STAFServiceInterfaceLevel2)serviceData.service;

                result = serviceL2.acceptRequest(
                    new STAFServiceInterfaceLevel2.RequestInfo(
                        reqInfo.localMachine, reqInfo.machine,
                        reqInfo.effectiveMachine, reqInfo.handleName,
                        reqInfo.handle, reqInfo.trustLevel, reqInfo.request));
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel3)
            {
                STAFServiceInterfaceLevel3 serviceL3 =
                    (STAFServiceInterfaceLevel3)serviceData.service;

                result = serviceL3.acceptRequest(
                    new STAFServiceInterfaceLevel3.RequestInfo(
                        reqInfo.localMachine, reqInfo.machine,
                        reqInfo.effectiveMachine, reqInfo.handleName,
                        reqInfo.handle, reqInfo.trustLevel, reqInfo.request));
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel4)
            {
                STAFServiceInterfaceLevel4 serviceL4 =
                    (STAFServiceInterfaceLevel4)serviceData.service;

                result = serviceL4.acceptRequest(
                    new STAFServiceInterfaceLevel4.RequestInfo(
                        reqInfo.localMachine, reqInfo.machine,
                        reqInfo.effectiveMachine, reqInfo.handleName,
                        reqInfo.handle, reqInfo.trustLevel,
                        reqInfo.diagEnabled, reqInfo.request));
            }
            else
            {
                result = new STAFResult(STAFResult.JavaError,
                                        "Unknown service interface type");
            }

            return result;
        }
        catch (Throwable t)
        {
            StringWriter stringWriter = new StringWriter();
            t.printStackTrace(new PrintWriter(stringWriter));
            return new STAFResult(STAFResult.JavaError, stringWriter.toString());
        }
    }

    private STAFResult termService(String service)
    {
        try
        {
            Object serviceDataObj = fServiceMap.get(service);
            int rc = STAFResult.UnknownError;

            if (serviceDataObj == null)
            {
                return new STAFResult(STAFResult.JavaError,
                                      "No such service implementation: " +
                                      service);
            }

            fServiceMap.remove(service);

            ServiceData serviceData = (ServiceData)serviceDataObj;

            // set context class loader

            Thread.currentThread().setContextClassLoader(serviceData.serviceClassLoader);

            if (serviceData.service instanceof STAFServiceInterfaceLevel1)
            {
                STAFServiceInterfaceLevel1 serviceL1 =
                    (STAFServiceInterfaceLevel1)serviceData.service;

                rc = serviceL1.term();
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel2)
            {
                STAFServiceInterfaceLevel2 serviceL2 =
                    (STAFServiceInterfaceLevel2)serviceData.service;

                rc = serviceL2.term();
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel3)
            {
                STAFServiceInterfaceLevel3 serviceL3 =
                    (STAFServiceInterfaceLevel3)serviceData.service;

                rc = serviceL3.term();
            }
            else if (serviceData.service instanceof STAFServiceInterfaceLevel4)
            {
                STAFServiceInterfaceLevel4 serviceL4 =
                    (STAFServiceInterfaceLevel4)serviceData.service;

                rc = serviceL4.term();
            }
            else
            {
                serviceData.serviceJar.close();

                return new STAFResult(STAFResult.JavaError,
                                      "Unknown service interface type");
            }

            serviceData.serviceJar.close();

            return new STAFResult(STAFResult.Ok);
        }
        catch (Throwable t)
        {
            StringWriter stringWriter = new StringWriter();
            t.printStackTrace(new PrintWriter(stringWriter));
            return new STAFResult(STAFResult.JavaError, stringWriter.toString());
        }
    }

    private String fJVMName;
    private Map fServiceMap = Collections.synchronizedMap(new HashMap());

    /************************/
    /* All the native stuff */
    /************************/

    private static native void initialize();
    private native void listen();

    // Static initializer - called first time class is loaded.
    static
    {
        System.loadLibrary("JSTAFSH");
        initialize();
    }
}
