/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

///////////////////////////////////////////////////////////////////////////////
#include "STAFOSTypes.h"
#include <stdio.h> 
#include <stdlib.h>
#include "STAF.h"       // must remain before perl.h on WIN32
#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>
///////////////////////////////////////////////////////////////////////////////

/* ************************************************************************* */
/* STAF_PERL_API_PROLOGUE must be the first call in each function, and it    */
/* takes the number of parametes the API expects and a message to display    */
/* if the function gets called incorrectly.                                  */
/* ************************************************************************* */
#define STAF_PERL_API_PROLOGUE(parms, message) \
    dXSARGS;                                   \
    if (items != parms) croak(message);        \
    SP -= items;                               \
    {

/* ************************************************************************* */
/* STAF_PERL_API_EPILOGUE must be the last call in each function, and it     */
/* takes the number of parametes the API returns.  No other point in the     */
/* function should return without previously pushing 'retvals' values in     */
/* the stack and calling this macro as the stack may be read wrongly  by     */
/* the PERL interpreter.                                                     */
/* ************************************************************************* */
#define STAF_PERL_API_EPILOGUE(retvals)        \
    XSRETURN(retvals);                         \
    PUTBACK;                                   \
    }

///////////////////////////////////////////////////////////////////////////////
static const unsigned int STAFPerlError = 4000;
static const char* STAF_HANDLE  = "STAF::Handle";
static const char* STAF_RESULT  = "STAF::Result";
static const char* STAF_RC      = "STAF::RC";
static const char* EMPTY_STRING = "";
///////////////////////////////////////////////////////////////////////////////

extern "C" 
{

#if defined(STAF_OS_NAME_HPUX) && !defined(__ia64)
    void _main();
#endif

#ifdef PERL5_6PLUS
void PLSTAF_Register(PerlInterpreter *myperl, CV *cv)
#else
void PLSTAF_Register(CV *cv)
#endif
{
    STAF_PERL_API_PROLOGUE(1, "Usage: rc = STAF::Register(name)");

    SV *pName = ST(0);

    // values to be returned
    unsigned int rc = 0;

    // check arguments
    if (!SvPOK(pName))
    {
        rc = STAFPerlError;
    }
    else
    {
        char   *name = (char *)SvPV(pName, PL_na);
        SV     *magicRC     = 0;
        SV     *magicHandle = 0;
        STAFHandle_t handle = 0;

        magicRC     = perl_get_sv((char *)STAF_RC,     1);
        magicHandle = perl_get_sv((char *)STAF_HANDLE, 1);

        if (magicRC == 0 || magicHandle == 0)
        {
            rc = STAFPerlError;
        }
        else
        {
            rc = STAFRegister(name, &handle);

            sv_setiv(magicRC, rc);

            if (rc == 0)
            {
                sv_setiv(magicHandle, handle);
            }
        }
    }

    // pass rc in stack as well
    XPUSHs(sv_2mortal(newSViv(rc)));

    STAF_PERL_API_EPILOGUE(1);
}

#ifdef PERL5_6PLUS
void PLSTAF_Submit(PerlInterpreter *myperl, CV *cv)
#else
void PLSTAF_Submit(CV *cv)
#endif
{
    STAF_PERL_API_PROLOGUE(3, "Usage: rc = STAF::Submit(where, service, " \
                                                        "request)");
    SV *pWhere   = ST(0);
    SV *pService = ST(1);
    SV *pRequest = ST(2);

    // values to be returned
    unsigned int rc = 0;

    char *result = 0;
    STAFHandle_t handle = 0;

    // check arguments
    if (!SvPOK(pWhere)  || !SvPOK(pService) || !SvPOK(pRequest))
    {
        rc = STAFPerlError;
    } 
    else
    {
        SV *magicHandle = perl_get_sv((char *)STAF_HANDLE, 0);
        SV *magicRC     = perl_get_sv((char *)STAF_RC,     0);
        SV *magicResult = perl_get_sv((char *)STAF_RESULT, 1);
        
        if (magicRC == 0 || magicHandle == 0 || magicResult == 0)
        {
            rc = STAFPerlError;
        }
        else
        {
            char *where         = SvPV(pWhere,   PL_na);
            char *service       = SvPV(pService, PL_na);
            char *request       = SvPV(pRequest, PL_na);
            unsigned int reslen = 0;

            handle = SvIV(magicHandle);

            rc = STAFSubmit(handle, where, service, request, 
                            strlen(request), &result, &reslen);
            
            sv_setiv(magicRC, rc);
            
            if (reslen == 0)
            {
                sv_setpv(magicResult, EMPTY_STRING);
            }
            else
            {
                sv_setpv(magicResult, result);
            }
        }
    }

    // pass rc in stack as well
    XPUSHs(sv_2mortal(newSViv(rc)));
    
    if (result && result != EMPTY_STRING)
    {
        STAFFree(handle, result);
    }
    
    STAF_PERL_API_EPILOGUE(1);
}
 
#ifdef PERL5_6PLUS
void PLSTAF_UnRegister(PerlInterpreter *myperl, CV *cv)
#else
void PLSTAF_UnRegister(CV *cv)
#endif
{
    STAF_PERL_API_PROLOGUE(0, "Usage: rc = STAF::UnRegister()");

    // values to be returned
    unsigned int rc = 0;

    SV *magicHandle = perl_get_sv((char *)STAF_HANDLE, 0);
    SV *magicRC     = perl_get_sv((char *)STAF_RC,     0);

    if (magicRC == 0 || magicHandle == 0)
    {
        rc = STAFPerlError;
    }
    else
    {
        STAFHandle_t handle = SvIV(magicHandle);
        rc = STAFUnRegister(handle);
        sv_setiv(magicRC, rc);
    }

    // pass rc in stack as well
    XPUSHs(sv_2mortal(newSViv(rc)));

    STAF_PERL_API_EPILOGUE(1);
}

static char sInfoBuffer[256]  = { 0 };
static int  sInfoBufferInited = 0;

char *STAFGetInformation()
{
   if (!sInfoBufferInited)
   {
       sprintf(sInfoBuffer, "%s STAF Perl support library version 1.0",
               #ifdef NDEBUG
                   "Retail");
               #else
                   "Debug");
               #endif
       sInfoBufferInited = 1;
   }

   return sInfoBuffer;
}

void boot_PLSTAF(CV *cv)
{
    #if defined(STAF_OS_NAME_HPUX) && !defined(__ia64)
        _main();
    #endif

    dXSARGS;
    static char* fileName = __FILE__;

    XS_VERSION_BOOTCHECK;

    newXS("STAF::Register",   PLSTAF_Register,   fileName);
    newXS("STAF::Submit",     PLSTAF_Submit,     fileName);
    newXS("STAF::UnRegister", PLSTAF_UnRegister, fileName);

    XSRETURN_YES;
}

} // extern C
