/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

////////////////////////////////////////////////////////////////////////////////
#include "STAFOSTypes.h"
#include "STAFServiceInterface.h"
#include "STAFError.h"
#include "STAFString.h"
#include <EXTERN.h>
#include <perl.h>
////////////////////////////////////////////////////////////////////////////////
#ifndef   MULTIPLE_INTERPRETERS
static PerlInterpreter *sPerl   = 0;
#endif // MULTIPLE_INTERPRETERS
////////////////////////////////////////////////////////////////////////////////
struct PerlServiceData
{
#ifdef    MULTIPLE_INTERPRETERS
    PerlInterpreter *perl;
#endif // MULTIPLE_INTERPRETERS

    STAFString       name;
    STAFString       exec;
    STAFString       args;
};
////////////////////////////////////////////////////////////////////////////////
/* ************************************************************************* */
/* STAF_2_PERL_PROLOGUE must be the first call in each interface, since it   */
/* makes sure the stack pointer is adjusted before we actually go into the   */
/* perl function.                                                            */
/* ************************************************************************* */
#define STAF_2_PERL_PROLOGUE()    \
    {                             \
        dSP;                      \
        ENTER;                    \
        SAVETMPS;

/* ************************************************************************* */
/* STAF_2_PERL_EPILOGUE must be the last call in each interface, since it    */
/* makes sure the stack pointer is adjusted after we actually get back from  */
/* the perl function.                                                        */
/* ************************************************************************* */
#define STAF_2_PERL_EPILOGUE()    \
        FREETMPS;                 \
        LEAVE;                    \
    }

/* ************************************************************************* */
/* These macros REALLY make the code more readable                           */
/* ************************************************************************* */
#define STAF_2_PERL_BEG_PARAMS() PUSHMARK(SP);
#define STAF_2_PERL_END_PARAMS() PUTBACK;
#define STAF_2_PERL_BEG_RETVAL() SPAGAIN;
#define STAF_2_PERL_END_RETVAL() PUTBACK;

/* ************************************************************************* */
/* Some prototypes/external references                                       */
/* ************************************************************************* */
extern "C" void boot_PLSTAF(void);  // in PLSTAF.cpp

/* ************************************************************************* */
/* Some constants                                                            */
/* ************************************************************************* */
static const unsigned int STAFPerlError = 4000;

////////////////////////////////////////////////////////////////////////////////

unsigned int STAFServiceGetLevelBounds(unsigned int levelID,
                                       unsigned int *minimum,
                                       unsigned int *maximum)
{
    switch (levelID)
    {
        case kServiceInfo:
        case kServiceInit:
        case kServiceAcceptRequest:
        {
            *minimum = 1;
            *maximum = 1;
            break;
        }
        case kServiceTerm:
        case kServiceDestruct:
        {
            *minimum = 0;
            *maximum = 0;
            break;
        }
        default:
        {
            return STAFError::kInvalidAPILevel;
        }
    }

    return 0;
}

unsigned int STAFServiceConstruct(STAFServiceHandle_t *pServiceHandle,
                                  void *pServiceInfo, unsigned int infoLevel,
                                  STAFString_t *pErrorBuffer)
{
    try
    {
        if (infoLevel != 1) return STAFError::kInvalidAPILevel;

        PerlInterpreter *perl = 0;
        PerlServiceData  data;

        STAFServiceInfoLevel1 *info =
            static_cast<STAFServiceInfoLevel1 *>(pServiceInfo);

        data.name = STAFString(info->name);
        data.exec = STAFString(info->exec);

        perl = perl_alloc();

        #ifdef    MULTIPLE_INTERPRETERS
        data.perl = perl;
        #else
        sPerl     = perl;
        #endif // MULTIPLE_INTERPRETERS

        if (perl == 0)
        {
            STAFString error("Out of memory in libPLSTAF.so");
            *pErrorBuffer = error.adoptImpl();
            return STAFPerlError;
        }

        perl_construct(perl);

        *pServiceHandle = new PerlServiceData(data);

        return STAFError::kOk;
    }
    catch (STAFException &e)
    { e.write(); }
    catch (...)
    { /* Do Nothing */ }

    return STAFError::kUnknownError;
}

unsigned int STAFServiceDestruct(STAFServiceHandle_t *pServiceHandle,
                                 void *pDestructInfo,
                                 unsigned int destructLevel,
                                 STAFString_t *pErrorBuffer)
{
    try
    {
        if (destructLevel != 0) return STAFError::kInvalidAPILevel;

        PerlInterpreter *perl = 0;
        PerlServiceData *data =
            static_cast<PerlServiceData *>(*pServiceHandle);

        #ifdef    MULTIPLE_INTERPRETERS
        perl = data->perl;
        #else
        perl = sPerl;
        #endif // MULTIPLE_INTERPRETERS

        perl_destruct(perl);
        perl_free(perl);

        delete data; 
        *pServiceHandle = 0;

        return STAFError::kOk;
    }
    catch (STAFException &e)
    { e.write(); }
    catch (...)
    { /* Do Nothing */ }

    return STAFError::kUnknownError;
}

unsigned int STAFServiceInit(STAFServiceHandle_t serviceHandle,
                             void *pInitInfo, unsigned int initLevel,
                             STAFString_t *pErrorBuffer)
{
    try
    {
        if (initLevel != 1) return STAFError::kInvalidAPILevel;

        PerlInterpreter *perl = 0;
        PerlServiceData *data =
            static_cast<PerlServiceData *>(serviceHandle);
        STAFServiceInitLevel1 *init =
            static_cast<STAFServiceInitLevel1 *>(pInitInfo);

        #ifdef    MULTIPLE_INTERPRETERS
        perl = data->perl;
        #else
        perl = sPerl;
        #endif // MULTIPLE_INTERPRETERS

        data->args = STAFString(init->parms);

        STAFStringBufferPtr name = data->name.toCurrentCodePage();
        STAFStringBufferPtr exec = data->exec.toCurrentCodePage();
        STAFStringBufferPtr args = data->args.toCurrentCodePage();
        STAFStringBufferPtr func = 
            STAFString(data->name + "::init").toCurrentCodePage();

        char *argv[] = { 0, (char *)exec->buffer(), 
                            (char *)args->buffer() };

        perl_parse(perl, boot_PLSTAF, 3, argv, 0);

        unsigned int count = 0;
        unsigned int rc    = 0;

        STAF_2_PERL_PROLOGUE();

        STAF_2_PERL_BEG_PARAMS();
        XPUSHs(sv_2mortal(newSVpv((char *)name->buffer(), 0)));
        XPUSHs(sv_2mortal(newSVpv((char *)args->buffer(), 0)));
        STAF_2_PERL_END_PARAMS();

        count = perl_call_pv((char *)func->buffer(), G_SCALAR);

        STAF_2_PERL_BEG_RETVAL();
        rc = POPi;
        STAF_2_PERL_END_RETVAL();

        STAF_2_PERL_EPILOGUE();

        return rc;
    }
    catch (STAFException &e)
    { e.write(); }
    catch (...)
    { /* Do Nothing */ }

    return STAFError::kUnknownError;
}

unsigned int STAFServiceTerm(STAFServiceHandle_t serviceHandle,
                             void *pTermInfo, unsigned int termLevel,
                             STAFString_t *pErrorBuffer)
{
    try
    {
        if (termLevel != 0) return STAFError::kInvalidAPILevel;

        PerlInterpreter *perl = 0;
        PerlServiceData *data = static_cast<PerlServiceData *>(serviceHandle);

        STAFStringBufferPtr func = 
            STAFString(data->name + "::term").toCurrentCodePage();

        #ifdef    MULTIPLE_INTERPRETERS
        perl = data->perl;
        #else
        perl = sPerl;
        #endif // MULTIPLE_INTERPRETERS

        unsigned int count = 0;
        unsigned int rc    = 0;

        STAF_2_PERL_PROLOGUE();

        STAF_2_PERL_BEG_PARAMS();
        // no parameters
        STAF_2_PERL_END_PARAMS();

        count = perl_call_pv((char *)func->buffer(), G_SCALAR);

        STAF_2_PERL_BEG_RETVAL();
        rc = POPi;
        STAF_2_PERL_END_RETVAL();

        STAF_2_PERL_EPILOGUE();
        
        return rc;
    }
    catch (STAFException &e)
    { e.write(); }
    catch (...)
    { /* Do Nothing */ }

    return STAFError::kUnknownError;
}

unsigned int STAFServiceAcceptRequest(STAFServiceHandle_t serviceHandle,
                                      void *pRequestInfo, unsigned int reqLevel,
                                      STAFString_t *pResultBuffer)
{
    try
    {
        if (reqLevel != 1) return STAFError::kInvalidAPILevel;

        PerlInterpreter *perl = 0;
        PerlServiceData *data = static_cast<PerlServiceData *>(serviceHandle);
        STAFServiceRequestLevel1 *info =
            static_cast<STAFServiceRequestLevel1 *>(pRequestInfo);

        #ifdef    MULTIPLE_INTERPRETERS
        perl = data->perl;
        #else
        perl = sPerl;
        #endif // MULTIPLE_INTERPRETERS

        STAFHandle_t handle = info->handle;
        STAFStringBufferPtr machine = 
            STAFString(info->machine).toCurrentCodePage();
        STAFStringBufferPtr process = 
            STAFString(info->processName).toCurrentCodePage();
        STAFStringBufferPtr request = 
            STAFString(info->request).toCurrentCodePage();
        STAFStringBufferPtr func = 
            STAFString(data->name + "::acceptRequest").toCurrentCodePage();

        unsigned int count = 0;
        unsigned int rc    = 0;
        char        *rb    = 0;

        STAF_2_PERL_PROLOGUE();

        STAF_2_PERL_BEG_PARAMS();
        XPUSHs(sv_2mortal(newSVpv((char *)machine->buffer(), 0)));
        XPUSHs(sv_2mortal(newSVpv((char *)process->buffer(), 0)));
        XPUSHs(sv_2mortal(newSViv(handle)));
        XPUSHs(sv_2mortal(newSVpv((char *)request->buffer(), 0)));
        STAF_2_PERL_END_PARAMS();

        count = perl_call_pv((char *)func->buffer(), G_ARRAY);

        STAF_2_PERL_BEG_RETVAL();
        rb = POPp;
        rc = POPi;
        STAF_2_PERL_END_RETVAL();

        STAFString result(rb);
        *pResultBuffer = result.adoptImpl();

        STAF_2_PERL_EPILOGUE();

        return rc;
    }
    catch (STAFException &e)
    { e.write(); }
    catch (...)
    { /* Do Nothing */ }

    return STAFError::kUnknownError;
}
////////////////////////////////////////////////////////////////////////////////
