/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2005                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.cron;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;
import org.python.util.PythonInterpreter;
import org.python.core.*;

public class CronService implements STAFServiceInterfaceLevel4, Runnable
{
    private STAFHandle fHandle;
    private final String kVersion = "1.2.5";
    private STAFCommandParser fParmsParser = new STAFCommandParser();
    private STAFCommandParser fRegisterParser;
    private STAFCommandParser fListParser;
    private Hashtable fRegTable;
    private static boolean DEBUG = true;
    private Thread fQueueThread;
    private STAFCommandParser fUnregisterParser;
    private int fCronID = 1;
    private String fLineSep;
    private String fCronServiceMachine = "";
    private String fCronServiceName = "";
    private PythonInterpreter fPythonInterpreter;
    private String fHashtableFileDirectory;
    private String fHashtableFileName;
    private String fLocalMachineName;
    private final String kMinute = "minute";
    private final String kHour = "hour";
    private final String kDay = "day";
    private final String kMonth = "month";
    private final String kWeekday = "weekday";
    private HashMap fSubmittedRequests = new HashMap();
    
    public CronService()
    {
    }    
    
    public int init(STAFServiceInterfaceLevel4.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
        
            fCronServiceMachine = (fHandle.submit2("local", "var", 
                            "resolve {STAF/Config/EffectiveMachine}")).result;
                            
            fCronServiceName = info.name;
            
            fLocalMachineName = fCronServiceMachine;
            
            if (info.serviceJar != null) 
            {
                // Create the Jython Library directory if it doesn't exist yet
                // from the jython library data in the service's jarfile.
                // Set the classpath to point to the service's jar file and to
                // the jython.jar.
                
                STAFResult res = STAFServiceSharedJython.setupJython(
                                     info.serviceJar, fHandle);
                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res.rc;
                }
                        
                // Initialize the PythonInterpreter's python.home to point to the
                // shared jython installation.

                Properties p = new Properties();
                p.setProperty("python.home",
                              STAFServiceSharedJython.getJythonDirName());
                PythonInterpreter.initialize(System.getProperties(), p, null);
            }

            // instantiate parsers as not case sensitive
            fRegisterParser = new STAFCommandParser(0, false);
            fListParser = new STAFCommandParser(0, false);
            fUnregisterParser = new STAFCommandParser(0, false);

            fRegisterParser.addOption("REGISTER", 1, 
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("MACHINE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONMACHINE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOptionGroup("MACHINE PYTHONMACHINE", 1, 1);

            fRegisterParser.addOption("SERVICE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("PYTHONSERVICE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOptionGroup("SERVICE PYTHONSERVICE", 1, 1);

            fRegisterParser.addOption("REQUEST", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("PYTHONREQUEST", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOptionGroup("REQUEST PYTHONREQUEST", 1, 1);
           
            fRegisterParser.addOption("PREPARE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("MINUTE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("HOUR", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("DAY", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("MONTH", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("WEEKDAY", 1, 
                                      STAFCommandParser.VALUEREQUIRED);

            // Require at least one time interval option to be specified
            fRegisterParser.addOptionGroup(
                "MINUTE HOUR DAY MONTH WEEKDAY", 1, 5);

            fUnregisterParser.addOption("UNREGISTER", 1, 
                                        STAFCommandParser.VALUENOTALLOWED);

            fUnregisterParser.addOption("ID", 1, 
                                        STAFCommandParser.VALUEREQUIRED);
                                        
            fUnregisterParser.addOptionNeed("UNREGISTER", "ID");

            fListParser.addOption("LIST", 1, 
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("MACHINE", 1, 
                                  STAFCommandParser.VALUEREQUIRED);
            
            fLineSep = (fHandle.submit2("local", "var", 
                        "resolve {STAF/Config/Sep/Line}")).result;
                        
            String fileSep = (fHandle.submit2("local", "var",
                               "resolve {STAF/Config/Sep/File}")).result;
            String stafRoot = (fHandle.submit2("local", "var",
                               "resolve {STAF/Config/STAFRoot}")).result;
                               
            String userHome = System.getProperty("user.home");
        
            if (!(userHome.endsWith(fileSep)))
            {
                userHome += fileSep;
            }
                               
            fHashtableFileDirectory = userHome + "crondata";
                                                            
            File dir = new File(fHashtableFileDirectory);
            
            if (!dir.exists())
            {                    
                dir.mkdirs();
            }
      
            fHashtableFileName = fHashtableFileDirectory + fileSep + 
                                 "cron.ser";
             
            loadHashtable();            
            
            fQueueThread = new Thread(this);
            fQueueThread.start(); // this calls the run() method
        }
        catch (STAFException e)
        {
            rc = STAFResult.STAFRegistrationError;
        }

        return rc;
    }
    
    private int handleParms(STAFServiceInterfaceLevel4.InitInfo info)
        throws STAFException
    {
        return 0;
    }
        
    public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo info)
    {               
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("register"))
        {
            return handleRegister(info);
        }
        else if (lowerRequest.startsWith("unregister"))
        {
            return handleUnregister(info);
        }
        else if (lowerRequest.startsWith("list"))
        {
            return handleList(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp();
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion();
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown Cron Request: " + 
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp()
    {
        return new STAFResult(STAFResult.Ok,
         "Cron Service Help" + fLineSep
         + "REGISTER   MACHINE <machine> | PYTHONMACHINE <machine>"
         + fLineSep + 
         "           SERVICE <service> | PYTHONSERVICE <machine>"
         + fLineSep + 
         "           REQUEST <request> | PYTHONREQUEST <request>" 
         + fLineSep + "           [PREPARE <script>]"
         + fLineSep + "           [MINUTE <minute>] [HOUR <hour>] "
         + fLineSep + "           [DAY <day>] [MONTH <month>] "
         + fLineSep + "           [WEEKDAY <weekday>]"
         + fLineSep + "UNREGISTER ID <registrationID>"
         + fLineSep + "LIST       [MACHINE <machine>]"
         + fLineSep + "VERSION" + fLineSep + "HELP");
    }

    private STAFResult handleList(STAFServiceInterfaceLevel4.RequestInfo info)
    {           
        StringBuffer result = new StringBuffer();
        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);
        
        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        String machineParm = "";
        String typeParm = "";

        try
        {
            STAFResult trust = 
                fHandle.submit2(info.localMachine, "trust", 
                                "get machine " + info.machine);
    
            if (trust.rc != 0)
            {
                return new STAFResult(trust.rc, trust.result);
            }

            if (Integer.parseInt(trust.result) < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                                  "Trust level 3 required for LIST request."
                                  + "Requesting machine's trust level: "
                                  + trust.result);
            }

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            if (info.diagEnabled == 1)
            {
                // Record diagnostics because the output for a LIST request
                // will change in STAF V3.0

                String trigger = new String("STAF/V3.0-Mig/CRON LIST");
                String source  = new String(info.processName + ";" +
                                            info.machine + ";" +
                                            info.handle);
                fHandle.submit2(
                    STAFHandle.ReqFireAndForget,
                    "local", "DIAG",
                    "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                    " SOURCE " + STAFUtil.wrapData(source));
            }

            int registrationTableSize = fRegTable.size();
            if (registrationTableSize > 0)
            {
                resolvedValue =
                    resolveVar(info.localMachine, 
                               parsedRequest.optionValue("machine"), 
                               info.handle);
                machineParm = resolvedValue.result;
               
                CronData theData;
                boolean match;
                Enumeration registrationEnum = fRegTable.elements();
                Enumeration keys = fRegTable.keys();
                
                for (Enumeration e = fRegTable.elements(); 
                     e.hasMoreElements();)
                {
                    Integer id = (Integer) keys.nextElement();
                    match = false;
                    theData = (CronData) e.nextElement();
                    
                    String minutes = "";
                    String hours = "";
                    String days = "";
                    String months = "";
                    String weekdays = "";

                    if (machineParm.equals("") | 
                        machineParm.equals(theData.fMachine))
                    {                                                
                        Enumeration minuteEnum = theData.fMinute.elements();
                        
                        while (minuteEnum.hasMoreElements())
                        {
                            minutes += (minuteEnum.nextElement()).toString();
                            
                            if (minuteEnum.hasMoreElements())
                                minutes += ",";
                        }
                                                                        
                        Enumeration hourEnum = theData.fHour.elements();
                        
                        while (hourEnum.hasMoreElements())
                        {
                            hours += (hourEnum.nextElement()).toString();
                            
                            if (hourEnum.hasMoreElements())
                                hours += ",";
                        }                        
                                                
                        Enumeration dayEnum = theData.fDay.elements();
                        
                        while (dayEnum.hasMoreElements())
                        {
                            days += (dayEnum.nextElement()).toString();
                            
                            if (dayEnum.hasMoreElements())
                                days += ",";
                        }                        
                                                
                        Enumeration monthEnum = theData.fMonth.elements();
                        
                        while (monthEnum.hasMoreElements())
                        {
                            months += (monthEnum.nextElement()).toString();
                            
                            if (monthEnum.hasMoreElements())
                                months += ",";
                        }                        
                                                
                        Enumeration weekdayEnum = theData.fWeekday.elements();
                        
                        while (weekdayEnum.hasMoreElements())
                        {
                            weekdays += (weekdayEnum.nextElement()).toString();
                            
                            if (weekdayEnum.hasMoreElements())
                                weekdays += ",";
                        }

                        result.append(id + ";");
                        result.append(theData.fMachine + ";");
                        result.append(theData.fPythonMachine + ";");
                        result.append(theData.fService + ";");
                        result.append(theData.fPythonService + ";");
                        result.append(theData.fRequest + ";");
                        result.append(theData.fPythonRequest + ";");
                        result.append(theData.fPrepare + ";");
                        result.append(minutes + ";");
                        result.append(hours + ";");
                        result.append(days + ";");
                        result.append(months + ";");
                        result.append(weekdays + ";");
                        result.append(fLineSep);
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError, 
                                  "Internal Java error.");
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }
    
    private STAFResult handleRegister(
                           STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String originMachine = "";
        String machine = "";
        boolean pythonMachine = true;
        String service = "";
        boolean pythonService = true;
        String request = "";
        boolean pythonRequest = true;
        String prepare = "";

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest = 
            fRegisterParser.parse(info.request);
            
        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the 
        // corresponding local variable to the user-specified value.

        try
        {
            STAFResult trust = 
                fHandle.submit2(info.localMachine, "trust", 
                                "get machine " + info.machine);

            if (trust.rc != 0)
            {
                return new STAFResult(trust.rc, trust.result);
            }

            if (Integer.parseInt(trust.result) < 5)
            {
                return new STAFResult(STAFResult.AccessDenied,
                              "Trust level 5 required for REGISTER request."
                              + " Requesting machine's trust level: "
                              + trust.result);
            }

            originMachine = info.machine;

            if (parsedRequest.optionTimes("machine") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("machine"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;

                machine = resolvedValue.result;
                pythonMachine = false;                
            }
            else if (parsedRequest.optionTimes("pythonmachine") > 0)
            {
                machine = parsedRequest.optionValue("pythonmachine");
                pythonMachine = true;
            }
            
            if (machine.equalsIgnoreCase("'local'"))
            {
                machine = "'" + fLocalMachineName + "'";
            }
            else if (machine.equalsIgnoreCase("local"))
            {
                machine = fLocalMachineName;
            }
            
            if (parsedRequest.optionTimes("service") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("service"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;

                service = resolvedValue.result;
                pythonService = false;
            }
            else if (parsedRequest.optionTimes("pythonservice") > 0)
            {
                service = parsedRequest.optionValue("pythonservice");
                pythonService = true;
            }
            
            if (parsedRequest.optionTimes("request") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("request"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;

                request = resolvedValue.result;
                pythonRequest = false;
            }
            else if (parsedRequest.optionTimes("pythonrequest") > 0)
            {
                request = parsedRequest.optionValue("pythonrequest");
                pythonRequest = true;
            }
            
            if (parsedRequest.optionTimes("prepare") > 0)
            {
                prepare = parsedRequest.optionValue("prepare");
            }
            
            Vector minute = new Vector();
            Vector hour = new Vector();
            Vector day = new Vector();
            Vector month = new Vector();
            Vector weekday = new Vector();
            
            if (parsedRequest.optionTimes(kMinute) > 0)
            {
                minute = cronParse(parsedRequest.optionValue(kMinute), kMinute);

                if (minute == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kMinute + ":  " +
                        parsedRequest.optionValue(kMinute));
                }
            }
            
            if (parsedRequest.optionTimes(kHour) > 0)
            {
                hour = cronParse(parsedRequest.optionValue(kHour), kHour);

                if (hour == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kHour + ":  " +
                        parsedRequest.optionValue(kHour));
                }
            }
            
            if (parsedRequest.optionTimes(kDay) > 0)
            {
                day = cronParse(parsedRequest.optionValue(kDay), kDay);

                if (day == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kDay + ":  " +
                        parsedRequest.optionValue(kDay));
                }
            }
            
            if (parsedRequest.optionTimes(kMonth) > 0)
            {
                month = cronParse(parsedRequest.optionValue(kMonth), kMonth);

                if (month == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kMonth + ":  " +
                        parsedRequest.optionValue(kMonth));
                }
            }
            
            if (parsedRequest.optionTimes(kWeekday) > 0)
            {
                weekday = cronParse(parsedRequest.optionValue(kWeekday),
                                    kWeekday);

                if (weekday == null)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for " + kWeekday + ":  " +
                        parsedRequest.optionValue(kWeekday));
                }
            }
            
            Integer index = new Integer(fCronID++);
                
            fRegTable.put(index, (Object) new 
                CronData(originMachine, machine, pythonMachine, service,
                pythonService, request, pythonRequest, prepare, minute,
                hour, day, month, weekday));
                                     
            sResult = new STAFResult(STAFResult.Ok, index.toString());
            saveHashtable();
            
            String message = "Register ID=" + index + " originMachine=" +
                originMachine + " machine=" + machine + " pythonMachine=" +
                pythonMachine + " service=" + service + " pythonService=" + 
                pythonService + " request=" + request + " pythonRequest=" +
                pythonRequest + " prepare=" + prepare + " minute=" + minute +
                " hour=" + hour + " day=" + day + " month=" + month + 
                " weekday=" + weekday;
            
            fHandle.submit2("local", "log", "log machine logname " +
                fCronServiceName + " level info message " + 
                STAFUtil.wrapData(message));
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError, 
                                  "Internal Java error.");
        }
    
        return sResult;
    }
    
    private STAFResult handleUnregister(
                           STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String reqMachine = "";
        String service;
        String request;
        Integer id;
        
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest = 
            fUnregisterParser.parse(info.request);
            
        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the 
        // corresponding local variable to the user-specified value.

        try
        {
            STAFResult trust = 
                fHandle.submit2(info.localMachine, "trust", 
                                "get machine " + info.machine);

            if (trust.rc != 0)
            {
                return new STAFResult(trust.rc, trust.result);
            }

            if (Integer.parseInt(trust.result) < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                              "Trust level 3 required for UNREGISTER request."
                              + " Requesting machine's trust level: "
                              + trust.result);
            }
                        
            reqMachine = info.machine;
            
            resolvedValue = resolveVar(info.localMachine, 
                                       parsedRequest.optionValue("id"), 
                                       info.handle);

            if (resolvedValue.rc != 0)
            {
                return resolvedValue;
            }

            id = new Integer(resolvedValue.result);
            
            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(STAFResult.DoesNotExist, 
                                      "ID " + id + " not found");
            }
            else
            {
                CronData data = (CronData)fRegTable.get(id);
                
                if (!(data.fOriginMachine.equals(reqMachine)))
                {
                    if (Integer.parseInt(trust.result) < 4)
                    {
                        return new STAFResult(STAFResult.AccessDenied,
                                      "Trust level 4 required for UNREGISTER " +
                                      " when the requesting machine did not " +
                                      " submit the REGISTER request. " 
                                      + " Requesting machine's trust level: "
                                      + trust.result);
                    }
                }

                fRegTable.remove(id);
                
                if (fRegTable.isEmpty())
                {
                    fCronID = 1;
                }

                saveHashtable();
                
                String message = "Unregister ID=" + id;

                fHandle.submit2("local", "log", "log machine logname " +
                    fCronServiceName + " level info message " + 
                    STAFUtil.wrapData(message));
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            return new STAFResult(STAFResult.JavaError, "Internal Java error.");
        }
    
        return sResult;
    }

    private STAFResult handleVersion()
    {
        return new STAFResult(STAFResult.Ok, kVersion);
    }
    
    private Vector cronParse(String msg, String type)
    {
        Vector result = new Vector();
        
        if (msg.equalsIgnoreCase("ANY") || msg.equals("*"))
        {
            result.add("ANY");
            return result;
        }
                
        StringTokenizer options = new StringTokenizer(msg, ",");
            
        while (options.hasMoreElements())
        {
            String element = ((String)options.nextElement()).trim();
        
            // Check if a range was specified:  <start>-<end>

            int firstDashIndex = element.indexOf("-");

            if (firstDashIndex == -1)
            {
                // Single value specified.  Verify that it is valid.
                    
                if (!isValidValue(element, type))
                    return null;

                result.add(element);
            }
            else
            {
                // Range was specified

                if (element.indexOf("-", firstDashIndex + 1) > -1)
                {
                    // Invalid.  Can only contain one dash.
                    return null;
                }

                StringTokenizer items = new StringTokenizer(element, "-");

                String start = null;
                String end = null;

                while (items.hasMoreElements())
                {
                    if (start == null)
                        start = ((String)items.nextElement()).trim();
                    else
                        end = ((String)items.nextElement()).trim();
                }

                // Verify that the values specified for the range are valid
                
                if (!isValidValue(start, type))
                    return null;

                if (!isValidValue(end, type))
                    return null;

                // Types other than weekday will always contain numeric values
                // A Weekday type can contain a numeric value or a string value

                if (!type.equals(kWeekday))
                {
                    // Type is kMonth, kDay, kHour, or kMinute

                    // Iterate through the numeric range

                    int startNum = new Integer(start).intValue();
                    int endNum = new Integer(end).intValue();

                    for (int i = startNum; i <= endNum; i++)
                    {
                        result.add(new String(new Integer(i).toString()));
                    }
                }
                else
                {
                    // Type is kWeekday

                    // Check if using string or numeric values for weekday.

                    boolean startIsNumeric = true;
                    boolean endIsNumeric = true;
                        
                    if (equalsWeekday(start)) startIsNumeric = false;
                    if (equalsWeekday(end)) endIsNumeric = false;
                        
                    int startNum = 0;
                    int endNum = 0;

                    // Get the corresponding numeric value for weekday if
                    // specified as a string, e.g. 0 for "Sunday".

                    if (startIsNumeric)
                        startNum = new Integer(start).intValue();
                    else
                        startNum = getNumericWeekday(start);

                    if (endIsNumeric)
                        endNum = new Integer(end).intValue();
                    else
                        endNum = getNumericWeekday(end);
                        
                    // Iterate throught the range, using the String or numeric
                    // representation based on the start value.

                    for (int i = startNum; i <= endNum; i++)
                    {
                        if (startIsNumeric)
                            result.add(new String(new Integer(i).toString()));
                        else
                            result.add(getStringWeekday(i));
                    }
                }
            }
        }

        return result;
    }
    
    boolean isValidValue(String value, String type)
    {
        if (value == null)
            return false;

        int valueNum = 0;

        if (type.equals(kWeekday))
        {
            if (equalsWeekday(value))
                return true;
        }
            
        try
        {
            valueNum = new Integer(value).intValue();
        }
        catch (NumberFormatException ex)
        {
            return false;
        }
        
        if (type.equals(kMinute))
        {
            if (valueNum >= 0 && valueNum <= 59)
                return true;
            else
                return false;
        }
        else if (type.equals(kHour))
        {
            if (valueNum >= 0 && valueNum <= 23)
                return true;
            else
                return false;
        }
        else if (type.equals(kDay))
        {
            if (valueNum >= 1 && valueNum <= 31)
                return true;
            else
                return false;
        }
        else if (type.equals(kMonth))
        {
            if (valueNum >= 1 && valueNum <= 12)
                return true;
            else
                return false;
        }
        else if (type.equals(kWeekday))
        {
            if (valueNum >= 0 && valueNum <= 6)
                return true;
            else
                return false;
        }
        else
        {
            return false;
        }
    }

    private boolean equalsWeekday(String msg)
    {
         if ((msg.equalsIgnoreCase("Sunday")) ||
             (msg.equalsIgnoreCase("Monday")) ||
             (msg.equalsIgnoreCase("Tuesday")) ||
             (msg.equalsIgnoreCase("Wednesday")) ||
             (msg.equalsIgnoreCase("Thursday")) ||
             (msg.equalsIgnoreCase("Friday")) ||
             (msg.equalsIgnoreCase("Saturday")))
         {
            return true;
         }
         else
         {
            return false;
         }
    }

    private int getNumericWeekday(String weekday)
    {
         if (weekday.equalsIgnoreCase("Sunday"))
             return 0;
         else if (weekday.equalsIgnoreCase("Monday"))
             return 1;
         else if (weekday.equalsIgnoreCase("Tuesday"))
             return 2;
         else if (weekday.equalsIgnoreCase("Wednesday"))
             return 3;
         else if (weekday.equalsIgnoreCase("Thursday"))
             return 4;
         else if (weekday.equalsIgnoreCase("Friday"))
             return 5;
         else if (weekday.equalsIgnoreCase("Saturday"))
             return 6;
         else
             return -1;
    }

    private String getStringWeekday(int weekday)
    {
         if (weekday == 0)
             return "Sunday";
         else if (weekday == 1)
             return "Monday";
         else if (weekday == 2)
             return "Tuesday";
         else if (weekday == 3)
             return "Wednesday";
         else if (weekday == 4)
             return "Thursday";
         else if (weekday == 5)
             return "Friday";
         else if (weekday == 6)
             return "Saturday";
         else
             return "";
    }

    private STAFResult integerCheck(String option, String optionValue, 
                                    boolean defaultOption)
    {
        char[] c = optionValue.toCharArray();
        for (int count = 0; count < c.length; count++)
        {
            if (!Character.isDigit(c[count]) && !defaultOption)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      "The value you entered for "
                                      + option + ", " + optionValue
                                      + ", is not a number. the value of "
                                      + option + " must be" + " numeric.");
            }
            else if (!Character.isDigit(c[count]))
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                        "The default value entered at STAF initialization for "
                        + option + ", " + optionValue + ", is not a number. "
                        + "STAF must be reinitialized with a numeric value"
                        + " for " + option + ".");
            }
        }

        return new STAFResult(STAFResult.Ok, optionValue);
    }

    private STAFResult resolveAndCheckInt(String machine, String option,
                                          String optionValue, int handle,
                                          boolean defaultOption)
    {
        STAFResult resolvedValue = resolveVar(machine, optionValue, handle);

        if (resolvedValue.rc != 0)
        {
            return resolvedValue;
        }

        return integerCheck(option, resolvedValue.result, defaultOption);
    }

    private STAFResult resolveVar(String machine, String optionValue, 
                                  int handle)
    {
        String value = "";
        STAFResult resolvedResult = null;

        if (optionValue.indexOf("{") != -1)
        {
            resolvedResult =
                fHandle.submit2(machine, "var", "handle " + handle + 
                                " resolve " + STAFUtil.wrapData(optionValue));

            if (resolvedResult.rc != 0)
            {
                return resolvedResult;
            }

            value = resolvedResult.result;
        }
        else
        {
            value = optionValue;
        }

        return new STAFResult(STAFResult.Ok, value);
    }

    public void run()
    {
        Calendar initialTimestamp = Calendar.getInstance();    

        int lastMinute = initialTimestamp.get(Calendar.MINUTE);
        int lastHour = initialTimestamp.get(Calendar.HOUR_OF_DAY);
        int lastDay = initialTimestamp.get(Calendar.DAY_OF_MONTH);
          // MONTH is 0-based, so add 1
        int lastMonth = initialTimestamp.get(Calendar.MONTH) + 1;
        int lastWeekday = initialTimestamp.get(Calendar.DAY_OF_WEEK);
        
        int currentMinute;
        int currentHour;
        int currentDay;
        int currentMonth;
        int currentWeekday;
        
        boolean minuteChanged = false;
        boolean hourChanged = false;
        boolean dayChanged = false;
        boolean monthChanged = false;
        boolean weekdayChanged = false;

        STAFResult queueGetResult;
        
        for (;;)
        {
            queueGetResult = fHandle.submit2("local", "queue", "get wait 10000");

            if (queueGetResult.rc == 0)
            {
                String message = queueGetResult.result;
                
                if (message.indexOf("Cron/End") > -1)
                {
                    // this message indicates that STAFProc is ending and
                    // the service is terminating 
                    
                    try
                    {
                        fHandle.unRegister();
                    }
                    catch (STAFException ex)
                    {
                        if (DEBUG)
                        {
                            ex.printStackTrace();
                        }
                    }

                    return;
                }
                else
                {
                    int priorityPos = message.indexOf(';');
                    int timestampPos = message.indexOf(';', priorityPos + 1);
                    int machinePos = message.indexOf(';', timestampPos + 1);
                    int processPos = message.indexOf(';', machinePos + 1);
                    int handlePos = message.indexOf(';', processPos + 1);        
                
                    message = message.substring(handlePos + 1);
                    
                    if (!message.startsWith("STAF/RequestComplete"))
                    {
                        return;
                    }
                    
                    int rcIndex = message.indexOf(";");
                    int resultIndex = message.indexOf(";", rcIndex + 1);
                    String reqNum = message.substring(21, rcIndex);
                    String rc = message.substring(rcIndex + 1, resultIndex);
                    String result = message.substring(resultIndex + 1);
                    
                    String level = "";
                    
                    if (rc.equals("0"))
                        level = "info";
                    else
                        level = "error";
                        
                    String request = (String)fSubmittedRequests.get(reqNum);

                    fHandle.submit2("local", "log", "log machine logname " +
                        fCronServiceName + " level " + level + " message " + 
                        STAFUtil.wrapData("Request #" + reqNum + ": RC=" + 
                        rc + ", Result=" + result + ", Request submitted: " +
                        request));
                        
                    fSubmittedRequests.remove(reqNum);
                }
            }
                        
            minuteChanged = false;
            hourChanged = false;
            dayChanged = false;
            monthChanged = false;
            weekdayChanged = false;
            
            Calendar currentTimestamp = Calendar.getInstance();
            
            currentMinute = currentTimestamp.get(Calendar.MINUTE);
            currentHour = currentTimestamp.get(Calendar.HOUR_OF_DAY);
            currentDay = currentTimestamp.get(Calendar.DAY_OF_MONTH);
              // MONTH is 0-based, so add 1
            currentMonth = currentTimestamp.get(Calendar.MONTH) + 1;
            currentWeekday = currentTimestamp.get(Calendar.DAY_OF_WEEK);

            if (currentMinute != lastMinute)
            {
                minuteChanged = true;
                lastMinute = currentMinute;
            }
            else
            {
                minuteChanged = false;
            }            
            
            if (currentHour != lastHour)
            {
                hourChanged = true;
                lastHour = currentHour;
            }
            else
            {
                hourChanged = false;
            }
            
            if (currentDay != lastDay)
            {
                dayChanged = true;
                lastDay = currentDay;
            }
            else
            {
                dayChanged = false;
            }
            
            if (currentMonth != lastMonth)
            {
                monthChanged = true;
                lastMonth = currentMonth;
            }
            else
            {
                monthChanged = false;
            }
            
            if (currentWeekday != lastWeekday)
            {
                weekdayChanged = true;
                lastWeekday = currentWeekday;
            }
            else
            {
                weekdayChanged = false;
            }
            
            if (minuteChanged || hourChanged || dayChanged || 
                monthChanged || weekdayChanged)
            {
                // create new Python Interpreter for each Cron
                fPythonInterpreter = new PythonInterpreter();
    
                CronData theData;
                
                for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
                {
                    Integer id = (Integer)e.nextElement();
                    
                    theData = (CronData) fRegTable.get(id);
                    
                    boolean minuteMatch = (theData.fMinute.contains(
                        String.valueOf(currentMinute)) ||
                        theData.fMinute.contains("ANY"));
    
                    boolean minuteMatchOrEmpty = theData.fMinute.isEmpty() ||
                        minuteMatch;
                        
                    boolean hourMatch = (theData.fHour.contains(
                        String.valueOf(currentHour)) ||
                        theData.fHour.contains("ANY"));
                       
                    boolean hourMatchOrEmpty = theData.fHour.isEmpty() ||
                        hourMatch;
                        
                    boolean dayMatch = (theData.fDay.contains(
                        String.valueOf(currentDay)) ||
                        theData.fDay.contains("ANY"));
                        
                    boolean dayMatchOrEmpty = theData.fDay.isEmpty() ||
                        dayMatch;
                        
                    boolean monthMatch = (theData.fMonth.contains(
                        String.valueOf(currentMonth)) ||
                        theData.fMonth.contains("ANY"));
                        
                    boolean monthMatchOrEmpty = theData.fMonth.isEmpty() ||
                        monthMatch;
                        
                    boolean weekdayMatch = (theData.fWeekday.contains(
                        String.valueOf(currentWeekday)) ||
                        theData.fWeekday.contains("ANY"));
                        
                    boolean weekdayMatchOrEmpty = theData.fWeekday.isEmpty() ||
                        weekdayMatch;
                    
                    boolean minuteTrigger = minuteChanged && minuteMatch;
                        
                    boolean hourTrigger = hourChanged && hourMatch;
                        
                    boolean dayTrigger = dayChanged && dayMatch;
                        
                    boolean monthTrigger = monthChanged && monthMatch;

                    boolean weekdayTrigger = weekdayChanged && weekdayMatch;
                        
                    // All 5 time units must match or be empty, and there 
                    // must be at least one trigger.
                    if ((minuteMatchOrEmpty && hourMatchOrEmpty && 
                         dayMatchOrEmpty && monthMatchOrEmpty && 
                         weekdayMatchOrEmpty) &&
                        (minuteTrigger || hourTrigger || dayTrigger || 
                         monthTrigger || weekdayTrigger))
                    {
                        pySetVar("STAFCronSubmit", "true");
                            
                        Object submit = pyGetVar("STAFCronSubmit");
                                                             
                        if ((submit != null) && 
                            (submit.toString().equals("true")))
                        {
                        
                            String prepare = theData.fPrepare;
                            pyExec(prepare);
                            
                            boolean pythonMachine = theData.fPythonMachine;
                            boolean pythonService = theData.fPythonService;
                            boolean pythonRequest = theData.fPythonRequest;

                            String machine = "";
                            String service = "";
                            String request = "";
                                          
                            if (pythonMachine)
                            {
                                machine = pyStringEval(theData.fMachine);
                            }
                            else
                            {
                                machine = theData.fMachine;
                            }
                            
                            if (pythonService)
                            {
                                service = pyStringEval(theData.fService);
                            }
                            else
                            {
                                service = theData.fService;
                            }
                           
                            if (pythonRequest)
                            {
                                request = pyStringEval(theData.fRequest);
                            }
                            else
                            {
                                request = theData.fRequest;
                            }
                                                        
                            STAFResult matchResult = fHandle.submit2(
                                STAFHandle.ReqQueue, machine, 
                                service, request);
                                
                            fSubmittedRequests.put(matchResult.result,
                                "machine=" + machine + " service=" + service +
                                " request=" + request);
                                
                            String triggers = "";
                            
                            if (minuteTrigger)
                            {
                                triggers += "minute[" + currentMinute + "],";
                            }
                            
                            if (hourTrigger || hourMatch)
                            {
                                triggers += "hour[" + currentHour + "],";
                            }
                            
                            if (dayTrigger || dayMatch)
                            {
                                triggers += "day[" + currentDay + "],";
                            }
                            
                            if (monthTrigger || monthMatch)
                            {
                                triggers += "month[" + currentMonth + "],";
                            }
                            
                            if (weekdayTrigger || weekdayMatch)
                            {
                                triggers += "weekday[" + currentWeekday + "],";
                            }
                            
                            if (triggers.endsWith(","))
                            {
                                triggers = triggers.substring(0, 
                                    triggers.length() - 1);
                            }
                                
                            String message = "Submitted request #" + 
                                matchResult.result + " ID=" + id.toString() + 
                                " triggers=" + triggers + 
                                " machine=" + machine + " service=" + service +
                                " request=" + request;

                            fHandle.submit2("local", "log", 
                                "log machine logname " +
                                fCronServiceName + " level info message " + 
                                STAFUtil.wrapData(message));
                        }
                    }
                }
            }
        }
    
    }
    
    public void loadHashtable()
    {
        try 
        {
            File fHashtableFile = new File(fHashtableFileName);
            if (!fHashtableFile.exists()) 
            {
                fRegTable = new Hashtable();
                return; 
            }
            else 
            {
                ObjectInputStream ois = 
                    new ObjectInputStream(new FileInputStream(fHashtableFile));
                fRegTable = (Hashtable)(ois.readObject());
                ois.close();		
                
                for (Enumeration e = fRegTable.elements(); 
                     e.hasMoreElements();)
                {
                    CronData data = (CronData) e.nextElement();
                    
                    // XXX Do something here?
                }
 
                fCronID = findNextCronID();                

            }	
        } 
        catch (IOException e) 
        {
            fRegTable = new Hashtable();
            return;
        }
        catch (ClassNotFoundException e) 
        {
            fRegTable = new Hashtable();
            return;
        }
    }
    
    public void saveHashtable()
    {
        try
        {            
            ObjectOutputStream oos = 
                new ObjectOutputStream(new 
                    FileOutputStream(fHashtableFileName));
            oos.writeObject(((Object)(fRegTable)));				
            oos.close();                         
        } 
        catch (IOException ex) 
        {
            ex.printStackTrace();
        }
    }
    
    public int findNextCronID()
    {
        int highestID = 0;
        for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
        {
            Integer id = (Integer) e.nextElement();
            if (id.intValue() > highestID)
            {
                highestID = id.intValue();
            }
        }
        
        return highestID + 1;
    }
    
    public int term()
    {
        fHandle.submit2(STAFHandle.ReqFireAndForget, "local",
                        "queue", "queue message Cron/End");
        
        return STAFResult.Ok;
    }   
     
    public String pyStringEval(String value)
    {
        synchronized(fPythonInterpreter)
        {
            PyObject result = null;
            try 
            {                
                result = fPythonInterpreter.eval(value);
            }
            catch (PyException ex)
            {
                ex.printStackTrace();
            }
            
            return result.toString();
        }
    }
    
    public void pySetVar(String varName, Object value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.set(varName, value);
            }
            catch (PyException e)
            {
                fHandle.submit2("local", "log", "log machine logname " +
                    fCronServiceName + " level error message " + 
                    STAFUtil.wrapData("PySetVar PyException caught setting " +
                    " var=" + varName + " to value=" + value + e.toString()));
            } 
        }
    }
    
    public Object pyGetVar(String varName)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                return fPythonInterpreter.get(varName);
            }
            catch (PyException e)
            {
                fHandle.submit2("local", "log", "log machine logname " +
                    fCronServiceName + " level error message " + 
                    STAFUtil.wrapData("PyGetVar PyException caught getting " +
                    " var=" + varName + " to value=" + varName + 
                    e.toString()));
                    
                return null;
            } 
        }
    }
    
    public void pyExec(String value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.exec(value);
            }
            catch (PyException e)
            {
                e.printStackTrace();
            } 
        }
    }
    
}
