/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.email;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.io.*;
import java.util.*;
import java.net.*;

public class EmailService implements STAFServiceInterfaceLevel3
{
    private STAFHandle fHandle;
    private final String kVersion = "1.1.5";
    private STAFCommandParser fSendParser;
    private STAFCommandParser fParmsParser;
    private String fLineSep;
    private static Socket fSocket;
    private static BufferedReader fInStream;
    private static PrintStream fOutStream;
    private static String fHostname;
    private static String fMailserver = "";
    private static int fMailport = 25;
    
    private static boolean DEBUG = false;
    
    public EmailService() {}    
    
    public int init(STAFServiceInterfaceLevel3.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
        }
        catch (STAFException e)
        {
            return STAFResult.STAFRegistrationError;
        }
        
            // Send parser
        fSendParser = new STAFCommandParser(0, false);

        fSendParser.addOption("SEND", 1, 
                              STAFCommandParser.VALUENOTALLOWED);

        fSendParser.addOption("TO", 0, 
                              STAFCommandParser.VALUEREQUIRED);
                              
        fSendParser.addOption("SUBJECT", 1, 
                              STAFCommandParser.VALUEREQUIRED);

        fSendParser.addOption("MESSAGE", 1, 
                              STAFCommandParser.VALUEREQUIRED);
                              
        fSendParser.addOption("FILE", 1, 
                              STAFCommandParser.VALUEREQUIRED);
                              
        fSendParser.addOption("MACHINE", 1, 
                              STAFCommandParser.VALUEREQUIRED);
                               
        fSendParser.addOptionGroup("MESSAGE FILE", 1, 1);
        
        fSendParser.addOptionNeed("MACHINE", "FILE");
                               
        fLineSep = (fHandle.submit2("local", "var", 
                    "resolve {STAF/Config/Sep/Line}")).result;
                    
        if (info.parms == null)
        {
            System.out.println("MAILSERVER is a required parameter for the " +
                "STAFEmail service");
            rc = STAFResult.ServiceConfigurationError;
                return rc;
        }
        else
        {
            fParmsParser = new STAFCommandParser(0, false);

            fParmsParser.addOption("MAILSERVER", 1, 
                STAFCommandParser.VALUEREQUIRED);
                               
            fParmsParser.addOption("PORT", 1, 
                STAFCommandParser.VALUEREQUIRED);
                               
            fParmsParser.addOption("DEBUG", 1, 
                STAFCommandParser.VALUENOTALLOWED);
                
            rc = handleParms(info);
                        
            if (rc != STAFResult.Ok)
            {
                rc = STAFResult.ServiceConfigurationError;
                return rc;
            }
        }

        return rc;
    }  
    
    private int handleParms(STAFServiceInterfaceLevel3.InitInfo info)
    {
        int rc = STAFResult.Ok;
        
        STAFCommandParseResult parseResult= fParmsParser.parse(info.parms);

        if (parseResult.rc != STAFResult.Ok)
        {
            return parseResult.rc;
        }

        if (parseResult.optionTimes("MAILSERVER") == 0)
        {
            System.out.println("MAILSERVER is a required parameter for the " +
                "STAFEmail service");
            rc = STAFResult.ServiceConfigurationError;
                return rc;
        }
        else
        {
            fMailserver = parseResult.optionValue("MAILSERVER");
        }
        
        if (parseResult.optionTimes("DEBUG") > 0)
        {
            DEBUG = true;
        }
        
        if (parseResult.optionTimes("PORT") > 0)
        {
            fMailport = 
                (new Integer(parseResult.optionValue("PORT"))).intValue();
        }
        
        return rc;
    }
        
    public STAFResult acceptRequest(STAFServiceInterfaceLevel3.RequestInfo info)
    {               
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("send"))
        {
            return handleSend(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp();
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion();
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown Email Request: " + 
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp()
    {
        return new STAFResult(STAFResult.Ok,
         "EmailService Service Help"          
         + fLineSep + "SEND TO <address> ... " + 
         "< MESSAGE <message> | FILE <file> [MACHINE <machine>] >"
         + fLineSep + "     [SUBJECT <subject>] "
         + fLineSep + "VERSION" + fLineSep + "HELP");
    }
    
    private STAFResult handleVersion()
    {
        return new STAFResult(STAFResult.Ok, kVersion);
    }

    private synchronized STAFResult handleSend(
                                    STAFServiceInterfaceLevel3.RequestInfo info)
    {
        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 3)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 2 required for SEND request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }    

        STAFResult result = new STAFResult(STAFResult.Ok, "");
        String resultString = "";
        STAFCommandParseResult parsedRequest = fSendParser.parse(info.request);
        Vector toList = new Vector();
        String messageOption = "";
        String subjectOption = "";
        STAFResult resolvedValue = null;

        try
        { 
            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            
            int toTimes = parsedRequest.optionTimes("to");
            
            for (int i = 0; i < toTimes; i++)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("to", i + 1), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;
               
                toList.add(resolvedValue.result);
            }
            
            if (parsedRequest.optionTimes("subject") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("subject"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;
                
                subjectOption = resolvedValue.result;
            }
            
            if (parsedRequest.optionTimes("message") == 1)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("message"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;
                    
                messageOption = resolvedValue.result;
            }
            else if (parsedRequest.optionTimes("file") == 1) 
            {
                String fileMachine = info.machine;
                
                if (parsedRequest.optionTimes("machine") == 1)
                {
                    resolvedValue = resolveVar(info.localMachine, 
                        parsedRequest.optionValue("machine"), info.handle);
    
                    if (resolvedValue.rc != 0)
                        return resolvedValue;
                        
                    fileMachine = resolvedValue.result;;
                }
                
                STAFResult res = fHandle.submit2(fileMachine, "fs", 
                    "get file " + parsedRequest.optionValue("file"));
                    
                if (res.rc == 0) 
                {
                    messageOption = res.result;
                } 
                else 
                {
                    messageOption = "Error getting email body from machine " +
                    fileMachine + ": RC=" + res.rc + " Result= " + res.result;
                }
            }
            
            String reply;
            
            try
            {
                try 
                {
                    InetAddress ipaddress = InetAddress.getLocalHost();
                                        
                    if ((fHostname = ipaddress.getHostName()) == null) 
                    {
                        fHostname = "localhost";
                    }
                    else
                    {                        
                        ipaddress = InetAddress.getByName(
                            ipaddress.getHostAddress());
                            
                        fHostname = ipaddress.getHostName();
                    }
                } 
                catch (UnknownHostException uhe) 
                {
                      fHostname = "localhost";
                }
                
                int index = messageOption.indexOf("\\n");
                String tmpStr = "";
                String CR = "";
                
                while (index >= 0) 
                {
                    tmpStr = tmpStr + CR + 
                        messageOption.substring(0, index);
                    CR = "\n";
                    messageOption = messageOption.substring(index+2);
                    index = messageOption.indexOf("\\n");
                }
                
                tmpStr = tmpStr + CR + messageOption;
                messageOption = tmpStr;
               
                for (int j = 0; j < toList.size(); j++)
                {
                    String to = (String)toList.elementAt(j);

                    fSocket = new Socket(fMailserver, fMailport);
                
                    fInStream = new BufferedReader(new 
                        InputStreamReader(fSocket.getInputStream()));
                
                    fOutStream = new PrintStream(fSocket.getOutputStream());
                
                    reply = fInStream.readLine();
                    if (DEBUG) System.out.println(reply);

                    fOutStream.println("HELO " + fHostname);
                    fOutStream.flush();
                
                    reply = fInStream.readLine();
                    if (DEBUG) System.out.println(reply);

                    fOutStream.println("MAIL FROM:  " +
                        "STAFEmailService@" + fHostname);
                    fOutStream.flush();

                    reply = fInStream.readLine();
                    if (DEBUG) System.out.println(reply);
                
                    fOutStream.println("RCPT TO:  " + to);
                    fOutStream.flush();
                 
                    reply = fInStream.readLine();
                    if (DEBUG) System.out.println(reply);
                
                    fOutStream.println("DATA");
                    fOutStream.flush();

                    // Some external SMTP servers require the following line
                    fInStream.readLine();
                    
                    fOutStream.println("To: " + to);
                    fOutStream.flush();
                    
                    fOutStream.println("Subject: " + subjectOption);
                    fOutStream.flush();
                    
                    fOutStream.println("From: STAFEmailService@" + fHostname);
                    fOutStream.flush();

                    String header = 
                        "*****************************************************" + 
                        "*************\n"+
                        "* DO NOT RESPOND TO THE SERVICE MACHINE THAT GENERATED " + 
                        "THIS NOTE *\n"+
                        "*****************************************************" +
                        "*************\n";
    
                    fOutStream.println(header);
                    fOutStream.flush();
    
                    fOutStream.println(messageOption);
                    fOutStream.flush();

                    fOutStream.println();
                    fOutStream.println(".");
                    fOutStream.println();
                    fOutStream.flush();

                    fOutStream.println("QUIT");
                    fOutStream.flush();

                    reply = fInStream.readLine();

                    if (DEBUG) System.out.println(reply);
                    fSocket.close();
                }
            }
            catch ( IOException ioe ) 
            {
                System.err.println("Socket - IO - Error:  " + ioe);
                System.err.println("Mailserver:  " + fMailserver);
                System.err.println("Port:  " + fMailport);
            }
        }
        catch (Exception e)
        {            
            return new STAFResult(STAFResult.JavaError, 
                                  "Internal Java error.");
        }

        return new STAFResult(STAFResult.Ok, resultString);
    }
    
    public int term()
    {
        // Un-register the service handle

        try
        {
            fHandle.unRegister();
        }
        catch (STAFException ex)
        {            
            return STAFResult.STAFRegistrationError;
        }
        
        return STAFResult.Ok;
    }

        // this method will resolve any STAF variables that
        // are contained within the Option Value
    private STAFResult resolveVar(String machine, String optionValue, 
                                  int handle)
    {
        String value = "";
        STAFResult resolvedResult = null;

        if (optionValue.indexOf("{") != -1)
        {
            resolvedResult =
                fHandle.submit2(machine, "var", "handle " + handle + 
                                " resolve " + STAFUtil.wrapData(optionValue));

            if (resolvedResult.rc != 0)
            {
                return resolvedResult;
            }

            value = resolvedResult.result;
        }
        else
        {
            value = optionValue;
        }

        return new STAFResult(STAFResult.Ok, value);
    }
}
