/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.event;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;

/**********************************************************************/
/* STAF Event Service                                                 */
/* By Doug McWhorter and Benjamin Vera-Tudela (C) IBM 2000            */
/* Version 0.90                                                       */
/**********************************************************************/
/* Supplied Methods                                                   */
/* ------------------                                                 */
/* public int init(String name, String parms)                         */
/*                                                                    */
/* public int term()                                                  */
/*                                                                    */
/* public STAFResult acceptRequest(String machine, String process,    */
/*                                     int handle, String request)    */
/*                                                                    */
/* synchronized private void serialize(String theFileName)            */
/*                                                                    */ 
/* synchronized private void deSerialize(String theFileName)          */
/*                                                                    */
/* private STAFResult resolveVar(String machine, String optionValue,  */
/*                                  int handle)                       */
/* private STAFResult resolveAndCheckInt(String machine, String       */  
/*     option, String optionValue, int handle)                        */
/*                                                                    */
/*                                                                    */
/**********************************************************************/

//  Class EventService administrates client requests to the Event Service.
//  Requests include:
//
//  REGISTER a process to be notified of an event of TYPE t, SUBTYPE s.
//  GENERATE an event of TYPE t, SUBTYPE s.
//  ACKNOWLEDGE that an event of TYPE t, SUBTYPE s has notified.
//  UNREGISTER a process that is currently registered for event notification.
//  QUERY information about process that are registered to be notified about 
//      events by TYPE.
//  QUERY information about processes that have generated an event by ENENTID #.
//  LIST information about TYPES and SUBTYPES of events that processes are
//      registered
//  to be notified about.
//  LIST information about events that have been generated by EVENTID #.
//  VERSION: display the installed version of Event Service.
//  HELP

public class EventService implements STAFServiceInterfaceLevel4
{ 
    static final long serialVersionUID = 0;
    transient static final boolean DEBUG = false;
    transient final String kVersion = "1.3.5";

    public static String lineSep;

    // global error constants
    transient static final int kNoAckPending = 4001;
    transient static final int kNoSuchID = 4002;
    transient static final int kNotRegisteredForType = 4003;
    transient static final int kNotRegisteredForSubtype = 4004;
    transient static final int kNoClientsForEvent = 4005;

    // Default service initialization parameters
    private int fMaxAttempts = 1;
    private int fAckTimeout = 60000;
    private int fPriority = 5;
    private int fPriorityDelta = 1;

    transient private STAFHandle fHandle = null;

    transient private String fServiceName = null;

    // stores & processes registration information
    transient RegistrationManager fRegManager = null;

    // stores & processes information about events
    transient GenerationManager fGenManager = null;

    // a STAFCommandParser for each request's options
    transient STAFCommandParser fListParser = null;
    transient STAFCommandParser fQueryParser = null;
    transient STAFCommandParser fRegisterParser = null;
    transient STAFCommandParser fGenerateParser = null;
    transient STAFCommandParser fAckParser = null;
    transient STAFCommandParser fResetParser = null;
    transient STAFCommandParser fUnregisterParser = null;
    transient STAFCommandParser fVersionParser = null;

    // parses any initialization arguments
    transient STAFCommandParser fParmsParser = null;

    // default constructor
    public EventService()
    { /* Do Nothing */ }

    /***************************************************************/
    /* init - Initializes Event Service with STAF.  Instantiates   */
    /* all parsers, fGenManager (deserializes fGenManager.EventID) */
    /* and fRegManager (Deserializes all registration data from    */
    /* file.                                                       */
    /*                                                             */
    /* accepts: STAFServiceInterfaceLevel4 initialization info     */
    /*                                                             */
    /* Returns: STAFResult.Ok or STAFResult.STAFRegistrationError  */
    /***************************************************************/
    public int init(STAFServiceInterfaceLevel4.InitInfo info)
    {
        try
        {
            // instantiate parsers as not case sensitive
            fRegisterParser = new STAFCommandParser(0, false);
            fGenerateParser = new STAFCommandParser(0, false);
            fAckParser = new STAFCommandParser(0, false);
            fResetParser = new STAFCommandParser(0, false);
            fUnregisterParser = new STAFCommandParser(0, false);
            fListParser = new STAFCommandParser(0, false);
            fQueryParser = new STAFCommandParser(0, false);
            fVersionParser = new STAFCommandParser(0, false);
            fParmsParser = new STAFCommandParser(0, false);

            //initialize all parsers to accept request options
            fParmsParser.addOption("MAXATTEMPTS", 1,
                                   STAFCommandParser.VALUEALLOWED);
            fParmsParser.addOption("ACKNOWLEDGETIMEOUT", 1,
                                   STAFCommandParser.VALUEALLOWED);
            fParmsParser.addOption("PRIORITY", 1,
                                   STAFCommandParser.VALUEALLOWED);
            fParmsParser.addOption("PRIORITYDELTA", 1,
                                   STAFCommandParser.VALUEALLOWED);

            fRegisterParser.addOption("REGISTER", 1,
                                      STAFCommandParser.VALUENOTALLOWED);
            fRegisterParser.addOption("TYPE", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fRegisterParser.addOption("SUBTYPE", 0,
                                      STAFCommandParser.VALUEREQUIRED);
            fRegisterParser.addOption("ACKNOWLEDGETIMEOUT", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fRegisterParser.addOption("MAXATTEMPTS", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fRegisterParser.addOption("PRIORITY", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fRegisterParser.addOption("PRIORITYDELTA", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fRegisterParser.addOption("BYHANDLE", 1,
                                      STAFCommandParser.VALUENOTALLOWED);
            fRegisterParser.addOption("BYNAME", 1,
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOptionGroup("BYHANDLE BYNAME", 0, 1);
            
            fRegisterParser.addOptionNeed("REGISTER", "TYPE");
    
            fGenerateParser.addOption("GENERATE", 1,
                                      STAFCommandParser.VALUENOTALLOWED);

            fGenerateParser.addOption("TYPE", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fGenerateParser.addOption("SUBTYPE", 1,
                                      STAFCommandParser.VALUEREQUIRED);
            fGenerateParser.addOption("PROPERTY", 0,
                                      STAFCommandParser.VALUEREQUIRED);

            fGenerateParser.addOptionNeed("GENERATE", "TYPE, SUBTYPE");

            fGenerateParser.addOptionGroup("TYPE SUBTYPE", 2, 2);

            fAckParser.addOption("ACKNOWLEDGE", 1,
                                 STAFCommandParser.VALUENOTALLOWED);

            fAckParser.addOption("EVENTID", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fAckParser.addOption("FORCE", 1,
                                 STAFCommandParser.VALUENOTALLOWED);
            fAckParser.addOption("MACHINE", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fAckParser.addOption("NAME", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fAckParser.addOption("HANDLE", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fAckParser.addOptionGroup("EVENTID", 1, 1);
            
            fAckParser.addOptionNeed("MACHINE", "FORCE");
            fAckParser.addOptionNeed("HANDLE", "FORCE");
            fAckParser.addOptionNeed("NAME", "FORCE");
            fAckParser.addOptionNeed("ACKNOWLEDGE", "EVENTID");
            
            fResetParser.addOption("RESET", 1,
                                 STAFCommandParser.VALUENOTALLOWED);

            fResetParser.addOption("GEN", 1,
                                 STAFCommandParser.VALUENOTALLOWED);
            fResetParser.addOption("REG", 1,
                                 STAFCommandParser.VALUENOTALLOWED);
            fResetParser.addOption("FORCE", 1,
                                 STAFCommandParser.VALUENOTALLOWED);
            fResetParser.addOptionGroup("GEN REG", 1, 1);
            
            fUnregisterParser.addOption("UNREGISTER", 1,
                                        STAFCommandParser.VALUENOTALLOWED);
            fUnregisterParser.addOption("TYPE", 1,
                                        STAFCommandParser.VALUEREQUIRED);
            fUnregisterParser.addOption("SUBTYPE", 0,
                                        STAFCommandParser.VALUEREQUIRED);
            fUnregisterParser.addOption("BYNAME", 1,
                                        STAFCommandParser.VALUENOTALLOWED);
            fUnregisterParser.addOption("FORCE", 1,
                                        STAFCommandParser.VALUENOTALLOWED);
            fUnregisterParser.addOption("MACHINE", 1,
                                        STAFCommandParser.VALUEREQUIRED);
            fUnregisterParser.addOption("NAME", 1,
                                        STAFCommandParser.VALUEREQUIRED);
            fUnregisterParser.addOption("HANDLE", 1,
                                        STAFCommandParser.VALUEREQUIRED);

            fUnregisterParser.addOptionGroup("NAME HANDLE", 0, 1);      
            fUnregisterParser.addOptionGroup("TYPE", 1, 1);
            fUnregisterParser.addOptionNeed("UNREGISTER", "TYPE");

            fUnregisterParser.addOptionNeed("SUBTYPE", "TYPE");
            fUnregisterParser.addOptionNeed("FORCE", "TYPE");
            fUnregisterParser.addOptionNeed("MACHINE", "FORCE");
            fUnregisterParser.addOptionNeed("HANDLE", "FORCE");
            fUnregisterParser.addOptionNeed("NAME", "FORCE");

            fUnregisterParser.addOptionGroup("FORCE", 0, 1);

            fListParser.addOption("LIST", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("TYPE", 1,
                                  STAFCommandParser.VALUEALLOWED);
            fListParser.addOption("EVENTID", 1,
                                  STAFCommandParser.VALUEREQUIRED);
            fListParser.addOption("REGCLIENTS", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("EVENTIDS", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("TYPES", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOptionGroup("EVENTID TYPE REGCLIENTS EVENTIDS TYPES",
                                       1, 1);

            fQueryParser.addOption("QUERY", 1,
                                   STAFCommandParser.VALUENOTALLOWED);

            fQueryParser.addOption("LONG", 1,
                                   STAFCommandParser.VALUENOTALLOWED);

            fQueryParser.addOption("TYPE", 1,
                                   STAFCommandParser.VALUEREQUIRED);
            fQueryParser.addOption("SUBTYPE", 0,
                                   STAFCommandParser.VALUEREQUIRED);
            fQueryParser.addOption("EVENTID", 1,
                                   STAFCommandParser.VALUEREQUIRED);
            fQueryParser.addOptionGroup("EVENTID TYPE", 1, 1);

            fQueryParser.addOptionNeed("SUBTYPE", "TYPE");
            fQueryParser.addOptionNeed("QUERY", "TYPE EVENTID");

            fVersionParser.addOption("VERSION", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
    
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
            fServiceName = info.name;
            
            // Get the line separator

            STAFResult res = fHandle.submit2("local", "var",
                             "resolve {STAF/Config/Sep/Line}");
            if (res.rc != 0)
            {
                System.out.println(res.result);
                return res.rc;
            }
            else
                lineSep = res.result;

            // Verify that the parameters specified for the service are valid

            int rc = handleParms(info);

            if (rc != STAFResult.Ok)
            {
                rc = STAFResult.ServiceConfigurationError;
                return rc;
            }

            // read registration data from file RegManager.out
            deSerialize("RegManager.out");

            fGenManager = new GenerationManager(fHandle, fServiceName,
                                                fRegManager);

            // read EventID from file GenManager.out
            deSerialize("GenManager.out");
            
            // Register Help Data

            registerHelpData(
                kNoAckPending,
                "No Acknowledgement Pending", 
                "Already received an acknowledgement for this event from " +
                "the acknowledging\nprocess.");

            registerHelpData(
                kNoSuchID,
                "No such event ID",
                "An acknowledgement was received but no event with the " +
                "submitted Event ID has\nbeen generated.");

            registerHelpData(
                kNotRegisteredForType,
                "Not registered for Type",
                "A process tried to unregister for an event Type that it " +
                "was not currently\nregistered for.");

            registerHelpData(
                kNotRegisteredForSubtype,
                "Not registered for Subtype",
                "A process tried to unregister for an event Subtype that " +
                "it was not currently\nregistered for.");

            registerHelpData(
                kNoClientsForEvent,
                "No clients registered",
                "No clients were registered for the event.");

            return STAFResult.Ok; 
        }
        catch(STAFException e)
        {
            if (DEBUG) e.printStackTrace();
            return STAFResult.STAFRegistrationError;
        }
    }


    private int handleParms(STAFServiceInterfaceLevel4.InitInfo info)
        throws STAFException
    {
        STAFCommandParseResult parsedResult= fParmsParser.parse(info.parms);

        String errmsg = "ERROR:  Service Configuration Error for Service " +
                        fServiceName + lineSep +
                        "EventService::handleParms() - ";

        if (parsedResult.rc != STAFResult.Ok)
        {
            System.out.println(
                errmsg + "PARMS parsing failed with RC=" +
                STAFResult.InvalidRequestString + 
                " Result=" + parsedResult.errorBuffer);

            return STAFResult.ServiceConfigurationError;
        }
        
        STAFResult resolvedResult;

        if (parsedResult.optionTimes("MAXATTEMPTS") > 0)
        {
            resolvedResult = resolveAndCheckInt(
                "local", "MAXATTEMPTS",
                parsedResult.optionValue("MAXATTEMPTS"), 0);
            
            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "RC=" + resolvedResult.rc +
                    " " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
            
            fMaxAttempts = Integer.parseInt(resolvedResult.result);
        }
        
        if (parsedResult.optionTimes("ACKNOWLEDGETIMEOUT") > 0)
        {
            resolvedResult = resolveAndCheckInt(
                "local", "ACKKNOWLEDGETIMEOUT",
                parsedResult.optionValue("ACKNOWLEDGETIMEOUT"), 0);

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "RC=" + resolvedResult.rc +
                    " " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            fAckTimeout = Integer.parseInt(resolvedResult.result);
        }
        
        // PRIORITY is an undocumented service parameter

        if (parsedResult.optionTimes("PRIORITY") > 0)
        {
            resolvedResult = resolveAndCheckInt(
                "local", "PRIORITY",
                parsedResult.optionValue("PRIORITY"), 0);

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "RC=" + resolvedResult.rc +
                    " " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            fPriority = Integer.parseInt(resolvedResult.result);
        }
        
        // PRIORITYDELTA is an undocumented service parameter

        if (parsedResult.optionTimes("PRIORITYDELTA") > 0)
        {
            resolvedResult = resolveAndCheckInt(
                "local", "PRIORITYDELTA",
                parsedResult.optionValue("PRIORITYDELTA"), 0);

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "RC=" + resolvedResult.rc +
                    " " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            fPriorityDelta = Integer.parseInt(resolvedResult.result);
        }
        
        return 0;
    }

    public int term()
    {
        // Un-register Help Data

        unregisterHelpData(kNoAckPending);
        unregisterHelpData(kNoSuchID);
        unregisterHelpData(kNotRegisteredForType);
        unregisterHelpData(kNotRegisteredForSubtype);
        unregisterHelpData(kNoClientsForEvent);

        // Un-register the service handle

        try
        {
            fHandle.unRegister();
        }
        catch (STAFException ex)
        {
            return STAFResult.STAFRegistrationError;
        }

        return STAFResult.Ok;
    }

    /***************************************************************************/
    /* acceptRequest- Calls appropriate methods to process a request from a    */
    /*                client process.                                          */
    /*                                                                         */
    /* accepts: STAFServiceInterfaceLevel4 request information                 */
    /*                                                                         */
    /* returns: STAFResult.rc = STAFResult.Ok, if successful; STAFResult.      */
    /*          InvalidRequestString, if unsuccessful;                         */
    /*                                                                         */
    /*          STAFResult.result contains the result returned by the called   */
    /*          method, if successful;                                         */
    /*          STAFResult.result contains the command, if unsuccessful        */
    /***************************************************************************/
    public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        String resolveString;

        // get the command and its options from the request string
        /*StringTokenizer stok = new StringTokenizer(request);
        String command = (stok.hasMoreTokens()) ?
                         stok.nextToken().toLowerCase() : "";
        String args = (stok.hasMoreTokens()) ?
               (request.substring(request.indexOf(stok.nextToken()))) : "";*/
               
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("register"))
            return handleRegister(info);
        else if (lowerRequest.startsWith("generate"))   
            return handleGenerate(info); 
        else if (lowerRequest.startsWith("acknowledge"))
            return handleAck(info);
        else if (lowerRequest.startsWith("reset"))
            return handleReset(info);
        else if (lowerRequest.startsWith("unregister"))
            return handleUnRegister(info);
        else if (lowerRequest.startsWith("query"))
            return handleQuery(info);
        else if (lowerRequest.startsWith("list"))
            return handleList(info);
        else if (lowerRequest.startsWith("help"))
            return handleHelp();
        else if (lowerRequest.startsWith("version"))
            return new STAFResult(STAFResult.Ok, kVersion);
        else 
            return new STAFResult(STAFResult.InvalidRequestString,
                                  info.request);
    }

    /***********************************************************************/
    /* serialize- object serializer that writes structured data to a file  */
    /* using the ObjectOutputStream.writeObject method().                  */
    /*                                                                     */
    /* accepts: the file name.                                             */
    /* returns: void.                                                      */
    /***********************************************************************/

    synchronized private void serialize(String theFileName)
    {
        try
        {
            if (theFileName == null) theFileName = "RegManager.out";

            if (theFileName.equals("RegManager.out"))
            {
                ObjectOutputStream out = new ObjectOutputStream(
                                         new FileOutputStream(theFileName));
                out.writeObject(fRegManager);
                out.close(); 
            }
            else fGenManager.serialize();
        } 
        catch(Exception e)
        {
           if (DEBUG) e.printStackTrace();
        }
    } 

    /***********************************************************************/
    /* deSerialize- object deSerializer that reads fGenManager.fEventID    */
    /* from a file using the ObjectInputStream.readObject method().        */
    /*                                                                     */
    /* accepts: the file name.                                             */
    /* returns: void.                                                      */
    /***********************************************************************/

    synchronized private void deSerialize(String theFileName)
    {
        try
        {
            ObjectInputStream in = new ObjectInputStream(
                                   new FileInputStream(theFileName));

            if (theFileName == null) theFileName = "RegManager.out";  

            if (theFileName.equals("RegManager.out"))
                fRegManager = (RegistrationManager)in.readObject(); 
            else
                fGenManager.deSerialize(); 
        }
        catch(Exception e)
        {
            if (theFileName.equals("RegManager.out"))
                fRegManager = new RegistrationManager();

            if (DEBUG) e.printStackTrace(); 
        }
    } 


    // Register error codes for the Event Service with the HELP service

    private void registerHelpData(int errorNumber, String info,
                                 String description)
    {
        STAFResult res = fHandle.submit2("local", "HELP",
                         "REGISTER SERVICE " + fServiceName +
                         " ERROR " + errorNumber +
                         " INFO " + STAFUtil.wrapData(info) +
                         " DESCRIPTION " + STAFUtil.wrapData(description));
    }

    // Un-register error codes for the Event Service with the HELP service

    private void unregisterHelpData(int errorNumber)
    {
        STAFResult res = fHandle.submit2("local", "HELP",
                         "UNREGISTER SERVICE " + fServiceName +
                         " ERROR " + errorNumber);
    }

    /***********************************************************************/
    /* resolveValue - Uses the var Service to resolve an option value.     */
    /*                                                                     */   
    /*                                                                     */
    /*                                                                     */
    /* accepts: the submitting machine name, an unresolved option value,   */
    /* the submitting process' handle                                      */
    /*                                                                     */   
    /* returns: STAFResult.rc = STAFResult.Ok;                             */
    /* STAFResult.result = the resolved option value                       */
    /***********************************************************************/

    private STAFResult resolveVar(String machine, String optionValue, int handle)
    {
        String value = "";
        STAFResult resolvedResult = null;

        if (optionValue.indexOf("{") != -1)
        {
            if (handle > 0)
            {
                resolvedResult = fHandle.submit2(
                    machine, "var", "handle " + handle + " resolve " + 
                    STAFUtil.wrapData(optionValue));
            }
            else
            {
                resolvedResult = fHandle.submit2(
                    machine, "var", "resolve " +
                    STAFUtil.wrapData(optionValue));
            }

            if (resolvedResult.rc != 0) return resolvedResult;
    
            value = resolvedResult.result;
        }
        else value = optionValue;
        
        return new STAFResult(STAFResult.Ok, value);
    }

    /***************************************************************************/
    /* resolveAndCheckInt- Uses the var Service to resolve an option value.    */
    /*                     Then checks if the resolved value is an integer.    */
    /*                                                                         */
    /* accepts: the submittind machine name, a command option name, a command  */
    /*          option value (unresolved), and the submiting process' handle   */
    /*                                                                         */
    /* returns: if the option value is a sequence of digits, STAFResult.rc     */
    /*          = STAFResult.Ok;STAFResult.result = the option value;          */
    /*          if the option value is not a sequence of digits,               */
    /*          STAFResult.result is an error message                          */
    /***************************************************************************/

    private STAFResult resolveAndCheckInt(String machine, String option,
                                          String optionValue,
                                          int handle)
    {
        STAFResult resolvedValue = resolveVar(machine, optionValue, handle);
        
        if (resolvedValue.rc != 0) return resolvedValue;
       
        try
        {
            Integer.parseInt(resolvedValue.result);
        }
        catch (NumberFormatException e)
        {
            return new STAFResult(STAFResult.InvalidValue,
                                  option + " value must be numeric.  " +
                                  option + "=" + resolvedValue.result);
        }

        return new STAFResult(STAFResult.Ok, resolvedValue.result);
    }

    /***************************************************************************/
    /* handleRegister - calls fRegManager.registerClient() to register a       */
    /*                  process for future notification when events occur.     */
    /*                                                                         */
    /* accepts: STAFServiceInterfaceLevel4 request information                 */
    /*          (including the submitting machine name, the submitting process,*/
    /*          the submitting process handle, a string containing any command */
    /*          options & option values).                                      */
    /*                                                                         */
    /* returns: if successful, STAFResult.rc = STAFResult.Ok (0);              */
    /*                         STAFResult.result is empty;                     */
    /*          if unsuccessful, STAFResult.rc   = error code > 0;             */
    /*                           STAFResult.result contains message            */
    /*                           correspond to error code in STAFResult.rc     */
    /***************************************************************************/
                                                                               
    private STAFResult handleRegister(
        STAFServiceInterfaceLevel4.RequestInfo info)
    {
        String type = "";
        String subtype = "";

        // Set to default values for the service
        int maxAttempts = fMaxAttempts;
        int timeout = fAckTimeout;
        int priority = fPriority;
        int priorityDelta = fPriorityDelta;

        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest =
            fRegisterParser.parse(info.request);

        try
        {
            if (info.trustLevel < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 3 required for REGISTER request. " +
                       "Requesting machine's trust level: " + info.trustLevel);
            }                                                               
            
            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            resolvedValue = resolveVar(info.machine,
                                       parsedRequest.optionValue("type"),
                                       info.handle);
         
            if (resolvedValue.rc != 0) return resolvedValue;
            
            type = resolvedValue.result;    

            if (parsedRequest.optionTimes("maxAttempts") > 0)
            {
                resolvedResult = resolveAndCheckInt(
                    info.machine, "maxAttempts",
                    parsedRequest.optionValue("maxAttempts"), info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                maxAttempts = Integer.parseInt(resolvedResult.result);
            }

            if (parsedRequest.optionTimes("acknowledgeTimeout") > 0)
            {
                resolvedResult = resolveAndCheckInt(
                    info.machine, "acknowledgeTimeout",
                    parsedRequest.optionValue("acknowledgeTimeout"),
                    info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                timeout = Integer.parseInt(resolvedResult.result);
            }

            if (parsedRequest.optionTimes("priority") > 0)
            {
                resolvedResult = resolveAndCheckInt(
                    info.machine, "priority",
                    parsedRequest.optionValue("priority"), info.handle);
                                             
                if (resolvedResult.rc != 0) return resolvedResult;    

                priority = Integer.parseInt(resolvedResult.result);
            }  

            if (parsedRequest.optionTimes("priorityDelta") > 0)
            {
                resolvedResult = resolveAndCheckInt(
                    info.machine, "priorityDelta",
                    parsedRequest.optionValue("priorityDelta"), info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;    

                priorityDelta = Integer.parseInt(resolvedResult.result);
            }

            int subTimes = parsedRequest.optionTimes("subtype");
            String[] subTypes = new String[subTimes];

            for (int count = 1; count <= subTimes; count++)
            {
                resolvedResult = resolveVar(
                    info.machine, parsedRequest.optionValue("subtype", count),
                    info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                subTypes[count - 1] = resolvedResult.result;
            }

            if ((parsedRequest.optionTimes("byHandle") == 0) &&
                (sResult.rc == 0))
            {
                fRegManager.registerClient(
                    new Client(info.machine, info.processName, 
                               maxAttempts, timeout, priority, 
                               priorityDelta), type, subTypes);
            }
            else if (sResult.rc == 0)
            {
                fRegManager.registerClient(
                    new Client(info.machine, info.handle,
                               maxAttempts, timeout, priority,
                               priorityDelta), type, subTypes);
            }

            serialize("RegManager.out");
        }
        catch (Exception e)
        {
            if (DEBUG) e.printStackTrace();

            return new STAFResult(STAFResult.JavaError, "Internal Java error."); 
        }

        return sResult;
    } 

    /***************************************************************************/
    /* handleQuery- If option TYPE, or TYPE and SUBTYPE, are passed in the arg */
    /*              string, lists, by TYPE, or by TYPE and SUBTYPE, the        */
    /*              following data about registered processes:                 */
    /*              machine name, process name, process handle, maxAttempts,   */
    /*              timeout, priority, priorityDelta.                          */
    /*                                                                         */
    /*              If option EVENTID is passed in the arg string,             */
    /*              lists, by EVENTID, the above data about registered         */
    /*              processes that have not yet acknowledged a generated event.*/
    /*              Also lists the following data about the process that       */
    /*              generated the event:                                       */
    /*                                                                         */
    /*              eventID#, generating machine name, generating process name,*/
    /*              generating process handle, generating, type of event,      */
    /*              subtype of event                                           */
    /*                                                                         */
    /* accepts: the submitting machine name, the submitting process, the sub-  */
    /*          mitting process handle, a string containing any command options*/
    /*          & option values.                                               */

    /*                                                                         */
    /* returns: if successful, STAFResult.rc = STAFResult.Ok (0);              */
    /*                         STAFResult.result contains the data described   */
    /*                         above.                                          */
    /*          if unsuccessful, STAFResult.rc   = error code > 0;             */
    /*                           STAFResult.result contains a message          */
    /*                           correspond to error code in STAFResult.rc     */
    /***************************************************************************/

    private STAFResult handleQuery(STAFServiceInterfaceLevel4.RequestInfo info) 
    {
        String type = null;
        String subtype = null;
        String eid = null;
        int id = 0;
        StringBuffer result = new StringBuffer();
        STAFCommandParseResult parsedRequest = fQueryParser.parse(info.request);
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;

        try
        {
            if (info.trustLevel < 2)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 2 required for QUERY request. Requesting " +
                       "machine's trust level: " + info.trustLevel);   
            }

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            if (info.diagEnabled == 1)
            {
                // Record diagnostics because the output for a QUERY request will
                // change in STAF V3.0

                String trigger = new String("STAF/V3.0-Mig/EVENT QUERY");
                String source  = new String(info.processName + ";" +
                                            info.machine + ";" +
                                            info.handle);
                fHandle.submit2(
                    STAFHandle.ReqFireAndForget,
                    "local", "DIAG",
                    "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                    " SOURCE " + STAFUtil.wrapData(source));
            }

            if (parsedRequest.optionTimes("type") > 0)
            {
                resolvedResult = resolveVar(info.machine,
                                 parsedRequest.optionValue("type"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult; 

                type = resolvedResult.result;

                if (parsedRequest.optionTimes("subtype") > 0)
                { 
                    resolvedResult = resolveVar(info.machine,
                                     parsedRequest.optionValue("subtype"),
                                     info.handle);

                    if (resolvedResult.rc != 0) return resolvedResult;

                    subtype = resolvedResult.result;
                }

                Vector crf = new Vector();
 
                if (subtype == null) 
                    crf = fRegManager.allClientsFor(type); 
                else
                    crf = fRegManager.clientsRegisteredFor(type, subtype);

                if ((crf == null) || (crf.size() == 0))
                {
                    return new STAFResult(subtype == null ?
                           kNotRegisteredForType : kNotRegisteredForSubtype,
                           "No processes are registered to be notified about " +
                           "events of type " + type + (subtype == null ? "." :
                           ", subtype " + subtype + "."));
                }

                if ((crf != null) && (parsedRequest.optionTimes("long") > 0))
                {                    
                    for (Enumeration e = crf.elements(); e.hasMoreElements();)
                    {
                        result.append(
                            ((Client)(e.nextElement())).longToString() +
                            "\n\n");
                    }
                }
                else if (crf != null)
                {                    
                    for (Enumeration e = crf.elements(); e.hasMoreElements();)
                    {
                        result.append(
                            ((Client)(e.nextElement())).toString() + "\n");
                    }
                }
            }
            else if (parsedRequest.optionTimes("eventID") > 0)
            {
                resolvedResult = resolveAndCheckInt(info.machine, "eventID",
                                 parsedRequest.optionValue("eventID"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                id = Integer.parseInt(resolvedResult.result);

                Vector clientsForEvent = fGenManager.getClientsForEvent(id);

                if (clientsForEvent != null && (clientsForEvent.size() > 0) &&
                    parsedRequest.optionTimes("long") > 0)
                {
                    for (Enumeration cfe = clientsForEvent.elements();
                         cfe.hasMoreElements();)
                    {
                        result.append((((GenerationManager.Notification)
                                        cfe.nextElement())).longToString() +
                                        "\n\n" );
                    }
                }
                else if (clientsForEvent != null && (clientsForEvent.size() > 0))
                {
                    for (Enumeration cfe = clientsForEvent.elements();
                         cfe.hasMoreElements();)
                    {
                             result.append(((cfe.nextElement())).toString() + 
                                           "\n");
                    }
                }
                else
                {
                    return new STAFResult(kNoClientsForEvent,
                           "No processes are waiting to be notified about " +
                           "eventID " + id + ".");
                }
            }
        }
        catch(Exception e)
        {
            if (DEBUG)e.printStackTrace();

            return new STAFResult(STAFResult.JavaError, "Internal Java error."); 
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }

    /***************************************************************************/
    /* handleList- If option TYPES is in the arg string,                       */
    /* lists all event types for which at least one process is currently reg-  */
    /* istered. Alternatively, lists a single type and all subtypes of that    */
    /* type if option TYPE is in the arg string.                               */
    /*                                                                         */
    /*                                                                         */
    /* If option EVENTID is passed in the arg string,                          */
    /*              Lists, by EVENTID, the following data about registered     */
    /*              processes that have not yet acknowledged a generated event:*/
    /*              machine name, process name, process handle, maxAttempts,   */
    /*              timeout, priority, priorityDelta.                          */
    /*              Also lists the following data about the process that       */
    /*              generated the event:                                       */
    /*                                                                         */
    /*              eventID#, generating machine name, generating process name,*/
    /*              generating process handle, generating, type of event,      */
    /*              subtype of event                                           */
    /*                                                                         */
    /* accepts: STAFServiceInterfaceLevel4 request information                 */
    /*          (including the submitting machine name, the submitting process,*/
    /*          the submitting process handle, a string containing any command */
    /*          options & option values).                                      */
    /*                                                                         */
    /* returns: if successful, STAFResult.rc = STAFResult.Ok (0);              */
    /*                         STAFResult.result contains the data described   */
    /*                         above.                                          */
    /*          if unsuccessful, STAFResult.rc   = error code > 0;             */
    /*                           STAFResult.result contains a message          */
    /*                           correspond to error code in STAFResult.rc     */
    /***************************************************************************/

    private STAFResult handleList(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        String type = null;
        String subtype = null;
        String eid = null;
        int id = 0;
        StringBuffer result = new StringBuffer();
        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;

        try
        {
            if (info.trustLevel < 2)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 2 required for LIST request. Requesting " +
                       "machine's trust level: " + info.trustLevel);
            }

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            if (info.diagEnabled == 1)
            {
                // Record diagnostics because the output for a LIST request will
                // change in STAF V3.0

                String trigger = new String("STAF/V3.0-Mig/EVENT LIST");
                String source  = new String(info.processName + ";" +
                                            info.machine + ";" +
                                            info.handle);
                fHandle.submit2(
                    STAFHandle.ReqFireAndForget,
                    "local", "DIAG",
                    "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                    " SOURCE " + STAFUtil.wrapData(source));
            }

            if (parsedRequest.optionTimes("REGCLIENTS") > 0)
            {
                for (Enumeration e = fRegManager.fRegistrationTable.keys();
                     e.hasMoreElements();)
                {
                    type = (String)e.nextElement();
                    RegistrationManager.Type typeObj = (RegistrationManager.Type)
                        (fRegManager.fRegistrationTable.get(type));
                    Vector typeClients = fRegManager.clientsRegisteredFor(type);
                                   
                    if (typeClients != null)
                    {
                        for (Enumeration vectorEnum = typeClients.elements();
                            vectorEnum.hasMoreElements();)
                        {
                            result.append(type + ";" + "all subtypes" + ";" +
                                ((Client)vectorEnum.nextElement()).toString() +
                                "\n");
                        }
                    }

                    for (Enumeration subEnum = typeObj.subtypeTable.keys();
                         subEnum.hasMoreElements();)
                    {
                        subtype = (String)subEnum.nextElement();
                        Vector subtypeClients =
                               fRegManager.clientsRegisteredFor(type, subtype);

                        if (subtypeClients != null)
                        {
                            for (Enumeration vectorEnum =
                                             subtypeClients.elements();
                                 vectorEnum.hasMoreElements();)
                            {
                                result.append(type + ";" + subtype + ";" +
                                    ((Client)vectorEnum.nextElement()).
                                    toString() + "\n");
                            }
                        }
                    }     
                }
            }

            if  ((parsedRequest.optionTimes("type") == 0) &&
                 (parsedRequest.optionTimes("TYPES") > 0)) 
            {
                result.append(fRegManager.typesToString("all", false,
                                                        false));
            }
            else if ((parsedRequest.optionTimes("type") > 0) &&
                     (parsedRequest.optionValue("type").equals("")))
            {
                result.append(fRegManager.typesToString("all", true, false));
            }
            else if ((parsedRequest.optionTimes("type") > 0) &&
                     (!parsedRequest.optionValue("type").equals("")))
            {
                resolvedResult = resolveVar(info.machine, parsedRequest.
                                            optionValue("type"), info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;
                 result.append(fRegManager.typesToString(
                    resolvedResult.result, true, true));
            }                           
            else if (parsedRequest.optionTimes("EVENTIDS") > 0)  
            {         
                if (!fGenManager.fEventTable.keys().hasMoreElements())
                {
                    return new STAFResult(kNoClientsForEvent,
                           "No events are currently waiting to be " +
                           "acknowledged");
                }

                for (Enumeration e = fGenManager.fEventTable.keys(); 
                     e.hasMoreElements();) 
                {
                    id = ((Integer)e.nextElement()).intValue();

                    Vector clientsForEvent =
                           fGenManager.getClientsForEvent(id);
                           
                    if (clientsForEvent.size() > 0)
                    {
                        for (Enumeration cfe = clientsForEvent.elements();
                             cfe.hasMoreElements();)
                        {
                            result.append(id + ";" +
                                          cfe.nextElement().toString() + "\n");
                        }
                    }
                }

                if (result.toString().equals(""))
                {
                    return new STAFResult(kNoClientsForEvent,
                           "No events are currently waiting to be " +
                           "acknowledged.");
                }
            }
            else if (parsedRequest.optionTimes("eventID") > 0)
            {
                resolvedResult = resolveAndCheckInt(info.machine, "eventID",
                                 parsedRequest.optionValue("eventID"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;    

                id = Integer.parseInt(resolvedResult.result);

                Vector clientsForEvent = fGenManager.getClientsForEvent(id);

                if (clientsForEvent != null && (clientsForEvent.size() > 0))
                {
                    for (Enumeration cfe = clientsForEvent.elements();
                         cfe.hasMoreElements();)
                    {
                         result.append(((cfe.nextElement())).toString() + 
                            "\n");
                    }
                }
                else
                {
                    return new STAFResult(kNoClientsForEvent,
                           "No processes are waiting to be notified about " +
                           "eventID " + id + ".");
                }
            }                 
            
        }
        catch(Exception e)
        {
            if (DEBUG) e.printStackTrace();

            return new STAFResult(STAFResult.JavaError, "Internal Java error.");
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }

    /***************************************************************************/
    /* handleGenerate - calls fGenManager.generateEvent() to notify all        */
    /*                  processes registered for events of the type and        */
    /*                  subtype contained in the "args" string. Includes a     */
    /*                  list of properties in the message sent to registered   */
    /*                  clients.                                               */
    /*                                                                         */
    /* accepts: STAFServiceInterfaceLevel4 request information                 */
    /*          (including the submitting machine name, the submitting process,*/
    /*          the submitting process handle, a string containing any command */
    /*          options & option values).                                      */
    /*                                                                         */
    /* returns: if successful, STAFResult.rc = STAFResult.Ok (0);              */
    /*                         STAFResult.result contains the data described   */
    /*                         above.                                          */
    /*          if unsuccessful, STAFResult.rc   = error code > 0;             */
    /*                           STAFResult.result contains a message          */
    /*                           correspond to error code in STAFResult.rc     */
    /***************************************************************************/

    private STAFResult handleGenerate(
        STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFCommandParseResult parsedRequest =
            fGenerateParser.parse(info.request);
        String type = null;     //parsedRequest.optionValue("type");
        String subtype = null;  //parsedRequest.optionValue("subtype");
        Vector v = new Vector();
        int propertyTimes = parsedRequest.optionTimes("property");
        int eventID;
        STAFResult resolvedResult = null;

        try
        {
            if (info.trustLevel < 4)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 4 required for GENERATE request. " +
                       "Requesting machine's trust level: " + info.trustLevel);
            }

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            if (parsedRequest.optionTimes("type") > 0)
            {
                resolvedResult = resolveVar(info.machine,
                                 parsedRequest.optionValue("type"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                type = resolvedResult.result;

                if (parsedRequest.optionTimes("subtype") > 0)
                {
                    resolvedResult = resolveVar(info.machine,
                                     parsedRequest.optionValue("subtype"),
                                     info.handle);

                    if (resolvedResult.rc != 0) return resolvedResult; 

                    subtype = resolvedResult.result;
                }
            }

            // Do NOT resolve PROPERTY option values

            for (int count = 1; count <= propertyTimes; count++)
            {
                String property = parsedRequest.optionValue("property", count);

                if (property.indexOf('=') == 0)
                {
                    return new STAFResult(
                        STAFResult.InvalidValue,
                        "Invalid value for PROPERTY " + property +
                        "\nThe format for a PROPERTY option value must be:" +
                        " <Name>[=<Value>]");
                }

                v.addElement(property);
            }

            String[] properties = new String[v.size()];
            int index = 0;

            for (Enumeration e = v.elements(); e.hasMoreElements();)
                properties[index++] = (String)e.nextElement();
 
            eventID = fGenManager.generateEvent(
                info.machine, info.processName, info.handle, type, subtype,
                properties, fRegManager.clientsFor(type, subtype));
                                                
            if (eventID == kNoClientsForEvent)
            {
                return new STAFResult(kNoClientsForEvent,
                       "No processes are registered to be notified about " +
                       "events of type " + type + ", subtype " + subtype + ".");
            }
        }
        catch(Exception e)
        { 
            if (DEBUG) e.printStackTrace();

            return new STAFResult(STAFResult.JavaError,"Internal Java error." );
        }

        serialize("GenManager.out");

        return new STAFResult(STAFResult.Ok, "" + eventID);
    }

    /***************************************************************************/
    /* handleUnRegister - calls fRegManager.unRegisterClient() to unregister   */
    /*                    the submitting process. The submitting process may   */
    /*                    also FORCE another process (including a process on   */
    /*                    another machine: Trust = 4 required) to unregister.  */
    /*                                                                         */
    /* accepts: STAFServiceInterfaceLevel4 request information                 */
    /*          (including the submitting machine name, the submitting process,*/
    /*          the submitting process handle, a string containing any command */
    /*          options & option values).                                      */
    /*                                                                         */
    /* returns: if successful, STAFResult.rc = STAFResult.Ok (0);              */
    /*                         STAFResult.result contains the data described   */
    /*                         above.                                          */
    /*          if unsuccessful, STAFResult.rc   = error code > 0;             */
    /*                           STAFResult.result contains a message          */
    /*                           correspond to error code in STAFResult.rc     */
    /***************************************************************************/

    private STAFResult handleUnRegister(
        STAFServiceInterfaceLevel4.RequestInfo info)
    {
        int fRC = 0;
        int subTimes = 0;
        String type = null;
        String subtype = null;
        String machine = info.machine;
        int handle = info.handle;
        String processName = info.processName;
        STAFCommandParseResult parsedRequest = null;
        STAFResult resolvedResult = null;

        try
        {
            if (info.trustLevel < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 3 required for UNREGISTER request. " +
                       "Requesting machine's trust level: " + info.trustLevel);
            }

            parsedRequest = fUnregisterParser.parse(info.request);

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            if (parsedRequest.optionTimes("FORCE") > 0)
            {
                if (info.trustLevel < 4)
                {
                    return new STAFResult(STAFResult.AccessDenied,
                           "Trust level 4 required for UNREGISTER FORCE " +
                           "request. Requesting machine's trust level: " +
                           info.trustLevel);
                }
            }

            if (parsedRequest.optionTimes("MACHINE") > 0)
            {
                resolvedResult = resolveVar(
                    info.machine, parsedRequest.optionValue("machine"),
                    info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;
                
                if (resolvedResult.result.toLowerCase().equals("local"))
                {
                    resolvedResult = resolveVar(
                        info.machine, "{STAF/Config/Machine}", info.handle);
                }

                if (resolvedResult.rc != 0) return resolvedResult;

                machine = resolvedResult.result;
            }

            if (parsedRequest.optionTimes("NAME") > 0)
            {
                resolvedResult = resolveVar(info.machine,
                                 parsedRequest.optionValue("name"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                processName = resolvedResult.result;

                handle = 0;
            }
            else if (parsedRequest.optionTimes("HANDLE") > 0)
            {
                resolvedResult = resolveAndCheckInt(info.machine, "handle",
                                 parsedRequest.optionValue("handle"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                handle = Integer.parseInt(resolvedResult.result);
            }

            if (parsedRequest.optionTimes("type") > 0)
            {
                resolvedResult = resolveVar(info.machine,
                                 parsedRequest.optionValue("type"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                type = resolvedResult.result;
            }

            subTimes = parsedRequest.optionTimes("subtype");

            String[] subTypes = new String[subTimes];

            for (int count = 1; count <= subTimes; count++)
            {
                resolvedResult = resolveVar(info.machine,
                                 parsedRequest.optionValue("subtype", count),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                subTypes[count - 1] = resolvedResult.result;
            }

            fRC = fRegManager.unRegisterClient(machine, processName,
                                               handle, type, subTypes);
            if (fRC != STAFResult.Ok)
                return new STAFResult(fRC, info.request); 
  
            serialize("RegManager.out");
        }
        catch (Exception e)
        {
            if (DEBUG) e.printStackTrace();

            return new STAFResult(STAFResult.InvalidRequestString,
                                  info.request);
        }

        return new STAFResult(STAFResult.Ok, "");
    }

    private STAFResult handleAck(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        String machine = info.machine;
        String processName = info.processName;
        int handle = info.handle;
        int eventID = 0;
        int rc;
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = new STAFResult(STAFResult.Ok, "");

        try
        {
            if (info.trustLevel < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 3 required for ACKNOWLEDGE request. " +
                       "Requesting machine's trust level: " + info.trustLevel);
            }

            STAFCommandParseResult parsedRequest =
                fAckParser.parse(info.request);

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            resolvedResult = resolveAndCheckInt(info.machine, "eventID",
                             parsedRequest.optionValue("eventID"),
                             info.handle);
                                         
            if (resolvedResult.rc != 0) return resolvedResult;

            eventID = Integer.parseInt(resolvedResult.result);

            if (parsedRequest.optionTimes("FORCE") > 0)
            {
                if (info.trustLevel < 4)
                {
                    return new STAFResult(STAFResult.AccessDenied,
                           "Trust level 4 required for ACKNOWLEDGE FORCE " +
                           "request. Requesting machine's trust level: " +
                           info.trustLevel);
                }
            }

            if (parsedRequest.optionTimes("MACHINE") > 0)
            {
                resolvedResult = resolveVar(
                    info.machine, parsedRequest.optionValue("machine"),
                    info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;
                
                if (resolvedResult.result.toLowerCase().equals("local"))
                {
                    resolvedResult = resolveVar(
                        info.machine, "{STAF/Config/Machine}", info.handle);
                }

                if (resolvedResult.rc != 0) return resolvedResult;

                machine = resolvedResult.result;
            }
            
            if (parsedRequest.optionTimes("NAME") > 0)
            {
                resolvedResult = resolveVar(info.machine,
                                 parsedRequest.optionValue("name"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;

                processName = resolvedResult.result;
            }
            else if (parsedRequest.optionTimes("HANDLE") > 0)
            {
                resolvedResult = resolveAndCheckInt(info.machine, "handle",
                                 parsedRequest.optionValue("handle"),
                                 info.handle);

                if (resolvedResult.rc != 0) return resolvedResult;
                
                handle = Integer.parseInt(resolvedResult.result);
            }
  
            rc = fGenManager.ackEvent(eventID, machine, processName, handle);

            if (rc == STAFResult.Ok)
                return new STAFResult (STAFResult.Ok, "" + eventID); 
        }
        catch(Exception e)
        {
            if (DEBUG) e.printStackTrace();
        
            return new STAFResult(STAFResult.JavaError, "Internal Java error." );
        }

        if (rc == kNoAckPending)
            return new STAFResult (kNoAckPending, "No ack pending");
        else if (rc == kNoSuchID)
            return new STAFResult (kNoSuchID, "No such event ID");
        else
            return new STAFResult(STAFResult.JavaError, "Internal Java error.");
    }
    
    private STAFResult handleReset(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        int eventID;
        int rc;
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = new STAFResult(STAFResult.Ok, "");

        try
        {
            if (info.trustLevel < 4)
            {
                return new STAFResult(STAFResult.AccessDenied,
                       "Trust level 4 required for RESET request. " +
                       "Requesting machine's trust level: " + info.trustLevel);
            }

            STAFCommandParseResult parsedRequest =
                fResetParser.parse(info.request);

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }
           
            if (parsedRequest.optionTimes("FORCE") == 0)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      "FORCE required");
            }
            else if (parsedRequest.optionTimes("REG") > 0)
            {
                fRegManager.handleReset();
                serialize("RegManager.out");
                
            }
            else if (parsedRequest.optionTimes("GEN") > 0)
            {
                fGenManager.handleReset();
                serialize("GenManager.out");
            }

            return new STAFResult (STAFResult.Ok, "");
        }
        catch(Exception e)
        {
            if (DEBUG) e.printStackTrace();
        
            return new STAFResult(STAFResult.JavaError, "Internal Java error." );
        }
    }
   
    private STAFResult handleHelp()
    {
        return new STAFResult(STAFResult.Ok,
        "EVENT Service Help \n\n" +
        "REGISTER          TYPE <Event Type> [SUBTYPE <Event SubType>]...\n" +
        "                  [ACKNOWLEDGETIMEOUT <Timeout>] [MAXATTEMPTS <Number>]\n" +
        "                  PRIORITY <Number>] [PRIORITYDELTA <Number>]\n" +
        "                  [BYNAME | BYHANDLE]\n\n" +
        "UNREGISTER        TYPE <Event Type> [SUBTYPE <Event Subtype>]...\n" +
        "                  [FORCE [MACHINE <Machine>] [HANDLE <Handle> | NAME <Name>]]\n\n" + 
        "ACKNOWLEDGE       EVENTID <Event ID>\n" +
        "                  [FORCE [MACHINE <Machine>] [HANDLE <Handle> | NAME <Name>]]\n\n" +
        "GENERATE          TYPE <Event Type> SUBTYPE <Event SubType>\n" +
        "                  [PROPERTY <Name>[=<Value>]]...\n\n" +
        "QUERY             [LONG] TYPE <Event Type> [SUBTYPE <Event Subtype>] |\n"+
        "                  [LONG] EVENTID <EventID>\n\n"  +
        "LIST              TYPES | TYPE [<Event Type>] | EVENTID <EventID> |\n" +
        "                  EVENTIDS | REGCLIENTS\n\n" +
        "RESET             REG | GEN FORCE\n\n" + 
        "VERSION\n\nHELP");
    }    
}
