/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.event;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;

public class GenerationManager
implements Serializable
{
    static final long serialVersionUID = 1; 
    transient double version = 0.9;
    transient double fileVersion;
    transient static int fEventID = 0;
    transient Hashtable fEventTable = new Hashtable();  
    transient PriorityQueue fQueue = new PriorityQueue();
    transient Thread fNotifierThread;
    transient STAFHandle eventSTAFHandle = null;
    transient String eventServiceName;
    transient RegistrationManager eventRegManager = null;

    public class EventID
    {
        transient int id;
        transient int generatingHandle;
        transient String generatingMachine;
        transient String generatingProcess;
        transient String type;
        transient String subType;
        transient String[] properties;
        transient Vector notificationList;

        EventID(int theID, int theGenHandle, String theGenMachine, 
                String theGenProcess, String theType, String theSubType,
                String[] theProperties, Vector theNotificationList)
        {
            id                = theID;
            generatingHandle  = theGenHandle;
            generatingMachine = theGenMachine;
            generatingProcess = theGenProcess;
            type              = theType;
            subType           = theSubType;
            properties        = theProperties;
            notificationList  = theNotificationList;
        }
                
        synchronized public String toString()
        {
            StringBuffer result = new StringBuffer();            

            result.append(id + ";" +  generatingMachine + ";" +   
                          generatingProcess + ";" + generatingHandle + ";" +
                          type + ";" + subType);

            for (int i = 0; i < properties.length; i++)
                result.append(";" + properties[i]); 
   
            return result.toString();
        }

        synchronized public String longToString()
        {
            StringBuffer result = new StringBuffer();            

            result.append("EventId:            " + id +
                          "\nGenerating Machine: " + generatingMachine +
                          "\nGenerating Process: " + generatingProcess +
                          "\nGenerating Handle:  " + generatingHandle + 
                          "\nType:               " + type +
                          "\nSubType:            " + subType); 

            for (int i = 0; i < properties.length; i++)
                result.append("\nProperty:           " + properties[i]); 
   
            return result.toString();
        }
    }	

    public class Notification
    {
        Client client;
        EventID eventID;
    	
        public Notification(Client client, EventID eventID)
        {
            this.client = client;
            this.eventID = eventID;
        }
		
        synchronized public String toString()
        {
            return client + ";" + eventID;
        }        

        synchronized public String longToString()
        {
            StringBuffer result = new StringBuffer();

            result.append(client.longToString() + "\n" + eventID.longToString()); 
            return result.toString();
        }
    }

    public GenerationManager(STAFHandle theSTAFHandle, String serviceName,
                             RegistrationManager regManager)
    {
        eventSTAFHandle = theSTAFHandle;
        eventServiceName = serviceName;
        eventRegManager = regManager;

        fNotifierThread = new Thread()
        {
            public void run() 
            {
                notificationThread();
            }
        };

        fNotifierThread.start();
    }

    synchronized public static void serializeEventID(ObjectOutputStream os)
    throws IOException 
    {
        os.writeInt(fEventID);
    }         

    synchronized public static void deSerializeEventID(ObjectInputStream os)
    throws IOException
    {
        fEventID = os.readInt();
    }        

    private void writeObject(ObjectOutputStream stream)
    throws IOException
    {
        stream.writeDouble(version);
        serializeEventID(stream);        
    } 

    private void readObject(ObjectInputStream stream)
    throws IOException, ClassNotFoundException
    {
        fileVersion = stream.readDouble();   
        deSerializeEventID(stream);
    }

    synchronized void serialize()
    {
        try
        {  
            ObjectOutputStream out = new ObjectOutputStream(                     
                new FileOutputStream("GenManager.out"));

            out.writeObject(this);
            out.close(); 
        } 
        catch(Exception e)
        {
           if (EventService.DEBUG) e.printStackTrace();
        }
    } 

    synchronized void deSerialize()
    {
        try
        {
            ObjectInputStream in = 
                new ObjectInputStream(
                new FileInputStream("GenManager.out")); 

                in.readObject();       
        }
        catch(Exception e)
        {
            if (EventService.DEBUG) e.printStackTrace();
        }  
    }

    synchronized public int generateEvent(String theGenMachine,
                                          String theGenProcess, int theGenHandle,
                                          String theType, String theSubtype, 
                                          String[] theProperties,
                                          Vector mergedClients)
    {
        if ((mergedClients == null) || (mergedClients.size() == 0))
            return EventService.kNoClientsForEvent;
 
        Integer nextID = new Integer(++fEventID);
        Vector notificationList = new Vector();
        EventID eventID = new EventID(nextID.intValue(), theGenHandle,
                                      theGenMachine, theGenProcess, theType,
                                      theSubtype, theProperties,
                                      notificationList);
               
        fEventTable.put(nextID, eventID);
 
        if ((mergedClients != null) && (mergedClients.size() != 0))
        {
            for (Enumeration e = mergedClients.elements(); e.hasMoreElements();)
            {
                Client client = new Client((Client)e.nextElement());
                Notification notification = new Notification(client, eventID);

                notificationList.addElement(notification);
                fQueue.enqueue(System.currentTimeMillis(), notification);
            }
        }
        else return EventService.kNoClientsForEvent;

        synchronized(fNotifierThread)
        {
            fNotifierThread.notify();  
        }           

        return nextID.intValue();
    }

    synchronized private int sendMessage(String propertyString,
                                         Notification notification)
    {
        String message = null;
        STAFResult fSTAFResult = null;
        TimeStamp now = new TimeStamp(); 
        
        if (notification.client.fWho.handle != 0)
        {
            message = "STAF/SERVICE/" + eventServiceName + "/;" + 
                      notification.eventID.id +
                      ";" + notification.eventID.generatingMachine + ";" + 
                      notification.eventID.generatingProcess + ";" +
                      notification.eventID.generatingHandle + ";" +
                      now.currentDate() + "-" + now.currentTime() + ";" +
                      notification.eventID.type + ";" +
                      notification.eventID.subType + ";" + propertyString; 

            fSTAFResult = eventSTAFHandle.submit2(
                          notification.client.getMachineName(), "queue",
                          "queue handle " + notification.client.fWho.handle +    
                          " priority " + notification.client.fHow.priority + 
                          " message :" + message.length() + ":" + message); 

            if (fSTAFResult.rc != fSTAFResult.Ok)
            {
                if (fSTAFResult.rc == STAFResult.HandleDoesNotExist ||
                    fSTAFResult.rc == STAFResult.NoPathToMachine ||
                    fSTAFResult.rc == STAFResult.CommunicationError)
                {
                    // Unregister the client (which was registered by handle)
                    // since the handle/machine is no longer available
            
                    String[] subTypes = new String[1];
                    subTypes[0] = notification.eventID.subType;
                
                    int rc = eventRegManager.unRegisterClient(
                        notification.client.getMachineName(),
                        notification.client.fWho.processName,
                        notification.client.fWho.handle,
                        notification.eventID.type, subTypes);
                }
            }
        }
        else
        {
            message = "STAF/SERVICE/" + eventServiceName + "/;" + 
                      notification.eventID.id + ";" +
                      notification.eventID.generatingMachine + ";" + 
                      notification.eventID.generatingProcess + ";" +
                      notification.eventID.generatingHandle  + ";" +
                      now.currentDate() + "-" + now.currentTime() + ";" +
                      notification.eventID.type + ";" +
                      notification.eventID.subType + ";" + propertyString;

            fSTAFResult = eventSTAFHandle.submit2(
                          notification.client.getMachineName(), "queue",
                          "queue name " + notification.client.fWho.processName +
                          " priority " + notification.client.fHow.priority + 
                          " message :" + message.length() + ":" + message);
        }               

        return fSTAFResult.rc;
    }

    synchronized public int ackEvent(int id, String machine, String processName,
                                     int handle)
    {
        EventID eventID = (EventID)fEventTable.get(new Integer(id));

        if (eventID == null) return EventService.kNoSuchID;
 
        Vector clientList = eventID.notificationList;

        if (clientList != null)
        {
            for (Enumeration e = clientList.elements(); e.hasMoreElements();)
            {
                Notification notification = (Notification)e.nextElement();

                if (((notification.client.fWho.handle == 0) &&
                     (notification.client.fWho.processName).
                      toLowerCase().equals(processName.toLowerCase())) ||
                    ((notification.client.fWho.handle != 0) &&
                     (notification.client.fWho.handle == handle)) &&
                    notification.client.fWho.machineName.equals(machine))
                {
                    notification.client.fHow.maxAttempts = 0;
                    clientList.removeElement(notification);

                    return STAFResult.Ok;  //kAckPending;
                }
            }

            return EventService.kNoAckPending;
        }

        return EventService.kNoSuchID;
    }

    synchronized public Vector getClientsForEvent(int id)
    {
        Vector result = null;
        EventID eventID = (EventID)fEventTable.get(new Integer(id));

        if (eventID != null) result = eventID.notificationList;

        return (result != null ? (Vector)result.clone() : new Vector());
    }
	
    synchronized public Vector getNotifications()
    {
        Vector result = new Vector();
		
        PriorityQueue.PriorityQueueEntry[] entries = fQueue.getQueueCopy();
		
        for (int i = 0; i < fQueue.count(); i++)
             result.addElement(entries[i]);
		
        return result;
    }
	
    void notificationThread()
    {
        while (true)
        {
            try
            {
                if (fQueue.front() == null)
                {
                    synchronized (fNotifierThread)
                    { fNotifierThread.wait(); }
                }
                else
                {
                    long waitTime = fQueue.topPriority() -
                                    System.currentTimeMillis();

                    if (waitTime  > 0)
                    {
                        synchronized (fNotifierThread)
                        { fNotifierThread.wait(waitTime); }
                    }

                    if (System.currentTimeMillis() >= fQueue.topPriority())
                        notify((Notification)fQueue.dequeue());
                }   
            }
            catch(InterruptedException e)
            {
                if (EventService.DEBUG) e.printStackTrace(); 
            } 
            catch(Exception e)
            {
                if (EventService.DEBUG) e.printStackTrace(); 
            } 
        }
    }

    synchronized private int notify(Notification notification)
    {
        Client client = notification.client;
        EventID eventID = notification.eventID;
        int id = eventID.id;
        int fReturnCode = STAFResult.Ok;
        String[] propertyArray = eventID.properties;
        int numberOfProperties = propertyArray.length;
        String propertyString = new String();

        if (fEventTable.get(new Integer(id)) == null)
            return EventService.kNoSuchID;

        if ((client.getMaxAttempts() <= 0))
        {
            Vector clientList =
                   ((EventID)fEventTable.get(new Integer(id))).notificationList;

            clientList.removeElement(notification);

            if (clientList.isEmpty())
                fEventTable.remove(new Integer(id)); 
        }
        else
        {
            for (int cnt = 0; cnt < numberOfProperties; cnt++)
                propertyString = propertyString + propertyArray[cnt] + ";"; 

            fReturnCode = sendMessage(propertyString, notification);

            if ((client.fHow.priority > 0) &&
                (client.fHow.priority >= client.fHow.priorityDelta))
            {
                    --client.fHow.priority;
            }
            else client.fHow.priority = 0;

            --client.fHow.maxAttempts;
            fQueue.enqueue(client.getTimeout() + System.currentTimeMillis(),
                           notification);
        }
       
        return fReturnCode;
    }
    
    public void handleReset()
    {
        fEventTable.clear();
    }
}
