/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2005                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.eventmanager;

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.util.*;
import java.io.*;
import org.python.util.PythonInterpreter;
import org.python.core.*;
import java.util.Properties;

public class EventManagerService implements STAFServiceInterfaceLevel4, Runnable
{
    private STAFHandle fHandle;
    private final String kVersion = "1.2.6";
    private STAFCommandParser fParmsParser = new STAFCommandParser();
    private STAFCommandParser fRegisterParser;
    private STAFCommandParser fListParser;
    private Hashtable fRegTable;
    private static boolean DEBUG = true;
    private Thread fQueueThread;
    private STAFCommandParser fUnregisterParser;
    private int fEventManagerID = 1;
    private String fLineSep;
    private String fEventServiceMachine = "";
    private String fEventServiceName = "";
    private String fEventManagerServiceMachine = "";
    private String fEventManagerServiceName = "eventmanager";
    private PythonInterpreter fPythonInterpreter = new PythonInterpreter();
    private String fHashtableFileDirectory;
    private String fHashtableFileName;
    private String fLocalMachineName;
    private HashMap fSubmittedRequests = new HashMap();
    
    public EventManagerService()
    {
    }    
    
    public int init(STAFServiceInterfaceLevel4.InitInfo info)
    {
        int rc = STAFResult.Ok;

        try
        {
            fHandle = new STAFHandle("STAF/SERVICE/" + info.name);
        }
        catch (STAFException e)
        {
            return STAFResult.STAFRegistrationError;
        }
        
        try 
        {
            fEventManagerServiceMachine = (fHandle.submit2("local", "var",
                "resolve {STAF/Config/EffectiveMachine}")).result;
            fEventManagerServiceName = info.name;
            
            fEventServiceMachine = "local";
            fEventServiceName = "event";
            
            fLocalMachineName = fEventManagerServiceMachine;

            if (info.serviceJar != null) 
            {
                // Create the Jython Library directory if it doesn't exist yet
                // from the jython library data in the service's jarfile.
                // Set the classpath to point to the service's jar file and to
                // the jython.jar.
                
                STAFResult res = STAFServiceSharedJython.setupJython(
                                     info.serviceJar, fHandle);
                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res.rc;
                }
                        
                // Initialize the PythonInterpreter's python.home to point to the
                // shared jython installation.

                Properties p = new Properties();
                p.setProperty("python.home",
                              STAFServiceSharedJython.getJythonDirName());
                PythonInterpreter.initialize(System.getProperties(), p, null);
            }

            // Parse PARMS if provided in the STAX configuration line

            if (info.parms != null)
            {
                fParmsParser.addOption("EVENTSERVICEMACHINE", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("EVENTSERVICENAME", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                rc = handleParms(info);
                
                if (rc != STAFResult.Ok)
                {
                    rc = STAFResult.STAFRegistrationError;
                }
            }

            // instantiate parsers as not case sensitive
            fRegisterParser = new STAFCommandParser(0, false);
            fListParser = new STAFCommandParser(0, false);
            fUnregisterParser = new STAFCommandParser(0, false);

            fRegisterParser.addOption("REGISTER", 1, 
                                      STAFCommandParser.VALUENOTALLOWED);

            fRegisterParser.addOption("MACHINE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOption("PYTHONMACHINE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOptionGroup("MACHINE PYTHONMACHINE", 1, 1);

            fRegisterParser.addOption("SERVICE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("PYTHONSERVICE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOptionGroup("SERVICE PYTHONSERVICE", 1, 1);

            fRegisterParser.addOption("REQUEST", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("PYTHONREQUEST", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOptionGroup("REQUEST PYTHONREQUEST", 1, 1);

            fRegisterParser.addOption("TYPE", 1, 
                                      STAFCommandParser.VALUEREQUIRED);
    
            fRegisterParser.addOption("SUBTYPE", 0, 
                                      STAFCommandParser.VALUEREQUIRED);
                                      
            fRegisterParser.addOption("PREPARE", 0, 
                                      STAFCommandParser.VALUEREQUIRED);

            fRegisterParser.addOptionNeed("REGISTER", "MACHINE PYTHONMACHINE");
            fRegisterParser.addOptionNeed("MACHINE PYTHONMACHINE", "REGISTER");

            fUnregisterParser.addOption("UNREGISTER", 1, 
                                        STAFCommandParser.VALUENOTALLOWED);

            fUnregisterParser.addOption("ID", 1, 
                                        STAFCommandParser.VALUEREQUIRED);
                                        
            fUnregisterParser.addOptionNeed("UNREGISTER", "ID");
                                               
            fListParser.addOption("LIST", 1, 
                                  STAFCommandParser.VALUENOTALLOWED);

            fListParser.addOption("MACHINE", 1, 
                                  STAFCommandParser.VALUEREQUIRED);

            fListParser.addOption("TYPE", 1, 
                                  STAFCommandParser.VALUEREQUIRED);
            
            fLineSep = (fHandle.submit2("local", "var", 
                        "resolve {STAF/Config/Sep/Line}")).result;
                        
            String fileSep = (fHandle.submit2("local", "var",
                               "resolve {STAF/Config/Sep/File}")).result;
            String stafRoot = (fHandle.submit2("local", "var",
                               "resolve {STAF/Config/STAFRoot}")).result;
                               
            fHashtableFileDirectory = stafRoot + fileSep + "data" + fileSep + 
                                      "eventmanager";
                                                            
            File dir = new File(fHashtableFileDirectory);
            
            if (!dir.exists())
            {                    
                dir.mkdirs();
            }
      
            fHashtableFileName = fHashtableFileDirectory + fileSep + 
                                 "eventman.ser";
             
            loadHashtable();            
            
            fQueueThread = new Thread(this);
            fQueueThread.start(); // this calls the run() method
        }
        catch (STAFException e)
        {
            rc = STAFResult.STAFRegistrationError;
        }

        return rc;
    }
    
    private int handleParms(STAFServiceInterfaceLevel4.InitInfo info)
        throws STAFException
    {
        STAFCommandParseResult parseResult= fParmsParser.parse(info.parms);

        if (parseResult.rc != STAFResult.Ok)
        {
            return parseResult.rc;
        }

        if (parseResult.optionTimes("EVENTSERVICEMACHINE") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok, 
                parseResult.optionValue("EVENTSERVICEMACHINE"));  

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var", 
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }
            
            if (resolvedResult.rc != STAFResult.Ok)
            {
                return resolvedResult.rc;
            }
            
            fEventServiceMachine = resolvedResult.result;
        }

        if (parseResult.optionTimes("EVENTSERVICENAME") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok, 
                parseResult.optionValue("EVENTSERVICENAME"));  

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var", 
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }
            
            if (resolvedResult.rc != STAFResult.Ok)
            {
                return resolvedResult.rc;
            }
            
            fEventServiceName = resolvedResult.result;
        }

        return 0;
    }
        
    public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo info)
    {               
        String lowerRequest = info.request.toLowerCase();

        // call the appropriate method to handle the command
        if (lowerRequest.startsWith("register"))
        {
            return handleRegister(info);
        }
        else if (lowerRequest.startsWith("unregister"))
        {
            return handleUnregister(info);
        }
        else if (lowerRequest.startsWith("list"))
        {
            return handleList(info);
        }
        else if (lowerRequest.startsWith("help"))
        {
            return handleHelp();
        }
        else if (lowerRequest.startsWith("version"))
        {
            return handleVersion();
        }
        else
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  "Unknown EventManager Request: " + 
                                  lowerRequest);
        }
    }

    private STAFResult handleHelp()
    {
        return new STAFResult(STAFResult.Ok,
         "EventManager Service Help" + fLineSep
         + "REGISTER   MACHINE <machine> | PYTHONMACHINE <machine>"
         + fLineSep + 
         "           SERVICE <service> | PYTHONSERVICE <machine>"
         + fLineSep + 
         "           REQUEST <request> | PYTHONREQUEST <request>"
         + fLineSep + "           TYPE <eventType> [SUBTYPE <eventSubType>]"
         + fLineSep + "           [PREPARE <script>]"
         + fLineSep + "UNREGISTER ID <registrationID>"
         + fLineSep + "LIST       [MACHINE <machine>] [TYPE <eventType>]"
         + fLineSep + "VERSION" + fLineSep + "HELP");
    }

    private STAFResult handleList(STAFServiceInterfaceLevel4.RequestInfo info)
    {           
        StringBuffer result = new StringBuffer();
        STAFCommandParseResult parsedRequest = fListParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }        

        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        String machineParm = "";
        String typeParm = "";

        try
        {
            STAFResult trust = 
                fHandle.submit2(info.localMachine, "trust", 
                                "get machine " + info.machine);
    
            if (trust.rc != 0)
            {
                return new STAFResult(trust.rc, trust.result);
            }

            if (Integer.parseInt(trust.result) < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                                  "Trust level 3 required for LIST request."
                                  + "Requesting machine's trust level: "
                                  + trust.result);
            }

            if (parsedRequest.rc != STAFResult.Ok)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      parsedRequest.errorBuffer);
            }

            if (info.diagEnabled == 1)
            {
                // Record diagnostics because the output for a LIST request
                // will change in STAF V3.0

                String trigger = new String("STAF/V3.0-Mig/EVENTMANAGER LIST");
                String source  = new String(info.processName + ";" +
                                            info.machine + ";" +
                                            info.handle);
                fHandle.submit2(
                    STAFHandle.ReqFireAndForget,
                    "local", "DIAG",
                    "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                    " SOURCE " + STAFUtil.wrapData(source));
            }

            int registrationTableSize = fRegTable.size();
            
            if (registrationTableSize > 0)
            {
                if (parsedRequest.optionTimes("machine") > 0)
                {
                    resolvedValue = resolveVar(
                        info.localMachine,
                        parsedRequest.optionValue("machine"), info.handle);

                    if (resolvedValue.rc != 0)
                        return resolvedValue;

                    machineParm = resolvedValue.result;
                }
    
                if (parsedRequest.optionTimes("type") > 0)
                {
                    resolvedValue = resolveVar(
                        info.localMachine,
                        parsedRequest.optionValue("type"), info.handle);

                    if (resolvedValue.rc != 0)
                        return resolvedValue;

                    typeParm = resolvedValue.result;
                }

                EventManagerData theData;
                boolean match;
                Enumeration registrationEnum = fRegTable.elements();
                Enumeration keys = fRegTable.keys();
                
                for (Enumeration e = fRegTable.elements(); 
                     e.hasMoreElements();)
                {
                    Integer id = (Integer) keys.nextElement();
                    match = false;
                    theData = (EventManagerData) e.nextElement();

                    if (machineParm.equals("") | 
                        machineParm.equalsIgnoreCase(theData.fMachine))
                    {
                        if (typeParm.equals("") | 
                            typeParm.equalsIgnoreCase(theData.fType))
                        {
                            match = true;
                        }
                    }

                    if (match)
                    {
                        result.append(id + ";");
                        result.append(theData.fMachine + ";");
                        result.append(theData.fPythonMachine + ";");
                        result.append(theData.fService + ";");
                        result.append(theData.fPythonService + ";");
                        result.append(theData.fRequest + ";");
                        result.append(theData.fPythonRequest + ";");
                        result.append(theData.fType + ";");
                        result.append(theData.fSubtype + ";");
                        result.append(theData.fPrepare + ";" + fLineSep);
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            
            return new STAFResult(STAFResult.JavaError, 
                                  "Internal Java error.");
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }
    
    private STAFResult handleRegister(
                           STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String originMachine = "";
        String machine = "";
        boolean pythonMachine = true;
        String service = "";
        boolean pythonService = true;
        String request = "";
        boolean pythonRequest = true;
        String type = "";
        String subtype = "";
        String prepare = "";

        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest = 
            fRegisterParser.parse(info.request);

        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the 
        // corresponding local variable to the user-specified value.

        try
        {
            STAFResult trust = 
                fHandle.submit2(info.localMachine, "trust", 
                                "get machine " + info.machine);

            if (trust.rc != 0)
            {
                return new STAFResult(trust.rc, trust.result);
            }

            if (Integer.parseInt(trust.result) < 5)
            {
                return new STAFResult(STAFResult.AccessDenied,
                              "Trust level 5 required for REGISTER request."
                              + " Requesting machine's trust level: "
                              + trust.result);
            }

            originMachine = info.machine;

            if (parsedRequest.optionTimes("machine") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("machine"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;

                machine = resolvedValue.result;
                pythonMachine = false;                
            }
            else if (parsedRequest.optionTimes("pythonmachine") > 0)
            {
                machine = parsedRequest.optionValue("pythonmachine");
                pythonMachine = true;
            }
            
            if (machine.equalsIgnoreCase("'local'"))
            {
                machine = "'" + fLocalMachineName + "'";
            }
            else if (machine.equalsIgnoreCase("local"))
            {
                machine = fLocalMachineName;
            }
            
            if (parsedRequest.optionTimes("service") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("service"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;

                service = resolvedValue.result;
                pythonService = false;
            }
            else if (parsedRequest.optionTimes("pythonservice") > 0)
            {
                service = parsedRequest.optionValue("pythonservice");
                pythonService = true;
            }
            
            if (parsedRequest.optionTimes("request") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("request"), info.handle);
    
                if (resolvedValue.rc != 0)
                    return resolvedValue;

                request = resolvedValue.result;
                pythonRequest = false;
            }
            else if (parsedRequest.optionTimes("pythonrequest") > 0)
            {
                request = parsedRequest.optionValue("pythonrequest");
                pythonRequest = true;
            }
            
            resolvedValue = resolveVar(info.localMachine, 
                                       parsedRequest.optionValue("type"),
                                       info.handle);
                
            if (resolvedValue.rc != 0)
                return resolvedValue;

            type = resolvedValue.result;
            
            if (parsedRequest.optionTimes("subtype") > 0)
            {
                resolvedValue = resolveVar(info.localMachine, 
                    parsedRequest.optionValue("subtype"), info.handle);

                if (resolvedValue.rc != 0)
                {
                    return resolvedValue;
                }

                subtype = resolvedValue.result;
            }
            
            if (parsedRequest.optionTimes("prepare") > 0)
            {
                prepare = parsedRequest.optionValue("prepare");
            }
            
            String registerRequest = "REGISTER TYPE " + type;
            
            if (!(subtype.equals("")))
            {
                registerRequest += " SUBTYPE " + subtype;
            }
             
            // Register with the Event service
            STAFResult registerResult = fHandle.submit2(fEventServiceMachine,
                                        fEventServiceName, registerRequest);
                                                         
            if (registerResult.rc != 0)
            {
                return registerResult;
            }
            else
            {
                Integer index = new Integer(fEventManagerID++);
                
                fRegTable.put(index, (Object) new 
                    EventManagerData(/*index.toString(), */originMachine, machine,
                    pythonMachine, service, pythonService, request, 
                    pythonRequest, type, subtype, prepare));
                                     
                sResult = new STAFResult(STAFResult.Ok, index.toString());
                saveHashtable();
                
                String message = "Register ID=" + index + " originMachine=" +
                    originMachine + " machine=" + machine + " pythonMachine=" +
                    pythonMachine + " service=" + service + " pythonService=" + 
                    pythonService + " request=" + request + " pythonRequest=" +
                    pythonRequest + " prepare=" + prepare + " type=" + type +
                    " subtype=" + subtype;
            
                fHandle.submit2("local", "log", "log machine logname " +
                    fEventManagerServiceName + " level info message " + 
                    STAFUtil.wrapData(message));
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            
            return new STAFResult(STAFResult.JavaError, 
                                  "Internal Java error.");
        }
    
        return sResult;
    }
    
    private STAFResult handleUnregister(
                           STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFResult sResult = new STAFResult(STAFResult.Ok, "");
        String reqMachine = "";
        String service;
        String request;
        Integer id;
        
        STAFResult resolvedResult = null;
        STAFResult resolvedValue = null;
        STAFCommandParseResult parsedRequest = 
            fUnregisterParser.parse(info.request);
            
        if (parsedRequest.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parsedRequest.errorBuffer);
        }

        // For each option, if the user has specified a value, set the 
        // corresponding local variable to the user-specified value.

        try
        {
            STAFResult trust = 
                fHandle.submit2(info.localMachine, "trust", 
                                "get machine " + info.machine);

            if (trust.rc != 0)
            {
                return new STAFResult(trust.rc, trust.result);
            }

            if (Integer.parseInt(trust.result) < 3)
            {
                return new STAFResult(STAFResult.AccessDenied,
                              "Trust level 3 required for UNREGISTER request."
                              + " Requesting machine's trust level: "
                              + trust.result);
            }
                        
            reqMachine = info.machine;
                        
            resolvedValue = resolveVar(info.localMachine, 
                                       parsedRequest.optionValue("id"), 
                                       info.handle);

            if (resolvedValue.rc != 0)
            {
                return resolvedValue;
            }

            id = new Integer(resolvedValue.result);
            
            if (!fRegTable.containsKey(id))
            {
                return new STAFResult(STAFResult.DoesNotExist, 
                                      "ID " + id + " not found");
            }
            else
            {
                EventManagerData data = (EventManagerData)fRegTable.get(id);
                
                if (!(data.fOriginMachine.equals(reqMachine)))
                {
                    if (Integer.parseInt(trust.result) < 4)
                    {
                        return new STAFResult(STAFResult.AccessDenied,
                                      "Trust level 4 required for UNREGISTER " +
                                      " when the requesting machine did not " +
                                      " submit the REGISTER request. " 
                                      + " Requesting machine's trust level: "
                                      + trust.result);
                    }
                }

                fRegTable.remove(id);
                
                if (fRegTable.isEmpty())
                {
                    fEventManagerID = 1;
                }

                saveHashtable();
                
                String message = "Unregister ID=" + id;

                fHandle.submit2("local", "log", "log machine logname " +
                    fEventManagerServiceName + " level info message " + 
                    STAFUtil.wrapData(message));
                
                String unregType = data.fType;
                String unregSubtype = data.fSubtype;
                boolean unregWithEventService = true;

                // Only unregister with the Event Service if there are no
                // other fRegTable entries with this type and subtype

                for (Enumeration e = fRegTable.elements();
                    e.hasMoreElements();)
                {
                     EventManagerData theData =
                        (EventManagerData) e.nextElement();

                    if (unregType.equalsIgnoreCase(theData.fType) &&
                        theData.fSubtype.equalsIgnoreCase(unregSubtype))
                    {
                        unregWithEventService = false;
                    }
                }
                
                if (unregWithEventService)
                {
                    String unRegisterRequest = "UNREGISTER TYPE " + unregType;

                    if (!(data.fSubtype.equals("")))
                    {
                        unRegisterRequest += " SUBTYPE " + unregSubtype;
                    }

                        // Unregister with the Event service
                    STAFResult unRegisterResult = fHandle.submit2(
                        fEventServiceMachine, fEventServiceName,
                        unRegisterRequest);
                
                    if (unRegisterResult.rc != 0)
                    {
                        return unRegisterResult;
                    }
                }
            }
        }
        catch (Exception e)
        {
            if (DEBUG)
            {
                e.printStackTrace();
            }
            
            return new STAFResult(STAFResult.JavaError, "Internal Java error.");
        }
    
        return sResult;
    }

    private STAFResult handleVersion()
    {
        return new STAFResult(STAFResult.Ok, kVersion);
    }    

    private STAFResult integerCheck(String option, String optionValue, 
                                    boolean defaultOption)
    {
        char[] c = optionValue.toCharArray();
        
        for (int count = 0; count < c.length; count++)
        {
            if (!Character.isDigit(c[count]) && !defaultOption)
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                                      "The value you entered for "
                                      + option + ", " + optionValue
                                      + ", is not a number. the value of "
                                      + option + " must be" + " numeric.");
            }
            else if (!Character.isDigit(c[count]))
            {
                return new STAFResult(STAFResult.InvalidRequestString,
                        "The default value entered at STAF initialization for "
                        + option + ", " + optionValue + ", is not a number. "
                        + "STAF must be reinitialized with a numeric value"
                        + " for " + option + ".");
            }
        }

        return new STAFResult(STAFResult.Ok, optionValue);
    }

    private STAFResult resolveAndCheckInt(String machine, String option,
                                          String optionValue, int handle,
                                          boolean defaultOption)
    {
        STAFResult resolvedValue = resolveVar(machine, optionValue, handle);

        if (resolvedValue.rc != 0)
        {
            return resolvedValue;
        }

        return integerCheck(option, resolvedValue.result, defaultOption);
    }

    private STAFResult resolveVar(String machine, String optionValue, 
                                  int handle)
    {
        String value = "";
        STAFResult resolvedResult = null;

        if (optionValue.indexOf("{") != -1)
        {
            resolvedResult =
                fHandle.submit2(machine, "var", "handle " + handle + 
                                " resolve " + STAFUtil.wrapData(optionValue));

            if (resolvedResult.rc != 0)
            {
                return resolvedResult;
            }

            value = resolvedResult.result;
        }
        else
        {
            value = optionValue;
        }

        return new STAFResult(STAFResult.Ok, value);
    }

    public void run()
    {
        STAFResult queueGetResult;
        
        for (;;)
        {
            queueGetResult = fHandle.submit2("local", "queue", "get wait");

            try
            {
            
            String message = queueGetResult.result;
            
            int priorityPos = message.indexOf(';');
            int timestampPos = message.indexOf(';', priorityPos + 1);
            int machinePos = message.indexOf(';', timestampPos + 1);
            int processPos = message.indexOf(';', machinePos + 1);
            int handlePos = message.indexOf(';', processPos + 1);        
                
            String queuedMessage = message.substring(handlePos + 1);
            
            if (queuedMessage.indexOf("EventManager/End") > -1)
            {
                // this message indicates that STAFProc is ending and
                // the service is terminating

                try
                {
                    fHandle.unRegister();
                }
                catch (STAFException ex)
                {
                    if (DEBUG)
                    {
                        ex.printStackTrace();
                    }
                }                

                return;
            }
            else if (queuedMessage.startsWith("STAF/RequestComplete"))
            {
                int rcIndex = queuedMessage.indexOf(";");
                int resultIndex = queuedMessage.indexOf(";", rcIndex + 1);
                String reqNum = queuedMessage.substring(21, rcIndex);
                String rc = queuedMessage.substring(rcIndex + 1, resultIndex);
                String result = queuedMessage.substring(resultIndex + 1);
                       
                String level = "";
                   
                if (rc.equals("0"))
                    level = "info";
                else
                    level = "error";
                      
                String request = (String)fSubmittedRequests.get(reqNum);

                fHandle.submit2("local", "log", "log machine logname " +
                    fEventManagerServiceName + " level " + level + " message "
                    + STAFUtil.wrapData("Request #" + reqNum + ": RC=" + 
                    rc + ", Result=" + result + ", Request submitted: " +
                    request));
                          
                fSubmittedRequests.remove(reqNum);
            }
            
            String upperMessage = message.toUpperCase();
            int index = upperMessage.indexOf(("STAF/SERVICE/" + 
                fEventServiceName).toUpperCase());
                
            if (index > -1)
            {            
                // create new Python Interpreter for each Event received
                fPythonInterpreter = new PythonInterpreter();
                
                Hashtable eventMap = new Hashtable();
                
                int queueTimestampPos = message.indexOf(";");
                int queueMachinePos = message.indexOf(";", queueTimestampPos + 1);
                int queueNamePos = message.indexOf(";", queueMachinePos + 1);
                int queueHandlePos = message.indexOf(";", queueNamePos + 1);
            
                int eventNamePos = message.indexOf(";", queueHandlePos + 1);
                int eventIDPos = message.indexOf(";", eventNamePos + 1);
                int genMachinePos = message.indexOf(';', eventIDPos + 1);
                int genProcessPos = message.indexOf(';', genMachinePos + 1);
                int genHandlePos = message.indexOf(';', genProcessPos + 1);
                int eventTimestampPos = message.indexOf(';', genHandlePos + 1);
                int eventTypePos = message.indexOf(';', eventTimestampPos + 1);
                int eventSubtypePos = message.indexOf(';', eventTypePos + 1);                
                int propertyPos = message.indexOf(';', eventSubtypePos + 1);
             
                String eventService = message.substring(eventNamePos + 1, 
                                                        eventIDPos);                
                pySetVar("eventservice", eventService);
                eventMap.put(Py.java2py("eventservice"), 
                    Py.java2py(eventService));
                
                String eventID = message.substring(eventIDPos + 1, 
                                                   genMachinePos);                
                pySetVar("eventid", eventID);
                eventMap.put(Py.java2py("eventid"), Py.java2py(eventID));
                
                String generatingMachine = message.substring(genMachinePos + 1,
                                                             genProcessPos);                
                pySetVar("generatingmachine", generatingMachine);
                eventMap.put(Py.java2py("generatingmachine"), 
                    Py.java2py(generatingMachine));
                
                String generatingProcess = message.substring(genProcessPos + 1,
                                                             genHandlePos);                
                pySetVar("generatingprocess", generatingProcess);
                eventMap.put(Py.java2py("generatingprocess"), 
                    Py.java2py(generatingProcess));
                
                String generatingHandle = message.substring(genHandlePos + 1,
                                                            eventTimestampPos);                
                pySetVar("generatinghandle", generatingHandle);                              
                eventMap.put(Py.java2py("generatinghandle"), 
                    Py.java2py(generatingHandle));
                
                String eventTimestamp = message.substring(eventTimestampPos 
                                                          + 1, eventTypePos);
                pySetVar("eventtimestamp", eventTimestamp);
                eventMap.put(Py.java2py("eventtimestamp"), 
                    Py.java2py(eventTimestamp));
                
                String eventType = message.substring(eventTypePos + 1, 
                                                     eventSubtypePos);                
                pySetVar("eventtype", eventType);
                eventMap.put(Py.java2py("eventtype"), Py.java2py(eventType));
                
                String eventSubtype = message.substring(eventSubtypePos + 1, 
                                                        propertyPos);                
                pySetVar("eventsubtype", eventSubtype);
                eventMap.put(Py.java2py("eventsubtype"), 
                    Py.java2py(eventSubtype));
                    
                HashMap eventProperties = new HashMap();
                                               
                if (propertyPos != -1)
                {
                    while ((propertyPos != -1) && 
                           (propertyPos != message.length() - 1))
                    {
                        int equalsIndex = message.indexOf("=", propertyPos);
                        String name = message.substring(propertyPos + 1, 
                                                        equalsIndex);
                        String value = message.substring(equalsIndex + 1,
                                          message.indexOf(';', equalsIndex));
                                              
                        eventProperties.put(name, value); 
                        pySetVar(name, value);
                        eventMap.put(Py.java2py(name), Py.java2py(value));
                        propertyPos = message.indexOf(';', propertyPos + 1);
                    }
                }
                       
                PyDictionary args = new PyDictionary(eventMap);
                pySetVar("eventinfo", args);
    
                EventManagerData theData;
                
                for (Enumeration e = fRegTable.keys(); 
                     e.hasMoreElements();)
                {
                    Integer id = (Integer)e.nextElement();
                    
                    theData = (EventManagerData) fRegTable.get(id);
                    
                    if (eventType.equalsIgnoreCase(theData.fType))
                    {
                        if (theData.fSubtype.equals("") || 
                            eventSubtype.equalsIgnoreCase(theData.fSubtype))
                        {
                            pySetVar("STAFEventManagerSubmit", "true");
                            
                            String prepare = theData.fPrepare;
                            pyExec(prepare);
                            
                            Object submit = 
                                pyGetVar("STAFEventManagerSubmit");
                                                             
                            if ((submit != null) && 
                                (submit.toString().equals("true")))
                            {
                                boolean pythonMachine = theData.fPythonMachine;
                                boolean pythonService = theData.fPythonService;
                                boolean pythonRequest = theData.fPythonRequest;

                                String machine = "";
                                String service = "";
                                String request = "";
                                              
                                if (pythonMachine)
                                {
                                    try
                                    {
                                        machine = 
                                            pyStringEval(theData.fMachine);
                                    }
                                    catch (PyException ex)
                                    {
                                        String msg = "Error: " + 
                                            "pythonmachine value is not valid"
                                            + " python code: " + 
                                            theData.fMachine + " exception=" +
                                            ex;
                                            
                                        fHandle.submit2("local", "log", 
                                            "log machine logname " +
                                            fEventManagerServiceName + 
                                            " level error message " +
                                            STAFUtil.wrapData(msg));

                                        continue;
                                    }
                                }
                                else
                                {
                                    machine = theData.fMachine;
                                }
                                
                                if (pythonService)
                                {
                                    try
                                    {
                                        service  = 
                                            pyStringEval(theData.fService);
                                    }
                                    catch (PyException ex)
                                    {
                                        String msg = "Error: " + 
                                            "pythonservice value is not valid"
                                            + " python code: " + 
                                            theData.fService + " exception=" +
                                            ex;
                                            
                                        fHandle.submit2("local", "log", 
                                            "log machine logname " +
                                            fEventManagerServiceName + 
                                            " level error message " +
                                            STAFUtil.wrapData(msg));
                                       
                                        continue;
                                    }
                                }
                                else
                                {
                                    service = theData.fService;
                                }
                           
                                if (pythonRequest)
                                {
                                    try
                                    {
                                        request = 
                                             pyStringEval(theData.fRequest);
                                    }
                                    catch (PyException ex)
                                    {
                                        String msg = "Error: " + 
                                            "pythonrequest value is not valid"
                                            + " python code: " + 
                                            theData.fRequest + " exception=" +
                                            ex;
                                            
                                        fHandle.submit2("local", "log", 
                                            "log machine logname " +
                                            fEventManagerServiceName + 
                                            " level error message " +
                                            STAFUtil.wrapData(msg));
                                            
                                         continue;
                                    }
                                    request = pyStringEval(theData.fRequest);
                                }
                                else
                                {
                                    request = theData.fRequest;
                                }
                                
                                STAFResult matchResult = fHandle.submit2(
                                    STAFHandle.ReqQueue,
                                    machine, service, request);
                                    
                                fSubmittedRequests.put(matchResult.result,
                                    "machine=" + machine + " service=" + 
                                    service + " request=" + request);
                              
                                message = "Submitted request #" + 
                                    matchResult.result + " ID=" + id.toString()
                                    + " type=" + eventType + 
                                    " subtype=" + eventSubtype +
                                    " prepare=" + theData.fPrepare +
                                    " eventservice=" + eventService +
                                    " eventid=" + eventID +
                                    " generatingmachine=" + generatingMachine +
                                    " generatingprocess=" + generatingProcess +
                                    " generatinghandle=" + generatingHandle +
                                    " eventtimestamp=" + eventTimestamp +
                                    " properties=" + eventProperties +
                                    " machine=" + machine + " service=" + 
                                    service + " request=" + request;
    
                                fHandle.submit2("local", "log", 
                                    "log machine logname " +
                                    fEventManagerServiceName + 
                                    " level info message " +
                                    STAFUtil.wrapData(message));
                            }
                        }
                    }
                }
            }
            }
            catch (Exception ex)
            {
                // Log an error and continue processing messages on the queue

                fHandle.submit2(
                    "local", "LOG", "LOG MACHINE LOGNAME " +
                    fEventManagerServiceName + " LEVEL Error MESSAGE " +
                    STAFUtil.wrapData(
                        "Exception handling queued message. Exception:\n" +
                        ex + "\nQueued Message:\n" + queueGetResult.result));
            }
        }
    
    }
    
    public void loadHashtable()
    {
        try 
        {
            File fHashtableFile = new File(fHashtableFileName);
            
            if (!fHashtableFile.exists()) 
            {
                fRegTable = new Hashtable();
                return; 
            }
            else 
            {
                ObjectInputStream ois = 
                    new ObjectInputStream(new FileInputStream(fHashtableFile));
                fRegTable = (Hashtable)(ois.readObject());
                ois.close();		
                
                for (Enumeration e = fRegTable.elements(); 
                     e.hasMoreElements();)
                {
                    EventManagerData data = (EventManagerData) e.nextElement();
                    STAFResult registerResult = 
                        fHandle.submit2(fEventServiceMachine,
                                        fEventServiceName,
                                        "REGISTER TYPE " + data.fType + " " 
                                        + "SUBTYPE " + data.fSubtype);
                
                    if (registerResult.rc != 0)
                    {
                        fHandle.submit2("local", "log", "log machine logname "
                            + fEventManagerServiceName + 
                            " level error message " + 
                            STAFUtil.wrapData("Error registering type " + 
                            data.fType + " subtype " + 
                            data.fSubtype + " with the "
                            + fEventServiceName + " service"));
                    }
                }
 
                fEventManagerID = findNextEventManagerID();                

            }	
        } 
        catch (IOException e) 
        {e.printStackTrace();
            fRegTable = new Hashtable();
            return;
        }
        catch (ClassNotFoundException e) 
        {e.printStackTrace();
            fRegTable = new Hashtable();
            return;
        }
    }
    
    public void saveHashtable()
    {
        try
        {            
            ObjectOutputStream oos = 
                new ObjectOutputStream(new 
                    FileOutputStream(fHashtableFileName));
            oos.writeObject(((Object)(fRegTable)));				
            oos.close();                         
        } 
        catch (IOException ex) 
        {
            ex.printStackTrace();
        }
    }
    
    public int findNextEventManagerID()
    {
        int highestID = 0;
        for (Enumeration e = fRegTable.keys(); e.hasMoreElements();)
        {
            Integer id = (Integer) e.nextElement();
            if (id.intValue() > highestID)
            {
                highestID = id.intValue();
            }
        }
        
        return highestID + 1;
    }
    
    public int term()
    {
        fHandle.submit2(STAFHandle.ReqFireAndForget, "local",
                        "queue", "queue message EventManager/End");
        
        // Un-register the service handle

        try
        {
            fHandle.unRegister();
        }
        catch (STAFException ex)
        {
            return STAFResult.STAFRegistrationError;
        }

        return STAFResult.Ok;
    }   
     
    public String pyStringEval(String value) throws PyException
    {
        synchronized(fPythonInterpreter)
        {
            PyObject result = null;
                     
            result = fPythonInterpreter.eval(value);
            
            if (result == null) 
            {
                return null;
            }
            else
            {
                return result.toString();
            }
        }
    }
    
    public void pySetVar(String varName, Object value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.set(varName, value);
            }
            catch (PyException e)
            {
                fHandle.submit2("local", "log", "log machine logname " +
                    fEventManagerServiceName + " level error message " + 
                    STAFUtil.wrapData("PySetVar PyException caught setting " +
                    " var=" + varName + " to value=" + value + e.toString()));
            } 
        }
    }
    
    public Object pyGetVar(String varName)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                return fPythonInterpreter.get(varName);
            }
            catch (PyException e)
            {
                fHandle.submit2("local", "log", "log machine logname " +
                    fEventManagerServiceName + " level error message " + 
                    STAFUtil.wrapData("PyGetVar PyException caught getting " +
                    " var=" + varName + " to value=" + varName + 
                    e.toString()));
                
                return null;
            } 
        }
    }
    
    public void pyExec(String value)
    {
        synchronized(fPythonInterpreter)
        {
            try
            {
                fPythonInterpreter.exec(value);
            }
            catch (PyException e)
            {
                e.printStackTrace();
            } 
        }
    }
    
}
