package com.ibm.staf.service.fsext;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Author: Chris Alkov                                                       */
/* Date: 12/2001                                                             */
/* Revisions:                                                                */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/* Class: FSEXT                                                              */
/* Description: This class provides the STAFService Interface and implements */
/*              the majority of the service function                         */
/*                                                                           */
/*****************************************************************************/

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import com.ibm.staf.service.utility.ServiceUtilities;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;
  
public class FSExt implements STAFServiceInterfaceLevel4 
{
    private STAFHandle sHandle;
    private String sServiceName;
    
    /* STAF Parsers */
    
    private STAFCommandParser procFileParser;
    private STAFCommandParser compDirParser;
    private STAFCommandParser fileContainsParser;
    private STAFCommandParser lineContainsParser;
    private STAFCommandParser waitForFileParser;
    private STAFCommandParser substituteParser;

    /* String Constants */

    private static final String FSEXTVERSION = "1.2.4";

    private static final String PROCESSFILE = "PROCESSFILE";
    private static final String MODE    = "MODE";
    private static final String CAPTURE = "CAPTURE";
    private static final String COMPARE = "COMPARE";
    private static final String FILE1   = "FILE1";
    private static final String FILE2   = "FILE2";
    private static final String SORT    = "SORT";

    private static final String COMPAREDIR = "COMPAREDIR";
    private static final String DIR     = "DIR";
    private static final String EXISTS  = "EXISTS";
    private static final String ATTEMPTS = "ATTEMPTS";
    private static final String POLL    = "INTERVAL";

    private static final String FILECONTAINS = "FILECONTAINS";
    private static final String LINECONTAINS = "LINECONTAINS";
    private static final String FILE    = "FILE";
    private static final String STRING = "STRING";
    private static final String NOT     = "NOT";
    private static final String IGNORECASE = "IGNORECASE";
    private static final String SAVEONFAILURE = "SAVEONFAILURE";
    private static final String LINENUMBER = "LINENUMBER";

    private static final String WAITFORFILE = "WAITFORFILE";
    private static final String TIMEOUT = "TIMEOUT";

    private static final String HELP = "HELP";
    private static final String VERSION = "VERSION";
    
    private static final String SUBSTITUTE = "SUBSTITUTE";
    private static final String TOFILE = "TOFILE";
    private static final String TOMACHINE = "TOMACHINE";
    private static final String FAILIFEXISTS = "FAILIFEXISTS";
    private static final String FAILIFNEW = "FAILIFNEW";

    /* Return Codes */

    private static final int COMPAREFAIL = 4001;
    private static final String COMPAREFAILInfo = "File Comparison Failed";
    private static final String COMPAREFAILDesc = 
        "The file comparison failed.";
    
    private static final int NOTDIRECTORY = 4002;
    private static final String NOTDIRECTORYInfo = "Not A Directory";
    private static final String NOTDIRECTORYDesc = 
        "The value specified for the DIR parameter is not a directory.";
    
    private static final int FILENOTFOUND = 4003;
    private static final String FILENOTFOUNDInfo = "File Not Found";
    private static final String FILENOTFOUNDDesc = 
        "The specified file was not found on the file system.";
    
    private static final int FILEEXISTS = 4004;
    private static final String FILEEXISTSInfo = "File Exists";
    private static final String FILEEXISTSDesc = 
        "The specified file exists. (The NOT parameter was specified.)";

    private static final int EXTRAFILES = 4005;
    private static final String EXTRAFILESInfo = "Extraneous Files Found";
    private static final String EXTRAFILESDesc = 
        "Extraneous files were found in the specified directory.";
    
    private static final int STRINGNOTFOUND = 4006;
    private static final String STRINGNOTFOUNDInfo = "String(s) Not Found";
    private static final String STRINGNOTFOUNDDesc = 
        "The specified String(s) were not found.";
    
    private static final int STRINGFOUND = 4007;
    private static final String STRINGFOUNDInfo = "String(s) Found";
    private static final String STRINGFOUNDDesc = 
        "The specified String(s) were found. " + 
        "(The NOT parameter was specified.)";

    private static final int LINEDOESNOTEXIST = 4008;
    private static final String LINEDOESNOTEXISTInfo = 
        "Line Number Does Not Exist";
    private static final String LINEDOESNOTEXISTDesc = 
        "The line number specified by the LINENUMBER parameter " + 
        "does not exist in the file.";
    
    private static final int IOERROR = 4100;
    private static final String IOERRORInfo = "Java IOError";
    private static final String IOERRORDesc = 
        "A Java IOError occurred during command execution. Check the result " +
        "buffer{STAF/Config/Sep/Line}for more information.";
    
/*****************************************************************************/
/*                                                                           */
/* Method: Constructor                                                       */
/* Description: Constructor method                                           */
/*                                                                           */
/*****************************************************************************/

public FSExt() 
{
    super();
}

/*****************************************************************************/
/*                                                                           */
/* Method: acceptRequest                                                     */
/* Description: required by interface STAFServiceInterfaceLevel4             */
/*              performs first parse of request                              */
/*                                                                           */
/*****************************************************************************/

public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo reqInfo)
{    
    String request = reqInfo.request.toUpperCase();
    STAFResult result = null;
    
    try
    {
        if (request.startsWith(PROCESSFILE)) 
        {
            result = processFile(reqInfo);
        }
        else if (request.startsWith(COMPAREDIR)) 
        {
            result = compareDir(reqInfo);
        }
        else if (request.startsWith(FILECONTAINS)) 
        {
            result = fileContains(reqInfo);
        }
        else if (request.startsWith(LINECONTAINS)) 
        {
            result = lineContains(reqInfo);
        }
        else if (request.startsWith(WAITFORFILE)) 
        {
            result = waitForFile(reqInfo);
        }
        else if(request.startsWith(SUBSTITUTE)) 
        {
            result = substitute(reqInfo);
        }
        else if (request.startsWith(HELP)) 
        {
            result = help();
        }
        else if (request.startsWith(VERSION)) 
        {
            result = version();
        }
        else 
        {
            result = new STAFResult(STAFResult.InvalidRequestString, 
                request);
        }
    }
    catch(Throwable t) 
    {
        result = new STAFResult(STAFResult.JavaError, t.toString());
    }

    return result;
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: compareDir                                                        */
/* Description: performs compareDir service function                         */
/* Parameters: reqInfo - RequestInfo passed to acceptRequest                 */
/* Returns: STAFResult                                                       */
/*                                                                           */
/*****************************************************************************/

private STAFResult compareDir(STAFServiceInterfaceLevel4.RequestInfo reqInfo) 
{
    /* Check for trust level of 3 or greater */

    if (reqInfo.trustLevel < 3) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel, "Minimum trust level of 3 required.");
    }

    /* Parse request string */

    STAFCommandParseResult pResult = compDirParser.parse(reqInfo.request);
    
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    /* Get Parameters */

    File dir;
    String[] expectedFiles;
    
    /* Parse DIR & FILES parms */
    
    try 
    {
        String dirName = 
            ServiceUtilities.getResolvedParm(reqInfo, pResult, DIR, sHandle);
        dir = new File(dirName);
        
        if (!dir.isDirectory()) 
        { //dir specified is not a directory
            return new STAFResult(NOTDIRECTORY, dirName);
        }

        /* Create a String array of fully qualified filenames from  
           list of relative filenames passed in via the FILE parm. */

        int numFiles = pResult.optionTimes(FILE);
        expectedFiles = new String[numFiles];
        
        for(int i=0; i < numFiles; i++) {
            String shortName = ServiceUtilities.getResolvedParm(reqInfo, 
                pResult, FILE, i+1, sHandle);
            File file = new File(dir, shortName);
            expectedFiles[i] = file.getPath();
        }
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, se.toString());
    }

    /* Parse EXISTS parm */

    boolean exact = true;

    if (pResult.optionTimes(EXISTS) == 1) 
    {
        exact = false;
    }
    else if (reqInfo.diagEnabled == 1)
    {
        // Record diagnostics because the output for a COMPAREDIR request
        // without the EXISTS option specified will change in STAF V3.0

        String trigger = new String(
            "STAF/V3.0-Mig/FSEXT COMPAREDIR (No EXISTS)");
        String source  = new String(reqInfo.processName + ";" +
                                    reqInfo.machine + ";" +
                                    reqInfo.handle);
        sHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "DIAG",
            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
            " SOURCE " + STAFUtil.wrapData(source));
    }

    /* Parse RETRY parm */

    int attempts = 1;
    int poll = 15000;
    
    try 
    {
        if (pResult.optionTimes(ATTEMPTS) == 1) 
        {
            attempts = Integer.parseInt(ServiceUtilities.getResolvedParm(
                reqInfo, pResult, ATTEMPTS, sHandle));
        }
        if (pResult.optionTimes(POLL) == 1) 
        {
            poll = Integer.parseInt(ServiceUtilities.getResolvedParm(
                reqInfo, pResult, POLL, sHandle));
        }
    }
    catch(STAFException se) 
    {
        return new STAFResult(se.rc, se.toString());
    }
    catch(NumberFormatException nfe) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, nfe.toString());
    }

    /* Start a large for loop to handle retries */

    for (int loop = 1; loop <= attempts; loop++) 
    {
        boolean success = true;

        /* Create a HashMap of all files in the specified directory. The filenames
           shall be the keys for the Map and the values will all initially be set 
           to false. As a filename in the FILES list is matched with a file in the 
           file system the value will be set to true. */
        
        HashMap files = ServiceUtilities.createMapOfFiles(dir);

        /* Loop through the list of expected files and verify that all of them
           exist in the specified directory */

        for(int i=0; i < expectedFiles.length; i++) 
        {
            if (files.containsKey(expectedFiles[i])) 
            { //file exists, set value to true
                files.put(expectedFiles[i], new Boolean(true));
            }
            else if (loop == attempts) 
            { // file does not exist, last retry so return failure
                return new STAFResult(FILENOTFOUND, expectedFiles[i]);
            }
            else 
            { // file does not exist, set success to false for this attempt
                success = false;
            }
        }

        /* If all the expected files were found and EXISTS was not specified, check
           that no extraneous files exist */

        if (exact && success) 
        { 
            if (files.containsValue(new Boolean(false))) 
            {// at least one extra file found
                if (loop == attempts) 
                { // last retry so return failure and list of extra files
                    StringBuffer fileList = new StringBuffer();
                    Iterator keys = files.keySet().iterator();
                    
                    while(keys.hasNext()) 
                    {// loop thru map and return all files with a value of false
                        String key = (String) keys.next();
                        boolean failed = 
                            !((Boolean) files.get(key)).booleanValue();
                            
                        if (failed) 
                        {
                            fileList.append(key+" ");
                        }
                    }
                    return new STAFResult(EXTRAFILES, fileList.toString());
                }
                else {// extra files found, set success to false for this attempt
                    success = false;
                }
            } // end if file.containsValue
        } // end if (exact && success)

        /* If all checks are successful break out of big loop and return OK. Otherwise
           delay and start check again at top of loop */

        if (success) 
        {
            break;
        }
        else 
        {
            sHandle.submit2("LOCAL", "DELAY", "DELAY "+poll);
        }

    } // end big for loop

    return new STAFResult(STAFResult.Ok);
}

/*****************************************************************************/
/*                                                                           */
/* Method: fileContains                                                      */
/* Description: performs fileContains service function                       */
/* Parameters: reqInfo - RequestInfo passed to acceptRequest                 */
/* Returns: STAFResult                                                       */
/*                                                                           */
/*****************************************************************************/

private STAFResult fileContains(STAFServiceInterfaceLevel4.RequestInfo reqInfo)
{    
    /* Check for trust level of 3 or greater */

    if (reqInfo.trustLevel < 3) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel, 
            "Minimum trust level of 3 required.");
    }

    /* Parse request String */

    STAFCommandParseResult pResult = fileContainsParser.parse(reqInfo.request);
    
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    /* Get FILE and STRING parameters */
    
    File file;
    Enumeration strings;
    
    try 
    {
        String filename = ServiceUtilities.getResolvedParm(reqInfo, 
            pResult, FILE, sHandle);
        file = new File(filename);

        /* Get all instances of STRING, store in a Vector and get the Enumeration */
        
        int numStrings = pResult.optionTimes(STRING);
        Vector stringVect = new Vector();
        
        for(int i=1; i <= numStrings; i++) 
        {
            String tempString = ServiceUtilities.getResolvedParm(
                reqInfo, pResult, STRING, i, sHandle);
            stringVect.add(tempString);
        }
        strings = stringVect.elements();
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, se.toString());
    }

    /* Parse NOT parameter */

    boolean fileContainsString = true;
    
    if (pResult.optionTimes(NOT) == 1) 
    {
        fileContainsString = false;

        if (reqInfo.diagEnabled == 1)
        {
            // Record diagnostics because the output for a FILECONTAINS request
            // with the NOT option specifed will change in STAF V3.0

            String trigger = new String("STAF/V3.0-Mig/FSEXT FILECONTAINS NOT");
            String source  = new String(reqInfo.processName + ";" +
                                        reqInfo.machine + ";" +
                                        reqInfo.handle);
            sHandle.submit2(
                STAFHandle.ReqFireAndForget,
                "local", "DIAG",
                "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                " SOURCE " + STAFUtil.wrapData(source));
        }
    }

    /* Parse IGNORECASE parameter */

    boolean ignoreCase = false;
    
    if (pResult.optionTimes(IGNORECASE) == 1) 
    {
        ignoreCase = true;
    }

    /* Parse SAVEONFAILURE parameter */

    boolean saveOnFail = false;
    
    if (pResult.optionTimes(SAVEONFAILURE) == 1) 
    {
        saveOnFail = true;
    }

    /* Read entire file and store as a String */

    FileReader in = null;
    String data = null;
    
    try 
    {
        in = new FileReader(file);
        int numBytes = (int) file.length();
        char[] fileData = new char[numBytes];
        in.read(fileData);
        data = new String(fileData);
        if (ignoreCase) 
        {
            data = data.toLowerCase();
        }
    }
    catch(IOException ioe) 
    {
        return new STAFResult(IOERROR, ioe.toString());
    }
    finally
    {
        try 
        {
            in.close();
        }
        catch(Throwable t) {} //don't care
    }
    
    /* Scan entire file for presence or absence of the specified Strings */

    if (fileContainsString) 
    { //check that file contains list of Strings
        while(strings.hasMoreElements()) 
        {
            String tempS = (String) strings.nextElement();
            if (ignoreCase) 
            {// IGNORECASE specified
                tempS = tempS.toLowerCase();
            }
            if (data.indexOf(tempS) == -1) 
            {// String not found
                if (saveOnFail) 
                {// capture the file for later comparison 
                    ExtFile capture = new ExtFile(file);
                    try 
                    {
                        capture.captureFile(file.getPath()+".fail");
                    }
                    catch(IOException ioe) 
                    {
                        return new STAFResult(IOERROR, ioe.toString());
                    }
                }

                return new STAFResult(STRINGNOTFOUND, tempS);
                
            }
        }
    }
    else 
    { //check that the file does NOT contain the list of Strings
        while(strings.hasMoreElements()) 
        {
            String tempS = (String) strings.nextElement();
            
            if (ignoreCase) 
            {// IGNORECASE specified
                tempS = tempS.toLowerCase();
            }
            
            int index = data.indexOf(tempS);
            
            if (index != -1) 
            {// String was found in the file
                if (saveOnFail) 
                {// capture the file for later comparison
                    ExtFile capture = new ExtFile(file);
                    try 
                    {
                        capture.captureFile(file.getPath()+".fail");
                    }
                    catch(IOException ioe) 
                    {
                        return new STAFResult(IOERROR, ioe.toString());
                    }
                }
                
                return new STAFResult(STRINGFOUND, index+":"+tempS);
                
            }
        }
    }               

    return new STAFResult(STAFResult.Ok);
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: help                                                              */
/* Description: returns service help information                             */
/* Parameters: none                                                          */
/* Returns: STAFResult.OK                                                    */
/*                                                                           */
/*****************************************************************************/

private STAFResult help() 
{    
    String help =
        sServiceName + " Service Help\n\n"+
        "COMPAREDIR DIR <directory> FILE <file1> [FILE <file 2> [...]] " + 
        "[EXISTS]\n"+
        "     [ATTEMPTS <num attempts> [INTERVAL <polling interval>]]\n\n"+
        "FILECONTAINS FILE <file> STRING <string1> [STRING <string2> [...]] " +
        "[NOT]\n"+
        "     [IGNORECASE] [SAVEONFAILURE]\n\n"+
        "LINECONTAINS FILE <file> STRING <string1> [STRING <string2> [...]] " +
        "[IGNORECASE]\n"+
        "     [SAVEONFAILURE] [LINENUMBER <line #>]\n\n"+
        "PROCESSFILE [MODE <capture | compare>] FILE1 <file1> FILE2 <file2> " +
        "[SORT]\n"+
        "     [SAVEONFAILURE]\n\n"+
        "SUBSTITUTE FILE <file> [TOFILE <file>] [TOMACHINE <machine>]\n"+
        "     [FAILIFNEW | FAILIFEXISTS]\n\n"+
        "WAITFORFILE FILE <filename> [TIMEOUT <max wait time>]\n"+
        "     [INTERVAL <polling interval>] [NOT]\n\n"+
        "VERSION\n\n"+
        "HELP";

    STAFResult result = new STAFResult(STAFResult.Ok, help);
    return result;
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: init                                                              */
/* Description: required by interface STAFServiceInterfaceLevel4             */
/*              creates parsers and registers with STAF                      */
/*                                                                           */
/*****************************************************************************/

public int init(STAFServiceInterfaceLevel4.InitInfo initInfo) 
{
    /* Generate STAF Parsers */

    /* Generate PROCESSFILE parser */

    procFileParser = new STAFCommandParser();
    
    procFileParser.addOption(PROCESSFILE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    procFileParser.addOption(MODE, 1, STAFCommandParser.VALUEREQUIRED);
    procFileParser.addOption(FILE1, 1, STAFCommandParser.VALUEREQUIRED);
    procFileParser.addOption(FILE2, 1, STAFCommandParser.VALUEREQUIRED);
    procFileParser.addOption(SORT, 1, STAFCommandParser.VALUENOTALLOWED);
    procFileParser.addOption(SAVEONFAILURE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    
    procFileParser.addOptionGroup(FILE1, 1, 1);
    procFileParser.addOptionGroup(FILE2, 1, 1);

    /* Generete COMPAREDIR parser */

    compDirParser = new STAFCommandParser();
    compDirParser.addOption(COMPAREDIR, 1, STAFCommandParser.VALUENOTALLOWED);
    compDirParser.addOption(DIR, 1, STAFCommandParser.VALUEREQUIRED);
    compDirParser.addOption(FILE, 0, STAFCommandParser.VALUEREQUIRED);
    compDirParser.addOption(EXISTS, 1, STAFCommandParser.VALUENOTALLOWED);
    compDirParser.addOption(ATTEMPTS, 1, STAFCommandParser.VALUEREQUIRED);
    compDirParser.addOption(POLL, 1, STAFCommandParser.VALUEREQUIRED);

    compDirParser.addOptionGroup(DIR, 1, 1);
    compDirParser.addOptionGroup(FILE, 1, 1);

    compDirParser.addOptionNeed(POLL, ATTEMPTS);

    /* Generate FILECONTAINS parser */

    fileContainsParser = new STAFCommandParser();

    fileContainsParser.addOption(FILECONTAINS, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    fileContainsParser.addOption(FILE, 1, STAFCommandParser.VALUEREQUIRED);
    fileContainsParser.addOption(STRING, 0, STAFCommandParser.VALUEREQUIRED);
    fileContainsParser.addOption(NOT, 1, STAFCommandParser.VALUENOTALLOWED);
    fileContainsParser.addOption(IGNORECASE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    fileContainsParser.addOption(SAVEONFAILURE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    
    fileContainsParser.addOptionGroup(FILE, 1, 1);
    fileContainsParser.addOptionGroup(STRING, 1, 1);

    /* Generate LINECONTAINS parser */

    lineContainsParser = new STAFCommandParser();

    lineContainsParser.addOption(LINECONTAINS, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    lineContainsParser.addOption(FILE, 1, STAFCommandParser.VALUEREQUIRED);
    lineContainsParser.addOption(STRING, 0, STAFCommandParser.VALUEREQUIRED);
    lineContainsParser.addOption(IGNORECASE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    lineContainsParser.addOption(SAVEONFAILURE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    lineContainsParser.addOption(LINENUMBER, 1, 
        STAFCommandParser.VALUEREQUIRED);
    
    lineContainsParser.addOptionGroup(FILE, 1, 1);
    lineContainsParser.addOptionGroup(STRING, 1, 1);

    /* Generate WAITFORFILE parser */

    waitForFileParser = new STAFCommandParser();

    waitForFileParser.addOption(WAITFORFILE, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    waitForFileParser.addOption(FILE, 1, STAFCommandParser.VALUEREQUIRED);
    waitForFileParser.addOption(TIMEOUT, 1, STAFCommandParser.VALUEREQUIRED);
    waitForFileParser.addOption(POLL, 1, STAFCommandParser.VALUEREQUIRED);
    waitForFileParser.addOption(NOT, 1, STAFCommandParser.VALUENOTALLOWED);

    waitForFileParser.addOptionGroup(FILE, 1, 1);

    /* Generate SUBSTITUTE parser */

    substituteParser = new STAFCommandParser();

    substituteParser.addOption(SUBSTITUTE, 1, STAFCommandParser.VALUENOTALLOWED);
    substituteParser.addOption(FILE, 1, STAFCommandParser.VALUEREQUIRED);
    substituteParser.addOption(TOFILE, 1, STAFCommandParser.VALUEREQUIRED);
    substituteParser.addOption(TOMACHINE, 1, STAFCommandParser.VALUEREQUIRED);
    substituteParser.addOption(FAILIFEXISTS, 1, STAFCommandParser.VALUENOTALLOWED);
    substituteParser.addOption(FAILIFNEW, 1, STAFCommandParser.VALUENOTALLOWED);

    substituteParser.addOptionGroup(FILE, 1, 1);
    substituteParser.addOptionGroup(FAILIFEXISTS+" "+FAILIFNEW, 0, 1);

    /* Set name */

    sServiceName = initInfo.name;
    
    /* Create STAFHandle */

    try 
    {
        sHandle = new STAFHandle(initInfo.name);
    }
    catch(STAFException se) 
    {
        se.printStackTrace();
        return STAFResult.STAFRegistrationError;
    }

    /* Register RCs with the HELP service */
    
    int rc = this.registerHelp(initInfo.name);

    /* Finished */

    return rc;
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: lineContains                                                      */
/* Description: performs lineContains service function                       */
/* Parameters: reqInfo - RequestInfo passed to acceptRequest                 */
/* Returns: STAFResult                                                       */
/*                                                                           */
/*****************************************************************************/

private STAFResult lineContains(STAFServiceInterfaceLevel4.RequestInfo reqInfo) 
{    
    /* Check for trust level of 3 or greater */

    if (reqInfo.trustLevel < 3) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel, 
            "Minimum trust level of 3 required.");
    }

    /* Parse request string */

    STAFCommandParseResult pResult = lineContainsParser.parse(reqInfo.request);
    
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    if (reqInfo.diagEnabled == 1)
    {
        // Record diagnostics because the output for a LINECONTAINS request
        // will change in STAF V3.0

        String trigger = new String("STAF/V3.0-Mig/FSEXT LINECONTAINS");
        String source  = new String(reqInfo.processName + ";" +
                                    reqInfo.machine + ";" +
                                    reqInfo.handle);
        sHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "DIAG",
            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
            " SOURCE " + STAFUtil.wrapData(source));
    }

    /* Get FILE and STRING parameters */
    
    Vector stringList = new Vector();
    File file;
    
    try 
    {
        String filename = ServiceUtilities.getResolvedParm(reqInfo, pResult, FILE, sHandle);
        file = new File(filename);

        /* Get all instances of STRING and store in a Vector */
        
        int numStrings = pResult.optionTimes(STRING);
        
        for(int i=1; i <= numStrings; i++) 
        {
            String tempString = ServiceUtilities.getResolvedParm(reqInfo, pResult, STRING, i, sHandle);
            stringList.add(tempString);
        }
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, se.toString());
    }

    /* Parse IGNORECASE parameter */

    boolean ignoreCase = false;
    
    if (pResult.optionTimes(IGNORECASE) == 1) 
    {
        ignoreCase = true;
    }

    /* Parse SAVEONFAILURE parameter */

    boolean saveOnFail = false;
    
    if (pResult.optionTimes(SAVEONFAILURE) == 1) 
    {
        saveOnFail = true;
    }

    /* Parse LINENUMBER parameter */

    boolean strictLine = false;
    int specificLineNumber = 0;
    
    if (pResult.optionTimes(LINENUMBER) == 1) 
    {
        strictLine = true;
        try 
        {
            String lineNumberString = ServiceUtilities.getResolvedParm(
                reqInfo, pResult, LINENUMBER, sHandle);
            specificLineNumber = Integer.parseInt(lineNumberString);
            
            if (specificLineNumber == 0) 
            {// 0 is not valid
                throw new NumberFormatException();
            }
        }
        catch(STAFException se) 
        {        
            return new STAFResult(STAFResult.InvalidRequestString, se.toString());
        }
        catch(NumberFormatException nfe) 
        {
            return new STAFResult(STAFResult.InvalidRequestString, 
                "The value specified for LINENUMBER is invalid.");
        }
    }

    /* Read file one line at a time and search each line for string list */

    BufferedReader in = null;
    int numStrings = stringList.size();
    int lineNumber = 0;
    Vector lineStore = new Vector();
    
    try 
    {
        in = new BufferedReader(new FileReader(file));
        
        while(true) 
        {
            lineNumber++; //inc lineNumber
            
            /* Read next line from file */  
            String line = in.readLine();
            
            if (line == null) 
            { // reached end of file
                break;
            }
            
            if (ignoreCase) 
            {
                line = line.toLowerCase();
            }
            
            if (strictLine) 
            {//simply store line for later comparison
                lineStore.add(line);
                continue;
            }

            boolean searchResult = ServiceUtilities.lineContainsStrings(line, 
                stringList, ignoreCase);
            
            if (searchResult) 
            { //found all strings in a single line
                return new STAFResult(STAFResult.Ok, lineNumber+":"+line);
            }
        }
    }
    catch(IOException ioe) 
    {
        return new STAFResult(IOERROR, ioe.toString());
    }
    finally
    {
        try 
        {
            in.close();
        }
        catch(Throwable t) {} //don't care
    }

    STAFResult result = new STAFResult(STRINGNOTFOUND); //default from here

    if (strictLine) 
    {// now do comparison
        int totalLines = lineStore.size();
        
        if (Math.abs(specificLineNumber) > totalLines) 
        {// line not in file
            result = new STAFResult(LINEDOESNOTEXIST, 
                String.valueOf(totalLines));
        }
        
        if (specificLineNumber > 0) 
        {
            String line = (String) lineStore.get(specificLineNumber-1);
            
            if (ServiceUtilities.lineContainsStrings(line, 
                stringList, ignoreCase))
            {
                return new STAFResult(STAFResult.Ok, 
                    specificLineNumber+":"+line);
            }
            else 
            {
                result = new STAFResult(STRINGNOTFOUND, 
                    specificLineNumber+":"+line);
            }
        }
        else 
        {
            String line = 
                (String) lineStore.get(totalLines+specificLineNumber);
                
            if (ServiceUtilities.lineContainsStrings(line, 
                stringList, ignoreCase)) 
            {
                return new STAFResult(STAFResult.Ok, 
                    (totalLines+specificLineNumber+1)+":"+line);
            }
            else 
            {
                result = new STAFResult(STRINGNOTFOUND, 
                    (totalLines+specificLineNumber+1)+":"+line);
            }
        }
    }

    /* If we get here query failed */

    if (saveOnFail) 
    {//Capture the file for later comparison
        ExtFile capture = new ExtFile(file);
        try 
        {
            capture.captureFile(file.getPath()+".fail");
        }
        catch(IOException ioe) 
        {
            return new STAFResult(IOERROR, ioe.toString());
        }
    }
    
    return result;
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: processFile                                                       */
/* Description: performs processFile service function                        */
/* Parameters: reqInfo - RequestInfo passed to acceptRequest                 */
/* Returns: STAFResult                                                       */
/*                                                                           */
/*****************************************************************************/

private STAFResult processFile(STAFServiceInterfaceLevel4.RequestInfo reqInfo) {
    
    /* Check for trust level of 3 or greater */

    if (reqInfo.trustLevel < 3) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel, 
            "Minimum trust level of 3 required.");
    }
    
    /* Parse request string */
    
    STAFCommandParseResult pResult = procFileParser.parse(reqInfo.request);

    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    /* Get parameters */

    boolean compare = true; //if true compare, if false capture
    boolean sort = false;
    String mode, file1, file2;

    /* Get FILE names */

    try 
    {
        file1 = ServiceUtilities.getResolvedParm(reqInfo,  
            pResult, FILE1, sHandle);
        file2 = ServiceUtilities.getResolvedParm(reqInfo, 
            pResult, FILE2, sHandle);
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, se.toString());
    }
    
    /* Get MODE */
    
    if (pResult.optionTimes(MODE) == 1) 
    {
        try 
        {
            mode = ServiceUtilities.getResolvedParm(reqInfo, 
                pResult, MODE, sHandle);
        }
        catch (STAFException se) 
        {
            return new STAFResult(STAFResult.InvalidRequestString, 
                se.toString());
        }
        
        if (mode.equalsIgnoreCase(CAPTURE)) 
        {
            compare = false;
        }
        else if (mode.equalsIgnoreCase(COMPARE)) 
        {
            compare = true;
        }
        else 
        {
            return new STAFResult(STAFResult.InvalidRequestString, 
                "Invalid value specified for MODE");
        }
    }

    /* Get SORT */

    if (pResult.optionTimes(SORT) == 1) 
    {
        sort = true;
    }

    /* Parse SAVEONFAILURE parameter */

    boolean saveOnFail = false;
    
    if (pResult.optionTimes(SAVEONFAILURE) == 1) 
    {
        saveOnFail = true;
    }

    /* Perform capture or compare */

    ExtFile sourceFile = new ExtFile(file1);
    sourceFile.setSaveFailures(saveOnFail);

    try 
    {
        if (compare) 
        {
            if (!sourceFile.compareFile(file2, sort)) 
            {
                return new STAFResult(COMPAREFAIL);
            }
        }
        else 
        {
            sourceFile.captureFile(file2);
        }
    }
    catch(IOException ioe) 
    {
        return new STAFResult(IOERROR, ioe.toString());
    }

    return new STAFResult(STAFResult.Ok);

}

/*****************************************************************************/
/*                                                                           */
/* Method: registerHelp                                                      */
/* Description: registers all FSEXT return codes with the HELP service       */
/* Parameters: serviceName - name which this service is registered as        */
/* Returns: int representation of STAFResult                                 */
/*                                                                           */
/*****************************************************************************/

private int registerHelp(String serviceName) 
{
    try 
    {
        String request = "REGISTER SERVICE "+serviceName+" ERROR "+COMPAREFAIL+
          " INFO "+COMPAREFAILInfo+" DESCRIPTION "+COMPAREFAILDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+NOTDIRECTORY+
          " INFO "+NOTDIRECTORYInfo+" DESCRIPTION "+NOTDIRECTORYDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+FILENOTFOUND+
          " INFO "+FILENOTFOUNDInfo+" DESCRIPTION "+FILENOTFOUNDDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+FILEEXISTS+
          " INFO "+FILEEXISTSInfo+" DESCRIPTION "+FILEEXISTSDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+EXTRAFILES+
          " INFO "+EXTRAFILESInfo+" DESCRIPTION "+EXTRAFILESDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+STRINGNOTFOUND+
          " INFO "+STRINGNOTFOUNDInfo+" DESCRIPTION "+STRINGNOTFOUNDDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+STRINGFOUND+
          " INFO "+STRINGFOUNDInfo+" DESCRIPTION "+STRINGFOUNDDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+LINEDOESNOTEXIST+
          " INFO "+LINEDOESNOTEXISTInfo+" DESCRIPTION "+LINEDOESNOTEXISTDesc;
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+serviceName+" ERROR "+IOERROR+
          " INFO "+IOERRORInfo+" DESCRIPTION "+IOERRORDesc;
        sHandle.submit("LOCAL", "HELP", request);
        
    }
    catch(STAFException se) 
    {
        return se.rc;
    }

    return STAFResult.Ok;
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: substitute                                                        */
/* Description: performs substitute service function                         */
/* Parameters: reqInfo - RequestInfo passed to acceptRequest                 */
/* Returns: STAFResult                                                       */
/*                                                                           */
/*****************************************************************************/
 
private STAFResult substitute(STAFServiceInterfaceLevel4.RequestInfo reqInfo) 
{
    /* Check for trust level of 4 or greater */

    if(reqInfo.trustLevel < 4) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel, 
            "Minimum trust level of 4 required.");
    }

    /* Parse request string */

    STAFCommandParseResult pResult = substituteParser.parse(reqInfo.request);
    
    if (pResult.rc != 0) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, 
            pResult.errorBuffer);
    }

    /* Get source file */

    String inputFile;
	
    try 
    {
        inputFile = ServiceUtilities.getResolvedParm(reqInfo, 
            pResult, FILE, sHandle);
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, se.toString());
    }

    /* Read file and resolve VARS then write out to tempFile */

    BufferedReader in = null;
    BufferedWriter out = null;
    File tempFile = null;

    try 
    {
        tempFile = File.createTempFile("fsext", null);
        in = new BufferedReader(new FileReader(inputFile));
        out = new BufferedWriter(new FileWriter(tempFile));
	
        while(true) 
        { 
            //outside loop to parse file line by line
            
            String line = in.readLine();
            
            if (line == null) 
            {
                //end of file

                break;
            }
             			
            while(true) 
            { 
                //inside loop to resolve multiple vars per line
                int beginIndex = line.indexOf("%{");
                
                if (beginIndex == -1) 
                {
                    //no vars to resolve
                    
                    break;
                } 
                
                int endIndex = line.indexOf("}%", beginIndex);
                String varToResolve = line.substring(beginIndex+1, endIndex+1);
                String resolvedVar;

                // Resolve using handle var if local request 
                // otherwise use global
				
                if (reqInfo.localMachine.equals(reqInfo.machine)) 
                {
                    resolvedVar = sHandle.submit("LOCAL", "VAR", "HANDLE " + 
                        String.valueOf(reqInfo.handle) + " RESOLVE " + 
                        varToResolve);
                }
                else 
                {
                    resolvedVar = sHandle.submit("LOCAL", "VAR", 
                        "GLOBAL RESOLVE " + varToResolve);
                }
				
                StringBuffer modString = new StringBuffer(line);
                modString = modString.replace(beginIndex, endIndex+2, 
                    resolvedVar);
                line = modString.toString();
            }
            
            out.write(line);
            out.newLine();
        }
    }
    catch(STAFException se) 
    {
        return new STAFResult(se.rc, se.getMessage());
    }
    catch(IOException ioe) 
    {
        return new STAFResult(IOERROR, ioe.getMessage());
    }
    finally 
    {
        try 
        {
            out.close();
            in.close();
        }
        catch(Exception e) {} //don't care
    }

    /* Resolve variables that will be required to call FS & generate FS request String */

    StringBuffer fsReq = new StringBuffer("COPY FILE "+tempFile.getPath());
    boolean toFileSet = false;
	
    if ((pResult.optionTimes(TOFILE)) == 1) 
    {
        fsReq.append(" TOFILE "+pResult.optionValue(TOFILE));
        toFileSet = true;
    }
    
    if ((pResult.optionTimes(TOMACHINE)) == 1) 
    {
        fsReq.append(" TOMACHINE "+pResult.optionValue(TOMACHINE));
        toFileSet = true;
    }
    
    if (!toFileSet) 
    {
        fsReq.append(" TOFILE "+inputFile);
    }
    
    if ((pResult.optionTimes(FAILIFEXISTS)) == 1) 
    {
        fsReq.append(" FAILIFEXISTS");    
    }
          
    if ((pResult.optionTimes(FAILIFNEW)) == 1) 
    {
        fsReq.append(" FAILIFNEW");
    }

    /* Send STAF FS Command */

    try 
    {
        String result = sHandle.submit("LOCAL", "FS", fsReq.toString());
        return new STAFResult(STAFResult.Ok, result);
    }
    catch(STAFException se) 
    {
        return new STAFResult(se.rc, se.getMessage());
    }
    finally 
    {
        tempFile.delete();
    }
}

/*****************************************************************************/
/*                                                                           */
/* Method: term                                                              */
/* Description: required by interface STAFServiceInterfaceLevel4             */
/*                                                                           */
/*****************************************************************************/

public int term() 
{
    // Un-register the service handle

    try
    {
        sHandle.unRegister();
    }
    catch (STAFException ex)
    {
        return STAFResult.STAFRegistrationError;
    }

    return 0;
}

/*****************************************************************************/
/*                                                                           */
/* Method: version                                                           */
/* Description: returns service version information                          */
/* Parameters: none                                                          */
/* Returns: STAFResult.OK                                                    */
/*                                                                           */
/*****************************************************************************/

private STAFResult version() 
{
    STAFResult result = new STAFResult(STAFResult.Ok, FSEXTVERSION);
    return result;
    
}

/*****************************************************************************/
/*                                                                           */
/* Method: waitForFile                                                       */
/* Description: performs waitForFile service function                        */
/* Parameters: reqInfo - RequestInfo passed to acceptRequest                 */
/* Returns: STAFResult                                                       */
/*                                                                           */
/*****************************************************************************/

private STAFResult waitForFile(STAFServiceInterfaceLevel4.RequestInfo reqInfo)
{    
    /* Check for trust level of 3 or greater */

    if (reqInfo.trustLevel < 3) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel, 
            "Minimum trust level of 3 required.");
    }

    /* Parse request string */

    STAFCommandParseResult pResult = waitForFileParser.parse(reqInfo.request);
    
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    /* Parse FILE, POLL, & TIMEOUT parameters */
    
    File file;
    String poll = "15000"; //default value
    long timeout = 
        Long.MAX_VALUE; // set to max value, for default infinite (almost) wait
    
    try 
    {
        String filename = ServiceUtilities.getResolvedParm(reqInfo, 
            pResult, FILE, sHandle);
        file = new File(filename);
        
        if (pResult.optionTimes(TIMEOUT) == 1) 
        {
            timeout = Long.parseLong(ServiceUtilities.getResolvedParm(
                reqInfo, pResult, TIMEOUT, sHandle));
        }
        if (pResult.optionTimes(POLL) == 1) 
        {
            poll = ServiceUtilities.getResolvedParm(reqInfo, 
                pResult, POLL, sHandle);
        }
    }
    catch(NumberFormatException nfe) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, nfe.toString());
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, se.toString());
    }

    /* Get not parameter */
    
    boolean not = false;
    
    if (pResult.optionTimes(NOT) == 1) 
    {
        not = true;
    }

    /* Get starting time */

    long start = (new Date()).getTime();
    Date current;

    /* Perform appropriate file check, if failed wait POLL ms and try again */
    /* Quit after TIMEOUT ms has passed */

    do 
    {    
        if ((file.exists() && !not) || (!file.exists() && not)) 
        {
            return new STAFResult(STAFResult.Ok);
        }
        
        sHandle.submit2("LOCAL", "DELAY", "DELAY "+poll);
        current = new Date();       
        
    }
    while ((current.getTime() - start) < timeout);

    /* If we get here timed out without finding file */

    if (not) 
    {
        return new STAFResult(FILEEXISTS, file.getPath());
    }
    else 
    {
        return new STAFResult(FILENOTFOUND, file.getPath());
    }
    
}
}
