/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAFOSTypes.h"
#include <map>
#include "STAF_fstream.h"
#include "STAF_iostream.h"
#include "STAFString.h"
#include "STAFError.h"
#include "STAFException.h"
#include "STAFRefPtr.h"
#include "STAFMutexSem.h"
#include "STAFCommandParser.h"
#include "STAFServiceInterface.h"
#include "STAFTimestamp.h"
#include "STAFUtil.h"
#include "STAFInternalUtil.h"
#include "STAFMonitorService.h"

typedef STAFRefPtr<STAFCommandParser> STAFCommandParserPtr;

typedef std::map<STAFHandle_t, STAFString> HandleMap;

struct MachineData
{
    STAFString realMachineName;
    HandleMap handleMap;
};

typedef std::map<STAFString, MachineData> MachineMap;

struct MonitorServiceData
{
    unsigned int fDebugMode;
    STAFString fShortName;
    STAFString fName;
    STAFHandlePtr fHandlePtr;
    unsigned int fResolveMessage;
    unsigned int fMaxRecordSize;    
    unsigned int fOldReturnCodes;    
    unsigned int fEnableResolveMessageVar;
    STAFCommandParserPtr fLogParser;
    STAFCommandParserPtr fQueryParser;
    STAFCommandParserPtr fListParser;
    STAFCommandParserPtr fDeleteParser;
    STAFCommandParserPtr fSetParser;
    STAFMutexSemPtr fMapSem;
    MachineMap fMap;
};

typedef STAFRefPtr<MonitorServiceData> MonitorServiceDataPtr;

static STAFString lineSep;
static const STAFString monitorVersion("2.1.3");
static const STAFString sMachine("MACHINE");
static const STAFString sHandle("HANDLE");
static const STAFString sResolveMessage("RESOLVEMESSAGE");
static const STAFString sNoResolveMessage("NORESOLVEMESSAGE");
static const STAFString sOldReturnCodes("OLDRETURNCODES");
static const STAFString sNewReturnCodes("NEWRETURNCODES");
static const STAFString sMaxRecordSize("MAXRECORDSIZE");
static const STAFString sEnableResolveMessageVar("ENABLERESOLVEMESSAGEVAR");
static const STAFString sDisableResolveMessageVar("DISABLERESOLVEMESSAGEVAR");
static const STAFString sLeftCurlyBrace(kUTF8_LCURLY);
static const STAFString sSlash(kUTF8_SLASH);
static const STAFString sSpace(kUTF8_SPACE);
static const STAFString sColon(kUTF8_COLON);
static const STAFString sEqual(kUTF8_EQUAL);
static const STAFString sHandleList("Handle");
static const STAFString sSize("Size");
static const STAFString sUsing("Using ");
static const STAFString sToday("today");
static const STAFString sBefore("before");
static const STAFString sMessage("message");
static const STAFString sLocal("local");
static const STAFString sSettings("settings");
static const STAFString sMachines("machines");
static const STAFString sVar("var");
static const STAFString sHelp("help");
static const STAFString sResStrHandle("HANDLE ");
static const STAFString sResStrResolve(" RESOLVE ");
static const STAFString sResStrGResolve("GLOBAL RESOLVE ");

static STAFResultPtr handleLog(STAFServiceRequestLevel2 *, MonitorServiceData *);
static STAFResultPtr handleQuery(STAFServiceRequestLevel2 *,
                                 MonitorServiceData *);
static STAFResultPtr handleList(STAFServiceRequestLevel2 *,
                                MonitorServiceData *);
static STAFResultPtr handleDelete(STAFServiceRequestLevel2 *,
                                  MonitorServiceData *);
static STAFResultPtr handleSet(STAFServiceRequestLevel2 *,
                               MonitorServiceData *);
static STAFResultPtr handleHelp(STAFServiceRequestLevel2 *,
                                MonitorServiceData *);
static STAFResultPtr handleVersion(STAFServiceRequestLevel2 *,
                                   MonitorServiceData *);
static STAFResultPtr resolveStr(STAFServiceRequestLevel2 *pInfo, 
                                MonitorServiceData *pData,
                                const STAFString &theString);
static STAFResultPtr resolveOp(STAFServiceRequestLevel2 *pInfo, 
                               MonitorServiceData *pData,
                               STAFCommandParseResultPtr &parsedResult,
                               const STAFString &fOption,
                               unsigned int optionIndex = 1);
static void registerHelpData(MonitorServiceData *pData,
                             unsigned int errorNumber,
                             const STAFString &shortInfo,
                             const STAFString &longInfo);
static void unregisterHelpData(MonitorServiceData *pData,
                               unsigned int errorNumber);

STAFRC_t STAFServiceGetLevelBounds(unsigned int levelID,
                                   unsigned int *minimum,
                                   unsigned int *maximum)
{
    switch (levelID)
    {
        case kServiceInfo:
        case kServiceInit:
        {
            *minimum = 1;
            *maximum = 1;
            break;
        }
        case kServiceAcceptRequest:
        {
            *minimum = 2;
            *maximum = 2;
            break;
        }
        case kServiceTerm:
        case kServiceDestruct:
        {
            *minimum = 0;
            *maximum = 0;
            break;
        }
        default:
        {
            return kSTAFInvalidAPILevel;
        }
    }

    return kSTAFOk;
}


STAFRC_t STAFServiceConstruct(STAFServiceHandle_t *pServiceHandle,
                              void *pServiceInfo, unsigned int infoLevel,
                              STAFString_t *pErrorBuffer)
{
    STAFRC_t rc = kSTAFUnknownError;

    try
    {
        if (infoLevel != 1) return kSTAFInvalidAPILevel;

        STAFServiceInfoLevel1 *pInfo =
            reinterpret_cast<STAFServiceInfoLevel1 *>(pServiceInfo);

        MonitorServiceData data;
        data.fDebugMode = 0;
        data.fMaxRecordSize = 1024;
        data.fResolveMessage = 0;        
        data.fOldReturnCodes = 0;        
        data.fEnableResolveMessageVar = 0;
        data.fShortName = pInfo->name;
        data.fName = "STAF/Service/";
        data.fName += pInfo->name;
        
        for (unsigned int i = 0; i < pInfo->numOptions; ++i)
        {            
            if (STAFString(pInfo->pOptionName[i]).upperCase() == "DEBUG")
            {
                data.fDebugMode = 1;
            }           
            else
            {
                STAFString optionError(pInfo->pOptionName[i]);
                *pErrorBuffer = optionError.adoptImpl();
                return kSTAFServiceConfigurationError;
            }
        }        

        // Set service handle

        *pServiceHandle = new MonitorServiceData(data);

        return kSTAFOk;
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                "STAFMonitorSerice.cpp: STAFServiceConstruct").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFMonitorService.cpp: STAFServiceConstruct: Caught "
                    "unknown exception in STAFServiceAcceptRequest()");
        *pErrorBuffer = error.adoptImpl();
    }    

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceInit(STAFServiceHandle_t serviceHandle,
                         void *pInitInfo, unsigned int initLevel,
                         STAFString_t *pErrorBuffer)
{
    STAFRC_t retCode = kSTAFUnknownError;

    try
    {    
        if (initLevel != 1) return kSTAFInvalidAPILevel;

        MonitorServiceData *pData =
            reinterpret_cast<MonitorServiceData *>(serviceHandle);
        
        STAFServiceInitLevel1 *pInfo =
            reinterpret_cast<STAFServiceInitLevel1 *>(pInitInfo);        

        retCode = STAFHandle::create(pData->fName, pData->fHandlePtr);
        
        if (retCode != kSTAFOk)
            return retCode;

        pData->fMapSem = STAFMutexSemPtr(new STAFMutexSem,
                                         STAFMutexSemPtr::INIT);    
        
        //LOG options
        pData->fLogParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fLogParser->addOption("LOG",              1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fLogParser->addOption("MESSAGE",          1,
                                     STAFCommandParser::kValueRequired);
        pData->fLogParser->addOption("RESOLVEMESSAGE",   1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fLogParser->addOption("NORESOLVEMESSAGE", 1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fLogParser->addOptionNeed("LOG", "MESSAGE");
        pData->fLogParser->addOptionGroup("RESOLVEMESSAGE NORESOLVEMESSAGE",
                                          0, 1);

        //QUERY options
        pData->fQueryParser = STAFCommandParserPtr(new STAFCommandParser,
                                                   STAFCommandParserPtr::INIT);
        pData->fQueryParser->addOption("QUERY",   1,
                                       STAFCommandParser::kValueNotAllowed);
        pData->fQueryParser->addOption("MACHINE", 1,
                                       STAFCommandParser::kValueRequired);
        pData->fQueryParser->addOption("HANDLE",  1,
                                       STAFCommandParser::kValueRequired);                
        pData->fQueryParser->addOptionGroup("MACHINE HANDLE", 2, 2);



        //LIST options
        pData->fListParser = STAFCommandParserPtr(new STAFCommandParser,
                                                  STAFCommandParserPtr::INIT);
        pData->fListParser->addOption("LIST",     1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOption("MACHINES", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOption("MACHINE",  1,
                                      STAFCommandParser::kValueRequired);
        pData->fListParser->addOption("SETTINGS", 1,
                                      STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOptionGroup("MACHINES MACHINE SETTINGS", 1, 1);


        //DELETE options
        pData->fDeleteParser = STAFCommandParserPtr(new STAFCommandParser,
                                                    STAFCommandParserPtr::INIT);
        pData->fDeleteParser->addOption("DELETE",  1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOption("CONFIRM", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOption("BEFORE",  1,
                                        STAFCommandParser::kValueRequired);

        pData->fDeleteParser->addOptionNeed("DELETE", "CONFIRM");

        //SET options
        pData->fSetParser = STAFCommandParserPtr(new STAFCommandParser,
                            STAFCommandParserPtr::INIT);
        pData->fSetParser->addOption("SET",                 1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOption("NORESOLVEMESSAGE",    1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOption("RESOLVEMESSAGE",      1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOptionGroup("NORESOLVEMESSAGE RESOLVEMESSAGE",
                                          0, 1);
        pData->fSetParser->addOption("OLDRETURNCODES",    1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOption("NEWRETURNCODES",      1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOptionGroup("OLDRETURNCODES NEWRETURNCODES",
                                          0, 1);           
        pData->fSetParser->addOption("MAXRECORDSIZE",    1,
                                     STAFCommandParser::kValueRequired);
        pData->fSetParser->addOption("ENABLERESOLVEMESSAGEVAR",    1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOption("DISABLERESOLVEMESSAGEVAR",      1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fSetParser->addOptionGroup(
                        "ENABLERESOLVEMESSAGEVAR DISABLERESOLVEMESSAGEVAR",
                        0, 1);

        // PARMS        
        STAFCommandParseResultPtr parsedParmsResult =
            pData->fSetParser->parse(pInfo->parms);

        if (parsedParmsResult->rc != kSTAFOk) 
        {
            return parsedParmsResult->rc;
        }
        
        if (parsedParmsResult->optionTimes(sResolveMessage))
        {
            pData->fResolveMessage = 1;
        }

        if (parsedParmsResult->optionTimes(sNoResolveMessage))
        {
            pData->fResolveMessage = 0;
        }
        
        if (parsedParmsResult->optionTimes(sOldReturnCodes))
        {
            pData->fOldReturnCodes = 1;
        }

        if (parsedParmsResult->optionTimes(sNewReturnCodes))
        {
            pData->fOldReturnCodes = 0;
        }

        if (parsedParmsResult->optionTimes(sMaxRecordSize))
        {
            pData->fMaxRecordSize = parsedParmsResult->optionValue(
                "MAXRECORDSIZE", 1).asUInt();
        }        
        
        if (parsedParmsResult->optionTimes(sEnableResolveMessageVar))
        {
            pData->fEnableResolveMessageVar = 1;
        }

        if (parsedParmsResult->optionTimes(sDisableResolveMessageVar))
        {
            pData->fEnableResolveMessageVar = 0;
        }

        // Get line separator
        STAFString getRequest = "RESOLVE {STAF/Config/Sep/Line}";
        STAFResultPtr result = pData->fHandlePtr->submit("LOCAL", "VAR",
                                                         getRequest);
        lineSep = result->result;

        STAFString invalidDirMessage("The STAF/Service/");

        invalidDirMessage += pData->fShortName;
        invalidDirMessage += "/Directory variable in the STAF configuration "
                             "file is{STAF/Config/Sep/Line}invalid";

        registerHelpData(pData, kSTAFMonitorInvalidDirectory,
            STAFString("Invalid monitor directory (deprecated)"),
            invalidDirMessage);

        registerHelpData(pData, kSTAFMonitorCreateDirectoryError,
            STAFString("Error creating directory (deprecated)"),
            STAFString("Unable to create the log file directory"));

        registerHelpData(pData, kSTAFMonitorInvalidFileFormat,
            STAFString("Invalid file format (deprecated)"),
            STAFString("An invalid/unknown record format was encountered while "
                       "reading the monitor{STAF/Config/Sep/Line}file"));
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                        "STAFMonitorSerice.cpp: STAFServiceInit").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFMonitorService.cpp: STAFServiceInit: Caught unknown "
                    "exception in STAFServiceAcceptRequest()");
        *pErrorBuffer = error.adoptImpl();
    }    

    return retCode;
}


STAFRC_t STAFServiceAcceptRequest(STAFServiceHandle_t serviceHandle,
                                  void *pRequestInfo, unsigned int reqLevel,
                                  STAFString_t *pResultBuffer)
{
    if (reqLevel != 2) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        STAFResultPtr result(new STAFResult(kSTAFOk, STAFString()),
                             STAFResultPtr::INIT);        

        STAFServiceRequestLevel2 *pInfo =
            reinterpret_cast<STAFServiceRequestLevel2 *>(pRequestInfo);

        MonitorServiceData *pData =
            reinterpret_cast<MonitorServiceData *>(serviceHandle);

        STAFString request(pInfo->request);
        STAFString action = request.subWord(0, 1).toLowerCase();
        
        //call functions for the request
        if (action == "log")
            result = handleLog(pInfo, pData);
        else if (action == "query")
            result = handleQuery(pInfo, pData);
        else if (action == "list")
            result = handleList(pInfo, pData);
        else if (action == "delete")
            result = handleDelete(pInfo, pData);
        else if (action == "set")
            result = handleSet(pInfo, pData);
        else if (action == "help")
            result = handleHelp(pInfo, pData);
         else if (action == "version")
            result = handleVersion(pInfo, pData);
        else
            result = STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                                   request.subWord(0, 1)), 
                                   STAFResultPtr::INIT);

        *pResultBuffer = result->result.adoptImpl();
        retCode = result->rc;
    }
    catch (STAFException &e)
    { 
        *pResultBuffer = getExceptionString(e,
                         "STAFMonitorService.cpp: "
                         "STAFServiceAcceptRequest").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFMonitorService.cpp: STAFServiceAcceptRequest: Caught "
                    "unknown exception in STAFServiceAcceptRequest()");
        *pResultBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFRC_t STAFServiceTerm(STAFServiceHandle_t serviceHandle,
                         void *pTermInfo, unsigned int termLevel,
                         STAFString_t *pErrorBuffer)
{
    if (termLevel != 0) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        retCode = kSTAFOk;

        MonitorServiceData *pData = 
            reinterpret_cast<MonitorServiceData *>(serviceHandle);

        // Un-register Help Data

        unregisterHelpData(pData, kSTAFMonitorInvalidDirectory);
        unregisterHelpData(pData, kSTAFMonitorCreateDirectoryError);
        unregisterHelpData(pData, kSTAFMonitorInvalidFileFormat);
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                        "STAFMonitorSerice.cpp: STAFServiceTerm").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFMonitorService.cpp: STAFServiceTerm: Caught unknown "
                    "exception in STAFServiceAcceptRequest()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFRC_t STAFServiceDestruct(STAFServiceHandle_t *serviceHandle,
                             void *pDestructInfo, unsigned int destructLevel,
                             STAFString_t *pErrorBuffer)
{
    if (destructLevel != 0) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        MonitorServiceData *pData =
            reinterpret_cast<MonitorServiceData *>(*serviceHandle);

        delete pData;
        *serviceHandle = 0;

        retCode = kSTAFOk;
    }
    catch (STAFException &e)
    { 
        *pErrorBuffer = getExceptionString(e,
                    "STAFMonitorSerice.cpp: STAFServiceDestruct").adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFMonitorService.cpp: STAFServiceDestruct: Caught "
                    "unknown exception in STAFServiceAcceptRequest()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFResultPtr handleLog(STAFServiceRequestLevel2 *pInfo, 
                        MonitorServiceData *pData)
{
    if (pInfo->trustLevel < 3)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult =
        pData->fLogParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }

    STAFString fMessage = STAFTimestamp::now().asString() + sSpace;

    // Determine if we need to resolve the message
    // Start with the PARMS/SET resolve value        
    unsigned int resolveMessage = pData->fResolveMessage;

    if ((parsedResult->optionTimes(sNoResolveMessage)) ||
        (parsedResult->optionTimes(sResolveMessage))) 
    {
        // Request contains [NO]RESOLVEMESSAGE, so use its resolve value
        if (parsedResult->optionTimes(sNoResolveMessage))
            resolveMessage = 0;
        else
            resolveMessage = 1;
    }
    else if (pData->fEnableResolveMessageVar)
    {                 
        // Check the STAF/Service/<Monitor-Name>/ResolveMessage variable
        STAFString varReq = "{STAF/Service/";
        varReq += pData->fShortName;
        varReq += "/ResolveMessage}";        
        
        STAFResultPtr varResult = resolveStr(pInfo, pData, varReq);
        if (varResult->rc == kSTAFOk) 
        {
            // Variable was found, so use its resolve value
            if (!varResult->result.isDigits())
            {
                return STAFResultPtr(new STAFResult(kSTAFInvalidValue, 
                                     STAFString()), STAFResultPtr::INIT);                
            }
            resolveMessage = (varResult->result).asUInt();                        
        }        
    }
    
    if (resolveMessage)
    {
        STAFResultPtr messageResult =
                      resolveOp(pInfo, pData, parsedResult, sMessage);
        
        if (messageResult->rc != 0) return messageResult;

        fMessage += messageResult->result;
    }
    else
    {
        fMessage += parsedResult->optionValue(sMessage);
    }
    
    if (fMessage.length() > pData->fMaxRecordSize) 
    {
        fMessage = fMessage.subString(0, pData->fMaxRecordSize, 
                                      STAFString::kChar);    
    }

    STAFMutexSemLock lock(*pData->fMapSem);
    
    STAFString upperCaseMachine = 
        ((STAFString)(pInfo->effectiveMachine)).toUpperCase();

    //Write message into the log
    if (pData->fMap[upperCaseMachine].realMachineName == "")
    {
        pData->fMap[upperCaseMachine].realMachineName = 
            pInfo->effectiveMachine;
    }
    pData->fMap[upperCaseMachine].handleMap[pInfo->handle] = fMessage;

    return STAFResultPtr(new STAFResult(kSTAFOk, STAFString()),
                         STAFResultPtr::INIT);
}


STAFResultPtr handleQuery(STAFServiceRequestLevel2 *pInfo, 
                          MonitorServiceData *pData)
{
    STAFString result;
    
    STAFRC_t rc = kSTAFOk;

    if (pInfo->trustLevel < 2)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult = 
        pData->fQueryParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
       
    if (pInfo->diagEnabled)
    {
        // Record diagnostics because the output of a QUERY request will
        // change in STAF V3.0

        static STAFString request("RECORD TRIGGER %C SOURCE %C");
        static STAFString trigger("STAF/V3.0-Mig/MONITOR QUERY");
        static STAFString sourceFormat("%s;%s;%d");
        STAFString source = STAFHandle::formatString(
            sourceFormat.getImpl(), pInfo->processName,
            pInfo->machine, pInfo->handle);

        pData->fHandlePtr->submit(
            sLocal, "DIAG",
            STAFHandle::formatString(
                request.getImpl(), trigger.getImpl(), source.getImpl()),
            kSTAFReqFireAndForget);
    }

    MachineMap::iterator machineIterator;
    HandleMap::iterator handleIterator;

    // Get and resolve MACHINE option

    STAFResultPtr res = resolveOp(pInfo, pData, parsedResult, sMachine);
    
    if (res->rc != kSTAFOk) return res;
    
    STAFString queryMachine = res->result;

    // Get and resolve HANDLE option

    res = resolveOp(pInfo, pData, parsedResult, sHandle);

    if (res->rc != kSTAFOk) return res;
    
    STAFString queryHandle = res->result;

    STAFError_e errorCode;

    if (!queryHandle.isDigits())
    {
        result = "Entry for handle " + queryHandle + " must be numeric" +
                 lineSep;
        errorCode = (STAFError_e)kSTAFInvalidValue;
        return STAFResultPtr(new STAFResult(errorCode, result),
                             STAFResultPtr::INIT);
    }
   
    STAFMutexSemLock lock(*pData->fMapSem);

    if ((machineIterator = pData->fMap.find(queryMachine.toUpperCase())) ==
        pData->fMap.end())
    {
        result = "Entry for machine " + STAFString(queryMachine) +
                 " not found" + lineSep;
        if (pData->fOldReturnCodes) 
            errorCode = kSTAFFileOpenError;
        else
            errorCode = (STAFError_e)kSTAFDoesNotExist;
        return STAFResultPtr(new STAFResult(errorCode, result),
                             STAFResultPtr::INIT);
    }
    else if ((handleIterator =
              machineIterator->second.handleMap.find(queryHandle.asUInt())) ==
              machineIterator->second.handleMap.end())
    {
        result = "Entry for handle " + queryHandle + " not found" + lineSep;
        if (pData->fOldReturnCodes) 
            errorCode = kSTAFFileOpenError;
        else
            errorCode = (STAFError_e)kSTAFDoesNotExist;
        return STAFResultPtr(new STAFResult(errorCode, result),
                             STAFResultPtr::INIT);
    }
    else
    {
        result = handleIterator->second;            
    }
   
    return STAFResultPtr(new STAFResult(kSTAFOk, result), 
                         STAFResultPtr::INIT);
}


STAFResultPtr handleList(STAFServiceRequestLevel2 *pInfo, 
                         MonitorServiceData *pData)
{    
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    if (pInfo->trustLevel < 2)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult =
        pData->fListParser->parse(pInfo->request);
    
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }    

    if (pInfo->diagEnabled)
    {
        // Record diagnostics because the output of a LIST request will
        // change in STAF V3.0

        static STAFString request("RECORD TRIGGER %C SOURCE %C");
        static STAFString trigger("STAF/V3.0-Mig/MONITOR LIST");
        static STAFString sourceFormat("%s;%s;%d");
        STAFString source = STAFHandle::formatString(
            sourceFormat.getImpl(), pInfo->processName,
            pInfo->machine, pInfo->handle);

        pData->fHandlePtr->submit(
            sLocal, "DIAG",
            STAFHandle::formatString(
                request.getImpl(), trigger.getImpl(), source.getImpl()),
            kSTAFReqFireAndForget);
    }

    if (parsedResult->optionTimes(sSettings))
    {
        result = "Max Record Size    : " + STAFString(pData->fMaxRecordSize)
                 + lineSep;
                  
        result += "Resolve Message    : ";

        if (pData->fResolveMessage != 0)
            result += "Enabled" + lineSep;
        else
            result += "Disabled" + lineSep;
            
        result += "Return codes       : ";

        if (pData->fOldReturnCodes != 0)
            result += "Old" + lineSep;
        else
            result += "New" + lineSep;
        
        result += "Resolve Message Var: ";

        if (pData->fEnableResolveMessageVar != 0)
            result += "Enabled" + lineSep;
        else
            result += "Disabled" + lineSep;
    }
    else if (parsedResult->optionTimes(sMachines) != 0)
    {
        STAFMutexSemLock lock(*pData->fMapSem);
        MachineMap::iterator machineIterator;

        //Test to see if anything has been logged        
        if (pData->fMap.begin() == pData->fMap.end())
        {
            result = "No monitor data found." + lineSep;
            return STAFResultPtr(new STAFResult(kSTAFDoesNotExist, result),
                                 STAFResultPtr::INIT);
        }
        else
        {
            //Write all the machine names to result
            for(machineIterator = pData->fMap.begin(); 
                machineIterator != pData->fMap.end(); ++machineIterator)
            {
                result += machineIterator->second.realMachineName + lineSep;
            }           
        }
    }

    else if (parsedResult->optionTimes(sMachine) != 0)
    {
        MachineMap::iterator machineIterator;
        STAFMutexSemLock lock(*pData->fMapSem);
        STAFString listMachine = parsedResult->optionValue(sMachine);

        if(listMachine.count(sLeftCurlyBrace) > 0)
        {
            STAFResultPtr machineResult = resolveOp(pInfo, pData, parsedResult,
                                                    "Machine");
            listMachine = machineResult->result;
        }

        if ((machineIterator = pData->fMap.find(listMachine.toUpperCase()))
            == pData->fMap.end())
        {
            result = "Machine " + listMachine + " has not " +
                     "logged any entries with the Monitor Service" + lineSep;            
        }
        else
        {
            HandleMap::iterator handleIterator;
            for(handleIterator = machineIterator->second.handleMap.begin();
                handleIterator != machineIterator->second.handleMap.end();
                ++handleIterator)
            {                
                    //Build the proper result string
                result += STAFString(handleIterator->first);

                if (handleIterator->first <= 9)
                    result += "........................  ";
                else if (handleIterator->first <= 99)
                    result += ".......................  ";
                else if (handleIterator->first <= 999)
                    result += "......................  ";
                else if (handleIterator->first <= 9999)
                    result += ".....................  ";
                else
                    result += "....................  ";
                result += handleIterator->second.subString(0,17) + 
                          "  Size=" +
                          STAFString(handleIterator->second.length()) + 
                          lineSep;        
            }
        }
    }

    return STAFResultPtr(new STAFResult(kSTAFOk, result), 
                         STAFResultPtr::INIT);
}


STAFResultPtr handleDelete(STAFServiceRequestLevel2 *pInfo, 
                           MonitorServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    if (pInfo->trustLevel < 4)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult = 
        pData->fDeleteParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }

    if (pData->fMap.begin() == pData->fMap.end())
    {
        result = "There are no entries in the Monitor Log." + lineSep;
        return STAFResultPtr(new STAFResult(kSTAFOk, result),
                             STAFResultPtr::INIT);
    } 
    else if (parsedResult->optionTimes(sBefore) != 0)
    {        
        MachineMap::iterator machineIterator;
        HandleMap::iterator handleIterator;        
        STAFMutexSemLock lock(*pData->fMapSem);
        STAFString beforeDateString = parsedResult->optionValue(sBefore);                                
        
        try
        {
            if (beforeDateString.isEqualTo(sToday, 
                                           STAFString::kCaseInsensitive))
            {   
                beforeDateString = 
                    (STAFTimestamp::now().asString()).subString(0,8);
            }
           
            STAFTimestamp beforeTimeStamp = STAFTimestamp(beforeDateString);            
           
            for(machineIterator = pData->fMap.begin();
                machineIterator != pData->fMap.end(); ++machineIterator)
            {
                for(handleIterator = machineIterator->second.handleMap.begin();
                    handleIterator != machineIterator->second.handleMap.end();
                    )                
                {                      
                    if (STAFTimestamp(handleIterator->second.subString(0,8)) <
                            beforeTimeStamp)
                     machineIterator->second.handleMap.erase(handleIterator++);
                    else                 
                        handleIterator++;                
                }
            }
        }
        catch(STAFTimestampInvalidDateException &e)
        {
            result = "Invalid Date format.  YYYYMMDD, MM/DD/YY, or ";
            result += "MM/DD/YYYY" + lineSep;
            result += getExceptionString(e) + lineSep;
            return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                 result), STAFResultPtr::INIT);
        }        
        catch (STAFException &e)
        { 
            result = getExceptionString(e,
                    "STAFMonitorSerice.cpp: Delete ").adoptImpl();
            return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                                 result), STAFResultPtr::INIT);                
        }
        catch (...)
        {
            result = "Unknown exception caught in STAFMonitorService ";            
            return STAFResultPtr(new STAFResult(kSTAFUnknownError,
                                 result), STAFResultPtr::INIT);
        }    
    }  
    else
    {
        pData->fMap.clear();
    }
  
    return STAFResultPtr(new STAFResult(kSTAFOk, STAFString()),
                         STAFResultPtr::INIT);
}


STAFResultPtr handleSet(STAFServiceRequestLevel2 *pInfo, 
                        MonitorServiceData *pData)
{
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    if (pInfo->trustLevel < 5)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult = 
        pData->fSetParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }
    
    if (pInfo->diagEnabled)
    {
        // Record diagnostics because the output of a SET request will change
        // in STAF V3.0

        static STAFString request("RECORD TRIGGER %C SOURCE %C");
        static STAFString trigger("STAF/V3.0-Mig/MONITOR SET");
        static STAFString sourceFormat("%s;%s;%d");
        STAFString source = STAFHandle::formatString(
            sourceFormat.getImpl(), pInfo->processName,
            pInfo->machine, pInfo->handle);

        pData->fHandlePtr->submit(
            sLocal, "DIAG",
            STAFHandle::formatString(
                request.getImpl(), trigger.getImpl(), source.getImpl()),
            kSTAFReqFireAndForget);
    }

    if (parsedResult->optionTimes(sMaxRecordSize) != 0)
    {          
        STAFResultPtr maxResult = resolveOp(pInfo, pData, parsedResult,
                                            sMaxRecordSize);

        if (maxResult->rc != kSTAFOk) return maxResult;

        if (!maxResult->result.isDigits())
        {
            maxResult->rc = kSTAFInvalidValue;
            return maxResult;
        }

        pData->fMaxRecordSize = maxResult->result.asUInt();

        result += "Max Record Size    : " + maxResult->result + lineSep;
    }
    
    if (parsedResult->optionTimes(sResolveMessage) != 0)
    {
        pData->fResolveMessage = 1;
        result += "Resolve Message    : Enabled" + lineSep;
    }
    else if (parsedResult->optionTimes(sNoResolveMessage) != 0)
    {
        pData->fResolveMessage = 0;
        result += "Resolve Message    : Disabled" + lineSep;
    }
    
    if (parsedResult->optionTimes(sNewReturnCodes) != 0)
    {
        pData->fOldReturnCodes = 0;
        result += "Return codes       : New" + lineSep;
    }
    else if (parsedResult->optionTimes(sOldReturnCodes) != 0)
    {
        pData->fOldReturnCodes = 1;
        result += "Return codes       : Old" + lineSep;
    }

    if (parsedResult->optionTimes(sEnableResolveMessageVar) != 0)
    {
        pData->fEnableResolveMessageVar = 1;
        result += "Resolve Message Var: Enabled" + lineSep;
    }
    else if (parsedResult->optionTimes(sDisableResolveMessageVar) != 0)
    {
        pData->fEnableResolveMessageVar = 0;
        result += "Resolve Message Var: Disabled" + lineSep;
    }    
    
    return STAFResultPtr(new STAFResult(kSTAFOk, result), 
                         STAFResultPtr::INIT);
}


STAFResultPtr handleHelp(STAFServiceRequestLevel2 *pInfo, 
                         MonitorServiceData *pData)
{    
    if (pInfo->trustLevel < 1)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFString help("MONITOR Service Help" + lineSep + lineSep);        

    help += "LOG     MESSAGE <Message> [RESOLVEMESSAGE | NORESOLVEMESSAGE]" ;
    help += lineSep;
    help += "QUERY   MACHINE <Machine> HANDLE <Handle>";
    help += lineSep;
    help += "LIST    <MACHINES | [MACHINE <Machine>] | SETTINGS>";
    help += lineSep;
    help += "DELETE  [BEFORE <Timestamp>] CONFIRM";
    help += lineSep;
    help += "SET     [RESOLVEMESSAGE | NORESOLVEMESSAGE]";
    help += lineSep;
    help += "        [OLDRETURNCODES | NEWRETURNCODE] [MAXRECORDSIZE <Size>]";
    help += lineSep;
    help += "        [ENABLERESOLVEMESSAGEVAR | DISABLERESOLVEMESSAGEVAR] ";
    help += lineSep;
    help += "VERSION" + lineSep;
    help += "HELP" + lineSep;

    return STAFResultPtr(new STAFResult(kSTAFOk, help), STAFResultPtr::INIT);
}


STAFResultPtr handleVersion(STAFServiceRequestLevel2 *pInfo, 
                            MonitorServiceData *pData)
{    
    if (pInfo->trustLevel < 1)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }   

    return STAFResultPtr(new STAFResult(kSTAFOk, monitorVersion), 
                         STAFResultPtr::INIT);
}


STAFResultPtr resolveStr(STAFServiceRequestLevel2 *pInfo,
                         MonitorServiceData *pData, 
                         const STAFString &theString)
{
    if (STAFString(pInfo->machine) == STAFString(pInfo->localMachine))
    {
        return pData->fHandlePtr->submit(sLocal, sVar, sResStrHandle +
                                     STAFString(pInfo->handle) +
                                     sResStrResolve +
                                     pData->fHandlePtr->wrapData(theString));
    }

    return pData->fHandlePtr->submit(sLocal, sVar, sResStrGResolve +
                                     pData->fHandlePtr->wrapData(theString));
}


STAFResultPtr resolveOp(STAFServiceRequestLevel2 *pInfo, 
                        MonitorServiceData *pData,
                        STAFCommandParseResultPtr &parsedResult,
                        const STAFString &fOption, unsigned int optionIndex)
{
    // ???: Would const STAFString & work here?
    STAFString optionValue = parsedResult->optionValue(fOption, optionIndex);

    if (optionValue.find(sLeftCurlyBrace) == STAFString::kNPos)
    {
        return STAFResultPtr(new STAFResult(kSTAFOk, optionValue),
                             STAFResultPtr::INIT);
    }

    return resolveStr(pInfo, pData, optionValue);
}


void registerHelpData(MonitorServiceData *pData, unsigned int errorNumber,
                      const STAFString &shortInfo, const STAFString &longInfo)
{
    static STAFString regString("REGISTER SERVICE %C ERROR %d INFO %C "
                                "DESCRIPTION %C");

    pData->fHandlePtr->submit(sLocal, sHelp, STAFHandle::formatString(
        regString.getImpl(), pData->fShortName.getImpl(), errorNumber,
        shortInfo.getImpl(), longInfo.getImpl()));
}


void unregisterHelpData(MonitorServiceData *pData, unsigned int errorNumber)
{
    static STAFString regString("UNREGISTER SERVICE %C ERROR %d");

    pData->fHandlePtr->submit(sLocal, sHelp, STAFHandle::formatString(
        regString.getImpl(), pData->fShortName.getImpl(), errorNumber));
}
