/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.namedcounter;

import com.ibm.staf.*;
import com.ibm.staf.service.*;

import java.util.HashMap;
import java.util.Iterator;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.FileNotFoundException;

/** STAF Named Counter Serivce, providing a mechanism for managing
 * Named Counters across a testing environment.
 * @author Karl Donald
 * @version 1.1.0
 * 
 * 02/21/2004 - V1.2.0: Changed from using old STAFServiceInterfaceLevel1 to
 *   using STAFServiceInterfaceLevel4 so that can access the diagEnabled flag
 *   to check if diagnostics are enabled when doing a LIST request.  If
 *   diagnostics are enabled, records diagnostics data.
 */
 
public class NamedCounterService implements STAFServiceInterfaceLevel4 
{
    /** The STAF Handle that the service uses itself
     */
    private STAFHandle stafHandle;

    /** Error code for a request for a counter which doesn't exist
     */
    static final int CounterNotExist = 4001;

    /** Return code for a counter which is at Zero and as such
     * can't be decremented
     */
    static final int CounterAlreadyZero = 4002;

    /** Return code for is for when a counter hits the maximum
     * value and is unable to increment
     */
    static final int CounterAtMax = 4003;

    /** The structure in which all of the named counters are stored
     */
    private HashMap counters;

    /** Flag to check whether to persist the named counters or not
     */
    private boolean persist;

    /** The name that the service has been assigned
     */
    private String serviceName;

    /** The Parser which is used to process the STAF request
     */
    STAFCommandParser parser = null;
    
    private final String serviceVersion = "1.2.1";

    /** Construct an instance of the NamedCounterService
     */
    public NamedCounterService() {
        counters = new HashMap();
    }

    /** Loads the counters from the file that they were persisted to
     */
    private void loadCounters() 
    {
        try 
        {
            FileInputStream fi = new FileInputStream(serviceName + ".ser");
            ObjectInputStream si = new ObjectInputStream(fi);
            counters = (HashMap) si.readObject();
        } 
        catch (FileNotFoundException e1) 
        {
            System.err.println(
                "NamedCounterService:Error: Named Counters " +
                "File Not Found for service " + serviceName);
            // Not a problem, will create a new one on exit.
            // The object has already been initialised
        } 
        catch (Exception e2) 
        {
            System.err.println(
                "NamedCounterService:Error: Exception caught " + 
                "whilst loading counters");
            System.err.println("                         : " + 
                e2.getMessage());
            e2.printStackTrace();
        }
    }

    /** Save the counters to a file
     * @return The STAFResult object which is the result of the request
     */
    private boolean saveCounters() 
    {
        boolean result = true;
        
        try 
        {
            FileOutputStream fo = new FileOutputStream(serviceName + ".ser");
            ObjectOutputStream so = new ObjectOutputStream(fo);
            so.writeObject(counters);
            so.flush();
        } 
        catch (Exception e) 
        {
            result = false;
            System.err.println(
                "NamedCounterService:Error: Exception caught " +
                "whilst saving counters");
            System.err.println("                         : " + 
                e.getMessage());
            e.printStackTrace();
        }
        
        return result;
    }


    // Register error codes for the service with the HELP service

    private void registerHelpData(int errorNumber, String info,
                                 String description)
    {
        STAFResult res = stafHandle.submit2("local", "HELP",
                         "REGISTER SERVICE " + serviceName +
                         " ERROR " + errorNumber +
                         " INFO " + STAFUtil.wrapData(info) +
                         " DESCRIPTION " + STAFUtil.wrapData(description));
    }

    // Un-register error codes for the service with the HELP service

    private void unregisterHelpData(int errorNumber)
    {
        STAFResult res = stafHandle.submit2("local", "HELP",
                         "UNREGISTER SERVICE " + serviceName +
                         " ERROR " + errorNumber);
    }


    /***************************************************************/
    /* init - Initializes NamedCounter Service with STAF.          */
    /* Instantiates all parsers.                                   */
    /* Registers help for service error codes.                     */
    /*                                                             */
    /* accepts: STAFServiceInterfaceLevel4 initialization info     */
    /*                                                             */
    /* Returns: STAFResult.Ok or STAFResult.STAFRegistrationError  */
    /***************************************************************/
    public int init(STAFServiceInterfaceLevel4.InitInfo info)
    {
        serviceName = info.name;
        
        try 
        {
            parser = new STAFCommandParser(2, false);

            // Add some options that are permitted in command statements 
            parser.addOption("INC", 1, STAFCommandParser.VALUEREQUIRED);
            parser.addOption("DEC", 1, STAFCommandParser.VALUEREQUIRED);
            parser.addOption("GET", 1, STAFCommandParser.VALUEREQUIRED);
            parser.addOption("RESET", 1, STAFCommandParser.VALUEREQUIRED);
            parser.addOption("DELETE", 1, STAFCommandParser.VALUEREQUIRED);
            parser.addOption("LIST", 1, STAFCommandParser.VALUENOTALLOWED);
            parser.addOption("HELP", 1, STAFCommandParser.VALUENOTALLOWED);
            parser.addOption("VERSION", 1, STAFCommandParser.VALUENOTALLOWED);

            // At least one and only one of these must appear 
            parser.addOptionGroup("INC DEC GET RESET DELETE LIST HELP VERSION",
                1, 1);

            // Get a STAFHandle for NamedCounterService <name>
            stafHandle = new STAFHandle("STAF/Service/" + serviceName);
            STAFResult persistRes =
                stafHandle.submit2(
                    "local",
                    "var",
                    "global get STAF/Service/" + serviceName + "/Persist");
                    
            if ((persistRes.rc == 0)
                & (persistRes.result.toUpperCase().equals("TRUE"))) 
            {
                persist = true;
                loadCounters();
            } 
            else 
            {
                persist = false;
            }

            // Register Help Data

            registerHelpData(
                CounterNotExist,
                "Counter Does Not Exist", 
                "The specified named counter could not be found.");

            registerHelpData(
                CounterAlreadyZero,
                "Counter Already Zero",
                "The named counter is currently at Zero.");

            registerHelpData(
                CounterAtMax,
                "Counter At Max",
                "The named counter is at its maximum possible value.");

            return STAFResult.Ok;
        } 
        catch (STAFException e) 
        {
            System.err.println(
                "NamedCounterService:Error: Exception caught " + 
                "whilst initializing service");
            System.err.println("                         : " + 
                e.getMessage());
            e.printStackTrace();
            
            return STAFResult.STAFRegistrationError;
        }
    }

    /** Process a HELP request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionHELP(STAFCommandParseResult myParsedRequest) 
    {
        return new STAFResult(
            STAFResult.Ok,
            "NamedCounter Service Help\n\n"
                + "INC <name>\n"
                + "DEC <name>\n"
                + "GET <name>\n"
                + "RESET <name>\n"
                + "DELETE <name>\n"
                + "LIST\n"
                + "VERSION\n"
                + "HELP");
    }
    
    /** Process a VERSION request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionVERSION(STAFCommandParseResult myParsedRequest) 
    {
        return new STAFResult(STAFResult.Ok, serviceVersion);
    }

    /** Process a GET request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionGET(STAFCommandParseResult myParsedRequest) 
    {
        synchronized(counters)
        {
            if (counters.containsKey(myParsedRequest.optionValue("GET")))
            {
                Integer i = (Integer) counters.get(
                    myParsedRequest.optionValue("GET"));
                    
                return new STAFResult(STAFResult.Ok, i.toString());
            } 
            else 
            {
                return new STAFResult(CounterNotExist, 
                    "Counter does not exist");
            }
        }
    }

    /** Process a LIST request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionLIST(STAFCommandParseResult myParsedRequest)
                                   
    {
        synchronized(counters)
        {
            // List all counters
            StringBuffer counterList = new StringBuffer();
            Iterator iter1 = counters.keySet().iterator();
            Iterator iter2 = counters.values().iterator();
        
            while (iter2.hasNext() && iter1.hasNext()) 
            {
                Object k = iter1.next();
                Object v = iter2.next();
                counterList.append(k.toString() + " = " + v.toString() + "\n");
            }
            
            return new STAFResult(STAFResult.Ok, counterList.toString());
        }
    }

    /** Process an INC request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionINC(STAFCommandParseResult myParsedRequest) 
    {
        synchronized(counters)
        {
            if (counters.containsKey(myParsedRequest.optionValue("INC"))) 
            {
                Integer i = (Integer) counters.get(
                    myParsedRequest.optionValue("INC"));
                
                if (Integer.MAX_VALUE == i.intValue()) 
                {
                    return new STAFResult(CounterAtMax, 
                        "Counter already at maximum");
                } 
                else 
                {
                    i = new Integer(i.intValue() + 1);
                    counters.put(myParsedRequest.optionValue("INC"), i);
                    
                    return new STAFResult(STAFResult.Ok, i.toString());
                }
            } 
            else 
            {
                Integer i = new Integer(1);
                counters.put(myParsedRequest.optionValue("INC"), i);
                
                return new STAFResult(STAFResult.Ok, i.toString());
            }
        }
    }

    /** Process a DEC request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionDEC(STAFCommandParseResult myParsedRequest) 
    {
        synchronized(counters)
        {
            if (counters.containsKey(myParsedRequest.optionValue("DEC"))) 
            {
                Integer i = (Integer) counters.get(
                    myParsedRequest.optionValue("DEC"));
                    
                if (i.intValue() >= 1) 
                {
                    i = new Integer(i.intValue() - 1);
                    counters.put(myParsedRequest.optionValue("DEC"), i);
                    
                    return new STAFResult(STAFResult.Ok, i.toString());
                } 
                else 
                {
                    return new STAFResult(CounterAlreadyZero, 
                        "Counter already at zero");
                }
            } 
            else 
            {
                return new STAFResult(CounterNotExist, "Counter does not exist");
            }
        }
    }

    /** Process a RESET request
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionRESET(STAFCommandParseResult myParsedRequest) 
    {
        synchronized(counters)
        {
            Integer i = new Integer(0);
            counters.put(myParsedRequest.optionValue("RESET"), i);
        
            return new STAFResult(STAFResult.Ok, i.toString());
        }
    }

    /** Process a DELETE reqest
     * @param myParsedRequest The pre-parsed result
     * @return The STAFResult object which is the result of the request
     */
    private STAFResult optionDELETE(STAFCommandParseResult myParsedRequest) 
    {
        synchronized(counters)
        {
            if (counters.containsKey(myParsedRequest.optionValue("DELETE"))) 
            {
                counters.remove(myParsedRequest.optionValue("DELETE"));
            }
        
            return new STAFResult(STAFResult.Ok);
        }
    }

    /***************************************************************************/
    /* acceptRequest- Calls appropriate methods to process a request from a    */
    /*                client process.                                          */
    /*                                                                         */
    /* accepts: STAFServiceInterfaceLevel4 request information                 */
    /*                                                                         */
    /* returns: STAFResult.rc = STAFResult.Ok, if successful; STAFResult.      */
    /*          InvalidRequestString, if unsuccessful;                         */
    /*                                                                         */
    /*          STAFResult.result contains the result returned by the called   */
    /*          method, if successful;                                         */
    /*          STAFResult.result contains the command, if unsuccessful        */
    /***************************************************************************/
    public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFResult result;
        
        try 
        {
            // Create a STAFCommandParseResult object to evaluate the 
            // arguments passed to the service. This is done using the 
            // STAFCommandParser, which was created in init(). 
            
            STAFCommandParseResult myParsedRequest = parser.parse(info.request);

            // If STAFEResult.rc is not 0, the argument string 
            // contained an invalid argument 
            
            if (myParsedRequest.rc != STAFResult.Ok) 
            {
                result = new STAFResult(STAFResult.InvalidRequestString, 
                    myParsedRequest.errorBuffer);
            } 
            else 
            {
                if (myParsedRequest.optionTimes("HELP") == 1) 
                {
                    result = optionHELP(myParsedRequest);
                } 
                else if (myParsedRequest.optionTimes("GET") == 1) 
                {
                    if (info.trustLevel < 1)
                    {
                         return new STAFResult(STAFResult.AccessDenied,
                             "Trust level 1 required for GET request. " +
                             "Requesting machine's trust level: " +
                             info.trustLevel);
                    }

                    result = optionGET(myParsedRequest);
                } 
                else if (myParsedRequest.optionTimes("LIST") == 1) 
                {
                    if (info.trustLevel < 1)
                    {
                         return new STAFResult(STAFResult.AccessDenied,
                             "Trust level 1 required for LIST request. " +
                             "Requesting machine's trust level: " +
                             info.trustLevel);
                    }

                    if (info.diagEnabled == 1)
                    {
                        // Record diagnostics because the output for a LIST
                        // request will change in STAF V3.0

                        String trigger = new String(
                            "STAF/V3.0-Mig/NAMEDCOUNTER LIST");
                        String source  = new String(info.processName + ";" +
                                                    info.machine + ";" +
                                                    info.handle);
                        stafHandle.submit2(
                            STAFHandle.ReqFireAndForget,
                            "local", "DIAG",
                            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                            " SOURCE " + STAFUtil.wrapData(source));
                    }

                    result = optionLIST(myParsedRequest);
                }
                else if (myParsedRequest.optionTimes("VERSION") == 1) 
                {
                    if (info.trustLevel < 1)
                    {
                         return new STAFResult(STAFResult.AccessDenied,
                             "Trust level 1 required for VERSION request. " +
                             "Requesting machine's trust level: " +
                             info.trustLevel);
                    }

                    result = optionVERSION(myParsedRequest);
                }
                else 
                {
                    //  STAFResult.result is always a String.  Strings of 
                    //  digits must be parsed to get the corresponding int
                    //  value. Return an error if Trust level is < 4.
                        
                    if (info.trustLevel < 4) 
                    {
                        return new STAFResult(STAFResult.AccessDenied,
                            "Trust level 4 required for INC, DEC, RESET or " +
                            "DELETE request. Requesting machine's trust " +
                            "level: " + info.trustLevel);
                    } 
                     
                    if (myParsedRequest.optionTimes("INC") == 1) 
                    {
                        result = optionINC(myParsedRequest);
                    } 
                    else if (myParsedRequest.optionTimes("RESET") == 1)
                    {
                        result = optionRESET(myParsedRequest);
                    } 
                    else if (myParsedRequest.optionTimes("DELETE") == 1)
                    {
                        result = optionDELETE(myParsedRequest);
                    } 
                    else if (myParsedRequest.optionTimes("DEC") == 1)
                    {
                        result = optionDEC(myParsedRequest);
                    } 
                    else 
                    {
                        // This should never be reached because it
                        // would mean that the parse had failed,
                        // however the following line is required
                        // so that the code can be compiled.
                        result = new 
                            STAFResult(STAFResult.UnknownError);
                    }
                }
            }
        } 
        catch (Exception e) 
        {
            System.err.println("NamedCounterService:Error: " + 
                "Exception caught whilst processing a request");
            System.err.println("                         : " + e.getMessage());
            e.printStackTrace();
            
            result = new STAFResult(STAFResult.JavaError, 
                "Internal Java error.");
        }
        
        return result;
    }

    /** Terminate the service, and save any existing counters if
     * the option to persist them is set
     * @return Ok if saving the counters succeeded, or JavaError if
     * the save failed (this is assuming that the service has
     * been setup to persist the counters)
     */
    public int term() 
    {
        boolean saveResult = true;
        
        // Un-register Help Data

        unregisterHelpData(CounterNotExist);
        unregisterHelpData(CounterAlreadyZero);
        unregisterHelpData(CounterAtMax);

        // If persisting counters
        if (persist) 
        {
            saveResult = saveCounters();
        }
        
        // Un-register the service handle

        try
        {
            stafHandle.unRegister();
        }
        catch (STAFException ex)
        {
            if (saveResult == true)
            {
                return STAFResult.STAFRegistrationError;
            }
        }

        // If the counters saved ok, (or we aren't persisting them)
        if (saveResult == true) 
        {
            return STAFResult.Ok;
            // Else something went wrong whilst saving the counters 
        } else 
        {
            return STAFResult.JavaError;
        }
    }
}
