/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002, 2005                                        */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import com.ibm.staf.*;
import com.ibm.staf.service.*;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.InputSource;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.io.StringReader;
import org.python.util.PythonInterpreter;
import java.util.Properties;
import java.util.jar.*;
import java.io.IOException;
import java.util.Map;
import java.io.BufferedReader;
import java.util.StringTokenizer;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

public class STAX implements STAFServiceInterfaceLevel4, STAXJobCompleteListener
{
    // 1.0  released 09/17/2001
    // 1.1.0 released 03/12/2002 
    // 1.2.0 released 08/05/2002
    // 1.3.0 released 09/23/2002
    // 1.3.1 released 10/09/2002
    // 1.3.2 released 12/05/2002
    // 1.3.3 released 03/18/2003
    // 1.4.0 released 06/04/2003
    // 1.4.1 released 09/18/2003
    // 1.5.0 released 01/29/2004
    // 1.5.2 released 04/01/2004
    // 1.5.5 released 11/04/2004

    static final String fVersion = "1.5.9";
    
    static boolean CACHE_PYTHON_CODE = true;

    static final String INLINE_DATA = "<inline data>";
    static final String sNOT_PROVIDED = "<Not Provided>";
    static final String sNONE = "<None>";
    static final String sNA = "<N/A>";

    static final String PACKAGE_NAME = "com.ibm.staf.service.stax.STAX";

    public static String lineSep;
    public static String fileSep;

    // STAF Service Interface Levels Supported for Generic Requests
    static final String INTERFACE_LEVEL_3 = 
        "com.ibm.staf.service.STAFServiceInterfaceLevel3";
    static final String INTERFACE_LEVEL_4 = 
        "com.ibm.staf.service.STAFServiceInterfaceLevel4";

    // STAX Service Error Code
    static final int ErrorSubmittingExecuteRequest = 4001;
    static final int BlockNotHeld = 4002;
    static final int BlockAlreadyHeld = 4003;

    static final String STAX_EXTENSION = new String("staf/stax/extension/");
    static final String STAX_MONITOR_EXTENSION =
        new String("staf/staxmonitor/extension/");
    static final String STAX_EXTENSION_INFO = 
        new String("staf/staxinfo/extension");

    public STAX() { /* Do Nothing */ }

    public int init(STAFServiceInterfaceLevel4.InitInfo info)
    {
        int rc = STAFResult.Ok;
        
        try
        {
            fServiceName = info.name;

            // Get the services STAF handle

            fHandle = new STAFHandle("STAF/Service/" + info.name);

            // Get the line separator

            STAFResult res = fHandle.submit2("local", "var",
                             "resolve {STAF/Config/Sep/Line}");
            if (res.rc != 0)
            {
                System.out.println(res.result);
                return res.rc;
            }
            else
                lineSep = res.result;

            // Get the effective machine name

            res = fHandle.submit2("local", "var", 
                  "resolve {STAF/Config/EffectiveMachine}");
            if (res.rc != 0)
            {
                System.out.println(res.result);
                return res.rc;
            }
            else
                fLocalMachineName = res.result;
            
            if (info.serviceJar != null)
            {
                // Create the Jython Library directory if it doesn't exist yet
                // from the jython library data in the service's jarfile.
                // Set the classpath to point to the service's jar file and to
                // the jython.jar.
                
                res = STAFServiceSharedJython.setupJython(info.serviceJar,
                                                          fHandle);
                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res.rc;
                }
                
                // Initialize the PythonInterpreter's python.home to point to
                // the shared jython installation.

                Properties p = new Properties();
                p.setProperty("python.home",
                              STAFServiceSharedJython.getJythonDirName());
                PythonInterpreter.initialize(System.getProperties(), p, null);
            }
            
            // Create a Xerces DOM parser to verify that Xerces is installed
            
            try
            {
                org.apache.xerces.parsers.DOMParser parser =
	              new org.apache.xerces.parsers.DOMParser();
            }
            catch (NoClassDefFoundError e)
            {
                rc = STAFResult.ServiceConfigurationError;
                System.out.println("ERROR: xercesImpl.jar is not in the " +
                                   "classpath.");
                return rc;
            }

            // Create a Python Interpreter to verify that Jython is installed

            try
            {
                STAXPythonInterpreter pyi = new STAXPythonInterpreter();
            }
            catch (NoClassDefFoundError e)
            {
                rc = STAFResult.ServiceConfigurationError;
                System.out.println("ERROR: jython.jar is not in the " +
                                   "classpath.");
                return rc;
            }
            
            // Verify that the version specified for the STAX service is valid

            try
            {
                STAXVersion serviceVersion = new STAXVersion(fVersion);
            }
            catch (NumberFormatException e)
            {
                rc = STAFResult.ServiceConfigurationError;
                System.out.println("ERROR: " + fVersion +
                                   " is an invalid STAX service version." +
                                   lineSep + e.toString());
                return rc;
            }

            // Assign the STAX service machine as the default EVENT service
            // machine and assign "Event" as the default EVENT service name.
            // These will be overridden if an EVENTSERVICEMACHINE or
            // EVENTSERVICENAME parameter is specified.

            fEventServiceMachine = fLocalMachineName;
            fEventServiceName = "Event";

            // Assign 5 as the default number of physical threads used by STAX
            fNumThreads = 5;

            // Parse PARMS if provided in the STAX configuration line

            if (info.parms != null)
            {
                fParmsParser.addOption("EVENTSERVICEMACHINE", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("EVENTSERVICENAME", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("NUMTHREADS", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("EXTENSIONXMLFILE", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("EXTENSIONFILE", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("EXTENSION", 0,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("PROCESSTIMEOUT", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("CLEARLOGS", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("LOGTCELAPSEDTIME", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("LOGTCNUMSTARTS", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOption("LOGTCSTARTSTOP", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                // CACHEPYTHONCODE is an undocumented parameter
                // Mainly there in case any problems arise with the caching
                // of Python code added in STAX V1.5.0.
                fParmsParser.addOption("CACHEPYTHONCODE", 1,
                                       STAFCommandParser.VALUEREQUIRED);
                fParmsParser.addOptionGroup(
                    "EXTENSIONFILE EXTENSIONXMLFILE", 0, 1);

                rc = handleParms(info);
                if (rc != STAFResult.Ok)
                {
                    rc = STAFResult.ServiceConfigurationError;
                    return rc;
                }
            }

            // Create the STAX Thread Queue

            fThreadQueue = new STAXThreadQueue(fNumThreads);

            // Prime the element pump

            // Note that elements that implement a STAX request handler pass
            // parameters "this" to the action factory constructor.

            fActionFactoryMap.put("process", 
                                  new STAXProcessActionFactory(this));
            fActionFactoryMap.put("stafcmd", 
                                  new STAXSTAFCommandActionFactory(this));
            fActionFactoryMap.put("job", 
                                  new STAXJobActionFactory(this));
            fActionFactoryMap.put("block", new STAXBlockActionFactory(this));
            fActionFactoryMap.put("testcase", 
                                  new STAXTestcaseActionFactory(this));

            fActionFactoryMap.put("function", new STAXFunctionActionFactory());
            fActionFactoryMap.put("signalhandler",
                                  new STAXSignalHandlerActionFactory());
            fActionFactoryMap.put("sequence", new STAXSequenceActionFactory());
            fActionFactoryMap.put("parallel", new STAXParallelActionFactory());
            fActionFactoryMap.put("call", new STAXCallActionFactory());
            fActionFactoryMap.put("call-with-list",
                                  new STAXCallWithListActionFactory());
            fActionFactoryMap.put("call-with-map",
                                  new STAXCallWithMapActionFactory());
            fActionFactoryMap.put("script", new STAXScriptActionFactory());
            fActionFactoryMap.put("if", new STAXIfActionFactory());
            fActionFactoryMap.put("nop", new STAXNopActionFactory());
            fActionFactoryMap.put("message", new STAXMessageActionFactory());
            fActionFactoryMap.put("log", new STAXLogActionFactory());
            fActionFactoryMap.put("loop", new STAXLoopActionFactory());
            fActionFactoryMap.put("iterate", new STAXIterateActionFactory());
            fActionFactoryMap.put("paralleliterate",
                                  new STAXParallelIterateActionFactory());
            fActionFactoryMap.put("terminate", 
                                  new STAXTerminateActionFactory());
            fActionFactoryMap.put("hold", new STAXHoldActionFactory());
            fActionFactoryMap.put("release", new STAXReleaseActionFactory());
            fActionFactoryMap.put("tcstatus", 
                                  new STAXTestcaseStatusActionFactory());
            fActionFactoryMap.put("break", new STAXBreakActionFactory());
            fActionFactoryMap.put("continue", new STAXContinueActionFactory());
            fActionFactoryMap.put("raise", new STAXRaiseActionFactory());
            fActionFactoryMap.put("try", new STAXTryActionFactory());
            fActionFactoryMap.put("throw", new STAXThrowActionFactory());
            fActionFactoryMap.put("rethrow", new STAXRethrowActionFactory());
            fActionFactoryMap.put("timer", new STAXTimerActionFactory());
            fActionFactoryMap.put("import", new STAXImportActionFactory());
            fActionFactoryMap.put("return", new STAXReturnActionFactory());

            // Prime the element pump with the Extension elements
           
            res = loadExtensions();

            if (res.rc != 0)
            {
                System.out.println(res.result);
                logToServiceLog("error", res.result);
                return res.rc;
            }

            // Log the registered extensions in the STAX Service Log

            synchronized (fExtensionsJarList)
            {
                Iterator iter = fExtensionsJarList.iterator();

                StringBuffer result = new StringBuffer(
                    lineSep + "Registered Extensions for " + fServiceName +
                    " Version " + fVersion + ":" + lineSep);

                while (iter.hasNext())
                {
                    STAXExtension ext = (STAXExtension)iter.next();

                    result.append(lineSep + queryExtension(ext));
                }

                logToServiceLog("info", result.toString());
                System.out.println(result.toString());
            }   

            // Execute parser

            fExecuteParser.addOption("EXECUTE", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
            fExecuteParser.addOption("FILE", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("MACHINE", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("DATA", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("JOBNAME", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("FUNCTION", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("ARGS", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("SCRIPT", 0,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("SCRIPTFILE", 0,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("SCRIPTFILEMACHINE", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("HOLD", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
            fExecuteParser.addOption("TEST", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
                                     
            // RETURNDETAILS option is un-documented
            fExecuteParser.addOption("RETURNDETAILS", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
                                     
            fExecuteParser.addOption("WAIT", 1,
                                     STAFCommandParser.VALUEALLOWED);
            fExecuteParser.addOption("RETURNRESULT", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
            fExecuteParser.addOption("CLEARLOGS", 1,
                                     STAFCommandParser.VALUEALLOWED);
            fExecuteParser.addOption("LOGTCELAPSEDTIME", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("LOGTCNUMSTARTS", 1,
                                     STAFCommandParser.VALUEREQUIRED);
            fExecuteParser.addOption("LOGTCSTARTSTOP", 1,
                                     STAFCommandParser.VALUEREQUIRED);

            fExecuteParser.addOptionGroup("FILE DATA", 0, 1);
            fExecuteParser.addOptionGroup("HOLD TEST WAIT", 0, 1);
            
            fExecuteParser.addOptionNeed("EXECUTE", "FILE DATA");
            fExecuteParser.addOptionNeed("MACHINE", "FILE");
            fExecuteParser.addOptionNeed("SCRIPTFILEMACHINE", "SCRIPTFILE");
            fExecuteParser.addOptionNeed("RETURNRESULT", "WAIT");
            fExecuteParser.addOptionNeed("RETURNDETAILS", "TEST");

            // List parser

            fListParser.addOption("LIST", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("JOBS", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("JOB", 1,
                                  STAFCommandParser.VALUEREQUIRED);
            fListParser.addOption("SETTINGS", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("EXTENSIONS", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("EXTENSIONJARFILES", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fListParser.addOption("THREADS", 1,
                                  STAFCommandParser.VALUENOTALLOWED);

            StringBuffer jobOptions = new StringBuffer("THREADS");

            // Add additional options from registered list handlers
            synchronized (fListRequestMap)
            {
                Iterator iter = fListRequestMap.keySet().iterator();
                while (iter.hasNext())
                {
                    String typeName = (String)iter.next();
                    fListParser.addOption(typeName.toUpperCase(), 1,
                                          STAFCommandParser.VALUENOTALLOWED);
                    jobOptions.append(" " + typeName.toUpperCase());
                }
            }

            String listJobOptions = new String(jobOptions);
            String listOptions = new String(
                "JOB JOBS SETTINGS EXTENSIONS EXTENSIONJARFILES");

            fListParser.addOptionGroup("LIST", 1, 1);
            fListParser.addOptionGroup(listOptions, 0, 1);
            fListParser.addOptionNeed("LIST", listOptions);
            fListParser.addOptionNeed(listOptions, "LIST");
            fListParser.addOptionNeed(listJobOptions, "JOB");
            fListParser.addOptionNeed("JOB", listJobOptions); 

            // Query parser

            fQueryParser.addOption("QUERY", 1,
                                   STAFCommandParser.VALUENOTALLOWED);
            fQueryParser.addOption("JOB", 1,
                                   STAFCommandParser.VALUEREQUIRED);
            fQueryParser.addOption("THREAD", 0,
                                   STAFCommandParser.VALUEREQUIRED);
            fQueryParser.addOption("EXTENSIONJARFILE", 1,
                                   STAFCommandParser.VALUEREQUIRED);
            fQueryParser.addOption("EXTENSIONJARFILES", 1,
                                   STAFCommandParser.VALUENOTALLOWED);
            
            jobOptions = new StringBuffer("THREAD");
            String fHelpQueryJobOptions = "THREAD <Thread ID>";

            // Add additional options from registered query handlers
            synchronized (fQueryRequestMap)
            {
                Iterator iter = fQueryRequestMap.keySet().iterator();
                while (iter.hasNext())
                {
                    String typeName = ((String)iter.next()).toUpperCase();
                    fQueryParser.addOption(typeName, 0,
                                           STAFCommandParser.VALUEREQUIRED);
                    jobOptions.append(" " + typeName);
                }
            }
            String queryJobOptions = new String(jobOptions);
            String queryOptions = "JOB EXTENSIONJARFILE EXTENSIONJARFILES";

            fQueryParser.addOptionGroup("QUERY", 1, 1);

            fQueryParser.addOptionGroup(queryOptions, 0, 1);
            fQueryParser.addOptionNeed(queryOptions, "QUERY");
            fQueryParser.addOptionNeed("QUERY", queryOptions);
            fQueryParser.addOptionNeed(queryJobOptions, "JOB");

            // Get parser

            fGetParser.addOption("GET", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fGetParser.addOption("DTD", 1, 
                                  STAFCommandParser.VALUENOTALLOWED);
            fGetParser.addOptionNeed("GET", "DTD");

            // Undocumented option STAX-EXTENSIONS-FILE - gets the DTD for
            // the STAX Extensions XML file
            fGetParser.addOption("STAX-EXTENSIONS-FILE", 1,
                                 STAFCommandParser.VALUENOTALLOWED);
            fGetParser.addOptionNeed("STAX-EXTENSIONS-FILE", "DTD");

            // Set parser

            fSetParser.addOption("SET", 1,
                                 STAFCommandParser.VALUENOTALLOWED);
            fSetParser.addOption("CLEARLOGS", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fSetParser.addOption("LOGTCELAPSEDTIME", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fSetParser.addOption("LOGTCNUMSTARTS", 1,
                                 STAFCommandParser.VALUEREQUIRED);
            fSetParser.addOption("LOGTCSTARTSTOP", 1,
                                 STAFCommandParser.VALUEREQUIRED);

            String setOptions = "CLEARLOGS LOGTCELAPSEDTIME " +
                                "LOGTCNUMSTARTS LOGTCSTARTSTOP";
            fSetParser.addOptionNeed("SET", setOptions);
            fSetParser.addOptionNeed(setOptions, "SET");

            // Version parser

            fVersionParser.addOption("VERSION", 1,
                                     STAFCommandParser.VALUENOTALLOWED);
            fVersionParser.addOptionGroup("VERSION", 1, 1);

            // Help parser

            fHelpParser.addOption("HELP", 1,
                                  STAFCommandParser.VALUENOTALLOWED);
            fHelpParser.addOptionGroup("HELP", 1, 1);

            // Create the DTD

            // Get the current date and time
            STAXTimestamp currTimestamp = new STAXTimestamp();

// Begin DTD chunk............................................................
            fDTD =
"<!--\n" +
"   STAf eXecution (STAX) Document Type Definition (DTD)\n" +
"\n" +
"   Generated Date: " + currTimestamp.getTimestampString() + "\n" +
"\n" +
"   This DTD module is identified by the SYSTEM identifier:\n" +
"\n" +
"     SYSTEM 'stax.dtd'\n" +
"\n" +
"-->\n" +
"\n" +
"<!-- Parameter entities referenced in Element declarations -->\n" +
"\n" +
"<!ENTITY % stax-elems 'function | script | signalhandler'>\n" +
"\n";
// End DTD chunk..............................................................

// Begin DTD chunk............................................................
String staxDTDsection =
"\n" +
"<!--================= STAX Job Definition ========================== -->\n" +
"<!--\n" +
"     The root element STAX contains all other elements.  It consists\n" +
"     of an optional defaultcall element and any number of function,\n" +
"     script, and/or signalhandler elements.\n" +
"-->\n" +
"<!ELEMENT stax         ((%stax-elems;)*, defaultcall?, (%stax-elems;)*)>\n" +
"\n" +
"<!--================= The Default Call Function Element ============ -->\n" +
"<!--\n" +
"     The defaultcall element defines the function to call by default\n" +
"     to start the job.  This can be overridden by the 'FUNCTION'\n" +
"     parameter when submitting the job to be executed.\n" +
"     The function attribute's value is a literal.\n" +
"-->\n" +
"<!ELEMENT defaultcall  (#PCDATA)>\n" +
"<!ATTLIST defaultcall\n" +
"          function     IDREF    #REQUIRED\n" +
">\n";
// End DTD chunk..............................................................

            StringBuffer taskEntitySection = 
                new StringBuffer("<!ENTITY % task       '");
            StringBuffer elemDTDsection = new StringBuffer();
            int numTasks = 0;

            synchronized (fActionFactoryMap)
            {
                Iterator factory_iter = fActionFactoryMap.keySet().iterator();
                while (factory_iter.hasNext())
                {
                    String actionFactoryName = (String)factory_iter.next();
                    STAXActionFactory factory = (STAXActionFactory)
                        fActionFactoryMap.get(actionFactoryName);

                    elemDTDsection.append(factory.getDTDInfo());

                    String taskName = factory.getDTDTaskName();

                    if (taskName != null)
                    {
                        // Put 3 actions in each line for the task entity.
                        if (numTasks == 0)
                            taskEntitySection.append(factory.getDTDTaskName());
                        else if (numTasks % 3 == 0)
                            taskEntitySection.append(" | \n" +
                                                 "                       " + 
                                                 factory.getDTDTaskName());
                        else
                            taskEntitySection.append(" | " + 
                                                 factory.getDTDTaskName());
                        numTasks++;
                    }
                }
            }
            
            taskEntitySection.append("'>\n");
            
            StringBuffer dtdBuff = new StringBuffer(fDTD);
            dtdBuff.append(taskEntitySection);
            dtdBuff.append(staxDTDsection);
            dtdBuff.append(elemDTDsection);
            fDTD = dtdBuff.toString();

            // Register Help Data

            registerHelpData(
                ErrorSubmittingExecuteRequest,
                "Error submitting execute request", 
                "Additional information about the error is put into the " +
                "STAF Result.");

            registerHelpData(
                BlockNotHeld,
                "Block not held",
                "Requested to release a block that is not held.");

            registerHelpData(
                BlockAlreadyHeld,
                "Block already held",
                "Requested to hold a block that is already held.");
        }
        catch (STAFException e)
        {
            rc = STAFResult.ServiceConfigurationError;
        }
        catch (Throwable t)
        {
            rc = STAFResult.ServiceConfigurationError;
            t.printStackTrace();
        }

        return rc;
    }

    public String getDTD() { return fDTD; }

    public String getServiceName() { return fServiceName; }

    public String getEventServiceMachine() { return fEventServiceMachine; }

    public String getEventServiceName() { return fEventServiceName; }

    public String getLocalMachineName() { return fLocalMachineName; }
    
    public int getProcessTimeout() { return fProcessTimeout; }

    public boolean getClearlogs() { return fClearLogs; }

    public String getClearLogsAsString()
    {
        if (fClearLogs)
            return "Enabled";
        else
            return "Disabled";
    }

    public boolean getLogTCElapsedTime() { return fLogTCElapsedTime; } 

    public String getLogTCElapsedTimeAsString()
    {
        if (fLogTCElapsedTime)
            return "Enabled";
        else
            return "Disabled";
    }

    public boolean getLogTCNumStarts() { return fLogTCNumStarts; }

    public String getLogTCNumStartsAsString()
    {
        if (fLogTCNumStarts)
            return "Enabled";
        else
            return "Disabled";
    }

    public boolean getLogTCStartStop() { return fLogTCStartStop; } 

    public String getLogTCStartStopAsString()
    {
        if (fLogTCStartStop)
            return "Enabled";
        else
            return "Disabled";
    }

    public int getNextJobNumber()
    {
        synchronized (fNextJobNumberSynch)
        {
            return fNextJobNumber++;
        }
    }
    public STAFHandle getSTAFHandle() { return fHandle; }

    public STAXActionFactory getActionFactory(String elementName)
    {
        return (STAXActionFactory)fActionFactoryMap.get(elementName);
    }

    public TreeMap getJobMap()
    {
        synchronized(this)
        {
            return fJobMap;
        }
    }

    public void setExtension(STAXExtension ext)
    {
        synchronized(fExtensionsJarList)
        {
            fExtensionsJarList.add(ext);
        }
    }

    public STAXThreadQueue getThreadQueue() { return fThreadQueue; }

    public STAXTimedEventQueue getTimedEventQueue() { return fTimedEventQueue; }

    public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        String lowerRequest = info.request.toLowerCase();

        if (lowerRequest.startsWith("execute"))
            return handleExecute(info, null);
        else if (lowerRequest.startsWith("list"))
            return handleList(info);
        else if (lowerRequest.startsWith("query"))
            return handleQuery(info);
        else if (lowerRequest.startsWith("help"))
            return handleHelp(info);
        else if (lowerRequest.startsWith("version"))
            return handleVersion(info);
        else if (lowerRequest.startsWith("get"))
            return handleGet(info);
        else if (lowerRequest.startsWith("set"))
            return handleSet(info);
        else
            return handleGenericRequest(info);
    }

    public int term()
    {
        // Issue a STAX TERMINATE JOB request for each job still running

        synchronized (fJobMap)
        {
            Iterator iter = fJobMap.values().iterator();

            while (iter.hasNext())
            {
                STAXJob thisJob = (STAXJob)iter.next();
                
                STAFResult result = acceptRequest(
                    new STAFServiceInterfaceLevel4.RequestInfo(
                    fLocalMachineName, fLocalMachineName, fLocalMachineName,
                    "STAF/Service/" + fServiceName, 1, 5,
                    0,  // XXX: Set to disable diagnostics since enableDiag
                        // isn't provided
                    "TERMINATE JOB " + thisJob.getJobNumber()));
                
                if (result.rc != 0)
                {
                    System.out.println("Error terminating STAX Job " +
                        thisJob.getJobNumber() + ".  RCc=" + result.rc +
                        ", result=" + result.result);
                }
            }
        }
            
        // Wait until all jobs are terminated or maximum wait time exceeded

        int maxLoops = 60;  // 60 seconds is the maximum wait time

        for (int i = 0; fJobMap.size() != 0 && i < maxLoops; i++)
        {
            try
            {                    
                Thread.sleep(1000);   // Sleep 1 second
            }
            catch (InterruptedException ex)
            { }
        }
        
        // Un-register Help Data

        unregisterHelpData(ErrorSubmittingExecuteRequest);
        unregisterHelpData(BlockNotHeld);
        unregisterHelpData(BlockAlreadyHeld);

        // Un-register the service handle

        try
        {
            fHandle.unRegister();
        }
        catch (STAFException ex)
        {
            return STAFResult.STAFRegistrationError;
        }

        return STAFResult.Ok;
    }

    private STAFResult handleExecute(STAFServiceInterfaceLevel4.RequestInfo info,
                                     STAXJobCompleteListener listener)
    {
        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 4)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 4 required for EXECUTE request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }
        
        STAXSubmitResult res = execute(info, listener);

        return res.getResult();
    }

    public STAXSubmitResult execute(STAFServiceInterfaceLevel4.RequestInfo info,
                                    STAXJobCompleteListener listener)
    {
        STAXParser parser = null;
        STAXJob job = null;
        
        STAFCommandParseResult parseResult = fExecuteParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAXSubmitResult(new STAFResult(
                STAFResult.InvalidRequestString, parseResult.errorBuffer));
        }
 
        try
        {
            parser = new STAXParser(this);

            // Parse and resolve the MACHINE value

            String machName;
            STAFResult resolvedValue;
            if (parseResult.optionTimes("MACHINE") > 0)
            {
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("MACHINE"), 
                                info.handle);
                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);
            
                machName = resolvedValue.result;    
            }
            else
                machName = info.machine;

            // Parse an XML document from a file.

            String fileName = "";
            if (parseResult.optionTimes("FILE") > 0)
            {
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("FILE"), info.handle);
                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);
            
                fileName = resolvedValue.result;    

                STAFResult copyResult = fHandle.submit2(machName, "fs", 
                    "get file " + fileName); 
                if (copyResult.rc != 0)           
                {   // There was an error in the FS service submission 
                    return new STAXSubmitResult(
                        new STAFResult(copyResult.rc, copyResult.result));
                }
                job = parser.parse(new InputSource(
                                   new StringReader(copyResult.result)));
            }

            // Parse an XML document provided in a String

            if (parseResult.optionTimes("DATA") > 0)
            {
                job = parser.parse(new InputSource(
                          new StringReader(parseResult.optionValue("DATA"))));
                fileName = INLINE_DATA;
            }
            
            job.setXmlFile(fileName);
            job.setXmlMachine(machName);
            job.setSourceMachine(info.machine);
            job.setSourceHandleName(info.processName);
            job.setSourceHandle(info.handle);

            if (parseResult.optionTimes("JOBNAME") > 0)
            {
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("JOBNAME"), 
                                info.handle);
                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);

                job.setJobName(resolvedValue.result);
            }

            if (parseResult.optionTimes("FUNCTION") > 0)
            {
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                            parseResult.optionValue("FUNCTION"),
                                            info.handle);
                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);
                
                job.setStartFunction(resolvedValue.result);

                // Make sure that the start function specified is valid

                if (!job.functionExists(resolvedValue.result))
                {
                    throw new STAXInvalidStartFunctionException(
                        resolvedValue.result);
                }
            }  

            if (parseResult.optionTimes("ARGS") > 0)
            {
                // Not resolving STAF variables for ARGS to support Python
                // dictionaries which use { } symbols as well.

                // Parse and compile starting function's arguments
                String args = STAXUtil.parseAndCompileForPython(
                    parseResult.optionValue("ARGS"), "\n  ARGS option");

                job.setStartFuncArgs(args);
            }  

            if (parseResult.optionTimes("HOLD") > 0)
            {
                job.setExecuteAndHold();
            }

            String scriptMachName = "";

            if (parseResult.optionTimes("SCRIPTFILEMACHINE") > 0)
            {
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("SCRIPTFILEMACHINE"),
                                info.handle);
                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);
                 
                scriptMachName = resolvedValue.result;
                job.setScriptFileMachine(resolvedValue.result);
            }
            else if (parseResult.optionTimes("SCRIPTFILE") > 0)
            {
                scriptMachName = machName;
                job.setScriptFileMachine(machName);
            }
            
            for (int i = 1; i <= parseResult.optionTimes("SCRIPTFILE"); i++)
            {
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("SCRIPTFILE", i),
                                info.handle);
                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);
                
                job.setScriptFile(resolvedValue.result);

                STAFResult copyResult = fHandle.submit2(scriptMachName, "fs", 
                    "get file " + resolvedValue.result); 
                if (copyResult.rc != 0)           
                {   // There was an error in the FS service submission 
                    return new STAXSubmitResult(
                        new STAFResult(copyResult.rc, copyResult.result));
                }

                // Parse and compile SCRIPTFILE contents containing Python code
                String code = STAXUtil.parseAndCompileForPython(
                    copyResult.result,
                    "\n  Contents of SCRIPTFILE " + resolvedValue.result);

                job.addDefaultAction(new STAXScriptAction(code));
            }

            for (int i = 1; i <= parseResult.optionTimes("SCRIPT"); i++)
            {
                // Parse and compile SCRIPT value containing Python code
                String code = STAXUtil.parseAndCompileForPython(
                    parseResult.optionValue("SCRIPT", i), "\n  SCRIPT option");
                
                job.addDefaultAction(new STAXScriptAction(code));
                job.setScript(code);
            }

            if (parseResult.optionTimes("CLEARLOGS") > 0)
            {
                // Resolve the value specified for CLEARLOGS
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("CLEARLOGS"),
                                info.handle);

                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);

                if (resolvedValue.result.equalsIgnoreCase("ENABLED") ||
                    resolvedValue.result.equals(""))
                {
                    job.setClearlogs(true);
                }
                else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                {
                    job.setClearlogs(false);
                }
                else
                {
                    return new STAXSubmitResult(new STAFResult(
                        STAFResult.InvalidValue, 
                        "CLEARLOGS value must be ENABLED or DISABLED.  " +
                        "Invalid value: " + resolvedValue.result));
                }
            }

            // Handle LOGTCELAPSEDTIME setting

            if (parseResult.optionTimes("LOGTCELAPSEDTIME") > 0)
            {
                // Resolve the value specified for LOGTCELAPSEDTIME
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("LOGTCELAPSEDTIME"),
                                info.handle);

                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);

                if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                {
                    job.setLogTCElapsedTime(true);
                }
                else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                {
                    job.setLogTCElapsedTime(false);
                }
                else
                {
                    return new STAXSubmitResult(new STAFResult(
                        STAFResult.InvalidValue, 
                        "LOGTCELAPSEDTIME value must be ENABLED or DISABLED.  " +
                        "Invalid value: " + resolvedValue.result));
                }
            }

            // Handle LOGTCNUMSTARTS setting

            if (parseResult.optionTimes("LOGTCNUMSTARTS") > 0)
            {
                // Resolve the value specified for LOGTCNUMSTARTS
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("LOGTCNUMSTARTS"),
                                info.handle);

                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);

                if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                {
                    job.setLogTCNumStarts(true);
                }
                else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                {
                    job.setLogTCNumStarts(false);
                }
                else
                {
                    return new STAXSubmitResult(new STAFResult(
                        STAFResult.InvalidValue, 
                        "LOGTCNUMSTARTS value must be ENABLED or DISABLED.  " +
                        "Invalid value: " + resolvedValue.result));
                }
            }

            // Handle LOGTCSTARTSTOP setting

            if (parseResult.optionTimes("LOGTCSTARTSTOP") > 0)
            {
                // Resolve the value specified for LOGTCSTARTSTOP
                resolvedValue = resolveValue(info.machine, info.localMachine,
                                parseResult.optionValue("LOGTCSTARTSTOP"),
                                info.handle);

                if (resolvedValue.rc != 0)
                    return new STAXSubmitResult(resolvedValue);

                if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                {
                    job.setLogTCStartStop(true);
                }
                else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                {
                    job.setLogTCStartStop(false);
                }
                else
                {
                    return new STAXSubmitResult(new STAFResult(
                        STAFResult.InvalidValue, 
                        "LOGTCSTARTSTOP value must be ENABLED or DISABLED.  " +
                        "Invalid value: " + resolvedValue.result));
                }
            }

            if (parseResult.optionTimes("TEST") > 0)
            {
                String result = "";
                
                if (parseResult.optionTimes("RETURNDETAILS") > 0)   
                {
                    if (info.diagEnabled == 1)
                    {
                        // Record diagnostics because output for an EXECUTE
                        // request with options TEST and RETURNDETAILS
                        // specified will change in STAF V3.0

                        String trigger = new String(
                            "STAF/V3.0-Mig/STAX EXECUTE TEST RETURNDETAILS");
                        String source  = new String(info.processName + ";" +
                                                    info.machine + ";" +
                                                    info.handle);
                        fHandle.submit2(
                            STAFHandle.ReqFireAndForget,
                            "local", "DIAG",
                            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                            " SOURCE " + STAFUtil.wrapData(source));
                    }

                    // Get a list of functions to return
                    
                    HashMap functionMap = job.getFunctionMap();
                    
                    java.util.Collection functionColl = functionMap.keySet();
                    TreeSet sortedFunctions = new TreeSet(functionColl);
                    Iterator iter = sortedFunctions.iterator();
                
                    result += "Default Call       : " + 
                        job.getStartFunction() + "\n\n";
                    
                    while (iter.hasNext())
                    {
                        String functionName = (String)iter.next();
                        
                        STAXFunctionAction function = 
                            (STAXFunctionAction)functionMap.get(functionName);
                        result += function.getArgumentInfo();
                        result += "\n";
                    }
                }
                
                return new STAXSubmitResult(
                        new STAFResult(STAFResult.Ok, result), job);
            }
            
            // Make sure that the start function is valid

            String startFunction = job.getStartFunction();

            if (!job.functionExists(startFunction))
            {
                throw new STAXInvalidStartFunctionException(startFunction);
            }

            job.setJobNumber(this.getNextJobNumber());


            synchronized (fJobMap)
            {
                fJobMap.put(job.getJobNumberAsInteger(), job);
            }
            
            job.addCompletionNotifiee(this);

            // If this is a sub-job (indicated by a non-null listener), make
            // sure the sub-job is also notified when the job completes.
            if (listener != null)
            {
                job.addCompletionNotifiee(listener);
            }

            job.startExecution();
        }
        catch (STAFException e)
        {
            return new STAXSubmitResult(
                new STAFResult(STAX.ErrorSubmittingExecuteRequest, 
                               "Caught " + e.getClass().getName() + ": " +
                               e.getMessage()));
        }
        catch (STAXException e)
        {
            return new STAXSubmitResult(
                new STAFResult(STAX.ErrorSubmittingExecuteRequest, 
                               "Caught " + e.getClass().getName() + ": " +
                               e.getMessage()));
        }
        catch (SAXNotRecognizedException e)
        {
            return new STAXSubmitResult( 
                new STAFResult(STAX.ErrorSubmittingExecuteRequest, 
                               "Caught SAXNRE: " + e.getMessage()));
        }
        catch (SAXNotSupportedException e)
        {
            return new STAXSubmitResult(
                new STAFResult(STAX.ErrorSubmittingExecuteRequest, 
                               "Caught SAXNSE: " + e.getMessage()));
        }
        catch (Exception e)
        {
            e.printStackTrace();

            return new STAXSubmitResult(
                new STAFResult(STAX.ErrorSubmittingExecuteRequest, 
                               "Caught Exception: " +
                               e.getClass().getName() + ": " +
                               e.getMessage()));
        }

        String jobNumber = Integer.toString(job.getJobNumber());
        
        if (parseResult.optionTimes("WAIT") > 0)
        {
            String waitRequest = "wait " + parseResult.optionValue("WAIT");
            
            synchronized (fExecuteWaitList)
            {
                fExecuteWaitList.add(jobNumber);
            }
            
            STAFResult getResult = fHandle.submit2("local", "queue", "get " +
                waitRequest + " contains STAX/Job/Wait/Complete/" + jobNumber);
                
            if (getResult.rc != 0)            
            {   // There was an error in the QUEUE service submission 

                if (getResult.rc == STAFResult.Timeout)
                {
                    return new STAXSubmitResult(
                        new STAFResult(getResult.rc, jobNumber), job);
                }            
                else
                {
                    return new STAXSubmitResult(
                        new STAFResult(getResult.rc, getResult.result), job);
                }
            }

            if (parseResult.optionTimes("RETURNRESULT") > 0)
            {
                if (info.diagEnabled == 1)
                {
                    // Record diagnostics because output for an EXECUTE request
                    // with options WAIT and RETURNRESULT specified will change
                    // in STAF V3.0

                    String trigger = new String(
                        "STAF/V3.0-Mig/STAX EXECUTE WAIT RETURNRESULT");
                    String source  = new String(info.processName + ";" +
                                                info.machine + ";" +
                                                info.handle);
                    fHandle.submit2(
                        STAFHandle.ReqFireAndForget,
                        "local", "DIAG",
                        "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                        " SOURCE " + STAFUtil.wrapData(source));
                }

                // Return the job number and the job result

                String result = jobNumber + ";" + job.getResult().toString();

                return new STAXSubmitResult(
                    new STAFResult(STAFResult.Ok, result), job);
            }
        }


        return new STAXSubmitResult(new STAFResult(STAFResult.Ok, jobNumber),
                                    job);
    }

    private STAFResult handleList(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFCommandParseResult parseResult= fListParser.parse(info.request);
        StringBuffer result = new StringBuffer();

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 2)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 2 required for LIST request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }

        if (info.diagEnabled == 1)
        {
            // Record diagnostics because the output for a QUERY request will
            // change in STAF V3.0

            String trigger = new String("STAF/V3.0-Mig/STAX LIST");
            String source  = new String(info.processName + ";" +
                                        info.machine + ";" +
                                        info.handle);
            fHandle.submit2(
                STAFHandle.ReqFireAndForget,
                "local", "DIAG",
                "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                " SOURCE " + STAFUtil.wrapData(source));
        }

        if (parseResult.optionTimes("JOBS") != 0)
        {
            synchronized (fJobMap)
            {
                Iterator iter = fJobMap.values().iterator();

                while (iter.hasNext())
                {
                    STAXJob job = (STAXJob)iter.next();

                    result.append(job.getJobNumber() + ";");

                    // Replace any newline characters in the job name and
                    // function argument values with a blank so that the
                    // STAX Monitor does not "get confused" when it lists jobs.

                    if (job.getJobName().equals(""))
                        result.append(sNA + ";");
                    else
                        result.append(job.getJobName().replace('\n', ' ') + ";");

                    result.append(job.getStartFunction() + ";");

                    if (job.getStartFuncArgs() == null)
                    {
                        result.append("None" + lineSep);
                    }
                    else
                    {
                        result.append(job.getStartFuncArgs().
                                      replace('\n', ' ') + lineSep);
                    }
                }
            } // end synchronized (fJobMap)
        }
        else if (parseResult.optionTimes("SETTINGS") != 0)
        {
            result.append("Event Machine      : " + fEventServiceMachine +
                          lineSep);
            result.append("Event Service Name : " + fEventServiceName +
                          lineSep);
            result.append("Number of Threads  : " + fNumThreads + lineSep);
            result.append("Process Timeout    : " + fProcessTimeout + lineSep);
            result.append("Clear Logs         : " +
                          getClearLogsAsString() + lineSep);
            result.append("Log TC Elapsed Time: " +
                          getLogTCElapsedTimeAsString() + lineSep);
            result.append("Log TC Num Starts  : " + 
                          getLogTCNumStartsAsString() + lineSep);
            result.append("Log TC Start/Stop  : " +
                          getLogTCStartStopAsString() + lineSep);
        }
        else if (parseResult.optionTimes("EXTENSIONS") != 0)
        {
            // LIST EXTENSIONS

            Iterator iter = fExtensionElementMap.keySet().iterator();

            while (iter.hasNext())
            {
                String elementName = (String)iter.next();
                result.append("<" + elementName + ">;");
                result.append(fExtensionElementMap.get(elementName) + lineSep);
            }
        }
        else if (parseResult.optionTimes("EXTENSIONJARFILES") != 0)
        {
            // LIST EXTENSIONJARFILES

            synchronized (fExtensionsJarList)
            {
                Iterator iter = fExtensionsJarList.iterator();

                while (iter.hasNext())
                {
                    STAXExtension ext = (STAXExtension)iter.next();

                    result.append(ext.getJarFileName() + ";");

                    String version = ext.getExtVersion();

                    if (version.equals(""))
                        version = sNOT_PROVIDED;

                    result.append(version + ";");

                    String description = ext.getExtDescription();

                    if (description.equals(""))
                         description = sNOT_PROVIDED;

                    result.append(description + lineSep);
                }
            }   
        }
        else
        {
            STAXJob job = null;

            try
            {
                job = (STAXJob)fJobMap.get(
                       new Integer(parseResult.optionValue("JOB")));

                if (job == null)
                {
                    return new STAFResult(STAFResult.DoesNotExist,
                                          parseResult.optionValue("JOB"));
                }
            }
            catch (NumberFormatException e)
            {
                return new STAFResult(STAFResult.InvalidValue,
                                      parseResult.optionValue("JOB"));
            }

            if (parseResult.optionTimes("THREADS") != 0)
            {
                STAXThread mainThread = job.getMainThread();

                if (mainThread == null)
                {
                    // No threads are running

                    return new STAFResult(STAFResult.Ok);
                }

                result.append(mainThread.getThreadNumber() + ";");
                result.append(mainThread.getStateAsString() + lineSep);

                LinkedList iterStack = new LinkedList();

                iterStack.add(mainThread.getChildrenThreads().values().
                              iterator());

                while (iterStack.size() != 0)
                {
                    Iterator thisIter = (Iterator)iterStack.getLast();

                    if (!thisIter.hasNext())
                    {
                        iterStack.removeLast();
                    }
                    else
                    {
                        STAXThread thisThread = (STAXThread)thisIter.next();

                        for (int i = 0; i < iterStack.size(); ++i)
                            result.append("..");

                        result.append(thisThread.getThreadNumber() + ";");
                        result.append(thisThread.getStateAsString() + lineSep);

                        iterStack.add(thisThread.getChildrenThreads().values().
                                      iterator());
                    }
                }
            }
            else
            {
                int typeInstanceNum = 0;   // 0 indicates type option not found

                for (int i = 1; i <= 3; i++)
                {
                    if (!parseResult.instanceName(i).equalsIgnoreCase("LIST") &&
                        !parseResult.instanceName(i).equalsIgnoreCase("JOB"))
                    {   // Found the type option
                        typeInstanceNum = i;
                        break;
                    }
                }

                String typeName = parseResult.instanceName(typeInstanceNum);

                STAXListRequestHandler handler;

                synchronized (fListRequestMap)
                {
                    if (!fListRequestMap.containsKey(typeName.toUpperCase()))
                        return new STAFResult(STAFResult.DoesNotExist,typeName);
                    else
                        handler = (STAXListRequestHandler)fListRequestMap.
                                   get(typeName.toUpperCase());
                }

                STAXRequestSettings settings = 
                   new STAXRequestSettings(lineSep);

                return handler.handleListRequest(typeName, job, settings);
            }
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }

    private STAFResult handleQuery(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFCommandParseResult parseResult= fQueryParser.parse(info.request);
        StringBuffer result = new StringBuffer();

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 2)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 2 required for QUERY request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }

        if (info.diagEnabled == 1)
        {
            // Record diagnostics because the output for a QUERY request will
            // change in STAF V3.0

            String trigger = new String("STAF/V3.0-Mig/STAX QUERY");
            String source  = new String(info.processName + ";" +
                                        info.machine + ";" +
                                        info.handle);
            fHandle.submit2(
                STAFHandle.ReqFireAndForget,
                "local", "DIAG",
                "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                " SOURCE " + STAFUtil.wrapData(source));
        }

        if (parseResult.optionTimes("EXTENSIONJARFILE") != 0)
        {
            // QUERY EXTENSIONJARFILE <Jar File Name>

            STAXExtension ext = null;
            String extJarFileName = parseResult.optionValue("EXTENSIONJARFILE");

            synchronized (fExtensionsJarList)
            {
                Iterator iter = fExtensionsJarList.iterator();

                while (iter.hasNext())
                {
                    STAXExtension extension = (STAXExtension)iter.next();

                    if (extension.getJarFileName().equals(extJarFileName))
                    {
                        ext = extension;
                        break;
                    }
                }
            }

            if (ext == null)
            {
                return new STAFResult(STAFResult.DoesNotExist, extJarFileName);
            }
            
            return new STAFResult(STAFResult.Ok, queryExtension(ext));
        }
        else if (parseResult.optionTimes("EXTENSIONJARFILES") != 0)
        {
            // QUERY EXTENSIONSJARFILES

            synchronized (fExtensionsJarList)
            {
                Iterator iter = fExtensionsJarList.iterator();

                while (iter.hasNext())
                {
                    STAXExtension ext = (STAXExtension)iter.next();

                    result.append(queryExtension(ext) + lineSep);
                }
            }

            return new STAFResult(STAFResult.Ok, result.toString());
        }
        
        STAXJob job = null;

        try
        {
            job = (STAXJob)fJobMap.get(
                   new Integer(parseResult.optionValue("JOB")));

            if (job == null)
            {
                return new STAFResult(STAFResult.DoesNotExist,
                                      parseResult.optionValue("JOB"));
            }
        }
        catch (NumberFormatException e)
        {
            return new STAFResult(STAFResult.InvalidValue,
                                  parseResult.optionValue("JOB"));
        }

        if (parseResult.optionTimes("THREAD") != 0)
        {
            // QUERY JOB <Job ID> THREAD <ThreadID>

            Integer threadID;

            try
            {
                threadID = new Integer(parseResult.optionValue("THREAD"));
            }
            catch (NumberFormatException e)
            {
                return new STAFResult(STAFResult.InvalidValue,
                                      parseResult.optionValue("THREAD"));
            }

            // XXX: We should probably maintain a set/map of all the threads
            //      in the job in STAXJob instead of having to go searching
            //      through all the children recursively

            boolean mainThread = false;
            STAXThread thread = job.getMainThread();

            if (thread == null)
            {
                // No threads are running

                return new STAFResult(STAFResult.DoesNotExist,
                                      parseResult.optionValue("THREAD"));
            }

            if (thread.getThreadNumberAsInteger().equals(threadID))
            {
                mainThread = true;
            }  
            else
            {
                LinkedList iterStack = new LinkedList();

                iterStack.add(thread.getChildrenThreads().values().
                              iterator());

                boolean threadFound = false;

                while (!threadFound && iterStack.size() != 0)
                {
                    Iterator thisIter = (Iterator)iterStack.getLast();

                    if (!thisIter.hasNext())
                    {
                        iterStack.removeLast();
                    }
                    else
                    {
                        thread = (STAXThread)thisIter.next();
                        if (thread.getThreadNumberAsInteger().equals(threadID))
                        {
                            threadFound = true;
                        } 
                        else
                            iterStack.add(thread.getChildrenThreads().
                                      values().iterator());
                    }
                }
                if (!threadFound)
                {
                    return new STAFResult(STAFResult.DoesNotExist,
                                          parseResult.optionValue("THREAD"));
                } 
            }

            result.append("Thread ID  : " + parseResult.optionValue("THREAD") +
                          lineSep);

            if (mainThread)
                result.append("Parent TID : " + lineSep);
            else
                result.append("Parent TID : " + thread.getParentThread().
                              getThreadNumber() + lineSep);

            result.append("Start Date : " +
                         thread.getStartTimestamp().getDateString() + lineSep);
            result.append("Start Time : " + 
                      thread.getStartTimestamp().getTimeString() + lineSep);
            
            result.append(lineSep + "Call Stack : " + lineSep + lineSep);

            StringBuffer buff = new StringBuffer();
            
            thread.visitActions(new STAXVisitorHelper(buff, lineSep)
            {
                public void visit(Object o, Iterator iter)
                {
                    STAXAction action = (STAXAction)o;
                    String className = action.getClass().getName();

                    if (className.startsWith(PACKAGE_NAME))
                        className = className.substring(PACKAGE_NAME.length());

                    String classSuffix = "Action";

                    if (className.endsWith(classSuffix))
                    {
                        className = className.substring(
                            0, className.length() - classSuffix.length());
                    }

                    StringBuffer buff = (StringBuffer)fData;

                    buff.append("  " + className + ": " + action.getInfo() +
                                (String)fData2);
                }
            });

            buff.append(lineSep + "Condition Stack : " + lineSep + lineSep);

            thread.visitConditions(new STAXVisitorHelper(buff, lineSep)
            {
                public void visit(Object o, Iterator iter)
                {
                    STAXCondition condition = (STAXCondition)o;
                    String className = condition.getClass().getName();

                    if (className.startsWith(PACKAGE_NAME))
                        className = className.substring(PACKAGE_NAME.length());
                    
                    String classSuffix = "Condition";

                    if (className.endsWith(classSuffix))
                    {
                        className = className.substring(
                            0, className.length() - classSuffix.length());
                    }

                    StringBuffer buff = (StringBuffer)fData;

                    buff.append("  " + className + ": " + "Source=" +
                                condition.getSource() + ", Priority=" +
                                condition.getPriority() + (String)fData2);
                }
            });

            result.append(buff);
        }
        else if (parseResult.numInstances() == 2)
        {   // QUERY JOB <Job ID>

            result.append("Job ID        : " + job.getJobNumber() + lineSep);
            if (job.getJobName().equals(""))
                result.append("Job Name      : " + sNA + lineSep);
            else 
                result.append("Job Name      : " + job.getJobName() + lineSep);

            result.append("XML File Name : " + job.getXmlFile() + lineSep);
            result.append("File Machine  : " + job.getXmlMachine() + lineSep);
            result.append("Function      : " + job.getStartFunction() +
                          lineSep);
            result.append("Arguments     : " + job.getStartFuncArgs() +
                          lineSep);

            Iterator iter = job.getScripts().iterator();
            int scriptNumber = 1;
            while (iter.hasNext())
                result.append("Script #" + scriptNumber++ + "     : " + 
                              (String)iter.next() + lineSep);

            iter = job.getScriptFiles().iterator();
            scriptNumber = 1;
            while (iter.hasNext())
                result.append("Script File #" + scriptNumber++ + ": " + 
                              (String)iter.next() + lineSep);

            if (! job.getScriptFileMachine().equals(""))
                result.append("Script Machine: " + job.getScriptFileMachine() +
                              lineSep);

            result.append("Start Date    : " + job.getStartTimestamp().
                          getDateString() + lineSep);
            result.append("Start Time    : " + job.getStartTimestamp().
                          getTimeString() + lineSep);
            result.append("Source Machine: " + job.getSourceMachine() +
                          lineSep);

            // Get information from registered query job request handlers

            STAFResult handlerResult = null;

            STAXRequestSettings settings = new STAXRequestSettings(lineSep);
            
            synchronized (fQueryJobRequestList)
            {
                Iterator listIter = fQueryJobRequestList.iterator();

                while(listIter.hasNext())
                {
                    handlerResult = ((STAXQueryRequestHandler)listIter.next()).
                        handleQueryJobRequest(job, settings);

                    if (handlerResult.rc != STAFResult.Ok)
                        return handlerResult;

                    result.append(handlerResult.result);
                } 
            }
            
            result.append("Clear Logs         : " +
                          job.getClearLogsAsString() + lineSep);
            result.append("Log TC Elapsed Time: " +
                          job.getLogTCElapsedTimeAsString() + lineSep);
            result.append("Log TC Num Starts  : " +
                          job.getLogTCNumStartsAsString() + lineSep);
            result.append("Log TC Start/Stop  : " +
                          job.getLogTCStartStopAsString() + lineSep);
        }
        else
        {   // QUERY JOB <Job ID> <TypeName> <TypeValue>
            int typeInstanceNum = 0;   // 0 indicates type option not found

            for (int i = 1; i <= 3; i++)
            {
                if (!parseResult.instanceName(i).equalsIgnoreCase("QUERY") &&
                    !parseResult.instanceName(i).equalsIgnoreCase("JOB"))
                {   // Found the type option
                   typeInstanceNum = i;
                   break;
                }
            }

            // Check to see if there is a handler for the type option.

            String typeName = parseResult.instanceName(typeInstanceNum);
            
            STAXQueryRequestHandler handler = null;

            synchronized (fQueryRequestMap)
            {
                if (!fQueryRequestMap.containsKey(typeName.toUpperCase()))
                    return new STAFResult(STAFResult.DoesNotExist, typeName);
                else
                    handler = (STAXQueryRequestHandler)fQueryRequestMap.
                               get(typeName.toUpperCase());
            }

            STAXRequestSettings settings = new STAXRequestSettings(lineSep);

            return handler.handleQueryRequest(typeName, 
                parseResult.instanceValue(typeInstanceNum), job, settings);
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }

    private String queryExtension(STAXExtension ext)
    {
        StringBuffer result = new StringBuffer("");

        result.append("Jar File Name         : " + ext.getJarFileName() +
                      lineSep);

        String version = ext.getExtVersion();
        if (version.equals("")) version = sNOT_PROVIDED;
        result.append("Version               : " + version + lineSep);

        String description = ext.getExtDescription();
        if (description.equals("")) description = sNOT_PROVIDED;
        result.append("Description           : " + description + lineSep);
        
        String serviceVersion = ext.getRequiredServiceVersion();

        if ((serviceVersion.equals(sNONE)) &&
            (ext.getSupportedElementMap().size() == 0) &&
            (ext.getUnsupportedElementMap().size() == 0))
        {
            // Don't show the "Service Version Prereq" line if no elements
        }
        else
        {
            result.append("Service Version Prereq: " + serviceVersion +
                          lineSep);
        }

        String monitorVersion = ext.getRequiredMonitorVersion();

        if ((monitorVersion.equals(sNONE)) &&
            (ext.getMonitorExtensionMap().size() == 0))
        {
            // Don't show the "Monitor Version Prereq" line
        }
        else
        {
            result.append("Monitor Version Prereq: " +
                          monitorVersion + lineSep);
        }

        HashMap parmMap = ext.getParmMap();
        Iterator iter = parmMap.keySet().iterator();
        int num = 1;

        while (iter.hasNext())
        {
            String name = (String)iter.next();
            String value = (String)parmMap.get(name);

            result.append("Parameter #" + num);

            if (num < 10)
                result.append("          : ");
            else if (num < 100)
                result.append("         : ");
            else
                result.append("        : ");

            result.append(name + "=" + value + lineSep);
            num++;
        }

        iter = ext.getSupportedElementMap().keySet().iterator();
        num = 1;

        while (iter.hasNext())
        {
            String name = (String)iter.next();

            result.append("Element Name #" + num);

            if (num < 10)
                result.append("       : ");
            else if (num < 100)      
                result.append("      : ");
            else
                result.append("     : ");

            result.append(name + lineSep);
            num++;
        }

        iter = ext.getUnsupportedElementMap().keySet().iterator();

        while (iter.hasNext())
        {
            String name = (String)iter.next();

            result.append("Element Name #" + num);

            if (num < 10)
                result.append("       : ");
            else if (num < 100)
                result.append("      : ");
            else
                result.append("     : ");

            result.append(sNA + " " + name + lineSep);
            num++;
        }

        iter = ext.getMonitorExtensionMap().keySet().iterator();
        num = 1;

        while (iter.hasNext())
        {
            String name = (String)iter.next();

            result.append("Monitor Extension #" + num);

            if (num < 10)
                result.append("  : ");
            else if (num < 100)
                result.append(" : ");
            else
                result.append(": ");

            result.append(name + lineSep);
            num++;
        }

        return result.toString();
    }

    private STAFResult handleVersion(STAFServiceInterfaceLevel4.
                                     RequestInfo info)
    {
        STAFCommandParseResult parseResult= fVersionParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 2)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 2 required for VERSION request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }

        return new STAFResult(STAFResult.Ok, fVersion);
    }

    private STAFResult handleGet(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFCommandParseResult parseResult= fGetParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 2)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 2 required for GET request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }

        if (parseResult.optionTimes("STAX-EXTENSIONS-FILE") > 0)
        {
            STAXExtensionFileParser parser = null;

            try
            {
                parser = new STAXExtensionFileParser(this);
            }
            catch (Exception e)
            {
                return new STAFResult(
                    STAFResult.JavaError, // XXX: What kind of error?
                    "Caught Exception: " + e.getMessage()); 
            }

            return new STAFResult(STAFResult.Ok, parser.getDTD());
        }

        return new STAFResult(STAFResult.Ok, fDTD);
    }

    private STAFResult handleSet(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFCommandParseResult parseResult= fSetParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 5)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 5 required for SET request. Requesting " +
                "machine's trust level: " +  info.trustLevel); 
        }

        if (info.diagEnabled == 1)
        {
            // Record diagnostics because the output for a SET request will
            // change in STAF V3.0

            String trigger = new String("STAF/V3.0-Mig/STAX SET");
            String source  = new String(info.processName + ";" +
                                        info.machine + ";" +
                                        info.handle);
            fHandle.submit2(
                STAFHandle.ReqFireAndForget,
                "local", "DIAG",
                "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
                " SOURCE " + STAFUtil.wrapData(source));
        }

        StringBuffer result = new StringBuffer("");
        STAFResult resolvedValue;

        // Handle CLEARLOGS setting
        
        if (parseResult.optionTimes("CLEARLOGS") > 0)
        {
            // Resolve the value specified for CLEARLOGS
            resolvedValue = resolveValue(info.machine, info.localMachine,
                            parseResult.optionValue("CLEARLOGS"),
                            info.handle);

            if (resolvedValue.rc != 0)
                return resolvedValue;

            if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                fClearLogs = true;
            else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                fClearLogs = false;
            else
            {
                return new STAFResult(
                    STAFResult.InvalidValue, 
                    "CLEARLOGS value must be ENABLED or DISABLED.  " +
                    "Invalid value: " + resolvedValue.result);
            }

            result.append("Clear Logs         : " +
                          getClearLogsAsString() + lineSep);
        }

        // Handle LOGTCELAPSEDTIME setting
        
        if (parseResult.optionTimes("LOGTCELAPSEDTIME") > 0)
        {
            // Resolve the value specified for LOGTCELAPSEDTIME
            resolvedValue = resolveValue(info.machine, info.localMachine,
                            parseResult.optionValue("LOGTCELAPSEDTIME"),
                            info.handle);

            if (resolvedValue.rc != 0)
                return resolvedValue;

            if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                fLogTCElapsedTime = true;
            else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                fLogTCElapsedTime = false;
            else
            {
                return new STAFResult(
                    STAFResult.InvalidValue, 
                    "LOGTCELAPSEDTIME value must be ENABLED or DISABLED.  " +
                    "Invalid value: " + resolvedValue.result);
            }

            result.append("Log TC Elapsed Time: " +
                          getLogTCElapsedTimeAsString() + lineSep);
        }

        // Handle LOGTCNUMSTARTS setting
        
        if (parseResult.optionTimes("LOGTCNUMSTARTS") > 0)
        {
            // Resolve the value specified for LOGTCNUMSTARTS
            resolvedValue = resolveValue(info.machine, info.localMachine,
                            parseResult.optionValue("LOGTCNUMSTARTS"),
                            info.handle);

            if (resolvedValue.rc != 0)
                return resolvedValue;

            if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                fLogTCNumStarts = true;
            else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                fLogTCNumStarts = false;
            else
            {
                return new STAFResult(
                    STAFResult.InvalidValue, 
                    "LOGTCNUMSTARTS value must be ENABLED or DISABLED.  " +
                    "Invalid value: " + resolvedValue.result);
            }

            result.append("Log TC Num Starts  : " +
                          getLogTCNumStartsAsString() + lineSep);
        }

        // Handle LOGTCSTARTSTOP setting
        
        if (parseResult.optionTimes("LOGTCSTARTSTOP") > 0)
        {
            // Resolve the value specified for LOGTCSTARTSTOP
            resolvedValue = resolveValue(info.machine, info.localMachine,
                            parseResult.optionValue("LOGTCSTARTSTOP"),
                            info.handle);

            if (resolvedValue.rc != 0)
                return resolvedValue;

            if (resolvedValue.result.equalsIgnoreCase("ENABLED"))
                fLogTCStartStop = true;
            else if (resolvedValue.result.equalsIgnoreCase("DISABLED"))
                fLogTCStartStop = false;
            else
            {
                return new STAFResult(
                    STAFResult.InvalidValue, 
                    "LOGTCSTARTSTOP value must be ENABLED or DISABLED.  " +
                    "Invalid value: " + resolvedValue.result);
            }

            result.append("Log TC Start/Stop  : " +
                          getLogTCStartStopAsString() + lineSep);
        }

        return new STAFResult(STAFResult.Ok, result.toString());
    }

    private STAFResult handleGenericRequest(STAFServiceInterfaceLevel4.
                                            RequestInfo info)
    {
        // Find first generic request handler registered for this request with
        // a matching parser.

        STAXGenericRequest genReq = null;
        STAXGenericRequestHandler handler = null;

        synchronized (fGenericRequestList)
        {
            Iterator listIter = fGenericRequestList.iterator();

            while(listIter.hasNext())
            {
                // XXX: When add a new Service Interface Level class (other than
                // STAFServiceInterfaceLevel4), need to add code here to pass
                // the right level of information to handleRequest.

                // Attempt to parse and handle the request with this handler
                STAFResult parseResult = ((STAXGenericRequest)listIter.next()).
                                         fHandler.handleRequest(info, this); 
                
                // Check if found matching parser and handler.
                // If parser returns an non-blank result, that indicates
                // it is the right parser, just an invalid request.
                if ((parseResult.rc != STAFResult.InvalidRequestString) ||
                    (! parseResult.result.equals("")))
                {
                    // Found matching parser and handler
                    return parseResult;
                }
            } 
        }
        
        return new STAFResult(STAFResult.InvalidRequestString,
                            "No element is registered to handle this request");
    }

    private STAFResult handleHelp(STAFServiceInterfaceLevel4.RequestInfo info)
    {
        STAFCommandParseResult parseResult= fHelpParser.parse(info.request);
        String result = new String();

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        result =  
          "STAX Service Help: " + lineSep + lineSep +
          "EXECUTE   < <FILE <XML File Name> [MACHINE <Machine Name>]> " +
          "| DATA <xml data> >" + lineSep +
          "          [JOBNAME <Job Name>] [FUNCTION <Function ID>] " +
          "[ARGS <Arguments>]" + lineSep +
          "          [SCRIPTFILE <File Name>... [SCRIPTFILEMACHINE " +
          "<machine name>]] " + lineSep +
          "          [SCRIPT <Python Code>]... " +
          "[CLEARLOGS [<Enabled | Disabled>]] " + lineSep +
          "          [ HOLD | TEST | WAIT [Timeout] [RETURNRESULT] ]" +
          lineSep +
          "          [LOGTCELAPSEDTIME <Enabled | Disabled>]" + lineSep +
          "          [LOGTCNUMSTARTS <Enabled | Disabled>]" + lineSep +
          "          [LOGTCSTARTSTOP <Enabled | Disabled>]" + lineSep +
          lineSep +
          "GET       DTD" + lineSep + lineSep +
          "LIST      JOBS | SETTINGS | EXTENSIONS | EXTENSIONJARFILES |" +
          lineSep +
          "          JOB <Job ID> <" + fHelpListJobOptions + ">" + 
          lineSep + lineSep +
          "QUERY     EXTENSIONJARFILE <Jar File Name> | EXTENSIONJARFILES |" +
          lineSep +  
          "          JOB <Job ID> [" + fHelpQueryJobOptions + "]" + 
          lineSep + lineSep +
          fHelpGenericRequests +
          "SET       [CLEARLOGS <Enabled | Disabled>]" + lineSep +
          "          [LOGTCELAPSEDTIME <Enabled | Disabled>]" + lineSep +
          "          [LOGTCNUMSTARTS <Enabled | Disabled>]" + lineSep +
          "          [LOGTCSTARTSTOP <Enabled | Disabled>]" + lineSep +
          lineSep +
          "VERSION" + lineSep + lineSep +
          "HELP" + lineSep;

        return new STAFResult(STAFResult.Ok, result);
    }

    private int handleParms(STAFServiceInterfaceLevel4.InitInfo info)
        throws STAFException
    {
        STAFCommandParseResult parseResult= fParmsParser.parse(info.parms);

        String errmsg = "ERROR:  Service Configuration Error for Service " +
                        fServiceName + lineSep + "STAX::handleParms() - ";

        if (parseResult.rc != STAFResult.Ok)
        {
            System.out.println(
                errmsg + "PARMS parsing failed with RC=" +
                STAFResult.InvalidRequestString +
                " Result=" + parseResult.errorBuffer);

            return STAFResult.ServiceConfigurationError;
        }

        if (parseResult.optionTimes("EVENTSERVICEMACHINE") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok, 
                parseResult.optionValue("EVENTSERVICEMACHINE"));  

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var", 
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }
            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving EVENTSERVICEMACHINE.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
            fEventServiceMachine = resolvedResult.result;
        }

        if (parseResult.optionTimes("EVENTSERVICENAME") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok, 
                parseResult.optionValue("EVENTSERVICENAME"));  

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var", 
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }
            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving EVENTSERVICENAME.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
            fEventServiceName = resolvedResult.result;
        }

        if (parseResult.optionTimes("NUMTHREADS") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("NUMTHREADS"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }
            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving NUMTHREADS value.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
            try
            {
                fNumThreads = (new Integer(resolvedResult.result)).
                                   intValue();
                if (fNumThreads < 2)
                {
                    System.out.println(
                        errmsg + "NUMTHREADS value must be > 1.  " +
                        "NUMTHREADS=" + fNumThreads);

                    return STAFResult.ServiceConfigurationError;
                }
            }
            catch (NumberFormatException e)
            {
                System.out.println(
                    errmsg + "NUMTHREADS value must be numeric.  " +
                    "NUMTHREADS=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("EXTENSIONXMLFILE") > 0)
        {
            STAFResult res = new STAFResult(STAFResult.Ok, 
                parseResult.optionValue("EXTENSIONXMLFILE"));  

            if (res.result.indexOf("{") >= 0)
            { 
                res = fHandle.submit2("local", "var", 
                    "resolve " + STAFUtil.wrapData(res.result));
            }
            if (res.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving EXTENSIONXMLFILE value. RC=" +
                    res.rc + " Result=" + res.result);

                return STAFResult.ServiceConfigurationError;
            }

            String extFileName = res.result;

            res = fHandle.submit2("local", "fs", "get file " + extFileName);

            if (res.rc != 0)           
            {   // There was an error in the FS service submission 
                System.out.println(
                    errmsg + "EXTENSIONXMLFILE " + extFileName +
                    " does not exist, RC: " + res.rc);

                return STAFResult.ServiceConfigurationError;
            }
            
            String extFileContents = res.result;
            STAXExtensionFileParser parser = null;

            try
            {
                parser = new STAXExtensionFileParser(this);

                parser.parse(new InputSource(
                    new StringReader(extFileContents)));
            }
            catch (STAXException e)
            {
                System.out.println(
                    errmsg + "Error parsing EXTENSIONXMLFILE " +
                    extFileName + lineSep +
                    "Caught " + e.getClass().getName() + ": " + lineSep +
                    e.getMessage());

                return STAFResult.ServiceConfigurationError;
            }
            catch (SAXNotRecognizedException e)
            {
                System.out.println(
                    errmsg + "Error parsing EXTENSIONXMLFILE " + extFileName +
                    lineSep + "Caught SAXNRE: " + e.getMessage());

                return STAFResult.ServiceConfigurationError;
            }
            catch (SAXNotSupportedException e)
            {
                System.out.println(
                    errmsg + "Error parsing EXTENSIONXMLFILE " + extFileName +
                    lineSep + "Caught SAXNSE: " + e.getMessage());

                return STAFResult.ServiceConfigurationError;
            }
            catch (Exception e)
            {
                System.out.println(
                    errmsg + "Error parsing EXTENSIONXMLFILE " + extFileName +
                    lineSep + "Caught Exception: " + e.getMessage());

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("EXTENSIONFILE") > 0)
        {
            STAFResult res = new STAFResult(STAFResult.Ok, 
                parseResult.optionValue("EXTENSIONFILE"));  

            if (res.result.indexOf("{") >= 0)
            { 
                res = fHandle.submit2("local", "var", 
                    "resolve " + STAFUtil.wrapData(res.result));
            }
            if (res.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving EXTENSIONFILE value.  RC=" +
                    res.rc + " Result=" + res.result);

                return STAFResult.ServiceConfigurationError;
            }

            String extFileName = res.result;

            res = fHandle.submit2("local", "fs", "get file " + extFileName);

            if (res.rc != 0)           
            {   // There was an error in the FS service submission 
                System.out.println(
                    errmsg + "EXTENSIONFILE " + extFileName +
                    " does not exist, RC: " + res.rc);

                return STAFResult.ServiceConfigurationError;
            }
            
            String extFileContents = res.result;

            res = readExtensionData(extFileContents);

            if (res.rc != 0)
            {
                System.out.println(res.result);
                return res.rc;
            }
        }
        
        if (parseResult.optionTimes("EXTENSION") > 0)
        {
            for (int i = 1; i <= parseResult.optionTimes("EXTENSION"); ++i)
            {   
                STAFResult res = new STAFResult(STAFResult.Ok, 
                    parseResult.optionValue("EXTENSION", i));  

                if (res.result.indexOf("{") >= 0)
                { 
                    res = fHandle.submit2("local", "var", 
                        "resolve " + STAFUtil.wrapData(res.result));
                }
                if (res.rc != STAFResult.Ok)
                {
                    System.out.println(
                        errmsg + "Error resolving EXTENSION value.  RC=" +
                        res.rc + " Result=" + res.result);

                    return STAFResult.ServiceConfigurationError;
                }
            
                String extValue = res.result;

                res = readExtensionData(extValue);

                if (res.rc != 0)
                {
                    System.out.println(res.result);
                    return res.rc;
                }
            }
        }
        
        if (parseResult.optionTimes("PROCESSTIMEOUT") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("PROCESSTIMEOUT"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }
            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving PROCESSTIMEOUT value.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
            try
            {
                fProcessTimeout = (new Integer(resolvedResult.result)).
                                   intValue();
                                   
                if (fProcessTimeout < 1000)
                {
                    System.out.println(
                        errmsg + "PROCESSTIMEOUT value must be at least " +
                        "1000 (1 second).  PROCESSTIMEOUT=" + fProcessTimeout);

                    return STAFResult.ServiceConfigurationError;
                }
            }
            catch (NumberFormatException e)
            {
                System.out.println(
                    errmsg + "PROCESSTIMEOUT value must be numeric.  " +
                    "PROCESSTIMEOUT=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("CLEARLOGS") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("CLEARLOGS"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving CLEARLOGS value.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            if (resolvedResult.result.equalsIgnoreCase("ENABLED"))
            {
                fClearLogs = true;
            }
            else if (resolvedResult.result.equalsIgnoreCase("DISABLED"))
            {
                fClearLogs = false;
            }
            else
            {
                System.out.println(
                    errmsg + "CLEARLOGS must be set to Enabled or Disabled. " +
                    " Invalid value: " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("LOGTCELAPSEDTIME") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("LOGTCELAPSEDTIME"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving LOGTCELAPSEDTIME value.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            if (resolvedResult.result.equalsIgnoreCase("ENABLED"))
            {
                fLogTCElapsedTime = true;
            }
            else if (resolvedResult.result.equalsIgnoreCase("DISABLED"))
            {
                fLogTCElapsedTime = false;
            }
            else
            {
                System.out.println(
                    errmsg + "LOGTCELAPSEDTIME must be set to Enabled or " +
                    "Disabled.  Invalid value: " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("LOGTCNUMSTARTS") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("LOGTCNUMSTARTS"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving LOGTCNUMSTARTS value.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            if (resolvedResult.result.equalsIgnoreCase("ENABLED"))
            {
                fLogTCNumStarts = true;
            }
            else if (resolvedResult.result.equalsIgnoreCase("DISABLED"))
            {
                fLogTCNumStarts = false;
            }
            else
            {
                System.out.println(
                    errmsg + "LOGTCNUMSTARTS must be set to Enabled or " +
                    "Disabled.  Invalid value: " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("LOGTCSTARTSTOP") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("LOGTCSTARTSTOP"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving LOGTCSTARTSTOP value.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            if (resolvedResult.result.equalsIgnoreCase("ENABLED"))
            {
                fLogTCStartStop = true;
            }
            else if (resolvedResult.result.equalsIgnoreCase("DISABLED"))
            {
                fLogTCStartStop = false;
            }
            else
            {
                System.out.println(
                    errmsg + "LOGTCSTARTSTOP must be set to Enabled or " +
                    "Disabled.  Invalid value: " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        if (parseResult.optionTimes("CACHEPYTHONCODE") > 0)
        {
            STAFResult resolvedResult = new STAFResult(STAFResult.Ok,
                parseResult.optionValue("CACHEPYTHONCODE"));

            if (resolvedResult.result.indexOf("{") >= 0)
            { 
                resolvedResult = fHandle.submit2("local", "var",
                    "resolve " + STAFUtil.wrapData(resolvedResult.result));
            }

            if (resolvedResult.rc != STAFResult.Ok)
            {
                System.out.println(
                    errmsg + "Error resolving CACHEPYTHONCODE.  RC=" +
                    resolvedResult.rc + " Result=" + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }

            if (resolvedResult.result.equalsIgnoreCase("ENABLED"))
            {
                CACHE_PYTHON_CODE = true;
            }
            else if (resolvedResult.result.equalsIgnoreCase("DISABLED"))
            {
                CACHE_PYTHON_CODE = false;
            }
            else
            {
                System.out.println(
                    errmsg + "CACHEPYTHONCODE must be set to Enabled or " +
                    "Disabled.  Invalid value: " + resolvedResult.result);

                return STAFResult.ServiceConfigurationError;
            }
        }

        return 0;
    }

    // STAXJobCompleteListener method

    public void jobComplete(STAXJob job)
    {
        String jobNumber = Integer.toString(job.getJobNumber());
        
        synchronized (fJobMap)
        {
            fJobMap.remove(job.getJobNumberAsInteger());
        }
        
        synchronized (fExecuteWaitList)
        {
            if (fExecuteWaitList.contains(jobNumber))
            {
                fExecuteWaitList.remove(fExecuteWaitList.indexOf(jobNumber));
                
                STAFResult Result = fHandle.submit2("local", "queue", 
                    "queue message STAX/Job/Wait/Complete/" + jobNumber);
            }
        }
    }
        
    private STAFResult readExtensionData(String value)
    {
        String line;
        BufferedReader br = new BufferedReader(new StringReader(value));
        StringTokenizer st;

        try
        {
            while ((line = br.readLine()) != null)
            {
                String jarFileName;
                int pos = line.indexOf('#');
                HashMap includeElementMap = new HashMap();

                if (pos == -1)
                {   
                    jarFileName = line.trim();
                }
                else if (pos == 0)
                {
                    continue;  // Skip comment indicated by # in first position
                }
                else
                {
                    jarFileName = line.substring(0, pos - 1).trim();

                    st = new StringTokenizer(line.substring(pos + 1));
                    
                    while (st.hasMoreTokens())
                    {
                        includeElementMap.put(st.nextToken(), null);
                    }
                }

                if (jarFileName.equals(""))
                    continue;  // Skip blank lines (& lines without a jar file)

                if (!jarFileName.toLowerCase().endsWith(".jar"))
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::readExtensionData() - " +
                        "Extension jar file: " + jarFileName + lineSep +
                        "Extension jar file name must end in .jar");
                }

                fExtensionsJarList.add(new STAXExtension(
                    jarFileName, includeElementMap));
            }
        }
        catch (Exception e)
        {
            return new STAFResult(STAFResult.InvalidValue, e.toString());
        }
        finally
        {
            try
            {
                br.close();
            }
            catch (IOException e)
            { /* Do nothing */ }
        }

        return new STAFResult(STAFResult.Ok, "");
    }

    private STAFResult loadExtensions()
    {
        // Map used to make sure no duplicate extensions are loaded
        Map extElementMap = new HashMap();  

        // Iterate through the Jar File Extensions List and load
        // the specified extension elements

        Iterator iter = fExtensionsJarList.iterator();

        while (iter.hasNext())
        {
            STAXExtension ext = (STAXExtension)iter.next();

            JarFile jarFile = null;
            String jarFileName = ext.getJarFileName();

            try
            {
                jarFile = new JarFile(jarFileName);
            }
            catch (IOException e)
            {
                return new STAFResult(
                    STAFResult.ServiceConfigurationError,
                    "ERROR:  Service Configuration Error for Service=" +
                    fServiceName + lineSep +
                    "STAX::loadExtensions() - Extension jar file: " +
                    jarFileName + lineSep +
                    "Unable to access extension jar file." + lineSep +
                    e.toString());
            }

            Manifest manifest = null;

            try
            {
                manifest = jarFile.getManifest();
            }
            catch (IOException e)
            {
                return new STAFResult(
                    STAFResult.ServiceConfigurationError,
                    "ERROR:  Service Configuration Error for Service=" +
                    fServiceName + lineSep +
                    "STAX::loadExtensions() - ERROR: " +
                    "Extension jar file: " + jarFileName + lineSep +
                    "Unable to access manifest in the extension jar file." +
                    lineSep + e.toString());
            }

            Map manifestEntryMap = manifest.getEntries();

            boolean allElements = true;

            Map includeElementMap = ext.getIncludeElementMap();
            Map excludeElementMap = ext.getExcludeElementMap();

            if (includeElementMap.size() != 0)
            {
                allElements = false;

                // Verify that all specified elements have a manifest entry
                // called staf/stax/extension/<element> in the jar file

                Iterator elemIter = includeElementMap.keySet().iterator();
                while (elemIter.hasNext())
                {
                    String elementName = (String)elemIter.next();
                    if (!manifestEntryMap.containsKey(
                        STAX_EXTENSION + elementName))
                    {
                        return new STAFResult(
                            STAFResult.ServiceConfigurationError,
                            "ERROR:  Service Configuration Error for Service=" +
                            fServiceName + lineSep +
                            "STAX::loadExtensions() - Extension jar file: " +
                            jarFileName + lineSep +
                            "Include element '" + elementName +
                            "' does not have a " + STAX_EXTENSION +
                            " manifest entry in the extension jar file.");
                    }
                }
            }
            else if (excludeElementMap.size() != 0)
            {
                allElements = false;

                // Verify that all specified elements have a manifest entry
                // called staf/stax/extension/<element> in the jar file

                Iterator elemIter = ext.getExcludeElementMap().
                    keySet().iterator();

                while (elemIter.hasNext())
                {
                    String elementName = (String)elemIter.next();
                    if (!manifestEntryMap.containsKey(
                        STAX_EXTENSION + elementName))
                    {
                        return new STAFResult(
                            STAFResult.ServiceConfigurationError,
                            "ERROR:  Service Configuration Error for Service=" +
                            fServiceName + lineSep +
                            "STAX::loadExtensions() - Extension jar file: " +
                            jarFileName + lineSep +
                            "Exclude element '" + elementName +
                            "' does not have a " + STAX_EXTENSION + 
                            " manifest entry in the extension jar file.");
                    }
                }
            }
            
            // Create a class loader used to load STAX so it can be
            // passed to the STAXExtensionClassLoader to load STAX classes

            Class c = this.getClass();
            ClassLoader parentClassLoader = c.getClassLoader();

            // Iterate through the staf/stax/extension entries in the manifest

            Iterator mfIter = manifestEntryMap.keySet().iterator();

            // Load the extension class

            ClassLoader loader = new STAXExtensionClassLoader(jarFile,
                                                     parentClassLoader);

            while (mfIter.hasNext())
            {
                String elementName = (String)mfIter.next();

                if (elementName.startsWith(STAX_EXTENSION_INFO))
                {
                    Attributes attrs = manifest.getAttributes(
                        STAX_EXTENSION_INFO);

                    String extVersion = attrs.getValue("Extension-Version");

                    if (extVersion != null) 
                    {
                        ext.setExtVersion(extVersion);
                    }

                    String description = attrs.getValue(
                        "Extension-Description");

                    if (description != null)
                    {
                        ext.setExtDescription(description);
                    }

                    String serviceVersion = attrs.getValue(
                        "Required-Service-Version");

                    if (serviceVersion != null)
                    {
                        ext.setRequiredServiceVersion(serviceVersion);

                        STAXVersion staxVersion = new STAXVersion(fVersion);
                        STAXVersion requiredVersion;

                        try
                        {
                            requiredVersion = new STAXVersion(serviceVersion);

                            if (staxVersion.compareTo(requiredVersion) < 0)
                            {
                                return new STAFResult(
                                    STAFResult.ServiceConfigurationError,
                                    "ERROR:  Service Configuration Error " +
                                    "for Service=" + fServiceName + lineSep +
                                    "STAX::loadExtensions() - " +
                                    "Extension jar file: " + jarFileName +
                                    lineSep + "Required-Service-Version: " +
                                    requiredVersion + lineSep +
                                    "STAX Service Version: " + staxVersion +
                                    lineSep + "The required STAX service " +
                                    "version is not installed.");
                            }
                        }
                        catch (NumberFormatException e)
                        {
                            return new STAFResult(
                                STAFResult.ServiceConfigurationError,
                                "ERROR:  Service Configuration Error for " +
                                "Service=" + fServiceName + lineSep +
                                "STAX::loadExtensions() - " +
                                "Extension jar file: " + jarFileName +
                                lineSep + "Invalid value specified for " +
                                "Required-Service-Version: " +
                                serviceVersion + lineSep + e.toString());
                        }
                    }

                    String monitorVersion = attrs.getValue(
                        "Required-Monitor-Version");

                    if (monitorVersion != null)
                    {
                        ext.setRequiredMonitorVersion(monitorVersion);
                    }

                    continue;
                }
                else if (elementName.startsWith(STAX_MONITOR_EXTENSION))
                {
                    String monitorExtName = elementName.substring(
                        STAX_MONITOR_EXTENSION.length());

                    Attributes attrs = manifest.getAttributes(
                        STAX_MONITOR_EXTENSION + monitorExtName);

                    String monitorExtClass = attrs.getValue(
                        "Extension-Class");

                    ext.setMonitorExtension(monitorExtName, monitorExtClass);

                    continue;  // No further processing is done on
                               // monitor extensions in STAX
                }
                else if (!elementName.startsWith(STAX_EXTENSION))
                    continue; // Ignore these entries
                
                elementName = elementName.substring(STAX_EXTENSION.length());

                // Put excluded elements in the unsupported element map
                // and continue

                if (!allElements)
                {
                    if (includeElementMap.size() != 0)
                    {
                        if (!includeElementMap.containsKey(elementName))
                        {
                            ext.setUnsupportedElement(elementName);
                            continue;  // Skip this element
                        }
                    }
                    else if (excludeElementMap.size() != 0)
                    {
                        if (excludeElementMap.containsKey(elementName))
                        {
                            ext.setUnsupportedElement(elementName);
                            continue;  // Skip this element
                        }
                    }
                }

                Attributes attrs = manifest.getAttributes(STAX_EXTENSION +
                                                          elementName);
                
                if (!attrs.containsKey(new Attributes.Name("Factory-Class")))
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Element '" + elementName + "' is missing attribute " +
                        "Factory-Class in the manifest for the " +
                        "extension jar file.");
                }

                String factoryClassName = attrs.getValue("Factory-Class");

                if (extElementMap.containsKey(elementName))
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Extension element '" + elementName +
                        "' is already registered with STAX and cannot " +
                        "be registerd twice.");
                }
                
                if (fActionFactoryMap.containsKey(elementName))
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Extension element '" + elementName + "'" +
                        " is already defined by STAX and cannot be replaced.");
                }

                extElementMap.put(elementName, factoryClassName);
                
                Class extFactoryClass = null;
                Object extFactoryObj = null;

                try
                {
                    extFactoryClass = loader.loadClass(factoryClassName);
                    
                    // Try using a constructor that accepts a STAX object
                    // and a Map object

                    Class [] parameterTypes = new Class[2];
                    parameterTypes[0] = this.getClass();
                    parameterTypes[1] = Class.forName("java.util.Map");
                    
                    try
                    {
                        Constructor construct =
                            extFactoryClass.getConstructor(parameterTypes);
                        Object [] initArgs = new Object[2];
                        initArgs[0] = this;
                        initArgs[1] = ext.getParmMap();
                        extFactoryObj = construct.newInstance(initArgs);
                    }
                    catch (NoSuchMethodException e)
                    {
                        if (ext.getParmMap().size() != 0)
                        {
                            return new STAFResult(
                                STAFResult.ServiceConfigurationError,
                                "ERROR:  Service Configuration Error for " +
                                "Service=" + fServiceName + lineSep +
                                "STAX::loadExtensions() - " +
                                "Extension jar file: " + jarFileName + lineSep +
                                "InvocationTargetException instantiating " +
                                "class: " + factoryClassName + lineSep +
                                "The factory class constructor does not " +
                                "support passing in a parameter map: " +
                                ext.getParmMap().toString() +
                                lineSep + e.toString());
                        }

                        // Try using a constructor that accepts a STAX object

                        parameterTypes = new Class[1];
                        parameterTypes[0] = this.getClass();

                        try
                        {
                            Constructor construct = extFactoryClass.
                                getConstructor(parameterTypes);
                            Object [] initArgs = new Object[1];
                            initArgs[0] = this;
                            extFactoryObj = construct.newInstance(initArgs);
                        }
                        catch (NoSuchMethodException e2)
                        {
                            // Extension does not have a constructor that
                            // accepts a STAX object, so try constructor 
                            // without parameters
                            extFactoryObj = extFactoryClass.newInstance();
                        }
                        catch (InvocationTargetException e2)
                        {
                            String errorMsg = "ERROR:  Service Configuration" +
                                " Error for Service=" + fServiceName + lineSep +
                                "STAX::loadExtensions() - " +
                                "Extension jar file: " + jarFileName + lineSep +
                                "InvocationTargetException instantiating " +
                                "class: " + factoryClassName;

                            if (e2.getTargetException() != null)
                            {
                                errorMsg += lineSep + "Caught " +
                                    e2.getTargetException().toString();
                            }

                            return new STAFResult(
                                STAFResult.ServiceConfigurationError, errorMsg);
                        }
                        catch (IllegalArgumentException e2)
                        {
                            return new STAFResult(
                                STAFResult.ServiceConfigurationError,
                                "ERROR:  Service Configuration Error for " +
                                "Service=" + fServiceName + lineSep +
                                "STAX::loadExtensions() - Extension jar file: " +
                                jarFileName + lineSep +
                                "IllegalArgumentException instantiating " +
                                "class: " + factoryClassName + lineSep +
                                e2.toString());
                        }
                    }
                    catch (InvocationTargetException e)
                    {
                        String errorMsg = "ERROR:  Service Configuration " +
                            "Error for Service=" + fServiceName + lineSep +
                            "STAX::loadExtensions() - " +
                            "Extension jar file: " + jarFileName + lineSep +
                            "Exception instantiating class: " +
                            factoryClassName;

                        if (e.getTargetException() != null)
                        {
                            errorMsg += lineSep + "Caught " +
                                        e.getTargetException().toString();
                        }
                        
                        return new STAFResult(
                            STAFResult.ServiceConfigurationError, errorMsg);
                    }
                    catch (IllegalArgumentException e)
                    {
                        return new STAFResult(
                            STAFResult.ServiceConfigurationError,
                            "ERROR:  Service Configuration Error for Service=" +
                            fServiceName + lineSep +
                            "STAX::loadExtensions() - Extension jar file: " +
                            jarFileName + lineSep +
                            "IllegalArgumentException instantiating " +
                            "class: " + factoryClassName + lineSep +
                            e.toString());
                    }
                }
                catch (ClassNotFoundException e)
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Cannot find class " + factoryClassName + lineSep +
                        e.toString());
                }
                catch (InstantiationException e)
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Could not instantiate class: " + factoryClassName +
                        lineSep + e.toString());
                }
                catch (IllegalAccessException e)
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Illegal access to class: " + factoryClassName +
                        lineSep + e.toString());
                }

                // Add to fActionFactoryMap
                String taskName = ((STAXActionFactory)extFactoryObj).
                                  getDTDTaskName();

                if ((taskName != null) && (!taskName.equals(elementName)))
                {
                    return new STAFResult(
                        STAFResult.ServiceConfigurationError,
                        "ERROR:  Service Configuration Error for Service=" +
                        fServiceName + lineSep +
                        "STAX::loadExtensions() - Extension jar file: " +
                        jarFileName + lineSep +
                        "Extension element '" + elementName +
                        "' does not match its getDTDTaskName() of '" +
                        taskName + "'.");
                }

                fActionFactoryMap.put(elementName, extFactoryObj);

                // Add to the support element list for this extension
                ext.setSupportedElement(elementName, factoryClassName);

                // Add to the map of all the registered extension elements
                fExtensionElementMap.put(elementName, jarFileName);
            }
        }

        return new STAFResult(STAFResult.Ok, "");
    }

    private STAFResult logToServiceLog(String level, String message)
    {
        String logName = fServiceName.toUpperCase() + "_Service";
        
        String logRequest = "LOG MACHINE LOGNAME " + 
                            STAFUtil.wrapData(logName) +
                            " LEVEL " + level + 
                            " MESSAGE " + STAFUtil.wrapData(message);
                                
        STAFResult result = fHandle.submit2(STAFHandle.ReqSync, 
                                            "LOCAL", "LOG", logRequest);

        // Check if the result was unsuccesful (except ignore RC 2 and 33)
        // Opened Bug #833816 for the RC 33 problem. 
        if (result.rc != 0 && result.rc != 2 && result.rc != 33)
        {
            System.out.println("STAX::logToServiceLog failed with RC " +
                               result.rc + " and Result " + result.result +
                               "  level: " + level +
                               "  logRequest: " + logRequest);
        }

        return result;
    }

    // Register error codes for the STAX Service with the HELP service

    private void registerHelpData(int errorNumber, String info,
                                 String description)
    {
        STAFResult res = fHandle.submit2("local", "HELP",
                         "REGISTER SERVICE " + fServiceName +
                         " ERROR " + errorNumber +
                         " INFO " + STAFUtil.wrapData(info) +
                         " DESCRIPTION " + STAFUtil.wrapData(description));
    }

    // Un-register error codes for the STAX Service with the HELP service

    private void unregisterHelpData(int errorNumber)
    {
        STAFResult res = fHandle.submit2("local", "HELP",
                         "UNREGISTER SERVICE " + fServiceName +
                         " ERROR " + errorNumber);
    }
    
    /***********************************************************************/
    /* resolveValue - Uses the var Service to resolve an option value.     */
    /*                                                                     */
    /* accepts: the submitting machine name, the local machine name,       */
    /*         an unresolved option value, the submitting process' handle  */
    /*                                                                     */   
    /* returns: STAFResult.rc = STAFResult.Ok;                             */
    /* STAFResult.result = the resolved option value                       */
    /***********************************************************************/

    private STAFResult resolveValue(String machine, String localMachine, 
                                    String optionValue, int handle)
    {
        STAFResult resolvedResult = new STAFResult(STAFResult.Ok, optionValue);  

        if (optionValue.indexOf("{") >= 0)
        { 
            if (machine.equals(localMachine))
            {  // Request was made locally
                resolvedResult = fHandle.submit2("local", "var", "handle " 
                    + handle + " resolve " + STAFUtil.wrapData(optionValue));
            }
            else
            {  // Request was made from another system
                resolvedResult = fHandle.submit2("local", "var", "global" 
                    + " resolve " + STAFUtil.wrapData(optionValue));
            }
        }
        
        return resolvedResult;
    }

    public void registerJobManagementHandler(STAXJobManagementHandler handler)
    {
        synchronized (fJobManagementHandlerSet)
        {
            fJobManagementHandlerSet.add(handler);
        }
    }

    public void unregisterJobManagementHandler(STAXJobManagementHandler handler)
    {
        synchronized (fJobManagementHandlerSet)
        {
            fJobManagementHandlerSet.remove(handler);
        }
    }

    public void visitJobManagementHandlers(STAXVisitor visitor)
    {
        synchronized (fJobManagementHandlerSet)
        {
            Iterator iter = fJobManagementHandlerSet.iterator();

            while (iter.hasNext())
            {
                visitor.visit(iter.next(), iter);
            }
        }
    }

    public void registerListHandler(String type, STAXListRequestHandler handler)
    {
        synchronized (fListRequestMap)
        {
            if (fListRequestMap.containsKey(type))
            {
                System.out.println("In STAX.registerListHandler: type=" + 
                    type + " is already registered with handler " + 
                    fListRequestMap.get(type));
            }
            fListRequestMap.put(type.toUpperCase(), handler);
            fHelpListJobOptions += " | " + type.toUpperCase();
        }
    }

    public void registerQueryHandler(String type, String typeValue,
                                     STAXQueryRequestHandler handler)
    {
        synchronized (fQueryRequestMap)
        {
            if (fQueryRequestMap.containsKey(type))
            {
                System.out.println("In STAX.registerQueryHandler: type=" + 
                    type + " is already registered with handler " + 
                    fQueryRequestMap.get(type));
            }
            fQueryRequestMap.put(type.toUpperCase(), handler);

            if (typeValue.equals("") || typeValue.equals(null))
                typeValue = "Value";

            fHelpQueryJobOptions += " | " + type.toUpperCase() + " <" +
                                    typeValue + ">";
        }
    }

    public void registerQueryJobHandler(STAXQueryRequestHandler handler)
    {
        synchronized (fQueryJobRequestList)
        {
            fQueryJobRequestList.add(handler);
        }
    }

    public int registerGenericRequestHandler(STAXGenericRequestHandler handler,
                                             Class serviceInterfaceClass)
    {
        synchronized (fGenericRequestList)
        {
            // Check if supports service interface level class specified

            if (!serviceInterfaceClass.getName().equals(INTERFACE_LEVEL_4))
            {
                return 1;
            }
                
            fGenericRequestList.add(new STAXGenericRequest(handler, 
                                                    serviceInterfaceClass));
        
            fHelpGenericRequests += handler.getHelpInfo(lineSep) +
                                    lineSep + lineSep;
        }

        return 0;
    }

    // Helper class for processing the Generic Request List

    class STAXGenericRequest
    {
        STAXGenericRequest(STAXGenericRequestHandler handler,
                           Class serviceInterfaceClass)
        {
            fHandler = handler;
            fServiceInterfaceClass = serviceInterfaceClass;
        }

        private STAXGenericRequestHandler fHandler;
        private Class fServiceInterfaceClass;
    }

    private String fServiceName;
    private String fEventServiceMachine = "";
    private String fEventServiceName = "";
    private String fLocalMachineName = "";
    private int fNumThreads;
    private int fProcessTimeout = 60000;
    private boolean fClearLogs = false;
    private boolean fLogTCElapsedTime = true;
    private boolean fLogTCNumStarts = true;
    private boolean fLogTCStartStop = false;
    private STAFHandle fHandle = null;
    private STAFCommandParser fParmsParser = new STAFCommandParser();
    private STAFCommandParser fExecuteParser = new STAFCommandParser();
    private STAFCommandParser fListParser = new STAFCommandParser();
    private STAFCommandParser fQueryParser = new STAFCommandParser();
    private STAFCommandParser fVersionParser = new STAFCommandParser();
    private STAFCommandParser fGetParser = new STAFCommandParser();
    private STAFCommandParser fSetParser = new STAFCommandParser();
    private STAFCommandParser fHelpParser = new STAFCommandParser();
    private Object fNextJobNumberSynch = new Object();
    private int fNextJobNumber = 1;
    private HashMap fActionFactoryMap = new HashMap();
    private TreeMap fJobMap = new TreeMap();
    private STAXThreadQueue fThreadQueue = null;
    private STAXTimedEventQueue fTimedEventQueue = new STAXTimedEventQueue();
    private String fDTD = "";
    private HashMap fListRequestMap = new HashMap();
    private HashMap fQueryRequestMap = new HashMap();
    private ArrayList fQueryJobRequestList = new ArrayList();
    private ArrayList fGenericRequestList = new ArrayList();
    private TreeSet fJobManagementHandlerSet =
                    new TreeSet(new STAXObjectComparator());
    private String fHelpListJobOptions = "THREADS";
    private String fHelpQueryJobOptions = "THREAD <Thread ID>";
    private String fHelpGenericRequests = "";
    private String fJythonLibName = "";  // Jython Library directory name
    private List fExtensionsJarList = new ArrayList();
    private Map fExtensionElementMap = new TreeMap();
    private ArrayList fExecuteWaitList = new ArrayList();
}
