/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

package com.ibm.staf.service.stax;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import com.ibm.staf.STAFResult;
import com.ibm.staf.STAFHandle;
import com.ibm.staf.STAFUtil;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.Iterator;
import com.ibm.staf.service.*;

public class STAXTestcaseActionFactory implements STAXActionFactory,
                                                  STAXListRequestHandler,
                                                  STAXQueryRequestHandler,
                                                  STAXGenericRequestHandler,
                                                  STAXJobManagementHandler
{
    static final String STAX_TESTCASE = new String("Testcase");

    private static String fDTDInfo =
"\n" +
"<!--================= The Testcase Element ========================= -->\n" +
"<!--\n" +
"     Defines a testcase.  Used in conjunction with the tcstatus\n" +
"     element to mark the status for a testcase.\n" +
"     The name attribute value is evaluated via Python.\n" +
"-->\n" +
"<!ELEMENT testcase   (%task;)>\n" +
"<!ATTLIST testcase\n" +
"          name       CDATA    #REQUIRED\n" +
"          mode       CDATA    \"'default'\"\n" +
">\n";

    public String getDTDInfo()
    {
        return fDTDInfo;
    }

    public String getDTDTaskName()
    {
        return "testcase";
    }

    public STAXTestcaseActionFactory(STAX staxService)
    {
        staxService.registerListHandler("TESTCASES", this);
        staxService.registerQueryHandler("TESTCASE", "Test Name", this);

        staxService.registerJobManagementHandler(this);

        // Set up Testcase Status Update parser

        fUpdateParser.addOption("UPDATE", 1,
                                STAFCommandParser.VALUENOTALLOWED);
        fUpdateParser.addOption("JOB", 1,
                                STAFCommandParser.VALUEREQUIRED);
        fUpdateParser.addOption("TESTCASE", 1,
                                STAFCommandParser.VALUEREQUIRED);
        fUpdateParser.addOption("STATUS", 1,
                                STAFCommandParser.VALUEREQUIRED);
        fUpdateParser.addOption("MESSAGE", 1, 
                                STAFCommandParser.VALUEREQUIRED);
        fUpdateParser.addOption("FORCE", 1,
                                STAFCommandParser.VALUENOTALLOWED);

        fUpdateParser.addOptionNeed("UPDATE", "JOB");
        fUpdateParser.addOptionNeed("JOB", "TESTCASE");
        fUpdateParser.addOptionNeed("TESTCASE", "STATUS");
        fUpdateParser.addOptionNeed("STATUS", "TESTCASE");
        fUpdateParser.addOptionNeed("MESSAGE", "STATUS");
        
        // Set up Testcase Start parser

        fStartParser.addOption("START", 1,
                               STAFCommandParser.VALUENOTALLOWED);
        fStartParser.addOption("JOB", 1,
                               STAFCommandParser.VALUEREQUIRED);
        fStartParser.addOption("TESTCASE", 1,
                               STAFCommandParser.VALUEREQUIRED);
        fStartParser.addOption("KEY", 1,
                               STAFCommandParser.VALUEREQUIRED);

        fStartParser.addOptionNeed("START", "JOB");
        fStartParser.addOptionNeed("JOB", "TESTCASE");
        fStartParser.addOptionNeed("KEY", "TESTCASE");

        // Set up Testcase Stop parser

        fStopParser.addOption("STOP", 1,
                              STAFCommandParser.VALUENOTALLOWED);
        fStopParser.addOption("JOB", 1,
                              STAFCommandParser.VALUEREQUIRED);
        fStopParser.addOption("TESTCASE", 1,
                              STAFCommandParser.VALUEREQUIRED);
        fStopParser.addOption("KEY", 1,
                              STAFCommandParser.VALUEREQUIRED);

        fStopParser.addOptionNeed("STOP", "JOB");
        fStopParser.addOptionNeed("JOB", "TESTCASE");
        fStopParser.addOptionNeed("KEY", "TESTCASE");

        // Register as a GenericRequestHandler
        try
        {
            // Assign STAFServiceInterfaceLevel class that this handler uses

            Class serviceInterfaceClass = Class.forName(STAX.INTERFACE_LEVEL_4);

            int rc = staxService.registerGenericRequestHandler(this,
                                                      serviceInterfaceClass);
            if (rc != 0)
            {
                System.out.println("STAXTestcaseActionFactory." +
                                   "registerGenericRequestHandler() failed");
            }
        }
        catch (ClassNotFoundException e)
        {
            System.out.println("STAXTestcaseActionFactory: " +
                               "registerGenericRequestHandler: " + e);
        }
    }

    public void initJob(STAXJob job)
    {
        if (!job.setData("testcaseMap", new TreeMap()))
        {
            String msg = "STAXTestcaseActionFactory.initJob: setData for " +
                         "testcaseMap failed.";
            job.log(STAXJob.JOB_LOG, "error", msg);
        } 

        if (!job.setData("activeTestcaseMap", new HashMap()))
        {
            String msg = "STAXTestcaseActionFactory.initJob: setData for " +
                         "activeTestcaseMap failed.";
            job.log(STAXJob.JOB_LOG, "error", msg);
        } 
    }

    public void terminateJob(STAXJob job)
    {
        // Check if any entries are left in the activeTestcaseMap and call
        // the stop testcase method to calculate the elapsed time.

        HashMap activeTestcaseMap = (HashMap)job.getData("activeTestcaseMap");
        TreeMap testcaseMap = (TreeMap)job.getData("testcaseMap");
        
        synchronized (activeTestcaseMap)
        {
            Iterator it = activeTestcaseMap.values().iterator();

            while (it.hasNext())
            {
                STAXActiveTestcase activeTest = (STAXActiveTestcase)it.next();

                String name = activeTest.getName();
                STAXTimestamp startTime = activeTest.getStartTime();
                
                // Stop the Testcase

                STAXTestcase theTest = (STAXTestcase)testcaseMap.get(name);
                
                if (theTest != null)
                {
                    theTest.stop(job, startTime);
                }
                else
                {
                    job.log(STAXJob.JOB_LOG, "warning",
                            "STAXTestcaseActionFactory::terminateJob() - " +
                            "Could not stop testcase: " + name +
                            ".  Not in active testcase map.");
                }
            }
        }

        // Log a summary of the testcase status for the job.

        synchronized (testcaseMap)
        {
            int totalTests = 0;
            int totalPasses = 0;
            int totalFails = 0;
            Iterator it = testcaseMap.keySet().iterator();

            while (it.hasNext())
            {
                String testName = (String)it.next();
                STAXTestcase test = (STAXTestcase)testcaseMap.get(testName);

                int passes = test.getNumPass();
                int fails  = test.getNumFail();

                if (passes == 0 && fails == 0 &&
                    test.getMode() == STAXTestcase.DEFAULT_MODE)
                {
                    // Do not log a status record for the testcase
                }
                else
                {
                    // Log a status record for the testcase

                    String msg = "Testcase: " + testName + 
                                 ", Pass: " + passes +
                                 ", Fail: " + fails; 

                    if (job.getLogTCElapsedTime())
                    {
                        msg += ", ElapsedTime: " + test.getElapsedTime();
                    }

                    if (job.getLogTCNumStarts())
                    {
                        msg += ", NumStarts: " + test.getNumStarts();
                    }

                    job.log(STAXJob.JOB_LOG, "status", msg);

                    totalTests += 1;
                    totalPasses += passes;
                    totalFails += fails;
                }
            }

            // Log testcase totals
            String msg = "Testcase Totals: Tests: " + totalTests +
                         ", Pass: " + totalPasses + ", Fail: " + totalFails;
            job.log(STAXJob.JOB_LOG, "status", msg);
        }
    }

    public STAFResult handleListRequest(String type, STAXJob job,
                                        STAXRequestSettings settings)
    {
        String lineSep = settings.getLineSeparator();
   
        if (type.equalsIgnoreCase("testcases"))
        {
            // LIST JOB <Job ID> TESTCASES

            String result = new String();

            TreeMap testcaseMap = (TreeMap)job.getData("testcaseMap");

            synchronized (testcaseMap)
            {
                Iterator iter = testcaseMap.values().iterator();

                while (iter.hasNext())
                {
                    STAXTestcase testcase = (STAXTestcase)iter.next();
                    int numPass = testcase.getNumPass();
                    int numFail = testcase.getNumFail();

                    if (testcase.getMode() == STAXTestcase.DEFAULT_MODE &&
                        numPass == 0 && numFail == 0)
                    {
                        // Don't show in list
                    }
                    else
                    {
                        result += testcase.getName() + ";";
                        result += numPass + ";";
                        result += numFail;

                        if (job.getLogTCElapsedTime())
                        {
                            result += ";" + testcase.getElapsedTime();
                        }

                        if (job.getLogTCNumStarts())
                        {
                            result += ";" + testcase.getNumStarts();
                        }

                        result += lineSep;
                    }
                }
            }
            return new STAFResult(STAFResult.Ok, result);
        }
        else
            return new STAFResult(STAFResult.DoesNotExist, type);
    }

    public STAFResult handleQueryRequest(String type, String key, STAXJob job,
                                        STAXRequestSettings settings)
    {
        String lineSep = settings.getLineSeparator();
    
        if (type.equalsIgnoreCase("testcase"))
        {
            // QUERY JOB <Job ID> TESTCASE <Testcase Name>

            TreeMap testcaseMap = (TreeMap)job.getData("testcaseMap");

            synchronized (testcaseMap)
            {
                if (!testcaseMap.containsKey(key))
                    return new STAFResult(STAFResult.DoesNotExist, key);
                else
                {
                    STAXTestcase testcase = (STAXTestcase)testcaseMap.get(key);

                    String result = "Testcase Name   : " + testcase.getName() +
                                    lineSep;
                    result += "Passes          : " + 
                              testcase.getNumPass() + lineSep;
                    result += "Fails           : " +
                              testcase.getNumFail() + lineSep;
                    result += "Last Status     : " +  
                              testcase.getLastStatus() + lineSep;

                    STAXTimestamp lastStatusTimestamp =
                        testcase.getLastStatusTimestamp();
                    
                    if (lastStatusTimestamp != null)
                    {
                        result += "Last Status Date: " + 
                                  testcase.getLastStatusTimestamp().
                                  getDateString() + lineSep;
                        result += "Last Status Time: " + 
                                  testcase.getLastStatusTimestamp().
                                  getTimeString() + lineSep;
                    }

                    result += "Elapsed Time    : " +
                              testcase.getElapsedTime() + lineSep;
                    result += "Number of Starts: " +
                              testcase.getNumStarts() + lineSep;
                    
                    return new STAFResult(STAFResult.Ok, result);
                }
            }
        }
        else
            return new STAFResult(STAFResult.DoesNotExist, type);
    }

    public STAFResult handleQueryJobRequest(STAXJob job,
                                            STAXRequestSettings settings)
    {
        return new STAFResult(STAFResult.Ok, "");
    }

    public STAXAction parseAction(STAX staxService, STAXJob job,
                                  org.w3c.dom.Node root) throws STAXException
    {
        String testName = new String();
        STAXAction testcaseAction = null;
        String testMode = new String();

        NamedNodeMap attrs = root.getAttributes();

        for (int i = 0; i < attrs.getLength(); ++i)
        {
            Node thisAttr = attrs.item(i);

            String errorInfo = "\n  Element: " + root.getNodeName() +
               "  Attribute: " + thisAttr.getNodeName();

            if (thisAttr.getNodeName().equals("name"))
            {
                testName = STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo);
            }
            else if (thisAttr.getNodeName().equals("mode"))
            {
                testMode = STAXUtil.parseAndCompileForPython(
                    thisAttr.getNodeValue(), errorInfo);
            }
            else
            {
                throw new STAXInvalidXMLAttributeException(
                          root.getNodeName() + ": " + thisAttr.getNodeName());
            }
        }

        NodeList children = root.getChildNodes();

        for (int i = 0; i < children.getLength(); ++i)
        {
            Node thisChild = children.item(i);

            if (thisChild.getNodeType() == Node.COMMENT_NODE)
            {
                /* Do nothing */
            }
            else if (thisChild.getNodeType() == Node.ELEMENT_NODE)
            {
                if (testcaseAction != null)
                {
                    throw new STAXInvalidXMLElementCountException(
                              thisChild.getNodeName());
                }

                STAXActionFactory factory = 
                    staxService.getActionFactory(thisChild.getNodeName());

                if (factory == null)
                {
                    throw new STAXInvalidXMLElementException(
                              thisChild.getNodeName());
                }

                testcaseAction = factory.parseAction(staxService, job,
                                                     thisChild);
            }
            else
            {
                throw new STAXInvalidXMLNodeTypeException(
                          Integer.toString(thisChild.getNodeType()));
            }
        }

        return new STAXTestcaseAction(testName, testcaseAction, testMode);
    }

    // STAXGenericRequestHandler Interface Methods

    public STAFResult handleRequest(Object infoObject, STAX staxService)
    {
        STAFServiceInterfaceLevel4.RequestInfo info =
            (STAFServiceInterfaceLevel4.RequestInfo)infoObject;
        
        String lowerRequest = info.request.toLowerCase();

        if (lowerRequest.startsWith("update"))
            return handleUpdateRequest(info, staxService);
        else if (lowerRequest.startsWith("start"))
            return handleStartRequest(info, staxService);
        else if (lowerRequest.startsWith("stop"))
            return handleStopRequest(info, staxService);
        else
        {   // Returning nothing in the result indicates that this parser
            // does not support this request.
            return new STAFResult(STAFResult.InvalidRequestString, "");
        }
    }

    private STAFResult handleUpdateRequest(
        STAFServiceInterfaceLevel4.RequestInfo info, STAX staxService)
    {
        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 3)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 3 required for UPDATE request." +
                " Requesting machine's trust level: " +  info.trustLevel); 
        }

        STAFCommandParseResult parseResult= fUpdateParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Resolve the value specified for JOB ID and get its integer value

        STAFResult resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("JOB"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        Integer jobID;

        try
        {
            jobID = new Integer(resolvedResult.result);    
        }
        catch (NumberFormatException e)
        {
            return new STAFResult(STAFResult.InvalidValue,
                                  parseResult.optionValue("JOB"));
        }
        
        // Verify that the JOB ID is a currently active job

        STAXJob job = null;

        job = (STAXJob)staxService.getJobMap().get(jobID);

        if (job == null)
        {
            return new STAFResult(STAFResult.DoesNotExist,
                                  parseResult.optionValue("JOB"));
        }

        // Resolve the value specified for Testcase name

        resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("TESTCASE"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        String testcaseName = resolvedResult.result;    

        // Resolve the value specified for the Status

        resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("STATUS"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        String status = resolvedResult.result;

        // Make sure that the status is either pass or fail

        if (!status.equalsIgnoreCase("pass") &&
            !status.equalsIgnoreCase("fail"))
        {   
            return new STAFResult(STAFResult.InvalidValue, status);
        }

        String message = new String("");

        if (parseResult.optionTimes("MESSAGE") != 0)
        {
            // Resolve the value specified for Message

            resolvedResult = resolveValue(
                info.machine, info.localMachine,
                parseResult.optionValue("MESSAGE"), info.handle,
                staxService.getSTAFHandle());

            if (resolvedResult.rc != 0) return resolvedResult;

            message = resolvedResult.result;    
        }

        boolean force = false;

        if (parseResult.optionTimes("FORCE") != 0)
        {
            force = true;
        }
        
        // Check if the testcase exists.  If not, and force is not specified,
        // return an error message. If not, and force is specified, add it
        // to the testcaseMap.

        TreeMap testcaseMap = (TreeMap)job.getData("testcaseMap");

        STAXTestcase theTest;

        synchronized (testcaseMap)
        {
            theTest = (STAXTestcase)testcaseMap.get(testcaseName);

            if (theTest == null)
            {
                if (! force)
                {
                    return new STAFResult(STAFResult.DoesNotExist,
                                          testcaseName);
                }
                else
                {
                    theTest = new STAXTestcase(testcaseName, STAXTestcase.STRICT_MODE);

                    synchronized(testcaseMap)
                    {
                        testcaseMap.put(testcaseName, theTest);
                    }
                }
            }
        }
        
        // Check if the testcase name is in the activeTestcaseMap.
        // If not, add it to the activeTestcaseMap, storing its start
        // timestamp, and start the testcase.

        HashMap activeTestcaseMap = (HashMap)job.getData("activeTestcaseMap");

        synchronized (activeTestcaseMap)
        {
            if (!activeTestcaseMap.containsKey(testcaseName))
            {
                STAXActiveTestcase activeTest =
                    new STAXActiveTestcase(testcaseName, new STAXTimestamp());

                activeTestcaseMap.put(testcaseName, activeTest);

                // Start the testcase
                theTest.start(job);
            }
        }

        // Update Testcase status

        theTest.updateStatus(status, message, job);
        
        return new STAFResult(STAFResult.Ok);
    }

    private STAFResult handleStartRequest(
        STAFServiceInterfaceLevel4.RequestInfo info, STAX staxService)
    {
        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 3)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 3 required for START request." +
                " Requesting machine's trust level: " +  info.trustLevel); 
        }
        
        STAFCommandParseResult parseResult= fStartParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }
        
        // Resolve the value specified for JOB ID and get its integer value

        STAFResult resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("JOB"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        Integer jobID;

        try
        {
            jobID = new Integer(resolvedResult.result);    
        }
        catch (NumberFormatException e)
        {
            return new STAFResult(STAFResult.InvalidValue,
                                  parseResult.optionValue("JOB"));
        }
        
        // Verify that the JOB ID is a currently active job

        STAXJob job = null;

        job = (STAXJob)staxService.getJobMap().get(jobID);

        if (job == null)
        {
            return new STAFResult(STAFResult.DoesNotExist,
                                  parseResult.optionValue("JOB"));
        }

        // Resolve the value specified for Testcase name

        resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("TESTCASE"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        String testcaseName = resolvedResult.result;    
        
        String key = new String("");

        if (parseResult.optionTimes("KEY") > 0)
        {
            // Resolve the value specified for the key if needed

            resolvedResult = resolveValue(
                info.machine, info.localMachine,
                parseResult.optionValue("KEY"), info.handle,
                staxService.getSTAFHandle());

            if (resolvedResult.rc != 0) return resolvedResult;

            key = resolvedResult.result;    
        }

        // Check if the testcase[:key] is in the activeTestcaseMap.  If not,
        // add it to the activeTestcaseMap, storing its start timestamp.

        String testcaseKey = testcaseName;

        if (!key.equals(""))
        {
            testcaseKey = testcaseName + ":" + key;
        }

        HashMap activeTestcaseMap = (HashMap)job.getData("activeTestcaseMap");

        synchronized (activeTestcaseMap)
        {
            if (activeTestcaseMap.containsKey(testcaseKey))
            {
                return new STAFResult(STAFResult.AlreadyExists, testcaseKey);
            }
            
            STAXActiveTestcase activeTest =
                new STAXActiveTestcase(testcaseName, new STAXTimestamp());

            activeTestcaseMap.put(testcaseKey, activeTest);
        }
        
        // Check if the testcase name is already in the testcaseMap.
        // If not, add it to the testcaseMap.  If the testcase already
        // exists, set its mode to strict if its not already strict.
        
        TreeMap testcaseMap = (TreeMap)job.getData("testcaseMap");

        STAXTestcase theTest;

        synchronized (testcaseMap)
        {
            theTest = (STAXTestcase)testcaseMap.get(testcaseName);

            if (theTest == null)
            {
                theTest = new STAXTestcase(testcaseName,
                                           STAXTestcase.STRICT_MODE);
                testcaseMap.put(testcaseName, theTest);
            }
            else if (theTest.getMode() != STAXTestcase.STRICT_MODE)
            {
                theTest.setMode(STAXTestcase.STRICT_MODE);
            }
        }

        // Start the testcase 
        theTest.start(job);

        return new STAFResult(STAFResult.Ok);
    }

    private STAFResult handleStopRequest(
        STAFServiceInterfaceLevel4.RequestInfo info, STAX staxService)
    {
        // Check whether Trust level is sufficient for this command.

        if (info.trustLevel < 3)
        {   
            return new STAFResult(STAFResult.AccessDenied, 
                "Trust level 3 required for START request." +
                " Requesting machine's trust level: " +  info.trustLevel); 
        }
        
        STAFCommandParseResult parseResult= fStopParser.parse(info.request);

        if (parseResult.rc != STAFResult.Ok)
        {
            return new STAFResult(STAFResult.InvalidRequestString,
                                  parseResult.errorBuffer);
        }

        // Resolve the value specified for JOB ID and get its integer value

        STAFResult resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("JOB"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        Integer jobID;

        try
        {
            jobID = new Integer(resolvedResult.result);    
        }
        catch (NumberFormatException e)
        {
            return new STAFResult(STAFResult.InvalidValue,
                                  parseResult.optionValue("JOB"));
        }
        
        // Verify that the JOB ID is a currently active job

        STAXJob job = null;

        job = (STAXJob)staxService.getJobMap().get(jobID);

        if (job == null)
        {
            return new STAFResult(STAFResult.DoesNotExist,
                                  parseResult.optionValue("JOB"));
        }

        // Resolve the value specified for Testcase name

        resolvedResult = resolveValue(
            info.machine, info.localMachine,
            parseResult.optionValue("TESTCASE"), info.handle,
            staxService.getSTAFHandle());

        if (resolvedResult.rc != 0) return resolvedResult;

        String testcaseName = resolvedResult.result;    
        
        String key = new String("");

        if (parseResult.optionTimes("KEY") > 0)
        {
            // Resolve the value specified for the key if needed

            resolvedResult = resolveValue(
                info.machine, info.localMachine,
                parseResult.optionValue("KEY"), info.handle,
                staxService.getSTAFHandle());

            if (resolvedResult.rc != 0) return resolvedResult;

            key = resolvedResult.result;    
        }
        
        // Check if the testcase:[key] is in the activeTestcaseMap.  If so,
        // get its start timestamp and remove it from the activeTestcaseMap.
        // If not, return a DoesNotExist error.

        String testcaseKey = testcaseName;

        if (!key.equals(""))
        {
            testcaseKey = testcaseName + ":" + key;
        }

        STAXTimestamp startTimestamp;

        HashMap activeTestcaseMap = (HashMap)job.getData("activeTestcaseMap");

        synchronized (activeTestcaseMap)
        {
            if (activeTestcaseMap.containsKey(testcaseKey))
            {

                STAXActiveTestcase activeTest = 
                    (STAXActiveTestcase)activeTestcaseMap.get(testcaseKey);

                startTimestamp = activeTest.getStartTime();

                activeTestcaseMap.remove(testcaseKey);
            }
            else
            {
                return new STAFResult(STAFResult.DoesNotExist, testcaseKey);
            }
        }

        // Stop the Testcase

        TreeMap testcaseMap = (TreeMap)job.getData("testcaseMap");

        STAXTestcase theTest;

        synchronized (testcaseMap)
        {
            theTest = (STAXTestcase)testcaseMap.get(testcaseName);
        }
            
        if (theTest == null)
        {
            return new STAFResult(STAFResult.DoesNotExist, testcaseName);
        }

        theTest.stop(job, startTimestamp);
        
        return new STAFResult(STAFResult.Ok);
    }

    public String getHelpInfo(String lineSep)
    {
        return "UPDATE    JOB <Job ID> TESTCASE <Testcase name> " +
               "STATUS <Status>" + lineSep +
               "          [MESSAGE <Message text>] [FORCE]" + lineSep +
               lineSep +
               "START     JOB <Job ID> TESTCASE <Testcase name> [KEY <Key>]" +
               lineSep + lineSep +
               "STOP      JOB <Job ID> TESTCASE <Testcase name> [KEY <Key>]";
    }
    
    /***********************************************************************/
    /* resolveValue - Uses the var Service to resolve an option value.     */
    /*                                                                     */
    /* accepts: the submitting machine name, the local machine name,       */
    /*         an unresolved option value, the submitting process' handle, */
    /*         and the STAX service's handle                               */
    /*                                                                     */   
    /* returns: STAFResult.rc = STAFResult.Ok;                             */
    /* STAFResult.result = the resolved option value                       */
    /***********************************************************************/

    private STAFResult resolveValue(String machine, String localMachine, 
                                    String optionValue, int handle,
                                    STAFHandle serviceHandle)
    {
        STAFResult resolvedResult = new STAFResult(STAFResult.Ok, optionValue);  

        if (optionValue.indexOf("{") >= 0)
        { 
            if (machine.equals(localMachine))
            {  // Request was made locally
                resolvedResult = serviceHandle.submit2(
                    "local", "var", "handle " + handle +
                    " resolve " + STAFUtil.wrapData(optionValue));
            }
            else
            {  // Request was made from another system
                resolvedResult = serviceHandle.submit2(
                    "local", "var", "global" + " resolve " +
                    STAFUtil.wrapData(optionValue));
            }
        }
        
        return resolvedResult;
    }


    private STAFCommandParser fUpdateParser = new STAFCommandParser();
    private STAFCommandParser fStartParser  = new STAFCommandParser();
    private STAFCommandParser fStopParser   = new STAFCommandParser();


    // STAXActiveTestcase class is the value stored in the activeTestcaseMap

    class STAXActiveTestcase
    {
        STAXActiveTestcase(String name, STAXTimestamp startTime)
        {
            fName = name;
            fStartTime = startTime;
        }

        String getName() { return fName; }
        STAXTimestamp getStartTime() { return fStartTime; }
        
        private String fName;
        private STAXTimestamp fStartTime;

    } // end class STAXActiveTestcase
}
