package com.ibm.staf.service.sxe;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Date;
import java.util.Calendar;
import java.util.StringTokenizer;

/**
 * This class provides the full function of the STAFSXE 
 * Service. 
 */

public class STAFSXE implements STAFServiceInterfaceLevel4 
{
    private STAFHandle sHandle;
    private String sServiceName;
    private STAFCommandParser sxeParser;
    private String logName = "SXELOG";

    /* static strings for parser */

    private final static String EXECUTE = "EXECUTE";
    private final static String VERSION = "VERSION";
    private final static String HELP = "HELP";
    private final static String FILE = "FILE";
    private final static String LOOP = "LOOP";
    private final static String FOREVER = "FOREVER";
    private final static String MINRUNTIME = "MINRUNTIME";
    private final static String LOGNAME = "LOGNAME";

    /* other static values */
    
    private final static String SXEVERSION = "2.1.3";
    
    private final static int START = 0;
    private final static int STOPPASS = 1;
    private final static int STOPFAIL = 2;
    
    private final static String LOGLEVELVAR = "{STAF/Service/SXE/LogLevel}";
    private final static String ELAPSEDTARGETVAR = 
        "{STAF/Service/SXE/ElapsedTarget}";
    private final static String ELAPSEDTOLVAR = 
        "{STAF/Service/SXE/ElapsedTolerance}";
    private final static String LOGNONE = "NONE";
    private final static String LOGFILEONLY = "FILE";

    /* SXE STAF return codes */

    public final static int FILEOPENERROR = 4001;
    public final static String FILEOPENERRORInfo = "File Open Error";
    public final static String FILEOPENERRORDesc = 
        "The specified file could not be opened by the file system.";
    public final static int IOERROR = 4002;
    public final static String IOERRORInfo = "I/O Error";
    public final static String IOERRORDesc = 
        "An I/O Error occurred accessing the specified file.";
    public final static int REGERROR = 4003;
    public final static String REGERRORInfo = "STAF Handle Registration Error";
    public final static String REGERRORDesc = 
        "An error occurred registering a STAF handle.";
    public final static int ELAPSEDTIMEFAIL = 4004;
    public final static String ELAPSEDTIMEFAILInfo = 
        "Elapsed Target Time Exceeded";
    public final static String ELAPSEDTIMEFAILDesc = 
        "The testcase elapsed time exceeded the elapsed target " + 
        "time by more than the\nelapsed tolerance value.";
    public final static int ELAPSEDTIMEERROR = 4005;
    public final static String ELAPSEDTIMEERRORInfo = "Elapsed Target Error";
    public final static String ELAPSEDTIMEERRORDesc = 
        "An error occurred resolving the elapsed target variables.";
    
/**
 * STAFSXE constructor
 */

public STAFSXE() 
{
    super();
}

/**
 * Method required by implemented STAF interface. This is the
 * entry point for a STAF request.
 */

public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo reqInfo)
{
    String request = reqInfo.request.toUpperCase();
    STAFResult result;
    
    if (request.startsWith(EXECUTE)) 
    {
        result = prepare2Execute(reqInfo);
    }
    else if (request.startsWith(VERSION)) 
    {
        result = version();
    }
    else if (request.startsWith(HELP)) 
    {
        result = help();
    }
    else 
    {
        result = new STAFResult(STAFResult.InvalidRequestString);
    }

    return result;    
}

/**
 * Checks if the elapsed time is within the specified parameters.
 * 
 * Returns false if the elapsed time exceeded the target+tolerance,
 * or true otherwise.
 */

private boolean checkElapsedTime(long startTime, long stopTime, 
                                 long elapsedTarget, int elapsedTolerance)
{
    /* Calculate the maxTime in millis:
       elapsedTarget + elapsedTarget * elapsedTolerance% */
       
    long maxTime = (long) (elapsedTarget + 
        elapsedTarget*(((double) elapsedTolerance)/100));

    if (stopTime - startTime > maxTime) 
    {
        return false;
    }
    else 
    {
        return true;
    }    
}

/**
 * Executes all commands in the specified file. One iteration
 * through this method executes the file once. 
 * If a command returns a non-zero return code, execution
 * will stop at that point.
 * Returns the STAFResult from executing the file. This will
 * be the STAFResult from a failed command if one failed.
 */

private STAFResult execute(STAFHandle fHandle, String filename, 
                           String logName, int loopNum) 
{
    long startTime;
    long stopTime = 0;

    /* Use 1 length arrays and a StringBuffer so that other called methods
        may directly modify these primitives/objects */
        
    long[] elapsedTarget = new long[1]; 
    int[] elapsedTolerance = new int[1];
    StringBuffer elapsedTargetSBuf = new StringBuffer();

    /* open file */

    BufferedReader in = null;
    STAFResult result;
    
    try 
    {   //big try block so we can close file in finally
        
        try 
        {
            in = new BufferedReader(new FileReader(new File(filename)));
        }
        catch(FileNotFoundException fnfe) 
        {
            return new 
                STAFResult(FILEOPENERROR, String.valueOf(loopNum)+filename);
        }

        /* iterate through file and execute STAF commands */

        int commandCount = 0;
        int lineCount = 0;

        /* Get startTime and log test start */

        startTime = new Date().getTime(); //get StartTime
        logFile(logName, START, fHandle, loopNum, startTime, stopTime, 
            elapsedTargetSBuf.toString(), elapsedTolerance[0]);
    
        while(true) 
        {
            String nextLine;

            /* Read a line from file */
            
            try 
            {
                nextLine = in.readLine();
            }
            catch(IOException ioe) 
            {
                stopTime = new Date().getTime(); //get stopTime, IOError
                
                result = new STAFResult(IOERROR, 
                    String.valueOf(loopNum)+"#"+String.valueOf(commandCount+1)+
                    "#"+String.valueOf(lineCount+1)+":"+filename);
                    
                logFile(logName, STOPFAIL, fHandle, loopNum, startTime, 
                    stopTime, elapsedTargetSBuf.toString(), 
                    elapsedTolerance[0]);
                break;
            }
            
            if (nextLine == null) 
            {   //at end of file, no failures
            
                stopTime = new Date().getTime(); //get stopTime, success
                boolean checkElapsed;
                
                try 
                {
                    checkElapsed = isCheckElapsed(fHandle, elapsedTargetSBuf, 
                        elapsedTarget, elapsedTolerance);
                }
                catch(Exception e) 
                {
                    result = new STAFResult(ELAPSEDTIMEERROR, 
                        "Error in ElapsedTime/ElapsedTolerance vars. Info: " + 
                        e.toString());
                        
                    logCommand(logName, fHandle, STOPFAIL, null, result);
                    
                    logFile(logName, STOPFAIL, fHandle, loopNum, startTime, 
                        stopTime, elapsedTargetSBuf.toString(), 
                        elapsedTolerance[0]);
                    break;
                }   
                
                if (checkElapsed && 
                    !checkElapsedTime(startTime, stopTime, elapsedTarget[0], 
                    elapsedTolerance[0])) 
                {
                    //elapsed time check failed
                    
                    result = new STAFResult(ELAPSEDTIMEFAIL, 
                        getElapsedTime(startTime, stopTime));
                    
                    logFile(logName, STOPFAIL, fHandle, loopNum, startTime, 
                        stopTime, elapsedTargetSBuf.toString(), 
                        elapsedTolerance[0]);
                }
                else 
                {
                    result = new STAFResult(STAFResult.Ok, 
                        String.valueOf(loopNum) + "#" + 
                        String.valueOf(commandCount));
                        
                    logFile(logName, STOPPASS, fHandle, loopNum, startTime, 
                        stopTime, elapsedTargetSBuf.toString(), 
                        elapsedTolerance[0]);
                }
                break;
            }
            
            /* If line is blank go to next line */
            nextLine = nextLine.trim(); //trim whitespace
            
            if (nextLine.equals(new String())) 
            {
                lineCount++;
                continue;
            }
            
            /* Goto next line if 1st non-whitespace character is # */
            if (nextLine.startsWith("#")) 
            {
                lineCount++;
                continue;
            }

            logCommand(logName, fHandle, START, nextLine, null);
            
            result = sendSTAFCommand(fHandle, nextLine);
            
            if (result.rc != STAFResult.Ok) 
            {
                stopTime = new Date().getTime(); //get stopTime, fail
                
                result = new STAFResult(result.rc, 
                    String.valueOf(loopNum)+"#"+String.valueOf(commandCount+1)+
                    "#"+String.valueOf(lineCount+1)+":["+nextLine+
                    "]:"+String.valueOf(result.rc)+":"+result.result);
                    
                logCommand(logName, fHandle, STOPFAIL, nextLine, result);
                
                logFile(logName, STOPFAIL, fHandle, loopNum, startTime, 
                    stopTime, elapsedTargetSBuf.toString(), 
                    elapsedTolerance[0]);
                break;
            }
            
            logCommand(logName, fHandle, STOPPASS, nextLine, result);
            commandCount++;
            lineCount++;
        }
    }
    finally 
    {   //make sure we always try to close file
    
        try 
        {
            in.close();
        }
        catch(Exception e) 
        {
            // do nothing??
        }
    }
    
    return result;
}

/**
 * Returns a String representation of the difference in
 * two times. The inputs are two times expressed in milliseconds
 * and stored in a long (see java.util.Date).
 * The returned format is: hh:mm:ss.mmm
 */

private String getElapsedTime(long start, long stop) 
{
    long time = stop - start;
    long hour = time / 3600000;
    time = time%3600000;
    long min = time / 60000;
    time = time%60000;
    long sec = time / 1000;
    long milli = time%1000;
    String milliString;

    /* Make sure that millis are represented properly */
    
    if(milli < 10) 
    {
        milliString = "00"+String.valueOf(milli);
    }
    else if (milli < 100) 
    {
        milliString = "0"+String.valueOf(milli);
    }
    else 
    {
        milliString = String.valueOf(milli);
    }

    String formattedTime = 
        String.valueOf(hour)+":"+String.valueOf(min)+":"+
        String.valueOf(sec)+"."+milliString;

    return formattedTime;
}

/**
 * Resolves any STAF variables in a command parameter.
 */

private String getResolvedParm(STAFServiceInterfaceLevel4.RequestInfo reqInfo,
                               STAFCommandParseResult parseResult, String parm)
                               throws STAFException 
{
    String value = parseResult.optionValue(parm);

    if (value.indexOf("{") == -1) 
    {
        return value;
    }

    String request;
    
    if (reqInfo.localMachine.equals(reqInfo.machine)) 
    {
        /* command came from local machine use handle resolve */
        request = "HANDLE "+String.valueOf(reqInfo.handle)+ 
            " RESOLVE \""+value+"\"";
    }
    else 
    {
        /* command came from remote machine use global pool */
        request = "GLOBAL RESOLVE \""+value+"\"";
    }

    String result = sHandle.submit("LOCAL", "VAR", request);
        
    return result;    
}

/**
 * Returns a formatted String representation of a time of day.
 * The time should be expressed in milliseconds and stored in 
 * a long (see java.util.Date).
 * The returned format is: hh:mm:ss.mmm
 */

private String getTime(long time) 
{
    Calendar cal = Calendar.getInstance();
    cal.setTime(new Date(time));

    int hour = cal.get(Calendar.HOUR_OF_DAY);
    int min = cal.get(Calendar.MINUTE);
    int sec = cal.get(Calendar.SECOND);
    int milli = cal.get(Calendar.MILLISECOND);
    String milliString;
    
    /* Make sure that millis are represented properly */
    
    if(milli < 10) 
    {
        milliString = "00"+String.valueOf(milli);
    }
    else if (milli < 100) 
    {
        milliString = "0"+String.valueOf(milli);
    }
    else 
    {
        milliString = String.valueOf(milli);
    }
    
    String formattedTime = String.valueOf(hour)+":"+String.valueOf(min)+":"+
        String.valueOf(sec)+"."+milliString;
    
    return formattedTime;    
}

/**
 * Given a String representation of time of day in the format
 * hh:mm:ss.mmm returns a long representing the number of milliseconds
 * since the beginning of that day (12 am).
 */

private long getTimeInMillis(String formattedTime) throws NumberFormatException
{
    StringTokenizer tok = new StringTokenizer(formattedTime, ":.");

    int hours = Integer.parseInt(tok.nextToken());
    int mins = Integer.parseInt(tok.nextToken());
    int secs = Integer.parseInt(tok.nextToken());
    String mills = tok.nextToken();
    
    // Make sure millis is correct (3 chars) 
    // otherwise 0:0:0.5 is 5 millis instead of 500
    
    while(true) 
    {
        if (mills.length() == 3) 
        {
            break;
        }
        
        mills = mills.concat("0");
    }
    
    int millis = Integer.parseInt(mills);

    long time = (hours*3600000)+(mins*60000)+(secs*1000)+millis;
    
    return time;
}

/**
 * Displays help output for the SXE service.
 */

private STAFResult help() 
{
    String LINESEP;
    
    try
    {
        LINESEP = sHandle.submit("LOCAL", "VAR", 
            "GLOBAL RESOLVE {STAF/Config/Sep/Line}");
    }
    catch (STAFException e) 
    {
        LINESEP = "\n";
    }

    String help =
        "SXE STAF Service"+LINESEP+LINESEP+
        "EXECUTE FILE <file> [LOOP <number of loops>] [MINRUNTIME <hh:mm>]" + 
        LINESEP+"VERSION"+LINESEP+"HELP";

    STAFResult result = new STAFResult(STAFResult.Ok, help);
    return result;    
}

/**
 * Method required by implemented STAF interface.
 * This method is called when STAF first initializes
 * the service.
 */

public int init(STAFServiceInterfaceLevel4.InitInfo initInfo) 
{    
    int result = STAFResult.Ok;

    /* Generate STAFParser */
    
    sxeParser = new STAFCommandParser();
    
    sxeParser.addOption(EXECUTE, 1, STAFCommandParser.VALUENOTALLOWED);
    sxeParser.addOption(VERSION, 1, STAFCommandParser.VALUENOTALLOWED);
    sxeParser.addOption(HELP, 1, STAFCommandParser.VALUENOTALLOWED);

    sxeParser.addOption(FILE, 1, STAFCommandParser.VALUEREQUIRED);
    sxeParser.addOption(LOOP, 1, STAFCommandParser.VALUEREQUIRED);
    sxeParser.addOption(MINRUNTIME, 1, STAFCommandParser.VALUEREQUIRED);

    sxeParser.addOptionGroup(FILE, 1, 1);
    sxeParser.addOptionGroup(LOOP+" "+MINRUNTIME, 0, 1);
    sxeParser.addOptionGroup(EXECUTE+" "+VERSION+" "+HELP, 1, 1);
    
    /* Create STAFHandle */

    try 
    {
        sHandle = new STAFHandle("STAF/Service/" + initInfo.name);
    }
    catch(STAFException se) 
    {
        se.printStackTrace();
        return STAFResult.STAFRegistrationError;
    }
    
    sServiceName = initInfo.name;

    /* Parse registration parms */
        
    STAFCommandParser parmsParser = new STAFCommandParser();
    parmsParser.addOption(LOGNAME, 1, STAFCommandParser.VALUEREQUIRED);
    
    STAFCommandParseResult pResult = parmsParser.parse(initInfo.parms);
    
    if (pResult.rc != 0) 
    {
        System.out.println("Error parsing "+initInfo.name+" service parms.");
        result = pResult.rc;
    }
    
    if (pResult.optionTimes(LOGNAME) == 1) 
    {
        logName = pResult.optionValue(LOGNAME);
    }   

    /* Register RCs with help service. */
    /* Do not replace result int if it already contained an error */
    
    int regResult = registerHelp(initInfo.name);
    
    if (regResult != 0) 
    {
        System.out.println("Error registering " + initInfo.name +
                           " return codes with Help service.");
        
        if(result == 0) 
        {
            result = regResult;
        }
    }

    return result;
}

/**
 * This method determines if elapsedTime checking should be performed
 * for this execution. If both STAF variables STAF/Service/SXE/ElapsedTarget
 * and STAF/Service/SXE/ElapsedTolerance can be resolved, then elapsed time
 * checking should be performed. Note that a StringBuffer and 1 length arrays
 * are used, so that once the variables are resolved they may be used by the
 * calling method.
 * Returns true if elapsed time checking should be performed or false
 * otherwise.
 */

private boolean isCheckElapsed(STAFHandle fHandle, 
                               StringBuffer elapsedTargetSBuf,
                               long[] elapsedTarget, int[] elapsedTolerance)
                               throws STAFException, NumberFormatException 
{
    /* Resolve parms for ELAPSEDTARGET and ELAPSEDTOLERANCE, if both resolve
       OK, do elapsedTime checking */

    boolean checkElapsed = false;

    String targetReq = "HANDLE "+String.valueOf(fHandle.getHandle())+
        " RESOLVE "+ELAPSEDTARGETVAR;
    String tolReq = "HANDLE "+String.valueOf(fHandle.getHandle())+
        " RESOLVE "+ELAPSEDTOLVAR;
    String elapsedTargetString;
    
    try 
    {
        elapsedTargetString = fHandle.submit("LOCAL", "VAR", targetReq);
    }
    catch(STAFException se) 
    { 
        //don't throw STAFResult.VariableDoesNotExist 
        
        if (se.rc == STAFResult.VariableDoesNotExist) 
        {
            return false;
        }
        else 
        {
            throw new STAFException(se.rc, se.getMessage());
        }
    }
    
    elapsedTarget[0] = getTimeInMillis(elapsedTargetString);
    elapsedTolerance[0] = 
        Integer.parseInt(fHandle.submit("LOCAL", "VAR", tolReq));
    elapsedTargetSBuf.replace(0, 
        elapsedTargetSBuf.length(), elapsedTargetString);
    checkElapsed = true; // no excpetions, so check elapsed time
    
    return checkElapsed;
}

/**
 * Logs the start and success or failure of an individual command.
 * If the STAF variable STAF/Service/SXE/LogLevel is set to "none" 
 * or "file" then this logging will not occur.
 */

private void logCommand(String logName, STAFHandle handle, int logType, 
                        String command, STAFResult result) 
{
    STAFResult llResult = 
        handle.submit2("LOCAL", "VAR", "RESOLVE "+LOGLEVELVAR);
    String logLevel = llResult.result;
    
    if (llResult.rc == STAFResult.VariableDoesNotExist ||
        logLevel.equalsIgnoreCase(LOGNONE) || 
        logLevel.equalsIgnoreCase(LOGFILEONLY)) 
    {
            return;
    }

    String where = "local";
    String service = "LOG";
    String request = null;
    String logMessage;

    switch(logType) 
    {
        case START:
            logMessage = "Begin: ["+command+"]";
            logMessage = ":"+logMessage.length()+":"+logMessage;
            request = "LOG MACHINE LOGNAME "+logName+
                " LEVEL info MESSAGE "+logMessage;
            break;
        case STOPPASS:
            logMessage = "End: ["+command+"], "+result.result;
            logMessage = ":"+logMessage.length()+":"+logMessage;
            request = "LOG MACHINE LOGNAME "+logName+
                " LEVEL info MESSAGE "+logMessage;
            break;
        case STOPFAIL:
            logMessage = "End: "+result.result;
            logMessage = ":"+logMessage.length()+":"+logMessage;
            request = "LOG MACHINE LOGNAME "+logName+
                " LEVEL error MESSAGE "+logMessage;
            break;
    }
    
    handle.submit2(where, service, request);
}

/**
 * Logs the start and success or failure of the execution of a file.
 * If the STAF variable STAF/Service/SXE/LogLevel is set to "none" 
 * then this logging will not occur.
 */

private void logFile(String logName, int logType, STAFHandle handle, 
                     int loopNum, long startTime, long stopTime, 
                     String elapsedTargetString, int elapsedTolerance) 
{
    STAFResult llResult = 
        handle.submit2("LOCAL", "VAR", "RESOLVE "+LOGLEVELVAR);
    String logLevel = llResult.result;
    
    if (logLevel.equalsIgnoreCase(LOGNONE)) 
    { 
        // create log entry
        return;
    }

    String where = "local";
    String service = "LOG";
    String request;
    String logMessage;
    String level = null;
    
    switch(logType) 
    {
        case START:     
            logMessage = new String(String.valueOf(loopNum)+
                "#StartTime: "+getTime(startTime));
            logMessage = ":"+logMessage.length()+":"+logMessage;
            request = "LOG MACHINE LOGNAME "+logName+
                " LEVEL start MESSAGE "+logMessage;
            handle.submit2(where, service, request);
            break;
        case STOPPASS:
            level = "PASS";
        case STOPFAIL:
            if(level == null) 
            {
                level = "FAIL";
            }
            
            /* Log Stop */
            logMessage = new String(String.valueOf(loopNum)+
                "#StopTime: "+getTime(stopTime));
            logMessage = ":"+logMessage.length()+":"+logMessage;
            request = "LOG MACHINE LOGNAME "+logName+
                " LEVEL stop MESSAGE "+logMessage;
            handle.submit2(where, service, request);
            
            /* Log Pass/Fail */
            logMessage = new String(String.valueOf(loopNum)+"#ElapsedTime: "+
                getElapsedTime(startTime, stopTime)+
                " ElapsedTarget: "+elapsedTargetString+" ElapsedTolerance: "
                +elapsedTolerance+"%");
            logMessage = ":"+logMessage.length()+":"+logMessage;
            request = "LOG MACHINE LOGNAME "+logName+" LEVEL "+
                level+" MESSAGE "+logMessage;
            handle.submit2(where, service, request);
            break;
    }
}

/**
 * Resolves command parameters and prepares to execute the specified file.
 * Controls number of times file is executed.
 */

private STAFResult prepare2Execute(
    STAFServiceInterfaceLevel4.RequestInfo reqInfo)
{
    /* Check for trust level of 3 or greater */

    if (reqInfo.trustLevel < 3) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }

    /* Parse command */
    
    STAFCommandParseResult pResult = sxeParser.parse(reqInfo.request);

    if (pResult.rc != 0) 
    {
        return new STAFResult(STAFResult.InvalidRequestString, 
            pResult.errorBuffer);
    }

    if (reqInfo.diagEnabled == 1)
    {
        // Record diagnostics because the output for an EXECUTE request will
        // change in STAF V3.0

        String trigger = new String("STAF/V3.0-Mig/SXE EXECUTE");
        String source  = new String(reqInfo.processName + ";" +
                                    reqInfo.machine + ";" +
                                    reqInfo.handle);
        sHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "DIAG",
            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
            " SOURCE " + STAFUtil.wrapData(source));
    }

    /* get filename to process */
    
    String filename;
    
    try 
    {
        filename = getResolvedParm(reqInfo, pResult, FILE);
    }
    catch (STAFException se) 
    {
        return new STAFResult(STAFResult.InvalidRequestString);
    }

    /* Get number of times to loop through file */

    int numLoops = 1; //default of 1
    boolean forever = false;

    if (pResult.optionTimes(LOOP) == 1) 
    {
        try 
        {
            numLoops = Integer.parseInt(getResolvedParm(reqInfo, 
                pResult, LOOP));
            
            if (numLoops <= 0) 
            {
                return new STAFResult(STAFResult.InvalidRequestString, 
                    "Invalid LOOP value");
            }
        }
        catch (STAFException se) 
        {
            return new STAFResult(STAFResult.InvalidRequestString);
        }
        catch (NumberFormatException nfe) 
        {
            if(nfe.getMessage().equalsIgnoreCase(FOREVER)) 
            {
                // see if forever specified
                forever = true;
            }
            else 
            {
                return new STAFResult(STAFResult.InvalidRequestString, 
                    "Invalid LOOP value");
            }
        }
    }

    /* Get minRuntime */

    boolean useMinRuntime = false;
    long minRuntime = 0;

    if (pResult.optionTimes(MINRUNTIME) == 1) 
    {
        try 
        {
            String minRuntimeString = getResolvedParm(reqInfo, 
                pResult, MINRUNTIME);
            minRuntime = getTimeInMillis(minRuntimeString+":0.0");
            useMinRuntime = true;
        }
        catch(NumberFormatException nfe) 
        {
            return new STAFResult(STAFResult.InvalidRequestString, 
                "Invalid MINRUNTIME value");
        }
        catch(STAFException se) 
        {
            return new STAFResult(se.rc, se.getMessage());          
        }
    }
    
    /* Create a new STAFHandle for each file passed to SXE */

    STAFHandle fHandle;
    
    try 
    {
        fHandle = new STAFHandle(filename);
    }
    catch(STAFException se) 
    {
        return new STAFResult(REGERROR, String.valueOf(se.rc));
    }

    STAFResult result = null;

    int loop;

    if (forever) 
    {
        loop = 1;
        
        while(true) 
        { 
            //loop forever
            setSXELoopVar(loop, fHandle);
            result = execute(fHandle, filename, logName, loop);
            
            if (result.rc != STAFResult.Ok) 
            {
                break;
            }
            
            loop++;
        }
    }
    else if (useMinRuntime) 
    { 
        //use the minRuntime parameter rather than loop        
        loop = 1;
        long tcStartTime = new Date().getTime();
        
        while(true) 
        {
            result = execute(fHandle, filename, logName, loop);
            long tcStopTime = new Date().getTime();
            
            if (result.rc != STAFResult.Ok) 
            {
                break;
            }
            
            if (tcStopTime - tcStartTime >= minRuntime) 
            {
                break;
            }//do checking for minruntime elapsed
            
            loop++;
        }
    }
    else 
    {
        for (loop = 1; loop <= numLoops; loop++) 
        {
            //loop set number of times
            setSXELoopVar(loop, fHandle);
            result = execute(fHandle, filename, logName, loop);
            
            if (result.rc != STAFResult.Ok) 
            {
                break;
            }
        }
    }

    try 
    {
        fHandle.unRegister();
    }
    catch(STAFException se) 
    {
        se.printStackTrace();
    }

    return result;    
}

/**
 * Registers return code information with the help service
 */

private int registerHelp(String name) 
{
    try 
    {
        String request = "REGISTER SERVICE "+name+" ERROR "+FILEOPENERROR+
            " INFO \""+FILEOPENERRORInfo+"\" DESCRIPTION \""+
            FILEOPENERRORDesc+"\"";
        sHandle.submit("LOCAL", "HELP", request);

        request = "REGISTER SERVICE "+name+" ERROR "+IOERROR+
            " INFO \""+IOERRORInfo+"\" DESCRIPTION \""+IOERRORDesc+"\"";
        sHandle.submit("LOCAL", "HELP", request);
        
        request = "REGISTER SERVICE "+name+" ERROR "+REGERROR+
            " INFO \""+REGERRORInfo+"\" DESCRIPTION \""+REGERRORDesc+"\"";
        sHandle.submit("LOCAL", "HELP", request);
        
        request = "REGISTER SERVICE "+name+" ERROR "+ELAPSEDTIMEFAIL+
            " INFO \""+ELAPSEDTIMEFAILInfo+"\" DESCRIPTION \""+
            ELAPSEDTIMEFAILDesc+"\"";
        sHandle.submit("LOCAL", "HELP", request);
        
        request = "REGISTER SERVICE "+name+" ERROR "+ELAPSEDTIMEERROR+
            " INFO \""+ELAPSEDTIMEERRORInfo+"\" DESCRIPTION \""+
            ELAPSEDTIMEERRORDesc+"\"";
        sHandle.submit("LOCAL", "HELP", request);
    }
    catch(STAFException se) 
    {
        return se.rc;
    }

    return STAFResult.Ok;
}


/**
 * Un-registers return code information with the help service
 */

private int unregisterHelp(String name) 
{
    try 
    {
        String request = "UNREGISTER SERVICE " + name + " ERROR " +
            FILEOPENERROR;
        sHandle.submit("LOCAL", "HELP", request);

        request = "UNREGISTER SERVICE " + name + " ERROR " + IOERROR;
        sHandle.submit("LOCAL", "HELP", request);
        
        request = "UNREGISTER SERVICE " + name + " ERROR " + REGERROR;
        sHandle.submit("LOCAL", "HELP", request);
        
        request = "UNREGISTER SERVICE " + name + " ERROR " + ELAPSEDTIMEFAIL;
        sHandle.submit("LOCAL", "HELP", request);
        
        request = "UNREGISTER SERVICE " + name + " ERROR " + ELAPSEDTIMEERROR;
        sHandle.submit("LOCAL", "HELP", request);
    }
    catch(STAFException se) 
    {
        return se.rc;
    }
    
    return STAFResult.Ok;
}

/**
 * Executes a staf command using the given handle.
 */

private STAFResult sendSTAFCommand(STAFHandle handle, String command) 
{
    STAFResult result;

    String where = command.substring(0, command.indexOf(" "));
    String service = command.substring(command.indexOf(" ")+1,
                    command.indexOf(" ", command.indexOf(" ")+1));
    String request = command.substring(command.indexOf(" ", 
        command.indexOf(" ")+1)+1);

    return handle.submit2(where, service, request);
}

/**
 * Sets a handle variable "sxeloop" to the specified integer.
 */

private void setSXELoopVar(int loopNum, STAFHandle fHandle) 
{    
    String request = "SET sxeloop="+String.valueOf(loopNum);
    fHandle.submit2("LOCAL", "VAR", request);   
   
}

/**
 * Required by implemented STAF interface.
 * This method is called when shutting the service down.
 */

public int term() 
{
    // Un-register Help data

    int rc = unregisterHelp(sServiceName);
    
    // Un-register the service handle

    try
    {
        sHandle.unRegister();
    }
    catch (STAFException ex)
    {
        if (rc == STAFResult.Ok)
        {
            rc = STAFResult.STAFRegistrationError;
        }
    }

    return rc;
}

/**
 * Outputs SXE Service version information.
 */

private STAFResult version() 
{
    return new STAFResult(STAFResult.Ok, SXEVERSION);
}
}
