package com.ibm.staf.service.timer;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/****************************************************************************/
//
// Class: Timer
//
// Logical Name: Timer.java
//
// Description: This class accepts requests from STAF.
//
//
// History:
//
// DATE       DEVELOPER   CHANGE DESCRIPTION
// ---------- ----------- -----------------------------------
// 02/01/2000 C Alkov     Original Program
// 02/21/2004 S. Lucas    Changed from STAFServiceInterfaceLevel1 to
//                        STAFServiceInterfaceLevel4 (Version 1.1.0)
// 
//
/****************************************************************************/

import com.ibm.staf.*;
import com.ibm.staf.wrapper.STAFLog;
import com.ibm.staf.service.STAFServiceInterfaceLevel4;

public class Timer implements STAFServiceInterfaceLevel4 
{    
    TimerRequestHandler reqHandler;
    private TimerCmdParser cmdParser;
    STAFHandle sHandle;
    STAFHandle wHandle;
    String sServiceName;
    String localMachine;
    STAFLog log;
    String tListFileName;
    String wListFileName;
    private String FS;
            
/****************************************************************************/
//
// Method: 
//   Constructor
//
// Description:
//   Constructor method for Timer class.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public Timer() 
{
    /* Create command parser */

    cmdParser = new TimerCmdParser(this);
        
}

/***************************************************************************/
/* acceptRequest- Calls appropriate methods to process a request from a    */
/*                client process.                                          */
/*                                                                         */
/* accepts: STAFServiceInterfaceLevel4 request information                 */
/*                                                                         */
/* returns: STAFResult.rc = STAFResult.Ok, if successful; STAFResult.      */
/*          InvalidRequestString, if unsuccessful;                         */
/*                                                                         */
/*          STAFResult.result contains the result returned by the called   */
/*          method, if successful;                                         */
/*          STAFResult.result contains the command, if unsuccessful        */
/***************************************************************************/
public STAFResult acceptRequest(STAFServiceInterfaceLevel4.RequestInfo info)
{
    STAFResult result = cmdParser.parse(info);
            
    return result;
}

/***************************************************************/
/* init - Initializes Timer Service with STAF.                 */
/*                                                             */
/* accepts: STAFServiceInterfaceLevel4 initialization info     */
/*          Parms includes a string representation of the      */
/*            directory to use for persistent storage.         */
/*                                                             */
/* Returns: STAFResult.Ok or STAFResult.STAFRegistrationError  */
/***************************************************************/
public int init(STAFServiceInterfaceLevel4.InitInfo info)
{
    /* Set service name */

    sServiceName = info.name;

    /* Register with STAF */
    
    try 
    {
        sHandle = new STAFHandle("STAF/Service/" + sServiceName);
    }
    catch (STAFException e) 
    {
        return e.rc;
    }

    /* Create a process for watch manager, this process will only be used as an exclusive
       queue for messages, all other requests will be handled by the sHandle process */

    try 
    {
        wHandle = new STAFHandle(sServiceName+"/watch");
    }
    catch (STAFException e) 
    {
        return e.rc;
    }

    /* Create STAFlog instance */

    log = new STAFLog(STAFLog.MACHINE, "STAFSERVICE_" + sServiceName, sHandle);

    /* Get file separator */

    try 
    {
        FS = sHandle.submit("LOCAL", "VAR", 
            "GLOBAL RESOLVE {STAF/Config/Sep/File}");
    }
    catch (STAFException e) 
    {
        log.log(STAFLog.Warning, 
            "Unexpected error attempting to resolve {STAF/Config/Sep/File}." +
            "Will not use persistent storage.");
        FS = null;
    }

    /* Create String to represent filenames for persistent storage */
    /* skip if could not resolve file separator */

    if (FS != null) 
    {
        if (info.parms.equals(new String())) 
        {
            /* No parm set, try to use {STAF/Config/STAFRoot}\bin */
            try 
            {
                String stafRoot = sHandle.submit("LOCAL", "VAR", 
                    "GLOBAL RESOLVE {STAF/Config/STAFRoot}");
                tListFileName = stafRoot + FS + "bin" + FS + "tlist.ser";
                wListFileName = stafRoot + FS + "bin" + FS + "wlist.ser";
            }
            catch (STAFException e) 
            {
                log.log(STAFLog.Warning, 
                    "Unexpected error attempting to resolve " +
                    "{STAF/Config/STAFRoot}." +
                    "Will not use persistent storage.");
                tListFileName = null;
                wListFileName = null;
            }
        }
        else 
        {
            tListFileName = info.parms + FS + "tlist.ser";
            wListFileName = info.parms + FS + "wlist.ser";
        }
    }
            
    /* Create request handler */

    reqHandler = new TimerRequestHandler(sServiceName, sHandle, this);

    /* Get local machine name */

    try 
    {
        localMachine = sHandle.submit("LOCAL", "VAR", 
            "GLOBAL RESOLVE {STAF/Config/Machine}");
    }
    catch (STAFException e) 
    {
        log.log(STAFLog.Warning, 
            "Unexpected error attempting to resolve local machine name.RC=" + 
            e.rc);
    }

    // Register Help Data

    registerHelpData(
        TimerUtils.INVALIDFREQUENCYRC,
        "Invalid Frequency", 
        "The FREQUENCY specified in the REGISTER command conflicts with " +
        "the minimum allowable frequency set by the variable MinFrequency.");
    
    registerHelpData(
        TimerUtils.INVALIDPRIORITYRC,
        "Invalid Priority",
        "The named counter is currently at Zero.");

    registerHelpData(
        TimerUtils.INVALIDTYPERC,
        "Invalid Type",
        "The Type specified in the UNREGISTER command does not exist for " +
        "this machine and process.");

    registerHelpData(
        TimerUtils.MACHINENOTWATCHEDRC,
        "Machine Not Watched",
        "The machihe specified in the UNWATCH command is not currently " +
        "being watched.");
   
    /* Get Timer variables from STAF using VAR service */

    STAFResult result = reqHandler.refresh();

    return result.rc;
    
}

/****************************************************************************/
//
// Method: 
//   term
//
// Description:
//   The termination method for the Timer service.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public int term() 
{
    // Un-register Help Data

    unregisterHelpData(TimerUtils.INVALIDFREQUENCYRC);
    unregisterHelpData(TimerUtils.INVALIDPRIORITYRC);
    unregisterHelpData(TimerUtils.INVALIDTYPERC);
    unregisterHelpData(TimerUtils.MACHINENOTWATCHEDRC);
    
    int rc = reqHandler.term();
    
    return rc;

}



    // Register error codes for the service with the HELP service

    private void registerHelpData(int errorNumber, String info,
                                 String description)
    {
        STAFResult res = sHandle.submit2("local", "HELP",
                         "REGISTER SERVICE " + sServiceName +
                         " ERROR " + errorNumber +
                         " INFO " + STAFUtil.wrapData(info) +
                         " DESCRIPTION " + STAFUtil.wrapData(description));
    }

    // Un-register error codes for the service with the HELP service

    private void unregisterHelpData(int errorNumber)
    {
        STAFResult res = sHandle.submit2("local", "HELP",
                         "UNREGISTER SERVICE " + sServiceName +
                         " ERROR " + errorNumber);
    }
}
