package com.ibm.staf.service.timer;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/****************************************************************************/
//
// Class: TimerCmdParser
//
// Logical Name: TimerCmdParser.java
//
// Description: This class parsers commands received from STAF.
//
//
// History:
//
// DATE       DEVELOPER   CHANGE DESCRIPTION
// ---------- ----------- -----------------------------------
// 02/01/2000 C Alkov     Original Program
// 04/05/2000 C Alkov     Fix error checking for unwatch parser (def:3610)
// 02/21/2004 S Lucas     Changed from STAFServiceInterfaceLevel1 to
//                        STAFServiceInterfaceLevel4
//
/****************************************************************************/

import com.ibm.staf.*;
import com.ibm.staf.service.*;
import com.ibm.staf.wrapper.STAFLog;
import java.util.Enumeration;

public class TimerCmdParser 
{
    private Timer fTimer;
    
    private STAFCommandParser registerParser;
    private STAFCommandParser unregisterParser;
    private STAFCommandParser listParser;
    private STAFCommandParser watchParser;
    private STAFCommandParser unwatchParser;
    
    private final static String REGISTER = "REGISTER";
    private final static String UNREGISTER = "UNREGISTER";
    private final static String LIST = "LIST";
    private final static String SP = " ";
    private final static String TYPE = "TYPE";
    private final static String FREQUENCY = "FREQUENCY";
    private final static String PRIORITY = "PRIORITY";
    private final static String MACHINE = "MACHINE";
    private final static String HANDLE = "HANDLE";
    private final static String TIMERS = "TIMERS";
    private final static String HELP = "HELP";
    private final static String VERSION = "VERSION";
    private final static String REFRESH = "REFRESH";
    private final static String UNREGONNOPATH = "UNREGONNOPATH";
    private final static String UNREGONNOHANDLE = "UNREGONNOHANDLE";
    private final static String PROCESS = "PROCESS";
    private final static String BYNAME = "BYNAME";
    private final static String FORMAT = "FORMAT";
    private final static String WATCHES = "WATCHES";
    private final static String WATCH = "WATCH";
    private final static String UNWATCH = "UNWATCH";
    private final static String MARGIN = "MARGIN";

    private final static String INVALIDOPTION =
      "You must have at least 1, but no more than 1 of the options,"+
      " REGISTER UNREGISTER LIST VERSION REFRESH HELP";

    private final static String BADFREQUENCY =
      "The value for FREQUENCY must be an integer.";

    private final static String BADMARGIN =
      "The value for MARGIN must be an integer.";  

    private final static String BADPRIORITY =
      "The value for PRIORITY must be an integer.";

    private final static String BADHANDLE =
      "The value for HANDLE must be an integer.";  
    
/****************************************************************************/
//
// Method: 
//   Constructor
//
// Description:
//   Constructor method for TimerCmdParser class.
//
// Input:
//   timer - The instance of the Timer service creating this CmdParser.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public TimerCmdParser(Timer timer) 
{
    fTimer = timer;         

    /* Create registerParser */

    registerParser = new STAFCommandParser();
    registerParser.addOption(REGISTER, 1, STAFCommandParser.VALUENOTALLOWED);
    registerParser.addOption(TYPE, 1, STAFCommandParser.VALUEREQUIRED);
    registerParser.addOption(FREQUENCY, 1, STAFCommandParser.VALUEREQUIRED);
    registerParser.addOption(PRIORITY, 1, STAFCommandParser.VALUEREQUIRED);
    registerParser.addOption(UNREGONNOPATH, 1, 
        STAFCommandParser.VALUEREQUIRED);
    registerParser.addOption(UNREGONNOHANDLE, 1, 
        STAFCommandParser.VALUEREQUIRED);
    registerParser.addOption(BYNAME, 1, STAFCommandParser.VALUENOTALLOWED);
    registerParser.addOptionGroup(REGISTER, 1, 1);
    registerParser.addOptionGroup(TYPE, 1, 1);
    registerParser.addOptionGroup(FREQUENCY, 1, 1);

    /* Create unregisterParser */

    unregisterParser = new STAFCommandParser();
    unregisterParser.addOption(UNREGISTER, 1, 
        STAFCommandParser.VALUENOTALLOWED);
    unregisterParser.addOption(TYPE, 1, STAFCommandParser.VALUEREQUIRED);
    unregisterParser.addOption(MACHINE, 1, STAFCommandParser.VALUEREQUIRED);
    unregisterParser.addOption(HANDLE, 1, STAFCommandParser.VALUEREQUIRED);
    unregisterParser.addOption(PROCESS, 1, STAFCommandParser.VALUEREQUIRED);
    unregisterParser.addOption(BYNAME, 1, STAFCommandParser.VALUENOTALLOWED);
    unregisterParser.addOptionGroup(UNREGISTER, 1, 1);
    unregisterParser.addOptionGroup(TYPE, 1, 1);
    unregisterParser.addOptionGroup(PROCESS+SP+HANDLE, 0, 1);
    unregisterParser.addOptionGroup(MACHINE+SP+BYNAME, 0, 1);
    unregisterParser.addOptionNeed(MACHINE, HANDLE+SP+PROCESS);
    unregisterParser.addOptionNeed(HANDLE, MACHINE);
    unregisterParser.addOptionNeed(PROCESS, MACHINE);

    /* Create listParser */
    /* This is trivial and not really neccessary for initial version.
       It has been included here because it will be needed when watches added */
       
    listParser = new STAFCommandParser();
    listParser.addOption(LIST,      1, STAFCommandParser.VALUENOTALLOWED);
    listParser.addOption(TIMERS,    1, STAFCommandParser.VALUENOTALLOWED);
    listParser.addOption(WATCHES,   1, STAFCommandParser.VALUENOTALLOWED);
    listParser.addOption(FORMAT, 1, STAFCommandParser.VALUENOTALLOWED);
    listParser.addOptionGroup(LIST, 1, 1);
    listParser.addOptionGroup(TIMERS+SP+WATCHES, 1, 1);

    /* Create watchParser */

    watchParser = new STAFCommandParser();
    watchParser.addOption(WATCH, 1, STAFCommandParser.VALUENOTALLOWED);
    watchParser.addOption(MACHINE, 1, STAFCommandParser.VALUEREQUIRED);
    watchParser.addOption(FREQUENCY, 1, STAFCommandParser.VALUEREQUIRED);
    watchParser.addOption(MARGIN, 1, STAFCommandParser.VALUEREQUIRED);
    watchParser.addOptionGroup(WATCH, 1, 1);
    watchParser.addOptionGroup(MACHINE, 1, 1);
    watchParser.addOptionGroup(FREQUENCY, 1, 1);

    /* Create unwatchParser */

    unwatchParser = new STAFCommandParser();
    unwatchParser.addOption(UNWATCH, 1, STAFCommandParser.VALUENOTALLOWED);
    unwatchParser.addOption(MACHINE, 1, STAFCommandParser.VALUEREQUIRED);
    unwatchParser.addOptionGroup(UNWATCH, 1, 1);
    unwatchParser.addOptionGroup(MACHINE, 1, 1);    
    
}

/****************************************************************************/
//
// Method: 
//   parse
//
// Description:
//   This method parses the first word of the request and handles
//   accordingly.
//
// Input:
//   machine - the machine name making the request
//   process - the name of the process making the request
//   handle - the handle of the process making the request
//   request - the request in String format
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult parse(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    info.request = info.request.toUpperCase();
    STAFResult result = null;

    if (info.request.startsWith(REGISTER)) 
    {
        result = parseRegister(info);
    }
    else if (info.request.startsWith(UNREGISTER)) 
    {
        result = parseUnregister(info);
    }
    else if (info.request.startsWith(WATCH)) 
    {
        result = parseWatch(info);
    }
    else if (info.request.startsWith(UNWATCH)) 
    {
        result = parseUnwatch(info);
    }
    else if (info.request.startsWith(LIST)) 
    {
        result = parseList(info);
    }
    else if (info.request.startsWith(HELP)) 
    {
        result = fTimer.reqHandler.help(fTimer.sServiceName);
    }
    else if (info.request.startsWith(VERSION)) 
    {
        result = fTimer.reqHandler.version(info);
    }
    else if (info.request.startsWith(REFRESH)) 
    {
        if (info.trustLevel < 3)
        {
            return new STAFResult(STAFResult.InvalidTrustLevel);
        }

        result = fTimer.reqHandler.refresh();
    }
    else 
    {
        result = new STAFResult(STAFResult.InvalidRequestString, INVALIDOPTION);
    }        
    
    return result;
}

/****************************************************************************/
//
// Method: 
//   parseList
//
// Description:
//   This method parses parameters for the List command.
//
// Input:
//   STAFServiceInterfaceLevel4 request information
//   (including the submitting machine name, the submitting process.
//    the submitting process handle, a string containing any command
//    options & option values).
//
// Exceptions Thrown:
//   none
//
// Notes:
//   private method
//
/****************************************************************************/

private STAFResult parseList(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    STAFCommandParseResult pResult = listParser.parse(info.request);

    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }
    
    if (info.trustLevel < 2)
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }
        
    boolean format = false;

    if (pResult.optionTimes(FORMAT) == 1) 
    {
        format = true;
    }

    if (pResult.optionTimes(TIMERS) == 1) 
    {
        /* Call List Timers request handler */
        return fTimer.reqHandler.listTimers(info, format);
    }
    else if (pResult.optionTimes(WATCHES) == 1) 
    {
        /* Call List Watches request handler */
        return fTimer.reqHandler.listWatches(info, format);
    }
    else 
    {
        /* Should not go here, but return InvalidRequestString */
        return new STAFResult(STAFResult.InvalidRequestString);
    }
}

/****************************************************************************/
//
// Method: 
//   parseRegister
//
// Description:
//   This method parses parameters for the Register command.
//
// Input:
//   STAFServiceInterfaceLevel4 request information
//   (including the submitting machine name, the submitting process.
//    the submitting process handle, a string containing any command
//    options & option values).
//
// Exceptions Thrown:
//   none
//
// Notes:
//   private method
//
/****************************************************************************/

private STAFResult parseRegister(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    STAFCommandParseResult pResult = registerParser.parse(info.request);
    int unregOnNoPath = -1;
    int unregOnNoHandle = -1;
    boolean byname = false;
        
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    if (info.trustLevel < 3)
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }

    long frequency;
    long priority = 5; // set default priority

    String name = TimerUtils.resolveLocalVar(fTimer.localMachine, 
        fTimer.sHandle, info.machine, info.handle, pResult.optionValue(TYPE), 
        fTimer);
                
    try 
    {
        frequency = 
            Long.parseLong(TimerUtils.resolveLocalVar(fTimer.localMachine,
            fTimer.sHandle, info.machine, info.handle, 
            pResult.optionValue(FREQUENCY), fTimer));
    }
    catch(NumberFormatException e) 
    {
        /* value not an int */
        return new STAFResult(STAFResult.InvalidRequestString,
                              BADFREQUENCY);
    }

    /* Get optional parameter if specified */
    
    if (pResult.numInstances() > 3) 
    {
        /* If priority option exists get value */
        if (pResult.optionTimes(PRIORITY) == 1) 
        {
            try 
            {
                priority = Long.parseLong(TimerUtils.resolveLocalVar(
                    fTimer.localMachine, fTimer.sHandle, info.machine,
                    info.handle, pResult.optionValue(PRIORITY), fTimer));
            }
            catch(NumberFormatException e) 
            {
                /* value not an int */
                return new STAFResult(STAFResult.InvalidRequestString, 
                                      BADPRIORITY);
            }
        }

        /* If unregOnNoPath option exists get value */
        if (pResult.optionTimes(UNREGONNOPATH) == 1) 
        {
            try 
            {
                unregOnNoPath = Integer.parseInt(TimerUtils.resolveLocalVar(
                    fTimer.localMachine, fTimer.sHandle, info.machine,
                    info.handle, pResult.optionValue(UNREGONNOPATH), fTimer));
            }
            catch(NumberFormatException e) 
            {
                fTimer.log.log(STAFLog.Warning, "Invalid value specified "+
                    "for UnregOnNoPath, no timer specific value set.");
            }
        }
            
        /* If unregOnNoHandle option exists get value */
        if (pResult.optionTimes(UNREGONNOHANDLE) == 1) 
        {
            try 
            {
                unregOnNoHandle = 
                    Integer.parseInt(TimerUtils.resolveLocalVar(
                    fTimer.localMachine, fTimer.sHandle, info.machine,
                    info.handle, pResult.optionValue(UNREGONNOHANDLE),
                    fTimer));
            }
            catch(NumberFormatException e) 
            {
                fTimer.log.log(STAFLog.Warning, "Invalid value specified "+
                    "for UnregOnNoHandle, no timer specific value set.");
            }
        }
            
        /* If byname option specified register byname */
        if (pResult.optionTimes(BYNAME) == 1) 
        {
            byname = true;
        }
    }
            
    /* Call Register request handler */

    return fTimer.reqHandler.register(
        name, info.machine, info.handle, info.processName, 
        frequency, priority, unregOnNoPath, unregOnNoHandle, byname);
}

/****************************************************************************/
//
// Method: 
//   parseUnregister
//
// Description:
//   This method parses parameters for the Unregister command.
//
// Input:
//   STAFServiceInterfaceLevel4 request information
//   (including the submitting machine name, the submitting process.
//    the submitting process handle, a string containing any command
//    options & option values).
//
// Exceptions Thrown:
//   none
//
// Notes:
//   private method
//   Trust level checking is done here because we must know the machine the
//   timer is running on and the machine submitting the request.
//
/****************************************************************************/

private STAFResult parseUnregister(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    STAFCommandParseResult pResult = unregisterParser.parse(info.request);
    int reqdTrustLevel = 3;
    boolean byname = false;

    /* Parse command */

    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    if (info.trustLevel < 3)
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }
        
    String name = TimerUtils.resolveLocalVar(fTimer.localMachine, 
        fTimer.sHandle, info.machine, info.handle, 
        pResult.optionValue(TYPE), fTimer);
    String fMachine = info.machine;
    int fHandle = info.handle;
    String fProcess = info.processName;

    /* Get optional parameters if specified */

    if (pResult.numInstances() > 2) 
    {
        if (pResult.optionTimes(MACHINE) == 1) 
        {
            /* User specified machine and handle or process to unregister */
                
            fMachine = TimerUtils.resolveLocalVar(fTimer.localMachine, 
                fTimer.sHandle, info.machine, info.handle, 
                pResult.optionValue(MACHINE), fTimer);
            
            /* Make sure we use long name even if short name was passed */          
            try 
            {
                String varRequest = "RESOLVE {STAF/Config/Machine}";
                fMachine = fTimer.sHandle.submit(fMachine, 
                    "VAR", varRequest);
            }
            catch (STAFException e) 
            {
                /* Cannot resolve: log error and continue with original parameter */
                fTimer.log.log(STAFLog.Warning, 
                    "Unexpected error attempting to resolve machine " + 
                    "name for machine:"+fMachine+" RC="+e.rc);
            }
    
            if(pResult.optionTimes(HANDLE) == 1) 
            {
                /* Use handle */
                
                try 
                {
                    fHandle = Integer.parseInt(TimerUtils.resolveLocalVar(
                        fTimer.localMachine, fTimer.sHandle, info.machine, 
                        info.handle, pResult.optionValue(HANDLE), fTimer));
                }
                catch (NumberFormatException e) 
                {
                    /* value not an int */
                    return new STAFResult(STAFResult.InvalidRequestString, 
                                         BADHANDLE);
                }
            }
            else 
            {
                /* Use process */
                fProcess = TimerUtils.resolveLocalVar(fTimer.localMachine, 
                    fTimer.sHandle, info.machine, info.handle, 
                    pResult.optionValue(PROCESS), fTimer);
                byname = true;
            }
        }

        if (pResult.optionTimes(BYNAME) == 1) 
        {
            /* User specified BYNAME */
            byname = true;
        }

        /* Modify reqdTrustLevel if necessary */

        if (!(fMachine.equals(info.machine))) 
        {
            reqdTrustLevel = 4;
        }
    }

    /* Check for proper trust level */

    if(info.trustLevel < reqdTrustLevel) 
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }

    /* Create proper timer string for timer to unregister */

    String tString;
  
    if (byname) 
    {
        tString = TimerUtils.createTimerString(name, fMachine, fProcess);
    }
    else 
    {
        tString = TimerUtils.createTimerString(name, fMachine, fHandle);
    }

    /* Call UnRegister request handler */
    
    return fTimer.reqHandler.unregister(tString);
}

/****************************************************************************/
//
// Method: 
//   parseUnwatch
//
// Description:
//   This method parses parameters for the Unwatch command.
//
// Input:
//   STAFServiceInterfaceLevel4 request information
//   (including the submitting machine name, the submitting process.
//    the submitting process handle, a string containing any command
//    options & option values).
//
// Exceptions Thrown:
//   none
//
// Notes:
//   private method
//   Trust level checking is done here because we must know the machine the
//   timer is running on and the machine submitting the request.
//
/****************************************************************************/

private STAFResult parseUnwatch(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    STAFCommandParseResult pResult = unwatchParser.parse(info.request); //3610
    
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);       //3610
    }

    if (info.trustLevel < 4)
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }

    String machineToWatch = pResult.optionValue(MACHINE);

    /* Make sure we use long name even if short name was passed */
    
    try 
    {
        String varRequest = "RESOLVE {STAF/Config/Machine}";
        machineToWatch = fTimer.sHandle.submit(machineToWatch, 
            "VAR", varRequest);
        machineToWatch = machineToWatch.toUpperCase();
    }
    catch (STAFException e) 
    {
        /* Cannot resolve: log error and continue with original parameter */
        fTimer.log.log(STAFLog.Warning, "Unexpected error attempting to " + 
            "resolve machine name for machine:"+machineToWatch+" RC="+e.rc);
    }

    return fTimer.reqHandler.unwatch(machineToWatch);
}

/****************************************************************************/
//
// Method: 
//   parseWatch
//
// Description:
//   This method parses parameters for the Watch command.
//
// Input:
//   STAFServiceInterfaceLevel4 request information
//   (including the submitting machine name, the submitting process.
//    the submitting process handle, a string containing any command
//    options & option values).
//
// Exceptions Thrown:
//   none
//
// Notes:
//   private method
//   Trust level checking is done here because we must know the machine the
//   timer is running on and the machine submitting the request.
//
/****************************************************************************/

private STAFResult parseWatch(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    STAFResult result;
    STAFCommandParseResult pResult = watchParser.parse(info.request);
            
    if (pResult.rc != 0) 
    {
        return new STAFResult(pResult.rc, pResult.errorBuffer);
    }

    if (info.trustLevel < 3)
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }

    long frequency;
    int margin = 0;
        
    String machineToWatch = TimerUtils.resolveLocalVar(
        fTimer.localMachine, fTimer.sHandle,
        info.machine, info.handle, 
        pResult.optionValue(MACHINE), fTimer);

    try 
    {
        frequency = Long.parseLong(TimerUtils.resolveLocalVar(
            fTimer.localMachine, fTimer.sHandle,
            info.machine, info.handle,
            pResult.optionValue(FREQUENCY), fTimer));
    }
    catch(NumberFormatException e) 
    {
        /* value not an int */
        return new STAFResult(STAFResult.InvalidRequestString,
                              BADFREQUENCY);
    }

    if (pResult.optionTimes(MARGIN) == 1) 
    {
        try 
        {
            margin = Integer.parseInt(TimerUtils.resolveLocalVar(
                fTimer.localMachine, fTimer.sHandle,
                info.machine, info.handle,
                pResult.optionValue(MARGIN), fTimer));
        }
        catch(NumberFormatException e) 
        {
            /* value not an int */
            return new STAFResult(STAFResult.InvalidRequestString, 
                                  BADMARGIN);
        }
    }

    return fTimer.reqHandler.watch(machineToWatch, frequency, margin);
}
}