package com.ibm.staf.service.timer;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/****************************************************************************/
//
// Class: TimerList
//
// Logical Name: TimerList.java
//
// Description: This class is a customized HashTable for storing the 
//              registered timers.
//
//
// History:
//
// DATE       DEVELOPER   CHANGE DESCRIPTION
// ---------- ----------- -----------------------------------
// 02/01/2000 C Alkov     Original Program
//
/****************************************************************************/

import java.util.Enumeration;
import java.util.Hashtable;
import java.io.Serializable;

public class TimerList extends Hashtable implements Serializable 
{
    /* Do not serialize (transient) this instance variable since
       reference will be different from run to run */

    transient TimerRequestHandler reqHandler;
    
/****************************************************************************/
//
// Method: 
//   Constructor
//
// Description:
//   Constructor method for TimerList class.
//
// Input:
//   freqHandler - The instance of the TimerRequestHandler creating 
//   this TimerList.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public TimerList(TimerRequestHandler freqHandler) 
{
    super();

    reqHandler = freqHandler;
    
}

/****************************************************************************/
//
// Method: 
//   containsTimer
//
// Description:
//   Determines if the specified timer exists in the list.
//
// Input:
//   tString - The unique String representation of the timer to 
//   look for.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public boolean containsTimer(String tString) 
{    
    return this.containsKey(tString);
    
}

/****************************************************************************/
//
// Method: 
//   getTimerObject
//
// Description:
//   Returns the TimerObject corresponding to the String timerString.
//
// Input:
//   timerString - The unique String representation of the timer.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public TimerObject getTimerObject(String timerString) 
{    
    return (TimerObject) this.get(timerString);
    
}
/****************************************************************************/
//
// Method: 
//   listFormattedTimers
//
// Description:
//   Returns a String list of all timers in this TimerList. Used in STAF
//   request LIST TIMERS FORMAT.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public String listFormattedTimers() 
{
    StringBuffer list = new StringBuffer();
    
    for (Enumeration e = this.elements(); e.hasMoreElements();) 
    {
        TimerObject to = (TimerObject) e.nextElement();
        list.append(to.getFormattedProperties()).append(reqHandler.LINESEP);
    }
        
    return list.toString();
}

/****************************************************************************/
//
// Method: 
//   listTimers
//
// Description:
//   Returns a String list of all timers in this TimerList.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public String listTimers() 
{
    StringBuffer list = new StringBuffer();
    
    for (Enumeration e = this.elements(); e.hasMoreElements();) 
    {
        TimerObject to = (TimerObject) e.nextElement();
        list.append(to.getProperties()).append(reqHandler.LINESEP);
    }
        
    return list.toString();
}

/****************************************************************************/
//
// Method: 
//   setReqHandler
//
// Description:
//   Sets the RequestHandler for the TimerList object, since that instance
//   variable cannot be serialized. (Reference changes from run to run.)  Then
//   goes through list and sets the RequestHandler for each TimerObject in list.
//
// Input:
//   aReqHandler - Reference to the current RequestHandler
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public void setReqHandler(TimerRequestHandler aReqHandler) 
{
    /* Set reqHandler for TimerList Object */
    
    reqHandler = aReqHandler;

    /* Set reqHandler for all TimerObjects referenced by this TimerList */
    
//  Enumeration keys = this.keys();
    Enumeration elements = this.elements();
//  String key;
    TimerObject to;

    if (!elements.hasMoreElements()) 
    {
        /* list is empty do not update TimerObjects (there are none) */
        return;
    }

    /* Go through list and set reqHandler on each Timer in list */
    
    do 
    {
//      key = (String) keys.nextElement();
//      to = (TimerObject) this.get(key);
        to = (TimerObject) elements.nextElement();
        to.setReqHandler(reqHandler);
//      this.put(key, to);
    }
    while (elements.hasMoreElements());
        
}
}
