package com.ibm.staf.service.timer;
import com.ibm.staf.service.STAFServiceInterfaceLevel4;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/****************************************************************************/
//
// Class: TimerRequestHandler
//
// Logical Name: TimerRequestHandler.java
//
// Description: This class performs the requested actions of STAF commands
//              after they have been parsed.
//
//
// History:
//
// DATE       DEVELOPER   CHANGE DESCRIPTION
// ---------- ----------- -----------------------------------
// 02/01/2000 C Alkov     Original Program
// 04/05/2000 C Alkov     Update version to 1.01 (def:3610,3622)
//
/****************************************************************************/

import com.ibm.staf.*;
import com.ibm.staf.wrapper.STAFLog;

public class TimerRequestHandler 
{    
    private int minFrequency;
    boolean unregisterOnNoHandle;
    boolean unregisterOnNoPath;
    int failCountLimit;
    
    String sName;
    STAFHandle sHandle;

    String LINESEP;
    private final static String UNEXPECTEDVARRESULT = 
        "Unexpected Result from call to VAR service";
    private final static String VERSION = "1.1.1";

    TimerList timerList;
    WatchList watchList;
    Timer timer;
    TimerManager tManager;
    WatchManager wManager;

/****************************************************************************/
//
// Method: 
//   Constructor
//
// Description:
//   Constructor method for the TimerRequestHandler class.
//
// Input:
//   name - The STAF name of this service.
//   handle - The STAFHandle of this service.
//   aTimer - reference to the main Timer object
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public TimerRequestHandler(String name, STAFHandle handle, Timer aTimer) 
{
    sName = name;
    sHandle = handle;
    timer = aTimer;

    try
    {
        LINESEP = sHandle.submit("LOCAL", "VAR", 
            "GLOBAL RESOLVE {STAF/Config/Sep/Line}");
    }
    catch (STAFException e) 
    {
        LINESEP = "\n";
    }

    /* Read in timer and watch list, except if we are not using persistent storage */
    if(!(timer.tListFileName == null)) 
    { //if tListFileName is null, so is wListFileName
        timerList = TimerUtils.readTimerList(this, timer.tListFileName);
        watchList = TimerUtils.readWatchList(this, timer.wListFileName);
    }
    

    /* No timer list create new list */
    
    if (timerList == null) 
    {
        timerList = new TimerList(this);
        watchList = new WatchList(this);
    }

    /* Start TimerManager thread */
    
    tManager = new TimerManager(timerList);
    tManager.start();

    /* Start WatchMangager thread */

    wManager = new WatchManager(watchList);
    wManager.start();
    
}

/****************************************************************************/
//
// Method: 
//   help
//
// Description:
//   Handler method for help request.
//
// Input:
//   machine - name of machine making the request
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult help(String serviceName) 
{
    String HELP =
      serviceName + " Service Help"+ LINESEP+LINESEP+
      "REGISTER   TYPE <Name> FREQUENCY <Milliseconds> [PRIORITY <Level>]"+
      LINESEP+
      "           [UNREGONNOPATH <1 | 0>] [UNREGONNOHANDLE <1 | 0>]"+LINESEP+
      "           [BYNAME]"+LINESEP+
      "UNREGISTER TYPE <Name> [[MACHINE <Machine> HANDLE <Handle> | "+LINESEP+
      "           PROCESS <Process>] | [BYNAME]]"+LINESEP+
      "WATCH      MACHINE <Machine> FREQUENCY <Frequency> [MARGIN <Margin>]"+
      LINESEP+
      "UNWATCH    MACHINE <Machine>"+LINESEP+
      "LIST       TIMERS | WATCHES [FORMAT]"+LINESEP+
      "REFRESH"+LINESEP+
      "VERSION"+LINESEP+
      "HELP";

    STAFResult result = new STAFResult(STAFResult.Ok, HELP);
    
    return result;
}

/****************************************************************************/
//
// Method: 
//   listTimers
//
// Description:
//   Handler method for list timers request.
//
// Input:
//   STAFServiceInterfaceLevel4
//   format - boolean specifying whether to format output
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult listTimers(STAFServiceInterfaceLevel4.RequestInfo info,
                             boolean format) 
{
    if (info.diagEnabled == 1)
    {
        // Record diagnostics because the output for a LIST request will
        // change in STAF V3.0

        String trigger = new String("STAF/V3.0-Mig/TIMER LIST");
        String source  = new String(info.processName + ";" +
                                    info.machine + ";" +
                                    info.handle);
        sHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "DIAG",
            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
            " SOURCE " + STAFUtil.wrapData(source));
    }

    String list;
    
    if (format) 
    {
        list = timerList.listFormattedTimers();
    }
    else 
    {
        list = timerList.listTimers();
    }

    return new STAFResult(STAFResult.Ok, list);
    
}

/****************************************************************************/
//
// Method: 
//   listWatches
//
// Description:
//   Handler method for list timers request.
//
// Input:
//   STAFServiceInterfaceLevel4
//   format - boolean specifying whether to format output
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult listWatches(STAFServiceInterfaceLevel4.RequestInfo info,
                              boolean format) 
{
    if (info.diagEnabled == 1)
    {
        // Record diagnostics because the output for a LIST request will
        // change in STAF V3.0

        String trigger = new String("STAF/V3.0-Mig/TIMER LIST");
        String source  = new String(info.processName + ";" +
                                    info.machine + ";" +
                                    info.handle);
        sHandle.submit2(
            STAFHandle.ReqFireAndForget,
            "local", "DIAG",
            "RECORD TRIGGER " + STAFUtil.wrapData(trigger) +
            " SOURCE " + STAFUtil.wrapData(source));
    }

    String list;
    
    if (format) 
    {
        list = watchList.listFormattedWatches();
    }
    else 
    {
        list = watchList.listWatches();
    }

    return new STAFResult(STAFResult.Ok, list);

}

/****************************************************************************/
//
// Method: 
//   refresh
//
// Description:
//   Handler method for refresh request.
//
// Input:
//   STAFServiceInterfaceLevel4.RequestInfo
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult refresh() 
{
    String where = "LOCAL";
    String service = "VAR";
    String request;
    STAFResult getVarResult;
    STAFResult result = new STAFResult(STAFResult.Ok);
    
    /* Get MinFrequency variable */
    
    request = "RESOLVE {STAF/Service/Timer/MinFrequency}";
    
    getVarResult = sHandle.submit2(where, service, request);

    if (getVarResult.rc == STAFResult.Ok) 
    {
        try 
        {
            minFrequency = Integer.parseInt(getVarResult.result);
            
            if (minFrequency < 0)
            {
                /* Set to default */
                minFrequency = 1000;
                timer.log.log(STAFLog.Warning, "Invalid value specified "+
                    "for MinFrequency, using default value (1000)");
            }
        }
        catch (NumberFormatException e) 
        {
            /* Set to default */
            minFrequency = 1000;
            timer.log.log(STAFLog.Warning, "Invalid value specified "+
                    "for MinFrequency, using default value (1000)");
        }
    }
    else if (getVarResult.rc == STAFResult.VariableDoesNotExist) 
    {
        /* Set to default */
        minFrequency = 1000;
    }
    else 
    {
        /* Unexpected Result shouldn't ever go here */
        result = new STAFResult(getVarResult.rc, UNEXPECTEDVARRESULT);
        timer.log.log(STAFLog.Warning, "Unexpected error attempting to "+
                    "resolve MinFrequency, using default value (1000)");
        minFrequency = 1000;
    } 

    /* Get UnregisterOnNoHandle variable */

    request = "RESOLVE {STAF/Service/Timer/UnregisterOnNoHandle}";
    
    getVarResult = sHandle.submit2(where, service, request);

    if (getVarResult.rc == STAFResult.Ok) 
    {
        try 
        {
            int iunregisterOnNoHandle = Integer.parseInt(getVarResult.result);
            
            if (iunregisterOnNoHandle == 0) 
            {
                unregisterOnNoHandle = false;
            }
            else 
            {
                unregisterOnNoHandle = true;
            }
        }
        catch (NumberFormatException e) 
        {
            /* Set to default */
            unregisterOnNoHandle = true;
            timer.log.log(STAFLog.Warning, "Invalid value specified "+
                "for UnregisterOnNoHandle, using default value (true)");
        }
    }
    else if (getVarResult.rc == STAFResult.VariableDoesNotExist) 
    {
        /* Set to default */
        unregisterOnNoHandle = true;
    }
    else 
    {
        /* Unexpected Result shouldn't ever go here */
        result = new STAFResult(getVarResult.rc, UNEXPECTEDVARRESULT);
        timer.log.log(STAFLog.Warning, "Unexpected error attempting to "+
            "resolve UnregisterOnNoHandle, using default value (true)");
        unregisterOnNoHandle = true;
    }

    /* Get UnregisterOnNoPath variable */

    request = "RESOLVE {STAF/Service/Timer/UnregisterOnNoPath}";
    
    getVarResult = sHandle.submit2(where, service, request);

    if (getVarResult.rc == STAFResult.Ok) 
    {
        try 
        {
            int iunregisterOnNoPath = Integer.parseInt(getVarResult.result);
            
            if (iunregisterOnNoPath == 0) 
            {
                unregisterOnNoPath = false;
            }
            else 
            {
                unregisterOnNoPath = true;
            }
        }
        catch (NumberFormatException e) 
        {
            /* Set to default */
            unregisterOnNoPath = true;
            timer.log.log(STAFLog.Warning, "Invalid value specified "+
                "for UnregisterOnNoPath, using default value (true)");
        }
    }
    else if (getVarResult.rc == STAFResult.VariableDoesNotExist) 
    {
        /* Set to default */
        unregisterOnNoPath = true;
    }
    else 
    { 
        /* Unexpected Result shouldn't ever go here */
        result = new STAFResult(getVarResult.rc, UNEXPECTEDVARRESULT);
        timer.log.log(STAFLog.Warning, "Unexpected error attempting to "+
            "resolve UnregisterOnNoPath, using default value (true)");
        unregisterOnNoPath = true;
    }

    /* Get TotalFireAttempts variable */

    request = "RESOLVE {STAF/Service/Timer/TotalFireAttempts}";
    
    getVarResult = sHandle.submit2(where, service, request);

    if (getVarResult.rc == STAFResult.Ok) 
    {
        try 
        {
            failCountLimit = Integer.parseInt(getVarResult.result);
            
            if (failCountLimit < 0) 
            {
                timer.log.log(STAFLog.Warning, "Negative value specified for "+
                    "TotalFireAttempts, using default value (3)"); 
                failCountLimit = 3;
            }
        }
        catch (NumberFormatException e) 
        {
            /* Set to default */
            failCountLimit = 3;
            timer.log.log(STAFLog.Warning, "Invalid value specified "+
                "for TotalFireAttempts, using default value (3)");
        }
    }
    else if (getVarResult.rc == STAFResult.VariableDoesNotExist) 
    {
        /* Set to default */
        failCountLimit = 3;
    }
    else 
    { 
        /* Unexpected Result shouldn't ever go here */
        result = new STAFResult(getVarResult.rc, UNEXPECTEDVARRESULT);
        timer.log.log(STAFLog.Warning, "Unexpected error attempting to "+
            "resolve TotalFireAttempts, using default value (3)");
        failCountLimit = 3;
    }
    
    return result;
}

/****************************************************************************/
//
// Method: 
//   register
//
// Description:
//   Handler method for register request.
//
// Input:
//   name - The String Name (Type) of this timer.
//   machine - The machine name this timer should notify.
//   handle - The Handle of the process on the machine this
//             timer should notify.
//   process - The name of the process this timer should notify.
//   frequency - The frequency of this timer.
//   priority - The priority of the Queue message this timer should
//               send to the registered machine.
//   fUnregOnNoPath - specifies a timer specific variable for unregisterOnNoPath
//                    a value of -1 means use the global variable (not timer specific)
//   fUnregOnNoHandle - specifies a timer specific variable for unregisterOnNoHandle
//                      a value of -1 means use the global variable (not timer specific)
//   byname - Boolean that specifies if the BYNAME parameter was specified.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult register(String name, String machine, int handle, 
                           String process, long frequency, long priority, 
                           int unregOnNoPath, int unregOnNoHandle, 
                           boolean byname) 
{
    /* Check for good frequency */

    if (frequency < minFrequency) 
    {
        return new STAFResult(TimerUtils.INVALIDFREQUENCYRC, 
            TimerUtils.INVALIDFREQUENCY);
    }

    /* Check for good priority */

    if (priority < 0 || priority > (Math.pow(2, 32)-1)) 
    {
        return new STAFResult(TimerUtils.INVALIDPRIORITYRC, 
            TimerUtils.INVALIDPRIORITY);
    }

    /* Choose which createTimerString method to call (byname) */

    String timerString;

    if (byname) 
    {
        timerString = TimerUtils.createTimerString(name, machine, process);
    }
    else 
    {
        timerString = TimerUtils.createTimerString(name, machine, handle);
    }

    /* See if timer already exists */
    
    if (timerList.containsTimer(timerString)) 
    {
        /* Timer exists, update timer */
        TimerObject timer = timerList.getTimerObject(timerString);
        timer.update(frequency, priority, unregOnNoPath, unregOnNoHandle);

    }
    else 
    {        
        /* Create new timer */
        TimerObject timer = new TimerObject (name, machine, handle, process, 
            frequency, priority, this, unregOnNoPath, unregOnNoHandle, byname);

    }

    return new STAFResult(STAFResult.Ok);
    
}

/****************************************************************************/
//
// Method: 
//   term
//
// Description:
//   Called when this service is stopped.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   Stops TimerManger thread, writes out TimerList, and unregisters the
//   STAFHandle.
//
/****************************************************************************/

public int term() 
{
    tManager.term();
    wManager.term();

    // Don't try to write if did not create a filename, no persistent storage
    if (!(timer.tListFileName == null)) 
    { // if tListFileName is null so is wListFileName
        TimerUtils.writeTimerList(timerList, timer.tListFileName);
        TimerUtils.writeWatchList(watchList, timer.wListFileName);
    }
        
    try 
    {
        sHandle.unRegister();
        timer.wHandle.unRegister();
    }
    catch (STAFException e) 
    {
        timer.log.log(STAFLog.Warning,
                      "Error during unregister; RC = " + e.rc);
        return e.rc;
    }

    return STAFResult.Ok;
    
}

/****************************************************************************/
//
// Method: 
//   unregister
//
// Description:
//   Handler method for unregister request.
//
// Input:
//   timerString - TimerString representing timer to unregister
//
// Exceptions Thrown:
//   none
//
// Notes:
//   The trust level checking is done in TimerCmdParser.parseUnregister() because 
//   we must know the machine the timer is running on and the machine submitting
//   the request.
//
/****************************************************************************/

public STAFResult unregister(String timerString) 
{
    /* Unregister timer */

    if(timerList.containsTimer(timerString)) 
    {
        timerList.remove(timerString);
        return new STAFResult(STAFResult.Ok);
    }
    else 
    {
        return new STAFResult(TimerUtils.INVALIDTYPERC, 
                             TimerUtils.INVALIDTYPE);
    }
}

/****************************************************************************/
//
// Method: 
//   unwatch
//
// Description:
//   Handler method for unregister request.
//
// Input:
//   machineToWatch - name of the machine to be unwatched
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult unwatch(String machineToWatch) 
{
    /* Unregister watch */
    
    if(watchList.containsWatch(machineToWatch))
    {
        try 
        {
            /* Unregister the timer with machineToWatch */

            String request = 
                "unregister type STAF/Service/" + sName + "/Watch byname";
            timer.wHandle.submit(machineToWatch, "TIMER", request);
        }
        catch (STAFException e) 
        {
            timer.log.log(STAFLog.Warning, 
                "Error unregistering watch. Could not unregister " +
                "corresponding timer with machine: " + machineToWatch +
                ". RC:"+e.rc);
        }
        
        watchList.remove(machineToWatch);
        return new STAFResult(STAFResult.Ok);
    }
    else 
    {
        return new STAFResult(TimerUtils.MACHINENOTWATCHEDRC,
                              TimerUtils.MACHINENOTWATCHED);
    }
}

/****************************************************************************/
//
// Method: 
//   version
//
// Description:
//   Handler method for version request.
//
// Input:
//   STAFServiceInterfaceLevel4.RequestInfo
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult version(STAFServiceInterfaceLevel4.RequestInfo info) 
{
    if (info.trustLevel < 1)
    {
        return new STAFResult(STAFResult.InvalidTrustLevel);
    }

    return new STAFResult(STAFResult.Ok, VERSION);
}

/****************************************************************************/
//
// Method: 
//   watch
//
// Description:
//   Handler method for the watch request.
//
// Input:
//   machineToWatch - the name of the machine to be watched
//   frequency - The frequency of this watch
//   margin - the allowable error in the response time of the machine being watched
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public STAFResult watch(String machineToWatch, long frequency, int margin) 
{    
    /* Make sure we use long name even if short name was passed */          
    try 
    {
        String varRequest = "RESOLVE {STAF/Config/Machine}";
        machineToWatch = sHandle.submit(machineToWatch, "VAR", varRequest);
    }
    catch (STAFException e) 
    {
        /* Cannot resolve: log error and continue with original parameter */
        timer.log.log(STAFLog.Error, 
            "Unexpected error attempting to resolve machine name for " + 
            "machine:"+machineToWatch+" RC="+e.rc);
        return new STAFResult(e.rc);
    }

    /* use uppercase for hashtable key */

    machineToWatch = machineToWatch.toUpperCase();

    /* See if watch already exists */
    
    if(watchList.containsWatch(machineToWatch)) 
    {
        /* Watch exists, update watch */
        WatchObject watch = watchList.getWatchObject(machineToWatch);
        watch.update(frequency, margin);

    }
    else 
    {
        try 
        {
            /* Create new watch */
            new WatchObject (machineToWatch, frequency, margin, this);
        }
        catch (STAFException e) 
        {
            timer.log.log(STAFLog.Error, 
                "Error registering a timer with machine: "+machineToWatch+
                " RC:"+e.rc);
            return new STAFResult(e.rc);
        }

    }
    
    return new STAFResult(STAFResult.Ok);
}
}
