package com.ibm.staf.service.timer;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/****************************************************************************/
//
// Class: WatchManager
//
// Logical Name: WatchManager.java
//
// Description: This class waits to receive notifications from the registered
//              watches and then records the notification.
//
//
// History:
//
// DATE       DEVELOPER   CHANGE DESCRIPTION
// ---------- ----------- -----------------------------------
// 02/01/2000 C Alkov     Original Program
//
/****************************************************************************/

import com.ibm.staf.STAFException;
import com.ibm.staf.wrapper.STAFLog;
import java.util.Date;
 
public class WatchManager extends Thread 
{
    private WatchList watchList;
    private volatile Thread wmThread;
    
/****************************************************************************/
//
// Method: 
//   Constructor
//
// Description:
//   Constructor method for WatchManager class.
//
// Input:
//   list - The WatchList instance this WatchManager should manage.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public WatchManager(WatchList list) 
{    super();

    watchList = list;
    
}

/****************************************************************************/
//
// Method: 
//   run
//
// Description:
//   The run method the WatchManager class.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   This object runs in its own thread.
//
/****************************************************************************/

public void run() 
{    
    Thread thisThread = Thread.currentThread();

    while (wmThread == thisThread) 
    {
        String result;
        String where = "LOCAL";
        String service = "QUEUE";
        String request = "GET WAIT";

        try 
        {
            result = watchList.reqHandler.timer.wHandle.submit(where, 
                service, request);
        }
        catch (STAFException e) 
        {
            watchList.reqHandler.timer.log.log(STAFLog.Error, 
                "Error retrieving watch message from "+
                "queue. RC="+e.rc);
            continue;
        }

        /* Extract the machine name and timestamp from the queued message */

        int beginTSIndex = result.indexOf(";") + 1; 
        // gets the index of beginning of timestamp
        int endTSIndex = result.indexOf(";", beginTSIndex); 
        // gets the index of end of timestamp + 1
        int beginMNameIndex = endTSIndex + 1; 
        // gets index of beginning of machine name
        int endMNameIndex = result.indexOf(";", beginMNameIndex); 
        // gets index of end of machine name + 1

        /* Get timestamp of queued message */
        /* We are going to use the timestamp from the queue service, rather than from the
           timer service on the requesting machine to avoid differences in the system clocks
           between machines */

        String timeStamp = result.substring(beginTSIndex, endTSIndex);
        Date thisFireTime = TimerUtils.parseDate(timeStamp);

        /* Get machine name from queued message */

        String machine = result.substring(beginMNameIndex, endMNameIndex); 
        // gets machine name from queued message

        /* Make sure we use long name even if short name was passed */
        try 
        {
            String varRequest = "RESOLVE {STAF/Config/Machine}";
            machine = watchList.reqHandler.sHandle.submit(machine, "VAR", 
                varRequest);
        }
        catch (STAFException e) 
        {
            /* Cannot resolve: log error */
            watchList.reqHandler.timer.log.log(STAFLog.Warning, 
                "Unexpected error attempting to resolve machine " + 
                "name from queued message.");
            continue;
        }

        /* Get watch object that this queued message is for */
        
        WatchObject wo = watchList.getWatchObject(machine.toUpperCase());
        
        try 
        {
            wo.setLastTimeStamp(thisFireTime);
        }
        catch (NullPointerException e) 
        {
            /* no corresponding watch object for this queued message, ignore message */
            continue;
        }
        
    }

}

/****************************************************************************/
//
// Method: 
//   start
//
// Description:
//   This method starts the WatchManager Thread and gives us a reference to
//   the Thread object.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   This object runs in its own thread.
//
/****************************************************************************/

public void start() 
{    
    if (wmThread == null) 
    {
        wmThread = new Thread(this);
        wmThread.start();
    }

}

/****************************************************************************/
//
// Method: 
//   term
//
// Description:
//   This method causes the WatchManager Thread to stop running.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   This object runs in its own thread.
//
/****************************************************************************/

public void term() 
{
    /* Send bogus queue message to watch manager to wake it from the QUEUE GET WAIT
       command. This will cause the thread to die quickly and stafproc will then be
       able to quickly shutdown */
       
    String request = "QUEUE NAME "+watchList.reqHandler.sName+
        "/watch MESSAGE QUIT";
    watchList.reqHandler.sHandle.submit2("LOCAL", "QUEUE", request);
    
    wmThread = null;
    
}
}
