package com.ibm.staf.service.timer;

/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2002                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

/****************************************************************************/
//
// Class: WatchObject
//
// Logical Name: WatchObject.java
//
// Description: This class contains information about each machine registered
//              to be watched.
//
//
// History:
//
// DATE       DEVELOPER   CHANGE DESCRIPTION
// ---------- ----------- -----------------------------------
// 02/01/2000 C Alkov     Original Program
// 02/12/2000 C Alkov     Register timers for corresponding watches using 
//                        unregonnopath 0 & unregonnohandle 0 (def:3622)
//
/****************************************************************************/

import java.io.Serializable;
import java.util.Date;
import com.ibm.staf.STAFException;
import com.ibm.staf.STAFResult;
import com.ibm.staf.wrapper.STAFLog;

public class WatchObject implements Serializable 
{
    private String machineToWatch;
    private long frequency;
    private int margin;

    /* Do not serialize (transient) this instance variable since
       reference will be different from run to run */
    private transient TimerRequestHandler reqHandler;
    
    private Date lastTimeStamp = null;
    
/****************************************************************************/
//
// Method: 
//   Constructor
//
// Description:
//   Constructor method for TimerObject class.
//
// Input:
//   aMachineToWatch - the name of the machine to be watched
//   aFrequency - the frequency at which the machine should send notifications
//                to the WatchManager
//   aMargin - the allowable error in the response time of the machine being
//             watched
//   aReqHandler - reference to the TimerRequestHandler object
//
// Exceptions Thrown:
//   STAFException
//
// Notes:
//   none
//
/****************************************************************************/

public WatchObject(String aMachineToWatch, long aFrequency, int aMargin, 
                   TimerRequestHandler aReqHandler) throws STAFException 
{
    super();

    machineToWatch = aMachineToWatch.toUpperCase();
    frequency = aFrequency;
    margin = aMargin;
    reqHandler = aReqHandler;

    /* Register a timer with machineToWatch */

    String request = "register type STAF/Service/"+reqHandler.sName+
        "/Watch frequency "+frequency+
        " unregonnopath 0 unregonnohandle 0 byname";

    reqHandler.timer.wHandle.submit(machineToWatch, "TIMER", request);

    /* Set lastTimeStamp to current time */

    setLastTimeStamp(new Date());

    /* Add this timer to watchList */

    reqHandler.watchList.put(machineToWatch, this);

}

/****************************************************************************/
//
// Method: 
//   getFormattedTimers
//
// Description:
//   Converts the Watch Object to String format for use
//   in the LIST WATCHES FORMAT request.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   synchronized method
//
/****************************************************************************/

public synchronized String getFormattedProperties() 
{
    String properties;
    String lastTime = TimerUtils.formatTime(lastTimeStamp);
    
    String machineStatus;
    
    if (lastNotificationReceived()) 
    {
        machineStatus = "OK";
    }
    else 
    {
        machineStatus = "Missed";
    }

    properties = "Machine To Watch   :"+machineToWatch+reqHandler.LINESEP+
                 "Machine Status     :"+machineStatus+reqHandler.LINESEP+
                 "Last Timestamp     :"+lastTime+reqHandler.LINESEP+
                 "Frequency          :"+frequency+reqHandler.LINESEP+
                 "Margin             :"+margin+reqHandler.LINESEP;

    return properties;
    
}

/****************************************************************************/
//
// Method: 
//   getProperties
//
// Description:
//   Returns the properties of this watch in a semi-colon delimited String
//   for use in the LIST WATCHES request.
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   synchronized method
//
/****************************************************************************/

public synchronized String getProperties() 
{
    String properties;
    String machineStatus;
    String lastTime = TimerUtils.formatTime(lastTimeStamp);

    if (lastNotificationReceived()) 
    {
        machineStatus = "OK";
    }
    else 
    {
        machineStatus = "Missed";
    }

    properties = machineToWatch+";"+machineStatus+";"+
        lastTime+";"+frequency+";"+margin+";";
     
    return properties;
    
}

/****************************************************************************/
//
// Method: 
//   lastNotificationReceived
//
// Description:
//   Private method for checking if the last notification was receieved
//   from the machine being watched
//
// Input:
//   none
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

private boolean lastNotificationReceived() 
{
    Date OKTime = new Date(lastTimeStamp.getTime() + frequency + margin);
    Date currentTime = new Date();

    if(currentTime.after(OKTime)) 
    {
        return false;
    }
    else 
    {
        return true;
    }
    
}

/****************************************************************************/
//
// Method: 
//   setLastTimeStamp
//
// Description:
//   Sets the value of lastTimeStamp (the last time a notification was
//   received from the machine being watched)
//
// Input:
//   newLastTimeStamp - Date representing the last time the machine being
//                      watched sent a notification message
//
// Exceptions Thrown:
//   none
//
// Notes:
//   synchronized method
//
/****************************************************************************/

public synchronized void setLastTimeStamp(Date newLastTimeStamp) 
{
    lastTimeStamp = newLastTimeStamp;
}

/****************************************************************************/
//
// Method: 
//   setReqHandler
//
// Description:
//   Sets the RequestHandler for the WatchObject.
//
// Input:
//   aReqHandler - Reference to the current RequestHandler
//
// Exceptions Thrown:
//   none
//
// Notes:
//   none
//
/****************************************************************************/

public void setReqHandler(TimerRequestHandler aReqHandler) 
{    
    reqHandler = aReqHandler;
    
}

/****************************************************************************/
//
// Method: 
//   update
//
// Description:
//   Updates the frequency and margin of the Watch.
//
// Input:
//   newFrequency - The new frequency of the watch.
//   newMargin - The new margin of the watch.
//
// Exceptions Thrown:
//   none
//
// Notes:
//   synchronized method
//
/****************************************************************************/

public synchronized STAFResult update(long newFrequency, int newMargin) 
{
    frequency = newFrequency;
    margin = newMargin;

    /* Update timer information with machineToWatch */

    String request = "register type STAF/Service/"+reqHandler.sName+
        "/Watch frequency "+frequency+" byanme";

    try 
    {
        reqHandler.sHandle.submit(machineToWatch, "TIMER", request);
    }
    catch(STAFException e) 
    {
        reqHandler.timer.log.log(STAFLog.Error, 
            "Error registering a timer with machine: "+machineToWatch+
            " RC:"+e.rc);
        return new STAFResult(e.rc);
    }
    
    return new STAFResult(STAFResult.Ok);
}
}
