/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2004                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFString.h"
#include "STAFTrace.h"
#include "STAFMutexSem.h"
#include "STAFCommandParser.h"
#include "STAFServiceInterface.h"

#include <vector>
#include <map>

#include "zlib.h"

#include "STAFZip.h"
#include "STAFZipUtil.h"
#include "STAFZipFileHeader.h"
#include "STAFZipFileAttribute.h"
#include "STAFZipCentralDirExtension.h"
#include "STAFZipLocalFileHeader.h"
#include "STAFZipCentralDirEndRecord.h"
#include "STAFZipCentralDir.h"
#include "STAFZipFile.h"
#include "STAFZipMutexLock.h"



// STAFZip Service Data
struct STAFZipServiceData
{
    unsigned int  fDebugMode;              // Debug Mode flag
    STAFString    fShortName;              // Short service name
    STAFString    fName;                   // Registered service name
    STAFHandlePtr fHandlePtr;              // STAFZip service's STAF handle
    STAFCommandParserPtr fZipParser;       // STAFZip ZIP command parser
    STAFCommandParserPtr fUnzipParser;     // STAFZip UNZIP command parser
    STAFCommandParserPtr fListParser;      // STAFZip LIST command parser
    STAFCommandParserPtr fDeleteParser;    // STAFZip DELETE command parser
    STAFCommandParserPtr fHelpParser;      // STAFZip HELP command parser
    STAFCommandParserPtr fVersionParser;   // STAFZip VERSION command parser
};

// static STAFZipMutexLock class
static STAFZipMutexLock zipLock;


// Static Variables

static STAFString sLineSep;
static const STAFString sVersionInfo("1.0.6");
// 1.0.0    Original release
// 1.0.1    Bug 1012202: Can't read permission info in the latest InfoZip archive    
// 1.0.2    Bug 1033654: Error inflate file    
// 1.0.3    Feature 1055682: Improve unzip's performance on large files
// 1.0.4    Bug 1076948: zero bytes when unzipping JAR archives

static const STAFString sLocal("local");
static const STAFString sHelp("help");
static const STAFString sVar("var");
static const STAFString sResStrHandle("HANDLE ");
static const STAFString sResStrResolve(" RESOLVE ");
static const STAFString sResStrGResolve("GLOBAL RESOLVE ");
static const STAFString sLeftCurlyBrace(kUTF8_LCURLY);

// Prototypes

static STAFResultPtr handleZip(STAFServiceRequestLevel2 *,
                               STAFZipServiceData *, STAFString *);
static STAFResultPtr handleUnzip(STAFServiceRequestLevel2 *,
                                  STAFZipServiceData *, STAFString *);
static STAFResultPtr handleList(STAFServiceRequestLevel2 *,
                                  STAFZipServiceData *, STAFString *);
static STAFResultPtr handleDelete(STAFServiceRequestLevel2 *,
                                  STAFZipServiceData *, STAFString *);
static STAFResultPtr handleHelp(STAFServiceRequestLevel2 *,
                                STAFZipServiceData *);
static STAFResultPtr handleVersion(STAFServiceRequestLevel2 *,
                                   STAFZipServiceData *);

static STAFResultPtr resolveStr(STAFServiceRequestLevel2 *pInfo,
                                STAFZipServiceData *pData,
                                const STAFString &theString);

static STAFResultPtr resolveOp(STAFServiceRequestLevel2 *pInfo,
                               STAFZipServiceData *pData,
                               STAFCommandParseResultPtr &parsedResult,
                               const STAFString &fOption,
                               unsigned int optionIndex = 1);

STAFResultPtr resolveOpLocal(STAFZipServiceData *pData,
                             STAFCommandParseResultPtr &parsedResult,
                             const STAFString &fOption,
                             unsigned int optionIndex = 1);

static void registerHelpData(STAFZipServiceData *pData,
                             unsigned int errorNumber,
                             const STAFString &shortInfo,
                             const STAFString &longInfo);

static void unregisterHelpData(STAFZipServiceData *pData,
                               unsigned int errorNumber);


// Begin implementation

STAFRC_t STAFServiceGetLevelBounds(unsigned int levelID,
                                   unsigned int *minimum,
                                   unsigned int *maximum)
{
    switch (levelID)
    {
        case kServiceInfo:
        case kServiceInit:
        {
            *minimum = 1;
            *maximum = 1;
            break;
        }        
        case kServiceAcceptRequest:
        {
            *minimum = 2;
            *maximum = 2;
            break;
        }
        case kServiceTerm:
        case kServiceDestruct:
        {
            *minimum = 0;
            *maximum = 0;
            break;
        }
        default:
        {
            return kSTAFInvalidAPILevel;
        }
    }

    return kSTAFOk;
}


STAFRC_t STAFServiceConstruct(STAFServiceHandle_t *pServiceHandle,
                              void *pServiceInfo, unsigned int infoLevel,
                              STAFString_t *pErrorBuffer)
{
    STAFRC_t rc = kSTAFUnknownError;

    try
    {
        if (infoLevel != 1) return kSTAFInvalidAPILevel;

        STAFServiceInfoLevel1 *pInfo =
            reinterpret_cast<STAFServiceInfoLevel1 *>(pServiceInfo);

        STAFZipServiceData data;
        data.fDebugMode = 0;
        data.fShortName = pInfo->name;
        data.fName = "STAF/Service/";
        data.fName += pInfo->name;

        for (unsigned int i = 0; i < pInfo->numOptions; ++i)
        {
            if (STAFString(pInfo->pOptionName[i]).upperCase() == "DEBUG")
            {
                data.fDebugMode = 1;
            }
            else
            {
                STAFString optionError(pInfo->pOptionName[i]);
                *pErrorBuffer = optionError.adoptImpl();
                return kSTAFServiceConfigurationError;
            }
        }

        // Set service handle

        *pServiceHandle = new STAFZipServiceData(data);

        return kSTAFOk;
    }
    catch (STAFException &e)
    {
        STAFString result;

        result += STAFString("In STAFZip.cpp: STAFServiceConstruct")
            + kUTF8_SCOLON;

        result += STAFString("Name: ") + e.getName() + kUTF8_SCOLON;
        result += STAFString("Location: ") + e.getLocation() + kUTF8_SCOLON;
        result += STAFString("Text: ") + e.getText() + kUTF8_SCOLON;
        result += STAFString("Error code: ") + e.getErrorCode() + kUTF8_SCOLON;

        *pErrorBuffer = result.adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFZip.cpp: STAFServiceConstruct: Caught "
                    "unknown exception in STAFServiceConstruct()");
        *pErrorBuffer = error.adoptImpl();
    }

    return kSTAFUnknownError;
}


STAFRC_t STAFServiceInit(STAFServiceHandle_t serviceHandle,
                         void *pInitInfo, unsigned int initLevel,
                         STAFString_t *pErrorBuffer)
{
    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        if (initLevel != 1) return kSTAFInvalidAPILevel;

        STAFZipServiceData *pData =
            reinterpret_cast<STAFZipServiceData *>(serviceHandle);

        STAFServiceInitLevel1 *pInfo =
            reinterpret_cast<STAFServiceInitLevel1 *>(pInitInfo);

        retCode = STAFHandle::create(pData->fName, pData->fHandlePtr);

        if (retCode != kSTAFOk)
            return retCode;


        //ZIP options
        pData->fZipParser = STAFCommandParserPtr(new STAFCommandParser,
                                                 STAFCommandParserPtr::INIT);
        pData->fZipParser->addOption("ZIP", 1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fZipParser->addOption("ZIPFILE", 1,
                                     STAFCommandParser::kValueRequired);
        pData->fZipParser->addOption("ADD", 1,
                                     STAFCommandParser::kValueNotAllowed);
        pData->fZipParser->addOption("FILE", 1,
                                     STAFCommandParser::kValueRequired);
        pData->fZipParser->addOption("DIRECTORY", 1,
                                     STAFCommandParser::kValueRequired);
        pData->fZipParser->addOption("RELATIVETO", 1,
                                     STAFCommandParser::kValueRequired);
        pData->fZipParser->addOption("RECURSE", 1,
                                        STAFCommandParser::kValueNotAllowed);

        pData->fZipParser->addOptionNeed("ZIP", "ZIPFILE");
        pData->fZipParser->addOptionNeed("ZIP", "ADD");
        pData->fZipParser->addOptionNeed("RELATIVETO", "ZIP");
        pData->fZipParser->addOptionNeed("RELATIVETO", "ZIPFILE");
        pData->fZipParser->addOptionNeed("RELATIVETO", "ADD");
        pData->fZipParser->addOptionNeed("RECURSE", "ADD");
        pData->fZipParser->addOptionNeed("RECURSE", "ZIP");
        pData->fZipParser->addOptionNeed("RECURSE", "ZIPFILE");
        pData->fZipParser->addOptionNeed("RECURSE", "DIRECTORY");
        pData->fZipParser->addOptionNeed("ADD", "FILE DIRECTORY");
        pData->fZipParser->addOptionNeed("FILE DIRECTORY", "ADD");
        pData->fZipParser->addOptionGroup("FILE DIRECTORY", 0, 1);

        //UNZIP options
        pData->fUnzipParser = STAFCommandParserPtr(new STAFCommandParser,
                                                   STAFCommandParserPtr::INIT);
        pData->fUnzipParser->addOption("UNZIP", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fUnzipParser->addOption("ZIPFILE", 1,
                                        STAFCommandParser::kValueRequired);
        pData->fUnzipParser->addOption("FILE", 0,
                                        STAFCommandParser::kValueRequired);
        pData->fUnzipParser->addOption("TODIRECTORY",  1,
                                        STAFCommandParser::kValueRequired);
        pData->fUnzipParser->addOption("REPLACE", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fUnzipParser->addOption("RESTOREPERMISSION", 1,
                                        STAFCommandParser::kValueNotAllowed);

        pData->fUnzipParser->addOptionNeed("UNZIP", "ZIPFILE");
        pData->fUnzipParser->addOptionNeed("UNZIP", "TODIRECTORY");
        pData->fUnzipParser->addOptionNeed("FILE", "UNZIP");
        pData->fUnzipParser->addOptionNeed("FILE", "ZIPFILE");
        pData->fUnzipParser->addOptionNeed("FILE", "TODIRECTORY");

        pData->fUnzipParser->addOptionNeed("REPLACE", "UNZIP");
        pData->fUnzipParser->addOptionNeed("REPLACE", "ZIPFILE");
        pData->fUnzipParser->addOptionNeed("REPLACE", "TODIRECTORY");

        pData->fUnzipParser->addOptionNeed("RESTOREPERMISSION", "UNZIP");
        pData->fUnzipParser->addOptionNeed("RESTOREPERMISSION", "ZIPFILE");
        pData->fUnzipParser->addOptionNeed("RESTOREPERMISSION", "TODIRECTORY");


        //LIST options
        pData->fListParser = STAFCommandParserPtr(new STAFCommandParser,
                                                   STAFCommandParserPtr::INIT);
        pData->fListParser->addOption("LIST", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fListParser->addOption("ZIPFILE", 1,
                                        STAFCommandParser::kValueRequired);
        pData->fListParser->addOptionNeed("LIST", "ZIPFILE");


        //DELETE options
        pData->fDeleteParser = STAFCommandParserPtr(new STAFCommandParser,
                                                   STAFCommandParserPtr::INIT);
        pData->fDeleteParser->addOption("DELETE", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOption("ZIPFILE", 1,
                                        STAFCommandParser::kValueRequired);
        pData->fDeleteParser->addOption("FILE",  0,
                                        STAFCommandParser::kValueRequired);
        pData->fDeleteParser->addOption("CONFIRM", 1,
                                        STAFCommandParser::kValueNotAllowed);
        pData->fDeleteParser->addOptionNeed("DELETE", "ZIPFILE");
        pData->fDeleteParser->addOptionNeed("DELETE", "FILE");
        pData->fDeleteParser->addOptionNeed("DELETE", "CONFIRM");


        //HELP options
        pData->fHelpParser = STAFCommandParserPtr(new STAFCommandParser,
                                                  STAFCommandParserPtr::INIT);
        pData->fHelpParser->addOption("HELP", 1,
                                      STAFCommandParser::kValueNotAllowed);

        //VERSION options
        pData->fVersionParser = STAFCommandParserPtr(new STAFCommandParser,
                                                  STAFCommandParserPtr::INIT);
        pData->fVersionParser->addOption("VERSION", 1,
                                         STAFCommandParser::kValueNotAllowed);

        // Get line separator
        STAFString getRequest = "RESOLVE {STAF/Config/Sep/Line}";
        STAFResultPtr result = pData->fHandlePtr->submit("LOCAL", "VAR",
                                                         getRequest);
        if (result->rc != 0)
        {
            *pErrorBuffer = result->result.adoptImpl();
            return result->rc;
        }
        else sLineSep = result->result;


        registerHelpData(pData, kZIPGeneralZipError,
            STAFString("General Zip Error"),
            STAFString("ZIP service returns a general error "));

        registerHelpData(pData, kZIPNotEnoughMemory,
            STAFString("Not Enought Memory"),
            STAFString("There is not enough memory in system "));

        registerHelpData(pData, kZIPChangeFileSizeError,
            STAFString("Change File Size Error"),
            STAFString("Change file size is unsuccessful "));

        registerHelpData(pData, kZIPErrorCreatingDir,
            STAFString("Error Creating Dir"),
            STAFString("Creating dir is unsuccessful "));

        registerHelpData(pData, kZIPInvalidZipFile,
            STAFString("Invalid Zip File"),
            STAFString("The Zip archive has an invalid format "));

        registerHelpData(pData, kZIPBadCRC,
            STAFString("Bad CRC"),
            STAFString("Bad CRC in the Zip archive "));

        registerHelpData(pData, kZIPInvalidOwnerGroup,
            STAFString("Invalid Owner Group"),
            STAFString("Owner or Group does not exist in system "));

        registerHelpData(pData, kZIPInvalidFileMode,
            STAFString("Invalid File Mode"),
            STAFString("Cannot set the file's mode (permission) in system "));

    }
    catch (STAFException &e)
    {
        STAFString result;

        result += STAFString("In STAFZip.cpp: STAFServiceInit")
            + kUTF8_SCOLON;

        result += STAFString("Name: ") + e.getName() + kUTF8_SCOLON;
        result += STAFString("Location: ") + e.getLocation() + kUTF8_SCOLON;
        result += STAFString("Text: ") + e.getText() + kUTF8_SCOLON;
        result += STAFString("Error code: ") + e.getErrorCode() + kUTF8_SCOLON;

        *pErrorBuffer = result.adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFZip.cpp: STAFServiceInit: Caught unknown "
                    "exception in STAFServiceServiceInit()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFRC_t STAFServiceAcceptRequest(STAFServiceHandle_t serviceHandle,
                                  void *pRequestInfo, unsigned int reqLevel,
                                  STAFString_t *pResultBuffer)
{
    if (reqLevel != 2) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    STAFString mutexName("STAFZIPMUTEX:");

    try
    {
        STAFResultPtr result(new STAFResult(),
                             STAFResultPtr::INIT);

        STAFServiceRequestLevel2 *pInfo =
            reinterpret_cast<STAFServiceRequestLevel2 *>(pRequestInfo);

        STAFZipServiceData *pData =
            reinterpret_cast<STAFZipServiceData *>(serviceHandle);

        STAFString request(pInfo->request);
        STAFString action = request.subWord(0, 1).toLowerCase();

        // Call functions for the request

        if (action == "zip")
            result = handleZip(pInfo, pData, &mutexName);
        else if (action == "unzip")
            result = handleUnzip(pInfo, pData, &mutexName);
        else if (action == "list")
            result = handleList(pInfo, pData, &mutexName);
        else if (action == "delete")
            result = handleDelete(pInfo, pData, &mutexName);
        else if (action == "help")
            result = handleHelp(pInfo, pData);
        else if (action == "version")
            result = handleVersion(pInfo, pData);
        else
        {
            result = STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                                   request.subWord(0, 1)),
                                   STAFResultPtr::INIT);
        }

        *pResultBuffer = result->result.adoptImpl();
        retCode = result->rc;
    }
    catch (STAFException &e)
    {
        STAFString result;

        result += STAFString("In STAFZip.cpp: STAFServiceAcceptRequest")
            + kUTF8_SCOLON;

        result += STAFString("Name: ") + e.getName() + kUTF8_SCOLON;
        result += STAFString("Location: ") + e.getLocation() + kUTF8_SCOLON;
        result += STAFString("Text: ") + e.getText() + kUTF8_SCOLON;
        result += STAFString("Error code: ") + e.getErrorCode() + kUTF8_SCOLON;

        *pResultBuffer = result.adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFZip.cpp: STAFServiceAcceptRequest: Caught "
                    "unknown exception in STAFServiceAcceptRequest()");
        *pResultBuffer = error.adoptImpl();
    }

    if (mutexName.length() > STAFString("STAFZIPMUTEX:").length())
    {
        // release the mutex
        zipLock.release(mutexName);
    }        

    return retCode;
}


STAFRC_t STAFServiceTerm(STAFServiceHandle_t serviceHandle,
                         void *pTermInfo, unsigned int termLevel,
                         STAFString_t *pErrorBuffer)
{
    if (termLevel != 0) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        retCode = kSTAFOk;

        STAFZipServiceData *pData =
            reinterpret_cast<STAFZipServiceData *>(serviceHandle);

        // Unregister help data

        unregisterHelpData(pData, kZIPGeneralZipError);
        unregisterHelpData(pData, kZIPNotEnoughMemory);
        unregisterHelpData(pData, kZIPChangeFileSizeError);
        unregisterHelpData(pData, kZIPErrorCreatingDir);
        unregisterHelpData(pData, kZIPInvalidZipFile);
        unregisterHelpData(pData, kZIPBadCRC);
        unregisterHelpData(pData, kZIPInvalidOwnerGroup);
        unregisterHelpData(pData, kZIPInvalidFileMode);

    }
    catch (STAFException &e)
    {
        STAFString result;

        result += STAFString("In STAFZip.cpp: STAFServiceTerm")
            + kUTF8_SCOLON;

        result += STAFString("Name: ") + e.getName() + kUTF8_SCOLON;
        result += STAFString("Location: ") + e.getLocation() + kUTF8_SCOLON;
        result += STAFString("Text: ") + e.getText() + kUTF8_SCOLON;
        result += STAFString("Error code: ") + e.getErrorCode() + kUTF8_SCOLON;

        *pErrorBuffer = result.adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFZip.cpp: STAFServiceTerm: Caught unknown "
                    "exception in STAFServiceTerm()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


STAFRC_t STAFServiceDestruct(STAFServiceHandle_t *serviceHandle,
                             void *pDestructInfo, unsigned int destructLevel,
                             STAFString_t *pErrorBuffer)
{
    if (destructLevel != 0) return kSTAFInvalidAPILevel;

    STAFRC_t retCode = kSTAFUnknownError;

    try
    {
        STAFZipServiceData *pData =
            reinterpret_cast<STAFZipServiceData *>(*serviceHandle);

        delete pData;
        *serviceHandle = 0;

        retCode = kSTAFOk;
    }
    catch (STAFException &e)
    {
        STAFString result;

        result += STAFString("In STAFZip.cpp: STAFServiceDestruct")
            + kUTF8_SCOLON;

        result += STAFString("Name: ") + e.getName() + kUTF8_SCOLON;
        result += STAFString("Location: ") + e.getLocation() + kUTF8_SCOLON;
        result += STAFString("Text: ") + e.getText() + kUTF8_SCOLON;
        result += STAFString("Error code: ") + e.getErrorCode() + kUTF8_SCOLON;

        *pErrorBuffer = result.adoptImpl();
    }
    catch (...)
    {
        STAFString error(
                    "STAFZip.cpp: STAFServiceDestruct: Caught "
                    "unknown exception in STAFServiceDestruct()");
        *pErrorBuffer = error.adoptImpl();
    }

    return retCode;
}


// Handles Zip requests

STAFResultPtr handleZip(STAFServiceRequestLevel2 *pInfo,
                        STAFZipServiceData *pData, STAFString *mutexName)
{
    STAFString result;
    STAFString resultBuffer;
    
    char zipFile[MAXFILENAME + 1] = "";
    char entry[MAXFILENAME + 1] = "";
    char excludePrefix[MAXFILENAME + 1] = "";
    char zipFileBackup[MAXFILENAME + 1] = "";

    unsigned int entryLength = 0;
    unsigned int prefixLength = 0;

    int replace = 0, recursive = 0;

    STAFZipFile *zf;
    STAFZipUtil util = STAFZipUtil(pData->fHandlePtr);


    STAFRC_t rc = kSTAFOk;

    if (pInfo->trustLevel < 4)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult =
        pData->fZipParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }


    // get "ZIPFILE" parameter value
    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, "ZIPFILE");

    if (resultPtr->rc != 0) return resultPtr;

    if (resultPtr->result.length() <= MAXFILENAME)
    {
        util.convertSTAFStringBuffer(resultPtr->result.toCurrentCodePage(), zipFile);
    }
    else
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                 STAFString("STAFZip::handleZip: ZIPFILE [")
                                 + zipFile
                                 + "] file name length exceeds "
                                 + MAXFILENAME
                                 + " charactors.\n"),
                                 STAFResultPtr::INIT);
    }

    resultPtr->result = resultPtr->result.replace(kUTF8_BSLASH, kUTF8_SLASH);

    // use zip file name to construct the mutex name
    *mutexName += resultPtr->result;
    
    *mutexName = (*mutexName).toUpperCase();
    

    // obtain mutex semphore on the zipfile
    zipLock.request(*mutexName);


    if (resultPtr->rc != 0) return resultPtr;

    
    // get "RELATIVETO" parameter value
    resultPtr = resolveOp(pInfo, pData, parsedResult, "RELATIVETO");

    if (resultPtr->rc != 0) return resultPtr;

    resultPtr->result = resultPtr->result.replace(kUTF8_BSLASH, kUTF8_SLASH);

    prefixLength = resultPtr->result.length();

    if (prefixLength > 0)
    {
        if (prefixLength <= MAXFILENAME)
        {
            util.convertSTAFStringBuffer(resultPtr->result.toCurrentCodePage(), excludePrefix);

            if (excludePrefix[prefixLength - 1] != '/')
            {
                excludePrefix[prefixLength] = '/';
                
                excludePrefix[prefixLength + 1] = kUTF8_NULL;
                
                prefixLength++;
            }
        }
        else
        {
            return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                     STAFString("STAFZip::handleZip:"
                                     " RELATIVETO [")
                                     + excludePrefix
                                     + "] file name length exceeds "
                                     + MAXFILENAME
                                     + " charactors.\n"),
                                     STAFResultPtr::INIT);
        }
    }


    // get "RECURSE" parameter value
    if (parsedResult->optionTimes("RECURSE"))
    {
        recursive = 1;
    }


    // make a backup copy of the current zip file
    int fileExist = util.checkFileExist(zipFile);

    if (fileExist)
    {
        strcpy(zipFileBackup, zipFile);

        strcat(zipFileBackup, ".ZIP");

        if(util.copyFile(zipFile, zipFileBackup) != kSTAFOk)
        {

            return STAFResultPtr(new STAFResult(kZIPGeneralZipError,
                                     STAFString("STAFZip::handleZip:"
                                     " Can't create file [")
                                     + zipFile
                                     + ".ZIP].\n"),
                                     STAFResultPtr::INIT);
        }
    }


    // get number of entries for FILE or DIRECTORY option
    unsigned int numFilesToAdd = parsedResult->optionTimes("FILE");
    
    unsigned int numDirsToAdd = parsedResult->optionTimes("DIRECTORY");
    
    
    // zipping the files
    for (unsigned int i = 1; i <= numFilesToAdd && rc == kSTAFOk; i++)
    {

        // resolve variables
        STAFString thisEntry = parsedResult->optionValue("FILE", i);

        resultPtr = resolveStr(pInfo, pData, thisEntry);

        if (resultPtr->rc != 0) 
        {
            rc = resultPtr->rc;

            result = resultPtr->result;
            
            break;
        }

        thisEntry = resultPtr->result;

        // replace all backward slashes with forward slashes
        thisEntry = thisEntry.replace(kUTF8_BSLASH, kUTF8_SLASH);

        entryLength = thisEntry.length();

        if (entryLength <= MAXFILENAME)
        {
            util.convertSTAFStringBuffer(thisEntry.toCurrentCodePage(), entry);
        }
        else
        {

            result = STAFString("STAFZip::handleZip: FILE [")
                     + entry
                     + "] file name length exceeds "
                     + MAXFILENAME
                     + " charactors.\n";

            rc = kSTAFInvalidValue;

            break;
        }


        if (prefixLength != 0)
        {
            // check if excludeprefix is shorter than entry name
            if (prefixLength > strlen(entry))
            {
                result = STAFString("STAFZip::handleZip: The length of"
                                    " RELATIVETO [")
                         + excludePrefix
                         + "] exceeds the length of FILE ["
                         + entry
                         + "].\n";

                rc = kSTAFInvalidValue;

                break;
            }

            // save the char of entry at the prefixLength
            char c = *(entry + prefixLength);

            // truncate the entry to be a identical c string as excludeprefix
            entry[prefixLength] = kUTF8_NULL;

            // check if excludeprefix is a part of the entry name
            if (util.myStrCmp(entry, excludePrefix))
            {
                result = STAFString("STAFZip::handleZip: RELATIVETO [")
                         + excludePrefix
                         + "] is not a leading sub-string of FILE ["
                         + entry
                         + "].\n";

                rc = kSTAFInvalidValue;

                break;
            }

            // restore the char of entry at the prefixLength
            *(entry + prefixLength) = c;
        }

        if (fileExist)
        {

            STAFTrace::trace(kSTAFTraceServiceResult,
                             STAFString("handleZip_CP1")
                             +" zipFileBackup ["
                             + zipFileBackup
                             + "]");

            // if appending to existing ZIP archive, use the backup file name
            zf = new STAFZipFile(pData->fHandlePtr, zipFileBackup,
                                 &rc, &resultBuffer, kSTAFZipReadWriteBinary);
        }
        else
        {
            STAFTrace::trace(kSTAFTraceServiceResult,
                             STAFString("handleZip_CP2")
                             + " zipFile ["
                             + zipFile
                             + "]");

            // if creating a new zip archive, use the original file name
            zf = new STAFZipFile(pData->fHandlePtr, zipFile,
                                 &rc, &resultBuffer, kSTAFZipCreateNewFile);
        }

        if (rc != kSTAFOk)
        {
            delete zf,

            result = resultBuffer;

            break;
        }

        if (STAFTrace::doTrace(kSTAFTraceServiceResult))
        {
            STAFTrace::trace(kSTAFTraceServiceResult,
                         STAFString("handleZip_CP3")
                         + " entry ["
                         + entry
                         + "] prefixLength ["
                         + prefixLength
                         + "] recursive ["
                         + recursive
                         + "]");
        }
        
        // adding the entry to the zip archive
        if ((rc = zf->zipFile(entry, prefixLength, recursive, &resultBuffer)) !=
             kSTAFOk)
        {
            result = resultBuffer;
        }

        delete zf;
    }


    // zipping the directories
    for (unsigned int j = 1; j <= numDirsToAdd && rc == kSTAFOk; j++)
    {

        // resolve variables
        STAFString thisEntry = parsedResult->optionValue("DIRECTORY", j);

        resultPtr = resolveStr(pInfo, pData, thisEntry);

        if (resultPtr->rc != 0) 
        {
            rc = resultPtr->rc;

            result = resultPtr->result;
            
            break;
        }

        thisEntry = resultPtr->result;


        // replace all backward slashes with forward slashes
        thisEntry = thisEntry.replace(kUTF8_BSLASH, kUTF8_SLASH);

        entryLength = thisEntry.length();

        if (entryLength <= MAXFILENAME)
        {
            util.convertSTAFStringBuffer(thisEntry.toCurrentCodePage(), entry);

            // append a '/' to the end of directory name
            if (entry[entryLength - 1] != '/')
            {
                entry[entryLength] = '/';

                entry[entryLength + 1] = kUTF8_NULL;
                
                entryLength++;
            }
        }
        else
        {

            result = STAFString("STAFZip::handleZip: DIRECTORY [")
                     + entry
                     + "] file name length exceeds "
                     + MAXFILENAME
                     + " charactors.\n";

            rc = kSTAFInvalidValue;

            break;
        }


        if (prefixLength != 0)
        {
            // check if excludeprefix is shorter than entry name
            if (prefixLength > strlen(entry))
            {
                result = STAFString("STAFZip::handleZip: The length of"
                                    " RELATIVETO [")
                         + excludePrefix
                         + "] exceeds the length of DIRECTORY ["
                         + entry
                         + "].\n";

                rc = kSTAFInvalidValue;

                break;
            }

            // save the char of entry at the prefixLength
            char c = *(entry + prefixLength);

            // truncate the entry to be a identical c string as excludeprefix
            entry[prefixLength] = kUTF8_NULL;

            // check if excludeprefix is a part of the entry name
            if (util.myStrCmp(entry, excludePrefix))
            {
                result = STAFString("STAFZip::handleZip: RELATIVETO [")
                         + excludePrefix
                         + "] is not a leading sub-string of DIRECTORY ["
                         + entry
                         + "].\n";

                rc = kSTAFInvalidValue;

                break;
            }

            // restore the char of entry at the prefixLength
            *(entry + prefixLength) = c;
        }

        if (fileExist)
        {

            STAFTrace::trace(kSTAFTraceServiceResult,
                             STAFString("handleZip_CP1")
                             +" zipFileBackup ["
                             + zipFileBackup
                             + "]");

            // if appending to existing ZIP archive, use the backup file name
            zf = new STAFZipFile(pData->fHandlePtr, zipFileBackup,
                                 &rc, &resultBuffer, kSTAFZipReadWriteBinary);
        }
        else
        {
            STAFTrace::trace(kSTAFTraceServiceResult,
                             STAFString("handleZip_CP2")
                             + " zipFile ["
                             + zipFile
                             + "]");

            // if creating a new zip archive, use the original file name
            zf = new STAFZipFile(pData->fHandlePtr, zipFile,
                                 &rc, &resultBuffer, kSTAFZipCreateNewFile);
        }

        if (rc != kSTAFOk)
        {
            delete zf,

            result = resultBuffer;

            break;
        }

        if (STAFTrace::doTrace(kSTAFTraceServiceResult))
        {
            STAFTrace::trace(kSTAFTraceServiceResult,
                         STAFString("handleZip_CP3")
                         + " entry ["
                         + entry
                         + "] prefixLength ["
                         + prefixLength
                         + "] recursive ["
                         + recursive
                         + "]");
        }
        
        // adding the entry to the zip archive
        if ((rc = zf->zipFile(entry, prefixLength, recursive, &resultBuffer)) !=
             kSTAFOk)
        {
            result = resultBuffer;
        }

        delete zf;
    }


    if (rc == kSTAFOk)
    {
        if (fileExist)
        {
            // if appending to existing archive
            // remove the original zip archive
            // since we are working on the backup copy
            if (remove(zipFile) != 0)
            {
                result = STAFString("STAFZip::handleZip: ")
                         + STAFString("Remove file fails: [")
                         + STAFString(zipFile)
                         + STAFString("].\n");

                rc = kSTAFFileDeleteError;
            }

            // rename the backup copy to be the zip archive name
            if (rename(zipFileBackup, zipFile) != 0)
            {
                result = STAFString("STAFZip::handleZip: ")
                         + STAFString("Rename file fails from [")
                         + STAFString(zipFileBackup)
                         + STAFString("] to [")
                         + STAFString(zipFile)
                         + STAFString("].\n");

                rc = kZIPGeneralZipError;
            }
        }
    }
    else
    {
        // if there was error occurred
        if (fileExist)
        {
            // if appending to existing zip archive,
            // remove the backup copy that we are working on
            // so that leaves the original untouched
            remove(zipFileBackup);
        }
        else
        {
            // if this is a new archive that we create
            // just remove it since there was an error during zipping
            remove(zipFile);
        }
    }


    return STAFResultPtr(new STAFResult(rc, result),
                             STAFResultPtr::INIT);

}


// Handles Unzip requests
STAFResultPtr handleUnzip(STAFServiceRequestLevel2 *pInfo,
                           STAFZipServiceData *pData, STAFString *mutexName)
{
    STAFString result;
    STAFString resultBuffer;

    char zipFile[MAXFILENAME + 1] = "";
    char file[MAXFILENAME + 1] = "";
    char toDir[MAXFILENAME + 1] = "";

    unsigned int toDirLength = 0;
    unsigned int fileLength = 0;
    int replace = 0;
    int restorePermission = 0;

    STAFZipFile *zf;
    STAFZipUtil util = STAFZipUtil(pData->fHandlePtr);

    STAFRC_t rc = kSTAFOk;


    if (pInfo->trustLevel < 4)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult =
        pData->fUnzipParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }


    // get "ZIPFILE" parameter value
    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, "ZIPFILE");

    if (resultPtr->rc != 0) return resultPtr;

    if (resultPtr->result.length() <= MAXFILENAME)
    {
        util.convertSTAFStringBuffer(resultPtr->result.toCurrentCodePage(), zipFile);
    }
    else
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                 STAFString("STAFZip::handleUnzip:"
                                            " ZIPFILE [")
                                 + zipFile
                                 + "] file name length exceeds "
                                 + MAXFILENAME
                                 + " charactors.\n"),
                                 STAFResultPtr::INIT);
    }

    resultPtr->result = resultPtr->result.replace(kUTF8_BSLASH, kUTF8_SLASH);

    // use zip file name to construct the mutex name
    *mutexName += resultPtr->result;
    
    *mutexName = (*mutexName).toUpperCase();

    // obtain mutex semphore on the zipfile
    zipLock.request(*mutexName);


    if (resultPtr->rc != 0) return resultPtr;


    // get "TODIRECTORY" parameter value
    resultPtr = resolveOp(pInfo, pData, parsedResult, "TODIRECTORY");

    if (resultPtr->rc != 0) return resultPtr;

    resultPtr->result = 
            STAFString(resultPtr->result.toCurrentCodePage()->buffer());

    resultPtr->result = resultPtr->result.replace(kUTF8_BSLASH, kUTF8_SLASH);

    toDirLength = resultPtr->result.length();

    if (toDirLength <= MAXFILENAME)
    {
        util.convertSTAFStringBuffer(resultPtr->result.toCurrentCodePage(), toDir);

        if (toDir[toDirLength - 1] != '/')
        {
            toDir[toDirLength] = '/';

            toDir[toDirLength + 1] = kUTF8_NULL;
            
            toDirLength++;
        }
    }
    else
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                     STAFString("STAFZip::handleUnzip:"
                                                " TODIRECTORY [")
                                     + toDir
                                     + "] file name length exceeds "
                                     + MAXFILENAME
                                     + " charactors.\n"),
                                     STAFResultPtr::INIT);
    }


    // get "REPLACE" parameter value
    if (parsedResult->optionTimes("REPLACE"))
    {
        replace = 1;
    }


    // get "RESTOREPERMISSION" parameter value
    if (parsedResult->optionTimes("RESTOREPERMISSION"))
    {
        restorePermission = 1;
    }


    STAFTrace::trace(kSTAFTraceServiceResult,
                     STAFString("handleUnzip_CP1")
                     +" zipFile ["
                     + zipFile
                     + "]");

    // open the ZIP archive
    zf = new STAFZipFile(pData->fHandlePtr, zipFile, &rc, &resultBuffer,
                         kSTAFZipReadOnlyBinary);

    if (rc != kSTAFOk)
    {
        delete zf;

        return STAFResultPtr(new STAFResult(rc, resultBuffer),
                                     STAFResultPtr::INIT);
    }

    unsigned int numFilesToUnzip = parsedResult->optionTimes("FILE");

    if (numFilesToUnzip == 0)
    {

        if (STAFTrace::doTrace(kSTAFTraceServiceResult))
        {
            STAFTrace::trace(kSTAFTraceServiceResult,
                         STAFString("handleUnzip_CP2")
                         + " toDir ["
                         + toDir
                         + "] replace ["
                         + replace
                         + "] restorePermission ["
                         + restorePermission
                         + "]");
        }

        // if not FILE option is provided, unzip all files in the archive
        if ((rc = zf->unzipFile(toDir, replace, restorePermission,
             &resultBuffer)) != kSTAFOk)
        {
            result = resultBuffer;
        }
    }
    else
    {
        // unzip the individual files one by one
        for (unsigned int i = 1; i <= numFilesToUnzip; i++)
        {
            // resolve variables
            STAFString thisFile = parsedResult->optionValue("FILE", i);

            resultPtr = resolveStr(pInfo, pData, thisFile);

            if (resultPtr->rc != 0) 
            {
                rc = resultPtr->rc;
    
                result = resultPtr->result;
            
                break;
            }

            thisFile = resultPtr->result;

            fileLength = thisFile.length();

            if (fileLength != 0)
            {
                if (fileLength <= MAXFILENAME)
                {
                    util.convertSTAFStringBuffer(thisFile.toCurrentCodePage(), file);
                }
                else
                {
                    result = STAFString("STAFZip::handleUnzip:"
                                        " FILE [")
                             + file
                             + "] file name length exceeds "
                             + MAXFILENAME
                             + " charactors.\n";

                    rc = kSTAFInvalidValue;

                    break;
                }
            }

            if (STAFTrace::doTrace(kSTAFTraceServiceResult))
            {
                STAFTrace::trace(kSTAFTraceServiceResult,
                             STAFString("handleUnzip_CP3")
                             + " file ["
                             + file
                             + "] toDir ["
                             + toDir
                             + "] replace ["
                             + replace
                             + "] restorePermission ["
                             + restorePermission
                             + "]");
            }
            
            
            // unzip the file entry
            if ((rc = zf->unzipFile(file, toDir, replace, restorePermission,
                &resultBuffer)) != kSTAFOk)
            {
                result = resultBuffer;

                // if fails, exit the loop
                break;
            }
        }
    }


    delete zf;


    return STAFResultPtr(new STAFResult(rc, result),
                         STAFResultPtr::INIT);

}


// Handles Delete requests
STAFResultPtr handleDelete(STAFServiceRequestLevel2 *pInfo,
                           STAFZipServiceData *pData, STAFString *mutexName)
{
    STAFString result;
    STAFString resultBuffer;

    char zipFile[MAXFILENAME + 1] = "";
    char deleteFile[MAXFILENAME + 1] = "";
    char zipFileBackup[MAXFILENAME + 1] = "";

    STAFZipFile *zf;
    STAFZipUtil util = STAFZipUtil(pData->fHandlePtr);

    STAFRC_t rc = kSTAFOk;


    if (pInfo->trustLevel < 4)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFCommandParseResultPtr parsedResult =
        pData->fDeleteParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }


    // get "ZIPFILE" parameter value
    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, "ZIPFILE");

    if (resultPtr->rc != 0) return resultPtr;

    if (resultPtr->result.length() <= MAXFILENAME)
    {
        util.convertSTAFStringBuffer(resultPtr->result.toCurrentCodePage(), zipFile);
    }
    else
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                            STAFString("STAFZip::handleDelete:"
                                            " ZIPFILE [")
                                            + zipFile
                                            + "] file name length exceeds "
                                            + MAXFILENAME
                                            + " charactors.\n"),
                                            STAFResultPtr::INIT);
    }


    resultPtr->result = resultPtr->result.replace(kUTF8_BSLASH, kUTF8_SLASH);

    // use zip file name to construct the mutex name
    *mutexName += resultPtr->result;
    
    *mutexName = (*mutexName).toUpperCase();

    // obtain mutex semphore on the zipfile
    zipLock.request(*mutexName);


    if (resultPtr->rc != 0) return resultPtr;


    int fileExist = util.checkFileExist(zipFile);

    if (!fileExist)
    {

        return STAFResultPtr(new STAFResult(kSTAFDoesNotExist,
                                          STAFString("STAFZip::handleDelete:"
                                          " Zip file does not exist [")
                                          + zipFile
                                          + "].\n"),
                                          STAFResultPtr::INIT);
    }        
        
    // backup the original zip archive
    strcpy(zipFileBackup, zipFile);

    strcat(zipFileBackup, ".ZIP");

    if(util.copyFile(zipFile, zipFileBackup) != kSTAFOk)
    {

        return STAFResultPtr(new STAFResult(kZIPGeneralZipError,
                                          STAFString("STAFZip::handleDelete:"
                                          " Can't create file [")
                                          + zipFile
                                          + ".ZIP].\n"),
                                          STAFResultPtr::INIT);
    }


    // loop to delete file entries from zip archive
    unsigned int numFilesToDelete = parsedResult->optionTimes("FILE");

    for (unsigned int i = 1; i <= numFilesToDelete; i++)
    {

        // resolve variables
        STAFString thisFile = parsedResult->optionValue("FILE", i);

        resultPtr = resolveStr(pInfo, pData, thisFile);

        if (resultPtr->rc != 0) 
        {
            rc = resultPtr->rc;
    
            result = resultPtr->result;
            
            break;
        }

        thisFile = resultPtr->result;

        if (thisFile.length() <= MAXFILENAME)
        {
            util.convertSTAFStringBuffer(thisFile.toCurrentCodePage(), deleteFile);
        }
        else
        {
            result = STAFString("STAFZip::handleDelete: FILE [")
                     + STAFString(deleteFile)
                     + "] file name length exceeds "
                     + MAXFILENAME
                     + " charactors.\n";

            rc = kSTAFInvalidValue;

            break;
        }


        // delete a file from zip archive

        uLong newsize;


        STAFTrace::trace(kSTAFTraceServiceResult,
                         STAFString("handleDelete_CP1")
                             + " zipFileBackup ["
                             + zipFileBackup
                             + "]");

        // open the backup zip archive
        zf = new STAFZipFile(pData->fHandlePtr, zipFileBackup, &rc,
                             &resultBuffer, kSTAFZipReadWriteBinary);

        if (rc != kSTAFOk)
        {
            delete zf;

            result = resultBuffer;

            break;
        }


        STAFTrace::trace(kSTAFTraceServiceResult,
                         STAFString("handleDelete_CP2")
                             + " deleteFile ["
                             + deleteFile
                             + "]");

        // delete one entry from the zip archive
        if ((rc = zf->deleteFile(deleteFile, &newsize, &resultBuffer))
            != kSTAFOk)
        {
            delete zf;

            result = resultBuffer;

            break;
        }

        delete zf;


        // to reduce file size after deleting file entry
        if (newsize == 0)
        {
            // if all files in the backup zip archive are removed,
            // remove the backup zip archive
            if ((rc = remove(zipFileBackup)) != 0)
            {
                result = STAFString("STAFZip::handleDelete: "
                                    "Remove file fails: [")
                         + STAFString(zipFileBackup)
                         + STAFString("].\n");

                rc = kSTAFFileDeleteError;

                break;
            }
        }
        else
        {
            // change file size to the new size
            // after deleting the file entry
            if ((rc = util.changeFileSize(zipFileBackup, newsize)) != 0)
            {
                result = STAFString("STAFZip::handleDelete:"
                         " Change file size fails: [")
                         + zipFileBackup
                         + STAFString("] to [")
                         + newsize
                         + STAFString("].\n");

                rc = kZIPGeneralZipError;

                break;
            }
        }
    }


    if (rc == kSTAFOk)
    {
        // if delete was successful,
        // remove the original zip archive
        // since we are working on the backup zip archive
        if (remove(zipFile) != 0)
        {
            result = STAFString("STAFZip::handleDelete:"
                                " Remove zipfile fails: [")
                     + zipFile
                     + STAFString("].\n");

            rc = kSTAFFileDeleteError;
        }

        // rename the backup zip archive to the original zip archive
        // since we are working on the backup zip archive
        if (rename(zipFileBackup, zipFile) != 0)
        {
            result = STAFString("STAFZip::handleDelete:"
                                " Rename zipfile fails: from [")
                     + zipFileBackup
                     + STAFString("] to [")
                     + zipFile
                     + STAFString("].\n");

            rc = kZIPGeneralZipError;
        }
    }
    else
    {
        // if delete was unsuccessful
        // remove the backup zip archive that we are working on
        remove(zipFileBackup);
    }


    return STAFResultPtr(new STAFResult(rc, result),
                         STAFResultPtr::INIT);

}


// Handles List requests
STAFResultPtr handleList(STAFServiceRequestLevel2 *pInfo,
                           STAFZipServiceData *pData, STAFString *mutexName)
{
    STAFString resultBuffer;

    char zipFile[MAXFILENAME + 1] = "";

    STAFZipFile *zf;
    STAFZipUtil util = STAFZipUtil(pData->fHandlePtr);

    STAFRC_t rc = kSTAFOk;


    if (pInfo->trustLevel < 3)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }


    STAFCommandParseResultPtr parsedResult =
        pData->fListParser->parse(pInfo->request);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidRequestString,
                             parsedResult->errorBuffer), STAFResultPtr::INIT);
    }


    // get "ZIPFILE" parameter value
    STAFResultPtr resultPtr = resolveOp(pInfo, pData, parsedResult, "ZIPFILE");

    if (resultPtr->rc != 0) return resultPtr;

    if (resultPtr->result.length() <= MAXFILENAME)
    {
        util.convertSTAFStringBuffer(resultPtr->result.toCurrentCodePage(), zipFile);
    }
    else
    {
        return STAFResultPtr(new STAFResult(kSTAFInvalidValue,
                                            STAFString("STAFZip::handleList:"
                                            " ZIPFILE [")
                                            + zipFile
                                            +"] file name length exceeds "
                                            + MAXFILENAME
                                            + " charactors.\n"),
                                            STAFResultPtr::INIT);
    }


    resultPtr->result = resultPtr->result.replace(kUTF8_BSLASH, kUTF8_SLASH);

    // use zip file name to construct the mutex name
    *mutexName += resultPtr->result;
    
    *mutexName = (*mutexName).toUpperCase();

    // obtain mutex semphore on the zipfile
    zipLock.request(*mutexName);


    if (resultPtr->rc != 0) return resultPtr;


    if (pInfo->diagEnabled)
    {
        // Record diagnostics because the output of a List request will
        // change in STAF V3.0

        static STAFString request("RECORD TRIGGER %C SOURCE %C");
        static STAFString trigger("STAF/V3.0-Mig/ZIP LIST");
        static STAFString sourceFormat("%s;%s;%d");
        STAFString source = STAFHandle::formatString(
            sourceFormat.getImpl(), pInfo->processName,
            pInfo->machine, pInfo->handle);

        pData->fHandlePtr->submit(
            sLocal, "DIAG",
            STAFHandle::formatString(
                request.getImpl(), trigger.getImpl(), source.getImpl()),
            kSTAFReqFireAndForget);
    }


    STAFTrace::trace(kSTAFTraceServiceResult,
                     STAFString("handleList_CP1")
                     + " zipFile ["
                     + zipFile
                     + "]");


    // open the Zip archive, don't create a new one
    zf = new STAFZipFile(pData->fHandlePtr, zipFile, &rc, &resultBuffer,
                         kSTAFZipReadOnlyBinary);

    if (rc != kSTAFOk)
    {
        delete zf;

        return STAFResultPtr(new STAFResult(rc, resultBuffer),
                         STAFResultPtr::INIT);
    }

    STAFTrace::trace(kSTAFTraceServiceResult,
                     STAFString("handleList_CP2"));

    STAFString buffer;

    rc = zf->listFile(&buffer, &resultBuffer);


    delete zf;

    if (rc == kSTAFOk)
    {
        resultBuffer = buffer;
    }

    return STAFResultPtr(new STAFResult(rc, resultBuffer),
                         STAFResultPtr::INIT);
}



STAFResultPtr handleHelp(STAFServiceRequestLevel2 *pInfo,
                         STAFZipServiceData *pData)
{
    if (pInfo->trustLevel < 1)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    STAFString help("ZIP Service Help" + sLineSep + sLineSep);

    help += "ZIP    ADD ZIPFILE <Name> < FILE <Name> | DIRECTORY <Name> [RECURSE] >";
    help += sLineSep;
    help += "       [RELATIVETO <Directory>]";
    help += sLineSep;
    help += "UNZIP  ZIPFILE <Name> TODIRECTORY <Name>";
    help += sLineSep;
    help += "       [FILE <Name>]... [RESTOREPERMISSION] [REPLACE]";
    help += sLineSep;
    help += "DELETE ZIPFILE <Name> FILE <Name> [FILE <Name>]... CONFIRM";
    help += sLineSep;
    help += "LIST   ZIPFILE <Name>";
    help += sLineSep;
    help += "VERSION" + sLineSep;
    help += "HELP" + sLineSep;

    return STAFResultPtr(new STAFResult(kSTAFOk, help), STAFResultPtr::INIT);
}


STAFResultPtr handleVersion(STAFServiceRequestLevel2 *pInfo,
                            STAFZipServiceData *pData)
{
    if (pInfo->trustLevel < 1)
    {
        return STAFResultPtr(new STAFResult(kSTAFAccessDenied, STAFString()),
                             STAFResultPtr::INIT);
    }

    return STAFResultPtr(new STAFResult(kSTAFOk, sVersionInfo),
                         STAFResultPtr::INIT);
}


STAFResultPtr resolveOp(STAFServiceRequestLevel2 *pInfo,
                        STAFZipServiceData *pData,
                        STAFCommandParseResultPtr &parsedResult,
                        const STAFString &fOption, unsigned int optionIndex)
{
    STAFString optionValue = parsedResult->optionValue(fOption, optionIndex);

    if (optionValue.find(sLeftCurlyBrace) == STAFString::kNPos)
    {
        return STAFResultPtr(new STAFResult(kSTAFOk, optionValue),
                             STAFResultPtr::INIT);
    }

    return resolveStr(pInfo, pData, optionValue);
}


STAFResultPtr resolveStr(STAFServiceRequestLevel2 *pInfo,
                         STAFZipServiceData *pData,
                         const STAFString &theString)
{
    if (STAFString(pInfo->machine) == STAFString(pInfo->localMachine))
    {
        return pData->fHandlePtr->submit(sLocal, sVar, sResStrHandle +
                                     STAFString(pInfo->handle) +
                                     sResStrResolve +
                                     pData->fHandlePtr->wrapData(theString));
    }

    return pData->fHandlePtr->submit(sLocal, sVar, sResStrGResolve +
                                     pData->fHandlePtr->wrapData(theString));
}


void registerHelpData(STAFZipServiceData *pData, unsigned int errorNumber,
                      const STAFString &shortInfo, const STAFString &longInfo)
{
    static STAFString regString("REGISTER SERVICE %C ERROR %d INFO %C "
                                "DESCRIPTION %C");

    pData->fHandlePtr->submit(sLocal, sHelp, STAFHandle::formatString(
        regString.getImpl(), pData->fShortName.getImpl(), errorNumber,
        shortInfo.getImpl(), longInfo.getImpl()));
}


void unregisterHelpData(STAFZipServiceData *pData, unsigned int errorNumber)
{
    static STAFString regString("UNREGISTER SERVICE %C ERROR %d");

    pData->fHandlePtr->submit(sLocal, STAFString("HELP"), STAFHandle::
    formatString(
        regString.getImpl(), pData->fShortName.getImpl(), errorNumber));
}


























