/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#ifndef STAF_Connection
#define STAF_Connection

#include "STAFString.h"
#include "STAFRefPtr.h"
#include "STAFUtil.h"
#include "STAFException.h"

// STAFConnection - An interface class used to read and write from an
//                  arbitrary source
//
// This class is meant to be a generic front end to various sources (such
// as Local Named Pipe, TCP/IP Sockets, NETBIOS Sessions) which are derived
// from this class.
//
// Facilities are provided to read/write unsigned ints, strings, and
// arbitrary buffers.
//
// A derived class must override the void * versions of read/write and
// should not override any of the other functions.
//
//
// STAFConnAllocInfo - This structure contains info on how to free a particular
//                     connection object

class STAFConnection;  // Forward declaration needed for typedef

typedef STAFRefPtr<STAFConnection> STAFConnPtr;
typedef void (*STAFConnectionFreeFunction)(STAFConnection *);


struct STAFConnAllocInfo
{
    STAFConnAllocInfo(STAFConnection *pConn,
                      STAFConnectionFreeFunction freeFunc)
        : pConnection(pConn), pFreeFunc(freeFunc)
    { /* Do Nothing */ }

    STAFConnection *pConnection;
    STAFConnectionFreeFunction pFreeFunc;
};


class STAFConnection
{
public:

    enum ConnectionType { kLocal = 1, kNetwork = 2};
    enum ConnectionMode { kServer = 1, kClient = 2};

    STAFConnection(ConnectionType conType, ConnectionMode conMode);

    ConnectionType connectionType() const { return fConnectionType; }
    ConnectionMode connectionMode() const { return fConnectionMode; }

    // Reads size bytes into buffer
    virtual void read(void *buffer, unsigned int size) = 0;

    unsigned int readUInt()
    {
        unsigned int data = 0;
        read((void *)&data, sizeof(unsigned int));
        return STAFUtilConvertNativeUIntToLE(data);
    }

    void readUInt(unsigned int &uint)
    {
        read((void *)&uint, sizeof(unsigned int));
        uint = STAFUtilConvertNativeUIntToLE(uint);
    }

    STAFString readString();
    void readString(STAFString &theString);

    // Writes size bytes from buffer
    virtual void write(void *buffer, unsigned int size) = 0;

    void writeUInt(unsigned int uint)
    {
        uint = STAFUtilConvertNativeUIntToLE(uint);
        write((void *)&uint, sizeof(unsigned int));
    }

    void writeString(const STAFString &theString);

    virtual ~STAFConnection();

private:

    // Don't allow copy construction or assignment
    STAFConnection(const STAFConnection &);
    STAFConnection &operator=(const STAFConnection &);

    ConnectionType fConnectionType;
    ConnectionMode fConnectionMode;
};

// Connection Exception declarations

STAF_EXCEPTION_DEFINITION(STAFConnectionException, STAFException);
STAF_EXCEPTION_DEFINITION(STAFConnectionConnectException,
                          STAFConnectionException);
STAF_EXCEPTION_DEFINITION(STAFConnectionConnectTimeoutException,
                          STAFConnectionConnectException);
STAF_EXCEPTION_DEFINITION(STAFConnectionIOException, STAFConnectionException);

#endif
