/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFSocketFuncs.h"
#include "STAFSocketConnection.h"
#include "STAFException.h"
#include <string.h>
#include <algorithm>

STAFConnAllocInfo STAFSocketConnection::create(STAFString host,
                                               unsigned short port,
                                               unsigned int timeout)
{
    return STAFConnAllocInfo(new STAFSocketConnection(host, port, timeout),
                             &freeSocketConnection);
}


STAFConnAllocInfo STAFSocketConnection::create(int socket)
{
    return STAFConnAllocInfo(new STAFSocketConnection(socket),
                             &freeSocketConnection);
}


STAFSocketConnection::STAFSocketConnection(STAFString host, unsigned short port,
                                           unsigned int timeout)
    : STAFConnection(kNetwork, kClient)
{
    struct sockaddr_in serverAddress = { 0 };

    serverAddress.sin_family = AF_INET;
    serverAddress.sin_port = STAFSocketFuncs::bswap(port);

    if (host.findFirstNotOf("1234567890.") != STAFString::kNPos)
    {
        // This is a hostname

        unsigned int osRC = 0;
        STAFRC_t rc = STAFSocketFuncs::getInAddrByName(
                      host.toCurrentCodePage()->buffer(),
                      &serverAddress.sin_addr, &osRC);

        if (rc != 0)
        {
            STAFConnectionConnectException hostnameError("getInAddrByName",
                                                         osRC);
            THROW_STAF_EXCEPTION(hostnameError);
        }
    }
    else
    {
        // This is an IP address

        if (host.count(kUTF8_PERIOD) != 3)
        {
            STAFString mesg = "Invalid IP Address, " + host;
            STAFConnectionConnectException badIPAddr(
                mesg.toCurrentCodePage()->buffer());
            THROW_STAF_EXCEPTION(badIPAddr);
        }

        serverAddress.sin_addr.s_addr =
            STAFSocketFuncs::inet_addr(host.toCurrentCodePage()->buffer());
    }

    fSocket = STAFSocketFuncs::socket(PF_INET, SOCK_STREAM, 0);

    if (!STAFUtilIsValidSocket(fSocket))
    {
        STAFConnectionConnectException noSocket("socket",
                                                STAFSocketFuncs::sock_errno());
        THROW_STAF_EXCEPTION(noSocket);
    }

    int modeRC = STAFSocketFuncs::setBlockingMode(fSocket,
                                                  STAFSocketFuncs::kNonBlocking);
    if (modeRC != 0)
    {
        STAFConnectionConnectException modeError("setBlockingMode(NonBlocking)",
                                                 modeRC);
        STAFSocketFuncs::soclose(fSocket);
        THROW_STAF_EXCEPTION(modeError);
    }

    int connectRC = STAFSocketFuncs::connect(fSocket,
                    (struct sockaddr *)&serverAddress, sizeof(serverAddress));

    if ((connectRC < 0) && (STAFSocketFuncs::sock_errno() != SOCEINPROGRESS) &&
       (STAFSocketFuncs::sock_errno() != SOCEWOULDBLOCK) &&
       (STAFSocketFuncs::sock_errno() != 0))
    {
        STAFConnectionConnectException connectError("connect",
                                       STAFSocketFuncs::sock_errno());
        STAFSocketFuncs::soclose(fSocket);
        THROW_STAF_EXCEPTION(connectError);
    }

    if (connectRC < 0)
    {
        fd_set writeSocks;
        FD_ZERO(&writeSocks);
        FD_SET(fSocket, &writeSocks);
        timeval theTimeout = { timeout / 1000, (timeout % 1000) * 1000 };
        int selectRC = STAFSocketFuncs::select(fSocket + 1, 0, &writeSocks, 0,
                                               (timeout == 0) ? 0 :
                                                                &theTimeout);
        if (selectRC < 0)
        {
            STAFConnectionConnectException selectError("select",
                                           STAFSocketFuncs::sock_errno());
            STAFSocketFuncs::soclose(fSocket);
            THROW_STAF_EXCEPTION(selectError);
        }
        else if (selectRC == 0)
        {
            STAFConnectionConnectTimeoutException timeoutError("select() "
                                                               "timeout");
            STAFSocketFuncs::soclose(fSocket);
            THROW_STAF_EXCEPTION(timeoutError);
        }

        // On AIX, select() returns successfully if the target system exists.
        // This is true, even if the connection to the particular port is
        // refused.  So, we put in a zero-byte read to make sure that the
        // socket is actually connected successfully.
        // On Windows, we need to check if errno is SOCEWOULDBLOCK
        int recvRC = STAFSocketFuncs::recv(fSocket, 0 , 0, STAF_MSG_NOSIGNAL);

        if ((recvRC < 0) && (STAFSocketFuncs::sock_errno() != SOCEWOULDBLOCK))
        {
            STAFConnectionConnectException recvError("recv",
                                           STAFSocketFuncs::sock_errno());
            STAFSocketFuncs::soclose(fSocket);
            THROW_STAF_EXCEPTION(recvError);
        }
    }

    modeRC = STAFSocketFuncs::setBlockingMode(fSocket,
                                              STAFSocketFuncs::kBlocking);
    if (modeRC != 0)
    {
        STAFConnectionConnectException modeError("setBlockingMode(Blocking)",
                                                 modeRC);
        STAFSocketFuncs::soclose(fSocket);
        THROW_STAF_EXCEPTION(modeError);
    }

}


STAFSocketConnection::STAFSocketConnection(int socket)
    : STAFConnection(kNetwork, kServer), fSocket(socket)
{
    /* Do Nothing */
}


void STAFSocketConnection::read(void *buffer, unsigned int size)
{
    int rc = 0;

    for(unsigned current = 0; current < size; current += rc)
    {
        // XXX: There should probably be a select() call here and in write

        int recvSize = (int)STAF_MIN((size_t)(size - current), 
                                     sizeof(fBuffer));
        rc = STAFSocketFuncs::recv(fSocket, fBuffer, recvSize,
                                   STAF_MSG_NOSIGNAL);
        if (rc < 0)
        {
            STAFConnectionIOException recvError("recv",
                                                STAFSocketFuncs::sock_errno());
            THROW_STAF_EXCEPTION(recvError);
        }
        else if (rc == 0)
        {
            STAFConnectionIOException closedError("Other side closed socket");
            THROW_STAF_EXCEPTION(closedError);
        }
        else
        {
            memcpy((char *)buffer + current, fBuffer, rc);
        }
    }
}


void STAFSocketConnection::write(void *buffer, unsigned int size)
{
    int rc = 0;

    for(unsigned int current = 0; current < size; current += rc)
    {
        int sendSize = (int)STAF_MIN((size_t)(size - current), 
                                     sizeof(fBuffer));
        memcpy(fBuffer, (char *)buffer + current, sendSize);

        rc = STAFSocketFuncs::send(fSocket, (char *)fBuffer, (int)sendSize, 0);

        if (rc < 0)
        {
            STAFConnectionIOException sendError("send",
                                                STAFSocketFuncs::sock_errno());
            THROW_STAF_EXCEPTION(sendError);
        }
    }
}


void STAFSocketConnection::freeSocketConnection(STAFConnection *pConn)
{
    delete pConn;
}


STAFSocketConnection::~STAFSocketConnection()
{
    // Note: "2" means to shutdown both the inbound and outbound traffic
    STAFSocketFuncs::soshutdown(fSocket, 2);
    STAFSocketFuncs::soclose(fSocket);
}
