/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAFOSTypes.h"
#include "STAFSocketFuncs.h"
#include "STAFLocalConnection.h"
#include "STAFException.h"
#include <string.h>
#include <algorithm>
#include <sys/time.h>
#include <sys/un.h>

STAFConnAllocInfo STAFLocalConnection::create(STAFString localName,
                                              unsigned int timeout)
{
    return STAFConnAllocInfo(new STAFLocalConnection(localName, timeout),
                             &freeLocalConnection);
}


STAFConnAllocInfo STAFLocalConnection::create(int socket)
{
    return STAFConnAllocInfo(new STAFLocalConnection(socket),
                             &freeLocalConnection);
}


STAFLocalConnection::STAFLocalConnection(STAFString localName, 
                                         unsigned int timeout)
    : STAFConnection(kLocal, kClient)
{
    // XXX: Major Hack to make sure socket pointers are available on the
    //      client side.  Yuck!
    if (STAFSocketFuncs::bswap == 0) STAFSocketFuncs::initSocketFuncs();

    struct sockaddr_un serverAddress = { 0 };

    serverAddress.sun_family = AF_UNIX;
    strcpy(serverAddress.sun_path, "/tmp/STAFIPC_");
    strcat(serverAddress.sun_path, localName.toCurrentCodePage()->buffer());
    // XXX: Linux doesn't support sun_len.  What to do?
    // serverAddress.sun_len = strlen(serverAddress.sun_path) + 1;

    fSocket = STAFSocketFuncs::socket(PF_UNIX, SOCK_STREAM, 0);
    if (!STAFUtilIsValidSocket(fSocket))
    {
        STAFConnectionConnectException noSocket("socket",
                                                STAFSocketFuncs::sock_errno());
        THROW_STAF_EXCEPTION(noSocket);
    }

    int modeRC = STAFSocketFuncs::setBlockingMode(fSocket,
                 STAFSocketFuncs::kNonBlocking);
    if (modeRC != 0)
    {
        STAFConnectionConnectException modeError("setBlockingMode(NonBlocking)",
                                                 modeRC);
        THROW_STAF_EXCEPTION(modeError);
    }

    int connectRC = STAFSocketFuncs::connect(fSocket,
                    (struct sockaddr *)&serverAddress, sizeof(serverAddress));

    if ((connectRC < 0) && (STAFSocketFuncs::sock_errno() != SOCEINPROGRESS) &&
       (STAFSocketFuncs::sock_errno() != SOCEWOULDBLOCK))
    {
        STAFConnectionConnectException connectError("connect",
                                       STAFSocketFuncs::sock_errno());
        THROW_STAF_EXCEPTION(connectError);
    }

    if (connectRC < 0)
    {
        fd_set writeSocks;
        FD_ZERO(&writeSocks);
        FD_SET(fSocket, &writeSocks);
        timeval theTimeout = { timeout / 1000, (timeout % 1000) * 1000 };
        int selectRC = STAFSocketFuncs::select(fSocket + 1, 0, &writeSocks, 0,
                                               (timeout == 0) ? 0 : &theTimeout);
        if (selectRC < 0)
        {
            STAFSocketFuncs::soclose(fSocket);
            STAFConnectionConnectException selectError("select",
                                           STAFSocketFuncs::sock_errno());
            THROW_STAF_EXCEPTION(selectError);
        }
        else if (selectRC == 0)
        {
            STAFSocketFuncs::soclose(fSocket);
            STAFConnectionConnectTimeoutException timeoutError("select() timeout");
            THROW_STAF_EXCEPTION(timeoutError);
        }
    }

    modeRC = STAFSocketFuncs::setBlockingMode(fSocket,
                                              STAFSocketFuncs::kBlocking);
    if (modeRC != 0)
    {
        STAFSocketFuncs::soclose(fSocket);
        STAFConnectionConnectException modeError("setBlockingMode(Blocking)",
                                                 modeRC);
        THROW_STAF_EXCEPTION(modeError);
    }

}


STAFLocalConnection::STAFLocalConnection(int socket)
    : STAFConnection(kLocal, kServer), fSocket(socket)
{
    /* Do Nothing */
}


void STAFLocalConnection::read(void *buffer, unsigned int size)
{
    int rc = 0;

    for(unsigned int current = 0; current < size; current += rc)
    {
        int recvSize = (int)std::min((size_t)(size - current), sizeof(fBuffer));
        rc = STAFSocketFuncs::recv(fSocket, fBuffer, recvSize,
                                   STAF_MSG_NOSIGNAL);
        if (rc < 0)
        {
            STAFConnectionIOException recvError("recv",
                                                STAFSocketFuncs::sock_errno());
            THROW_STAF_EXCEPTION(recvError);
        }
        else if (rc == 0)
        {
            STAFConnectionIOException closedError("Other side closed socket");
            THROW_STAF_EXCEPTION(closedError);
        }
        else
        {
            memcpy((char *)buffer + current, fBuffer, rc);
        }
    }
}


void STAFLocalConnection::write(void *buffer, unsigned int size)
{
    int rc = 0;

    for(unsigned int current = 0; current < size; current += rc)
    {
        int sendSize = (int)std::min((size_t)(size - current), sizeof(fBuffer));
        memcpy(fBuffer, (char *)buffer + current, sendSize);

        rc = STAFSocketFuncs::send(fSocket, (char *)fBuffer, (int)sendSize, 0);

        if (rc < 0)
        {
            STAFConnectionIOException sendError("send",
                                                STAFSocketFuncs::sock_errno());
            THROW_STAF_EXCEPTION(sendError);
        }
    }
}


void STAFLocalConnection::freeLocalConnection(STAFConnection *pConn)
{
    delete pConn;
}


STAFLocalConnection::~STAFLocalConnection()
{
    STAFSocketFuncs::soclose(fSocket);
}
