/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAFOSTypes.h"
#include "STAF_iostream.h"
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "STAFSocketFuncs.h"
#include "STAFException.h"
#include "STAFString.h"

// Prototypes

static unsigned short staf_bswap(unsigned short);
static unsigned long staf_lswap(unsigned long);
static int staf_sock_errno(void);

// Define Function Pointers

STAFSocket_t (* SYSLINK STAFSocketFuncs::accept)(STAFSocket_t, struct sockaddr *,
                                                 int *) = 0;
int (* SYSLINK STAFSocketFuncs::bind)(STAFSocket_t, struct sockaddr *, int) = 0;
int (* SYSLINK STAFSocketFuncs::connect)(STAFSocket_t, struct sockaddr *, int) = 0;
int (* SYSLINK STAFSocketFuncs::gethostname)(char *, int);
int (* SYSLINK STAFSocketFuncs::listen)(STAFSocket_t, int) = 0;
int (* SYSLINK STAFSocketFuncs::recv)(STAFSocket_t, char *, int, int) = 0;
int (* SYSLINK STAFSocketFuncs::send)(STAFSocket_t, char *, int, int) = 0;
int (* SYSLINK STAFSocketFuncs::select)(int, fd_set *, fd_set *, fd_set *,
                                        struct timeval *) = 0;
int (* SYSLINK STAFSocketFuncs::sock_errno)(void) = 0;
STAFSocket_t (* SYSLINK STAFSocketFuncs::socket)(int, int, int) = 0;
int (* SYSLINK STAFSocketFuncs::soclose)(STAFSocket_t) = 0;
int (* SYSLINK STAFSocketFuncs::soshutdown)(STAFSocket_t, int) = 0;

unsigned short   (* SYSLINK STAFSocketFuncs::bswap)(unsigned short) = 0;
struct hostent * (* SYSLINK STAFSocketFuncs::gethostbyaddr)(const char *, 
                                                            int, int) = 0;
struct hostent * (* SYSLINK STAFSocketFuncs::gethostbyname)(const char *) = 0;
unsigned long    (* SYSLINK STAFSocketFuncs::inet_addr)(const char *) = 0;
unsigned long    (* SYSLINK STAFSocketFuncs::lswap)(unsigned long) = 0;


int STAFSocketFuncs::initSocketFuncs()
{
    STAFSocketFuncs::accept = (STAFSocket_t (*)(STAFSocket_t, struct sockaddr *,
                                                int *))::accept;
    STAFSocketFuncs::bind = (int (*)(STAFSocket_t, struct sockaddr *, int))::bind;
    STAFSocketFuncs::connect = (int (*)(STAFSocket_t, struct sockaddr *,
                                        int))::connect;
    STAFSocketFuncs::gethostname = (int (*)(char *, int))::gethostname;
    STAFSocketFuncs::listen = (int (*)(STAFSocket_t,int))::listen;
    STAFSocketFuncs::recv = (int (*)(STAFSocket_t, char *, int, int))::recv;
    STAFSocketFuncs::select = (int (*)(int, fd_set *, fd_set *, fd_set *,
                                       struct timeval *))::select;
    STAFSocketFuncs::send = (int (*)(STAFSocket_t, char *, int, int))::send;
    STAFSocketFuncs::sock_errno = (int (*)())staf_sock_errno;
    STAFSocketFuncs::socket = (STAFSocket_t (*)(int, int, int))::socket;
    STAFSocketFuncs::soclose = (int (*)(STAFSocket_t))::close;
    STAFSocketFuncs::soshutdown = (int (*)(STAFSocket_t, int))::shutdown;
    STAFSocketFuncs::bswap = (unsigned short (*)(unsigned short))staf_bswap;
    STAFSocketFuncs::gethostbyaddr = (hostent * (*)(const char *, int,
                                                    int))::gethostbyaddr;
    STAFSocketFuncs::gethostbyname = 
                                 (hostent * (*)(const char *))::gethostbyname;
    STAFSocketFuncs::inet_addr = (unsigned long (*)(const char *))::inet_addr;
    STAFSocketFuncs::lswap = (unsigned long (*)(unsigned long))staf_lswap;

    return 0;
}


int STAFSocketFuncs::getRealHostName(STAFString &hostname)
{
    char buffer[256];

    if (STAFSocketFuncs::gethostname(buffer, sizeof(buffer)) == SOCKET_ERROR)
    {
        cout << "Error on gethostname()" << endl;
        return 1;
    }

    struct hostent *host = STAFSocketFuncs::gethostbyname(buffer);

    if (host == 0)
    {
        cout << "Error on gethostbyname()" << endl;
        return 1;
    }

    STAFString hostname1(host->h_name);

    if (hostname1.count(kUTF8_PERIOD) > 2)
    {
        hostname = hostname1;
        return 0;
    }

    int addr = *(int *)host->h_addr;

    host = STAFSocketFuncs::gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);

    if (host == 0)
    {
        cout << "Error on gethostbyaddr()" << endl;
        return 1;
    }

    STAFString hostname2(host->h_name);

    if (hostname2.count(kUTF8_PERIOD) > 2)
    {
        hostname = hostname1;
        return 0;
    }

    if (hostname1.length() > hostname2.length())
        hostname = hostname1;
    else
        hostname = hostname2;

    return 0;
}


unsigned short staf_bswap(unsigned short swap) 
{
    return(htons(swap));
}

unsigned long staf_lswap(unsigned long swap) 
{
    return(htonl(swap));
}

int staf_sock_errno(void) 
{
    return(errno);
}

int STAFSocketFuncs::setBlockingMode(STAFSocket_t theSocket,
                                     STAFSocketFuncs::BlockingMode blockingMode)
{
    int fd_flags = fcntl(theSocket, F_GETFL, 0);

    if (fd_flags == -1)
    {
        cout << "Error getting file descriptor flags, RC: " << errno << endl;
        return errno;
    }

    if (blockingMode == kBlocking)
        fd_flags &= ~O_NONBLOCK;
    else
        fd_flags |= O_NONBLOCK;

    if (fcntl(theSocket, F_SETFL, fd_flags) == -1)
    {
        cout << "Error setting socket flags, RC: " << errno << endl;
        return errno;
    }

    return 0;
}

STAFRC_t STAFSocketFuncs::getInAddrByName(const char *name, in_addr *addr,
                                          unsigned int *osRC)
{
#if defined(STAF_GETHOSTBYNAME_R_6PARM)
    struct hostent hostent_data = { 0 };
    struct hostent *hostname = &hostent_data;
    char hostdata_buff[2048] = { 0 };
    struct hostent *hostname2 = 0;
    int host_error = 0;

    int rc = gethostbyname_r(name, hostname, hostdata_buff,
                             sizeof(hostdata_buff), &hostname2, &host_error);

    if (host_error != 0)
    {
        if (osRC) *osRC = host_error;
        return kSTAFBaseOSError;
    }
#elif defined(STAF_GETHOSTBYNAME_R_5PARM)
    struct hostent hostent_data = { 0 };
    struct hostent *hostname = &hostent_data;
    char hostdata_buff[2048] = { 0 };
    int host_error = 0;

    gethostbyname_r(name, hostname, hostdata_buff, sizeof(hostdata_buff),
                    &host_error);

    if (host_error != 0)
    {
        if (osRC) *osRC = host_error;
        return kSTAFBaseOSError;
    }
#elif defined(STAF_GETHOSTBYNAME_R_3PARM)
    struct hostent_data hostname_hostent_data = { 0 };
    struct hostent hostname_data = { 0 };
    struct hostent *hostname = &hostname_data;
    int rc = gethostbyname_r(name, hostname, &hostname_hostent_data);

    if (rc != 0)
    {
        if (osRC) *osRC = rc;
        return kSTAFBaseOSError;
    }
#elif defined(STAF_NO_GETHOSTBYNAME_R)
    struct hostent *hostname = gethostbyname(name);

    if (hostname == 0)
    {
        if (osRC) *osRC = h_errno;
        return kSTAFBaseOSError;
    }
#endif

    //addr->s_addr = *reinterpret_cast<unsigned long *>(hostname->h_addr);
    addr->s_addr = *reinterpret_cast<unsigned int *>(hostname->h_addr);

    return kSTAFOk;
}
