/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <map>
#include "STAFUtil.h"
#include "STAFLocalConnection.h"
#include "STAFException.h"
#include "STAFMutexSem.h"
#include "STAFUtilWin32.h"
#include <aclapi.h>     // Include for SetSecurityInfo()

#define THROWCONNECT(string) {\
        STAFConnectionConnectException connectError(string, GetLastError());\
        THROW_STAF_EXCEPTION(connectError);}

#define THROWIO(string) {\
        STAFConnectionIOException ioError(string, GetLastError());\
        THROW_STAF_EXCEPTION(ioError);}

// Typedefs for function pointers in ADVAPI32.DLL
typedef BOOL (STDMETHODCALLTYPE FAR * LPFNSETSECURITYINFO) (
   HANDLE hToken,                // user token
   SE_OBJECT_TYPE ObjectType, // type of object 
   SECURITY_INFORMATION SecurityInfo, // type of security information to set 
   PSID psidOwner, // pointer to the new owner SID 
   PSID psidGroup, // pointer to the new primary group SID 
   PACL pDacl, // pointer to the new DACL 
   PACL pSacl // pointer to the new SACL
);

static HMODULE                 sAdvApi32Lib                 = NULL;
static LPFNSETSECURITYINFO     sSetSecurityInfoFunc         = NULL;

// End of definitions for SetSecurityInfo() stuff

STAFConnAllocInfo STAFLocalConnection::create(STAFString pipeName)
{
    return STAFConnAllocInfo(new STAFLocalConnection(pipeName),
                             &freeLocalConnection);
}


STAFConnAllocInfo STAFLocalConnection::create(HANDLE readHandle,
                                              HANDLE writeHandle)
{
    return STAFConnAllocInfo(new STAFLocalConnection(readHandle, writeHandle),
                             &freeLocalConnection);
}

struct PipeData
{
    HANDLE clientMutex;
    HANDLE serverSynchSem;
    HANDLE clientSynchSem;
    DWORD *pidSwap;
    HANDLE *readHandlePtr;
    HANDLE *writeHandlePtr;
    HANDLE serverProcessHandle;
};

typedef std::map<STAFString, PipeData> PipeMap;

static STAFMutexSem sPipeMapSem;
static PipeMap sPipeMap;

//static BOOL InitAdvApi32();

// Note: You should already own sPipeMapSem before calling this function
static void GetPipeData(const STAFString &pipeName, PipeData &pipeData)
{
    PipeMap::iterator iter;

    if ((iter = sPipeMap.find(pipeName)) == sPipeMap.end())
    {
        HANDLE fileMapHandle = 0;
        void *sharedMemAddr = 0;
        DWORD *serverPID = 0;

        SECURITY_ATTRIBUTES sa;
        PSECURITY_DESCRIPTOR sd = 0;

        try
        {
            STAFString globalStr = "";

            if (STAFUtilWin32GetWinType() & kSTAFWin2KPlus)
            {
                // Open Process to get process handle and SetSecurityInfo
                HANDLE pHandle = OpenProcess(PROCESS_ALL_ACCESS, FALSE,
                                             GetCurrentProcessId());
                if (pHandle == 0) THROWCONNECT("OpenProcess1");
        
                int rc = sSetSecurityInfoFunc(pHandle, SE_KERNEL_OBJECT,
                                              DACL_SECURITY_INFORMATION, 
                                              0, 0, 0, 0);
                if (rc != 0)
                {
                    STAFConnectionConnectException connectErr("SetSecurityInfo",
                                                              rc);
                    THROW_STAF_EXCEPTION(connectErr);
                }
  
                // To support Fast User Switching and Terminal Server need
                // to prepend Global\\ for Win2K and above.
                globalStr = "Global\\";
            }

            STAFString sharedMemName = globalStr + pipeName + "/SharedMemory";

            // Obtain Shared Memory
            fileMapHandle = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE,
                            sharedMemName.toCurrentCodePage()->buffer());

            if (fileMapHandle == 0) THROWCONNECT("OpenFileMapping");

            sharedMemAddr = MapViewOfFile(fileMapHandle,
                                          FILE_MAP_ALL_ACCESS, 0, 0, 0);

            if (sharedMemAddr == 0) THROWCONNECT("MapViewOfFile");

            // Set up security attributes
            sa.nLength = sizeof(SECURITY_ATTRIBUTES);
            sa.bInheritHandle = TRUE;
            sa.lpSecurityDescriptor = sd;

            // Create Null Security Descriptor to allow access to everyone
            if (STAFUtilWin32CreateNullSD(&sd))
                THROWCONNECT("STAFUtilWin32CreateNullSD");

            sa.lpSecurityDescriptor = sd;

            // Get initialization semaphores
            STAFString clientMutexName = globalStr + pipeName + "/ClientMutex";
            STAFString serverSynchName = globalStr + pipeName + "/ServerSynch";
            STAFString clientSynchName = globalStr + pipeName + "/ClientSynch";

            pipeData.clientMutex = CreateMutex(&sa, FALSE,
                clientMutexName.toCurrentCodePage()->buffer());
            if (pipeData.clientMutex == 0) THROWCONNECT("CreateMutex");

            pipeData.serverSynchSem = CreateEvent(&sa, TRUE, FALSE, 
                serverSynchName.toCurrentCodePage()->buffer());
            if (pipeData.serverSynchSem == 0) THROWCONNECT("CreateEvent");
                
            pipeData.clientSynchSem = CreateEvent(&sa, TRUE, FALSE, 
                clientSynchName.toCurrentCodePage()->buffer());
            if (pipeData.clientSynchSem == 0) THROWCONNECT("CreateEvent");

            // Set up some pointers into the shared memory area
            pipeData.pidSwap = (DWORD *)sharedMemAddr;
            serverPID = (DWORD *)((char *)sharedMemAddr + sizeof(DWORD));
            pipeData.readHandlePtr = (HANDLE *)((char *)sharedMemAddr +
                                                (2 *sizeof(DWORD)));
            pipeData.writeHandlePtr = (HANDLE *)((char *)sharedMemAddr +
                                                 (2 * sizeof(DWORD)) +
                                                 sizeof(HANDLE));

            // Obtain handle to Factory process
            pipeData.serverProcessHandle = OpenProcess(PROCESS_ALL_ACCESS,
                                                       FALSE, *serverPID);

            if (pipeData.serverProcessHandle == 0)
                THROWCONNECT("OpenProcess2");

            sPipeMap[pipeName] = pipeData;

            if (sd != 0) STAFUtilWin32DeleteNullSD(&sd);
        }
        catch (...)
        {
            // Release shared memory and handles if necessary

            if (sd != 0) STAFUtilWin32DeleteNullSD(&sd);
            if (sharedMemAddr != 0) UnmapViewOfFile(sharedMemAddr);
            if (fileMapHandle != 0) CloseHandle(fileMapHandle);
            if (pipeData.clientMutex != 0)
                CloseHandle(pipeData.clientMutex);
            if (pipeData.serverSynchSem != 0)
                CloseHandle(pipeData.serverSynchSem);
            if (pipeData.clientSynchSem != 0)
                CloseHandle(pipeData.clientSynchSem);

            throw;
        }
    }
    else pipeData = iter->second;
}


// True = successfully connected, False = server process disappeared
// Note: This function can still throw exceptions
static bool ConnectToPipe(PipeData &pipeData, HANDLE &readHandle,
                          HANDLE &writeHandle)
{
    // Obtain the client mutex on this "named pipe", while we swap
    // process information with the STAFLocalConnectionFactory
    WaitForSingleObject(pipeData.clientMutex, INFINITE);

    *pipeData.pidSwap = GetCurrentProcessId();

    ResetEvent(pipeData.clientSynchSem);
    SetEvent(pipeData.serverSynchSem);

    HANDLE waitArray[2] = { pipeData.clientSynchSem,
                            pipeData.serverProcessHandle };
    DWORD rc = WaitForMultipleObjects(2, waitArray, FALSE, INFINITE);
    unsigned int signaledHandleIndex = 0;

    if ((rc >= WAIT_OBJECT_0) && (rc <= (WAIT_OBJECT_0 + 2 - 1)))
    {
        signaledHandleIndex = (unsigned int)(rc - WAIT_OBJECT_0);
    }
    else if ((rc >= WAIT_ABANDONED_0) &&
             (rc <= (WAIT_ABANDONED_0 + 2 - 1)))
    {
        signaledHandleIndex = (unsigned int)(rc - WAIT_ABANDONED_0);
    }
    else
    {
        // We either got WAIT_FAILED, WAIT_TIMEOUT or some other return
        // code which we weren't expecting, so spit out an error message
        // and continue

        ReleaseMutex(pipeData.clientMutex);

        STAFConnectionConnectException connectError("Other side closed "
                                                    "connection", rc);
        THROW_STAF_EXCEPTION(connectError);
    }

    if (signaledHandleIndex != 0)
    {
        ReleaseMutex(pipeData.clientMutex);

        return false;
    }

    if (pipeData.pidSwap == 0)
    {
        ReleaseMutex(pipeData.clientMutex);

        STAFConnectionConnectException connectError("Other side closed "
                                                    "connection");
        THROW_STAF_EXCEPTION(connectError);
    }

    readHandle = *pipeData.readHandlePtr;
    writeHandle = *pipeData.writeHandlePtr;

    ReleaseMutex(pipeData.clientMutex);

    return true;
}


STAFLocalConnection::STAFLocalConnection(STAFString pipeName)
    : STAFConnection(kLocal, kClient)
{
    InitConnection();
    
    PipeData pipeData = { 0 };

    {
        STAFMutexSemLock lock(sPipeMapSem);
        GetPipeData(pipeName, pipeData);
    }

    bool connected = ConnectToPipe(pipeData, fReadHandle, fWriteHandle);

    if (!connected)
    {
        // This case means that the server we were attached to has disappeared.
        // So, let's flush the cache and try again.

        {
            STAFMutexSemLock lock(sPipeMapSem);

            sPipeMap.erase(pipeName);

            GetPipeData(pipeName, pipeData);
        }

        connected = ConnectToPipe(pipeData, fReadHandle, fWriteHandle);

        // If the problem still exists then go ahead and throw the error

        if (!connected)
        {
            STAFConnectionConnectException connectError("Other side closed "
                                                        "connection (process "
                                                        "ended)");
            THROW_STAF_EXCEPTION(connectError);
        }
    }
}


STAFLocalConnection::STAFLocalConnection(HANDLE readHandle, HANDLE writeHandle)
                  : STAFConnection(kLocal, kServer), fReadHandle(readHandle),
                    fWriteHandle(writeHandle)
{
    /* Do Nothing */
}


void STAFLocalConnection::read(void *buffer, unsigned int size)
{
    BOOL rc = TRUE;
    DWORD actual = 0;

    for(DWORD current = 0; current < size; current += actual)
    {
        rc = ReadFile(fReadHandle, (LPVOID)((char *)buffer + current),
                      (DWORD)(size - current), &actual, 0);

        if (rc == FALSE) THROWIO("ReadFile");

        if ((actual == 0) && (size != 0))
        {
            STAFConnectionIOException noData("No data available on pipe");
            THROW_STAF_EXCEPTION(noData);
        }
    }
}


void STAFLocalConnection::write(void *buffer, unsigned int size)
{
    BOOL rc = TRUE;
    DWORD actual = 0;

    for(DWORD current = 0; current < size; current += actual)
    {
        rc = WriteFile(fWriteHandle, (LPVOID)((char *)buffer + current),
                      (DWORD)(size - current), &actual, 0);

        if (rc == FALSE) THROWIO("WriteFile");

        if ((actual == 0) && (size != 0))
        {
            STAFConnectionIOException noData("No data written to pipe");
            THROW_STAF_EXCEPTION(noData);
        }
    }
}


void STAFLocalConnection::freeLocalConnection(STAFConnection *pConn)
{   
    delete pConn;
}


void STAFLocalConnection::InitConnection()
{
    static STAFMutexSem initSem;
    static bool alreadyInited = false;

    if (alreadyInited) return;

    STAFMutexSemLock initLock(initSem);

    if (alreadyInited) return;
    
    if (STAFUtilWin32GetWinType() & kSTAFWinNTPlus)
    {
        // Set ADVAPI32.DLL function pointers for SetSecurityInfo()
        // if running on Windows NT/2000/XP or above.

        int rc = InitAdvApi32();

        if (rc != TRUE)
        {
            STAFConnectionIOException initAdvApi32(
                "InitAdvApi32() failed in InitConnection()");
            THROW_STAF_EXCEPTION(initAdvApi32);
        }
    }
    
    alreadyInited = true;
}


/*******************************************************************************/
/* InitAdvApi32 - This function dynamically links to ADVAPI32.DLL and sets up  */
/*                the required function pointers                               */
/*                                                                             */
/* Accepts: Nothing                                                            */
/*                                                                             */
/* Returns: TRUE , if successful                                               */
/*          FALSE, otherwise                                                   */
/*******************************************************************************/
BOOL STAFLocalConnection::InitAdvApi32()
{
   sAdvApi32Lib = LoadLibrary("advapi32.dll");

   if (!sAdvApi32Lib)
   {
       STAFConnectionIOException loadLibrary(
           "LoadLibrary(advapi32.dll) in InitAdvApi32()");
       THROW_STAF_EXCEPTION(loadLibrary);
   }

   sSetSecurityInfoFunc = (LPFNSETSECURITYINFO)GetProcAddress(
                           sAdvApi32Lib, "SetSecurityInfo" );
   if (!sSetSecurityInfoFunc)
   {
       STAFConnectionIOException getProcAddress(
           "GetProcAddress(sSetSecurityInfoFunc) in InitAdvApi32()");
       THROW_STAF_EXCEPTION(getProcAddress);
   }

   return TRUE;
}


STAFLocalConnection::~STAFLocalConnection()
{   
    CloseHandle(fReadHandle);
    CloseHandle(fWriteHandle);
}
