/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFUtil.h"
#include "STAFString.h"
#include "STAFException.h"
#include "STAFLocalConnectionFactory.h"
#include "STAFUtilWin32.h"
#include <aclapi.h>          // Include for SetSecurityInfo()

#define THROWCONNECT(string) {\
        STAFConnectionConnectException connectError(string, GetLastError());\
        THROW_STAF_EXCEPTION(connectError);}

// Typedefs for function pointers in ADVAPI32.DLL
typedef BOOL (STDMETHODCALLTYPE FAR * LPFNSETSECURITYINFO) (
   HANDLE hToken,                // user token
   SE_OBJECT_TYPE ObjectType, // type of object 
   SECURITY_INFORMATION SecurityInfo, // type of security information to set 
   PSID psidOwner, // pointer to the new owner SID 
   PSID psidGroup, // pointer to the new primary group SID 
   PACL pDacl, // pointer to the new DACL 
   PACL pSacl // pointer to the new SACL
);

static HMODULE                 sAdvApi32Lib                 = NULL;
static LPFNSETSECURITYINFO     sSetSecurityInfoFunc         = NULL;

// End of definitions for SetSecurityInfo() stuff

// Note, that between the Factory constructor and the acceptConnection()
// method, there is always one outstanding Pipe.  This is so that a
// client can never try to open a non-existant pipe.  In other words, all
// of the servers pipes will not be closed at the same time.

STAFLocalConnectionFactory::STAFLocalConnectionFactory(STAFString localName,
    unsigned char instanceCount) : fLocalName(localName), fClientMutex(),
    fInstanceCount(instanceCount)
{
    SECURITY_ATTRIBUTES sa;
    PSECURITY_DESCRIPTOR sd = 0;

    InitConnectionFactory();
    
    try
    {
        fFileMapHandle = 0;
        fSharedMemAddr = 0;

        STAFString globalStr = "";

        // To support Fast User Switching/Terminal Server need Global\\
        // prepended for WinXP, Win2K and above.
        if (STAFUtilWin32GetWinType() & kSTAFWin2KPlus)
            globalStr = "Global\\";

        STAFString sharedMemoryName = globalStr + fLocalName + "/SharedMemory";

        // Set up security attributes 
        sa.nLength = sizeof(SECURITY_ATTRIBUTES);
        sa.bInheritHandle = TRUE;
        sa.lpSecurityDescriptor = sd;

        // Create a Null Security Descriptor to allow access to everyone.
        if (STAFUtilWin32CreateNullSD(&sd))
            THROWCONNECT("STAFUtilWin32CreateNullSD");
        
        sa.lpSecurityDescriptor = sd;
        
        // Create events for server and client passing in security attributes
        fServerSynch = CreateEvent(&sa, TRUE, FALSE, STAFString(globalStr +
            localName + "/ServerSynch").toCurrentCodePage()->buffer());
        fClientSynch = CreateEvent(&sa, TRUE, FALSE, STAFString(globalStr +
            localName + "/ClientSynch").toCurrentCodePage()->buffer());
       
        // Obtain Shared Memory with security
        fFileMapHandle = CreateFileMapping(INVALID_HANDLE_VALUE, &sa,
                         PAGE_READWRITE, 0,
                         2 * sizeof(DWORD) + 2 * sizeof(HANDLE),
                         sharedMemoryName.toCurrentCodePage()->buffer());

        if (fFileMapHandle == 0) THROWCONNECT("CreateFileMapping");

        fSharedMemAddr = MapViewOfFile(fFileMapHandle, FILE_MAP_ALL_ACCESS, 0,
                                       0, 0);
        if (fSharedMemAddr == 0) THROWCONNECT("MapViewOfFile");

        DWORD *serverPID = (DWORD *)((char *)fSharedMemAddr + sizeof(DWORD));
        *serverPID = GetCurrentProcessId();

        if (STAFUtilWin32GetWinType() & kSTAFWin2KPlus)
        {
            // Open the Process to get its handle to pass to SetSecurityInfo
            HANDLE pHandle = OpenProcess(PROCESS_ALL_ACCESS, FALSE,
                                         *serverPID);
            if (pHandle == 0) THROWCONNECT("OpenProcess1");
           
            int rc = sSetSecurityInfoFunc(pHandle, SE_KERNEL_OBJECT,
                                          DACL_SECURITY_INFORMATION, 
                                          0, 0, 0, 0);

            if (rc != 0)
            {
                STAFConnectionConnectException connectErr(
                    "SetSecurityInfo", rc);
                THROW_STAF_EXCEPTION(connectErr);
            }
        }
            
        if (sd != 0) STAFUtilWin32DeleteNullSD(&sd);
    }
    catch (...)
    {
        if (sd != 0) STAFUtilWin32DeleteNullSD(&sd);
        if (fSharedMemAddr != 0) UnmapViewOfFile(fSharedMemAddr);
        if (fFileMapHandle != 0) CloseHandle(fFileMapHandle);

        throw;
    }
}


STAFConnAllocInfo STAFLocalConnectionFactory::acceptConnection()
{
    HANDLE readHandle = 0;
    HANDLE clientReadHandle = 0;
    HANDLE writeHandle = 0;
    HANDLE clientWriteHandle = 0;
    HANDLE clientProcess = 0;

    // Set up variables into shared memory
    DWORD *pidSwap = (DWORD *)fSharedMemAddr;
    HANDLE *readHandlePtr = (HANDLE *)((char *)fSharedMemAddr +
                                       (2 * sizeof(DWORD)));
    HANDLE *writeHandlePtr = (HANDLE *)((char *)fSharedMemAddr +
                                        (2 * sizeof(DWORD)) + sizeof(HANDLE));

    try
    {
        WaitForMultipleObjects(1, &fServerSynch, TRUE, INFINITE);
        // XXX: Temp until auto-reset event sem done
        ResetEvent(fServerSynch);

        clientProcess = OpenProcess(PROCESS_ALL_ACCESS, FALSE,
                                    *pidSwap);

        if (clientProcess == 0) THROWCONNECT("OpenProcess2");

        // Create ServerRead/ClientWrite pipe
        if (!CreatePipe(&readHandle, &clientWriteHandle, 0, 0x00000400))
            THROWCONNECT("CreatePipe");

        // Create ServerWrite/ClientRead pipe
        if (!CreatePipe(&clientReadHandle, &writeHandle, 0, 0x00000400))
            THROWCONNECT("CreatePipe");

        if (!DuplicateHandle(GetCurrentProcess(), clientReadHandle,
                             clientProcess, readHandlePtr, 0, FALSE,
                             DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE))
        {
            THROWCONNECT("DuplicateHandle(readHandle)");
        }

        if (!DuplicateHandle(GetCurrentProcess(), clientWriteHandle,
                             clientProcess, writeHandlePtr, 0, FALSE,
                             DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE))
        {
            THROWCONNECT("DuplicateHandle(writeHandle)");
        }

        CloseHandle(clientProcess);
        SetEvent(fClientSynch);
    }
    catch (...)
    {
        // Close any open handles
        if (clientWriteHandle != 0) CloseHandle(clientWriteHandle);
        if (writeHandle != 0) CloseHandle(writeHandle);
        if (clientReadHandle != 0) CloseHandle(clientReadHandle);
        if (readHandle != 0) CloseHandle(readHandle);
        if (clientProcess != 0) CloseHandle(clientProcess);

        // Inform the client that an error occurred
        *pidSwap = 0;
        SetEvent(fClientSynch);

        throw;
    }

    return STAFLocalConnection::create(readHandle, writeHandle);
}


void STAFLocalConnectionFactory::InitConnectionFactory()
{
    static STAFMutexSem initSem;
    static bool alreadyInited = false;

    if (alreadyInited) return;

    STAFMutexSemLock initLock(initSem);

    if (alreadyInited) return;
    
    if (STAFUtilWin32GetWinType() & kSTAFWinNTPlus)
    {
        // Set ADVAPI32.DLL function pointers for SetSecurityInfo()
        // if running on Windows NT/2000/XP or above.

        int rc = InitAdvApi32();

        if (rc != TRUE)
        {
            STAFConnectionIOException initAdvApi32(
                "InitAdvApi32() failed in InitConnection()");
            THROW_STAF_EXCEPTION(initAdvApi32);
        }
    }
    
    alreadyInited = true;
}


/*******************************************************************************/
/* InitAdvApi32 - This function dynamically links to ADVAPI32.DLL and sets up  */
/*                the required function pointers                               */
/*                                                                             */
/* Accepts: Nothing                                                            */
/*                                                                             */
/* Returns: TRUE , if successful                                               */
/*          FALSE, otherwise                                                   */
/*******************************************************************************/
BOOL STAFLocalConnectionFactory::InitAdvApi32()
{
   sAdvApi32Lib = LoadLibrary("advapi32.dll");

   if (!sAdvApi32Lib)
   {
       STAFConnectionIOException loadLibrary(
           "LoadLibrary(advapi32.dll) in InitAdvApi32()");
       THROW_STAF_EXCEPTION(loadLibrary);
   }

   sSetSecurityInfoFunc = (LPFNSETSECURITYINFO)GetProcAddress(
                            sAdvApi32Lib, "SetSecurityInfo" );
   if (!sSetSecurityInfoFunc)
   {
       STAFConnectionIOException getProcAddress(
           "GetProcAddress(sSetSecurityInfoFunc) in InitAdvApi32()");
       THROW_STAF_EXCEPTION(getProcAddress);
   }

   return TRUE;
}


STAFLocalConnectionFactory::~STAFLocalConnectionFactory()
{
    UnmapViewOfFile(fSharedMemAddr);
    CloseHandle(fFileMapHandle);
}

