/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAF_iostream.h"
#include "STAF_fstream.h"
#include "STAFString.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFConfig.h"
#include "STAFUtil.h"
#include "STAFCommandParser.h"
#include "STAFException.h"
#include "STAFDelegatedService.h"
#include "STAFExternalService.h"
#include "STAFNotificationList.h"
#include "STAFTrustManager.h"
#include "STAFDiagManager.h"
#include "STAFServiceManager.h"
#include "STAFThreadManager.h"
#include "STAFProcessService.h"
#include "STAFTrace.h"

unsigned int handleVarConfig(STAFString line);
unsigned int handleMachineConfig(STAFString line);
unsigned int handleInterfaceConfig(STAFString line);
unsigned int handleTrustConfig(STAFString line);
unsigned int handleNotifyConfig(STAFString line);
unsigned int handleTraceConfig(STAFString line);
unsigned int handleServiceConfig(STAFString line);
unsigned int handleServiceLoaderConfig(STAFString line);
unsigned int handleSetConfig(STAFString line);

unsigned int fSLSNum = 1;
const unsigned int maxLineLength = 2048;

unsigned int readConfigFile(const STAFString &filename)
{
    unsigned int rc = 0;

    ifstream config(filename.toCurrentCodePage()->buffer());

    if (!config)
    {
        cout << "Could not find config file " << filename << endl;
        return 1;
    }

    STAFString whiteSpace("\x20\x09\x0A\x0D");
    STAFString line;
    char configLine[maxLineLength];

    for(; config.good() && !rc;)
    {
        configLine[0] = 0;
        config.getline(configLine, maxLineLength);

        if (!config.good())
        {
            if (!config.eof())
            {
                cout << "Line in configuration file exceeds maximum length of "
                     << STAFString(maxLineLength) << " characters" << endl;
                rc = 1;
                break;
            }

            if (configLine == "") break;
        }

        STAFString thisLine(configLine);

        if ((thisLine == "") || 
            (thisLine.findFirstNotOf(whiteSpace) == STAFString::kNPos) ||
            (thisLine.subString(0, 1) == kUTF8_POUND))
        {
            if (line.length() != 0)
            {
                cout << "Invalid continuation line, " << thisLine << endl;
                rc = 1;
            }

            continue;
        }
        else if (thisLine.subString(thisLine.length() - 1, 1) == kUTF8_BSLASH)
        {
            line += thisLine.subString(0,
                    thisLine.length() - 1).strip(STAFString::kFront);
            continue;
        }
        else
        {
            line += thisLine.strip(STAFString::kFront);
        }

        STAFString firstLowerWord = line.subWord(0, 1).lowerCase();

        if (firstLowerWord == "var") 
            rc = handleVarConfig(line);
        else if (firstLowerWord == "machine") 
            rc = handleMachineConfig(line);
        else if (firstLowerWord == "interface") 
            rc = handleInterfaceConfig(line);
        else if (firstLowerWord == "trust") 
            rc = handleTrustConfig(line);
        else if (firstLowerWord == "notify") 
            rc = handleNotifyConfig(line);
        else if (firstLowerWord == "trace") 
            rc = handleTraceConfig(line);
        else if (firstLowerWord == "service") 
            rc = handleServiceConfig(line);
        else if (firstLowerWord == "serviceloader") 
            rc = handleServiceLoaderConfig(line);
        else if (firstLowerWord == "set") 
            rc = handleSetConfig(line);
        else
        {
            cout << "Unknown config command, " << line << endl;
            rc = 1;
        }

        // Reset line for the next statement
        line = STAFString();
    }

    return rc;
}

unsigned int handleTrustConfig(STAFString line)
{
    STAFCommandParser fTrustParser;

    // trust config options
 
    fTrustParser.addOption("TRUST",   1,
        STAFCommandParser::kValueNotAllowed);
    fTrustParser.addOption("LEVEL",   1,
        STAFCommandParser::kValueRequired);
    fTrustParser.addOption("MACHINE", 0,
        STAFCommandParser::kValueRequired);
    fTrustParser.addOption("DEFAULT", 1,
        STAFCommandParser::kValueNotAllowed);
 
    // trust config groups
 
    fTrustParser.addOptionGroup("LEVEL",           1, 1);
    fTrustParser.addOptionGroup("MACHINE DEFAULT", 1, 1);

    STAFCommandParseResultPtr parsedResult = fTrustParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on TRUST definition, " << line << endl 
             << parsedResult->errorBuffer << endl;
 
        return 1;
    }                                      

    STAFString unresLevelString = parsedResult->optionValue("LEVEL");
    STAFString levelString;

    STAFRC_t rc = (*gGlobalVariablePoolPtr)->resolve(unresLevelString, 0,
                                                          levelString);
    if (rc != kSTAFOk)
    {
        cout << "Error on TRUST definition, " << line << endl
             << "Error resolving " << unresLevelString << ", RC: " << rc 
             << endl;
        return 1;
    }

    if (!levelString.isDigits())
    {
        cout << "Error on TRUST definition, " << line << endl
             << "The <Level> must be a non-negative integer" << endl;
        return 1;
    }

    unsigned int level = levelString.asUInt();

    if (parsedResult->optionTimes("DEFAULT") != 0)
    {
        rc = gTrustManagerPtr->setDefaultTrusteeLevel(level);

        if (rc != kSTAFOk)
        {
            cout << "Error setting DEFAULT TRUST definition, " << line 
                 << endl << "Error code: " << rc << endl;
            return 1;
        }
    }
    else
    {
        for(int i = 1; i <= parsedResult->optionTimes("MACHINE"); ++i)
        {
            STAFString unresMachine = parsedResult->optionValue("MACHINE", i);
            STAFString machine;

            rc = (*gGlobalVariablePoolPtr)->resolve(unresMachine, 0, machine);

            if (rc != kSTAFOk)
            {
                cout << "Error on TRUST definition, " << line << endl
                     << "Error resolving " << unresMachine << ", RC: " 
                     << rc << endl;
                return 1;
            }

            rc = gTrustManagerPtr->setTrusteeLevel(machine, level);

            if (rc != kSTAFOk)
            {
                cout << "Error adding TRUST definition, " << line 
                     << endl << "Error code: " << rc << endl;
                return 1;
            }
        }
    }

    return 0;
}


unsigned int handleInterfaceConfig(STAFString line)
{
    STAFCommandParser fInterfaceParser;

    // interface config options

    fInterfaceParser.addOption("INTERFACE", 1,
        STAFCommandParser::kValueNotAllowed);
    fInterfaceParser.addOption("TCPIP",     1,
        STAFCommandParser::kValueAllowed);

    // interface config groups
 
    fInterfaceParser.addOptionGroup("TCPIP", 1, 1);

    STAFCommandParseResultPtr parsedResult = fInterfaceParser.parse(line);

    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    if (gSpecifiedMachineConfig)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << "You may not specify an INTERFACE definition and a "
             << "MACHINE definition" << endl;
        return 1;
    }

    // When we offer other communication interfaces, we will have
    // to differintiate between them here

    gUseTCPIPInterface = 1;

    STAFString unresPort = parsedResult->optionValue("TCPIP");
    STAFString port;

    STAFRC_t rc = (*gGlobalVariablePoolPtr)->resolve(unresPort, 0, port);

    if (rc != kSTAFOk)
    {
        cout << "Error on INTERFACE definition, " << line << endl
             << "Error resolving " << unresPort << ", RC: " << rc
             << endl;
        return 1;
    }

    if (port != "")
    {
        if (!port.isDigits())
        {
            cout << "Error on INTERFACE definition, " << line << endl
                 << "The <Port> must be a non-negative integer" 
                 << endl;
            return 1;
        }
        else gTCPIPPort = (unsigned short)port.asUInt();
    }

    return 0;
}


unsigned int handleVarConfig(STAFString line)
{
    STAFCommandParser fVarParser;

    // var config options
 
    fVarParser.addOption("VAR", 1,
        STAFCommandParser::kValueRequired);

    STAFCommandParseResultPtr parsedResult = fVarParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on variable definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;

    STAFString nameAndValue(parsedResult->optionValue("VAR"));
    unsigned int equalPos = nameAndValue.find(kUTF8_EQUAL);

    if (equalPos == STAFString::kNPos)
    {
        cout << "Error on variable definition, " << line << endl
             << errorBuffer << endl;
        return 1;
    }

    (*gGlobalVariablePoolPtr)->set(nameAndValue.subString(0, equalPos),
                                   nameAndValue.subString(equalPos +
                                   nameAndValue.sizeOfChar(equalPos)));

    return 0;
}


unsigned int handleMachineConfig(STAFString line)
{
    STAFCommandParser fMachineParser;

    // machine config options
 
    fMachineParser.addOption("MACHINE", 1,
        STAFCommandParser::kValueRequired);

    STAFCommandParseResultPtr parsedResult = fMachineParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on MACHINE definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    if (gUseTCPIPInterface)
    {
        cout << "Error on MACHINE definition, " << line << endl
             << "You may not specify a MACHINE definition and an "
             << "INTERFACE definition" << endl;
        return 1;
    }

    STAFString unresMachine = parsedResult->optionValue("MACHINE");
    STAFString machine;

    STAFRC_t rc = (*gGlobalVariablePoolPtr)->resolve(unresMachine, 0,
                                                          machine);

    if (rc != kSTAFOk)
    {
        cout << "Error on MACHINE definition, " << line << endl
             << "Error resolving " << unresMachine << ", RC: " << rc
             << endl;
        return 1;
    }

    gSpecifiedMachineConfig = 1;
    *gMachineNamePtr = machine;

    return 0;
}


unsigned int handleNotifyConfig(STAFString line)
{
    STAFCommandParser fNotifyParser;

    // notify config options
 
    fNotifyParser.addOption("NOTIFY",     1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("ONSTART",    1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("ONSHUTDOWN", 1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyParser.addOption("NAME",       1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("HANDLE",     1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("MACHINE",    1,
        STAFCommandParser::kValueRequired);
    fNotifyParser.addOption("PRIORITY",   1,
        STAFCommandParser::kValueRequired);
 
    // notify config groups
 
    fNotifyParser.addOptionGroup("NAME HANDLE",        1, 1);
    fNotifyParser.addOptionGroup("ONSTART ONSHUTDOWN", 1, 1);
    fNotifyParser.addOptionGroup("MACHINE",            1, 1);
    fNotifyParser.addOptionGroup("NOTIFY",             1, 1);

    STAFCommandParseResultPtr parsedResult = fNotifyParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {    
        cout << "Error on NOTIFY definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;

    unsigned int onStart = parsedResult->optionTimes("ONSTART");
    unsigned int notifyName = parsedResult->optionTimes("NAME");
    unsigned int priority = 5;
    unsigned int handle = 0;
    STAFString name;
    STAFString machine;
    STAFRC_t rc = resolveUIntIfExists(parsedResult, "PRIORITY", 0, 
                                           priority, errorBuffer);

    if (!rc) rc = resolveUIntIfExists(parsedResult, "HANDLE", 0, handle, 
                  errorBuffer);

    if (!rc) rc = resolveStringIfExists(parsedResult, "NAME", 0, name, 
                  errorBuffer);
    if (!rc)
    {
        rc = resolveString(parsedResult->optionValue("MACHINE"), 0, machine,
                           errorBuffer);
    }

    if (rc != kSTAFOk)
    {
        cout << "Error on NOTIFY definition, " << line << endl
             << "Error with, " << errorBuffer << ", RC: " << rc << endl;
        return 1;
    }

    STAFNotificationList *notificationList = gNotifyOnStartPtr;

    if (!onStart) notificationList = gNotifyOnShutdownPtr;

    if (notifyName) rc = notificationList->reg(machine, name, priority);
    else rc = notificationList->reg(machine, handle, priority);

    return rc;
}


unsigned int handleTraceConfig(STAFString line)
{
    STAFCommandParser fTraceParser;

    // trace config options
 
    fTraceParser.addOption("TRACE",               1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ON",                  1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICES",            1, 
        STAFCommandParser::kValueAllowed);
    fTraceParser.addOption("TO",                  1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("STDOUT",              1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("STDERR",              1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("FILE",                1,
        STAFCommandParser::kValueRequired);
    fTraceParser.addOption("SERVICEREQUEST",      1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICEACCESSDENIED", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICEERROR",        1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICERESULT",       1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICEMANAGEMENT",   1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("REGISTRATION",        1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("REMOTEREQUESTS",      1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ERROR",               1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("WARNING",             1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("INFO",                1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("DEPRECATED",          1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ALL",                 1,
        STAFCommandParser::kValueNotAllowed);

    // trace config groups
 
    fTraceParser.addOptionGroup("TRACE",              1, 1);
    fTraceParser.addOptionGroup("TO ON SERVICES",     1, 1);
    fTraceParser.addOptionGroup("STDOUT STDERR FILE", 0, 1);
 
    // trace config needs
 
    fTraceParser.addOptionNeed("STDOUT STDERR FILE", "TO");
    fTraceParser.addOptionNeed("SERVICEREQUEST SERVICEACCESSDENIED "
        "SERVICEERROR SERVICERESULT SERVICEMANAGEMENT REGISTRATION "
        "REMOTEREQUESTS ERROR WARNING INFO DEPRECATED", "ON");
    fTraceParser.addOptionNeed("ALL", "ON SERVICES");
    fTraceParser.addOptionNeed("ON", "SERVICEREQUEST "
        "SERVICEACCESSDENIED SERVICEERROR SERVICERESULT "
        "SERVICEMANAGEMENT REGISTRATION REMOTEREQUESTS ERROR WARNING "
        "INFO DEPRECATED ALL");

    STAFCommandParseResultPtr parsedResult = fTraceParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {  
        cout << "Error on TRACE definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;

    STAFString filename;
    STAFRC_t rc = resolveStringIfExists(parsedResult, "FILE", 0, filename,
                                             errorBuffer);

    if (rc != kSTAFOk)
    {
        cout << "Error on TRACE definition, " << line << endl
             << "Error with, " << errorBuffer << ", RC: " << rc << endl;
        return 1;
    }

    if (parsedResult->optionTimes("TO") != 0)
    {
        if (parsedResult->optionTimes("STDOUT"))
            STAFTrace::setTraceDestination(kSTAFTraceToStdout);
        else if (parsedResult->optionTimes("STDERR"))
            STAFTrace::setTraceDestination(kSTAFTraceToStderr);
        else
            STAFTrace::setTraceDestination(kSTAFTraceToFile, filename);
    }
    else if (parsedResult->optionTimes("ON") != 0)
    {
        if (parsedResult->optionTimes("SERVICEREQUEST") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceRequest);
        if (parsedResult->optionTimes("SERVICEACCESSDENIED") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceAccessDenied);
        if (parsedResult->optionTimes("SERVICEERROR") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceError);
        if (parsedResult->optionTimes("SERVICERESULT") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceResult);
        if (parsedResult->optionTimes("SERVICEMANAGEMENT") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceManagement);
        if (parsedResult->optionTimes("REGISTRATION") != 0)
            STAFTrace::traceOn(kSTAFTraceRegistration);
        if (parsedResult->optionTimes("REMOTEREQUESTS") != 0)
            STAFTrace::traceOn(kSTAFTraceRemoteRequests);
        if (parsedResult->optionTimes("ERROR") != 0)
            STAFTrace::traceOn(kSTAFTraceError);
        if (parsedResult->optionTimes("WARNING") != 0)
            STAFTrace::traceOn(kSTAFTraceWarning);
        if (parsedResult->optionTimes("INFO") != 0)
            STAFTrace::traceOn(kSTAFTraceInfo);
        if (parsedResult->optionTimes("DEPRECATED") != 0)
            STAFTrace::traceOn(kSTAFTraceDeprecated);
        if (parsedResult->optionTimes("ALL") != 0)
            STAFTrace::traceOn(kSTAFTraceAll);
    }
    else
    {
        if (parsedResult->optionTimes("ALL") != 0)
        {
            STAFServiceManager::traceAllServices();
        }
        else
        {
            STAFServiceManager::traceServices(
                parsedResult->optionValue("SERVICES"));
        }
    }

    return rc;
}


unsigned int handleServiceConfig(STAFString line)
{
    STAFCommandParser fServiceParser;

    // SERVICE <Name>
    // LIBRARY <Library> [EXECUTE <Exec>] {OPTION <OpName[=OpValue]>}
    //     [PARMS <Parms>]
    // or
    // DELEGATE <Machine> [TONAME <Service Name>]

    // service config options
 
    fServiceParser.addOption("SERVICE",  1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("LIBRARY",    1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("EXECUTE",    1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("OPTION",    0,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("PARMS",    1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("DELEGATE", 1,
        STAFCommandParser::kValueRequired);
    fServiceParser.addOption("TONAME",   1,
        STAFCommandParser::kValueRequired);
 
    // service config groups
 
    fServiceParser.addOptionGroup("SERVICE",                    1, 1);
    fServiceParser.addOptionGroup("LIBRARY DELEGATE", 1, 1);
 
    // service config needs
 
    fServiceParser.addOptionNeed("EXECUTE OPTION PARMS",  "LIBRARY");
    fServiceParser.addOptionNeed("TONAME", "DELEGATE");
                  
    STAFCommandParseResultPtr parsedResult = fServiceParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on Service definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;
    STAFString name;
    STAFString unresValue = parsedResult->optionValue("SERVICE");
    STAFRC_t rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, name);

    if (rc)
    {
        cout << "Error on Service definition, " << line << endl
             << "Error resolving " << unresValue << ", RC: " << rc << endl;
        return 1;
    }

    try
    {
        STAFServicePtr service;

        if (parsedResult->optionTimes("LIBRARY") != 0)
        {
            // Setup external service data
            STAFString library;
            STAFString exec;
            STAFString parms;
            STAFExternalService::OptionList options;
            
            unresValue = parsedResult->optionValue("LIBRARY");
            rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, library);

            if (!rc)
            {            
                unresValue = parsedResult->optionValue("EXECUTE");
                rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, exec);
            }

            if (!rc)
            {            
                unresValue = parsedResult->optionValue("PARMS");
                rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, parms);
            }

            for (int i = 1; !rc && (i <= parsedResult->optionTimes("OPTION"));
                 ++i)
            {
                STAFString option;
                unresValue = parsedResult->optionValue("OPTION", i);
                rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, option);
                options.push_back(option);
            }

           if (rc)
           {
               cout << "Error on Service definition, " << line << endl
                    << "Error resolving " << unresValue << ", RC: " << rc
                    << endl;
               return 1;
           }

           service = STAFServicePtr(new STAFExternalService(name, library,
                                    exec, options, parms),
                                    STAFServicePtr::INIT);
        }
        else
        {
            // This is a Delegated service

            STAFString toMachine;
            STAFString toName = name;

            unresValue = parsedResult->optionValue("DELEGATE");
            rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, toMachine);

            if (!rc && (parsedResult->optionTimes("TONAME") != 0))
            {
                unresValue = parsedResult->optionValue("TONAME");

                rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, toName);
            }

            if (rc)
            {
                cout << "Error on Service definition, " << line << endl
                     << "Error resolving " << unresValue << ", RC: " << rc
                     << endl;
                return 1;
            }

            if (gUseTCPIPInterface == 0)
            {
                cout << "Error on Service definition, " << line  << endl
                     << "Cannot delegate a Service to an unconfigured "
                     << "interface" << endl;
                return 1;
            }

            service = STAFServicePtr(new STAFDelegatedService(name, toMachine, 
                                     toName), STAFServicePtr::INIT);
        }

        rc = gServiceManagerPtr->add(service);

        if (rc)
        {
            cout << "Error, RC: " << rc << ", adding Service definition, "
                 << line << endl;
            return 1;
        }
    }
    catch (STAFServiceCreateException &e)
    {
        cout << "Error on Service definition, " << line << endl
             << e.getErrorCode() << ":" << e.getText() << endl;
        return 1;
    }
    catch (STAFException &e)
    {
        e.write();
        return 1;
    }
    catch (...)
    {
        // XXX: Why is this here?
        cout << "Exception caught. Need to get this out (STAFConfig.cpp)"
             << endl;
        return 1;
    }

    return 0;
}


unsigned int handleServiceLoaderConfig(STAFString line)
{
    STAFCommandParser fServiceLoaderParser;

    // SERVICELOADER
    // LIBRARY <Library> [EXECUTE <Exec>] {OPTION <OpName[=OpValue]>}
    //     [PARMS <Parms>]
    
    // serviceloader config options
 
    fServiceLoaderParser.addOption("SERVICELOADER",  1,
        STAFCommandParser::kValueNotAllowed);
    fServiceLoaderParser.addOption("LIBRARY",    1,
        STAFCommandParser::kValueRequired);
    fServiceLoaderParser.addOption("EXECUTE",    1,
        STAFCommandParser::kValueRequired);
    fServiceLoaderParser.addOption("OPTION",    0,
        STAFCommandParser::kValueRequired);
    fServiceLoaderParser.addOption("PARMS",    1,
        STAFCommandParser::kValueRequired);
 
    // service config groups
 
    fServiceLoaderParser.addOptionGroup("SERVICELOADER", 1, 1);
    fServiceLoaderParser.addOptionGroup("LIBRARY", 1, 1);
   
    // service config needs
 
    fServiceLoaderParser.addOptionNeed("EXECUTE OPTION PARMS",  "LIBRARY");
                  
    STAFCommandParseResultPtr parsedResult = fServiceLoaderParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on ServiceLoader definition, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFString errorBuffer;
    STAFString unresValue;
    STAFRC_t rc;

    try
    {
        STAFServicePtr service;
        
        // Setup external service data
        STAFString library;
        STAFString exec;
        STAFString parms;
        STAFExternalService::OptionList options;

        unresValue = parsedResult->optionValue("LIBRARY");
        rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, library);

        if (!rc)
        {            
            unresValue = parsedResult->optionValue("EXECUTE");
            rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, exec);
        }

        if (!rc)
        {            
            unresValue = parsedResult->optionValue("PARMS");
            rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, parms);
        }

        for (int i = 1; !rc && (i <= parsedResult->optionTimes("OPTION"));
             ++i)
        {
            STAFString option;
            unresValue = parsedResult->optionValue("OPTION", i);
            rc = (*gGlobalVariablePoolPtr)->resolve(unresValue, 0, option);
            options.push_back(option);
        }

        if (rc)
        {
            cout << "Error on ServiceLoader definition, " << line << endl
                 << "Error resolving " << unresValue << ", RC: " << rc
                 << endl;
            return 1;
        }
        
        service = STAFServicePtr(new 
            STAFExternalService("STAF/ServiceLoader/" + STAFString(fSLSNum), library,
            exec, options, parms), STAFServicePtr::INIT);
            
        fSLSNum++;
        
        rc = gServiceManagerPtr->addSLS(service);

        if (rc)
        {
            cout << "Error, RC: " << rc 
                << ", adding ServiceLoader definition, "
                << line << endl;
            return 1;
        }
    }
    catch (STAFServiceCreateException &e)
    {
        cout << "Error on ServiceLoader definition, " << line << endl
             << e.getErrorCode() << ":" << e.getText() << endl;
        return 1;
    }
    catch (STAFException &e)
    {
        e.write();
        return 1;
    }
    catch (...)
    {
        // XXX: Why is this here?
        cout << "Exception caught. Need to get this out (STAFConfig.cpp)"
             << endl;
        return 1;
    }

    return 0;
}


unsigned int handleSetConfig(STAFString line)
{
    STAFCommandParser fSetParser;

    // set config options

    // XXX: Note ALLOWMULTIREG is deprecated.  It is currently allowed by the
    //      parser, but ignored everywhere else, for backward compatibility.
    //      It will be removed entirely in STAF V3.0
 
    fSetParser.addOption("SET",                   1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("CONNECTTIMEOUT",        1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("CONNECTATTEMPTS",       1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("CONNECTRETRYDELAY",     1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("MAXFILES",              1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("USELONGNAMES",          1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("MAXQUEUESIZE",          1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("INITIALTHREADS",        1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("THREADGROWTHDELTA",     1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("ALLOWMULTIREG",         1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("DEFAULTSTOPUSING",      1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTNEWCONSOLE",     1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("DEFAULTSAMECONSOLE",    1,
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("DEFAULTAUTHDISABLEDACTION", 1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTAUTHUSERNAME",   1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTAUTHPASSWORD",   1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("PROCESSAUTHMODE",       1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTSHELL",          1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTNEWCONSOLESHELL", 1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULTSAMECONSOLESHELL", 1,
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("ENABLEDIAGS", 1,
        STAFCommandParser::kValueNotAllowed);
 
    // set config groups
 
    fSetParser.addOptionGroup("SET", 1 , 1);
    fSetParser.addOptionGroup("DEFAULTNEWCONSOLE DEFAULTSAMECONSOLE", 0 , 1);
    
    STAFCommandParseResultPtr parsedResult = fSetParser.parse(line);
 
    if (parsedResult->rc != kSTAFOk)
    {
        cout << "Error on SET statement, " << line << endl
             << parsedResult->errorBuffer << endl;
        return 1;
    }

    STAFRC_t rc = kSTAFOk;

    if (parsedResult->optionTimes("CONNECTTIMEOUT") != 0)
    {
        // Handle CONNECTTIMEOUT

        STAFString unresConnectTimeout = 
            parsedResult->optionValue("CONNECTTIMEOUT");
        STAFString connectTimeout;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresConnectTimeout, 0,
                                                connectTimeout);
        if (rc != kSTAFOk)
        {
            cout << "Error on CONNECTTIMEOUT setting, " << line << endl
                 << "Error resolving " << unresConnectTimeout << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!connectTimeout.isDigits())
        {
            cout << "Error on CONNECTTIMEOUT setting, " << line << endl
                 << "CONNECTTIMEOUT must be a non-negative integer" << endl;
            return 1;
        }

        gTCPIPConnectionTimeout = connectTimeout.asUInt();
    }

    if (parsedResult->optionTimes("CONNECTATTEMPTS") != 0)
    {
        // Handle CONNECTATTEMPTS

        STAFString unresConnectAttempts = 
            parsedResult->optionValue("CONNECTATTEMPTS");
        STAFString connectAttempts;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresConnectAttempts, 0,
                                                connectAttempts);
        if (rc != kSTAFOk)
        {
            cout << "Error on CONNECTATTEMPTS setting, " << line << endl
                 << "Error resolving " << unresConnectAttempts << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!connectAttempts.isDigits())
        {
            cout << "Error on CONNECTATTEMPTS setting, " << line << endl
                 << "CONNECTATTEMPTS must be a positive integer" << endl;
            return 1;
        }

        gTCPIPConnectionAttempts = connectAttempts.asUInt();

        if (gTCPIPConnectionAttempts < 1)
        {
            cout << "Error on CONNECTATTEMPTS setting, " << line << endl
                 << "CONNECTATTEMPTS must be a positive integer" << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("CONNECTRETRYDELAY") != 0)
    {
        // Handle CONNECTRETRYDELAY

        STAFString unresConnectRetryDelay = 
            parsedResult->optionValue("CONNECTRETRYDELAY");
        STAFString connectRetryDelay;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresConnectRetryDelay, 0,
                                                connectRetryDelay);
        if (rc != kSTAFOk)
        {
            cout << "Error on CONNECTRETRYDELAY setting, " << line << endl
                 << "Error resolving " << unresConnectRetryDelay << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!connectRetryDelay.isDigits())
        {
            cout << "Error on CONNECTRETRYDELAY setting, " << line << endl
                 << "CONNECTRETRYDELAY must be a non-negative integer" << endl;
            return 1;
        }

        gConnectionRetryDelay = connectRetryDelay.asUInt();
    }

    if (parsedResult->optionTimes("MAXFILES") != 0)
    {
        // Handle MAXFILES

        STAFString unresMaxFiles = parsedResult->optionValue("MAXFILES");
        STAFString maxFiles;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresMaxFiles, 0, maxFiles);

        if (rc != kSTAFOk)
        {
            cout << "Error on MAXFILES setting, " << line << endl
                 << "Error resolving " << unresMaxFiles << ", RC: " << rc
                 << endl;
            return 1;
        }

        if (!maxFiles.isDigits())
        {
            cout << "Error on MAXFILES setting, " << line << endl
                 << "MAXFILES must be a non-negative integer" << endl;
            return 1;
        }

        gMaxFiles = (long)maxFiles.asUInt();
    }

    // Handle USELONGNAMES

    if (parsedResult->optionTimes("USELONGNAMES") != 0)
        gUseLongNames = 1;

    if (parsedResult->optionTimes("MAXQUEUESIZE") != 0)
    {
        // Handle MAXQUEUESIZE

        STAFString unresMaxQueueSize = 
            parsedResult->optionValue("MAXQUEUESIZE");
        STAFString maxQueueSize;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresMaxQueueSize, 0,
                                                maxQueueSize);
        if (rc != kSTAFOk)
        {
            cout << "Error on MAXQUEUESIZE setting, " << line << endl
                 << "Error resolving " << unresMaxQueueSize << ", RC: " 
                 << rc << endl;
            return 1;
        }

        if (!maxQueueSize.isDigits())
        {
            cout << "Error on MAXQUEUESIZE setting, " << line << endl
                 << "MAXQUEUESIZE must be a non-negative integer" << endl;
            return 1;
        }

        gMaxQueueSize = maxQueueSize.asUInt();
    }

    if (parsedResult->optionTimes("INITIALTHREADS") != 0)
    {
        // Handle INITIALTHREADS

        STAFString unresInitialThreads = 
            parsedResult->optionValue("INITIALTHREADS");
        STAFString initialThreads;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresInitialThreads, 0,
                                                initialThreads);
        if (rc != kSTAFOk)
        {
            cout << "Error on INITIALTHREADS setting, " << line  << endl
                 << "Error resolving " << unresInitialThreads << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!initialThreads.isDigits() || 
            (initialThreads.asUInt() < 1))
        {
            cout << "Error on INITIALTHREADS setting, " << line << endl
                 << "INITIALTHREADS must be a positive integer" << endl;
            return 1;
        }

        gNumInitialThreads = initialThreads.asUInt();
    }

    if (parsedResult->optionTimes("THREADGROWTHDELTA") != 0)
    {
        // Handle THREADGROWTHDELTA

        STAFString unresThreadDelta = 
            parsedResult->optionValue("THREADGROWTHDELTA");
        STAFString threadDelta;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresThreadDelta, 0,
                                                threadDelta);
        if (rc != kSTAFOk)
        {
            cout << "Error on THREADGROWTHDELTA setting, " << line << endl
                 << "Error resolving " << unresThreadDelta << ", RC: "
                 << rc << endl;
            return 1;
        }

        if (!threadDelta.isDigits() || (threadDelta.asUInt() < 1))
        {
            cout << "Error on THREADGROWTHDELTA setting, " << line << endl 
                 << "THREADGROWTHDELTA must be a positive integer" << endl;
            return 1;
        }

        gThreadManagerPtr->setGrowthDelta(threadDelta.asUInt());
    }

    if (parsedResult->optionTimes("DEFAULTSTOPUSING") != 0)
    {
        STAFString unresStopString =
                   parsedResult->optionValue("DEFAULTSTOPUSING");
        STAFString stopString;
        STAFProcessStopMethod_t stopMethod;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresStopString, 0,
                                                stopString);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSTOPUSING setting, " << line << endl
                 << "Error resolving " << unresStopString << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::getStopMethodFromString(stopMethod,
                                                         stopString);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSTOPUSING setting, " << line << endl
                 << "Invalid value, " << stopString << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultStopMethod(stopMethod); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSTOPUSING setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("PROCESSAUTHMODE") != 0)
    {
        STAFString unresAuthModeString =
                   parsedResult->optionValue("PROCESSAUTHMODE");
        STAFString authModeString;
        STAFProcessAuthenticationMode_t authMode;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresAuthModeString, 0,
                                                authModeString);
        if (rc != kSTAFOk)
        {
            cout << "Error on PROCESSAUTHMODE setting, " << line << endl
                 << "Error resolving " << unresAuthModeString << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::getAuthModeFromString(authMode,
                                                       authModeString);

        if (rc != kSTAFOk)
        {
            cout << "Error on PROCESSAUTHMODE setting, " << line << endl
                 << "Invalid value, " << authModeString << ", RC: "
                 << rc << endl;
            return 1;
        }

        unsigned int osRC = 0;

        rc = STAFProcessService::setAuthMode(authMode, osRC);

        if (rc != kSTAFOk)
        {
            cout << "Error on PROCESSAUTHMODE setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHUSERNAME") != 0)
    {
        STAFString unresAuthUsernameString =
                   parsedResult->optionValue("DEFAULTAUTHUSERNAME");
        STAFString authUsernameString;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresAuthUsernameString, 0,
                                                authUsernameString);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHUSERNAME setting, " << line << endl
                 << "Error resolving " << unresAuthUsernameString << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultAuthUsername(authUsernameString); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHUSERNAME setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHPASSWORD") != 0)
    {
        STAFString unresAuthPasswordString =
                   parsedResult->optionValue("DEFAULTAUTHPASSWORD");
        STAFString authPasswordString;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresAuthPasswordString, 0,
                                                authPasswordString);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHPASSWORD setting, " << line << endl
                 << "Error resolving " << unresAuthPasswordString << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultAuthPassword(authPasswordString); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHPASSWORD setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTAUTHDISABLEDACTION") != 0)
    {
        STAFString unresDefaultAuthDisabledActionString =
                   parsedResult->optionValue("DEFAULTAUTHDISABLEDACTION");
        STAFString defaultAuthDisabledActionString;
        STAFProcessDisabledAuthAction_t defaultAuthDisabledAction;

        rc = (*gGlobalVariablePoolPtr)->resolve(
             unresDefaultAuthDisabledActionString, 0, 
             defaultAuthDisabledActionString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHDISABLEDACTION setting, " << line 
                 << endl
                 << "Error resolving " << unresDefaultAuthDisabledActionString 
                 << ", RC: " << rc << endl;
            return 1;
        }

        rc = STAFProcessService::getDefaultDisabledAuthActionFromString(
             defaultAuthDisabledAction, defaultAuthDisabledActionString);

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHDISABLEDACTION setting, " << line
                 << endl
                 << "Invalid value, " << defaultAuthDisabledActionString 
                 << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultDisabledAuthAction(
             defaultAuthDisabledAction); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTAUTHDISABLEDACTION setting, " << line
                 << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    if (parsedResult->optionTimes("DEFAULTNEWCONSOLE") != 0)
        gDefaultConsoleMode = kSTAFProcessNewConsole;
    else if (parsedResult->optionTimes("DEFAULTSAMECONSOLE") != 0)
        gDefaultConsoleMode = kSTAFProcessSameConsole;
    
    if (parsedResult->optionTimes("DEFAULTSHELL") != 0)
    {
        STAFString unresShellCommand =
                   parsedResult->optionValue("DEFAULTSHELL");
        STAFString shellCommand;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresShellCommand, 0,
                                                shellCommand);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSHELL setting, " << line << endl
                 << "Error resolving " << unresShellCommand << ", RC: "
                 << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultShellCommand(shellCommand); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSHELL setting, " << line << endl
                 << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }
    
    if (parsedResult->optionTimes("DEFAULTSAMECONSOLESHELL") != 0)
    {
        STAFString unresShellCommand =
                   parsedResult->optionValue("DEFAULTSAMECONSOLESHELL");
        STAFString shellCommand;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresShellCommand, 0,
                                                shellCommand);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSAMECONSOLESHELL setting, " << line
                 << endl << "Error resolving " << unresShellCommand
                 << ", RC: " << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultSameConsoleShell(shellCommand); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTSAMECONSOLESHELL setting, " << line
                 << endl << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }
        
    if (parsedResult->optionTimes("DEFAULTNEWCONSOLESHELL") != 0)
    {
        STAFString unresShellCommand =
                   parsedResult->optionValue("DEFAULTNEWCONSOLESHELL");
        STAFString shellCommand;

        rc = (*gGlobalVariablePoolPtr)->resolve(unresShellCommand, 0,
                                                shellCommand);
        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTNEWCONSOLESHELL setting, " << line
                 << endl << "Error resolving " << unresShellCommand
                 << ", RC: " << rc << endl;
            return 1;
        }

        rc = STAFProcessService::setDefaultNewConsoleShell(shellCommand); 

        if (rc != kSTAFOk)
        {
            cout << "Error on DEFAULTNEWCONSOLESHELL setting, " << line
                 << endl << "Error changing the default, RC: " << rc << endl;
            return 1;
        }
    }

    // Handle ENABLEDIAGS

    if (parsedResult->optionTimes("ENABLEDIAGS") != 0)
    {
        rc = gDiagManagerPtr->enable();

        if (rc != kSTAFOk)
        {
            cout << "Error on ENABLEDIAGS settings, " << line 
                 << endl << "RC: " << rc << endl;
            return 1;
        }
    }

    return 0;
}
