/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFString.h"
#include "STAFProc.h"
#include "STAFException.h"
#include "STAFConnectionManager.h"
#include "STAFLocalConnection.h"
#include "STAFSocketConnection.h"
#include "STAFThreadManager.h"
#include "STAFTrace.h"

STAFConnectionManager::STAFConnectionManager()
{
    /* Do Nothing */
}


STAFRC_t STAFConnectionManager::makeConnection(const STAFString &where,
                                                    STAFConnPtr &connection,
                                                    STAFString &errorBuffer)
{
    STAFRC_t rc = kSTAFOk;
    STAFString lowerWhere = where.toLowerCase();

    // Attempt to make a connection up to the number of times specified by
    // gTCIPConnectionAttempts until a connection is made.

    for (int attempt = 1; attempt <= gTCPIPConnectionAttempts; ++attempt)
    {
        rc = kSTAFOk;

        STAFConnAllocInfo allocInfo(0, 0);

        try
        {
            if ((lowerWhere == gMachineNetNamePtr->toLowerCase()) ||
                (lowerWhere == gMachineNamePtr->toLowerCase()) ||
                (lowerWhere == "local"))
            {
                allocInfo = STAFLocalConnection::create("STAFProc");

            }
            else if (gUseTCPIPInterface)
            {
                allocInfo = STAFSocketConnection::create(where, gTCPIPPort,
                                                         gTCPIPConnectionTimeout);
            }
            else rc = kSTAFNoPathToMachine;

            connection = STAFConnPtr(allocInfo.pConnection, STAFConnPtr::INIT,
                                     STAFConnPtr::CUSTOM, allocInfo.pFreeFunc);
        }
        catch (STAFConnectionConnectException &e)
        {
            rc = kSTAFNoPathToMachine;
            errorBuffer = e.getText() + STAFString(": ") +
                          STAFString(e.getErrorCode()) + ", Machine: " + where;
        }
        catch (...)
        {
            rc = kSTAFUnknownError;
        }

        if (rc == kSTAFOk) return rc;  // Worked, so don't retry

        int delayTime = 0;

        // Sleep (except after last connection attempt)
        if (attempt < gTCPIPConnectionAttempts)
        {
            // Sleep a random number of milliseconds between 1 and the
            // gConnectionRetryDelay to reduce the # of connection attempts
            // needed when getting RC 16 recv: 111 errors that occur when the
            // queue for pending socket requests is full (Bug #915342)

            delayTime = rand() % gConnectionRetryDelay;
            STAFThreadManager::sleepCurrentThread(delayTime);
        }
        
        STAFTrace::trace(
            kSTAFTraceWarning,
            "STAFConnectionManager::makeConnection - Attempt #" +
            STAFString(attempt) + " of " + gTCPIPConnectionAttempts +
            " (Delay " + delayTime + " milliseconds), RC: " + rc +
            ", Result: " + errorBuffer);
    }

    return rc;
}


STAFConnectionManager::~STAFConnectionManager()
{
    /* Do Nothing */
}
