/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2004                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFDiagService.h"
#include "STAFDiagManager.h"
#include "STAFCommandParser.h"
#include <set>
#include <algorithm>
#include <vector>

// RECORD TRIGGER <Trigger> SOURCE <Source>
// LIST   [TRIGGER <Trigger> | SOURCE <Source> | TRIGGERS | SOURCES | SETTINGS]
// RESET  FORCE
// ENABLE
// DISABLE
// HELP

typedef std::set<STAFString> STAFStringList;
typedef std::vector<STAFDiagManager::DiagData> DiagDataVector;

// Descending Count Sorting Function for DiagData

static bool descendingCountSort(STAFDiagManager::DiagData lhs,
                                STAFDiagManager::DiagData rhs)
{
    return (lhs.count > rhs.count);
}

// Ascending Source Sorting Function for DiagData

struct AscendingSourceSort
{
    bool operator()(STAFDiagManager::DiagData lhs,
                    STAFDiagManager::DiagData rhs)
    {
        unsigned int comp = 0;

        STAFStringCompareTo(lhs.source.toUpperCase().getImpl(),
                            rhs.source.toUpperCase().getImpl(),
                            &comp, 0);
        
        return (comp == 1);
    }
};

static const STAFString sNotApplicable = "<N/A>";

STAFDiagService::STAFDiagService() : STAFService("DIAG")
{
    // Record options
 
    fRecordParser.addOption("RECORD", 1, STAFCommandParser::kValueNotAllowed);
    fRecordParser.addOption("TRIGGER", 1, STAFCommandParser::kValueRequired);
    fRecordParser.addOption("SOURCE", 1, STAFCommandParser::kValueRequired);
 
    // Record groups
 
    fRecordParser.addOptionGroup("TRIGGER", 1, 1);
    fRecordParser.addOptionGroup("SOURCE", 1, 1);
    fRecordParser.addOptionNeed("TRIGGER", "SOURCE");
    fRecordParser.addOptionNeed("SOURCE", "TRIGGER");
 
    // List options

    fListParser.addOption("LIST", 1, STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("TRIGGER", 1, STAFCommandParser::kValueRequired);
    fListParser.addOption("SOURCE", 1, STAFCommandParser::kValueRequired);
    fListParser.addOption("TRIGGERS", 1, STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SOURCES", 1, STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SETTINGS", 1, STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SORTBYCOUNT", 1,
                          STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SORTBYTRIGGER", 1,
                          STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SORTBYSOURCE", 1,
                          STAFCommandParser::kValueNotAllowed);
 
    // List groups
 
    fListParser.addOptionGroup(
        "TRIGGER SOURCE TRIGGERS SOURCES SETTINGS", 0, 1);
    fListParser.addOptionGroup(
        "SORTBYCOUNT SORTBYTRIGGER SORTBYSOURCE SETTINGS", 0, 1);
 
    // Reset options
 
    fResetParser.addOption("RESET",  1, STAFCommandParser::kValueNotAllowed);
    fResetParser.addOption("FORCE", 1, STAFCommandParser::kValueNotAllowed);
 
    // Reset groups
 
    fResetParser.addOptionGroup("FORCE", 1, 1);

    // Enable options
    
    fEnableParser.addOption("ENABLE", 1, STAFCommandParser::kValueNotAllowed);

    // Disable options

    fDisableParser.addOption("DISABLE", 1,
                             STAFCommandParser::kValueNotAllowed);
}


STAFDiagService::~STAFDiagService()
{
    /* Do Nothing */
}


STAFString STAFDiagService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFDiagService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if (action == "record")
        return handleRecord(requestInfo);
    else if (action == "list")
        return handleList(requestInfo);
    else if (action == "reset")
        return handleReset(requestInfo);
    else if (action == "enable")
        return handleEnable(requestInfo);
    else if (action == "disable")
        return handleDisable(requestInfo);
    else if (action == "help") 
        return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString, "");
}


STAFServiceResult STAFDiagService::handleRecord(
    const STAFServiceRequest &requestInfo)
{
    if (!isLocalMachine(requestInfo.fMachine))
        return kSTAFAccessDenied;

    // This is redundant given that RECORD is only valid locally
    //
    // if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult = 
        fRecordParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    // Resolve any STAF variables specified in values for TRIGGER and SOURCE
 
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString errorBuffer;

    STAFString trigger;
    STAFString source;

    STAFRC_t rc = resolveStringIfExists(parsedResult, "TRIGGER", handlePool,
                                        trigger, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = resolveStringIfExists(parsedResult, "SOURCE", handlePool,
                               source, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);
    
    // Add to diagnostics map

    return STAFServiceResult(gDiagManagerPtr->record(trigger, source));
}


STAFServiceResult STAFDiagService::handleList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult = 
        fListParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString result("");

    if (parsedResult->optionTimes("SETTINGS") != 0)
    {
        // List the settings for the diagnostic service

        result += "Diagnostics               : " +
                  gDiagManagerPtr->getEnabledAsString() +
                  *gLineSeparatorPtr +
                  "Last Disabled             : " +
                  gDiagManagerPtr->getDisabledTimestamp().asString() +
                  *gLineSeparatorPtr +
                  "Last Reset / First Enabled: ";
        
        if (gDiagManagerPtr->getResetTimestamp() == 0)
            result += sNotApplicable;
        else
            result += gDiagManagerPtr->getResetTimestamp().asString();

        result += *gLineSeparatorPtr;

        return STAFServiceResult(kSTAFOk, result);
    }

    // Get a copy of the Diagnostics Map

    STAFDiagManager::DiagMap diagMap = gDiagManagerPtr->getDiagMapCopy();

    // Determine Start Time - If never been enabled/reset, assign "<N/A>".

    STAFString startTime;

    if (gDiagManagerPtr->getResetTimestamp() == 0)
        startTime = sNotApplicable;
    else
        startTime = gDiagManagerPtr->getResetTimestamp().asString();

    // Determine Stop Time - If enabled, use current time; otherwise, use
    // time last disabled.

    STAFString stopTime;

    if (gDiagManagerPtr->getEnabled())
        stopTime = STAFTimestamp::now().asString();
    else
        stopTime = gDiagManagerPtr->getDisabledTimestamp().asString();
    
    // Determine Elapsed Time.  If startTime == "<N/A>", assign "<N/A>".
    
    STAFString elapsedTime;

    if (startTime == sNotApplicable)
        elapsedTime = sNotApplicable;
    else if (STAFTimestamp(startTime) > STAFTimestamp(stopTime))
        elapsedTime = sNotApplicable;
    else
    {
        elapsedTime = STAFTimestamp::getElapsedTime(STAFTimestamp(stopTime) -
                                                    STAFTimestamp(startTime));
    }

    // Create prolog for result

    result = "From Date/Time    : " + startTime +
             *gLineSeparatorPtr +
             "To Date/Time      : " + stopTime +
             *gLineSeparatorPtr + 
             "Elapsed Time      : " + elapsedTime +
             *gLineSeparatorPtr;

    // Determine how to sort the results

    STAFDiagSortBy_t sortBy = kSTAFDiagSortByCount; // Default Sort Option

    if (parsedResult->optionTimes("SORTBYTRIGGER") != 0)
        sortBy = kSTAFDiagSortByTrigger;
    else if (parsedResult->optionTimes("SORTBYSOURCE") != 0)
        sortBy = kSTAFDiagSortBySource;

    // Determine the LIST option which dictates the results to generate.
    
    if (parsedResult->optionTimes("TRIGGER") != 0)
    {
        // List all the sources for the specified trigger

        // Resolve any STAF variables specified in values for TRIGGER

        STAFVariablePoolPtr handlePool = getClientVariablePool(
            requestInfo.fMachine, requestInfo.fHandle);

        STAFString errorBuffer;
        STAFString trigger;

        STAFRC_t rc = resolveStringIfExists(parsedResult, "TRIGGER",
                                            handlePool, trigger, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFString triggerLowerCase = trigger.toLowerCase();
        unsigned int sourceCount  = 0;
        DiagDataVector diagDataVector;
        STAFDiagManager::DiagMap::iterator mIter;
       
        for (mIter = diagMap.begin(); mIter != diagMap.end(); mIter++)
        {
            if (triggerLowerCase == mIter->second.trigger.toLowerCase())
            {
                diagDataVector.push_back(mIter->second);
                sourceCount++;
            }
        }

        // Sort the diagDataVector (already sorted by trigger)

        switch (sortBy)
        {
            case kSTAFDiagSortByCount:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            case kSTAFDiagSortByTrigger:
            {
                // Sort by count since only one trigger

                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            case kSTAFDiagSortBySource:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          AscendingSourceSort());
                break;
            }
             
            default: break;
        }

        // Iterate thru the sorted diagDataVector and generate the result

        DiagDataVector::iterator vIter;
        STAFString iterResult = STAFString("");
        
        for (vIter = diagDataVector.begin(); vIter != diagDataVector.end();
             vIter++)
        {
            iterResult += STAFString(vIter->count) + ";" + vIter->source +
                *gLineSeparatorPtr;
        }

        result += "Trigger           : " + trigger +
                  *gLineSeparatorPtr +
                  "Number of Sources : " + STAFString(sourceCount) +
                  *gLineSeparatorPtr + *gLineSeparatorPtr +
                  iterResult;
    }
    else if (parsedResult->optionTimes("SOURCE") != 0)
    {
        // List all the triggers for the specified source

        // Resolve any STAF variables specified in values for SOURCE

        STAFVariablePoolPtr handlePool = getClientVariablePool(
            requestInfo.fMachine, requestInfo.fHandle);

        STAFString errorBuffer;
        STAFString source;

        STAFRC_t rc = resolveStringIfExists(parsedResult, "SOURCE",
                                            handlePool, source, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFString sourceLowerCase = source.toLowerCase();
        unsigned int triggerCount  = 0;
        DiagDataVector diagDataVector;
        STAFDiagManager::DiagMap::iterator mIter;
       
        for (mIter = diagMap.begin(); mIter != diagMap.end(); mIter++)
        {
            if (sourceLowerCase == mIter->second.source.toLowerCase())
            {
                diagDataVector.push_back(mIter->second);
                triggerCount++;
            }
        }

        // Sort the diagDataVector (already sorted by trigger)

        switch (sortBy)
        {
            case kSTAFDiagSortByCount:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            case kSTAFDiagSortBySource:
            {
                // Source by count since only only source

                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }
             
            default: break;
        }

        // Iterate thru the sorted diagDataVector and generate the result

        DiagDataVector::iterator vIter;
        STAFString iterResult = STAFString("");
        
        for (vIter = diagDataVector.begin(); vIter != diagDataVector.end();
             vIter++)
        {
            iterResult += STAFString(vIter->count) + ";" + vIter->trigger +
                *gLineSeparatorPtr;
        }

        result += "Source            : " + source +
                  *gLineSeparatorPtr +
                  "Number of Triggers: " + STAFString(triggerCount) +
                  *gLineSeparatorPtr + *gLineSeparatorPtr +
                  iterResult;
    }
    else if (parsedResult->optionTimes("TRIGGERS") != 0)
    {
        // List all the triggers

        unsigned int triggerCount = 0;
        STAFDiagManager::DiagMap triggerMap;
        STAFDiagManager::DiagMap::iterator mIter;
       
        // Create a map of unique triggers, accumulating counts by trigger

        for (mIter = diagMap.begin(); mIter != diagMap.end(); mIter++)
        {
            if (triggerMap.find(mIter->second.trigger.toLowerCase()) ==
                triggerMap.end())
            {
                triggerMap[mIter->second.trigger.toLowerCase()] =
                    mIter->second;
            }
            else
            {
                triggerMap[mIter->second.trigger.toLowerCase()].count +=
                    mIter->second.count;
            }
        }
        
        // Iterate thru the TriggerMap and create a vector of triggers

        DiagDataVector diagDataVector;
       
        for (mIter = triggerMap.begin(); mIter != triggerMap.end(); mIter++)
        {
            diagDataVector.push_back(mIter->second);
        }

        // Sort the diagDataVector (already sorted by trigger)

        switch (sortBy)
        {
            case kSTAFDiagSortByCount:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            case kSTAFDiagSortBySource:
            {
                // Sort by Count since source will not be in the result

                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }
             
            default: break;
        }

        // Iterate thru the sorted diagDataVector and generate the result

        DiagDataVector::iterator vIter;
        STAFString iterResult = STAFString("");
        
        for (vIter = diagDataVector.begin(); vIter != diagDataVector.end();
             vIter++)
        {
            iterResult += STAFString(vIter->count) + ";" + vIter->trigger +
                *gLineSeparatorPtr;

            triggerCount++;
        }

        result += "Number of Triggers: " + STAFString(triggerCount) +
                  *gLineSeparatorPtr + *gLineSeparatorPtr +
                  iterResult;
    }
    else if (parsedResult->optionTimes("SOURCES") != 0)
    {
        // List all the sources

        unsigned int sourceCount = 0;
        STAFDiagManager::DiagMap sourceMap;
        STAFDiagManager::DiagMap::iterator mIter;
       
        // Create a map of unique sources, accumulating counts by source

        for (mIter = diagMap.begin(); mIter != diagMap.end(); mIter++)
        {
            if (sourceMap.find(mIter->second.source.toLowerCase()) ==
                sourceMap.end())
            {
                sourceMap[mIter->second.source.toLowerCase()] =
                    mIter->second;
            }
            else
            {
                sourceMap[mIter->second.source.toLowerCase()].count +=
                    mIter->second.count;
            }
        }
        
        // Iterate thru the SourceMap and create a vector of sources

        DiagDataVector diagDataVector;
       
        for (mIter = sourceMap.begin(); mIter != sourceMap.end(); mIter++)
        {
            diagDataVector.push_back(mIter->second);
        }

        // Sort the diagDataVector (already sorted by trigger)

        switch (sortBy)
        {
            case kSTAFDiagSortByCount:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            case kSTAFDiagSortBySource:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          AscendingSourceSort());
                break;
            }

            case kSTAFDiagSortByTrigger:
            {
                // Sort by Count since trigger will not be in the result

                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            default: break;
        }

        // Iterate thru the sorted diagDataVector and generate the result

        DiagDataVector::iterator vIter;
        STAFString iterResult = STAFString("");
        
        for (vIter = diagDataVector.begin(); vIter != diagDataVector.end();
             vIter++)
        {
            iterResult += STAFString(vIter->count) + ";" + vIter->source +
                *gLineSeparatorPtr;

            sourceCount++;
        }

        result += "Number of Sources : " + STAFString(sourceCount) +
                  *gLineSeparatorPtr + *gLineSeparatorPtr +
                  iterResult;
    }
    else
    {
        // List all diagnostic data

        STAFStringList triggerList;
        STAFStringList sourceList;
        DiagDataVector diagDataVector;
        STAFDiagManager::DiagMap::iterator mIter;
       
        for (mIter = diagMap.begin(); mIter != diagMap.end(); mIter++)
        {
            diagDataVector.push_back(mIter->second);

            // Generate a list of all unique triggers

            if (triggerList.find(mIter->second.trigger.toLowerCase()) ==
                triggerList.end())
            {
                triggerList.insert(mIter->second.trigger.toLowerCase());
            }

            // Generate a list of all unique sources

            if (sourceList.find(mIter->second.source.toLowerCase()) ==
                sourceList.end())
            {
                sourceList.insert(mIter->second.source.toLowerCase());
            }
        }

        // Sort the diagDataVector (already sorted by trigger)

        switch (sortBy)
        {
            case kSTAFDiagSortByCount:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          descendingCountSort);
                break;
            }

            case kSTAFDiagSortBySource:
            {
                std::sort(diagDataVector.begin(), diagDataVector.end(),
                          AscendingSourceSort());
                break;
            }
             
            default: break;
        }

        // Iterate thru the sorted diagDataVector and generate the result

        DiagDataVector::iterator vIter;
        STAFString iterResult = STAFString("");
        
        for (vIter = diagDataVector.begin(); vIter != diagDataVector.end();
             vIter++)
        {
            iterResult += STAFString(vIter->count) + ";" +
                vIter->trigger + ";" + vIter->source + 
                 *gLineSeparatorPtr;
        }

        result += "Number of Triggers: " + STAFString(triggerList.size()) +
                  *gLineSeparatorPtr +
                  "Number of Sources : " + STAFString(sourceList.size()) +
                  *gLineSeparatorPtr + *gLineSeparatorPtr +
                  iterResult;
    }
    
    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFDiagService::handleReset(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;
 
    STAFCommandParseResultPtr parsedResult = 
        fResetParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    return STAFServiceResult(gDiagManagerPtr->reset());
}


STAFServiceResult STAFDiagService::handleEnable(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;
    
    STAFCommandParseResultPtr parsedResult = 
        fEnableParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    return STAFServiceResult(gDiagManagerPtr->enable());
}


STAFServiceResult STAFDiagService::handleDisable(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;
    
    STAFCommandParseResultPtr parsedResult = 
        fDisableParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    return STAFServiceResult(gDiagManagerPtr->disable());
}


STAFServiceResult STAFDiagService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString result("DIAG service help" + *gLineSeparatorPtr +
                      *gLineSeparatorPtr);

    result += "RECORD TRIGGER <Trigger> SOURCE <Source>" +
              *gLineSeparatorPtr +
              "LIST   < [TRIGGER <Trigger> | SOURCE <Source> | " +
              "TRIGGERS | SOURCES]" +
              *gLineSeparatorPtr +
              "         [SORTBYCOUNT | SORTBYTRIGGER | SORTBYSOURCE] > |" +
              *gLineSeparatorPtr +
              "       SETTINGS" + *gLineSeparatorPtr +
              "RESET  FORCE" + *gLineSeparatorPtr +
              "ENABLE"  + *gLineSeparatorPtr +
              "DISABLE" + *gLineSeparatorPtr +
              "HELP"    + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
