/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFExternalService.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFTrustManager.h"

STAFExternalService::STAFExternalService(STAFString name, STAFString library,
                                         STAFString exec, OptionList options,
                                         STAFString parms)
    : STAFService(name, 1), fLibName(library),
      fImplLib(library.toCurrentCodePage()->buffer()), fExec(exec),
      fParms(parms)
{
    // Obtain service's interface addresses and levels

    fGetBounds = (STAFServiceGetLevelBounds_t)fImplLib.getAddress(
        "STAFServiceGetLevelBounds");

    fLevelConstruct = getLevel(kServiceInfo, "ServiceInfo", 1, 2);
    fConstruct = (STAFServiceConstruct_t)fImplLib.getAddress(
        "STAFServiceConstruct");

    fLevelInit = getLevel(kServiceInit, "ServiceInit", 1, 2);
    fInit = (STAFServiceInit_t)fImplLib.getAddress("STAFServiceInit");

    fLevelAcceptRequest = getLevel(kServiceAcceptRequest,
                                   "ServiceAcceptRequest", 1, 2);
    fAcceptRequest = (STAFServiceAcceptRequest_t)fImplLib.getAddress(
                     "STAFServiceAcceptRequest");

    fLevelTerm = getLevel(kServiceTerm, "ServiceTerm", 0, 0);
    fTerm = (STAFServiceTerm_t)fImplLib.getAddress("STAFServiceTerm");

    fLevelDestruct = getLevel(kServiceDestruct, "ServiceDestruct", 0, 0);
    fDestruct = (STAFServiceDestruct_t)fImplLib.getAddress(
                "STAFServiceDestruct");

    // Build option name and value arrays

    STAFString *optionNameStrings = new STAFString[options.size()];
    STAFString *optionValueStrings = new STAFString[options.size()];
    unsigned int i = 0;

    for (OptionList::iterator iter = options.begin(); iter != options.end();
         ++iter, ++i)
    {
        unsigned int equalPos = iter->find(kUTF8_EQUAL);
        optionNameStrings[i] = iter->subString(0, equalPos);
        if (equalPos != STAFString::kNPos)
            optionValueStrings[i] = iter->subString(equalPos + 1);
    }

    STAFString_t *optionNames = new STAFString_t[options.size()];
    STAFString_t *optionValues = new STAFString_t[options.size()];

    for (i = 0; i < options.size(); ++i)
    {
        optionNames[i] = optionNameStrings[i].getImpl();
        optionValues[i] = optionValueStrings[i].getImpl();
    }

    // Construct service

    STAFString_t errorBuffer = 0;
    unsigned int constructRC = kSTAFUnknownError;

    if (fLevelConstruct == 1)
    {
        STAFServiceInfoLevel1 info = { name.getImpl(), exec.getImpl(),
                                       options.size(), optionNames,
                                       optionValues };

        constructRC = fConstruct(&fServiceHandle, &info, 1, &errorBuffer);
    }
    else
    {
        STAFServiceInfoLevel2 info = { name.getImpl(), exec.getImpl(),
                                       gSTAFWriteLocationPtr->getImpl(),
                                       options.size(), optionNames,
                                       optionValues };

        constructRC = fConstruct(&fServiceHandle, &info, 2, &errorBuffer);
    }

    // Cleanup

    delete [] optionNameStrings;
    delete [] optionValueStrings;

    if (constructRC != 0)
    {
        STAFString errorString;

        if (errorBuffer)
            errorString = STAFString(errorBuffer, STAFString::kShallow);

        STAFServiceCreateException
            error(errorString.toCurrentCodePage()->buffer(), constructRC);

        THROW_STAF_EXCEPTION(error);
    }
}

unsigned int STAFExternalService::getLevel(STAFServiceLevelID levelID,
                                           const char *levelString,
                                           unsigned int stafMin,
                                           unsigned int stafMax)
{
    unsigned int serviceMin = 0;
    unsigned int serviceMax = 0;
    unsigned int rc = 0;

    // Get the service's supported levels

    if ((rc = fGetBounds(levelID, &serviceMin, &serviceMax) != 0))
    {
        STAFString message(STAFString("Error getting level bounds for ") +
                           levelString);
        STAFServiceCreateException error(message.toCurrentCodePage()->buffer());
        THROW_STAF_EXCEPTION(error);
    }

    // Check for non-overlapping ranges

    if ((serviceMin > stafMax) || (stafMin > serviceMax))
    {
        STAFString message(levelString);
        message += STAFString(" level validation failure, RC = " + rc) +
                   ".  Service supports "        +
                   serviceMin + ":" + serviceMax +
                   ".  STAF supports " + stafMin +
                   ":" + stafMax + ".";
        STAFServiceCreateException
            error(message.toCurrentCodePage()->buffer());
        THROW_STAF_EXCEPTION(error);
    }

    return STAF_MIN(serviceMax, stafMax);
}


STAFExternalService::~STAFExternalService()
{
    STAFString_t errorBuffer = 0;
    unsigned int rc = fDestruct(&fServiceHandle, 0, 0, &errorBuffer);
    unsigned int osRC = 0;

    if (errorBuffer) STAFStringDestruct(&errorBuffer, &osRC);
}


STAFString STAFExternalService::info(unsigned int) const
{
    STAFString data(name() + ": External library " + fLibName);

    if (fExec.length() != 0)
        data += ", exec " + fExec;

    return data;
}


STAFRC_t STAFExternalService::init()
{
    STAFString_t errorBuffer = 0;
    STAFRC_t rc = kSTAFUnknownError;

    if (fLevelInit == 1)
    {
        STAFServiceInitLevel1 initData = { fParms.getImpl() };
        rc = fInit(fServiceHandle, &initData, 1, &errorBuffer);
    }
    else
    {
        STAFServiceInitLevel2 initData = { fParms.getImpl(),
                                           gSTAFWriteLocationPtr->getImpl() };
        rc = fInit(fServiceHandle, &initData, 2, &errorBuffer);
    }

    unsigned int osRC = 0;

    if (errorBuffer) STAFStringDestruct(&errorBuffer, &osRC);

    return rc;
}


STAFServiceResult STAFExternalService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString_t errorBuffer = 0;
    STAFRC_t rc = kSTAFUnknownError;

    if (fLevelAcceptRequest == 1)
    {
        STAFServiceRequestLevel1 requestData = {
            gMachineNetNamePtr->getImpl(),
            requestInfo.fMachine.getImpl(), 0,
            requestInfo.fHandleName.getImpl(),
            requestInfo.fHandle,
            requestInfo.fTrustLevel,
            requestInfo.fRequest.getImpl() };

        STAFString effectiveMachine = effectiveMachineName(
            requestInfo.fMachine);
        requestData.effectiveMachine = effectiveMachine.getImpl();

        STAFString_t resultBuffer;
        STAFRC_t rc = fAcceptRequest(fServiceHandle, &requestData, 1,
                                     &resultBuffer);

        return STAFServiceResult(rc, STAFString(resultBuffer,
                                                STAFString::kShallow));
    }
    else
    {
        STAFServiceRequestLevel2 requestData = {
            gMachineNetNamePtr->getImpl(),
            requestInfo.fMachine.getImpl(), 0,
            requestInfo.fHandleName.getImpl(),
            requestInfo.fHandle,
            requestInfo.fTrustLevel,
            requestInfo.fDiagEnabled,
            requestInfo.fRequest.getImpl() };
        
        STAFString effectiveMachine = effectiveMachineName(
            requestInfo.fMachine);
        requestData.effectiveMachine = effectiveMachine.getImpl();

        STAFString_t resultBuffer;
        STAFRC_t rc = fAcceptRequest(fServiceHandle, &requestData, 2,
                                     &resultBuffer);

        return STAFServiceResult(rc, STAFString(resultBuffer,
                                                STAFString::kShallow));
    }
}


STAFRC_t STAFExternalService::term()
{
    STAFString_t errorBuffer = 0;
    STAFRC_t rc = fTerm(fServiceHandle, 0, 0, &errorBuffer);
    unsigned int osRC = 0;

    if (errorBuffer) STAFStringDestruct(&errorBuffer, &osRC);

    return rc;
}
