/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAF_iostream.h"
#include "STAF_fstream.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFString.h"
#include "STAFRefPtr.h"
#include "STAFUtil.h"
#include "STAFFSService.h"
#include "STAFTrustManager.h"
#include "STAFConnectionManager.h"
#include "STAFVariablePool.h"
#include "STAFException.h"
#include "STAFFileSystem.h"
#include "STAFTrace.h"
#include "STAFConverter.h"
#include "STAFDiagManager.h"
#include "STAFThreadManager.h"
#include <set>
#include <string>

typedef std::set<STAFString> SET_STAFString;

static STAFString sColon(kUTF8_COLON);
static STAFString sSemiColon(kUTF8_SCOLON);
static const unsigned int sMaxReadAttempts = 20;
static const unsigned int sReadRetryDelay = 500;  // 1/2 second (500ms)

// This table allows for lookup of a hex character (in UTF-8)
static const char HEX_TABLE[] =
{
    // Here's the corresponding non-UTF-8 hex representation
    // '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    // 'A', 'B', 'C', 'D', 'E', 'F'

    0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39,
    0x41, 0x42, 0x43, 0x44, 0x45, 0x46
};

STAFString getTypeString(STAFFSEntryType_t type)
{
    switch (type)
    {
        case kSTAFFSFile:              return "F";
        case kSTAFFSSpecialDirectory: // Fall through
        case kSTAFFSDirectory:         return "D";
        case kSTAFFSPipe:              return "P";
        case kSTAFFSSocket:            return "S";
        case kSTAFFSSymLink:           return "L";
        case kSTAFFSBlkDev:            return "B";
        case kSTAFFSCharDev:           return "C";
        case kSTAFFSOther:             return "O";
        default:                       return "?";
    }
}


unsigned int getTypeMaskFromString(const STAFString &typeString,
                                   bool includeSpecial)
{
    STAFString upperTypeString = typeString.toUpperCase();
    unsigned int entryTypesUInt = kSTAFFSNone;

    for (int i = 0; i < upperTypeString.length(STAFString::kChar); ++i)
    {
        STAFString thisChar = upperTypeString.subString(i, 1, STAFString::kChar);

        if ((thisChar == "!") && includeSpecial)
        {
            entryTypesUInt |= kSTAFFSSpecialDirectory;
        }
        else if (thisChar == "F") entryTypesUInt |= kSTAFFSFile;
        else if (thisChar == "D") entryTypesUInt |= kSTAFFSDirectory;
        else if (thisChar == "P") entryTypesUInt |= kSTAFFSPipe;
        else if (thisChar == "S") entryTypesUInt |= kSTAFFSSocket;
        else if (thisChar == "L") entryTypesUInt |= kSTAFFSSymLink;
        else if (thisChar == "B") entryTypesUInt |= kSTAFFSBlkDev;
        else if (thisChar == "C") entryTypesUInt |= kSTAFFSCharDev;
        else if (thisChar == "O") entryTypesUInt |= kSTAFFSOther;
    }

    return entryTypesUInt;
}


void updateResultString(STAFString &result, STAFFSEntryPtr &entry,
                        STAFRC_t rc, unsigned int osRC)
{
    if (rc != kSTAFOk)
    {
        result += entry->path().asString() + sSemiColon + rc;

        if (rc == kSTAFBaseOSError)
            result += sColon + osRC;

        result += *gLineSeparatorPtr;
    }
}


STAFRC_t readFile(fstream &inFile, char *fileBuffer, unsigned int readLength,
                  const STAFString fromFile, const STAFString toMachine,
                  unsigned int fileLength)
{
    // Read a block of data from the file
    
    unsigned int currentPos = (unsigned int)inFile.tellg();
    inFile.read(fileBuffer, readLength);
    
    if (inFile.good())
        return kSTAFOk;  // Read was successful

    // If get eof condition reading the file, make sure that all bytes have
    // really been read to make sure it's not a "pre-mature" eof condition.
    //
    // Note:  A "pre-mature" eof condition can occur on Windows when copying
    // from a file on a mapped drive (in text mode) and the drive is
    // disconnected.

    if (inFile.eof() && (currentPos + inFile.gcount() == fileLength))
        return kSTAFOk;  // Completed reading the file

    // The read failed.  Retry the read up to sMaxReadAttempts times with a
    // delay between each attempt.

    for (int readAttempt = 1;
         !inFile.good() && readAttempt <= sMaxReadAttempts;
         readAttempt++)
    {
        if (inFile.fail())
        {
            // Recoverable read error

            // Log a warning tracepoint message

            STAFString warningMsg(
                "STAFFSService::readFile() - Read attempt #" +
                STAFString(readAttempt) + " failed while copying file " +
                fromFile + " to machine " + toMachine);

            STAFTrace::trace(kSTAFTraceWarning, warningMsg);

            // Delay and retry read after clearing any error flags and
            // repositioning the file pointer

            STAFThreadManager::sleepCurrentThread(
                sReadRetryDelay);

            inFile.clear();

            if (readAttempt == sMaxReadAttempts)
            {
                // Before the last read attempt, try closing the file and
                // reopening it first to see if that fixes the problem

                inFile.close();
                inFile.open(fromFile.toCurrentCodePage()->buffer(),
                            ios::in | STAF_ios_binary);
            }

            inFile.seekg(currentPos, ios::beg);
            inFile.read(fileBuffer, readLength);
        }
        else if (inFile.bad())
        {
            // Unrecoverable read error.
            break;
        }
    }

    if (!inFile.good())
    {
        // Unrecoverable read failure

        return kSTAFFileReadError;
    }

    return kSTAFOk;
}


STAFRC_t copyDirectory(STAFFSEnumPtr childEnum, STAFString fromDir, 
    STAFConnPtr connection, SET_STAFString textExtList, STAFString currentEOL,
    STAFString newEOL, bool doCodepageConvert, int level,
    STAFFSCaseSensitive_t caseSensitive, STAFString &result,
    const STAFString &toMachine)
{
    unsigned int osRC = 0;
    STAFRC_t rc = kSTAFOk;
    STAFRC_t ack = kSTAFOk;
    STAFString ackResult;
    STAFFSEntryPtr fileEntry;
    STAFString toFile;
    bool textTransfer;
    SET_STAFString::iterator i;

    bool doConvert = doCodepageConvert;
    int currentEOLLength=currentEOL.length();

    for (; childEnum->isValid(); childEnum->next())
    {
        fileEntry = childEnum->entry();
        STAFString fromFile = fileEntry->path().asString();
        fromFile = fromFile.replace(kUTF8_BSLASH, kUTF8_SLASH);
        toFile = fromFile.subString(fromDir.length());

        fstream inFile(fromFile.toCurrentCodePage()->buffer(),
                    ios::in | STAF_ios_binary);

        // determine if this file is to be text or binary
        textTransfer = false;
        unsigned int isMatch;

        for(i = textExtList.begin(); i != textExtList.end(); i++)
        {
            isMatch = STAFFileSystem::matchesWildcards(fileEntry->path().
                                                extension(), *i, caseSensitive);
            textTransfer |= (isMatch > 0);
        } 

        doCodepageConvert = doConvert && textTransfer;

        textTransfer &= (!newEOL.isEqualTo(currentEOL));

        if (!inFile)
        {
            updateResultString(result, fileEntry,
                               kSTAFFileOpenError, osRC);
            continue;
        }

        connection->writeUInt(kSTAFFSContinueCopy);

        connection->writeUInt(kSTAFFSFile);
        connection->writeString(toFile);

        if (level > 1)
        {
            if (doCodepageConvert)
                connection->writeUInt(kSTAFFSTextConvert);
            else if (textTransfer)
                connection->writeUInt(kSTAFFSTextNoConvert);
            else
                connection->writeUInt(kSTAFFSBinary);
        }
            
        char fileBuffer[4000] = {0};
        unsigned int writeLength = 0;
            
        STAFRC_t ack = static_cast<STAFRC_t>(connection->readUInt());
        ackResult = connection->readString();

        if (ack != kSTAFOk) 
            updateResultString(result, fileEntry, ack, osRC);

        /*  This if clause is used to perform a text transfer with 
            conversion of eol chars and a codepage conversion
            It runs considerably slower than the binary    transfer.
            This is the type of transfer that will occur if the TEXT
            option is specified and the codepages are different.
            Data is sent as a series of bufers of UTF8 chars
        */
        else if (doCodepageConvert)
        {
            // Get file length

            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();
            inFile.seekg(0, ios::beg);

            connection->writeString(currentEOL);
            
            STAFStringBufferPtr newlineStr = currentEOL.toCurrentCodePage();
            
            // How many bytes in the end-of-line sequence
            unsigned int newlineBufferSize = newlineStr->length();  
            
            // Points to a buffer containing end-of-line sequence
            char * newlineBuffer = new char [newlineBufferSize];  
            memcpy(newlineBuffer, newlineStr->buffer(), newlineBufferSize);  
            
            // Last byte of end-of-line sequence
            char newlineKey     = newlineBuffer[newlineBufferSize - 1];  
            bool done = false;
            const unsigned int defaultBufferSize = 4096;

            STAFRefPtr<char> buffer = STAFRefPtr<char>(new char[defaultBufferSize],
                                                       STAFRefPtr<char>::INIT,
                                                       STAFRefPtr<char>::ARRAY);
            unsigned int bufferSize = defaultBufferSize;
            unsigned int bufferReadOffset = 0;

            while (!done)
            {
                rc = readFile(
                    inFile, static_cast<char *>(buffer + bufferReadOffset),
                    bufferSize - bufferReadOffset, fromFile, toMachine,
                    fileLength);

                if (rc != kSTAFOk)
                {
                    updateResultString(result, fileEntry, rc, osRC);
                    break;
                }

                unsigned int bytesInBuffer = inFile.gcount() + bufferReadOffset;

                if (bytesInBuffer < bufferSize) done = true;

                // Find a newline.  Make sure we don't underrun the buffer;

                unsigned int i = bytesInBuffer - 1;  // Last NewLine index        
                unsigned int bufferGuardIndex = newlineBufferSize - 1;

                while (((buffer[i] != newlineKey) ||
                        !memcmp(buffer + i - newlineBufferSize,
                                newlineBuffer, newlineBufferSize)) &&
                       (i > bufferGuardIndex))
                { --i; }

                while ((i == bufferGuardIndex) && !done)
                {
                    // We have a line bigger than our buffer.
                    // Note: the beginning of the buffer may be a lone newline, but we
                    //       ignore that for this algorithm (as the next line is likely
                    //       larger than the buffer any way

                    // First, create a buffer that is double our current size, and copy
                    // our existing buffer data into it

                    STAFRefPtr<char> tmpBuffer =
                                     STAFRefPtr<char>(new char[bufferSize * 2],
                                                      STAFRefPtr<char>::INIT,
                                                      STAFRefPtr<char>::ARRAY);
                    memcpy(tmpBuffer, buffer, bufferSize);
                    buffer = tmpBuffer;
                    bufferSize *= 2;

                    // Now, read in enough data to fill the remainder of the buffer

                    rc = readFile(inFile, buffer + (bufferSize / 2),
                                  bufferSize / 2, fromFile, toMachine,
                                  fileLength);

                    if (rc != kSTAFOk)
                    {
                        updateResultString(result, fileEntry, rc, osRC);
                        break;
                    }

                    bytesInBuffer += inFile.gcount();

                    // Finally, let's check to make sure that this buffer was big enough,
                    // by finding a newline.  Otherwise, let's run the loop again.

                    if (bytesInBuffer < bufferSize) done = true;

                    i = bytesInBuffer - 1;  // Last NewLine index
                    bufferGuardIndex = (bufferSize / 2) - newlineBufferSize;

                    while (((buffer[i] != newlineKey) ||
                            !memcmp(buffer + i - newlineBufferSize,
                                    newlineBuffer, newlineBufferSize)) &&
                           (i > bufferGuardIndex))
                    { --i; }

                } // while ((i == bufferGuardIndex) && !done)

                // We now have the last newline in the buffer

                if (!done)
                {
                    connection->writeUInt(kSTAFFSContinueCopy);
                    connection->writeString(STAFString(buffer, i + 1, 
                                                       STAFString::kCurrent));
                    memmove(buffer, buffer + i + 1, bufferSize - i - 1);
                    bufferReadOffset = bufferSize - i - 1;
                }
                else
                {
                    connection->writeUInt(kSTAFFSContinueCopy);
                    connection->writeString(STAFString(buffer,
                                                       bytesInBuffer,
                                                       STAFString::kCurrent));
                }

            } // while (!done)

            free (newlineBuffer);
            connection->writeUInt(kSTAFFSFinishedCopy);

            inFile.close();    

            if (rc == kSTAFOk)
            {
                // Read an ack, so that we know the file is closed

                ack = connection->readUInt();

                if (ack != kSTAFOk)
                {
                    updateResultString(result, fileEntry, ack, osRC);
                }
            }
        }

        /*  This if clause is used to perform a text transfer with 
            conversion of eol chars without a codepage conversion
            It runs considerably faster than the codepage conversion 
            transfer.
            This is the type of transfer that will occur if the NOCONVERT
            option is enabled
            This is the type of transfer that will occur if the codepages 
            are the same and a text transfer has been specified
        */
        else if (textTransfer)
        {
            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();
            inFile.seekg(0, ios::beg);

            STAFString transferString;
            int bufferSize = 3000;
            unsigned int transferLen = 0;
            char * buffer = new char[bufferSize]; 
            
            connection->writeString(currentEOL);
            connection->writeUInt(bufferSize);
            
            while ((fileLength > 0) && (inFile.good()))
            {
                transferLen = STAF_MIN(bufferSize, fileLength);
            
                rc = readFile(inFile, buffer, transferLen, fromFile,
                              toMachine, fileLength);
                
                if (rc != kSTAFOk)
                {
                    updateResultString(result, fileEntry, rc, osRC);
                    fileLength = 0;
                    break;
                }

                connection->writeUInt(transferLen);
                connection->write(buffer, transferLen);
                
                fileLength -= transferLen;
             }

             connection->writeUInt(kSTAFFSFinishedCopy);

             free(buffer);
             inFile.close();    

             if (rc == kSTAFOk)
             {
                 // Read an ack, so that we know the file is closed

                 ack = connection->readUInt();
                 
                 if (ack != kSTAFOk)
                 {
                     updateResultString(result, fileEntry, ack, osRC);
                     fileLength = 0;
                 }
             }
        } 
        else
        {
            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();

            connection->writeUInt(fileLength);
            inFile.seekg(0, ios::beg);

            while ((fileLength > 0) && (inFile.good()))
            {
                writeLength = STAF_MIN(sizeof(fileBuffer), fileLength);
                
                rc = readFile(inFile, reinterpret_cast<char *>(fileBuffer),
                              writeLength, fromFile, toMachine, fileLength);

                if (rc != kSTAFOk)
                {
                    updateResultString(result, fileEntry, rc, osRC);
                    fileLength = 0;
                    connection->writeUInt(kSTAFFileReadError);
                    break;
                }

                connection->writeUInt(kSTAFOk);
                connection->write(fileBuffer, writeLength);
                fileLength -= writeLength;

                ack = connection->readUInt();

                if (ack != kSTAFOk)
                {
                    updateResultString(result, fileEntry, ack, osRC);
                    fileLength = 0;
                    break;
                }
            }
            
            inFile.close(); 
        }
    }

    return kSTAFOk;
}


STAFRC_t recurseCopyDir(STAFFSEntryPtr entry, const STAFString &namePattern, 
    const STAFString &extPattern, STAFString fromDir, bool keepEmptyDir, 
    bool onlyDir, unsigned int entryTypesUInt, STAFConnPtr connection, 
    STAFFSCaseSensitive_t caseSensitive, SET_STAFString textExtList ,
    STAFString currentEOL, STAFString newEOL, int level, bool doCodepageConvert,
    STAFString &result, const STAFString &toMachine)
{
    unsigned int osRC = 0;
    STAFRC_t rc = kSTAFOk;

    // Enumerate the files in the current entry
    STAFFSEnumPtr fileEnum = entry->enumerate(namePattern, extPattern,
        STAFFSEntryType_t(entryTypesUInt & ~kSTAFFSDirectory),
        kSTAFFSNoSort, caseSensitive);

    STAFString thisDir = entry->path().asString();

    // relative toDirectory
    STAFString toDir = thisDir.subString(fromDir.length());
    toDir = toDir.replace(kUTF8_BSLASH, kUTF8_SLASH);

    // If the thisDir is equal to fromDir, then the directory already exists
    // If there is at least one file to be copied or KEEPEMPTYDIRECTORIES
    // or ONLYDIRECTORIES option was used, then create the current directory.
    if ((thisDir != fromDir) && 
        (onlyDir || keepEmptyDir || (fileEnum->isValid())))
    {
        connection->writeUInt(kSTAFFSContinueCopy);
        connection->writeUInt(kSTAFFSDirectory);
        connection->writeString(toDir);

        //...wait for STAFProc's ack
        STAFRC_t ack = static_cast<STAFRC_t>(connection->readUInt());
        STAFString ackResult = connection->readString();

        if (ack != kSTAFOk) 
            updateResultString(result, entry, ack, osRC);
    }

    // copy current directory
    if (!onlyDir && fileEnum->isValid())
        copyDirectory(fileEnum, fromDir, connection, textExtList, currentEOL, 
                      newEOL, doCodepageConvert, level, caseSensitive, result,
                      toMachine);

    // Enumerate the directories in the current entry
    STAFFSEnumPtr directoryEnum = entry->enumerate(kUTF8_STAR, kUTF8_STAR,
        STAFFSEntryType_t(entryTypesUInt & ~kSTAFFSFile),
        kSTAFFSNoSort, caseSensitive);

    // copy sub-directories in current directory
    for (; directoryEnum->isValid(); directoryEnum->next())
    {
        recurseCopyDir(directoryEnum->entry(), namePattern, extPattern, 
                       fromDir, keepEmptyDir, onlyDir, entryTypesUInt,
                       connection, caseSensitive, textExtList, currentEOL, 
                       newEOL, level, doCodepageConvert, result, toMachine);
    }
    
    return kSTAFOk;
}


STAFRC_t removeChildren(STAFFSEntryPtr entry, const STAFString &namePattern,
                        const STAFString &extPattern,
                        unsigned int entryTypesUInt,
                        STAFFSCaseSensitive_t caseSensitive, STAFString &result)
{
    STAFFSEnumPtr childEnum = entry->enumerate(namePattern, extPattern,
                                               STAFFSEntryType_t(entryTypesUInt),
                                               kSTAFFSNoSort, caseSensitive);
    unsigned int osRC = 0;
    STAFRC_t rc = kSTAFOk;

    for (; childEnum->isValid(); childEnum->next())
    {
        STAFFSEntryPtr entry = childEnum->entry();
        rc = entry->remove(&osRC);
        updateResultString(result, entry, rc, osRC);
    }

    return kSTAFOk;
}


STAFRC_t recurseRemove(STAFFSEntryPtr entry, const STAFString &namePattern,
                       const STAFString &extPattern, unsigned int entryTypesUInt,
                       STAFFSCaseSensitive_t caseSensitive, STAFString &result)
{
    STAFFSEnumPtr childDirEnum = entry->enumerate(kUTF8_STAR, kUTF8_STAR,
                                                  kSTAFFSDirectory);
    STAFRC_t rc = kSTAFOk;

    for (; childDirEnum->isValid(); childDirEnum->next())
    {
        STAFFSEntryPtr childDirEntry = childDirEnum->entry();
        
        // If child directory entry is a link, skip it

        if (!childDirEntry->isLink())
        {
            rc = recurseRemove(childDirEntry, namePattern, extPattern,
                               entryTypesUInt, caseSensitive, result);
        }
    }

    rc = removeChildren(entry, namePattern, extPattern, entryTypesUInt,
                        caseSensitive, result);

    return kSTAFOk;
}


STAFServiceResult convertToHex(const char *buffer,
                               unsigned int fileLength)
{
    // Convert the buffer to a hex format

    STAFBuffer<char> hexBuffer(new char[fileLength * 2], 
                               STAFBuffer<char>::INIT, 
                               STAFBuffer<char>::ARRAY);

    for (unsigned int i = 0, j = 0; i < fileLength; i++)
    {
        hexBuffer[j++] = HEX_TABLE[(buffer[i] >> 4) & 0xF];
        hexBuffer[j++] = HEX_TABLE[buffer[i] & 0xF];
    }

    // Return new result in hex format
    return STAFServiceResult(kSTAFOk, STAFString(hexBuffer, fileLength * 2,
                                                 STAFString::kUTF8));
}


STAFServiceResult convertLineEndings(const char *buffer,
                                     unsigned int fileLength,
                                     const STAFString &textFormat,
                                     const STAFString &orgMachine,
                                     bool testFlag)
{
    // Convert the line ending characters in the buffer

    STAFString windowsLEChar = STAFString(kUTF8_CR) + STAFString(kUTF8_LF);
    STAFString unixLEChar = STAFString(kUTF8_LF);

    // Determine the new line ending character(s) to use

    STAFString newLEChar = STAFString("");

    if (textFormat.toUpperCase() == "NATIVE")
    {
        if (!isLocalMachine(orgMachine))
        {
            // Try to get the line separator for the originating machine.
            // If VAR resolve request fails (probably due to insufficient
            // trust), default to target system's line separator.

            STAFResultPtr lineSepResult = gSTAFProcHandlePtr->submit(
                orgMachine, "VAR", "RESOLVE " +
                STAFHandle::wrapData("{STAF/Config/Sep/Line}"));

            if (lineSepResult->rc == kSTAFOk)
            {
                newLEChar = lineSepResult->result;
            }
            else if (lineSepResult->rc == kSTAFInvalidRequestString)
            {
                // The machine is probably a STAF V3.0 system.  Try
                // using the VAR RESOLVE STRING request instead
                lineSepResult = gSTAFProcHandlePtr->submit(
                    orgMachine, "VAR", "RESOLVE STRING " +
                    STAFHandle::wrapData("{STAF/Config/Sep/Line}"));

                if (lineSepResult->rc == kSTAFOk)
                    newLEChar = lineSepResult->result;
            }
        }

        if (newLEChar == STAFString(""))
        {
            // Get the line ending character(s) for the system where the
            // file resides.

            STAFConfigInfo configInfo;
            STAFString_t errorBufferT;
            unsigned int osRC = 0;

            if (STAFUtilGetConfigInfo(&configInfo, &errorBufferT, &osRC) 
                != kSTAFOk)
            {
                 return STAFServiceResult(
                     kSTAFBaseOSError, STAFString(errorBufferT) +
                     ", RC: " + osRC);
            }

            // Default to the target system's line separator if target
            // system is the same as the orginating system (e.g. is local) or
            // if can't get the originating system's line separator.

            newLEChar = configInfo.lineSeparator;
        }
    }
    else if (textFormat.toUpperCase() == "WINDOWS")
        newLEChar = windowsLEChar;
    else if (textFormat.toUpperCase() == "UNIX")
        newLEChar = unixLEChar;
    else
        newLEChar = textFormat;
    
    STAFString result = STAFString(buffer, fileLength);

    // Check which line endings the file actually contains and assign
    // to currLEChar.  The first line-ending found for Windows (0D0A)
    // or Unix (0A) determines the line endings assumed for the file.
    // If no line-endings are found in the file, default to newLEChar 
    // so that no line-endings conversion is done.

    STAFString currLEChar = newLEChar;

    unsigned int windowsLEIndex = result.find(windowsLEChar);
    unsigned int unixLEIndex = result.find(unixLEChar);

    if (windowsLEIndex < unixLEIndex)
        currLEChar = windowsLEChar;
    else if (unixLEIndex < windowsLEIndex)
        currLEChar = unixLEChar;

    // If debug flag is true, prints line ending characters in hex

    bool debug = false;

    if (debug)
    {
        cout << "currLEChar in Hex: ";

        const char *buffer = currLEChar.buffer();

        for (unsigned int y = 0; y < currLEChar.length(); ++y)
        {
            unsigned int currChar = static_cast<unsigned char>(buffer[y]);
            if (currChar < 16) cout << "0";
            cout << hex << currChar << dec << " ";
        }

        cout << endl << "newLEChar in Hex:  ";

        buffer = newLEChar.buffer();

        for (unsigned int i = 0; i < newLEChar.length(); ++i)
        {
            unsigned int currChar = static_cast<unsigned char>(buffer[i]);
            if (currChar < 16) cout << "0";
            cout << hex << currChar << dec << " ";
        }

        cout << endl << endl;
    }

    // If conversion of line ending character(s) is needed, replace the
    // current line ending character(s) with the new line ending character(s)

    if (currLEChar != newLEChar)
        result = result.replace(currLEChar, newLEChar);

    // If testFlag, convert the result to hex to verify that the line-ending
    // characters were converted as expected.

    if (testFlag)
        return convertToHex(result.buffer(), result.length());

    // Return new result in hex format
    return STAFServiceResult(kSTAFOk, result);
}


STAFFSService::STAFFSService() : STAFService("FS")
{
    // COPY FILE parser setup

    fCopyFileParser.addOption("COPY",                 1,
        STAFCommandParser::kValueNotAllowed);
    fCopyFileParser.addOption("FILE",                 1,
        STAFCommandParser::kValueRequired);
    fCopyFileParser.addOption("TOFILE",               1,
        STAFCommandParser::kValueRequired);
    fCopyFileParser.addOption("TOMACHINE",            1,
        STAFCommandParser::kValueRequired);
    fCopyFileParser.addOption("FORMAT",               1,
        STAFCommandParser::kValueRequired);
    fCopyFileParser.addOption("TEXT",                 1, 
        STAFCommandParser::kValueNotAllowed);
    // Remove comment when enabling convert/noconvert option
    /*fCopyFileParser.addOption("CONVERT",              1, 
        STAFCommandParser::kValueNotAllowed);
    fCopyFileParser.addOption("NOCONVERT",            1, 
        STAFCommandParser::kValueNotAllowed);*/
    fCopyFileParser.addOption("FAILIFEXISTS",         1, 
        STAFCommandParser::kValueNotAllowed);
    fCopyFileParser.addOption("FAILIFNEW",            1,
        STAFCommandParser::kValueNotAllowed);

    fCopyFileParser.addOptionGroup("FAILIFEXISTS FAILIFNEW", 0, 1);
    // Remove comment when enabling convert/noconvert option
    //fCopyFileParser.addOptionGroup("CONVERT NOCONVERT", 0, 1);

    fCopyFileParser.addOptionNeed("FORMAT", "TEXT");
    
    // Remove comment when enabling convert/noconvert option
    //fCopyFileParser.addOptionNeed("CONVERT", "TEXT");
    //fCopyFileParser.addOptionNeed("NOCONVERT", "TEXT");
    fCopyFileParser.addOptionNeed("COPY", "FILE");

    // COPY DIRECTORY parser setup

    fCopyDirParser.addOption("COPY",                 1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("DIRECTORY",            1,
        STAFCommandParser::kValueRequired);
    fCopyDirParser.addOption("TODIRECTORY",          1,
        STAFCommandParser::kValueRequired);
    fCopyDirParser.addOption("TOMACHINE",            1,
        STAFCommandParser::kValueRequired);
    fCopyDirParser.addOption("NAME",                 1,
        STAFCommandParser::kValueRequired);
    fCopyDirParser.addOption("EXT",                  1,
        STAFCommandParser::kValueRequired);
    fCopyDirParser.addOption("TEXTEXT",              0,
        STAFCommandParser::kValueRequired);
    fCopyDirParser.addOption("FORMAT",               1,
        STAFCommandParser::kValueRequired);
    // Remove comment when enabling convert/noconvert option
    /*fCopyDirParser.addOption("CONVERT",              1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("NOCONVERT",            1,
        STAFCommandParser::kValueNotAllowed);*/
    fCopyDirParser.addOption("RECURSE",              1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("CASESENSITIVE",        1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("CASEINSENSITIVE",      1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("KEEPEMPTYDIRECTORIES", 1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("ONLYDIRECTORIES",      1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("IGNOREERRORS",         1,
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("FAILIFEXISTS",         1, 
        STAFCommandParser::kValueNotAllowed);
    fCopyDirParser.addOption("FAILIFNEW",            1,
        STAFCommandParser::kValueNotAllowed);

    fCopyDirParser.addOptionGroup("CASESENSITIVE CASEINSENSITIVE", 0, 1);
    fCopyDirParser.addOptionGroup("KEEPEMPTYDIRECTORIES ONLYDIRECTORIES", 0, 1);
    fCopyDirParser.addOptionGroup("FAILIFEXISTS FAILIFNEW", 0, 1);
    // Remove comment when enabling convert/noconvert option
    //fCopyDirParser.addOptionGroup("CONVERT NOCONVERT", 0, 1);

    fCopyDirParser.addOptionNeed("FORMAT", "TEXTEXT");
    // Remove comment when enabling convert/noconvert option
    //fCopyDirParser.addOptionNeed("CONVERT", "TEXTEXT");
    //fCopyDirParser.addOptionNeed("NOCONVERT", "TEXTEXT");
    fCopyDirParser.addOptionNeed("KEEPEMPTYDIRECTORIES ONLYDIRECTORIES",
                              "RECURSE");
    fCopyDirParser.addOptionNeed("COPY", "DIRECTORY");

    // GET parser setup

    fGetParser.addOption("GET",              1,
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("FILE",             1,
        STAFCommandParser::kValueRequired);

    fGetParser.addOption("TEXT",             1,
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("BINARY",           1,
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("FORMAT",           1,
        STAFCommandParser::kValueRequired);

    // Note that TEST is an undocumented option that we can use for testing
    // that the line ending character really got converted correctly
    fGetParser.addOption("TEST",           1,
     STAFCommandParser::kValueNotAllowed);

    fGetParser.addOption("ENTRY",            1,
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("TYPE",             1,
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("SIZE",             1,
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("MODTIME",             1,
        STAFCommandParser::kValueNotAllowed);

    fGetParser.addOptionGroup("FILE ENTRY", 1, 1);

    fGetParser.addOptionGroup("TEXT BINARY", 0, 1);
    fGetParser.addOptionNeed("TEXT BINARY", "FILE");
    fGetParser.addOptionNeed("FORMAT", "TEXT BINARY");
    fGetParser.addOptionNeed("TEST", "FILE");
    
    fGetParser.addOptionNeed("GET", "FILE ENTRY");
    fGetParser.addOptionNeed("TYPE SIZE MODTIME", "ENTRY");
    fGetParser.addOptionNeed("ENTRY", "TYPE SIZE MODTIME");

    // LIST parser setup

    fListParser.addOption("LIST",             1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("DIRECTORY",        1,
        STAFCommandParser::kValueRequired);
    fListParser.addOption("NAME",             1,
        STAFCommandParser::kValueRequired);
    fListParser.addOption("EXT",              1,
        STAFCommandParser::kValueRequired);
    fListParser.addOption("LONG",             1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("FORMAT",           1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SORTBYNAME",       1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SORTBYSIZE",       1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SORTBYMODTIME",    1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("CASESENSITIVE",    1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("CASEINSENSITIVE",  1,
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("TYPE",             1,
        STAFCommandParser::kValueRequired);

    fListParser.addOptionGroup("SORTBYNAME SORTBYSIZE SORTBYTIME", 0, 1);
    fListParser.addOptionGroup("CASESENSITIVE CASEINSENSITIVE", 0, 1);

    fListParser.addOptionNeed("LIST", "DIRECTORY");
    fListParser.addOptionNeed("FORMAT NAME EXT SORTBYNAME", "DIRECTORY");
    fListParser.addOptionNeed("FORMAT", "LONG");
    fListParser.addOptionNeed("CASESENSITIVE CASEINSENSITIVE",
                              "NAME EXT SORTBYNAME");

    // CREATE parser setup

    fCreateParser.addOption("CREATE",           1,
        STAFCommandParser::kValueNotAllowed);
    fCreateParser.addOption("DIRECTORY",        1,
        STAFCommandParser::kValueRequired);
    fCreateParser.addOption("FULLPATH",         1,
        STAFCommandParser::kValueNotAllowed);
    fCreateParser.addOption("FAILIFEXISTS",     1,
        STAFCommandParser::kValueNotAllowed);

    fCreateParser.addOptionNeed("CREATE", "DIRECTORY");

    // DELETE parser setup

    fDeleteParser.addOption("DELETE",           1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("ENTRY",            1,
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("CHILDREN",         1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("NAME",             1,
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("EXT",              1,
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("TYPE",             1,
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("CASESENSITIVE",    1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("CASEINSENSITIVE",  1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("RECURSE",          1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("CONFIRM",          1,
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("IGNOREERRORS",     1,
        STAFCommandParser::kValueNotAllowed);

    fDeleteParser.addOptionGroup("CASESENSITIVE CASEINSENSITIVE", 0, 1);

    fDeleteParser.addOptionNeed("DELETE", "ENTRY");
    fDeleteParser.addOptionNeed("DELETE", "CONFIRM");
    fDeleteParser.addOptionNeed("NAME EXT TYPE", "CHILDREN");
    fDeleteParser.addOptionNeed("CASESENSITIVE CASEINSENSITIVE", "NAME EXT");
    fDeleteParser.addOptionNeed("IGNOREERRORS", "CHILDREN RECURSE");

    // QUERY parser setup

    fQueryParser.addOption("QUERY",            1,
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("ENTRY",            1,
        STAFCommandParser::kValueRequired);

    fQueryParser.addOptionNeed("QUERY", "ENTRY");
}


STAFFSService::~STAFFSService()
{
    /* Do Nothing */
}


STAFString STAFFSService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFFSService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).toLowerCase();
    STAFString subAction = requestInfo.fRequest.subWord(1, 1).toLowerCase();

    if (action == "copy")
    {
        if (subAction == "file")
            return handleCopyFile(requestInfo);
        else if (subAction == "directory")
            return handleCopyDir(requestInfo);
        else return STAFServiceResult(kSTAFInvalidRequestString, "");
    }
    else if (action == "get")    return handleGet(requestInfo);
    else if (action == "list")   return handleList(requestInfo);
    else if (action == "create") return handleCreate(requestInfo);
    else if (action == "delete") return handleDelete(requestInfo);
    else if (action == "query")  return handleQuery(requestInfo);
    else if (action == "help")   return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString, "");
}


STAFServiceResult STAFFSService::handleCopyFile(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fCopyFileParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFString fromMachine = requestInfo.fMachine;
    STAFHandle_t handle = requestInfo.fHandle;
    STAFString request = requestInfo.fRequest;
    STAFString errorBuffer;
    STAFVariablePoolPtr handlePool = getClientVariablePool(fromMachine, handle);
    STAFRC_t rc = kSTAFOk;
    STAFString unresFromFile = parsedResult->optionValue("FILE");
    STAFString toFile = unresFromFile; // toFile defaulted to unresFromFile
    STAFString fromFile;
    STAFString toMachine = fromMachine;

    rc = resolveString(unresFromFile, handlePool, fromFile, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = resolveStringIfExists(parsedResult, "TOMACHINE", handlePool, toMachine,
                               errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    // if user specified TOFILE value
    if (parsedResult->optionTimes("TOFILE") != 0)
        toFile = parsedResult->optionValue("TOFILE");

    if (isLocalMachine(fromMachine) && isLocalMachine(toMachine, 1))
    {
        // If we are copying to and from ourself then we need to resolve
        // any handle specific variables here.  STAFProc is only responsible
        // for resolving variables in the case where the file came from
        // a different machine.

        STAFString unresToFile = toFile;

        rc = resolveString(unresToFile, handlePool, toFile, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
    }
    
    // Check if the from file exists

    STAFFSPath fromPath(fromFile);

    if (!fromPath.exists())
        return STAFServiceResult(kSTAFDoesNotExist, fromFile);

    // Remove trailing slashes, if any, in the file name
    try
    {
        fromFile = fromPath.getEntry()->path().asString();
    }
    catch (STAFBaseOSErrorException &e)
    {
        STAFString result = e.getText() + STAFString(": ") +
            e.getErrorCode() + "\nError on from file: " + fromFile;

        return STAFServiceResult(kSTAFBaseOSError, result);
    }

    // determine if transfer is text or binary
    bool doText = false;
    bool doCodepage = false; 
    STAFString newEOL;

    if (parsedResult->optionTimes("TEXT") > 0)
    {
        doText = true;
        doCodepage = true;
    
        // if text transfer get the new EOL marker
        if (parsedResult->optionTimes("FORMAT") > 0)
        {       
            STAFString unresFormat = parsedResult->optionValue("FORMAT");
            rc = resolveString(unresFormat, handlePool, newEOL, errorBuffer);
            if (rc) return STAFServiceResult(rc, errorBuffer);
        }
        else
            newEOL = "Native";        

        // If enable NOCONVERT option, uncomment the  following line
        //doCodepage = (parsedResult->optionTimes("NOCONVERT") == 0);
    }

    fstream inFile;

    // open the file as binary
    inFile.open(fromFile.toCurrentCodePage()->buffer(),
                 ios::in | STAF_ios_binary);

    if (!inFile)
        return STAFServiceResult(kSTAFFileOpenError, fromFile);

    unsigned char fileBuffer[4000] = { 0 };
    unsigned int writeLength = 0;
    STAFConnPtr connection;
    STAFString result;
    unsigned int useNewAPI = 1;
    unsigned int levelToUse = 0;
    unsigned int flags = 0;

    if (parsedResult->optionTimes("FAILIFEXISTS") != 0)
        flags |= 0x00000001;
    else if (parsedResult->optionTimes("FAILIFNEW") != 0)
        flags |= 0x00000002;

    try
    {
        rc = gConnectionManagerPtr->makeConnection(toMachine, connection,
                                                   result);
        if (rc) return STAFServiceResult(rc, result);

        // First, lets try the new API

        connection->writeUInt(5);    // New API number
        connection->writeUInt(0);    // Dummy level

        STAFRC_t ack = connection->readUInt();

        if (ack != kSTAFOk)
        {
            // They don't support the new API, so try the old API
            useNewAPI = 0;

            rc = gConnectionManagerPtr->makeConnection(
                toMachine, connection, result);

            if (rc) return STAFServiceResult(rc, result);

            connection->writeUInt(4);    // API Number
            connection->writeUInt(0);    // API Level

            ack = connection->readUInt();

            if (ack != kSTAFOk) return ack;
        }
        else
        {
            // Now find out the specific level to use
            unsigned int minLevel = 1;
            unsigned int maxLevel = 2;

            connection->writeUInt(minLevel);
            connection->writeUInt(maxLevel);

            levelToUse = connection->readUInt();

            if (levelToUse == 0) return kSTAFInvalidAPILevel;
        }

        // if the other machine can do a text transfer specify text or binary
        if (useNewAPI && (levelToUse > 1))
            if (doText) 
            {
                if (doCodepage)
                    connection->writeUInt(kSTAFFSTextConvert);
                else
                    connection->writeUInt(kSTAFFSTextNoConvert);
            }
            else 
            {
                connection->writeUInt(kSTAFFSBinary);
            }

        connection->writeString(*gMachineNetNamePtr);
        connection->writeString(fromMachine);
        connection->writeString(toFile);

        if (!(useNewAPI && (levelToUse > 1)) && doText)
        {
            // fail and abort
            flags |= (0x00000001 | 0x00000002);
        
            // send bad flags 0x1|0x2 // fail if new | fail if exist
            // to kill other side
            connection->writeUInt(flags);
            
            // read kill confirmation
            ack = static_cast<STAFRC_t>(connection->readUInt());
            result = connection->readString();
            inFile.close();
            return kSTAFInvalidAPILevel;
        }
        
        connection->writeUInt(flags);

        ack = static_cast<STAFRC_t>(connection->readUInt());
        result = connection->readString();

        if (ack != kSTAFOk) return STAFServiceResult(ack, result);

        STAFString currentEOL;
        if (doText)
        {
            // determine current EOL
            STAFConfigInfo sysInfo;
            STAFString_t errorBuffer;
            unsigned int osRC;
            
            if (STAFUtilGetConfigInfo(&sysInfo, &errorBuffer, &osRC) 
                != kSTAFOk)
            {
                return STAFServiceResult(kSTAFBaseOSError, 
                                        "Get current EOL failure");
            }
            currentEOL = sysInfo.lineSeparator;
            
            // send the new EOL marker across
            connection->writeString(newEOL);
            
            // verify ok new eol
            if (connection->readUInt() == kSTAFFSStopCopy)
                return STAFServiceResult(kSTAFBaseOSError, 
                                        "Get current EOL failure");

            newEOL = connection->readString();

            /*
                This clause allows the use of the binary transfer if
                the eol on both sides are the same and no convert is needed.
                This allows for a faster transfer.
            */
            if (newEOL.isEqualTo(currentEOL))
            {
                connection->writeUInt(kSTAFFSBinary);
                doText = false;
            }
            else
                connection->writeUInt(kSTAFFSTextNoConvert);
        }
        /*
           This clause allows the use of the text noconvert routine if
           the codepages on both sides are the same.
           This allows for a faster transfer.
        */
        if (doCodepage)
        {
            // get this codepage id
            STAFString cPage;
            
            rc = resolveString("{STAF/Config/CodePage}", 0, cPage, errorBuffer);

            // get other codepage id
            STAFString otherCPage = connection->readString();

            if (cPage.isEqualTo(otherCPage))
            {
                // change this line on force convert
                connection->writeUInt(kSTAFFSTextConvert);

                // comment this line to disable the codepage filter and force 
                //  codepage conversion on all text transfers
                //doCodepage = false;
                //cout<<"Codepages equal"<<endl;
            }
            else
                connection->writeUInt(kSTAFFSTextConvert);
        }

        /*  This if clause is used to perform a text transfer with 
            conversion of eol chars and a codepage conversion
            It runs considerably slower than the binary    transfer.
            This is the type of transfer that will occur if the TEXT
            option is specified and the codepages are different.
            Data is sent as a series of bufers of UTF8 chars
        */
        if (doCodepage)
        {
            // Get the length of the file

            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();
            inFile.seekg(0, ios::beg);

            connection->writeString(currentEOL);

            STAFStringBufferPtr newlineStr = currentEOL.toCurrentCodePage();
            
            // How many bytes in the end-of-line sequence
            unsigned int newlineBufferSize = newlineStr->length();  
            
            // Points to a buffer containing end-of-line sequence
            char * newlineBuffer = new char [newlineBufferSize];  
            memcpy(newlineBuffer, newlineStr->buffer(), newlineBufferSize);  
            
            // Last byte of end-of-line sequence
            char newlineKey     = newlineBuffer[newlineBufferSize - 1];  
            bool done = false;
            const unsigned int defaultBufferSize = 4096;

            STAFRefPtr<char> buffer = STAFRefPtr<char>(new char[defaultBufferSize],
                                                       STAFRefPtr<char>::INIT,
                                                       STAFRefPtr<char>::ARRAY);
            unsigned int bufferSize = defaultBufferSize;
            unsigned int bufferReadOffset = 0;

            while (!done)
            {
                rc = readFile(inFile,
                              static_cast<char *>(buffer + bufferReadOffset),
                              bufferSize - bufferReadOffset, fromFile,
                              toMachine, fileLength);
                
                if (rc != kSTAFOk)
                {
                    result = "Unrecoverable read error occurred while"
                             " copying file " + fromFile + " in text mode";
                    break;
                }

                unsigned int bytesInBuffer = inFile.gcount() + bufferReadOffset;

                if (bytesInBuffer < bufferSize) done = true;

                // Find a newline.  Make sure we don't underrun the buffer;

                unsigned int i = bytesInBuffer - 1;  // Last NewLine index
                
                unsigned int bufferGuardIndex = newlineBufferSize - 1;

                while (((buffer[i] != newlineKey) ||
                        !memcmp(buffer + i - newlineBufferSize,
                                newlineBuffer, newlineBufferSize)) &&
                       (i > bufferGuardIndex))
                { --i; }

                while ((i == bufferGuardIndex) && !done)
                {
                    // We have a line bigger than our buffer.
                    // Note: the beginning of the buffer may be a lone newline, but we
                    //       ignore that for this algorithm (as the next line is likely
                    //       larger than the buffer any way

                    // First, create a buffer that is double our current size, and copy
                    // our existing buffer data into it

                    STAFRefPtr<char> tmpBuffer =
                                     STAFRefPtr<char>(new char[bufferSize * 2],
                                                      STAFRefPtr<char>::INIT,
                                                      STAFRefPtr<char>::ARRAY);
                    memcpy(tmpBuffer, buffer, bufferSize);
                    buffer = tmpBuffer;
                    bufferSize *= 2;

                    // Now, read in enough data to fill the remainder of the buffer

                    rc = readFile(inFile, buffer + (bufferSize / 2),
                                  bufferSize / 2, fromFile, toMachine,
                                  fileLength);

                    if (rc != kSTAFOk)
                    {
                        result = "Unrecoverable read error occurred while"
                                 " copying file " + fromFile + " in text mode";
                        break;
                    }

                    bytesInBuffer += inFile.gcount();

                    // Finally, let's check to make sure that this buffer was
                    // big enough by finding a newline.  Otherwise, let's run
                    // the loop again.

                    if (bytesInBuffer < bufferSize) done = true;

                    i = bytesInBuffer - 1;  // Last NewLine index
                    bufferGuardIndex = (bufferSize / 2) - newlineBufferSize;

                    while (((buffer[i] != newlineKey) ||
                            !memcmp(buffer + i - newlineBufferSize,
                                    newlineBuffer, newlineBufferSize)) &&
                           (i > bufferGuardIndex))
                    { --i; }

                } // while ((i == bufferGuardIndex) && !done)

                // We now have the last newline in the buffer

                if (rc == kSTAFOk)
                {
                    if (!done)
                    {
                        connection->writeUInt(kSTAFFSContinueCopy);
                        connection->writeString(
                            STAFString(buffer, i + 1, STAFString::kCurrent));
                        memmove(buffer, buffer + i + 1, bufferSize - i - 1);
                        bufferReadOffset = bufferSize - i - 1;
                    }
                    else
                    {
                        connection->writeUInt(kSTAFFSContinueCopy);
                        connection->writeString(
                            STAFString(buffer, bytesInBuffer,
                                       STAFString::kCurrent));
                    }
                }
            } // while (!done)

            free (newlineBuffer);
            connection->writeUInt(kSTAFFSFinishedCopy);

            inFile.close();    

            if (rc == kSTAFOk)
            {
                // Read an ack, so that we know the file is closed and
                // if the file was copied successfully
                rc = connection->readUInt();
            }
        }
        /*  This if clause is used to perform a text transfer with 
            conversion of eol chars without a codepage conversion
            It runs considerably faster than the codepage conversion 
            transfer.
            This is the type of transfer that will occur if the NOCONVERT
            option is enabled
            This is the type of transfer that will occur if the codepages 
            are the same and a text transfer has been specified
        */
        else if (doText)
        {
            // Text file copy without codepage conversion

            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();
            inFile.seekg(0, ios::beg);
            
            STAFString transferString;
            int bufferSize = 3000;
            unsigned int transferLen = 0;
            char * buffer = new char[bufferSize]; 
            
            connection->writeString(currentEOL);
            connection->writeUInt(bufferSize);
            
            while ((fileLength > 0) && (inFile.good()))
            {
                transferLen = STAF_MIN(bufferSize, fileLength);
            
                rc = readFile(inFile, buffer, transferLen, fromFile,
                              toMachine, fileLength);
                    
                if (rc != kSTAFOk)
                {
                    result = "Unrecoverable read error occurred while"
                         " copying file " + fromFile +
                         " in text (no codepage convert) mode";
                    break;
                }

                connection->writeUInt(transferLen);
                connection->write(buffer,transferLen);
                
                fileLength -= transferLen;
            }

            connection->writeUInt(kSTAFFSFinishedCopy);
            free(buffer);

            inFile.close();    

            if (rc == kSTAFOk)
            {
                // Read an ack, so that we know the file is closed and
                // if the file was copied successfully
                rc = connection->readUInt();
            }
        }
        else if (useNewAPI)
        {
            // Binary file copy (using new API)
            
            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();
            connection->writeUInt(fileLength);
            inFile.seekg(0, ios::beg);
            
            while ((fileLength > 0) && inFile.good())
            {
                writeLength = STAF_MIN(sizeof(fileBuffer), fileLength);

                rc = readFile(inFile, reinterpret_cast<char *>(fileBuffer),
                              writeLength, fromFile, toMachine, fileLength);

                if (rc != kSTAFOk)
                {
                    result = "Unrecoverable read error occurred while"
                         " copying file " + fromFile +
                         " in binary mode";
                    break;
                }
                
                connection->write(fileBuffer, writeLength);
                fileLength -= writeLength;
             }
             
             inFile.close();

             if (rc == kSTAFOk)
             {
                 // Read an ack, so that we know the file is closed and
                 // if the file was copied successfully
                 rc = connection->readUInt();
             }
        }
        else
        {
            // Binary file copy (using old API)

            // Get the length of the file

            inFile.seekg(0, ios::end);
            unsigned int fileLength = (unsigned int)inFile.tellg();
            inFile.seekg(0, ios::beg);

            do
            {
                rc = readFile(inFile, reinterpret_cast<char *>(fileBuffer),
                              sizeof(fileBuffer), fromFile, toMachine,
                              fileLength);

                if (rc != kSTAFOk)
                {
                    result = "Unrecoverable read error occurred while"
                         " copying file " + fromFile +
                         " in binary mode";
                    break;
                }

                writeLength = inFile.gcount();
                connection->writeUInt(writeLength);

                if ((writeLength == 0) && !inFile.eof())
                {
                   return STAFServiceResult(kSTAFFileReadError, fromFile);
                }
                else if (writeLength != 0)
                {
                    connection->write(fileBuffer, writeLength);
                    ack = connection->readUInt();
                    if (ack) result = connection->readString();
                }
            }while ((writeLength != 0) && !ack && inFile.good());

            inFile.close(); 

            if (ack != kSTAFOk)
            {
                return STAFServiceResult(ack, result);
            }
            else if (writeLength != 0)
            {
                 // Tell the other side we are finished

                 connection->writeUInt(0);

                  // Read the final ack

                  try
                  {
                      ack = connection->readUInt();
                  }
                  catch (STAFConnectionIOException)
                  {
                     /* Older clients will close the connection instead of
                       acking so we need to ignore Connection IO exceptions here
                      */
                      ack = kSTAFOk;
                  }

                  if (ack != kSTAFOk)
                     return STAFServiceResult(ack, result);
              }
        }
    }
    catch (STAFConnectionConnectException &e)
    {
        rc = kSTAFNoPathToMachine;
        result = e.getText() + STAFString(": ") + STAFString(e.getErrorCode());
    }
    catch (STAFConnectionIOException &e)
    {
        rc = kSTAFCommunicationError;
        result = e.getText() + STAFString(": ") + STAFString(e.getErrorCode());
    }
    catch (...)
    {
        cout << "Caught unknown exception in STAFFSService::handleCopyFile()"
             << endl;
    }

    return STAFServiceResult(rc, result);
}


STAFServiceResult STAFFSService::handleCopyDir(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fCopyDirParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFString fromMachine = requestInfo.fMachine;
    STAFHandle_t handle = requestInfo.fHandle;
    STAFString request = requestInfo.fRequest;
    STAFString errorBuffer;
    STAFVariablePoolPtr handlePool = getClientVariablePool(fromMachine, handle);
    STAFRC_t rc = kSTAFOk;
    bool recurse = (parsedResult->optionTimes("RECURSE") > 0);
    bool keepEmptyDir = (parsedResult->optionTimes("KEEPEMPTYDIRECTORIES") > 0);
    bool onlyDir = (parsedResult->optionTimes("ONLYDIRECTORIES") > 0);
    bool ignoreErrors = (parsedResult->optionTimes("IGNOREERRORS") > 0);
    STAFString namePattern(kUTF8_STAR);
    STAFString extPattern(kUTF8_STAR);
    unsigned int entryTypesUInt = kSTAFFSFile | kSTAFFSDirectory;
    STAFFSCaseSensitive_t caseSensitive = kSTAFFSCaseDefault;
    STAFString unresFromDir = parsedResult->optionValue("DIRECTORY");
    STAFString toDir = unresFromDir;
    STAFString fromDir;
    STAFString toMachine = fromMachine;
    STAFString typeString;
    STAFString result;
    STAFString ackResult;

    if (!ignoreErrors && requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a COPY DIRECTORY request
        // without the IGNOREERRORS option specified will change in STAF V3.0

        gDiagManagerPtr->record(
            "STAF/V3.0-Mig/FS COPY DIRECTORY (No IGNOREERRORS)",
            requestInfo.fHandleName,
            requestInfo.fMachine,
            requestInfo.fHandle);
    }
    
    rc = resolveString(unresFromDir, handlePool, fromDir, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = resolveStringIfExists(parsedResult, "TOMACHINE", handlePool, 
                               toMachine, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    // if user specified TODIRECTORY value
    if (parsedResult->optionTimes("TODIRECTORY") != 0)
        toDir = parsedResult->optionValue("TODIRECTORY");

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "NAME", handlePool,
                                   namePattern, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "EXT", handlePool,
                                   extPattern, errorBuffer);
    }

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (parsedResult->optionTimes("CASESENSITIVE") != 0)
        caseSensitive = kSTAFFSCaseSensitive;
    else if (parsedResult->optionTimes("CASEINSENSITIVE") != 0)
        caseSensitive = kSTAFFSCaseInsensitive;

    // Obtain directory path from the value of DIRECTORY
    STAFFSPath directoryPath(fromDir);

    if (!directoryPath.exists()) 
    {
        ackResult = "Directory, " + fromDir + ", does Not Exist";
        return STAFServiceResult(kSTAFDoesNotExist, ackResult);
    }

    STAFFSEntryPtr entry;

    try
    {
        entry = directoryPath.getEntry();

        // Remove trailing slashes in the directory name, if any
        fromDir = entry->path().asString();
    }
    catch (STAFBaseOSErrorException &e)
    {
        result = e.getText() + STAFString(": ") + e.getErrorCode() +
            "\nError on Source Directory: " + fromDir;

        return STAFServiceResult(kSTAFBaseOSError, result);
    }

    // if the value of DIRECTORY is not a directory, then stop here
    if (entry->type() != kSTAFFSDirectory)
    {
        ackResult = fromDir + " is Not a Directory";
        return STAFServiceResult(kSTAFInvalidValue, ackResult);
    }
    
    STAFConnPtr connection;
    unsigned int levelToUse = 0;
    unsigned int flags = 0;

    if (parsedResult->optionTimes("FAILIFEXISTS") != 0)
        flags |= 0x00000001;
    else if (parsedResult->optionTimes("FAILIFNEW") != 0)
        flags |= 0x00000002;

    STAFString newEOL;
    if (parsedResult->optionTimes("FORMAT") != 0)
    {
        STAFString unresFormat = parsedResult->optionValue("FORMAT");
        rc = resolveString(unresFormat, handlePool,newEOL, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
    }
    else
        newEOL = "Native";

    /************** Start transferring the directory ***************/

    try
    {
        // Make a connection to the toMachine

        rc = gConnectionManagerPtr->makeConnection(toMachine, connection,
                                                   result);
        if (rc) return STAFServiceResult(rc, result);

        // First, lets try the new API

        connection->writeUInt(6);    // New API number
        connection->writeUInt(0);    // Dummy level

        STAFRC_t ack = connection->readUInt();

        // The client don't support the new API
        if (ack != kSTAFOk)
            return STAFServiceResult(ack, "No Support for Directory Copy API");
        
        // Now find out the specific level to use
        unsigned int minLevel = 1;
        unsigned int maxLevel = 2;

        connection->writeUInt(minLevel);
        connection->writeUInt(maxLevel);

        levelToUse = connection->readUInt();

        if (levelToUse == 0)
            return STAFServiceResult(kSTAFInvalidAPILevel, 
                                     "Invalid API Level");
        
        int numTextExtentions = parsedResult->optionTimes("TEXTEXT");


        // this prevents attempts to transfer if invalid api
        if ((levelToUse < 2) && ( numTextExtentions != 0))
            return STAFServiceResult(kSTAFInvalidAPILevel, 
                                     "Invalid API Level");
        
        STAFString currentEOL;
        
        // determine current EOL
        STAFConfigInfo sysInfo;
        STAFString_t eBuff;
        unsigned int osRC;
        if (STAFUtilGetConfigInfo(&sysInfo, &eBuff, &osRC) 
             != kSTAFOk)
        {
            // kill transfer
            return STAFServiceResult(kSTAFBaseOSError,
                                         "Get current EOL failure");
        }
        currentEOL = sysInfo.lineSeparator;

        connection->writeString(*gMachineNetNamePtr);
        connection->writeString(fromMachine);
        connection->writeString(toDir);   // write the root of toDir
        connection->writeUInt(flags);

        if (levelToUse > 1)
            connection->writeString(newEOL);
        ack = static_cast<STAFRC_t>(connection->readUInt());
        ackResult = connection->readString();

        if (ack != kSTAFOk) 
            return STAFServiceResult(ack, ackResult);

        bool doCodepageConvert = false;

        if (levelToUse > 1)
        {
            newEOL = connection->readString();

            if (numTextExtentions > 0)
            {   
                doCodepageConvert = true;
                // If enable NOCONVERT option, uncomment the following line
                //doCodepageConvert = (parsedResult->optionTimes("NOCONVERT") == 0);
                connection->writeUInt(kSTAFFSTextConvert);
            }
            else
            {
                connection->writeUInt(kSTAFFSBinary);
            }
        }

        /*
           This clause allows the use of the text noconvert routine if
           the codepages on both sides are the same.
           This allows for a faster transfer.
        */
        if (doCodepageConvert)
        {
            // get this codepage id
            STAFString cPage;
            rc = resolveString("{STAF/Config/CodePage}", 0, cPage, errorBuffer);

            // get other codepage id
            STAFString otherCPage = connection->readString();

            if (cPage.isEqualTo(otherCPage))
            {
                // comment this line to disable the codepage filter
                //  force codepage conversion if convert requested
                //doCodepageConvert = false;
            }
        }

        // add the text extensions to a list
        STAFString unresExt, resExt;
        SET_STAFString TextExtensionList;
        for (int i = 0; i < numTextExtentions; i++)
        {
            unresExt = parsedResult->optionValue("TEXTEXT", i + 1);  
            rc = resolveString(unresExt, handlePool, resExt, errorBuffer);
            if (!rc)
                TextExtensionList.insert(resExt);
        }

        if (!recurse)
        {
            STAFFSEnumPtr fileEnum = entry->enumerate(namePattern, extPattern,
                STAFFSEntryType_t(entryTypesUInt & ~kSTAFFSDirectory), 
                kSTAFFSNoSort, caseSensitive);

            if (fileEnum->isValid())
                copyDirectory(fileEnum, fromDir, connection, 
                            TextExtensionList, currentEOL, newEOL, 
                            doCodepageConvert, levelToUse, caseSensitive, 
                            result, toMachine);
        }
        else
        {
            recurseCopyDir(entry, namePattern, extPattern, fromDir,
                          keepEmptyDir, onlyDir, entryTypesUInt,
                          connection, caseSensitive, TextExtensionList,
                          currentEOL, newEOL, levelToUse, doCodepageConvert,
                          result, toMachine);
        }

        // Stop copying files/directories
        connection->writeUInt(kSTAFFSStopCopy);

        if (!ignoreErrors && (result.length() != 0))
            rc = kSTAFDirectoryCopyError;
        else
        {
            rc = kSTAFOk;
            result = STAFString();
        }
    }
    catch (STAFConnectionConnectException &e)
    {
        rc = kSTAFNoPathToMachine;
        result = e.getText() + STAFString(": ") + STAFString(e.getErrorCode());
    }
    catch (STAFConnectionIOException &e)
    {
        rc = kSTAFCommunicationError;
        result = e.getText() + STAFString(": ") + STAFString(e.getErrorCode());
    }
    catch (...)
    {
        rc = kSTAFDirectoryCopyError;
        result = "Caught unknown exception in STAFFSService::handleCopyDir() "
            "while copying to directory " + toDir + " from machine " +
            fromMachine + "\n";
    }

    return STAFServiceResult(rc, result);
}

STAFServiceResult STAFFSService::handleGet(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult =
                              fGetParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                                                      

    STAFVariablePoolPtr handlePool =
        getClientVariablePool(requestInfo.fMachine, requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    STAFString errorBuffer;
    STAFString result;

    if (parsedResult->optionTimes("FILE") != 0)
    {
        // GET FILE request

        if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;    

        // Determine if Text or Binary

        bool convertToText = true;         // Defaults to text

        if (parsedResult->optionTimes("BINARY") != 0) convertToText = false;

        // Determine the format

        STAFString format;
        
        if (parsedResult->optionTimes("FORMAT") != 0)
        {
            // Resolve the value specified for the FORMAT option

            STAFString unresFormat = parsedResult->optionValue("FORMAT");
 
            rc = resolveString(unresFormat, handlePool, format, errorBuffer);

            if (rc) return STAFServiceResult(rc, errorBuffer);

            // If Binary, verify that the format specified is Hex

            if (!convertToText && format.toUpperCase() != "HEX")
                return STAFServiceResult(kSTAFInvalidValue, format);
        }
        else
        {
            if (convertToText)
                format = "NATIVE";  // Default Text format is Native
            else
                format = "HEX";     // Default Binary format is Hex
        }
        
        // Check if the undocumented TEST option was specified

        bool testFlag = false;

        if (parsedResult->optionTimes("TEST") != 0)
            testFlag = true;

        // Resolve the value specified for FILE

        STAFString unresFile = parsedResult->optionValue("FILE");
        STAFString file;

        rc = resolveString(unresFile, handlePool, file, errorBuffer);

        if (rc) return STAFServiceResult(rc, errorBuffer);

        // Check if the file exists

        STAFFSPath path(file);

        if (!path.exists())
            return STAFServiceResult(kSTAFDoesNotExist, file);

        // Remove trailing slashes, if any, in the file name
        try
        {
            file = path.getEntry()->path().asString();
        }
        catch (STAFBaseOSErrorException &e)
        {
            result = e.getText() + STAFString(": ") +
                e.getErrorCode() + "\nError on file: " + file;

            return STAFServiceResult(kSTAFBaseOSError, result);
        }

        // Open the file to read it

        fstream inFile(file.toCurrentCodePage()->buffer(),
                       ios::in | STAF_ios_binary);

        if (!inFile)
            return STAFServiceResult(kSTAFFileOpenError, file);

        // Figure out how big the file is

        inFile.seekg(0, ios::end);
        unsigned int fileLength = (unsigned int)inFile.tellg();

        // Initialize the output buffer and read in the file

        STAFBuffer<char> buffer(new char[fileLength], STAFBuffer<char>::INIT, 
                                STAFBuffer<char>::ARRAY);

        inFile.seekg(0, ios::beg);
        inFile.read(buffer, fileLength);
        
        if (!convertToText)
        {
            // Convert the file contents to a hex format
            return convertToHex(buffer, fileLength);
        }

        // Don't convert line-endings in the text file if format is "AsIs"

        if (format.toUpperCase() == "ASIS")
        {
            // If the undocumented TEST option was specified.
        
            if (testFlag)
            {
                // Return the result in hex to verify that the line-ending
                // characters were not converted.
                return convertToHex(buffer, fileLength);
            }

            return STAFServiceResult(rc, STAFString(buffer, fileLength));
        }

        // Convert the line ending characters in the text file

        return convertLineEndings(buffer, fileLength, format, 
                                  requestInfo.fMachine, testFlag);
    }
    else
    {
        // GET ENTRY request

        if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;    

        STAFString unresEntryName = parsedResult->optionValue("ENTRY");
        STAFString entryName;

        rc = resolveString(unresEntryName, handlePool, entryName, errorBuffer);

        if (rc) return STAFServiceResult(rc, errorBuffer);

        STAFFSPath path(entryName);

        if (!path.exists())
            return STAFServiceResult(kSTAFDoesNotExist, entryName);

        STAFFSEntryPtr entry = path.getEntry();

        if (parsedResult->optionTimes("TYPE") != 0)
        {
            result = getTypeString(entry->type());
        }
        else if (parsedResult->optionTimes("SIZE") != 0)
        {
            if (requestInfo.fDiagEnabled)
            {
                // Record diagnostics because output for a GET ENTRY SIZE
                // request will change in STAF V3.0

                gDiagManagerPtr->record("STAF/V3.0-Mig/FS GET ENTRY SIZE",
                                        requestInfo.fHandleName,
                                        requestInfo.fMachine,
                                        requestInfo.fHandle);
            }

            result = STAFString(entry->size().first) + kUTF8_COLON +
                STAFString(entry->size().second);
        }
        else  // MODTIME
        {
            result = entry->modTime().asString();
        }
    
        return STAFServiceResult(rc, result);
    }
}


STAFServiceResult STAFFSService::handleList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fListParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                                                      

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a LIST request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/FS LIST",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool =
        getClientVariablePool(requestInfo.fMachine, requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    
    bool showLong = (parsedResult->optionTimes("LONG") > 0);
    bool format = (parsedResult->optionTimes("FORMAT") > 0);
    STAFString namePattern(kUTF8_STAR);
    STAFString extPattern(kUTF8_STAR);
    STAFFSSortBy_t sortBy = kSTAFFSNoSort;
    STAFFSCaseSensitive_t caseSensitive = kSTAFFSCaseDefault;
    unsigned int entryTypesUInt = kSTAFFSNormal;
    STAFString unresDir = parsedResult->optionValue("DIRECTORY");
    STAFString dir;
    STAFString typeString;
    STAFString result;

    rc = resolveString(unresDir, handlePool, dir, errorBuffer);

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "NAME", handlePool,
                                   namePattern, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "EXT", handlePool,
                                   extPattern, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "TYPE", handlePool,
                                   typeString, errorBuffer);
    }

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (parsedResult->optionTimes("SORTBYNAME") != 0)
        sortBy = kSTAFFSSortByName;
    else if (parsedResult->optionTimes("SORTBYSIZE") != 0)
        sortBy = kSTAFFSSortBySize;
    else if (parsedResult->optionTimes("SORTBYMODTIME") != 0)
        sortBy = kSTAFFSSortByModTime;

    if (parsedResult->optionTimes("CASESENSITIVE") != 0)
        caseSensitive = kSTAFFSCaseSensitive;
    else if (parsedResult->optionTimes("CASEINSENSITIVE") != 0)
        caseSensitive = kSTAFFSCaseInsensitive;

    if (typeString.toUpperCase() == "ALL")
        entryTypesUInt = kSTAFFSAll;
    else if (typeString.length() != 0)
        entryTypesUInt = getTypeMaskFromString(typeString, true);

    STAFFSPath dirPath(dir);

    if (!dirPath.exists()) return STAFServiceResult(kSTAFDoesNotExist, dir);

    STAFFSEnumPtr dirEnum =
                  dirPath.getEntry()->enumerate(namePattern, extPattern,
                  STAFFSEntryType_t(entryTypesUInt), sortBy, caseSensitive);
    STAFFSEntryPtr entry;

    for (; dirEnum->isValid(); dirEnum->next())
    {
        entry = dirEnum->entry();

        STAFString theName =
            entry->path().setRoot(STAFString()).clearDirList().asString();

        if (showLong && !format)
        {
            result += theName;
            result += kUTF8_SCOLON;
            result += getTypeString(entry->type());
            result += kUTF8_SCOLON;
            result += entry->size().first;
            result += kUTF8_COLON;
            result += entry->size().second;
            result += kUTF8_SCOLON;
            result += entry->modTime().asString();
        }
        else if (showLong && format)
        {
            result += getTypeString(entry->type());
            result += kUTF8_SPACE;

            STAFString sizeInfo;

            if (entry->size().first == 0)
            {
                if (entry->size().second > (99999 * 1024))
                {
                    sizeInfo = (entry->size().second / (1024 * 1024));
                    sizeInfo += "M";
                }
                else if (entry->size().second > 99999)
                {
                    sizeInfo = (entry->size().second / 1024);
                    sizeInfo += "K";
                }
                else sizeInfo = entry->size().second;
            }
            else sizeInfo = "*****";

            // 5 is the maximum length of sizeInfo + 1

            for (int i = 5 - sizeInfo.length(); i > 0; --i)
                result += kUTF8_SPACE;

                result += sizeInfo;
                result += kUTF8_SPACE;
                result += entry->modTime().asString();
                result += kUTF8_SPACE;
                result += theName;
        }
        else result += theName;

        result += *gLineSeparatorPtr;
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFFSService::handleCreate(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fCreateParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                                                      

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool =
        getClientVariablePool(requestInfo.fMachine, requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    unsigned int osRC = 0;
    STAFFSDirectoryCreateMode_t createMode = kSTAFFSCreateDirOnly;
    STAFString unresDir = parsedResult->optionValue("DIRECTORY");
    STAFString dir;
    STAFString result;

    rc = resolveString(unresDir, handlePool, dir, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (parsedResult->optionTimes("FULLPATH") != 0)
        createMode = kSTAFFSCreatePath;

    rc = STAFFSCreateDirectory(dir.getImpl(), createMode, &osRC);

    if (rc == kSTAFBaseOSError) result = osRC;
    else if ((rc == kSTAFAlreadyExists) &&
             (parsedResult->optionTimes("FAILIFEXISTS") == 0))
    {
        rc = kSTAFOk;
    }
    else if (rc == kSTAFAlreadyExists) result = dir;

    return STAFServiceResult(rc, result);
}


STAFServiceResult STAFFSService::handleDelete(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fDeleteParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                                                      

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool =
        getClientVariablePool(requestInfo.fMachine, requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    bool children = (parsedResult->optionTimes("CHILDREN") > 0);
    bool recurse = (parsedResult->optionTimes("RECURSE") > 0);
    bool ignoreErrors = (parsedResult->optionTimes("IGNOREERRORS") > 0);
    STAFString namePattern(kUTF8_STAR);
    STAFString extPattern(kUTF8_STAR);
    STAFFSCaseSensitive_t caseSensitive = kSTAFFSCaseDefault;
    unsigned int entryTypesUInt = kSTAFFSAll & ~kSTAFFSSpecialDirectory;
    STAFString unresEntry = parsedResult->optionValue("ENTRY");
    STAFString entryString;
    STAFString typeString;
    STAFString result;

    if (recurse && (requestInfo.fTrustLevel < 5))
        return kSTAFAccessDenied;

    if (!ignoreErrors && requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a DELETE request without
        // the IGNOREERRORS option specified will change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/FS DELETE (No IGNOREERRORS)",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    rc = resolveString(unresEntry, handlePool, entryString, errorBuffer);

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "NAME", handlePool,
                                   namePattern, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "EXT", handlePool,
                                   extPattern, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "TYPE", handlePool,
                                   typeString, errorBuffer);
    }

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (parsedResult->optionTimes("CASESENSITIVE") != 0)
        caseSensitive = kSTAFFSCaseSensitive;
    else if (parsedResult->optionTimes("CASEINSENSITIVE") != 0)
        caseSensitive = kSTAFFSCaseInsensitive;

    if (typeString.toUpperCase() == "ALL")
        entryTypesUInt = kSTAFFSAll;
    else if (typeString.length() != 0)
        entryTypesUInt = getTypeMaskFromString(typeString, false);

    STAFFSPath entryPath(entryString);

    if (!entryPath.exists())
        return STAFServiceResult(kSTAFDoesNotExist, entryString);

    STAFFSEntryPtr entry = entryPath.getEntry();
    unsigned int osRC = 0;

    if (recurse)
    {
        recurseRemove(entry, namePattern, extPattern, entryTypesUInt,
                      caseSensitive, result);

        if (!children)
        {
            rc = entry->remove(&osRC);
            updateResultString(result, entry, rc, osRC);
        }
    }
    else if (children)
    {
        removeChildren(entry, namePattern, extPattern, entryTypesUInt,
                       caseSensitive, result);
    }
    else
    {
        rc = entry->remove(&osRC);
        updateResultString(result, entry, rc, osRC);
    }

    if (recurse || children)
    {
        if (result.length() != 0)
            rc = kSTAFFileDeleteError;
        
        if (ignoreErrors)
            result = STAFString();
    }
    else
    {
        if (rc == kSTAFBaseOSError)
            rc = kSTAFFileDeleteError;
    }

    return STAFServiceResult(rc, result);
}


STAFServiceResult STAFFSService::handleQuery(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fQueryParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                                                      

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a QUERY request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/FS QUERY",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool =
        getClientVariablePool(requestInfo.fMachine, requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    STAFString unresEntryName = parsedResult->optionValue("ENTRY");
    STAFString entryName;
    STAFString result;

    rc = resolveString(unresEntryName, handlePool, entryName, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFFSPath path(entryName);

    if (!path.exists()) return STAFServiceResult(kSTAFDoesNotExist, entryName);

    STAFFSEntryPtr entry = path.getEntry();

    result += "Name    : ";
    result += entry->path().asString();
    result += *gLineSeparatorPtr;
    result += "Type    : ";
    result += getTypeString(entry->type());
    result += *gLineSeparatorPtr;
    result += "Size    : ";
    result += entry->size().first;
    result += kUTF8_COLON;
    result += entry->size().second;
    result += *gLineSeparatorPtr;
    result += "Modified: ";
    result += entry->modTime().asString();
    result += *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFFSService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString help("FS Service Help" + *gLineSeparatorPtr + 
                    *gLineSeparatorPtr);

    help += "COPY   FILE <Name> [TOFILE <Name>] [TOMACHINE <Machine>]" +
            *gLineSeparatorPtr;
            // Remove comment when enabling convert/noconvert option
    help += "       [TEXT [FORMAT <Format>]]"/* [CONVERT | NOCONVERT]]" */
         + *gLineSeparatorPtr;
    help += "       [FAILIFEXISTS | FAILIFNEW]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "COPY   DIRECTORY <Name> [TODIRECTORY <Name>] "
            "[TOMACHINE <Machine>]" + *gLineSeparatorPtr;
    help += "       [NAME <Pattern>] [EXT <Pattern>] "
            "[CASESENSITIVE | CASEINSENSITIVE]" + *gLineSeparatorPtr;
    help += "       [TEXTEXT <Pattern>... [FORMAT <Format>]] ";
            // Remove comment when enabling convert/noconvert option
    help += /*"[CONVERT | NOCONVERT]]" +*/ *gLineSeparatorPtr;
    help += "       [RECURSE [KEEPEMPTYDIRECTORIES | ONLYDIRECTORIES]]" +
            *gLineSeparatorPtr;
    help += "       [IGNOREERRORS] [FAILIFEXISTS | FAILIFNEW]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "GET    FILE <Name> [[TEXT | BINARY] [FORMAT <Format>]]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "GET    ENTRY <Name> <TYPE | SIZE | MODTIME>" + *gLineSeparatorPtr +
            *gLineSeparatorPtr;
    help += "QUERY  ENTRY <Name>" + *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "CREATE DIRECTORY <Name> [FULLPATH] [FAILIFEXISTS]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "LIST   DIRECTORY <Name> [NAME <Pattern>] [EXT <Pattern>] "
            "[TYPE <Types>]" + *gLineSeparatorPtr;
    help += "       [SORTBYNAME | SORTBYSIZE | SORTBYMODTIME]" +
            *gLineSeparatorPtr;
    help += "       [CASESENSITIVE | CASEINSENSITIVE] [LONG [FORMAT]]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "DELETE ENTRY <Name> [CHILDREN [NAME <Pattern>] [EXT <Pattern>] "
            "[TYPE <Types>]" + *gLineSeparatorPtr;
    help += "       [CASESENSITIVE | CASEINSENSITIVE]] [RECURSE] [IGNOREERRORS]"
            " CONFIRM" + *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, help);
}
