/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFHandleManager.h"
#include "STAFProcess.h"
#include "STAFTrace.h"

static STAFHandleManager *pHandleManager = 0;


STAFHandleManager::STAFHandleManager() : fNextHandle(1)
{
    pHandleManager = this;
}


void STAFHandleManager::processTerminated(STAFProcessID_t pid,
                                           STAFProcessHandle_t procHandle,
                                           unsigned int rc, void *)
{
    pHandleManager->handleProcessTerminated(pid, procHandle);
}


void STAFHandleManager::handleProcessTerminated(STAFProcessID_t pid,
                                                STAFProcessHandle_t procHandle)
{
    STAFMutexSemLock handleLock(fHandleListSem);
    HandleList::iterator handleIter;

    // Basically, we loop through the list finding a registered handle with
    // the specified pid and procHandle.  We only have to worry about registered
    // handles here.  Pending handles are removed by the Process service, and we
    // never have static or in-process handles monitored by the ProcessManager.

    for(handleIter = fHandleList.begin();
        (handleIter != fHandleList.end()) &&
            ((handleIter->second.pid != pid) ||
             (handleIter->second.procHandle != procHandle) ||
             (handleIter->second.state != kRegistered));
        ++handleIter)
    { /* Do Nothing */ }

    if (handleIter != fHandleList.end())
        fHandleList.erase(handleIter);
}


STAFString STAFHandleManager::name(STAFHandle_t handle)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) != fHandleList.end())
        return fHandleList[handle].name;

    return STAFString();
}


STAFRC_t STAFHandleManager::variablePool(STAFHandle_t handle,
                                         STAFVariablePoolPtr &pool)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) != fHandleList.end())
        pool = fHandleList[handle].variablePoolPtr;
    else rc = kSTAFHandleDoesNotExist;

    return rc;
}


STAFRC_t STAFHandleManager::handleQueue(STAFHandle_t handle,
                                        STAFHandleQueuePtr &queue)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) != fHandleList.end())
        queue = fHandleList[handle].handleQueuePtr;
    else rc = kSTAFHandleDoesNotExist;

    return rc;
}


std::vector<STAFHandle_t> STAFHandleManager::handlesWithName(STAFString name)
{
    name.lowerCase();

    std::vector<STAFHandle_t> handles;
    STAFMutexSemLock handleLock(fHandleListSem);
    HandleList::iterator handleIter;

    // XXX: This should be able to be done with some form of copy_if

    for(handleIter = fHandleList.begin(); handleIter != fHandleList.end();
        ++handleIter)
    {
        if (name == handleIter->second.name.toLowerCase())
            handles.push_back(handleIter->second.handle);
    }

    return handles;
}

STAFRC_t STAFHandleManager::updateTimestamp(STAFHandle_t handle,
                                            STAFProcessID_t pid)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
        return kSTAFHandleDoesNotExist;

    HandleData &theHandle = fHandleList[handle];

    if ((theHandle.state != kStatic) && (theHandle.pid != pid))
        return kSTAFInvalidHandle;

    theHandle.lastUsedTimestamp = STAFTimestamp::now();

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addAndGetPendingHandle(STAFHandle_t &handle,
    STAFProcessID_t pid, STAFProcessHandle_t procHandle,
    STAFVariablePoolPtr pool)
{
    // Note: This method should only be used by STAFProcessService.  It will
    //       obtain a lock on our fHandleListSem before calling this method.
    //       Should anyone else need to call this method, they must also lock
    //       our fHandleListSem before calling this method.

    HandleData theHandle(fNextHandle++, pid, procHandle, kPending,
                         STAFTimestamp::now(), STAFString(), pool);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addAndGetStaticHandle(STAFHandle_t &handle,
                                                  const STAFString &name)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    HandleData theHandle(fNextHandle++, 0, 0, kStatic,
                         STAFTimestamp::now(), name);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::addAndGetStaticHandle(STAFHandle_t &handle,
                                                  const STAFString &name,
                                                  STAFVariablePoolPtr pool)
{
    STAFMutexSemLock handleLock(fHandleListSem);

    HandleData theHandle(fNextHandle++, 0, 0, kStatic,
                         STAFTimestamp::now(), name, pool);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    return kSTAFOk;
}


STAFRC_t STAFHandleManager::registerHandle(STAFHandle_t &handle,
                                           STAFProcessID_t pid,
                                           const STAFString &name)
{
    STAFString lowerName(name.toLowerCase());
    STAFMutexSemLock handleLock(fHandleListSem);
    HandleList::iterator handleIter;
    STAFProcessHandle_t procHandle = 0;

    for (handleIter = fHandleList.begin(); handleIter != fHandleList.end();
         ++handleIter)
    {

        if (handleIter->second.pid != pid) continue;

        if (pid == gSTAFProcPID)
        {
            if (lowerName == handleIter->second.lowerName)
            {
                // This is an existing kInProcess handle

                if (handleIter->second.state != kInProcess)
                {
                    STAFTrace::trace(kSTAFTraceError,
                                     "In STAFHandleManager::registerHandle(), "
                                     "found handle with STAFProc PID, but not "
                                     "state == kInProcess");
                }

                handle = handleIter->second.handle;

                return kSTAFOk;
            }
            else continue;
        }

        if (!STAFProcess::isRunning(handleIter->second.procHandle))
        {
            // This is a program that is no longer running, but which has not
            // been removed from the list yet.  Either our callback hasn't
            // been called for a regular kRegistered process, or the
            // ProcessManager has not told us to remove the kPending(Registered)
            // handle.  Either way, this handle is not a match.

            continue;
        }

        if (handleIter->second.state == kPending)
        {
            // This is an existing kPending handle, and the process is
            // just now registering.

            handle = handleIter->second.handle;
            handleIter->second.name = name;
            handleIter->second.state = kPendingRegistered;
            handleIter->second.lastUsedTimestamp = STAFTimestamp::now();

            return kSTAFOk;
        }

        procHandle = handleIter->second.procHandle;
    }

    // This is a valid new handle so register it

    if (procHandle == 0)
    {
        unsigned int osRC = 0;
        STAFRC_t rc = STAFProcess::getProcessHandleFromID(
                      pid, procHandle, osRC);
    }

    if (procHandle == 0)
    {
        STAFTrace::trace(kSTAFTraceError,
                         "In STAFHandleManager::registerHandle: received zero "
                         " process handle from getProcessHandleFromID()");
    }

    HandleData theHandle(fNextHandle++, pid, procHandle,
                         (pid == gSTAFProcPID) ? kInProcess : kRegistered,
                         STAFTimestamp::now(), name);

    fHandleList[theHandle.handle] = theHandle;
    handle = theHandle.handle;

    if (pid == gSTAFProcPID) return kSTAFOk;

    STAFProcessEndCallbackLevel1 callback = { processTerminated, 0 };

    return STAFProcess::registerForProcessTermination(pid,
           theHandle.procHandle, callback);
}


STAFRC_t STAFHandleManager::unRegister(STAFHandle_t handle,
                                       STAFProcessID_t pid)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        rc = kSTAFHandleDoesNotExist;
    }
    else
    {
        HandleData &theHandle = fHandleList[handle];

        if (theHandle.pid != pid)
            rc = kSTAFInvalidHandle;
        else if (theHandle.state == kPendingRegistered)
            theHandle.state = kPending;
        else if (theHandle.state == kRegistered)
            fHandleList.erase(handle);
    }

    return rc;
}


STAFRC_t STAFHandleManager::removePendingHandle(STAFHandle_t handle,
                                                STAFProcessID_t pid)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        rc = kSTAFHandleDoesNotExist;
    }
    else
    {
        const HandleData &theHandle = fHandleList[handle];

        if (theHandle.pid != pid)
        {
            rc = kSTAFInvalidHandle;
        }
        else if ((theHandle.state == kPending) ||
                 (theHandle.state == kPendingRegistered))
        {
            fHandleList.erase(handle);
        }
        else
        {
            rc = kSTAFInvalidHandle;
        }
    }

    return rc;
}


STAFRC_t STAFHandleManager::removeStaticHandle(STAFHandle_t handle)
{
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock handleLock(fHandleListSem);

    if (fHandleList.find(handle) == fHandleList.end())
    {
        rc = kSTAFHandleDoesNotExist;
    }
    else
    {
        const HandleData &theHandle = fHandleList[handle];

        if (theHandle.state == kStatic)
            fHandleList.erase(handle);
        else
            rc = kSTAFInvalidHandle;
    }

    return rc;
}


STAFHandleManager::HandleList STAFHandleManager::getHandleListCopy()
{
    STAFMutexSemLock handleLock(fHandleListSem);

    return fHandleList;
}


STAFMutexSem &STAFHandleManager::getHandleManagerSem()
{
    return fHandleListSem;
}
