/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#ifndef STAF_HandleManager
#define STAF_HandleManager

#include <map>
#include <vector>
#include "STAFString.h"
#include "STAFRefPtr.h"
#include "STAF.h"
#include "STAFProc.h"
#include "STAFTimestamp.h"
#include "STAFVariablePool.h"
#include "STAFMutexSem.h"
#include "STAFHandleQueue.h"


// STAFHandleManager - This class manages all the process handles used by
//                     STAFProc.  It handles the semantics of adding,
//                     registering, unregistering, and removing the various
//                     types of handles.  It also provides facilities to
//                     query information about a given handle, as well as to
//                     retrieve a set of handles.


class STAFHandleManager
{
public:

    // Public types

    // kInProcess         - Represents a handle for a service executing with
    //                      the STAFProc process
    // KPending           - Represents an unregistered handle for a process
    //                      that was started via STAF
    // kPendingRegistered - Represents a registered handle for a process that
    //                      was started via STAF
    // kRegistered        - Represents a registered handle for a process that
    //                      was not started via STAF
    // kStatic            - Represents a handle shared across multiple
    //                      native processes

    enum HandleState { kInProcess,  kPending, kPendingRegistered,
                       kRegistered, kStatic };

    struct HandleData
    {
        HandleData() : lastUsedTimestamp(STAFTimestamp::now())
        { /* Do Nothing */ }


        // Note: We need two constructors here, because VC++ V6 doesn't like the
        //       declaration of default values for variablePoolPtr and
        //       handleQueuePtr

        HandleData(STAFHandle_t aHandle, STAFProcessID_t aPID,
                   STAFProcessHandle_t aProcHandle, HandleState aState,
                   STAFTimestamp aTimestamp, const STAFString &aName)
            : handle(aHandle), pid(aPID), procHandle(aProcHandle),
              state(aState), lastUsedTimestamp(aTimestamp), name(aName),
              lowerName(aName.toLowerCase()),
              variablePoolPtr(new STAFVariablePool, STAFVariablePoolPtr::INIT),
              handleQueuePtr(new STAFHandleQueue(gMaxQueueSize),
                             STAFHandleQueuePtr::INIT)
        { /* Do Nothing */ }

        HandleData(STAFHandle_t aHandle, STAFProcessID_t aPID,
                   STAFProcessHandle_t aProcHandle, HandleState aState,
                   STAFTimestamp aTimestamp, const STAFString &aName,
                   STAFVariablePoolPtr varPool)
             : handle(aHandle), pid(aPID), procHandle(aProcHandle),
               state(aState), lastUsedTimestamp(aTimestamp), name(aName),
               lowerName(aName.toLowerCase()), variablePoolPtr(varPool),
               handleQueuePtr(new STAFHandleQueue(gMaxQueueSize),
                              STAFHandleQueuePtr::INIT)
        { /* Do Nothing */ }


        STAFHandle_t handle;
        STAFProcessID_t pid;
        STAFProcessHandle_t procHandle;
        HandleState state;
        STAFTimestamp lastUsedTimestamp;
        STAFString name;
        STAFString lowerName;
        STAFVariablePoolPtr variablePoolPtr;
        STAFHandleQueuePtr handleQueuePtr;
    };

    typedef std::map<STAFHandle_t, HandleData> HandleList;

    STAFHandleManager();

    STAFString name(STAFHandle_t handle);
    STAFRC_t variablePool(STAFHandle_t handle, STAFVariablePoolPtr &pool);
    STAFRC_t handleQueue(STAFHandle_t handle, STAFHandleQueuePtr &queue);

    std::vector<STAFHandle_t> handlesWithName(STAFString name);

    STAFRC_t updateTimestamp(STAFHandle_t handle, STAFProcessID_t pid);
    STAFRC_t addAndGetPendingHandle(STAFHandle_t &handle,
                                    STAFProcessID_t pid,
                                    STAFProcessHandle_t procHandle,
                                    STAFVariablePoolPtr pool);
    STAFRC_t addAndGetStaticHandle(STAFHandle_t &handle, const STAFString &name);
    STAFRC_t addAndGetStaticHandle(STAFHandle_t &handle, const STAFString &name,
                                   STAFVariablePoolPtr pool);
    STAFRC_t registerHandle(STAFHandle_t &handle, STAFProcessID_t pid,
                            const STAFString &name);
    STAFRC_t unRegister(STAFHandle_t handle, STAFProcessID_t pid);
    STAFRC_t removePendingHandle(STAFHandle_t handle, STAFProcessID_t pid);
    STAFRC_t removeStaticHandle(STAFHandle_t handle);

    HandleList getHandleListCopy();

    // Caution: The getHandleManagerSem() API is only intended for the use of
    //          the process service.  Do not use this API unless you REALLY
    //          know what you are doing.

    STAFMutexSem &getHandleManagerSem();

private:

    // Don't allow copy construction or assignment
    STAFHandleManager(const STAFHandleManager &);
    STAFHandleManager &operator=(const STAFHandleManager &);

    // This is the callback
    static void processTerminated(STAFProcessID_t pid,
                                  STAFProcessHandle_t procHandle,
                                  unsigned int rc,
                                  void *);

    // This does the actual work
    void handleProcessTerminated(STAFProcessID_t pid,
                                 STAFProcessHandle_t procHandle);

    STAFHandle_t fNextHandle;
    STAFMutexSem fHandleListSem;
    HandleList fHandleList;

};

#endif
