/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFHandleService.h"
#include "STAFHandleManager.h"
#include "STAFTrustManager.h"
#include "STAFProcUtil.h"
#include "STAFDiagManager.h"

// QUERY [ALL | NAME <Handle Name>] [PENDING] [REGISTERED] [INPROCESS] [STATIC]
// Default is REGISTERED, INPROCESS, and STATIC
//
// CREATE HANDLE NAME <Handle Name>
//
// DELETE HANDLE <Number>


STAFHandleService::STAFHandleService() : STAFService("HANDLE")
{
    // query options

    fQueryParser.addOption("QUERY",      1, 
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("ALL",        1, 
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("PENDING",    1, 
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("REGISTERED", 1, 
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("INPROCESS",  1, 
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("STATIC",     1, 
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("NAME",       1,
        STAFCommandParser::kValueRequired);

    // query groups

    fQueryParser.addOptionGroup("QUERY",    1, 1);
    fQueryParser.addOptionGroup("ALL NAME", 1, 1);

    // create options

    fCreateParser.addOption("CREATE",      1, 
        STAFCommandParser::kValueNotAllowed);
    fCreateParser.addOption("HANDLE",      1, 
        STAFCommandParser::kValueNotAllowed);
    fCreateParser.addOption("NAME",        1, 
        STAFCommandParser::kValueRequired);

    // create needs

    fCreateParser.addOptionNeed("CREATE", "HANDLE");
    fCreateParser.addOptionNeed("CREATE", "NAME");

    // delete options

    fDeleteParser.addOption("DELETE",      1, 
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("HANDLE",      1, 
        STAFCommandParser::kValueRequired);

    // delete needs

    fDeleteParser.addOptionNeed("DELETE", "HANDLE");
}


STAFHandleService::~STAFHandleService()
{
    /* Do Nothing */
}


STAFString STAFHandleService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFHandleService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();

    if      (action == "query")  return handleQuery(requestInfo);
    else if (action == "create") return handleCreate(requestInfo);
    else if (action == "delete") return handleDelete(requestInfo);
    else if (action == "help")   return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString, "");

}


STAFServiceResult STAFHandleService::handleCreate(
    const STAFServiceRequest &requestInfo)
{
    if (!isLocalMachine(requestInfo.fMachine))
        return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fCreateParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }
 
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString name;
    STAFString errorBuffer;
    STAFRC_t rc = resolveStringIfExists(parsedResult, "NAME", handlePool, name, 
                                        errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFHandle_t theHandle = 0;

    rc = gHandleManagerPtr->addAndGetStaticHandle(theHandle, name);

    if (rc) return rc;

    return STAFServiceResult(kSTAFOk, theHandle);
}


STAFServiceResult STAFHandleService::handleDelete(
    const STAFServiceRequest &requestInfo)
{
    if (!isLocalMachine(requestInfo.fMachine))
        return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fDeleteParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }
 
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFHandle_t theHandle = 0;
    STAFString errorBuffer;
    STAFRC_t rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool,
                                      theHandle, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = gHandleManagerPtr->removeStaticHandle(theHandle);

    if (rc) return STAFServiceResult(rc, theHandle);

    return kSTAFOk;
}


STAFServiceResult STAFHandleService::handleQuery(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fQueryParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }
 
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a QUERY request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/HANDLE QUERY",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    int getPending = 0;
    int getRegistered = 0;
    int getInProcess = 0;
    int getStatic = 0;

    if ((parsedResult->optionTimes("PENDING") == 0) &&
        (parsedResult->optionTimes("REGISTERED") == 0) &&
        (parsedResult->optionTimes("INPROCESS") == 0) &&
        (parsedResult->optionTimes("STATIC") == 0))
    {
        getPending = 0;
        getRegistered = 1;
        getInProcess = 1;
        getStatic = 1;
    }
    else
    {
        if (parsedResult->optionTimes("PENDING") != 0)
            getPending = 1;
        if (parsedResult->optionTimes("REGISTERED") != 0)
            getRegistered = 1;
        if (parsedResult->optionTimes("INPROCESS") != 0)
            getInProcess = 1;
        if (parsedResult->optionTimes("STATIC") != 0)
            getStatic = 1;
    }

    int getAll = parsedResult->optionTimes("ALL");

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString name;
    STAFString errorBuffer;
    STAFRC_t rc = resolveStringIfExists(parsedResult, "NAME", handlePool, name,
                                        errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFHandleManager::HandleList handleList =
                                  gHandleManagerPtr->getHandleListCopy();
    STAFHandleManager::HandleList::iterator iter;
    STAFString result;

    for(iter = handleList.begin(); iter != handleList.end(); iter++)
    {
        STAFHandleManager::HandleData handleData = iter->second;

        if (((getAll != 0) || (handleData.name == name)) &&
           (((getPending == 1) && (handleData.state ==
                                   STAFHandleManager::kPending)) ||
            ((getRegistered == 1) && ((handleData.state ==
                                   STAFHandleManager::kRegistered) ||
                                  (handleData.state ==
                                   STAFHandleManager::kPendingRegistered))) ||
            ((getInProcess == 1) && (handleData.state ==
                                   STAFHandleManager::kInProcess)) ||
            ((getStatic == 1) && (handleData.state ==
                                   STAFHandleManager::kStatic))))
        {
            result += STAFString(handleData.handle) + " ";

            if (handleData.state == STAFHandleManager::kPending)
            {
                if (handleData.name.length() == 0)
                    result += "<Unknown> Pending";
                else
                    result += handleData.name + " Pending";
            }
            else if ((handleData.state == STAFHandleManager::kRegistered) ||
                     (handleData.state == STAFHandleManager::kPendingRegistered))
            {
                result += handleData.name + " Registered";
            }
            else if (handleData.state == STAFHandleManager::kInProcess)
            {
                result += handleData.name + " InProcess";
            }
            else if (handleData.state == STAFHandleManager::kStatic)
            {
                result += handleData.name + " Static";
            }
            else
            {
                result += "<Unknown> <Unknown>";
            }

            result += " " + handleData.lastUsedTimestamp.asString();
            result += *gLineSeparatorPtr;
        }
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFHandleService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString result("HANDLE service help");

    result += *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "CREATE HANDLE NAME <Handle Name>" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "DELETE HANDLE <Number>" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "QUERY  [ALL | NAME <Handle Name>] [PENDING] [REGISTERED] ";
    result += "[INPROCESS] [STATIC]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
