/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFHelpService.h"
#include "STAFTrustManager.h"
#include "STAFUtil.h"
#include "STAFDiagManager.h"

// REGISTER SERVICE <Name> ERROR <#> INFO <String> DESCRIPTION <String>
//
// UNREGISTER SERVICE <Name> ERROR <#>
//
// [SERVICE <Name>] ERROR <#>
//
// LIST [SERVICE <Name>] [ERRORS]
//
// HELP

struct HelpMessageInfo
{
    STAFRC_t rc;
    const char *info;
    const char *description;
};

STAFHelpService::STAFHelpService() : STAFService("HELP")
{
    fRegParser.addOption("REGISTER",    1, STAFCommandParser::kValueNotAllowed);
    fRegParser.addOption("UNREGISTER",  1, STAFCommandParser::kValueNotAllowed);
    fRegParser.addOption("SERVICE",     1, STAFCommandParser::kValueRequired);
    fRegParser.addOption("ERROR",       1, STAFCommandParser::kValueRequired);
    fRegParser.addOption("INFO",        1, STAFCommandParser::kValueRequired);
    fRegParser.addOption("DESCRIPTION", 1, STAFCommandParser::kValueRequired);

    fRegParser.addOptionGroup("REGISTER UNREGISTER", 0, 1);
    fRegParser.addOptionGroup("UNREGISTER INFO", 0, 1);
    fRegParser.addOptionGroup("UNREGISTER DESCRIPTION", 0, 1);

    fRegParser.addOptionNeed("REGISTER", "INFO");
    fRegParser.addOptionNeed("REGISTER", "DESCRIPTION");

    fListParser.addOption("LIST",     1, STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SERVICE",  1, STAFCommandParser::kValueRequired);
    fListParser.addOption("ERRORS",   1, STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("SERVICES", 1, STAFCommandParser::kValueNotAllowed);

    fListParser.addOptionGroup("SERVICES ERRORS", 0, 1);
    fListParser.addOptionGroup("SERVICE SERVICES", 0, 1);

    fErrorParser.addOption("SERVICE", 1, STAFCommandParser::kValueRequired);
    fErrorParser.addOption("ERROR",   1, STAFCommandParser::kValueRequired);

    HelpMessageInfo helpMsgs[] =
    {
        { kSTAFOk, "No error",
          "No error" },
        { kSTAFInvalidAPI, "Invalid API",
          "This indicates that a process has tried to call an invalid "
          "internal STAF API.{STAF/Config/Sep/Line}If this error occurs, "
          "report it to the authors.{STAF/Config/Sep/Line}" },
        { kSTAFUnknownService, "Unknown service",
          "You have tried to submit a request to a service that is "
          "unknown to STAFProc.{STAF/Config/Sep/Line}Verify that you "
          "have correctly registered the service.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidHandle, "Invalid handle",
          "You are passing an invalid handle to a STAF API.  Ensure that "
          "you are using{STAF/Config/Sep/Line}the handle you received "
          "when you registered with STAF.{STAF/Config/Sep/Line}" },
        { kSTAFHandleAlreadyExists, "Handle already exists",
          "This indicates an internal STAF error. If this error occurs, "
          "report it to{STAF/Config/Sep/Line}the authors."
          "{STAF/Config/Sep/Line}" },
        { kSTAFHandleDoesNotExist, "Handle does not exist",
          "You are trying to perform an operation on a handle that does "
          "not exist.{STAF/Config/Sep/Line}For example, you may be "
          "trying to stop a process, but you are specifying the"
          "{STAF/Config/Sep/Line}wrong handle.{STAF/Config/Sep/Line}" },
        { kSTAFUnknownError, "Unknown error",
          "An unknown error has occurred.  This error is usually an "
          "indication of an{STAF/Config/Sep/Line}internal STAF error.  "
          "If this error occurs, report it the authors."
          "{STAF/Config/Sep/Line}" },
        { kSTAFInvalidRequestString, "Invalid request string",
          "You have submitted an improperly formatted request to a "
          "service.  See the{STAF/Config/Sep/Line}appropriate section "
          "in STAF User's Guide for the syntax of the service's"
          "{STAF/Config/Sep/Line}requests, or contact the provider of "
          "the service.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidServiceResult, "Invalid service result",
          "This indicates an internal error with the service to which a "
          "request was{STAF/Config/Sep/Line}submitted.  If this error "
          "occurs, report it to the authors and the service"
          "{STAF/Config/Sep/Line}provider.{STAF/Config/Sep/Line}" },
        { kSTAFREXXError, "REXX Error",
          "This indicates an internal error in an external Rexx "
          "service.  If this error{STAF/Config/Sep/Line}occurs, report "
          "it to the authors and the service provider."
          "{STAF/Config/Sep/Line}" },
        { kSTAFBaseOSError, "Base operating system error",
          "This indicates that a base operating system error was "
          "encountered.{STAF/Config/Sep/Line}The actual base operating "
          "system error code will be returned in the{STAF/Config/Sep/Line}"
          "result passed back from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFProcessAlreadyComplete, "Process already complete",
          "You are trying to stop a specific process that has either "
          "already been{STAF/Config/Sep/Line}stopped or has finished "
          "execution on its own.{STAF/Config/Sep/Line}" },
        { kSTAFProcessNotComplete, "Process not complete",
          "You are trying to free process information for a process that "
          "is still{STAF/Config/Sep/Line}executing.{STAF/Config/Sep/Line}" },
        { kSTAFVariableDoesNotExist, "Variable does not exist",
          "You are trying to resolve, get or delete a variable that does "
          "not exist.{STAF/Config/Sep/Line}" },
        { kSTAFUnResolvableString, "Unresolvable string",
          "You have requested to resolve a string that cannot be resolved.  "
          "This{STAF/Config/Sep/Line}indicates that you have "
          "exceeded the resolution depth of the VAR service."
          "{STAF/Config/Sep/Line}The most common cause of this is recursive "
          "variables definitions.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidResolveString, "Invalid resolve string",
          "The string you requested to be resolved has a non-matching left "
          "or right{STAF/Config/Sep/Line}curly brace.  Ensure that all "
          "variable references have both left and right"
          "{STAF/Config/Sep/Line}curly braces.{STAF/Config/Sep/Line}" },
        { kSTAFNoPathToMachine, "No path to machine",
          "This indicates that STAFProc was not able to submit the "
          "request to the{STAF/Config/Sep/Line}requested machine.  This "
          "error usually indicates that you have not configured"
          "{STAF/Config/Sep/Line}any network interfaces, or that STAF is "
          "not running on the target machine.{STAF/Config/Sep/Line}" },
        { kSTAFFileOpenError, "File open error",
          "This indicates that there was an error opening the requested "
          "file.  Some{STAF/Config/Sep/Line}possible explanations are that "
          "the file/path does not exist, contains{STAF/Config/Sep/Line}"
          "invalid characters, or is locked.{STAF/Config/Sep/Line}" },
        { kSTAFFileReadError, "File read error",
          "This indicates that there was an error while trying to read "
          "data from a file.{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}"
          "Note: Additional information regarding which file could not be"
          "read may be{STAF/Config/Sep/Line}provided in the result passed back "
          "from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFFileWriteError, "File write error",
          "This indicates that there was an error while trying to write "
          "data to a file.{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}Note: "
          "Additional information regarding which file could not be written to "
          "may{STAF/Config/Sep/Line}be provided in the result passed back from "
          "the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFFileDeleteError, "File delete error",
          "This indicates that there was an error while trying to delete "
          "a file or{STAF/Config/Sep/Line}directory.{STAF/Config/Sep/Line}"
          "{STAF/Config/Sep/Line}Note: Additional information regarding which "
          "file or directory could not be{STAF/Config/Sep/Line}deleted may be "
          "provided in the result passed back from the submit call."
          "{STAF/Config/Sep/Line}" },
        { kSTAFNotRunning, "STAF not running",
          "This indicates that STAF is not running on the local machine.  "
          "Verify that{STAF/Config/Sep/Line}STAFProc is running."
          "{STAF/Config/Sep/Line}" },
        { kSTAFCommunicationError, "Communication error",
          "This indicates an error transmitting data across the network, "
          "or to the local{STAF/Config/Sep/Line}STAF process.  For "
          "example, you would receive this error if STAFProc.exe was"
          "{STAF/Config/Sep/Line}terminated in the middle of a service "
          "request, or if a bridge went down in the{STAF/Config/Sep/Line}"
          "middle of a remote service request.{STAF/Config/Sep/Line}" },
        { kSTAFTrusteeDoesNotExist, "Trusteee does not exist",
          "You have requested to delete a trustee, and the trustee does "
          "not exist.{STAF/Config/Sep/Line}Verify that you have specified "
          "the correct trustee.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidTrustLevel, "Invalid trust level",
          "You have attempted to set a machine or default trust level to "
          "an invalid{STAF/Config/Sep/Line}level.  The valid trust levels "
          "are from zero to five.{STAF/Config/Sep/Line}" },
        { kSTAFAccessDenied, "Access denied",
          "You have submitted a request from a machine which is not "
          "authorized to{STAF/Config/Sep/Line}perform the request."
          "{STAF/Config/Sep/Line}" },
        { kSTAFRegistrationError, "Registration error",
          "This indicates an error with the configuration of an external "
          "service.  If this{STAF/Config/Sep/Line}error occurs, report it "
          "the service's authors.{STAF/Config/Sep/Line}" },
        { kSTAFServiceConfigurationError, "Service configuration error",
          "This indicates an error with the configuration of an external "
          "service.  This{STAF/Config/Sep/Line}usually stems from a "
          "problem with the PARMS specified when configuring the"
          "{STAF/Config/Sep/Line}service.  Consult the appropriate "
          "section in this document to verify the{STAF/Config/Sep/Line}"
          "service configuration parameters, or contact the service "
          "provider.{STAF/Config/Sep/Line}" },
        { kSTAFQueueFull, "Queue full",
          "This indicates that you are trying to queue a message to a "
          "handle's queue,{STAF/Config/Sep/Line}but the queue is full.  "
          "The maximum queue size can be increased by using the"
          "{STAF/Config/Sep/Line}MAXQUEUESIZE statement in the STAF "
          "Configuration  File.{STAF/Config/Sep/Line}" },
        { kSTAFNoQueueElement, "No queue element",
          "This indicates that you tried to GET or PEEK a particular "
          "element in a queue,{STAF/Config/Sep/Line}but no such element "
          "exists, or the queue is empty.{STAF/Config/Sep/Line}" },
        { kSTAFNotifieeDoesNotExist, "Notifiee does not exist",
          "This indicates that you are trying to remove a message "
          "notification for a{STAF/Config/Sep/Line}machine/process/priority "
          "combination which does not exist in the notification"
          "{STAF/Config/Sep/Line}list.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidAPILevel, "Invalid API level",
          "This indicates that a process has tried to call an invalid "
          "level of an{STAF/Config/Sep/Line}internal STAF API.  If this "
          "error occurs, report it to the authors.{STAF/Config/Sep/Line}" },
        { kSTAFServiceNotUnregisterable, "Service not unregisterable",
          "This indicates that you are trying to unregister a service "
          "that is not{STAF/Config/Sep/Line}unregisterable."
          "{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}Note: Internal "
          "services are not unregisterable.{STAF/Config/Sep/Line}" },
        { kSTAFServiceNotAvailable, "Service not available",
          "This indicates that the service you requested is not currently able "
          "to accept{STAF/Config/Sep/Line}requests.  The service may be in "
          "the process of initializing or terminating.{STAF/Config/Sep/Line}" },
        { kSTAFSemaphoreDoesNotExist, "Semaphore does not exist",
          "This indicates that you are trying to release, query, or "
          "delete a semaphore{STAF/Config/Sep/Line}that does not exist."
          "{STAF/Config/Sep/Line}" },
        { kSTAFNotSemaphoreOwner, "Not sempahore owner",
          "This indicates that you are trying to release a semaphore for "
          "which your{STAF/Config/Sep/Line}process is not the current "
          "owner.{STAF/Config/Sep/Line}" },
        { kSTAFSemaphoreHasPendingRequests, "Semaphore has pending requests",
          "This indicates that you are trying to delete either a mutex "
          "semaphore that{STAF/Config/Sep/Line}is currently owned or an "
          "event semaphore that has waiting processes."
          "{STAF/Config/Sep/Line}" },
        { kSTAFTimeout, "Timeout",
          "This indicates that you submitted a request with a timeout "
          "value and the{STAF/Config/Sep/Line}request did not complete "
          "within the requested time.{STAF/Config/Sep/Line}" },
        { kSTAFJavaError, "Java error",
          "This indicates an error performing a Java native method call.  "
          "A description{STAF/Config/Sep/Line}of the error will be "
          "returned in the result passed back from the submit call."
          "{STAF/Config/Sep/Line}" },
        { kSTAFConverterError, "Converter error",
          "This indicates an error performing a codepage conversion.  The "
          "most likely{STAF/Config/Sep/Line}cause of this error is that "
          "STAF was not properly installed.  However, it is"
          "{STAF/Config/Sep/Line}possible that you are currently using a "
          "codepage that was not present or{STAF/Config/Sep/Line}specified "
          "during STAF installation.{STAF/Config/Sep/Line}" },
        { kSTAFServiceAlreadyExists, "Service already exists (deprecated)",
          "This indicates that the specified Service already exists."
          "{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}Note: Additional "
          "information regarding the service which already exists may"
          "{STAF/Config/Sep/Line}be provided in the result passed back "
          "from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidObject, "Invalid object",
          "This indicates that an invalid object was specified to a STAF "
          "API.  If you{STAF/Config/Sep/Line}receive this return code via "
          "a standard STAFSubmit call, report it to the{STAF/Config/Sep/Line}"
          "authors and the service provider.{STAF/Config/Sep/Line}" },
        { kSTAFInvalidParm, "Invalid parm",
          "This indicates that an invalid parameter was specified to a "
          "STAF API.  If you{STAF/Config/Sep/Line}receive this return code "
          "via a standard STAFSubmit call, report it to the"
          "{STAF/Config/Sep/Line}authors and the service provider."
          "{STAF/Config/Sep/Line}" },
        { kSTAFRequestNumberNotFound, "Request number not found",
          "This indicates that the specified request number was not found.  "
          "The specified{STAF/Config/Sep/Line}request number may be "
          "invalid, or the request's information may no longer be"
          "{STAF/Config/Sep/Line}available from the Service Service (for "
          "example, if the SERVICE FREE command{STAF/Config/Sep/Line}had "
          "previously been issued for the request number)."
          "{STAF/Config/Sep/Line}" },
        { kSTAFInvalidAsynchOption, "Invalid asynchronous option",
          "This indicates that an invalid asynchronous submit option was "
          "specified.{STAF/Config/Sep/Line}" },
        { kSTAFRequestNotComplete, "Request not complete",
          "This indicates that the specified request is not complete.  "
          "This error code{STAF/Config/Sep/Line}would be returned, for "
          "example, if you requested the result of a request which"
          "{STAF/Config/Sep/Line}has not yet completed."
          "{STAF/Config/Sep/Line}" },
        { kSTAFProcessAuthenticationDenied, "Process authentication denied",
          "This indicates that the userid/password you specified could not be"
          "{STAF/Config/Sep/Line}authenticated.  The userid/password may not "
          "be valid or authentication{STAF/Config/Sep/Line}may be disabled."
          "{STAF/Config/Sep/Line}" },
        { kSTAFInvalidValue, "Invalid value",
          "This indicates that an invalid value was specified.  This is "
          "closely related{STAF/Config/Sep/Line}to the Invalid Request "
          "String return code, but indicates that a specific"
          "{STAF/Config/Sep/Line}value in the request is invalid.  For "
          "example, you may not have specified{STAF/Config/Sep/Line}a number "
          "where a number was expected.{STAF/Config/Sep/Line}"
          "{STAF/Config/Sep/Line}Note: Additional information regarding the "
          "invalid value may be provided in{STAF/Config/Sep/Line}the result "
          "passed back from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFDoesNotExist, "Does not exist",
          "This indicates that the item you specified does not exist."
          "{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}Note: Additional "
          "information regarding the item which could not be found "
          "may be{STAF/Config/Sep/Line}provided in the result passed back "
          "from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFAlreadyExists, "Already exists",
          "This indicates that the item you specified already exists."
          "{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}Note: Additional "
          "information regarding the item which already exists may "
          "be{STAF/Config/Sep/Line}provided in the result passed back "
          "from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFDirectoryNotEmpty, "Directory Not Empty",
          "This indicates that you have tried to delete a directory, but that "
          "directory{STAF/Config/Sep/Line}is not empty.{STAF/Config/Sep/Line}"
          "{STAF/Config/Sep/Line}Note: Additional information specifying the "
          "directory which could not be{STAF/Config/Sep/Line}deleted may be "
          "provided in the result passed back from the submit call."
          "{STAF/Config/Sep/Line}" },
        { kSTAFDirectoryCopyError, "Directory Copy Error",
          "This indicates that you have tried to copy a directory, but that "
          "errors{STAF/Config/Sep/Line}occurred during the copy."
          "{STAF/Config/Sep/Line}{STAF/Config/Sep/Line}Note: Additional "
          "information specifying the entries which could not be"
          "{STAF/Config/Sep/Line}copied may be provided in the result "
          "passed back from the submit call.{STAF/Config/Sep/Line}" },
        { kSTAFDiagnosticsNotEnabled, "Diagnostics Not Enabled",
          "This indicates that you tried to record diagnostics data, but "
          "diagnostics{STAF/Config/Sep/Line}have not been enabled.  You must "
          "enable diagnostics before you can record{STAF/Config/Sep/Line}"
          "diagnostics data.{STAF/Config/Sep/Line}" }
    };

    for (int i = 0; i < (sizeof(helpMsgs) / sizeof(HelpMessageInfo)); ++i)
    {
        ServiceHelpData helpData(helpMsgs[i].info, helpMsgs[i].description);
        fErrorMap[helpMsgs[i].rc]["INTERNAL"] = helpData;
        fServiceMap["INTERNAL"][helpMsgs[i].rc] = helpData;
    }
}


STAFString STAFHelpService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFHelpService::~STAFHelpService()
{
    /* Do Nothing */
}


STAFServiceResult STAFHelpService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if (action == "register")
        return handleRegister(true, requestInfo);
    else if (action == "unregister")
        return handleRegister(false, requestInfo);
    else if (action == "list")
        return handleList(requestInfo);
    else if ((action == "service") || (action == "error"))
        return handleError(requestInfo);
    else if (action == "help")
        return handleHelp(requestInfo);
    
    return kSTAFInvalidRequestString;
}



STAFServiceResult STAFHelpService::handleRegister(
    bool isRegister, const STAFServiceRequest &requestInfo)
{
    if (!isLocalMachine(requestInfo.fMachine))
        return kSTAFAccessDenied;
 
    STAFCommandParseResultPtr parsedResult =
                              fRegParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }
 
    STAFString errorBuffer;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString info = parsedResult->optionValue("INFO");
    STAFString description = parsedResult->optionValue("DESCRIPTION");
    STAFString service;
    unsigned int errorNumber;

    STAFRC_t rc = resolveUIntIfExists(parsedResult, "ERROR", handlePool,
                                      errorNumber, errorBuffer);

    if (!rc) rc = resolveStringIfExists(parsedResult, "SERVICE", handlePool, 
                                        service, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString upperService = service.toUpperCase();

    if (isRegister)
    {
        ServiceHelpData helpData = ServiceHelpData(info, description);
        STAFMutexSemLock lock(fMapSem);

        fErrorMap[errorNumber][upperService] = helpData;
        fServiceMap[upperService][errorNumber] = helpData;
    }
    else
    {
        STAFMutexSemLock lock(fMapSem);

        if (fErrorMap.find(errorNumber) == fErrorMap.end())
            return STAFServiceResult(kSTAFDoesNotExist, errorNumber);

        if (fErrorMap[errorNumber].find(upperService) ==
            fErrorMap[errorNumber].end())
        {
            return STAFServiceResult(kSTAFDoesNotExist, service);
        }

        fErrorMap[errorNumber].erase(upperService);
        fServiceMap[upperService].erase(errorNumber);
    }

    return kSTAFOk;
}


STAFServiceResult STAFHelpService::handleList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fListParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }
 
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a LIST request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/HELP LIST",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString service = "INTERNAL";
    STAFString errorBuffer;
    STAFString result;
    STAFRC_t rc = resolveStringIfExists(parsedResult, "SERVICE", handlePool, 
                                        service, errorBuffer);
    STAFString upperService = service.toUpperCase();
    STAFMutexSemLock lock(fMapSem);

    if (parsedResult->optionTimes("SERVICES") != 0)
    {
        HelpServiceMap::iterator iter;

        for (iter = fServiceMap.begin(); iter != fServiceMap.end(); ++iter)
        {
            if (iter->first != "INTERNAL")
                result += iter->first + *gLineSeparatorPtr;
        }
    }
    else
    {
        if (fServiceMap.find(upperService) == fServiceMap.end())
            return STAFServiceResult(kSTAFDoesNotExist, service);

        HelpErrorToInfoMap::iterator iter;

        STAFString spaces("     ");

        for (iter = fServiceMap[upperService].begin();
             iter != fServiceMap[upperService].end(); ++iter)
        {
            STAFString errorNumberString(iter->first);

            result += spaces.subString(0, 5 -
                      errorNumberString.length(STAFString::kChar)) +
                      errorNumberString + STAFString(kUTF8_SPACE) +
                      STAFString(kUTF8_SPACE) + iter->second.info +
                      *gLineSeparatorPtr;
        }

        if (upperService == "INTERNAL")
            result += "4000+  Service specific errors" + *gLineSeparatorPtr;

    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFHelpService::handleError(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fErrorParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for an ERROR request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/HELP ERROR",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString service;
    STAFString errorBuffer;
    STAFString result;
    unsigned int errorNumber;
    STAFRC_t rc = resolveStringIfExists(parsedResult, "SERVICE", handlePool, 
                                        service, errorBuffer);

    if (!rc) rc = resolveUIntIfExists(parsedResult, "ERROR", handlePool, 
                                      errorNumber, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString upperService = service.toUpperCase();
    STAFMutexSemLock lock(fMapSem);

    if (service.length() != 0)
    {
        if (fServiceMap.find(upperService) == fServiceMap.end())
            return STAFServiceResult(kSTAFDoesNotExist, service);

        if (fServiceMap[upperService].find(errorNumber) ==
            fServiceMap[upperService].end())
        {
            return STAFServiceResult(kSTAFDoesNotExist, errorNumber);
        }
    }

    if (fErrorMap.find(errorNumber) == fErrorMap.end())
        return STAFServiceResult(kSTAFDoesNotExist, errorNumber);

    HelpServiceToInfoMap::iterator iter;

    for(iter = fErrorMap[errorNumber].begin();
        iter != fErrorMap[errorNumber].end(); ++iter)
    {
        // If they specified a particular service, then don't list other
        // services with the same error code

        if ((upperService.length() != 0) && (upperService != iter->first))
            continue;

        STAFString info;
        STAFString detail;

        STAFRC_t rc = resolveString(iter->second.info, handlePool, info,
                                    errorBuffer);
        if (!rc)
        {
            rc = resolveString(iter->second.detail, handlePool, detail,
                               errorBuffer);
        }

        if (rc) return STAFServiceResult(rc, errorBuffer);

        // if this is not the first entry add some spacing

        if (result.length() != 0)
            result += *gLineSeparatorPtr + *gLineSeparatorPtr;

        // If this is an "internal" error code, then don't prepend the
        // service name

        if ((iter->first == "INTERNAL") || (upperService.length() != 0))
        {
            result += iter->second.info + *gLineSeparatorPtr +
                      *gLineSeparatorPtr;
        }
        else
        {
            result += iter->first + " service: " + info +
                      *gLineSeparatorPtr + *gLineSeparatorPtr;
        }

        result += detail + *gLineSeparatorPtr;
    }

    return STAFServiceResult(kSTAFOk, result);
}

STAFServiceResult STAFHelpService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

   STAFString result("HELP Service Help");

   result += *gLineSeparatorPtr;
   result += *gLineSeparatorPtr;
   result += "REGISTER   SERVICE <Name> ERROR <Number> INFO <String> "
             "DESCRIPTION <String>"; 
   result += *gLineSeparatorPtr;
   result += *gLineSeparatorPtr;
   result += "UNREGISTER SERVICE <Name> ERROR <Number>"; 
   result += *gLineSeparatorPtr;
   result += *gLineSeparatorPtr;
   result += "[SERVICE <Name>] ERROR <Number>";
   result += *gLineSeparatorPtr;
   result += *gLineSeparatorPtr;
   result += "LIST SERVICES | [SERVICE <Name>] ERRORS";
   result += *gLineSeparatorPtr;
   result += *gLineSeparatorPtr;
   result += "HELP";
   result += *gLineSeparatorPtr;

   return STAFServiceResult(kSTAFOk, result);
}
