/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFUtil.h"
#include "STAFTrustManager.h"
#include "STAFTrace.h"
#include "STAFServiceManager.h"
#include "STAFMiscService.h"
#include "STAFThreadManager.h"
#include "STAFDiagManager.h"

// MACHINE <Machine>
// VERSION
// TRACE <ON | OFF> [ServiceRequest] [ServiceResult] [ServiceError]
//                  [ServiceAccessDenied] [ServiceManagement]
//                  [Error] [Warning] [Info] [Deprecated] [All]
// TRACE TO <STDOUT | STDERR | FILE <Filename> >
// TRACE LIST
// THREAD INFO
// HELP


// When adding a new trace level you need to add it to the traceOptions,
// to the traceNeeds, the ON/OFF/LIST processing in handleTrace() and the
// output of handleHelp()

STAFMiscService::STAFMiscService() : STAFService("MISC")
{
    // machine options

    fMachineParser.addOption("MACHINE", 1, STAFCommandParser::kValueRequired);

    // machine groups
 
    fMachineParser.addOptionGroup("MACHINE", 1, 1);                 

    // trace options

    fTraceParser.addOption("TRACE",               1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ON",                  1,          
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("OFF",                 1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICES",            1, 
        STAFCommandParser::kValueAllowed);
    fTraceParser.addOption("LIST",                1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("TO",                  1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("STDOUT",              1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("STDERR",              1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("FILE",                1, 
        STAFCommandParser::kValueRequired);
    fTraceParser.addOption("SERVICEREQUEST",      1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICEACCESSDENIED", 1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICEERROR",        1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("SERVICERESULT",       1, 
        STAFCommandParser::kValueNotAllowed);     
    fTraceParser.addOption("SERVICEMANAGEMENT",   1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("REGISTRATION",        1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("REMOTEREQUESTS",      1,
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("ERROR",               1, 
        STAFCommandParser::kValueNotAllowed);     
    fTraceParser.addOption("WARNING",             1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("INFO",                1, 
        STAFCommandParser::kValueNotAllowed);
    fTraceParser.addOption("DEPRECATED",          1, 
        STAFCommandParser::kValueNotAllowed);     
    fTraceParser.addOption("ALL",                 1, 
        STAFCommandParser::kValueNotAllowed);
 
    // trace groups
 
    fTraceParser.addOptionGroup("TRACE", 1, 1);     
    fTraceParser.addOptionGroup("TO ON OFF LIST SERVICES", 1, 1);
    fTraceParser.addOptionGroup("STDOUT STDERR FILE", 0, 1);
 
    // trace needs
 
    fTraceParser.addOptionNeed("STDOUT STDERR FILE", "TO");
    fTraceParser.addOptionNeed("SERVICEREQUEST SERVICEACCESSDENIED "
        "SERVICEERROR SERVICERESULT SERVICEMANAGEMENT REGISTRATION "
        "REMOTEREQUESTS ERROR WARNING INFO DEPRECATED", "ON OFF");
    fTraceParser.addOptionNeed("ALL", "ON OFF SERVICES");
    fTraceParser.addOptionNeed("ON OFF", "SERVICEREQUEST "
        "SERVICEACCESSDENIED SERVICEERROR SERVICERESULT "
        "SERVICEMANAGEMENT REGISTRATION REMOTEREQUESTS ERROR WARNING "
        "INFO DEPRECATED ALL");

    // thread options

    fThreadParser.addOption("THREAD", 1,
        STAFCommandParser::kValueNotAllowed);
    fThreadParser.addOption("INFO", 1,
        STAFCommandParser::kValueNotAllowed);

    // thread groups

    fThreadParser.addOptionGroup("THREAD", 1, 1);     
    fThreadParser.addOptionGroup("INFO", 1, 1);     
}


STAFServiceResult STAFMiscService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if      (action == "machine") return handleMachine(requestInfo);
    else if (action == "version") return handleVersion(requestInfo);
    else if (action == "trace")   return handleTrace(requestInfo);
    else if (action == "thread")  return handleThread(requestInfo);
    else if (action == "help")    return handleHelp(requestInfo);
    else return kSTAFInvalidRequestString;
}


STAFServiceResult STAFMiscService::handleMachine(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fMachineParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because the MACHINE request will no longer
        // exist in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/MISC MACHINE",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }
    
    STAFString errorBuffer;                  
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString machine;
    STAFRC_t rc = resolveString(parsedResult->optionValue("MACHINE"), 
                                handlePool, machine, errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    machine = effectiveMachineName(machine);

    return STAFServiceResult(kSTAFOk, machine);
}


STAFServiceResult STAFMiscService::handleVersion(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    return STAFServiceResult(kSTAFOk, gVersion);
}


STAFServiceResult STAFMiscService::handleThread(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

    unsigned int numThreads = gThreadManagerPtr->getThreadPoolSize();
    unsigned int numReady = gThreadManagerPtr->getNumReadyThreads();
    unsigned int numWork = gThreadManagerPtr->getNumWorkingThreads();

    STAFString data("Total threads  : ");

    data += numThreads + *gLineSeparatorPtr + "Working threads: ";
    data += numWork + *gLineSeparatorPtr + "Idle threads   : ";
    data += numReady + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, data);
}


STAFServiceResult STAFMiscService::handleTrace(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult =
                              fTraceParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFString errorBuffer;

    if (((parsedResult->optionTimes("LIST") != 0) &&
         (requestInfo.fTrustLevel < 2)) ||
        (requestInfo.fTrustLevel < 5))
    {
        return kSTAFAccessDenied;
    }

    if (requestInfo.fDiagEnabled)
    {
        if (parsedResult->optionTimes("LIST") != 0)
        {
            // Record diagnostics because output for a TRACE LIST request will
            // change in STAF V3.0 and the syntax for all TRACE requests will
            // change in STAF V3.0 with the addition of a new TRACE service.

            gDiagManagerPtr->record("STAF/V3.0-Mig/MISC TRACE LIST",
                                    requestInfo.fHandleName,
                                    requestInfo.fMachine,
                                    requestInfo.fHandle);
        }
        else
        {
            // Record diagnostics because the syntax for all TRACE requests
            // will change in STAF V3.0 with the addition of a new TRACE
            // service (instead of TRACE being part of the MISC service).

            gDiagManagerPtr->record("STAF/V3.0-Mig/MISC TRACE",
                                    requestInfo.fHandleName,
                                    requestInfo.fMachine,
                                    requestInfo.fHandle);
        }
    }

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString result;
    STAFString filename;
    STAFRC_t rc = resolveStringIfExists(parsedResult, "FILE", handlePool,
                                        filename, errorBuffer);

    if (rc) return STAFServiceResult(kSTAFInvalidRequestString, errorBuffer);

    if (parsedResult->optionTimes("TO") != 0)
    {
        if (parsedResult->optionTimes("STDOUT")) 
            STAFTrace::setTraceDestination(kSTAFTraceToStdout);
        else if (parsedResult->optionTimes("STDERR")) 
            STAFTrace::setTraceDestination(kSTAFTraceToStderr);
        else
            STAFTrace::setTraceDestination(kSTAFTraceToFile, filename);
    }
    else if (parsedResult->optionTimes("ON") != 0)
    {
        if (parsedResult->optionTimes("SERVICEREQUEST") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceRequest);
        if (parsedResult->optionTimes("SERVICEACCESSDENIED") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceAccessDenied);
        if (parsedResult->optionTimes("SERVICEERROR") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceError);
        if (parsedResult->optionTimes("SERVICERESULT") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceResult);
        if (parsedResult->optionTimes("SERVICEMANAGEMENT") != 0)
            STAFTrace::traceOn(kSTAFTraceServiceManagement);
        if (parsedResult->optionTimes("REGISTRATION") != 0)
            STAFTrace::traceOn(kSTAFTraceRegistration);
        if (parsedResult->optionTimes("REMOTEREQUESTS") != 0)
            STAFTrace::traceOn(kSTAFTraceRemoteRequests);
        if (parsedResult->optionTimes("ERROR") != 0)
            STAFTrace::traceOn(kSTAFTraceError);
        if (parsedResult->optionTimes("WARNING") != 0)
            STAFTrace::traceOn(kSTAFTraceWarning);
        if (parsedResult->optionTimes("INFO") != 0)
            STAFTrace::traceOn(kSTAFTraceInfo);
        if (parsedResult->optionTimes("DEPRECATED") != 0)
            STAFTrace::traceOn(kSTAFTraceDeprecated);
        if (parsedResult->optionTimes("ALL") != 0)
            STAFTrace::traceOn(kSTAFTraceAll);
    }
    else if (parsedResult->optionTimes("OFF") != 0)
    {
        if (parsedResult->optionTimes("SERVICEREQUEST") != 0)
            STAFTrace::traceOff(kSTAFTraceServiceRequest);
        if (parsedResult->optionTimes("SERVICEACCESSDENIED") != 0)
            STAFTrace::traceOff(kSTAFTraceServiceAccessDenied);
        if (parsedResult->optionTimes("SERVICEERROR") != 0)
            STAFTrace::traceOff(kSTAFTraceServiceError);
        if (parsedResult->optionTimes("SERVICERESULT") != 0)
            STAFTrace::traceOff(kSTAFTraceServiceResult);
        if (parsedResult->optionTimes("SERVICEMANAGEMENT") != 0)
            STAFTrace::traceOff(kSTAFTraceServiceManagement);
        if (parsedResult->optionTimes("REGISTRATION") != 0)
            STAFTrace::traceOff(kSTAFTraceRegistration);
        if (parsedResult->optionTimes("REMOTEREQUESTS") != 0)
            STAFTrace::traceOff(kSTAFTraceRemoteRequests);
        if (parsedResult->optionTimes("ERROR") != 0)
            STAFTrace::traceOff(kSTAFTraceError);
        if (parsedResult->optionTimes("WARNING") != 0)
            STAFTrace::traceOff(kSTAFTraceWarning);
        if (parsedResult->optionTimes("INFO") != 0)
            STAFTrace::traceOff(kSTAFTraceInfo);
        if (parsedResult->optionTimes("DEPRECATED") != 0)
            STAFTrace::traceOff(kSTAFTraceDeprecated);
        if (parsedResult->optionTimes("ALL") != 0)
            STAFTrace::traceOff(kSTAFTraceAll);
    }
    else if (parsedResult->optionTimes("SERVICES") != 0)
    {
        if (parsedResult->optionTimes("ALL") != 0)
        {
            STAFServiceManager::traceAllServices();
        }
        else
        {
            STAFServiceManager::traceServices(
                parsedResult->optionValue("SERVICES"));
        }
    }
    else
    {
        STAFString traceFilename;
        STAFTraceDestination_t traceDestination =
                               STAFTrace::getTraceDestination(traceFilename);

        result = "Tracing to: ";

        if (traceDestination == kSTAFTraceToStdout)
            result += "STDOUT";
        else if (traceDestination == kSTAFTraceToStderr)
            result += "STDERR";
        else
            result += "File, " + traceFilename;

        result += *gLineSeparatorPtr + *gLineSeparatorPtr;
        result += "Trace points enabled:" + *gLineSeparatorPtr +
                  *gLineSeparatorPtr;

        if (STAFTrace::getTraceMask() == 0)
        {
            result += "None" + *gLineSeparatorPtr;
        }
        else
        {
            if (STAFTrace::doTrace(kSTAFTraceServiceRequest))
                result += "ServiceRequest" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceServiceAccessDenied))
                result += "ServiceAccessDenied" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceServiceError))
                result += "ServiceError" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceServiceResult))
                result += "ServiceResult" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceServiceManagement))
                result += "ServiceManagement" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceRegistration))
                result += "Registration" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceRemoteRequests))
                result += "RemoteRequests" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceError))
                result += "Error" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceWarning))
                result += "Warning" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceInfo))
                result += "Info" + *gLineSeparatorPtr;
            if (STAFTrace::doTrace(kSTAFTraceDeprecated))
                result += "Deprecated" + *gLineSeparatorPtr;
        }

        result += *gLineSeparatorPtr + "Services to trace:" +
                  *gLineSeparatorPtr + *gLineSeparatorPtr;

        STAFServiceManager::ServiceStringList serviceList =
            STAFServiceManager::getTraceServices();

        if (serviceList.size() == 0)
        {
            result += "All" + *gLineSeparatorPtr;
        }
        else
        {
            for (STAFServiceManager::ServiceStringList::iterator iter =
                     serviceList.begin();
                 iter != serviceList.end(); ++iter)
            {
                result += *iter + *gLineSeparatorPtr;
            }
        }
    }

    return STAFServiceResult(rc, result);
}


STAFServiceResult STAFMiscService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString result("MISC service help" + *gLineSeparatorPtr +
                      *gLineSeparatorPtr);

    result += "VERSION" + *gLineSeparatorPtr;
    result += "MACHINE <Machine>" + *gLineSeparatorPtr;
    result += "TRACE <ON | OFF> [ServiceRequest] [ServiceResult] "
              "[ServiceError]" + *gLineSeparatorPtr;
    result += "                 [ServiceAccessDenied] [ServiceManagement] " +
              *gLineSeparatorPtr;
    result += "                 [Registration] [RemoteRequests] " +
              *gLineSeparatorPtr;
    result += "                 [Error] [Warning] [Info] [Deprecated] [All]" +
              *gLineSeparatorPtr;
    result += "TRACE [ALL] SERVICES [Service list]" + *gLineSeparatorPtr;
    result += "TRACE TO <STDOUT | STDERR | FILE <Filename> >" +
              *gLineSeparatorPtr;
    result += "TRACE LIST" + *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}


STAFString STAFMiscService::info(unsigned int) const
{
    return name() + ": Internal";
}

STAFMiscService::~STAFMiscService()
{
    /* Do Nothing */
}
