/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFNotificationList.h"
#include "STAFProc.h"

STAFNotificationList::STAFNotificationList()
{
    /* Do Nothing */
}


STAFNotificationList::~STAFNotificationList()
{
    /* Do Nothing */
}


STAFRC_t STAFNotificationList::reg(const STAFString &machine,
                                   const STAFHandle_t handle,
                                   unsigned int priority)
{
    unsigned int foundIt = 0;
    STAFMutexSemLock semLock(fNotifieeListSem);
    NotifieeList::iterator listIter;

    for(listIter = fNotifieeList.begin();
        !foundIt && (listIter != fNotifieeList.end());)
    {
        Notifiee notifiee(*listIter);

        if ((notifiee.nameOrHandle == Notifiee::kHandle) &&
            (notifiee.machine == machine) && (notifiee.handle == handle) &&
            (notifiee.priority == priority))
        {
            foundIt = 1;
        }
        else
        {
            ++listIter;
        }
    }

    if (!foundIt) fNotifieeList.push_back(Notifiee(machine, handle, priority));

    return kSTAFOk;
}


STAFRC_t STAFNotificationList::reg(const STAFString &machine,
                                   const STAFString &process,
                                   unsigned int priority)
{
    unsigned int foundIt = 0;
    STAFMutexSemLock semLock(fNotifieeListSem);
    NotifieeList::iterator listIter;

    for(listIter = fNotifieeList.begin();
        !foundIt && (listIter != fNotifieeList.end());)
    {
        Notifiee notifiee(*listIter);

        if ((notifiee.nameOrHandle == Notifiee::kName) &&
            (notifiee.machine == machine) && (notifiee.process == process) &&
            (notifiee.priority == priority))
        {
            foundIt = 1;
        }
        else
        {
            ++listIter;
        }
    }

    if (!foundIt) fNotifieeList.push_back(Notifiee(machine, process, priority));

    return kSTAFOk;
}


STAFRC_t STAFNotificationList::unregister(const STAFString &machine,
                                          const STAFHandle_t handle,
                                          unsigned int priority)
{
    unsigned int foundIt = 0;
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock semLock(fNotifieeListSem);
    NotifieeList::iterator listIter;

    for(listIter = fNotifieeList.begin();
        !foundIt && (listIter != fNotifieeList.end());)
    {
        Notifiee notifiee(*listIter);

        if ((notifiee.nameOrHandle == Notifiee::kHandle) &&
            (notifiee.machine == machine) && (notifiee.handle == handle) &&
            (notifiee.priority == priority))
        {
            foundIt = 1;
        }
        else
        {
            ++listIter;
        }
    }

    if (foundIt) fNotifieeList.erase(listIter);
    else rc = kSTAFNotifieeDoesNotExist;

    return rc;
}


STAFRC_t STAFNotificationList::unregister(const STAFString &machine,
                                          const STAFString &process,
                                          unsigned int priority)
{
    unsigned int foundIt = 0;
    STAFRC_t rc = kSTAFOk;
    STAFMutexSemLock semLock(fNotifieeListSem);
    NotifieeList::iterator listIter;

    for(listIter = fNotifieeList.begin();
        !foundIt && (listIter != fNotifieeList.end());)
    {
        Notifiee notifiee(*listIter);

        if ((notifiee.nameOrHandle == Notifiee::kName) &&
            (notifiee.machine == machine) && (notifiee.process == process) &&
            (notifiee.priority == priority))
        {
            foundIt = 1;
        }
        else
        {
            ++listIter;
        }
    }

    if (foundIt) fNotifieeList.erase(listIter);
    else rc = kSTAFNotifieeDoesNotExist;

    return rc;
}


STAFRC_t STAFNotificationList::sendNotificationToHandle(
    const STAFString &where, const STAFHandle_t handle, unsigned int priority,
    const STAFString &message)
{
    STAFString queueMessage("QUEUE HANDLE " + STAFString(handle) +
                            " PRIORITY " + STAFString(priority) + " MESSAGE :" +
                            STAFString(message.length(STAFString::kChar))
                            + ":" + message);
    STAFString result;

    return gSTAFProcHandlePtr->submit(where, "QUEUE", queueMessage)->rc;
}


STAFRC_t STAFNotificationList::sendNotificationToName(
    const STAFString &where, const STAFString &name, unsigned int priority,
    const STAFString &message)
{
    STAFString queueMessage("QUEUE NAME :" +
                            STAFString(name.length(STAFString::kChar)) +
                            ":" + name + " PRIORITY " +
                            STAFString(priority) + " MESSAGE :" +
                            STAFString(message.length(STAFString::kChar)) + ":" +
                            message);
    STAFString result;

    return gSTAFProcHandlePtr->submit(where, "QUEUE", queueMessage)->rc;
}


unsigned int STAFNotificationList::sendNotification(const STAFString &message)
{
    unsigned int numSent = 0;
    NotifieeList notifieeList;

    // Lock the list temporarily to get a copy
    {
        STAFMutexSemLock semLock(fNotifieeListSem);
        notifieeList = fNotifieeList;
    }

    NotifieeList::iterator listIter;

    for(listIter = notifieeList.begin(); listIter != notifieeList.end();
        ++listIter)
    {
        Notifiee notifiee = *listIter;

        if (notifiee.nameOrHandle == Notifiee::kHandle)
        {
            if (sendNotificationToHandle(notifiee.machine, notifiee.handle,
                notifiee.priority, message) == kSTAFOk)
            {
                ++numSent;
            }
        }
        else
        {
            if (sendNotificationToName(notifiee.machine, notifiee.process,
                notifiee.priority, message) == kSTAFOk)
            {
                ++numSent;
            }
        }
    }

    return numSent;
}


STAFNotificationList::NotifieeList STAFNotificationList::getNotifieeListCopy()
{
    STAFMutexSemLock semLock(fNotifieeListSem);

    return fNotifieeList;
}
