/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAF_iostream.h"
#include "STAF_fstream.h"
#include "STAF_exception.h"
#include "STAFString.h"
#include "STAFProc.h"
#include "STAFConfig.h"
#include "STAFRefPtr.h"
#include "STAFLocalConnectionFactory.h"
#include "STAFSocketFuncs.h"
#include "STAFSocketConnectionFactory.h"
#include "STAFSimpleServices.h"
#include "STAFProcessService.h"
#include "STAFHandleService.h"
#include "STAFVariableService.h"
#include "STAFFSService.h"
#include "STAFServiceService.h"
#include "STAFTrustService.h"
#include "STAFDiagService.h"
#include "STAFQueueService.h"
#include "STAFShutdownService.h"
#include "STAFMiscService.h"
#include "STAFSemService.h"
#include "STAFHelpService.h"
#include "STAFEventSem.h"
#include "STAFHandleManager.h"
#include "STAFRequestManager.h"
#include "STAFTrustManager.h"
#include "STAFDiagManager.h"
#include "STAFConnectionManager.h"
#include "STAFServiceManager.h"
#include "STAFNotificationList.h"
#include "STAFException.h"
#include "STAFUtil.h"
#include "STAFProcUtil.h"
#include "STAFProcOSUtil.h"
#include "STAFTrace.h"
#include "STAFThreadManager.h"
#include "STAFFileSystem.h"
#include "STAFConverter.h"

//
// Prototypes
//

unsigned int Receive(void *factoryPtr);
unsigned int HandleRequest(void *connectionPtr);
STAFServiceResult submitRemoteRequest(STAFString &where, STAFHandle_t &handle,
                              STAFConnPtr connection, STAFString &service, 
                              STAFSyncOption_t syncOption,
                              STAFRequestManager::STAFReqNum_t requestNumber,
                              STAFString &request);
void handleLocalServiceRequestAPI(unsigned int level, STAFConnPtr connection,
                                  unsigned int &doShutdown);
void handleRemoteServiceRequestAPI(unsigned int level, STAFConnPtr connection,
                                   unsigned int &doShutdown);
STAFServiceResult submitLocalRequest(STAFConnPtr connection, 
                               STAFString serviceName, 
                               STAFString machineName,
                               STAFString processName,
                               STAFSyncOption_t syncOption,
                               STAFRequestManager::STAFReqNum_t requestNumber,
                               STAFHandle_t handle, STAFString request);
void handleProcessRegistrationAPI(unsigned int level, STAFConnPtr connection,
                                  unsigned int &);
void handleProcessUnRegistrationAPI(unsigned int level, STAFConnPtr connection,
                                    unsigned int &);
void handleFileTransferAPI(unsigned int level, STAFConnPtr connection,
                           unsigned int &);
void handleFileTransferAPI2(unsigned int level, STAFConnPtr connection,
                           unsigned int &);
void handleDirectoryCopyAPI(unsigned int level, STAFConnPtr connection,
                           unsigned int &);
void handleFileTransfer(unsigned int apiType, unsigned int level,
                        STAFConnPtr connection);
STAFRC_t writeFile(fstream &outFile, const char *fileBuffer,
                   unsigned int writeLength, 
                   const STAFString toFile, const STAFString fromMachine);
void stafProcTerminate();
void replaceStringInBuffer (char* current, const char* searchStr, const char* replaceStr,
              int currentLen, int searchStrLen, int replaceStrLen, int* newLen, 
              int* lastReplace, char* newBuffer);
void rollBuffer(char * buffer, int bufferLen, int rollLen);


// type definitions

typedef void (*STAFAPIHandler)(unsigned int level, STAFConnPtr connection,
                               unsigned int &doShutdown);

const unsigned int kOldAPI = 0;
const unsigned int kNewAPI = 1;

struct STAFAPIDescriptor
{
    unsigned int apiType;       // 0 = old API type, 1 = new API type
    unsigned int minLevel;
    unsigned int maxLevel;
    unsigned int connMask;      // kLocal and/or kNetwork
    STAFAPIHandler handler;
};


//
// Global variables
//

// Note: New APIs must have a minimum/maximum level >= 1

STAFAPIDescriptor gAPITable[] =
{
    { kOldAPI, 2, 2, STAFConnection::kLocal, handleLocalServiceRequestAPI },
    { kOldAPI, 0, 0, STAFConnection::kNetwork, handleRemoteServiceRequestAPI },
    { kOldAPI, 0, 0, STAFConnection::kLocal, handleProcessRegistrationAPI },
    { kOldAPI, 0, 0, STAFConnection::kLocal, handleProcessUnRegistrationAPI },
    { kOldAPI, 0, 0, STAFConnection::kLocal | STAFConnection::kNetwork,
      handleFileTransferAPI },
    { kNewAPI, 1, 2, STAFConnection::kLocal | STAFConnection::kNetwork,
      handleFileTransferAPI2 },
    { kNewAPI, 1, 2, STAFConnection::kLocal | STAFConnection::kNetwork,
      handleDirectoryCopyAPI }
};

unsigned int gMaxAPINumber = (sizeof(gAPITable) / sizeof(STAFAPIDescriptor)) - 1;

STAFString gVersion("2.6.10");
STAFThreadManager gThreadManager(0, 1);
STAFThreadManager *gThreadManagerPtr = &gThreadManager;
STAFHandleManager gHandleManager;
STAFHandleManager *gHandleManagerPtr = &gHandleManager;
STAFRequestManager gRequestManager;
STAFRequestManager *gRequestManagerPtr = &gRequestManager;
STAFTrustManager gTrustManager(5);
STAFTrustManager *gTrustManagerPtr = &gTrustManager;
STAFConnectionManager gConnectionManager;
STAFConnectionManager *gConnectionManagerPtr = &gConnectionManager;
STAFServiceManager gServiceManager;
STAFServiceManager *gServiceManagerPtr = &gServiceManager;
STAFDiagManager gDiagManager(0);
STAFDiagManager *gDiagManagerPtr = &gDiagManager;
STAFEventSemPtr gShutdownSemaphore;
STAFEventSemPtr *gShutdownSemaphorePtr = &gShutdownSemaphore;
STAFProcessID_t gSTAFProcPID = 0;
STAFHandle_t gSTAFProcHandle = 0;
STAFHandlePtr gSTAFProcHandlePtr;
STAFEnvMap gEnvMap;
STAFEnvMap *gEnvMapPtr = &gEnvMap;
unsigned char *gEnvBuffer = 0;
int gEnvSize = 0;
STAFMutexSem gDirectorySem;
STAFMutexSem *gDirectorySemPtr = &gDirectorySem;
STAFString gMachineName;
STAFString *gMachineNamePtr = &gMachineName;
STAFString gMachineNetName;
STAFString *gMachineNetNamePtr = &gMachineNetName;
STAFString gSTAFWriteLocation;
STAFString *gSTAFWriteLocationPtr = &gSTAFWriteLocation;
STAFVariablePoolPtr gGlobalVariablePool(new STAFVariablePool,
                                        STAFVariablePoolPtr::INIT);
STAFVariablePoolPtr *gGlobalVariablePoolPtr = &gGlobalVariablePool;
unsigned short gTCPIPPort = 6500;
unsigned int gMaxQueueSize = 100;
unsigned int gNumInitialThreads = 10;
STAFProcessConsoleMode_t gDefaultConsoleMode = kSTAFProcessNewConsole;
long gMaxFiles = 500;
int gSpecifiedMachineConfig = 0;
int gUseTCPIPInterface = 0;
unsigned int gTCPIPConnectionTimeout = 5000;
unsigned int gTCPIPConnectionAttempts = 2;
unsigned int gConnectionRetryDelay = 1000;
int gUseLongNames = 0;
STAFNotificationList gNotifyOnStart;
STAFNotificationList *gNotifyOnStartPtr = &gNotifyOnStart;
STAFNotificationList gNotifyOnShutdown;
STAFNotificationList *gNotifyOnShutdownPtr = &gNotifyOnShutdown;
STAFEventSem gInterfaceSem;
STAFString gLineSeparator;
STAFString *gLineSeparatorPtr = &gLineSeparator;
STAFString gFileSeparator;
STAFString *gFileSeparatorPtr = &gFileSeparator;
STAFString gPathSeparator;
STAFString *gPathSeparatorPtr = &gPathSeparator;
STAFString gCommandSeparator;
STAFString *gCommandSeparatorPtr = &gCommandSeparator;

const STAFString sQueueHandle("QUEUE HANDLE ");
const STAFString sMessage(" MESSAGE ");
const STAFString sRequestComplete("STAF/RequestComplete ");
const STAFString sSemiColon(";");
const STAFString sColon(":");
const STAFString sLocal("LOCAL");
const STAFString sQueue("QUEUE");

int main(int argc, char **argv, char **envp)
{
#ifdef STAF_OS_NAME_ZOS
    extern char **environ;
    envp = environ;
#endif

    // Register our termination handler.  It will simply tell us if we have
    // ended abruptly due to an uncaught or unexpected exception.

    STAF_set_terminate(stafProcTerminate);

    gSTAFProcPID = STAFUtilGetPID();

    // Initialize Platform specific code 

    STAFString errorBuffer = "";
    unsigned int osRC = 0;
    int stafOSInitRC = STAFProcOSInit(errorBuffer, osRC);

    if (stafOSInitRC != 0)
    {
        cout << "Error during platform-specific initialization"  << endl;
        cout << errorBuffer << ", RC: " << osRC << endl;
        return 1;
    }

    // Default tracing to standard output

    STAFTrace::setTraceDestination(kSTAFTraceToStdout);

    // Initialize the global variable pool
    //
    // Note: The variable "Machine" is set after the config is read

    STAFString configHead("STAF/Config/");
    STAFConfigInfo configInfo;
    STAFString_t errorBufferT;

    STAFRC_t rc = STAFUtilGetConfigInfo(&configInfo, &errorBufferT, &osRC);
    
    if (rc != kSTAFOk)
    {
        cout << STAFString(errorBufferT);
        cout << ", STAF rc: " << rc << ", OS rc: " << osRC << endl;
        return 1;
    }

    gGlobalVariablePool->set("STAF/Version", gVersion);
    gGlobalVariablePool->set(configHead + "BootDrive", configInfo.bootDrive);
    gGlobalVariablePool->set(configHead + "OS/Name", configInfo.osName);
    gGlobalVariablePool->set(configHead + "OS/MajorVersion",
                             configInfo.osMajorVersion);
    gGlobalVariablePool->set(configHead + "OS/MinorVersion",
                             configInfo.osMinorVersion);
    gGlobalVariablePool->set(configHead + "OS/Revision",
                             configInfo.osRevision);
    gGlobalVariablePool->set(configHead + "Mem/Physical/Bytes",
                             STAFString(configInfo.physicalMemory));
    gGlobalVariablePool->set(configHead + "Mem/Physical/KB",
                             STAFString(configInfo.physicalMemory / 1024));
    gGlobalVariablePool->set(configHead + "Mem/Physical/MB",
                             STAFString(configInfo.physicalMemory /
                                           (1024 * 1024)));
    gGlobalVariablePool->set(configHead + "STAFRoot", configInfo.exePath);
    gGlobalVariablePool->set(configHead + "Sep/Line",
                             configInfo.lineSeparator);
    gGlobalVariablePool->set(configHead + "Sep/File",
                             configInfo.fileSeparator);
    gGlobalVariablePool->set(configHead + "Sep/Path",
                             configInfo.pathSeparator);
    gGlobalVariablePool->set(configHead + "Sep/Command",
                             configInfo.commandSeparator);
    gGlobalVariablePool->set(configHead + "CodePage",
                             STAFConverter::determineCodePage());

    gLineSeparator = configInfo.lineSeparator;
    gFileSeparator = configInfo.fileSeparator;
    gPathSeparator = configInfo.pathSeparator;
    gCommandSeparator = configInfo.commandSeparator;

    // Set writeable STAF location.  Default is {STAF/Config/STAFRoot}/data.

    STAFFSPath writeLocation;

    writeLocation.setRoot(configInfo.exePath);
    writeLocation.addDir("data");

    gSTAFWriteLocation = writeLocation.asString();

    // Set the config file to use

    STAFString configFile;

    if (argc == 2)
    {
        configFile = argv[1];
    }
    else if (argc > 2)
    {
        cout << "Usage: STAFProc [Config file]" << endl;
        return 1;
    }
    else
    {
        configFile = STAFString(configInfo.exePath) +
                     STAFString(configInfo.fileSeparator) + "bin" +
                     STAFString(configInfo.fileSeparator) + "STAF.cfg";
    }

    // Set environment buffer for use by Process service

    STAFString envHead("STAF/Env/");

    for (int envi = 0; envp[envi] != 0; ++envi)
    {
        STAFString envVar(envp[envi]);
        unsigned int equalPos = envVar.find(kUTF8_EQUAL);
        STAFString envName(envVar.subString(0, equalPos));
        STAFString envValue;

        if (equalPos != STAFString::kNPos)
            envValue = envVar.subString(equalPos + 1);

        gEnvMap[envName] = envValue;

        gGlobalVariablePool->set(envHead + envName, envValue);
    }

    int envc = 0;

    for(; envp[envc] != 0; ++envc)
        gEnvSize += strlen(envp[envc]) + 1;

    gEnvBuffer = new unsigned char[++gEnvSize];
    memset(gEnvBuffer, 0, gEnvSize);

    for(gEnvSize = 0, envc = 0; envp[envc] != 0; ++envc)
    {
        int length = strlen(envp[envc]);
        memcpy(gEnvBuffer + gEnvSize, envp[envc], length);
        gEnvSize += length + 1;
    }

    STAFRC_t regRC = gHandleManager.registerHandle(
                          gSTAFProcHandle, gSTAFProcPID, "STAF_Process");
    if (regRC)
    {
        cout << "Error registering STAF Process handle, RC: " << regRC << endl;
        return 1;
    }

    STAFRC_t createRC = STAFHandle::create(gSTAFProcHandle, gSTAFProcHandlePtr);

    if (createRC)
    {
        cout << "Error creating STAF Process handle, RC: " << createRC << endl;
        return 1;
    }

    try
    {
        // Add all the internal services

        gServiceManager.add(STAFServicePtr(new STAFDelayService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFDiagService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFEchoService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFHelpService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFFSService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFHandleService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFMiscService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFPingService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFProcessService(),
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFQueueService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFSemService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFServiceService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFShutdownService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFTrustService(), 
                                           STAFServicePtr::INIT));
        gServiceManager.add(STAFServicePtr(new STAFVariableService(), 
                                           STAFServicePtr::INIT));

        // Process the config file

        gGlobalVariablePool->set(configHead + "ConfigFile", configFile);

        unsigned int configRC = readConfigFile(configFile);

        if (configRC != 0)
        {
            cout << "Error reading config file" << endl;
            return 1;
        }

        // Initialize the TCP/IP interface, if necessary

        if (gUseTCPIPInterface != 0)
        {
            // Get machine name

            if (STAFSocketFuncs::initSocketFuncs() != 0)
            {
                cout << "Error initializing socket functions" << endl;
                return 1;
            }

            STAFString machineName;

            if (STAFSocketFuncs::getRealHostName(machineName) != 0)
            {
                cout << "Could not determine hostname" << endl;
                return 1;
            }

            gMachineNetName = machineName;

            cout << "HostName: " << machineName << endl;

            if ((gUseLongNames == 0) &&
                (machineName.find(kUTF8_PERIOD) != 
                 STAFString::kNPos))
            {
                machineName = machineName.subString(0,
                              machineName.find(kUTF8_PERIOD));
            }

            if (gMachineName != "")
            {
                if (gMachineName.toLowerCase() != machineName.toLowerCase())
                {
                    cout << "When using the TCPIP interface, the MACHINE "
                         << "name and the TCPIP hostname must match" << endl;
                    return 1;
                }
            }
            else gMachineName = machineName;

        }  // end if using tcpip
        else
        {
            // Only using local interface

            gMachineNetName = gMachineName;
        }

        if (gMachineName == STAFString())
        {
            cout << "Unknown machine name" << endl;
            return 1;
        }

        // Set the Machine and EffectiveMachine variables

        gGlobalVariablePool->set(configHead + "Machine", gMachineNetName);
        gGlobalVariablePool->set(configHead + "EffectiveMachine",
                                 gMachineName);

        cout << "Machine name: " << gMachineName << endl;

        // Set initial number of threads
        gThreadManagerPtr->growThreadPool(gNumInitialThreads);

        // Make performance adjustments

        STAFPerfInfo perfInfo = { (unsigned int)gMaxFiles };

        if (makePerformanceAdjustments(perfInfo, errorBuffer, osRC) != 0)
        {
            cout << errorBuffer << ", RC:" << osRC << endl;
            return 1;
        }

        gShutdownSemaphore = STAFEventSemPtr(new STAFEventSem(),
                                             STAFEventSemPtr::INIT);
        // Start local interface

        gInterfaceSem.reset();

        STAFConnectionFactory *localFactory =
            new STAFLocalConnectionFactory("STAFProc");

        gThreadManagerPtr->dispatch(Receive, localFactory);

        gInterfaceSem.wait();

        // This sleep is here to "guarantee" the local interface has
        // initialized completely

        gThreadManagerPtr->sleepCurrentThread(1000);
        
        // Initialize the service loader services

        STAFServiceManager::ServiceList slsList =
                                        gServiceManager.getSLSListCopy();

        STAFServiceManager::ServiceList::iterator slsIter;

        for(slsIter = slsList.begin(); 
            slsIter != slsList.end(); ++slsIter)
        {
            unsigned int rc = (*slsIter)->initialize();

            if (rc)
            {
                cout << "Error initializing service loader service, "
                     << (*slsIter)->name() << ", RC: " << rc << endl;
            }
        }

        // Don't hold on to service pointers we will never use
        slsList = STAFServiceManager::ServiceList();

        // Initialize the services.  Note that the internal services were
        // already initialized, so this will only initialize the external
        // services.

        STAFServiceManager::ServiceList serviceList =
                                        gServiceManager.getServiceListCopy();

        STAFServiceManager::ServiceList::iterator serviceIter;

        for(serviceIter = serviceList.begin(); 
            serviceIter != serviceList.end(); ++serviceIter)
        {
            unsigned int rc = (*serviceIter)->initialize();

            if (rc)
            {
                cout << "Error initializing service, "
                     << (*serviceIter)->name() << ", RC: " << rc << endl;
                     
                gServiceManager.remove((*serviceIter)->name());
            }
        }

        // Don't hold on to service pointers we will never use
        serviceList = STAFServiceManager::ServiceList();

        // Start network interface(s)

        if (gUseTCPIPInterface != 0)
        {
            gInterfaceSem.reset();

            STAFConnectionFactory *socketFactory =
                new STAFSocketConnectionFactory(gTCPIPPort);

            gThreadManagerPtr->dispatch(Receive, socketFactory);

            gInterfaceSem.wait(); 
        }

        cout << endl << "STAFProc version " << gVersion << " initialized"
             << endl;

        // Send start notifications

        gNotifyOnStart.sendNotification("STAF/START");

        // Call registration program if necessary

        STAFString infFileName = STAFString(configInfo.exePath) +
                                 STAFString(configInfo.fileSeparator) +
                                 "STAFReg.inf";
        STAFString cmpFileName = STAFString(configInfo.exePath) +
                                 STAFString(configInfo.fileSeparator) +
                                 "STAFReg.cmp";

        if (STAFFSPath(infFileName).exists() &&
            !STAFFSPath(cmpFileName).exists())
        {
            system("STAFReg");
        }

        // Wait for signal to end

        gShutdownSemaphore->wait();

        cout << "STAFProc ending normally" << endl;

        // Send shutdown notifications

        gNotifyOnShutdown.sendNotification("STAF/SHUTDOWN");

        // Terminate services

        serviceList = gServiceManager.getServiceListCopy();

        for(STAFServiceManager::ServiceList::reverse_iterator serviceIter2 =
                serviceList.rbegin();
            serviceIter2 != serviceList.rend(); ++serviceIter2)
        {
            STAFServicePtr theService = *serviceIter2;

            gServiceManager.remove(theService->name());
            theService->terminate();
        }
        
        // Terminate service loader services

        slsList = gServiceManager.getSLSListCopy();

        for(STAFServiceManager::ServiceList::reverse_iterator slsIter2 =
                slsList.rbegin();
            slsIter2 != slsList.rend(); ++slsIter2)
        {
            STAFServicePtr theService = *slsIter2;

            gServiceManager.removeSLS(theService);
            theService->terminate();
        }
        
        // Terminate any remaining services that might have been re-initialized
        // by a service loader during another service's terminiation

        serviceList = gServiceManager.getServiceListCopy();

        for(STAFServiceManager::ServiceList::reverse_iterator serviceIter3 =
                serviceList.rbegin();
            serviceIter3 != serviceList.rend(); ++serviceIter3)
        {
            STAFServicePtr theService = *serviceIter3;

            gServiceManager.remove(theService->name());
            theService->terminate();
        }

        // XXX: Without this sleep, the system doesn't get time to cleanup
        // everything before the main thread ends, resulting in an extra
        // message stating that terminate() was called.  I should probably
        // try to debug this further, but this masks the problem sufficiently
        // at the moment.

        gThreadManagerPtr->sleepCurrentThread(1000);

        // clean up any OS specific stuff.

        int stafOSTermRC = STAFProcOSTerm(errorBuffer, osRC);

        if (stafOSTermRC != 0)
        {
            cout << "Error during platform-specific termination"  << endl;
            cout << errorBuffer << ", RC: " << osRC << endl;
            return 1;
        }
    }
    catch (STAFException &se)
    { se.write("main()"); }
    catch (...)
    { cout << "Caught unknown exception in main()" << endl; }

    return 0;
}

unsigned int Receive(void *factoryPtr)
{
    try
    {
        gInterfaceSem.post();

        STAFConnFactoryPtr factory((STAFConnectionFactory *)factoryPtr,
                                   STAFConnFactoryPtr::INIT);
        for(;;)
        {
            STAFConnAllocInfo allocInfo = factory->acceptConnection();

            gThreadManagerPtr->dispatch(HandleRequest,
                                        new STAFConnAllocInfo(allocInfo));
        }
    }
    catch (STAFException &se)
    { se.write("Receive()"); }
    catch (...)
    { cout << "Caught unknown exception in Receive()" << endl; }

    gShutdownSemaphore->post();

    return 0;
}


unsigned int HandleRequest(void *allocInfoPtr)
{
    unsigned int doShutdown = 0;

    try
    {
        STAFConnAllocInfo *pAllocInfo =
                          static_cast<STAFConnAllocInfo *>(allocInfoPtr);
        STAFConnPtr connection(pAllocInfo->pConnection, STAFConnPtr::INIT,
                               STAFConnPtr::CUSTOM, pAllocInfo->pFreeFunc);
        delete pAllocInfo;

        unsigned int readBuffer[2];

        connection->read(readBuffer, sizeof(readBuffer));

        unsigned int apiNum = STAFUtilConvertLEUIntToNative(readBuffer[0]);
        unsigned int apiLevel = STAFUtilConvertLEUIntToNative(readBuffer[1]);
        unsigned int minLevel = apiLevel;
        unsigned int maxLevel = apiLevel;

        if ((apiNum <= gMaxAPINumber) && (apiLevel == 0) &&
            (gAPITable[apiNum].apiType == kNewAPI))
        {
            // Need to send the ok and then read the min and max levels

            connection->writeUInt(kSTAFOk);
            minLevel = connection->readUInt();
            maxLevel = connection->readUInt();
        }

        if (apiNum > gMaxAPINumber)
        {
            if (STAFTrace::doTrace(kSTAFTraceError))
            {
                STAFString message("Invalid API number (");
                message += STAFString(apiNum) + ") received";
                STAFTrace::trace(kSTAFTraceError, message);
            }

            connection->writeUInt(kSTAFInvalidAPI);
        }
        else if ((maxLevel < gAPITable[apiNum].minLevel) ||
                 (minLevel > gAPITable[apiNum].maxLevel))
        {
            if (STAFTrace::doTrace(kSTAFTraceError))
            {
                STAFString message("Invalid API level (");
                message += STAFString(apiLevel) + ") received for API number " +
                           apiNum;
                STAFTrace::trace(kSTAFTraceError, message);
            }

            connection->writeUInt(kSTAFInvalidAPILevel);
        }
        else if (!(connection->connectionType() & gAPITable[apiNum].connMask))
        {
            if (STAFTrace::doTrace(kSTAFTraceError))
            {
                STAFString message("Invalid connection type (");
                message += STAFString(connection->connectionType()) +
                           ") received for API number" + apiNum;
                STAFTrace::trace(kSTAFTraceError, message);
            }

            connection->writeUInt(kSTAFInvalidAPI);
        }
        else
        {
            if (gAPITable[apiNum].apiType == kOldAPI)
            {
                connection->writeUInt(kSTAFOk);
            }
            else
            {
                apiLevel = STAF_MIN(maxLevel, gAPITable[apiNum].maxLevel);
                connection->writeUInt(apiLevel);
            }

            gAPITable[apiNum].handler(apiLevel, connection, doShutdown);
        }
    }
    catch (STAFConnectionIOException)
    {
        // A client has unexpectedly terminated the connection
        // Do nothing, cleanup is automatic

        STAFTrace::trace(kSTAFTraceError, "HandleRequest(): Connection "
                                          "terminated unexpectedly");
    }
    catch (STAFException &se)
    { se.write("HandleRequest()"); }
    catch (...)
    { cout << "Caught unknown exception in HandleRequest()" << endl; }

    if (doShutdown) gShutdownSemaphore->post();

    return 0;
}


void handleLocalServiceRequestAPI(unsigned int level, STAFConnPtr connection,
                                  unsigned int &doShutdown)
{
    STAFString errorBuffer;

    unsigned int readBuffer[6];

    connection->read(readBuffer, sizeof(readBuffer));

    STAFSyncOption_t syncOption = STAFUtilConvertLEUIntToNative(readBuffer[0]);
    STAFProcessID_t pid = STAFUtilConvertLEUIntToNative(readBuffer[1]);
    STAFHandle_t handle = STAFUtilConvertLEUIntToNative(readBuffer[2]);
    unsigned int whereLength = STAFUtilConvertLEUIntToNative(readBuffer[3]);
    unsigned int serviceLength = STAFUtilConvertLEUIntToNative(readBuffer[4]);
    unsigned int requestLength = STAFUtilConvertLEUIntToNative(readBuffer[5]);
    unsigned int buffSize = whereLength + serviceLength + requestLength;
    STAFBuffer<char> buffer(new char[buffSize], STAFBuffer<char>::INIT);

    connection->read(buffer, buffSize);

    STAFString unresWhere(buffer, whereLength, STAFString::kUTF8);
    STAFString unresServiceName(buffer + whereLength, serviceLength,
                                STAFString::kUTF8);
    STAFString request(buffer + whereLength + serviceLength, requestLength,
                       STAFString::kUTF8);
    STAFString where;
    STAFServiceResult serviceResult;
    STAFVariablePoolPtr handlePool;
    STAFRC_t rc = gHandleManager.updateTimestamp(handle, pid);

    if (!rc) rc = gHandleManager.variablePool(handle, handlePool);
    if (!rc) rc = resolveString(unresWhere, handlePool, where, errorBuffer);

    if (rc)
    {
        // XXX: Add a trace statement here.  Differentiate between invalid
        //      handle and bad "where"

        connection->writeUInt(rc);
        connection->writeString(errorBuffer);
    }
    else
    {
        // Resolved string, where, is stripped to allow begining and/or
        // trailing white spaces

        where = where.strip(STAFString::kBoth);

        STAFString targetMachine;
        if (!isLocalMachine(where, 1))
            targetMachine = where;            
        else
            targetMachine = gMachineName;

        STAFRequestManager::RequestData requestData = 
                    STAFRequestManager::RequestData(
                                    effectiveMachineName(gMachineName), 
                                    gHandleManager.name(handle), handle,
                                    STAFTimestamp::now(),
                                    effectiveMachineName(targetMachine), 
                                    unresServiceName, request);

        STAFRequestManager::STAFReqNum_t requestNumber;

        rc = gRequestManager.add(requestData, &requestNumber);   

        try
        {
            if (!isLocalMachine(where, 1))
            {            
                // Service request to a remote system            
            
                serviceResult = submitRemoteRequest(where, handle, connection,
                                                    unresServiceName,
                                                    syncOption, requestNumber,
                                                    request);            
            }
            else
            {
                // Local service request
            
                STAFString serviceName;            

                rc = resolveString(unresServiceName, handlePool, serviceName,
                                   errorBuffer);

                if (rc)
                {
                    serviceResult.fServicePerformedRC = rc;
                    serviceResult.fResult = errorBuffer;
                }
                else
                {   
                    // Resolved string, serviceName, is stripped to allow
                    // begining and/or trailing white spaces

                    serviceName = serviceName.strip(STAFString::kBoth);

                    serviceResult = submitLocalRequest(connection, 
                                                   serviceName, 
                                                   gMachineNetName,               
                                                   gHandleManager.name(handle),
                                                   syncOption, requestNumber, 
                                                   handle, request);
                }
            }                

            if (syncOption == kSTAFReqSync) 
            {
                connection->writeUInt(serviceResult.fServicePerformedRC);    
                connection->writeString(serviceResult.fResult);            
            }

            doShutdown = serviceResult.fDoShutdown; 
        }
        catch (STAFException &e)
        {
            if ((syncOption == kSTAFReqRetain) || 
                (syncOption == kSTAFReqQueueRetain))
            {        
                STAFServiceResult serviceResult;
                serviceResult.fServicePerformedRC = e.getErrorCode();
                serviceResult.fResult = e.getText();
                gRequestManager.requestCompleted(requestNumber, serviceResult); 
            } 
            else
            {
                if (gRequestManager.requestExists(requestNumber))
                    gRequestManager.deleteRequest(requestNumber);
            }

            throw;           
        }
        catch (...)
        {
            if (gRequestManager.requestExists(requestNumber))
                gRequestManager.deleteRequest(requestNumber);

            throw;
        }    
    }
}


void handleRemoteServiceRequestAPI(unsigned int level, STAFConnPtr connection,
                                   unsigned int &doShutdown)
{
    STAFString machineName = connection->readString();
    STAFHandle_t handle = connection->readUInt();
    STAFString processName = connection->readString();
    STAFString unresServiceName = connection->readString();
    STAFString request = connection->readString();

    STAFString errorBuffer;
    STAFServiceResult serviceResult;

    // Note: We can only resolve global variables here, as this request
    //       came from a remote system

    STAFString serviceName;
    STAFRC_t rc = resolveString(unresServiceName, 0, serviceName,
                                     errorBuffer);


    if (rc)
    {
        serviceResult.fServicePerformedRC = rc;
        serviceResult.fResult = errorBuffer;        
    }
    else
    {
        // Resolved string, serviceName, is stripped to allow begining and/or
        // trailing white spaces

        serviceName = serviceName.strip(STAFString::kBoth);

        STAFRequestManager::RequestData requestData = 
            STAFRequestManager::RequestData(effectiveMachineName(machineName), 
                                            processName, handle, 
                                            STAFTimestamp::now(),
                                            effectiveMachineName(gMachineName), 
                                            unresServiceName, request);
        STAFRequestManager::STAFReqNum_t requestNumber;
        rc = gRequestManager.add(requestData, &requestNumber);    

        try 
        {
            serviceResult = submitLocalRequest(connection, serviceName,
                                               machineName, processName,
                                               // Remote is always synchronous
                                               kSTAFReqSync,  
                                               requestNumber, handle, request);
        }
        catch (...)
        {
            if (gRequestManager.requestExists(requestNumber))
                gRequestManager.deleteRequest(requestNumber);
            throw;
        }    
    }
    
    connection->writeUInt(serviceResult.fServicePerformedRC);
    connection->writeString(serviceResult.fResult);
    doShutdown = serviceResult.fDoShutdown; 
}

    
STAFServiceResult submitLocalRequest(STAFConnPtr connection, 
                               STAFString serviceName, 
                               STAFString machineName,
                               STAFString processName,
                               STAFSyncOption_t syncOption,
                               STAFRequestManager::STAFReqNum_t requestNumber,
                               STAFHandle_t handle, STAFString request)
{
    STAFServicePtr service;
    STAFServiceResult serviceResult;    
    STAFServiceResult requestResult;

    STAFRC_t rc = gServiceManager.get(serviceName, service);

    if (rc)
    {
        if ((syncOption != kSTAFReqRetain) && 
            (syncOption != kSTAFReqQueueRetain)) 
        { 
            gRequestManager.requestCompleted(requestNumber, serviceResult); 
            gRequestManager.freeRequest(requestNumber, &requestResult); 
        }
       
        if (rc == kSTAFDoesNotExist)
            rc = kSTAFUnknownService;

        if (syncOption != kSTAFReqSync)                     
        {
            // pass back the request number
            
            serviceResult = STAFServiceResult(kSTAFOk, requestNumber);
            connection->writeUInt(serviceResult.fServicePerformedRC);
            connection->writeString(serviceResult.fResult);

            // send the rc/result to the requester's queue
            if ((syncOption == kSTAFReqQueue) || 
                (syncOption == kSTAFReqQueueRetain))
            {
                STAFString resultMessage(sRequestComplete);

                resultMessage += requestNumber;
                resultMessage += sSemiColon;
                resultMessage += rc;
                resultMessage += sSemiColon;
                resultMessage += serviceName;

                STAFString queueMessage(sQueueHandle);

                queueMessage += handle;
                queueMessage += sMessage;            
                queueMessage += sColon;
                queueMessage += resultMessage.length(STAFString::kChar);
                queueMessage += sColon;
                queueMessage += resultMessage;
                
                gSTAFProcHandlePtr->submit(sLocal, sQueue, queueMessage,
                                           kSTAFReqFireAndForget);
            }
            return STAFServiceResult(kSTAFOk);
        }
        else
        {
            return STAFServiceResult(rc, serviceName);
        }
    }
    else
    {               
        // Let the submitter continue by sending back the service result
        // and then submit the actual request

        if (syncOption != kSTAFReqSync)                     
        {                        
            // pass back the request number
            serviceResult = STAFServiceResult(kSTAFOk, requestNumber);
            connection->writeUInt(serviceResult.fServicePerformedRC);
            connection->writeString(serviceResult.fResult);
        }
        
        STAFServiceRequest requestInfo;

        requestInfo.fMachine = machineName;
        requestInfo.fHandle = handle;
        requestInfo.fHandleName = processName;
        requestInfo.fRequest = request;
        requestInfo.fTrustLevel = gTrustManager.trustLevel(machineName);
        requestInfo.fDiagEnabled = gDiagManager.getEnabled();

        serviceResult = service->submitRequest(requestInfo);
    }

    gRequestManager.requestCompleted(requestNumber, serviceResult);              

    if ((syncOption == kSTAFReqQueue) || 
        (syncOption == kSTAFReqQueueRetain))
    {    
        STAFString resultMessage(sRequestComplete);

        resultMessage += requestNumber;
        resultMessage += sSemiColon;
        resultMessage += serviceResult.fServicePerformedRC;
        resultMessage += sSemiColon;            
        resultMessage += serviceResult.fResult;

        STAFString queueMessage(sQueueHandle);

        queueMessage += handle;
        queueMessage += sMessage;            
        queueMessage += sColon;
        queueMessage += resultMessage.length(STAFString::kChar);
        queueMessage += sColon;
        queueMessage += resultMessage;                        

        gSTAFProcHandlePtr->submit(sLocal, sQueue, queueMessage,
                                   kSTAFReqFireAndForget);
    }   
    
    if ((syncOption != kSTAFReqRetain) && (syncOption != kSTAFReqQueueRetain))
    {        
        gRequestManager.freeRequest(requestNumber, &requestResult);    
    }
    
    return serviceResult;
}


STAFServiceResult submitRemoteRequest(STAFString &where, STAFHandle_t &handle,
                                  STAFConnPtr connection, STAFString &service, 
                                  STAFSyncOption_t syncOption,
                                  STAFRequestManager::STAFReqNum_t requestNumber,
                                  STAFString &request)
{
    // Note: This routine only handles sending remote requests which
    //       originated on this machine

    // Let the submitter continue and then submit the request


    if (syncOption != kSTAFReqSync)
    {                        
        // pass back the request number
        STAFServiceResult asyncServiceResult =
                          STAFServiceResult(kSTAFOk, requestNumber);
        connection->writeUInt(asyncServiceResult.fServicePerformedRC);
        connection->writeString(asyncServiceResult.fResult);
    }

    if (STAFTrace::doTrace(kSTAFTraceServiceRequest |
                           kSTAFTraceRemoteRequests) &&
        STAFServiceManager::doTraceService(service))
    {
        STAFString data("Remote ");

        data += service + " Service Request - To: " + where +
                ", from Handle: " + STAFString(handle) + ", Process: " +
                gHandleManager.name(handle) + ", Request: " + request;

        STAFTrace::trace(kSTAFTraceServiceRequest, data);
    }

    if (gUseTCPIPInterface == 0)
        return STAFServiceResult(kSTAFNoPathToMachine, STAFString());

    STAFRC_t rc = kSTAFUnknownError;
    STAFString result;

    try
    {
        STAFConnPtr connection;
        
        rc = gConnectionManager.makeConnection(where, connection, result);
       
        if (rc == kSTAFOk)
        {            
            connection->writeUInt(1);    // API Number
            connection->writeUInt(0);    // API Level
            
            STAFRC_t ack = static_cast<STAFRC_t>(connection->readUInt());
            if (ack != kSTAFOk) return ack;

            connection->writeString(gMachineNetName);
            connection->writeUInt(handle);
            connection->writeString(gHandleManager.name(handle));
            connection->writeString(service);
            connection->writeString(request);
            
            rc = static_cast<STAFRC_t>(connection->readUInt());
            result = connection->readString();            
        }
    }
    catch (STAFConnectionConnectException &e)
    {
        rc = kSTAFNoPathToMachine;
        result = e.getText() + STAFString(": ") + STAFString(e.getErrorCode());
    }
    catch (STAFConnectionIOException &e)
    {        
        rc = kSTAFCommunicationError;
        result = e.getText() + STAFString(": ") + STAFString(e.getErrorCode());
    }
    catch (STAFException &se)
    { se.write("submitRemoteRequest()"); }
    catch (...)
    { cout << "Caught unknown exception in submitRemoteRequest()" << endl; }

    if (STAFTrace::doTrace(kSTAFTraceServiceResult |
                           kSTAFTraceRemoteRequests) &&
        STAFServiceManager::doTraceService(service))
    {
        STAFString data("Remote ");

        data += service + " Service Result (" + rc + ") - To: " + where +
                ", from Handle: " + STAFString(handle) + ", Process: " +
                gHandleManager.name(handle) + ", Request: " + request +
                 ", Result: " + ((result.length() == 0) ?
                STAFString("<No Result>") : result);

        STAFTrace::trace(kSTAFTraceServiceResult, data);
    }
    else if ((rc != 0) &&
             STAFTrace::doTrace(kSTAFTraceServiceError |
                                kSTAFTraceRemoteRequests) &&
             STAFServiceManager::doTraceService(service))
    {
        STAFString data("Remote ");

        data += service + " Service Non-Zero (" + rc + ") Return Code - To: " +
                where + ", from Handle: " + STAFString(handle) + ", Process: " +
                gHandleManager.name(handle) + ", Request: " + request +
                ", Result: " + ((result.length() == 0) ?
                STAFString("<No Result>") : result);

        STAFTrace::trace(kSTAFTraceServiceError, data);
    }
    else if ((rc == kSTAFAccessDenied) &&
             STAFTrace::doTrace(kSTAFTraceServiceAccessDenied |
                                kSTAFTraceRemoteRequests) &&
             STAFServiceManager::doTraceService(service))
    {
        STAFString data("Remote ");

        data += service + " Service Access Denied - To: " + where +
                ", from Handle: " + STAFString(handle) + ", Process: " +
                 gHandleManager.name(handle) + ", Request: " + request;

        STAFTrace::trace(kSTAFTraceServiceAccessDenied, data);
    }

    STAFServiceResult serviceResult = STAFServiceResult(rc, result);

    gRequestManager.requestCompleted(requestNumber, serviceResult);    

    if ((syncOption == kSTAFReqQueue) | 
        (syncOption == kSTAFReqQueueRetain))
    {    
        STAFString resultMessage(sRequestComplete);

        resultMessage += requestNumber;
        resultMessage += sSemiColon;
        resultMessage += serviceResult.fServicePerformedRC;
        resultMessage += sSemiColon;            
        resultMessage += serviceResult.fResult;

        STAFString queueMessage(sQueueHandle);

        queueMessage += handle;
        queueMessage += sMessage;            
        queueMessage += sColon;
        queueMessage += resultMessage.length(STAFString::kChar);
        queueMessage += sColon;
        queueMessage += resultMessage;                        
            
        gSTAFProcHandlePtr->submit(sLocal, sQueue, queueMessage,
                                   kSTAFReqFireAndForget);
    }     
    
    STAFServiceResult requestResult;
    if ((syncOption != kSTAFReqRetain) && (syncOption != kSTAFReqQueueRetain))
    {        
        gRequestManager.freeRequest(requestNumber, &requestResult);    
    }

    return serviceResult;
}


void handleProcessRegistrationAPI(unsigned int level, STAFConnPtr connection,
                                  unsigned int &)
{
    STAFHandle_t handle = 0;
    STAFProcessID_t pid = connection->readUInt();
    STAFString name = connection->readString();
    STAFRC_t rc = gHandleManager.registerHandle(handle, pid, name);

    if (STAFTrace::doTrace(kSTAFTraceRegistration))
    {
        STAFString message;

        if (rc == 0)
        {
            message += STAFString("Registered Handle: ") + handle +
                       ",PID: " + pid + ", Name: " + name;
        }
        else
        {
            message += STAFString("Error registering handle with PID: ") + pid +
                       "Name: " + name + ", RC: " + rc;
        }

        STAFTrace::trace(kSTAFTraceRegistration, message);
    }

    connection->writeUInt(rc);
    connection->writeUInt(handle);
}


void handleProcessUnRegistrationAPI(unsigned int level, STAFConnPtr connection,
                                    unsigned int &)
{
    STAFProcessID_t pid = connection->readUInt();
    STAFHandle_t handle = connection->readUInt();

    connection->writeUInt(gHandleManager.unRegister(handle, pid));

    if (STAFTrace::doTrace(kSTAFTraceRegistration))
    {
        STAFString message("Unregistered Handle: ");

        message += STAFString(handle) + ", PID: " + pid;

        STAFTrace::trace(kSTAFTraceRegistration, message);
    }
}


void handleFileTransferAPI(unsigned int level, STAFConnPtr connection,
                           unsigned int &)
{
    handleFileTransfer(kOldAPI, level, connection);
}


void handleFileTransferAPI2(unsigned int level, STAFConnPtr connection,
                            unsigned int &)
{
    handleFileTransfer(kNewAPI, level, connection);
}


void handleDirectoryCopyAPI(unsigned int level, STAFConnPtr connection, 
                            unsigned int &)
{
    // fromMachine is the machine that the files/directories are coming from
    // orgMachine is the machine that originated the request

    STAFString fromMachine = connection->readString();
    STAFString orgMachine = connection->readString();
    STAFString unresToDirectory = connection->readString();
    unsigned int flags = connection->readUInt();
    STAFString newEOL = STAFString(kUTF8_NULL);

    if (level > 1)
        newEOL = connection->readString();

    STAFString toDirectory = unresToDirectory;
    STAFString errorBuffer;
    STAFString result;
    STAFRC_t rc = kSTAFOk;
    unsigned int osRC = 0;

    // Only perform the access check when the request did not originate from
    // this machine.

    if (!isLocalMachine(orgMachine))
    {
        if (gTrustManager.trustLevel(fromMachine) < 4)
        {
            rc = kSTAFAccessDenied;
            result = fromMachine;
        }
        else if (gTrustManager.trustLevel(orgMachine) < 4)
        {
            rc = kSTAFAccessDenied;
            result = orgMachine;
        }
    }

    if (rc == kSTAFOk)
    {
        // Note: If the request is to and from the same machine then the
        //       TODIRECTORY string will already be resolved by the FS service,
        //       so we only have to worry about global resolutions for
        //       machines other than ourselves
        rc = resolveString(unresToDirectory, 0, toDirectory, errorBuffer);
        if (rc) result = errorBuffer;
    }

    if (rc)
    {
        connection->writeUInt(rc);
        connection->writeString(result);

        return;
    }

    unsigned int theDirectoryExists = STAFFSPath(toDirectory).exists();

    if ((flags & 0x00000001) && theDirectoryExists) 
    {
        connection->writeUInt(kSTAFAlreadyExists);
        connection->writeString(toDirectory);

        return;
    }

    if ((flags & 0x00000002) && !theDirectoryExists)
    {
        connection->writeUInt(kSTAFDoesNotExist);
        connection->writeString(toDirectory);

        return;
    }

    // Establish eol
    STAFString tmpEOL = newEOL;

    // Determine new EOL
    if (newEOL.lowerCase() == "native")
    {
        STAFConfigInfo sysInfo;
        STAFString_t errorBuffer;
        unsigned int osRC;
        
        if (STAFUtilGetConfigInfo(&sysInfo, &errorBuffer, &osRC) != kSTAFOk)
        {    
            // Kill transfer
            connection->writeUInt(kSTAFBaseOSError);
            connection->writeString("Get EOL failed");
            return;
        }
        newEOL=sysInfo.lineSeparator;
    }
    else if (newEOL.lowerCase() == "windows")
    {
        newEOL = STAFString(kUTF8_CR) + STAFString(kUTF8_LF);// \r\n
    }
    else if (newEOL.lowerCase() == "unix")
    {
        newEOL = STAFString(kUTF8_LF);// \n
    }
    else
        newEOL = tmpEOL;


    rc = kSTAFOk;
    result = STAFString();

    // Create directory if !theDirectoryExists
    if (!theDirectoryExists)
    {
        rc = STAFFSCreateDirectory(toDirectory.getImpl(), kSTAFFSCreatePath, 
                                   &osRC);

        if (rc == kSTAFBaseOSError) 
            result = "OSRC: " + STAFString(osRC) + 
                     ", Error on Creating a Directory: " + toDirectory;
    }

    connection->writeUInt(rc);
    connection->writeString(result);

    if (rc != kSTAFOk) return;

    if (level > 1)
    {
        connection->writeString(newEOL);
 
        // Allow the use of the text noconvert routine if the codepages on
        // both sides are the same.  This allows for a faster transfer.
        bool doCodepageConvert = (connection->readUInt() == kSTAFFSTextConvert);

        if (doCodepageConvert)
        {
            STAFString cPage;
            rc = resolveString("{STAF/Config/CodePage}", 0, cPage, errorBuffer);

            // Send this codepage id
            connection->writeString(cPage);
        }  
    }

    rc = kSTAFOk;
    result = STAFString();

    /************** Start transferring the directory ***************/

    for (unsigned int stop = connection->readUInt(); !stop;
         stop = connection->readUInt())
    {
        unsigned int isFile = connection->readUInt();

        if (isFile == kSTAFFSDirectory)
        {
            // Current entry is a directory

            STAFString subDirectory = connection->readString();
            STAFString newDirectory = toDirectory + subDirectory;

            rc = STAFFSCreateDirectory(newDirectory.getImpl(), 
                                       kSTAFFSCreatePath, &osRC);

            if (rc == kSTAFBaseOSError) result = osRC;
            else if (rc == kSTAFAlreadyExists)
            {
                // We will override the current subDirectory
                rc = kSTAFOk;
                result = STAFString();
            }

            connection->writeUInt(rc);
            connection->writeString(result);
        }
        else if (isFile == kSTAFFSFile)
        {
            // Current entry is a file.  Copy it.

            STAFString newFile = connection->readString();
            STAFString toFile = toDirectory + newFile;
            
            bool textTransfer = false;
            bool doCodepageConvert = false; 
            unsigned int transferType;

            //  IF API appropriate check for text transfer
            //  level > 1
            if (level > 1) 
            { 
                // Determine transfer type
                transferType= connection->readUInt();
                textTransfer = (transferType != kSTAFFSBinary);
                doCodepageConvert = (transferType == kSTAFFSTextConvert);
            }

            fstream outFile(toFile.toCurrentCodePage()->buffer(),
                            STAF_ios_binary | ios::out);

            if (!outFile)
            {
                connection->writeUInt(kSTAFFileOpenError);
                connection->writeString(toFile);

                continue; 
            }

            // outFile is ok
            connection->writeUInt(kSTAFOk);
            connection->writeString(STAFString());

            if (doCodepageConvert)
            {
                // Perform a text transfer with conversion of eol chars and a
                // codepage conversion.  It runs considerably slower than the
                // binary transfer.
                // - This is the type of transfer that will occur if the TEXT
                //   option is specified and the codepages are different.

                STAFString currentEOL = connection->readString();
                STAFString transferStr;
                STAFStringBufferPtr convertedStr;
                int continueCopy = connection->readUInt();

                // If an error occurs writing to the file, continue

                while (continueCopy != kSTAFFSFinishedCopy)
                {
                    transferStr = connection->readString();
                    transferStr = transferStr.replace(currentEOL, newEOL);
                    convertedStr = transferStr.toCurrentCodePage();

                    rc = writeFile(
                        outFile, convertedStr->buffer(),
                        convertedStr->length(), toFile, fromMachine);

                    continueCopy = connection->readUInt();
                }

                outFile.close();
        
                // Indicate whether the copy was successful
                connection->writeUInt(rc);
            }
            else if (textTransfer)
            {
                // Perform a text transfer with conversion of eol chars without
                // a codepage conversion.  It runs considerably faster than the
                // codepage conversion transfer.
                // -  This is the type of transfer that will occur if the
                //    NOCONVERT option is enabled.
                // -  This is the type of transfer that will occur if the
                //    codepages are the same and a text transfer has been
                //    specified.

                STAFString currentEOL = connection->readString();
                unsigned int transferLen = connection->readUInt();
                unsigned int bufferLen = connection->readUInt();

                STAFStringBufferPtr convertedNewEOL = newEOL.toCurrentCodePage();
                STAFStringBufferPtr convertedCurrentEOL = currentEOL.toCurrentCodePage();
                int newEOLLength = convertedNewEOL->length();
                int len;
                int offset = 0;
                int endBuffer;
                int strLen = 1;
                int currentEOLLength = convertedCurrentEOL->length();
                int currentEOLOffset = currentEOLLength - 1;
                
                // This is the largest a buffer can grow to during an eol
                // replace.
                unsigned int currentBuffer =
                    (1 + abs(currentEOLLength - newEOLLength)) * transferLen +
                    currentEOLLength + 1;
                
                // This pointer marks the location to start loading characters 
                // into the buffer.
                char * offsetBuffer;
                
                // This is the buffer to store the incoming data        
                char * buffer = new char[transferLen + currentEOLLength + 1];
                
                // This is the buffer to store the data that has had the eol 
                // replaced.
                char * newBuffer = new char[currentBuffer];
                
                offsetBuffer = buffer;

                while (bufferLen != kSTAFFSFinishedCopy)
                {
                    // Update current buffer length
                    strLen += bufferLen;

                    connection->read(offsetBuffer,bufferLen);
                    
                    // Mark the end of the string with a null
                    offsetBuffer[bufferLen] = '\0';

                    replaceStringInBuffer(buffer, convertedCurrentEOL->buffer(), 
                        convertedNewEOL->buffer(), strLen, currentEOLLength, 
                        newEOLLength, &len, &endBuffer, newBuffer);
                    
                    // Get length of the next transfer
                    bufferLen = connection->readUInt();

                    if (bufferLen == kSTAFFSFinishedCopy)
                    {
                        // Last line to be output

                        rc = writeFile(
                            outFile, newBuffer, len - 1, toFile, fromMachine);
                    }
                    else 
                    {
                        // Buffer must roll

                        if (endBuffer != 0xffffffff)
                        {
                            // There is an eol marker in the area to be rolled.
                            // The roll will take place starting at the end of
                            //  the eol

                            endBuffer = endBuffer + newEOLLength;
                            rc = writeFile(
                                outFile, newBuffer, endBuffer,
                                toFile, fromMachine);
                           
                            // offset = buffer length - null char - ending 
                            //   index of eol
                            offset = len - 1 - endBuffer;
                           
                            rollBuffer(buffer, strLen, offset + 1);
                            offsetBuffer = buffer + offset;
                           
                            // Set string length = offset + null
                            strLen = offset + 1;  
                        }
                        else
                        {
                            // Wrap whole buffer
                            offset = currentEOLOffset;
                           
                            // The end of the buffer to write is length of the
                            // buffer - offset - null
                            endBuffer = len - offset - 1;
                           
                            rc = writeFile(
                                outFile, newBuffer, endBuffer,
                                toFile, fromMachine);

                            rollBuffer(buffer, strLen, offset + 1);
                            offsetBuffer = buffer + offset;

                            // Set string length = offset + null
                            strLen = offset + 1;
                        }
                    }
                }

                free(buffer);
                free(newBuffer);
                outFile.close();

                // Indicate whether the copy was successful
                connection->writeUInt(rc);
            }
            else
            {
                // Perform a binary transfer of the file

                unsigned char fileBuffer[4000] = { 0 };
                unsigned int writeLength = 0;

                unsigned int fileLength = connection->readUInt();

                while ((fileLength > 0) && (outFile.good()))
                {
                    rc = connection->readUInt();

                    if (rc == kSTAFOk)
                    {
                        writeLength = STAF_MIN(sizeof(fileBuffer), fileLength);
                        connection->read(fileBuffer, writeLength);

                        rc = writeFile(
                            outFile, reinterpret_cast<char *>(fileBuffer),
                            writeLength, toFile, fromMachine);
                         
                        connection->writeUInt(rc);

                        if (rc != kSTAFOk)
                        {
                            fileLength = 0;
                            break;
                        }

                        fileLength -= writeLength;
                    }
                    else
                    {
                        // File read on other side of connection failed
                        fileLength = 0;
                        break;
                    }
                }

                outFile.close();
            }
        }   
        else
        {
            // Should Never Get Here
        }
    } // end of for loop
}

void handleFileTransfer(unsigned int apiType, unsigned int level,
                        STAFConnPtr connection)
{
    bool textTransfer = false;
    bool doCodepageConvert = false;

    // If the text transfer can be requested it will be done here

    if (apiType == kNewAPI  && level > 1)
    {
        // Get transfer type
        unsigned int transferType = connection->readUInt();
        textTransfer = (transferType != kSTAFFSBinary);
        doCodepageConvert = (transferType == kSTAFFSTextConvert);
    }

    // fromMachine is the machine that the file is coming from
    // orgMachine is the machine that originated the request
    STAFString fromMachine = connection->readString();
    STAFString orgMachine = connection->readString();
    STAFString unresToFile = connection->readString();

    unsigned int flags = connection->readUInt();

    STAFString toFile;
    STAFString errorBuffer;
    STAFString result;
    STAFRC_t rc = kSTAFOk;

    // Only perform the access check when the request did not originate from
    // this machine.

    if (!isLocalMachine(orgMachine))
    {
        if (gTrustManager.trustLevel(fromMachine) < 4)
        {
            rc = kSTAFAccessDenied;
            result = fromMachine;
        }
        else if (gTrustManager.trustLevel(orgMachine) < 4)
        {
            rc = kSTAFAccessDenied;
            result = orgMachine;
        }
    }

    if (rc == kSTAFOk)
    {
        // Note: If the request is to and from the same machine then the
        //       TOFILE string will already be resolved by the FS service,
        //       so we only have to worry about global resolutions for
        //       machines other than ourselves

        rc = resolveString(unresToFile, 0, toFile, errorBuffer);
        if (rc) result = errorBuffer;
    }

    if (rc)
    {
        connection->writeUInt(rc);
        connection->writeString(result);

        return;
    }
    
    // Check if the to file already exists

    STAFFSPath toPath(toFile);

    unsigned int theFileExists = toPath.exists();
    
    if (((flags & 0x00000001) && theFileExists) ||
       ((flags & 0x00000002) && !theFileExists))
    {
        connection->writeUInt(kSTAFFileOpenError);
        connection->writeString(toFile);

        return;
    }

    if (theFileExists)
    {
        // Remove trailing slashes, if any, in the file name
        toFile = toPath.getEntry()->path().asString();
    }
    
    fstream outFile(toFile.toCurrentCodePage()->buffer(), 
                        STAF_ios_binary | ios::out);

    if (!outFile)
    {
        connection->writeUInt(kSTAFFileOpenError);
        connection->writeString(toFile);

        return;
    }

    connection->writeUInt(kSTAFOk);
    connection->writeString(STAFString());

    STAFString newEOL;

    if (textTransfer)
    {
        // Establish the new eol marker.  It causes the use of binary transfer
        // if the eol on both sides are the same and no codepage conversion 
        // is to be done.  This allows for a faster transfer.
        
        // Get EOL chars
        newEOL = connection->readString();
        STAFString tmpEOL = newEOL;

        // Determine new EOL
        if (newEOL.lowerCase() == "native")
        {
            STAFConfigInfo sysInfo;
            STAFString_t errorBuffer;
            unsigned int osRC;

            if (STAFUtilGetConfigInfo(&sysInfo, &errorBuffer, &osRC) != kSTAFOk)
            {    
                // Kill the file transfer
                outFile.close();
                connection->writeUInt(kSTAFFSStopCopy);
                return;
            }
            newEOL = sysInfo.lineSeparator;
        }
        else if (newEOL.lowerCase() == "windows")
        {
            newEOL = STAFString(kUTF8_CR) + STAFString(kUTF8_LF);// \r\n
        }
        else if (newEOL.lowerCase() == "unix")
        {
            newEOL = STAFString(kUTF8_LF);// \n
        }
        else
            newEOL = tmpEOL;

        // Send new eol ok
        connection->writeUInt(kSTAFFSContinueCopy);

        // Send new EOL
        connection->writeString(newEOL);

        textTransfer = (connection->readUInt() == kSTAFFSTextNoConvert);
    }

    if (doCodepageConvert)
    {
        // This clause allows the use of the text noconvert routine if the
        // codepages on both sides are the same.  This allows for a faster
        // transfer.

        STAFString cPage;

        // Get local codepage
        rc = resolveString("{STAF/Config/CodePage}", 0, cPage, errorBuffer);

        // Send this codepage id
        connection->writeString(cPage);
        doCodepageConvert = (connection->readUInt() == kSTAFFSTextConvert);

        // XXX: Uncomment out next line to disable the codepage filter
        //      force the codepage conversion if requested
        //doCodepageConvert = true;
    }    

    /************** Start transferring the file ***************/

    try
    {
        if (doCodepageConvert)
        {
            // Perform a text transfer with conversion of eol chars and a 
            // codepage conversion.  It runs considerably slower than the
            // binary transfer.
            // - This is the type of transfer that will occur if the TEXT 
            //   option is specified and the codepages are different.
            
            STAFString currentEOL = connection->readString();

            STAFString transferStr;
            STAFStringBufferPtr convertedStr;
            int continueCopy = connection->readUInt();

            while (continueCopy != kSTAFFSFinishedCopy)
            {
                transferStr = connection->readString();
                transferStr = transferStr.replace(currentEOL, newEOL);
                convertedStr = transferStr.toCurrentCodePage();

                rc = writeFile(
                    outFile, convertedStr->buffer(), convertedStr->length(),
                    toFile, fromMachine);

                continueCopy = connection->readUInt();
            }

            outFile.close();
        }
        else if (textTransfer)
        {
            // Perform a text transfer with conversion of eol chars without a
            // codepage conversion.  It runs considerably faster than the 
            // codepage conversion transfer.
            // - This is the type of transfer that will occur if the NOCONVERT
            //   option is enabled.
            // - This is the type of transfer that will occur if the codepages 
            //   are the same and a text transfer has been specified.
        
            STAFString currentEOL = connection->readString();
            unsigned int transferLen = connection->readUInt();
            unsigned int bufferLen = connection->readUInt();

            STAFStringBufferPtr convertedNewEOL = newEOL.toCurrentCodePage();
            STAFStringBufferPtr convertedCurrentEOL = currentEOL.
                toCurrentCodePage();
            int newEOLLength = convertedNewEOL->length();
            int len;
            int offset = 0;
            int endBuffer;
            int strLen = 1;
            int currentEOLLength = convertedCurrentEOL->length();
            int currentEOLOffset = currentEOLLength - 1;
        
            // This is the largest a buffer can grow to during an eol replace.
            unsigned int currentBuffer = 
                (1 + abs(currentEOLLength - newEOLLength)) * transferLen +
                currentEOLLength + 1;
        
            // This pointer marks the location to start loading characters
            // into the buffer.
            char * offsetBuffer;
        
            // This is the buffer to store the incoming data 
            char * buffer = new char[transferLen + currentEOLLength + 1];
        
            // This is the buffer to store the data with the EOLs replaced.
            char * newBuffer = new char[currentBuffer];
        
            offsetBuffer = buffer;

            while (bufferLen != kSTAFFSFinishedCopy)
            {
                // Update the current buffer length
                strLen += bufferLen;

                connection->read(offsetBuffer,bufferLen);
            
                // Mark the end of the string with a null
                offsetBuffer[bufferLen] = '\0';

                replaceStringInBuffer(buffer, convertedCurrentEOL->buffer(), 
                    convertedNewEOL->buffer(), strLen, currentEOLLength, 
                    newEOLLength, &len, &endBuffer, newBuffer);

                // Get length of the next transfer
                bufferLen = connection->readUInt();

                if (bufferLen == kSTAFFSFinishedCopy)
                {
                    // Last line to be output
                    rc = writeFile(outFile, newBuffer, len - 1,
                                   toFile, fromMachine);
                }
                else 
                {
                    // Buffer must roll

                    if (endBuffer != 0xffffffff)
                    {
                        // There is an eol marker in the area to be rolled.
                        // The end of the buffer is the starting index of
                        // the last occurence of the new eol + the length of
                        // the new eol.
                   
                        endBuffer = endBuffer + newEOLLength;

                        rc = writeFile(outFile, newBuffer, endBuffer,
                                       toFile, fromMachine);

                        // offset = buffer length - null char - ending index
                        // of eol
                        offset = len - 1 - endBuffer;

                        rollBuffer(buffer, strLen, offset + 1);
                        offsetBuffer = buffer + offset;

                        // Set string length = offset + null
                        strLen = offset + 1;  
                    }
                    else
                    {
                        // Wrap whole buffer
                        offset = currentEOLOffset;

                        // The end of the buffer to write is length of the
                        // buffer - offset - null
                        endBuffer = len - offset - 1;

                        rc = writeFile(outFile, newBuffer, endBuffer,
                                       toFile, fromMachine);

                        rollBuffer(buffer, strLen, offset + 1);
                        offsetBuffer = buffer + offset;

                        // Set string length = offset + null
                        strLen = offset + 1;
                    }
                }
            }

            free(buffer);
            free(newBuffer);
            outFile.close();
        }
        else
        {
            // Binary file copy

            unsigned char fileBuffer[4000] = { 0 };
            unsigned int writeLength = 0;

            if (apiType == kNewAPI)
            {
                unsigned int fileLength = connection->readUInt();

                while (fileLength > 0)
                {
                    writeLength = STAF_MIN(sizeof(fileBuffer), fileLength);

                    connection->read(fileBuffer, writeLength);

                    rc = writeFile(outFile, reinterpret_cast<char *>(fileBuffer),
                              writeLength, toFile, fromMachine);

                    fileLength -= writeLength;
                }

                outFile.close();
            }
            else
            {
                do
                {
                    connection->readUInt(writeLength);

                    if (writeLength != 0)
                    {
                        connection->read(fileBuffer, writeLength);

                        rc = writeFile(
                            outFile, reinterpret_cast<char *>(fileBuffer),
                            writeLength, toFile, fromMachine);

                        if (rc == kSTAFOk)
                        {
                            connection->writeUInt(0);
                        }
                        else
                        {
                            connection->writeUInt(kSTAFFileWriteError);
                            connection->writeString(toFile);
                        }
                    }
                } while ((writeLength != 0) && (outFile.good()));

                outFile.close();
            }
        }

        // Write the final ack indicating if the file copy was successful

        try
        {
            connection->writeUInt(rc);
        }
        catch (STAFConnectionIOException)
        {
            // Older clients will have already closed the connection, so we
            // need to ignore Conneciton IO exceptions at this point
        }

    } /****************** End try file transfer block *******************/
    catch (STAFConnectionIOException)
    {
        // A client has unexpectedly terminated the connection
        // Do nothing, cleanup is automatic

        STAFTrace::trace(
            kSTAFTraceError, "HandleFileTransfer(): Connection terminated "
            "unexpectedly while copying to file " + toFile +
            " from machine " + fromMachine);
    }
    catch (STAFException &se)
    {
        se.write("HandleFileTransfer()");
    }
    catch (...)
    {
        cout << "Caught unknown exception in HandleFileTransfer() "
             << "while copying to file " << toFile
             << " from machine " << fromMachine << endl;
    }
}


STAFRC_t writeFile(fstream &outFile, const char *fileBuffer, unsigned int writeLength,
                   const STAFString toFile, const STAFString fromMachine)
{
    // If an error already occurred writing to the file, don't keep
    // writing to it

    if (!outFile.good())
        return kSTAFFileWriteError;

    // Write a block of data to the file

    unsigned int currentPos = (unsigned int)outFile.tellp();
    outFile.write(fileBuffer, writeLength);

    if (outFile.good())
        return kSTAFOk;  // Write was successful
    
    // The write failed.  Retry write up to 20 times with a delay between
    // each attempt.

    for (int writeAttempt = 1;
         !outFile.good() && writeAttempt <= 20;
         writeAttempt++)
    {
        if (outFile.fail())
        {
            // Recoverable write error

            // Log a warning tracepoint message

            STAFString warningMsg(
                "STAFProc::writeFile() - Write attempt #" +
                STAFString(writeAttempt) + " failed while copying to file " +
                toFile + " from machine " + fromMachine);

            STAFTrace::trace(kSTAFTraceWarning, warningMsg);

            // Delay 1/2 second and retry write after clearing any error
            // flags and repositioning the file pointer

            STAFThreadManager::sleepCurrentThread(500);

            outFile.clear();
            outFile.seekp(currentPos, ios::beg);
            outFile.write(fileBuffer, writeLength);
        }
        else if (outFile.bad())
        {
            // Unrecoverable write error.
            break;
        }
    }

    if (!outFile.good())
    {
        // Note:  If recovery fails, could try closing the file and retrying
        //        one more time

        // Unrecoverable write failure

        return kSTAFFileWriteError;
    }

    return kSTAFOk;
}


void stafProcTerminate()
{
    cout << "STAFProc ending abnormally: terminate() called" << endl;
}


/******************************************************************************/
/* replaceStringInBuffer -  This is a replace routine for preallocated buffers*/
/*    This function replaces all occurences of searchStr in current with      */
/*    replaceStr storing the result in newBuffer.  It updates newLen to the   */
/*    last copied character in newBuffer.  It updates lastReplace to the index*/
/*    of the first character of the last occurence of replaceStr or -1 if no  */
/*    replacements were performed.                                            */
/*                                                                            */
/*    This function may be usable else where and may be moved                 */
/*                                                                            */
/* Accepts: (In)  A pointer to a char buffer of the string to be searched     */
/*          (In)  A pointer to a char buffer of the pattern to match          */
/*          (In)  A pointer to a char buffer of the pattern to replace with   */
/*          (In)  The length of the string to be searched                     */
/*          (In)  The length of the string that is to be matched              */
/*          (In)  The length of the string that is to replace                 */
/*          (Out) The length of the modified string                           */
/*          (Out) The starting index of the last replacement,                 */
/*                -1 if no replacement done                                   */
/*          (Out) A pointer to a char buffer of the modified string           */
/*                !!!! This string must be allocated prior to                 */
/*                     calling the function !!!!                              */
/*                                                                            */
/* Returns: void                                                              */
/******************************************************************************/

void replaceStringInBuffer (char* current, const char* searchStr, const char* 
              replaceStr, int currentLen, int searchStrLen, int replaceStrLen,
              int* newLen, int* lastReplace, char* newBuffer)
{
    *lastReplace = -1;
    *newLen = 0;
    int offset = currentLen - searchStrLen;
    int i, j, last, numEOL;
    last = -1;
    numEOL = 0;
    bool isMatch;

    for(i = 0; i < offset; i++)
    {
        isMatch = true;
    
        for(j = 0; j < searchStrLen; j++)
            isMatch &= (current[j] == searchStr[j]);
        
        if (isMatch)
        {
            last = i;
        
            for(j = 0; j < replaceStrLen; j++)
                newBuffer[j] = replaceStr[j];
            
            newBuffer += replaceStrLen;
            current += searchStrLen;
            i += (searchStrLen-1);
            numEOL++;
        }
        else
        {
            *newBuffer = *current;
            newBuffer++;
            current++;
        }
    }

    *lastReplace = last + ((numEOL - 1) * (replaceStrLen - searchStrLen));

    if ((i == offset) || (last == -1))
        *lastReplace = -1;

    j = searchStrLen - (i - offset);
    
    for (i = 0; i < j; i++)
        newBuffer[i] = current[i];

    *newLen = currentLen + (numEOL * (replaceStrLen - searchStrLen));
    
    return;
}


/******************************************************************************/
/* rollBuffer - This is a buffer rolling routine for preallocated buffers.    */
/*              This function copies the the last (rollLen) characters to the */
/*              front of buffer.                                              */
/*                                                                            */
/* Accepts: (Out)  A pointer to a char buffer of the string to be rolled      */
/*          (In)  The length of the string to be rolled                       */
/*          (In)  The length of the roll                                      */
/*                                                                            */
/* Returns: void                                                              */
/******************************************************************************/

void rollBuffer(char * buffer, int bufferLen, int rollLen)
{
    int endIndx = bufferLen - rollLen;
    char * end = buffer + endIndx;
    int i;

    for (i = 0; i < rollLen; i++)
        buffer[i] = end[i];

    return;
}
