/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFString.h"
#include "STAFProcUtil.h"
#include "STAFUtil.h"
#include "STAFException.h"
#include "STAFVariablePool.h"
#include "STAFProc.h"
#include "STAFConnection.h"
#include "STAFConnectionManager.h"
#include "STAFHandleManager.h"
#include "STAFCommandParser.h"

STAFRC_t resolveUIntIfExists(STAFCommandParseResultPtr &parseResult,
    const STAFString &optionName, STAFVariablePool *primaryPool,
    unsigned int &result, STAFString &errorBuffer, unsigned int defaultValue)
{
    if (parseResult->optionTimes(optionName) == 0)
        return kSTAFOk;

    return resolveUInt(parseResult->optionValue(optionName), primaryPool,
                       result, errorBuffer, defaultValue);
}


STAFRC_t resolveUInt(const STAFString &unresolved,
     STAFVariablePool *primaryPool, unsigned int &result,
     STAFString &errorBuffer, unsigned int defaultValue)
{
    STAFString resultString;
    STAFRC_t rc = kSTAFOk;

    if (primaryPool != 0)
    {
        rc = primaryPool->resolve(unresolved, *gGlobalVariablePoolPtr,
                                  resultString);
    }
    else
    {
        rc = (*gGlobalVariablePoolPtr)->resolve(unresolved, primaryPool,
                                                resultString);
    }

    if (rc != kSTAFOk)
    {
        errorBuffer = unresolved;
        return rc;
    }

    if (resultString.length() == 0)
    {
        result = defaultValue;
    }
    else if (!resultString.isDigits())
    {
        errorBuffer = resultString;
        return kSTAFInvalidRequestString;
    }
    else result = (unsigned int)resultString.asUInt();

    return rc;
}


STAFRC_t resolveStringIfExists(STAFCommandParseResultPtr &parseResult,
    const STAFString &optionName, STAFVariablePool *primaryPool,
    STAFString &result, STAFString &errorBuffer)
{
    if (parseResult->optionTimes(optionName) == 0)
        return kSTAFOk;

    return resolveString(parseResult->optionValue(optionName), primaryPool,
                         result, errorBuffer);
}


STAFRC_t resolveString(const STAFString &unresolved,
                            STAFVariablePool *primaryPool,
                            STAFString &result, STAFString &errorBuffer)
{
    STAFRC_t rc = kSTAFOk;

    if (primaryPool != 0)
    {
        rc = primaryPool->resolve(unresolved, *gGlobalVariablePoolPtr, result);
    }
    else
    {
        rc = (*gGlobalVariablePoolPtr)->resolve(unresolved, primaryPool,
                                                result);
    }

    if (rc != kSTAFOk) errorBuffer = unresolved;

    return rc;
}


STAFString effectiveMachineName(const STAFString &machine)
{
    if ((gUseLongNames == 0) && (machine.find(kUTF8_PERIOD) != 
                                 STAFString::kNPos))
    {
        return machine.subString(0, machine.find(kUTF8_PERIOD));
    }

    return machine;
}


STAFVariablePoolPtr getClientVariablePool(const STAFString &client,
                                          STAFHandle_t handle)
{
    STAFVariablePoolPtr thePool;

    if (client.toLowerCase() == gMachineNetNamePtr->toLowerCase())
        gHandleManagerPtr->variablePool(handle, thePool);

    return thePool;
}


bool isLocalMachine(const STAFString &machine, unsigned int fullCompare)
{
    static STAFString sLocal = "local";

    if (fullCompare)
    {
        if (machine.isEqualTo(sLocal, STAFString::kCaseInsensitive) ||
            machine.isEqualTo(*gMachineNetNamePtr,
                              STAFString::kCaseInsensitive) ||
            machine.isEqualTo(*gMachineNamePtr, STAFString::kCaseInsensitive)) 
        {
            return true;
        }
        else return false;
    }

    return machine.isEqualTo(*gMachineNetNamePtr, STAFString::kCaseInsensitive);
}

