/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include <deque>
#include "STAF_iostream.h"
#include "STAF_fstream.h"
#include "STAFString.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFException.h"
#include "STAFProcessService.h"
#include "STAFProcess.h"
#include "STAFHandleManager.h"
#include "STAFTrustManager.h"
#include "STAFVariablePool.h"
#include "STAFUtil.h"
#include "STAFThreadManager.h"
#include "STAFTrace.h"
#include "STAFInternalProcess.h"
#include "STAFDiagManager.h"

STAFProcessAuthenticationMode_t
    STAFProcessService::fAuthMode = kSTAFProcessAuthDisabled;
STAFProcessStopMethod_t
    STAFProcessService::fDefaultStopMethod = kSTAFProcessStopWithSigKill;
STAFProcessDisabledAuthAction_t
    STAFProcessService::fAuthAction = kSTAFProcessDisabledAuthIgnore;
STAFString STAFProcessService::fDefaultAuthUsername;
STAFString STAFProcessService::fDefaultAuthPassword;
STAFString STAFProcessService::fDefaultShellCommand;
STAFString STAFProcessService::fDefaultSameConsoleShell;
STAFString STAFProcessService::fDefaultNewConsoleShell;
STAFEventSem STAFProcessService::fNotifySem;
STAFProcessService::ProcessPtr STAFProcessService::fNotifyProcess;
unsigned int STAFProcessService::fEnvVarCaseSensitive = 1;

// Note: This is a mild hack to allow the callback routine to work.
// In a perfect world, this class (and all the Service classes) would
// be Singletons and I would create a more generic callback mechanism.

static STAFProcessService *processService = 0;

STAFServiceResult readFileIntoString(const STAFString &file)
{
    fstream inFile(file.toCurrentCodePage()->buffer(),
                   ios::in | STAF_ios_binary);

    if (!inFile) return STAFServiceResult(kSTAFFileOpenError, file);

    // Figure out how big the file is

    inFile.seekg(0, ios::end);
    unsigned int fileLength = (unsigned int)inFile.tellg();

    // Initialize the output buffer and read in the file

    STAFRefPtr<char> buffer(new char[fileLength], STAFRefPtr<char>::INIT, 
                            STAFRefPtr<char>::ARRAY);

    inFile.seekg(0, ios::beg);
    inFile.read(buffer, fileLength);

    return STAFServiceResult(kSTAFOk, STAFString(buffer, fileLength));
}


STAFServiceResult deleteTempFile(const STAFString &tempFileName)
{
    STAFRC_t rc = kSTAFBaseOSError;
    unsigned int osRC = 0;
    STAFString errorBuffer = STAFString("");
    unsigned int retryAttempts = 10;

    // Retry to delete the temporary file up to 10 times until successful
    // because sometimes the file is still in use (OSRC=32 on Windows)
    for (unsigned int i = 1; (rc != kSTAFOk) && (i <= retryAttempts); ++i)
    {
        try
        {
            rc = STAFFSPath(tempFileName).getEntry()->remove(&osRC);
            
            if (rc != kSTAFOk)
            {
                if (i == retryAttempts)
                {
                    errorBuffer =
                        STAFString("STAFProcessService::deleteTempFile()"
                                   " - Failed for " + tempFileName +
                                   " with OSRC=" + osRC);

                    STAFTrace::trace(kSTAFTraceWarning, errorBuffer);
                }
                else
                {
                    // Sleep a short time before retrying to delete the file
                    STAFThreadManager::sleepCurrentThread(500);
                }
            }
        }
        catch (STAFException &se)
        {
            // Don't retry deleting the temp file
            rc = kSTAFBaseOSError;
            errorBuffer = STAFString("STAFProcessService::deleteTempFile()"
                                     " - Failed for " + tempFileName +
                                     " Text: " + se.getText());
            STAFTrace::trace(kSTAFTraceWarning, errorBuffer);
        }
        catch (...)
        {
            // Don't retry deleting the temp file
            rc = kSTAFBaseOSError;
            errorBuffer = STAFString("STAFProcessService::deleteTempFile()"
                                     " - Caught unknown exception deleting " +
                                     tempFileName);
            STAFTrace::trace(kSTAFTraceWarning, errorBuffer);
        }
    }
    
    return STAFServiceResult(rc, errorBuffer);
}


STAFProcessService::STAFProcessService() : STAFService("PROCESS")
{
    processService = this;

    // START options
 
    fStartParser.addOption("START",             1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("WORKLOAD",          1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("COMMAND",           1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("PARMS",             1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("SHELL",             1,
        STAFCommandParser::kValueAllowed);
    fStartParser.addOption("WORKDIR",           1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("TITLE",             1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("WAIT",              1,
        STAFCommandParser::kValueAllowed);
    fStartParser.addOption("ASYNC",             1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("VAR",               0,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("ENV",               0,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("USEPROCESSVARS",    0,
        STAFCommandParser::kValueNotAllowed);
    // fStartParser.addOption("FOREGROUND",     1,
    //    STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("NOTIFY",            1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("ONEND",             1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("NAME",              1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("HANDLE",            1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("MACHINE",           1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("PRIORITY",          1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("KEY",          1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STOPUSING",         1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("NEWCONSOLE",        1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("SAMECONSOLE",       1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("USERNAME",          1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("PASSWORD",          1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STDIN",             1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STDOUT",            1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STDERR",            1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STDOUTAPPEND",      1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STDERRAPPEND",      1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("STDERRTOSTDOUT", 0,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("DISABLEDAUTHISERROR", 0,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("IGNOREDISABLEDAUTH",  0,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("STATICHANDLENAME",  1,
        STAFCommandParser::kValueRequired);
    fStartParser.addOption("RETURNSTDOUT",      1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("RETURNSTDERR",      1,
        STAFCommandParser::kValueNotAllowed);
    fStartParser.addOption("RETURNFILE",        0,
        STAFCommandParser::kValueRequired);
 
    // START option groups
 
    fStartParser.addOptionGroup("COMMAND", 1, 1);
    fStartParser.addOptionGroup("WAIT ASYNC",  0, 1);
    fStartParser.addOptionGroup("HANDLE NAME", 0, 1);
    fStartParser.addOptionGroup("NEWCONSOLE SAMECONSOLE", 0, 1);
    fStartParser.addOptionGroup("DISABLEDAUTHISERROR IGNOREDISABLEDAUTH", 0, 1);
    fStartParser.addOptionGroup("STDOUT STDOUTAPPEND", 0, 1);
    fStartParser.addOptionGroup("STDERR STDERRAPPEND STDERRTOSTDOUT", 0, 1);

    // START option needs

    fStartParser.addOptionNeed("NOTIFY", "ONEND");
    fStartParser.addOptionNeed("PASSWORD", "USERNAME");
    fStartParser.addOptionNeed("NAME HANDLE MACHINE PRIORITY KEY", "NOTIFY");
    fStartParser.addOptionNeed("STDERRTOSTDOUT", "STDOUT STDOUTAPPEND RETURNSTDOUT");

    // QUERY options

    fQueryParser.addOption("QUERY",    1,
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("FREE",    1,
        STAFCommandParser::kValueNotAllowed);
    fQueryParser.addOption("WORKLOAD", 1,
        STAFCommandParser::kValueRequired);
    fQueryParser.addOption("HANDLE",   1,
        STAFCommandParser::kValueRequired);
    fQueryParser.addOption("ALL",      1,
        STAFCommandParser::kValueNotAllowed);

    // QUERY groups

    fQueryParser.addOptionGroup("QUERY FREE",          1, 1);
    fQueryParser.addOptionGroup("WORKLOAD HANDLE ALL", 1, 1);

    // STOP options

    fStopParser.addOption("STOP",     1,
        STAFCommandParser::kValueNotAllowed);
    fStopParser.addOption("WORKLOAD", 1,
        STAFCommandParser::kValueRequired);
    fStopParser.addOption("HANDLE",   1,
        STAFCommandParser::kValueRequired);
    fStopParser.addOption("ALL",      1,
        STAFCommandParser::kValueNotAllowed);
    fStopParser.addOption("CONFIRM",  1,
        STAFCommandParser::kValueNotAllowed);
    fStopParser.addOption("USING",    1,
        STAFCommandParser::kValueRequired);

    // STOP needs

    fStopParser.addOptionNeed("ALL", "CONFIRM");

    // NOTIFY registration options

    fNotifyRegistrationParser.addOption("NOTIFY",        1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyRegistrationParser.addOption("REGISTER",      1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyRegistrationParser.addOption("UNREGISTER",    1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyRegistrationParser.addOption("ONENDOFHANDLE", 1,
        STAFCommandParser::kValueRequired);
    fNotifyRegistrationParser.addOption("NAME",          1,
        STAFCommandParser::kValueRequired);
    fNotifyRegistrationParser.addOption("HANDLE",        1,
        STAFCommandParser::kValueRequired);
    fNotifyRegistrationParser.addOption("MACHINE",       1,
        STAFCommandParser::kValueRequired);
    fNotifyRegistrationParser.addOption("PRIORITY",      1,
        STAFCommandParser::kValueRequired);

    // NOTIFY registration groups
 
    fNotifyRegistrationParser.addOptionGroup("ONENDOFHANDLE",       1, 1);
    fNotifyRegistrationParser.addOptionGroup("HANDLE NAME",         0, 1);
 
    // NOTIFY LIST options
 
    fNotifyListParser.addOption("NOTIFY",        1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyListParser.addOption("LIST",          1,
        STAFCommandParser::kValueNotAllowed);
    fNotifyListParser.addOption("ONENDOFHANDLE", 1,
        STAFCommandParser::kValueRequired);
 
    // NOTIFY LIST groups

    fNotifyListParser.addOptionGroup("ONENDOFHANDLE", 1, 1);
    
    // Set the default stop method and the environment variable case
    // sensitive flag
    
    STAFConfigInfo configInfo;
    STAFString_t errorBufferT;
    unsigned int osRC = 0;

    if (STAFUtilGetConfigInfo(&configInfo, &errorBufferT, &osRC) != kSTAFOk)
    {
        cout << STAFString(errorBufferT) << ", RC: " << osRC << endl;
    }
    else
    {
        STAFProcessService::fDefaultStopMethod = 
            configInfo.defaultProcessStopMethod;
        STAFProcessService::fEnvVarCaseSensitive =
            configInfo.envVarCaseSensitive;
    }
}


STAFProcessService::~STAFProcessService()
{
    /* Do Nothing */
}


STAFString STAFProcessService::info(unsigned int) const
{
    return name() + ": Internal";
}


void STAFProcessService::processTerminationCallback(
         STAFProcessID_t pid, STAFProcessHandle_t procHandle,
         unsigned int retCode, void *)
{
    processService->handleProcessTermination(pid, retCode);
}


unsigned int STAFProcessService::sendNotificationCallback(void *)
{
    try
    {
        processService->sendNotification();
    }
    catch (STAFException &se)
    { se.write("STAFProcessService::sendNotificationCallback()"); }
    catch (...)
    {
        cout << "Caught unknown exception in "
             << "STAFProcessService::sendNotificationCallback()" << endl;
    }

    return 0;
}


STAFRC_t STAFProcessService::getStopMethodFromString(
    STAFProcessStopMethod_t &stopMethod,
    const STAFString &methodString)
{
    static STAFString sSigKill("SIGKILL");
    static STAFString sSigKillAll("SIGKILLALL");
    static STAFString sSigTerm("SIGTERM");
    static STAFString sSigInt("SIGINT");
    static STAFString sWM_CLOSE("WM_CLOSE");
    STAFString upperMode = methodString.toUpperCase();

    if (upperMode == sSigKill)
        stopMethod = kSTAFProcessStopWithSigKill;
    else if (upperMode == sSigKillAll)
        stopMethod = kSTAFProcessStopWithSigKillAll;
    else if (upperMode == sSigTerm)
        stopMethod = kSTAFProcessStopWithSigTerm;
    else if (upperMode == sSigInt)
        stopMethod = kSTAFProcessStopWithSigInt;
    else if (upperMode == sWM_CLOSE)
        stopMethod = kSTAFProcessStopWithWM_CLOSE;
    else return kSTAFInvalidRequestString;

    return kSTAFOk;
}


STAFRC_t STAFProcessService::setDefaultStopMethod(
    STAFProcessStopMethod_t stopMethod)
{
    if (STAFProcess::isValidStopMethod(stopMethod) != kSTAFOk)
        return kSTAFInvalidValue;

    fDefaultStopMethod = stopMethod;

    return kSTAFOk;
}


STAFProcessStopMethod_t STAFProcessService::getDefaultStopMethod()
{
    return fDefaultStopMethod;
}


STAFRC_t STAFProcessService::getAuthModeFromString(
    STAFProcessAuthenticationMode_t &authMode, const STAFString &modeString)
{
    static STAFString sDisabled("DISABLED");
    static STAFString sNone("NONE");
    static STAFString sWindows("WINDOWS");
    STAFString upperMode = modeString.toUpperCase();

    if      (upperMode == sDisabled) authMode = kSTAFProcessAuthDisabled;
    else if (upperMode == sNone)     authMode = kSTAFProcessAuthNone;
    else if (upperMode == sWindows)  authMode = kSTAFProcessAuthWindows;
    else return kSTAFInvalidRequestString;

    return kSTAFOk;
}


STAFRC_t STAFProcessService::setAuthMode(
    STAFProcessAuthenticationMode_t authMode, unsigned int &osRC)
{
    if (STAFProcess::isValidAuthMode(authMode) != kSTAFOk)
        return kSTAFInvalidValue;

    fAuthMode = authMode;

    return kSTAFOk;
}


STAFProcessAuthenticationMode_t STAFProcessService::getAuthMode()
{
    return fAuthMode;
}


STAFRC_t STAFProcessService::getDefaultDisabledAuthActionFromString(
    STAFProcessDisabledAuthAction_t &authAction, const STAFString &actionString)
{
    static STAFString sIgnore("IGNORE");
    static STAFString sError("ERROR");
 
    STAFString upperAction = actionString.toUpperCase();
 
    if (upperAction == sIgnore) authAction = kSTAFProcessDisabledAuthIgnore;
    else if (upperAction == sError) authAction = kSTAFProcessDisabledAuthError;
    else return kSTAFInvalidRequestString;

    return kSTAFOk;
}
 
 
STAFRC_t STAFProcessService::setDefaultDisabledAuthAction(
    STAFProcessDisabledAuthAction_t authAction)
{
    fAuthAction = authAction;
    return kSTAFOk;
}
                                                                               

STAFProcessDisabledAuthAction_t
STAFProcessService::getDefaultDisabledAuthAction()
{
    return fAuthAction;
}


STAFRC_t STAFProcessService::setDefaultAuthUsername(const STAFString &username)
{
    fDefaultAuthUsername = username;
    return kSTAFOk;
}

                                                                               
STAFString STAFProcessService::getDefaultAuthUsername()
{
    return fDefaultAuthUsername;
}


STAFRC_t STAFProcessService::setDefaultAuthPassword(const STAFString &password)
{
    fDefaultAuthPassword = password;
    return kSTAFOk;
}

                                                                              
STAFString STAFProcessService::getDefaultAuthPassword()
{
    return fDefaultAuthPassword;
}


STAFRC_t isShellCommandValid(const STAFString &shellCommand)
{
    return STAFProcessValidateShellSubstitutionChars(shellCommand);
}


STAFRC_t STAFProcessService::setDefaultShellCommand(
    const STAFString &shellCommand)
{
    if (isShellCommandValid(shellCommand) != kSTAFOk)
        return kSTAFInvalidValue;

    fDefaultShellCommand = shellCommand;

    return kSTAFOk;
}


STAFString STAFProcessService::getDefaultShellCommand()
{
    return fDefaultShellCommand;
}


STAFRC_t STAFProcessService::setDefaultSameConsoleShell(
    const STAFString &shellCommand)
{
    if (isShellCommandValid(shellCommand) != kSTAFOk)
        return kSTAFInvalidValue;

    fDefaultSameConsoleShell = shellCommand;

    return kSTAFOk;
}


STAFString STAFProcessService::getDefaultSameConsoleShell()
{
    return fDefaultSameConsoleShell;
}


STAFRC_t STAFProcessService::setDefaultNewConsoleShell(
    const STAFString &shellCommand)
{
    if (isShellCommandValid(shellCommand) != kSTAFOk)
        return kSTAFInvalidValue;

    fDefaultNewConsoleShell = shellCommand;

    return kSTAFOk;
}


STAFString STAFProcessService::getDefaultNewConsoleShell()
{
    return fDefaultNewConsoleShell;
}


void STAFProcessService::handleProcessTermination(STAFProcessID_t pid,
                                                  unsigned int retCode)
{
    try
    {
        ProcessList::iterator iter;

        STAFMutexSemLock processLock(fProcessListSem);

        for (iter = fProcessList.begin(); (iter != fProcessList.end()) &&
             (iter->second->pid != pid) ||
             (iter->second->state != kRunning); iter++)
        { /* Do Nothing */ }

        if (iter == fProcessList.end())
        {
            STAFTrace::trace(kSTAFTraceError, "PID was not in the list in "
                             "handleProcessTermination()");
            return;
        }

        ProcessPtr process = iter->second;

        process->RC = retCode;
        process->endStamp = STAFTimestamp::now();
        process->state = kComplete;

        gHandleManagerPtr->unRegister(process->handle, process->pid);

        if (process->notify) process->notify->post();

        // Send out the notification list

        fNotifySem.reset();
        fNotifyProcess = process;

        gThreadManagerPtr->dispatch(sendNotificationCallback, 0);

        fNotifySem.wait();
        fNotifyProcess = ProcessPtr();

    }
    catch (STAFException &se)
    { se.write("STAFProcessService::handleProcessTermination()"); }
    catch (...)
    {
        cout << "Caught unknown exception in "
             << "STAFProcessService::handleProcessTermination()" << endl;
    }
}

void STAFProcessService::sendNotification()
{
    ProcessPtr process(fNotifyProcess);

    fNotifySem.post();

    STAFString message("STAF/PROCESS/END " + STAFString(process->handle) + 
                       ";" + process->endStamp.asString() + ";" + 
                       STAFString(process->RC));

    if (process->key.length() != 0)
    {
        message += ";";
        message += "KEY=";
        message += process->key;
    }

    if (process->retFileList.size() != 0)
    {
        message += ";";
        message += STAFString(process->retFileList.size());
        message += ";";

        for (STAFProcessService::Process::FileList::iterator iter =
                 process->retFileList.begin();
             iter != process->retFileList.end();
             ++iter)
        {            
            STAFServiceResult result = readFileIntoString(*iter);

            message += STAFString(result.fServicePerformedRC);
            message += ":";
            message += STAFString(result.fResult.length(STAFString::kChar));
            message += ":";
            message += STAFString(result.fResult);
        }
    }

    process->notificationList->sendNotification(message);

    // Delete the temporary StdOut/StdErr files if used and available for
    // deletion.
    deleteTempFiles(*process);
}


STAFServiceResult STAFProcessService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    // select word 1 as the action and the rest of the request as rest
 
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if      (action == "start") return handleStart(requestInfo);
    else if (action == "stop")  return handleStop(requestInfo);
    else if (action == "query") return handleQuery(requestInfo);
    else if (action == "free")  return handleFree(requestInfo);
    else if (action == "notify")
    {
        STAFString subAction = requestInfo.fRequest.subWord(1, 1).lowerCase();

        if (subAction == "register")
            return handleNotifyRegistration(1, requestInfo);
        else if (subAction == "unregister")
            return handleNotifyRegistration(0, requestInfo);
        else if (subAction == "list")
            return handleNotifyList(requestInfo);
        else return STAFServiceResult(kSTAFInvalidRequestString);
    }
    else if (action == "help") return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString);
}


STAFServiceResult STAFProcessService::handleStart(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fStartParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString errorBuffer;                  

    ProcessSync startMode = kAsync;

    if (parsedResult->optionTimes("WAIT") != 0)
    {
        startMode = kWait;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the output for a START request
            // with the WAIT option specified will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/PROCESS START WAIT", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }
    }

    STAFProcessStartInfoLevel1 startData = { 0 };
    ProcessPtr process(new Process(), ProcessPtr::INIT);
    unsigned int doNotify = parsedResult->optionTimes("NOTIFY");
    unsigned int useName = parsedResult->optionTimes("NAME");
    unsigned int priority = 5;
    STAFHandle_t theHandle = requestInfo.fHandle;
    STAFString machine = requestInfo.fMachine;
    STAFString name;
    STAFString password = "";
    STAFString stdInp = "";
    STAFString stdOut = "";
    STAFString stdErr = "";
    STAFString stopMethodString;
    unsigned int timeout = 0;
    STAFString staticHandleName;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);

    if (parsedResult->optionTimes("STATICHANDLENAME") != 0)
        process->handleType = kStatic;
    else
        process->handleType = kPending;

    if (parsedResult->optionTimes("SHELL") != 0)
        process->processType = kShell;
    else
        process->processType = kCommand;

    // Set up variable pool

    STAFVariablePoolPtr varPool(new STAFVariablePool,
                                STAFVariablePoolPtr::INIT);

    if (parsedResult->optionTimes("VAR") != 0)
    {
        for (int i = 1; i <= parsedResult->optionTimes("VAR"); ++i)
        {
            STAFString nameAndValue = parsedResult->optionValue("VAR", i);
            unsigned int equalPos = nameAndValue.find(kUTF8_EQUAL);

            if (equalPos == STAFString::kNPos)
            {
                return STAFServiceResult(kSTAFInvalidRequestString,
                                         nameAndValue);
            }

            varPool->set(nameAndValue.subString(0, equalPos),
                         nameAndValue.subString(equalPos +
                         nameAndValue.sizeOfChar(equalPos)));
        }
    }

    // If specified, set the primary variable resolution pool to this
    // new process' pool.

    if (parsedResult->optionTimes("USEPROCESSVARS") != 0)
        handlePool = varPool;

    // Resolve the rest of the options

    STAFRC_t rc = resolveStringIfExists(parsedResult, "WORKLOAD", 
                       handlePool, process->workload, errorBuffer);

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "SHELL", handlePool,
                                   process->shellCommand, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "COMMAND", handlePool,
                                   process->command, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "WORKDIR", handlePool,
                                   process->workdir, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "PARMS", handlePool, 
                                   process->parms, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "TITLE", handlePool, 
                                   process->title, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveUIntIfExists(parsedResult, "PRIORITY", handlePool, 
                                 priority, errorBuffer);
    }
    
    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "KEY", handlePool, 
                                   process->key, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool, 
                                 theHandle, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "NAME", handlePool,
                                   name, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "MACHINE", handlePool, 
                                   machine, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveUIntIfExists(parsedResult, "WAIT", handlePool,
                                 timeout, errorBuffer, (unsigned int)-1);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "STOPUSING", handlePool,
                                   stopMethodString, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "USERNAME", handlePool, 
                                   process->username, errorBuffer);
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "PASSWORD", handlePool, 
                                   password, errorBuffer);
    }

    if (!rc)
    {
        if (parsedResult->optionTimes("STDIN") != 0)
        {
            startData.stdinMode = kSTAFProcessIOReadFile;
            rc = resolveStringIfExists(parsedResult, "STDIN", handlePool, 
                                       stdInp, errorBuffer);
        }
    }

    if (!rc)
    {
        if (parsedResult->optionTimes("STDOUT") != 0)
        {
            startData.stdoutMode = kSTAFProcessIOReplaceFile;
            rc = resolveStringIfExists(parsedResult, "STDOUT", handlePool, 
                                       stdOut, errorBuffer);
        }
        else if (parsedResult->optionTimes("STDOUTAPPEND") != 0)
        {
            startData.stdoutMode = kSTAFProcessIOAppendFile;
            rc = resolveStringIfExists(parsedResult, "STDOUTAPPEND", 
                                       handlePool, stdOut, errorBuffer);
        }
    }

    if (!rc)
    {
        if (parsedResult->optionTimes("STDERR") != 0)
        {
            startData.stderrMode = kSTAFProcessIOReplaceFile;
            rc = resolveStringIfExists(parsedResult, "STDERR", handlePool, 
                                       stdErr, errorBuffer);
        }
        else if (parsedResult->optionTimes("STDERRAPPEND") != 0)
        {
            startData.stderrMode = kSTAFProcessIOAppendFile;
            rc = resolveStringIfExists(parsedResult, "STDERRAPPEND", 
                                       handlePool, stdErr, errorBuffer);
        }
    }

    if (!rc)
    {
        // Make sure files specified for stdout and stderr are not the same
        if ((stdOut.length() != 0) && (stdErr.length() != 0) &&
            (stdOut == stdErr))
        {
            // XXX: Would be nice to have an API to call that checks if the
            // two files are the same in a platform independent way and that
            // would check for other variations like / vs \ and \ vs \\. For 
            // now, do a case-insensitive check in Win32/STAFProcess.
            return STAFServiceResult(kSTAFInvalidValue, stdErr, 0);
        }
    }

    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "STATICHANDLENAME",
                                   handlePool, staticHandleName, errorBuffer);
    }
    
    if (!rc && (parsedResult->optionTimes("RETURNSTDOUT") != 0))
    {
        if (stdOut.length() == 0)
        {
            // STDOUT name not specified so create a temp file.
            // Use Replace File mode by default.

            startData.stdoutMode = kSTAFProcessIOReplaceFile;

            // Get a temporary file name

            STAFString_t tempFileNameT;
            STAFString_t errorBufferT;
            unsigned int osRC = 0;

            rc = STAFUtilCreateTempFile(&tempFileNameT, &errorBufferT, &osRC);

            if (rc)
            {
                return STAFServiceResult(rc, STAFString(errorBufferT) +
                                         " for STDOUT, OSRC: " + osRC);
            }
            
            stdOut = STAFString(tempFileNameT, STAFString::kShallow);

            TempFileInfo tempFileInfo;
            tempFileInfo.name = stdOut;

            if (startMode == kWait)
            {
                // Delete the temp file after waiting for the process to end
                tempFileInfo.deleteFlag = false;
            }
            else
            {
                // Delete the temp file after the queue has been updated
                tempFileInfo.deleteFlag = true;
            }

            process->stdoutTempFileInfo = tempFileInfo;
        }

        process->retFileList.push_back(stdOut);
    }

    if (!rc && (parsedResult->optionTimes("RETURNSTDERR") != 0))
    {
        if (stdErr.length() == 0)
        {
            // STDERR name not specified so create a temp file.
            // Use Replace File mode by default.

            startData.stderrMode = kSTAFProcessIOReplaceFile;

            // Get a temporary file name

            STAFString_t tempFileNameT;
            STAFString_t errorBufferT;
            unsigned int osRC = 0;

            rc = STAFUtilCreateTempFile(&tempFileNameT, &errorBufferT, &osRC);

            if (rc)
            {
                // Delete temp file created for Stdout before returning
                if (process->stdoutTempFileInfo.name.length() != 0)
                    deleteTempFile(process->stdoutTempFileInfo.name);

                return STAFServiceResult(rc, STAFString(errorBufferT) +
                                         " for STDERR, OSRC: " + osRC);
            }
            
            stdErr = STAFString(tempFileNameT, STAFString::kShallow);

            TempFileInfo tempFileInfo;
            tempFileInfo.name = stdErr;

            if (startMode == kWait)
            {
                // Delete the temp file after waiting for the process to end
                tempFileInfo.deleteFlag = false;
            }
            else
            {
                // Delete the temp file after the queue has been updated
                tempFileInfo.deleteFlag = true;
            }

            process->stderrTempFileInfo = tempFileInfo;
        }

        process->retFileList.push_back(stdErr);
    }

    if (parsedResult->optionTimes("STDERRTOSTDOUT") != 0)
        startData.stderrMode = kSTAFProcessIOStdout;

    for (int fileIndex = 1;
         !rc && (fileIndex <= parsedResult->optionTimes("RETURNFILE"));
         ++fileIndex)
    {
        STAFString unresReturnFile =
                   parsedResult->optionValue("RETURNFILE", fileIndex);
        STAFString returnFile;

        rc = resolveString(unresReturnFile, handlePool, returnFile,
                           errorBuffer);

        if (!rc) process->retFileList.push_back(returnFile);
    }

    if (!rc && (process->shellCommand.length() != 0))
    {
        if (isShellCommandValid(process->shellCommand) != kSTAFOk)
        {
            rc = kSTAFInvalidValue;
            errorBuffer = process->shellCommand;
        }
    }

    if (!rc)
    {
        if (stopMethodString.length() > 0)
        {
            rc = getStopMethodFromString(process->stopMethod, stopMethodString);

            if (rc)
            {
                rc = kSTAFInvalidRequestString;
                errorBuffer = stopMethodString;
            }
        }
        else
        {
            process->stopMethod = STAFProcessService::getDefaultStopMethod();
        }
    }

    if (rc)
    {
        // Delete the temporary StdOut/StdErr files if used before returning
        deleteTempFiles(*process, true);

        return STAFServiceResult(rc, errorBuffer);
    }

    process->startStamp = STAFTimestamp::now();
    process->state = kRunning;
    process->notificationList = STAFNotificationListPtr(
                                new STAFNotificationList(), 
                                STAFNotificationListPtr::INIT);
    if (doNotify && useName)
        process->notificationList->reg(machine, name, priority);
    else if (doNotify)
        process->notificationList->reg(machine, theHandle, priority);

    if (startMode == kWait)
        process->notify = STAFEventSemPtr(new STAFEventSem(), 
                                          STAFEventSemPtr::INIT);

    std::deque<STAFString> userEnvList;
    STAFRefPtr<STAFStringConst_t> userEnvArray;
    char *envBuf = 0;

    // Note:  Due to a MS compiler bug, had to reverse the order of the
    // conditions in the following if.  The static check had to be first.
    if ((process->handleType == kStatic) ||
        (parsedResult->optionTimes("ENV") != 0))
    {
        // Create a STAFProcessEnvMap from the global environment map so that
        // the key (the environment variable name) can be case-insensitive
        // if designated by the operating system.

        STAFProcessEnvMap envMap;

        for (STAFEnvMap::iterator emIter = gEnvMapPtr->begin();
             emIter != gEnvMapPtr->end(); ++emIter)
        {
            STAFString envName = emIter->first;
            STAFString envNameKey = envName;
            
            if (!fEnvVarCaseSensitive)
                envNameKey.upperCase();

            STAFString envValue = emIter->second;

            envMap[envNameKey] = STAFProcessEnvData(envName, envValue);
        }
        
        // Add all the user specified variables to the envMap

        for (int i = 1; i <= parsedResult->optionTimes("ENV"); ++i)
        {
            STAFString unresEnvVar = parsedResult->optionValue("ENV", i);
            STAFString envVar;

            rc = resolveString(unresEnvVar, handlePool, envVar, errorBuffer);

            if (rc)
            {
                // Delete the temporary StdOut/StdErr files if used
                deleteTempFiles(*process, true);

                return STAFServiceResult(rc, errorBuffer);
            }

            unsigned int equalPos = envVar.find(kUTF8_EQUAL);
            STAFString envName(envVar.subString(0, equalPos));
            STAFString envNameKey = envName;
            STAFString envValue;

            if (equalPos != STAFString::kNPos)
                envValue = envVar.subString(equalPos + 1);

            if (!fEnvVarCaseSensitive)
            {
                envNameKey.upperCase();
            }
            
            // Retain original env variable name if already exists
            if (envMap.find(envNameKey) != envMap.end())
                envMap[envNameKey].envValue = envValue;
            else
                envMap[envNameKey] = STAFProcessEnvData(envName, envValue);

            // Store only the user specified variables here for use by
            // CreateProcessAsUser() called in Win32's STAFProcess.
            userEnvList.push_back(envVar);
        }

        if (process->handleType == kStatic)
        {
            // Allocate the handle and set the environment variable
            
            gHandleManagerPtr->addAndGetStaticHandle(process->handle,
                                                     staticHandleName,
                                                     varPool);

            envMap["STAF_STATIC_HANDLE"] = STAFProcessEnvData(
                "STAF_STATIC_HANDLE", process->handle);

            userEnvList.push_back(STAFString("STAF_STATIC_HANDLE=") +
                                  process->handle);
        }

        // Now, walk through the list and combine the entries back into
        // Name=Value form, get the current code page representation and 
        // figure out how big a buffer we need

        int size = 0;
        std::deque<STAFStringBufferPtr> envList;

        for (STAFProcessEnvMap::iterator iter2 = envMap.begin();
            iter2 != envMap.end(); ++iter2)
        {
            STAFProcessEnvData envData = iter2->second;

            STAFString envCombo = envData.envName + kUTF8_EQUAL +
                envData.envValue;

            STAFStringBufferPtr envComboPtr = envCombo.toCurrentCodePage();

            size += envComboPtr->length() + 1;  // Add one for the null byte
            envList.push_back(envComboPtr);
        }

        // Allocate the buffer

        size += 1;                             // Add one for the trailing null
        envBuf = new char[size];

        // Walk the list and add the entries to the buffer

        size = 0;

        for (std::deque<STAFStringBufferPtr>::iterator iter = envList.begin();
             iter != envList.end(); ++iter)
        {
            memcpy(envBuf + size, (*iter)->buffer(), (*iter)->length());
            envBuf[size + (*iter)->length()] = 0;
            size += (*iter)->length() + 1;
        }

        // Add the trailing null

        envBuf[size] = 0;

        startData.environment = (char *)envBuf;

        // Now set up the user environment variables

        userEnvArray = STAFRefPtr<STAFStringConst_t>(
                       new STAFStringConst_t[userEnvList.size()],
                       STAFRefPtr<STAFStringConst_t>::INIT);

        for (unsigned int userEnvIndex = 0; userEnvIndex < userEnvList.size();
             ++userEnvIndex)
        {
            userEnvArray[userEnvIndex] = userEnvList[userEnvIndex].getImpl();
        }

        startData.userEnvList = userEnvArray;
        startData.userEnvCount = userEnvList.size();
    }
    else startData.environment = 0;
    
    if (parsedResult->optionTimes("NEWCONSOLE") != 0)
        startData.consoleMode = kSTAFProcessNewConsole;
    else if (parsedResult->optionTimes("SAMECONSOLE") != 0)
        startData.consoleMode = kSTAFProcessSameConsole;
    else
        startData.consoleMode = gDefaultConsoleMode;

    STAFString defaultShellCommand;
    STAFString defaultSameConsoleShell;
    STAFString defaultNewConsoleShell;

    if (process->processType == kCommand)
    {
        startData.commandType = kSTAFProcessCommand;
        startData.shellCommand = 0;
        process->shellCommand = "<No Shell>";
    }
    else
    {   // process->processType == kShell
        startData.commandType = kSTAFProcessShell;
        if (process->shellCommand.length() != 0)
        {
            startData.shellCommand = process->shellCommand.getImpl();
        }
        else
        {
             // Assign the default shell command
             defaultShellCommand = STAFProcessService::getDefaultShellCommand();

             if (startData.consoleMode == kSTAFProcessSameConsole)
             {
                 defaultSameConsoleShell = STAFProcessService::getDefaultSameConsoleShell();

                 if (defaultSameConsoleShell.length() != 0)
                     startData.shellCommand = defaultSameConsoleShell.getImpl();
                 else if (defaultShellCommand.length() != 0)
                     startData.shellCommand = defaultShellCommand.getImpl();
                 else
                     startData.shellCommand = 0;
             }
             else
             {
                 // startData.consoleMode = kSTAFProcessNewConsole
                 defaultNewConsoleShell = STAFProcessService::getDefaultNewConsoleShell();

                 if (defaultNewConsoleShell.length() != 0)
                     startData.shellCommand = defaultNewConsoleShell.getImpl();
                 else if (defaultShellCommand.length() != 0)
                     startData.shellCommand = defaultShellCommand.getImpl();
                 else
                     startData.shellCommand = 0;
             }

             // Assign shell command so query a process handle shows it
             if (startData.shellCommand == 0)
                 process->shellCommand = "<Default Shell>";
             else
                 process->shellCommand = startData.shellCommand;
        }
    }

    // XXX: What to do about these conditionals?

    startData.command        = process->command.getImpl();
    startData.parms          = (process->parms.length() != 0)
                                   ? process->parms.getImpl(): 0;
    startData.title          = (process->title.length() != 0)
                                   ? process->title.getImpl() : 0;
    startData.workdir        = (process->workdir.length() != 0)
                                   ? process->workdir.getImpl() : 0;
    startData.workload       = (process->workload.length() != 0)
                                   ? process->workload.getImpl() : 0;

    STAFString defaultUsername;
    STAFString defaultPassword;

    if (process->username.length() != 0)
        startData.username = process->username.getImpl();
    else
    {
        defaultUsername = STAFProcessService::getDefaultAuthUsername();
        if (defaultUsername.length() != 0)
            startData.username = defaultUsername.getImpl();
        else
            startData.username = 0;
    }
    
    if (password.length() != 0)
        startData.password = password.getImpl();
    else
    {
        defaultPassword = STAFProcessService::getDefaultAuthPassword();
        if (defaultPassword.length() != 0)
            startData.password = defaultPassword.getImpl();
        else
            startData.password = 0;
    }

    startData.stdinRedirect  = stdInp.getImpl();
    startData.stdoutRedirect = stdOut.getImpl();
    startData.stderrRedirect = stdErr.getImpl();

    if (parsedResult->optionTimes("DISABLEDAUTHISERROR"))
        startData.disabledAuthAction = kSTAFProcessDisabledAuthError;
    else if (parsedResult->optionTimes("IGNOREDISABLEDAUTH"))
        startData.disabledAuthAction = kSTAFProcessDisabledAuthIgnore;
    else
        startData.disabledAuthAction = STAFProcessService::getDefaultDisabledAuthAction();
     
    startData.authMode = STAFProcessService::getAuthMode();

    if (parsedResult->optionTimes("FOREGROUND") != 0)
        startData.consoleFocus = kSTAFProcessForeground;

    STAFProcessEndCallbackLevel1 callback = { processTerminationCallback, 0 };

    startData.callback = &callback;

    unsigned int osRC = 0;

    // This block is here to make sure that the process is started and on
    // the list before processTerminationCallback tries to remove it from
    // the list.  We also get the HandleManager's lock so that we atomically
    // start the process and add the pending handle.

    {
        STAFMutexSemLock processLock(fProcessListSem);
        STAFMutexSemLock handleLock(gHandleManagerPtr->getHandleManagerSem());

        rc = STAFProcess::startProcess(
                 startData, process->pid, process->procHandle, osRC);
       
        if (rc == kSTAFOk)
        {
            if (process->handleType == kPending)
            {
                gHandleManagerPtr->addAndGetPendingHandle(process->handle,
                    process->pid, process->procHandle, varPool);
            }

            fProcessList[process->handle] = process;
        }
    }
    
    if (rc != kSTAFOk)
    {
        // The process failed to start, but Stdout/Stderr files (temporary or
        // not) may have been created and should be deleted before returning
        try
        {
            if (stdOut.length() != 0)
                STAFFSPath(stdOut).getEntry()->remove(&osRC);

            if (stdErr.length() != 0)
                STAFFSPath(stdErr).getEntry()->remove(&osRC);
        }
        catch (...)
        { /* Ignore any errors */ }
    }
    
    delete [] envBuf;

    if ((rc != kSTAFOk) && (process->handleType == kStatic))
        gHandleManagerPtr->removeStaticHandle(process->handle);

    if (rc == kSTAFBaseOSError) return STAFServiceResult(rc, STAFString(osRC));
    else if (rc != kSTAFOk) return rc;

    STAFString result(process->handle);

    if (startMode == kWait)
    {
        process->notify->wait(timeout);

        if (process->notify->query())
        {
            // We didn't time out waiting for the process to end
            // So remove the process from the list, remove the handle,
            // and return the RC
             
            STAFMutexSemLock processLock(fProcessListSem);

            if (process->handleType == kPending)
            {
                gHandleManagerPtr->removePendingHandle(process->handle,
                                                       process->pid);
            }
            else
            {
                gHandleManagerPtr->removeStaticHandle(process->handle);
            }

            fProcessList.erase(process->handle);
            result = STAFString(process->RC);

            // Now, tack on any files that the user wanted returned

            if (process->retFileList.size() != 0)
            {
                result += ";";
                result += STAFString(process->retFileList.size());
                result += ";";

                for (STAFProcessService::Process::FileList::iterator iter =
                         process->retFileList.begin();
                     iter != process->retFileList.end();
                     ++iter)
                {
                    STAFServiceResult fileResult = readFileIntoString(*iter);

                    result += STAFString(fileResult.fServicePerformedRC);
                    result += ":";
                    result += STAFString(fileResult.fResult.length(STAFString::kChar));
                    result += ":";
                    result += STAFString(fileResult.fResult);
                }
            }

            // Delete the temporary StdOut/Stderr files if used, with the
            // force argument set to true so it doesn't check the deleteFlags
            deleteTempFiles(*process, true);
        }
        else 
        {   
            // STAF timed out waiting for the process to end
            
            rc = kSTAFTimeout;

            // XXX: Is there a timing issue where the notify could have
            // happened before this lock such that the temp files don't
            // get deleted?

            // Set the delete flag for the temporary Stdout/Stderr files
            // if used so that the temp files will be deleted in the
            // sendNotfication() method.

            STAFMutexSemLock lock(*process->tempFileMutex);

            if (process->stdoutTempFileInfo.name.length() != 0)
                process->stdoutTempFileInfo.deleteFlag = true;

            if (process->stderrTempFileInfo.name.length() != 0)
                process->stderrTempFileInfo.deleteFlag = true;
        }
    }

    return STAFServiceResult(rc, result);
}


STAFServiceResult STAFProcessService::handleStop(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fStopParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString errorBuffer;                      
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFHandle_t theHandle = 0;
    STAFString stopMethodString;
    STAFProcessStopMethod_t stopMethod;
    bool userSpecifiedStopMethod = false;
    
    STAFRC_t rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool,
                                           theHandle, errorBuffer);
    if (!rc)
    {
        rc = resolveStringIfExists(parsedResult, "USING", handlePool,
                                   stopMethodString, errorBuffer);
    }

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (stopMethodString.length() > 0)
    {
        STAFRC_t typeRC =
            STAFProcessService::getStopMethodFromString(stopMethod, 
                                                        stopMethodString);

        if (typeRC) return STAFServiceResult(kSTAFInvalidRequestString,
                                             stopMethodString);
        userSpecifiedStopMethod = true;
    }

    STAFString result;
    STAFMutexSemLock processLock(fProcessListSem);

    if (parsedResult->optionTimes("HANDLE") != 0)
    {
        if (fProcessList.find(theHandle) == fProcessList.end())
        {
            return STAFServiceResult(kSTAFHandleDoesNotExist, 
                                     STAFString());
        }

        ProcessPtr process = fProcessList[theHandle];

        if (process->state == kComplete)
            return STAFServiceResult(kSTAFProcessAlreadyComplete, "");

        unsigned int osRC = 0;

        rc = STAFProcess::stopProcess(process->pid,
             userSpecifiedStopMethod ? stopMethod : process->stopMethod, osRC);

        if (rc == kSTAFBaseOSError)
            return STAFServiceResult(rc, STAFString(osRC));
        else if (rc != kSTAFOk)
            return rc;

        gHandleManagerPtr->unRegister(theHandle, process->pid);

    }
    else
    {
        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because output for a STOP ALL/WORKLOAD
            // request will change in STAF V3.0

            gDiagManagerPtr->record("STAF/V3.0-Mig/PROCESS STOP ALL|WORKLOAD",
                                    requestInfo.fHandleName,
                                    requestInfo.fMachine,
                                    requestInfo.fHandle);
        }

        int stopAll = parsedResult->optionTimes("ALL");
        STAFString unresWorkload = parsedResult->optionValue("WORKLOAD");
        STAFString workload;
        STAFRC_t rc = resolveString(unresWorkload, handlePool, workload,
                                         errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        workload.lowerCase();

        ProcessList::iterator iter;
        int countTotal = 0;
        int countStopped = 0;

        for (iter = fProcessList.begin(); iter != fProcessList.end(); ++iter)
        {
            ProcessPtr process = iter->second;

            if ((stopAll != 0) || (process->workload.toLowerCase() ==
                                   workload))
            {
                // Note: When stopping ALL or a workload, it is not an error
                //       to stop a kComplete process.  The implication is
                //       that only kRunning processes will be stopped.

                unsigned int osRC = 0;

                ++countTotal;

                if (process->state == kRunning)
                {
                    rc = STAFProcess::stopProcess(
                         process->pid,
                         userSpecifiedStopMethod ?
                             stopMethod :
                             process->stopMethod,
                         osRC);
                    ++countStopped;                
                }

                if (rc == kSTAFBaseOSError)
                    return STAFServiceResult(rc, STAFString(osRC));
                else if (rc != kSTAFOk)
                    return rc;

                gHandleManagerPtr->unRegister(process->handle, process->pid);
            }

        }  // end while iter valid

        result = STAFString(countStopped) + " " + STAFString(countTotal);

    }  // end if ALL or WORKLOAD

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFProcessService::handleQuery(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFString result;

    ProcessList::iterator iter;

    STAFMutexSemLock processLock(fProcessListSem);

    STAFCommandParseResultPtr parsedResult =
                              fQueryParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a QUERY request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/PROCESS QUERY",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    int queryAll = parsedResult->optionTimes("ALL");

    if ((queryAll != 0) || (parsedResult->optionTimes("WORKLOAD") != 0))
    {
        STAFString unresWorkload = parsedResult->optionValue("WORKLOAD");
        STAFString workload;
        STAFRC_t rc = resolveString(unresWorkload, handlePool, workload,
                                         errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        workload.lowerCase();

        result += "Handle  WorkLoad     Command      Status" +
                  *gLineSeparatorPtr;
        result += "------- ------------ ------------ "
                  "---------------------------------------------" +
                  *gLineSeparatorPtr;

        for (iter = fProcessList.begin(); iter != fProcessList.end(); iter++)
        {
            ProcessPtr process = iter->second;

            if ((queryAll != 0) || (process->workload.toLowerCase() ==
                                    workload))
            {
                STAFString handleString(process->handle);

                // XXX: Check these string creation statements
                result += STAFString(STAFString("        ").subString(0, 7 - 
                          handleString.length(STAFString::kChar)) +
                          handleString).subString(0, 7) + " ";

                if (process->workload == "")
                {
                    result += "<Unknown>    ";
                }
                else
                    result += STAFString(process->workload + 
                              STAFString("            ").subString(0, 12 -
                              process->workload.length(STAFString::kChar))).
                              subString(0, 12) + " ";

                result += STAFString(process->command + 
                          STAFString("            ").subString(0, 12 - 
                          process->command.length())).subString(0, 12) + " ";

                if (process->state == kComplete)
                {
                    result += "Completed at  ";
                    result += process->endStamp.asString();
                    result += " RC: " + STAFString(process->RC) +
                              *gLineSeparatorPtr;
                }
                else
                {
                    result += "Running since ";
                    result += process->startStamp.asString() +
                              *gLineSeparatorPtr;
                }
            }
        }
    }
    else  // Specific handle query
    {
        STAFString unresHandleValue = parsedResult->optionValue("HANDLE");
        STAFString handleValue;

        STAFRC_t rc = resolveString(unresHandleValue, handlePool,
                                         handleValue, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        if ((parsedResult->optionTimes("HANDLE") != 0) && 
            !handleValue.isDigits())
        {
            return STAFServiceResult(kSTAFInvalidRequestString, 
                handleValue);
        }

        STAFHandle_t handle = static_cast<STAFHandle_t>(handleValue.asUInt());

        if (fProcessList.find(handle) == fProcessList.end())
            return STAFServiceResult(kSTAFHandleDoesNotExist, "");

        ProcessPtr process = fProcessList[handle];

        result += "Handle    : " + STAFString(handle) + *gLineSeparatorPtr;
        result += "Name      : " + gHandleManagerPtr->name(handle) +
                  *gLineSeparatorPtr;
        result += "Title     : " + process->title + *gLineSeparatorPtr;
        result += "Workload  : " + process->workload + *gLineSeparatorPtr;
        result += "Shell     : " + process->shellCommand + *gLineSeparatorPtr;
        result += "Command   : " + process->command + *gLineSeparatorPtr;
        result += "Parms     : " + process->parms + *gLineSeparatorPtr;
        result += "WorkDir   : " + process->workdir + *gLineSeparatorPtr;

        if (process->notify)
            result += "Start Mode: Wait" + *gLineSeparatorPtr;
        else
            result += "Start Mode: Async" + *gLineSeparatorPtr;

        result += "Start Date: " +
                  process->startStamp.asDateString() +
                  *gLineSeparatorPtr;
        result += "Start Time: " + 
                  process->startStamp.asTimeString() +
                  *gLineSeparatorPtr;

        if (process->state == kComplete)
        {
            result += "End Date  : " +
                      process->endStamp.asDateString() +
                      *gLineSeparatorPtr;
            result += "End Time  : " + 
                      process->endStamp.asTimeString() +
                      *gLineSeparatorPtr;
            result += "RC        : " + STAFString(process->RC) +
                      *gLineSeparatorPtr;
        }
        else
        {
            result += "End Date  : " + *gLineSeparatorPtr + "End Time  : " +
                      *gLineSeparatorPtr + "RC        : " + *gLineSeparatorPtr;
        }
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFProcessService::handleFree(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fQueryParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
                                       
    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFString unresHandleValue = parsedResult->optionValue("HANDLE");
    STAFString handleValue;
    STAFRC_t rc = resolveString(unresHandleValue, handlePool, handleValue,
                                     errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    if ((parsedResult->optionTimes("HANDLE") != 0) && !handleValue.isDigits())
        return STAFServiceResult(kSTAFInvalidRequestString, handleValue);

    STAFString result;
   
    STAFMutexSemLock processLock(fProcessListSem);

    if (parsedResult->optionTimes("HANDLE") != 0)
    {
        STAFHandle_t handle = static_cast<STAFHandle_t>(handleValue.asUInt());

        if (fProcessList.find(handle) == fProcessList.end())
            return STAFServiceResult(kSTAFHandleDoesNotExist, "");

        if (fProcessList[handle]->state == kRunning)
            return STAFServiceResult(kSTAFProcessNotComplete, "");

        if (fProcessList[handle]->handleType == kPending)
        {
            gHandleManagerPtr->removePendingHandle(handle,
                                                   fProcessList[handle]->pid);
        }
        else
        {
            gHandleManagerPtr->removeStaticHandle(handle);
        }

        fProcessList.erase(handle);
    }
    else
    {
        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because output for a FREE ALL | WORKLOAD
            // request will change in STAF V3.0

            gDiagManagerPtr->record("STAF/V3.0-Mig/PROCESS FREE ALL|WORKLOAD",
                                    requestInfo.fHandleName,
                                    requestInfo.fMachine,
                                    requestInfo.fHandle);
        }

        int freeAll = parsedResult->optionTimes("ALL");
        STAFString unresWorkload = parsedResult->optionValue("WORKLOAD");
        STAFString workload;
        STAFRC_t rc = resolveString(unresWorkload, handlePool, workload,
                                         errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        workload.lowerCase();

        ProcessList::iterator iter;
        int countTotal = fProcessList.size();
        int countFreed = 0;

        for (int freedOne = 1; freedOne == 1;)
        {
            for (freedOne = 0, iter = fProcessList.begin();
                 !freedOne && iter != fProcessList.end();)
            {
                ProcessPtr process = iter->second;

                if ((freeAll != 0) || (process->workload.toLowerCase() ==
                    workload))
                {
                    // Note: When freeing ALL or a workload, it is not an error
                    //       to free a kRunning process.  The implication is
                    //       that only kComplete processes will be freed.

                    if (process->state == kComplete)
                    {
                        if (process->handleType == kPending)
                        {
                            gHandleManagerPtr->removePendingHandle(
                                process->handle, process->pid);
                        }
                        else
                        {
                            gHandleManagerPtr->removeStaticHandle(
                                process->handle);
                        }

                        fProcessList.erase(process->handle);
                        freedOne = 1;
                        ++countFreed;
                        iter = fProcessList.begin();
                    }
                }

                if (!freedOne) iter++;

            }  // end while iter valid or not freed one
        }  // end while we can still free more

        result = STAFString(countFreed) + " " + STAFString(countTotal);

    }  // end if free ALL or WORKLOAD

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFProcessService::handleNotifyRegistration(
    unsigned int isRegister, const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;
 
    STAFCommandParseResultPtr parsedResult = 
        fNotifyRegistrationParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    unsigned int useName = parsedResult->optionTimes("NAME");
    unsigned int priority = 5;
    STAFHandle_t theHandle = requestInfo.fHandle;
    STAFHandle_t processHandle = 0;
    STAFString machine = requestInfo.fMachine;
    STAFString name;

    STAFRC_t rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool,
                                           theHandle, errorBuffer);

    if (!rc) rc = resolveUIntIfExists(parsedResult, "PRIORITY", handlePool, 
                                      priority, errorBuffer);

    if (!rc) rc = resolveUIntIfExists(parsedResult, "ONENDOFHANDLE",
                                      handlePool, processHandle, errorBuffer);

    if (!rc) rc = resolveStringIfExists(parsedResult, "MACHINE", handlePool,
                                        machine, errorBuffer);

    if (!rc) rc = resolveStringIfExists(parsedResult, "NAME", handlePool, name,
                                        errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFMutexSemLock processLock(fProcessListSem);

    if (fProcessList.find(processHandle) == fProcessList.end())
        return STAFServiceResult(kSTAFHandleDoesNotExist);

    ProcessPtr process = fProcessList[processHandle];

    if (process->state == kComplete)
        return STAFServiceResult(kSTAFProcessAlreadyComplete);

    if (isRegister)
    {
        if (useName)
            rc = process->notificationList->reg(machine, name, priority);
        else
            rc = process->notificationList->reg(machine, theHandle, priority);
    }
    else
    {
        if (useName)
        {
            rc = process->notificationList->unregister(machine, name,
                                                       priority);
        }
        else
        {
            rc = process->notificationList->unregister(machine, theHandle,
                                                       priority);
        }
    }

    return rc;
}


STAFServiceResult STAFProcessService::handleNotifyList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fNotifyListParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
                                   
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a NOTIFY LIST request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/PROCESS NOTIFY LIST",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFHandle_t processHandle;

    STAFRC_t rc = resolveUIntIfExists(parsedResult, "ONENDOFHANDLE", 
                                           handlePool, processHandle, 
                                           errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    ProcessPtr process;

    // Just lock the list long enough to get a pointer to the process

    {
        STAFMutexSemLock processLock(fProcessListSem);

        if (fProcessList.find(processHandle) == fProcessList.end())
            return STAFServiceResult(kSTAFHandleDoesNotExist);

        process = fProcessList[processHandle];
    }

    STAFNotificationList::NotifieeList notifieeList(
        process->notificationList->getNotifieeListCopy());

    STAFString result;
    STAFNotificationList::NotifieeList::iterator iter;

    for (iter = notifieeList.begin(); iter != notifieeList.end(); iter++)
    {
        STAFNotificationList::Notifiee notifiee = *iter;

        result += STAFString(notifiee.priority) + ":" + notifiee.machine +
                  ":" + ((notifiee.nameOrHandle ==
                  STAFNotificationList::Notifiee::kName) ? ("NAME:" +
                  notifiee.process) : ("HANDLE:" +
                  STAFString(notifiee.handle))) + *gLineSeparatorPtr;
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFProcessService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString help("PROCESS Service Help" + *gLineSeparatorPtr +
                 *gLineSeparatorPtr);

    help += "START [SHELL [<Shell>]] COMMAND <Command> [PARMS <Parms>] "
            "[WORKDIR <Directory>]" + *gLineSeparatorPtr;
    help += "      [VAR <Variable>=<Value>] [ENV <Env. Var.>=<Value>] "
            "[USEPROCESSVARS]" + *gLineSeparatorPtr;
    help += "      [WORKLOAD <Name>] [TITLE <Title>] [WAIT [Timeout] | "
            "ASYNC]" + *gLineSeparatorPtr;
    help += "      [STOPUSING <Method>] [NEWCONSOLE | SAMECONSOLE]" +
            *gLineSeparatorPtr;
    help += "      [USERNAME <User Name> [PASSWORD <Password>]]" +
            *gLineSeparatorPtr;
    help += "      [DISABLEDAUTHISERROR | IGNOREDISABLEDAUTH] "
            "[STATICHANDLENAME <Name>]" + *gLineSeparatorPtr;
    help += "      [STDIN <File>] [STDOUT <File> | STDOUTAPPEND <File>]" +
            *gLineSeparatorPtr;
    help += "      [STDERR <File> | STDERRAPPEND <File> | STDERRTOSTDOUT]" +
            *gLineSeparatorPtr;
    help += "      [RETURNSTDOUT] [RETURNSTDERR] [RETURNFILE <File>]..." +
            *gLineSeparatorPtr;
    help += "      [NOTIFY ONEND [HANDLE <Handle> | NAME <Name>]" +
            *gLineSeparatorPtr;
    help += "      [MACHINE <Machine>] [PRIORITY <Priority>] [KEY <Key>]]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "STOP  <ALL CONFIRM | WORKLOAD <Name> | HANDLE <Handle>> "; 
    help += "[USING <Method>]" + *gLineSeparatorPtr;
    help += "QUERY <ALL | WORKLOAD <Name> | HANDLE <Handle>>" +
            *gLineSeparatorPtr;
    help += "FREE  <ALL | WORKLOAD <Name> | HANDLE <Handle>>" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "NOTIFY REGISTER   ONENDOFHANDLE <Handle> [HANDLE <Handle> | "
            "NAME <Name>]" + *gLineSeparatorPtr;
    help += "                  [MACHINE <Machine>] [PRIORITY <Priority>]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "NOTIFY UNREGISTER ONENDOFHANDLE <Handle> [HANDLE <Handle> | "
            "NAME <Name>]" + *gLineSeparatorPtr;
    help += "                  [MACHINE <Machine>] [PRIORITY <Priority>]" +
            *gLineSeparatorPtr + *gLineSeparatorPtr;
    help += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, help);
}


STAFServiceResult STAFProcessService::deleteTempFiles(
    Process &process, bool forceFlag)
{
    STAFServiceResult result = STAFServiceResult(kSTAFOk);
    STAFServiceResult deleteResult;

    STAFMutexSemLock lock(*process.tempFileMutex);

    if ((process.stdoutTempFileInfo.name.length() != 0) &&
        (forceFlag || process.stdoutTempFileInfo.deleteFlag))
    {
        // Set flag to prevent this file from being deleted again
        process.stdoutTempFileInfo.deleteFlag = false;
        
        deleteResult = deleteTempFile(process.stdoutTempFileInfo.name);

        if (deleteResult.fServicePerformedRC != kSTAFOk)
        {
            result = deleteResult;
        }
    }

    if ((process.stderrTempFileInfo.name.length() != 0) &&
        (forceFlag || process.stderrTempFileInfo.deleteFlag))
    {
        // Set flag to prevent this file from being deleted again
        process.stderrTempFileInfo.deleteFlag = false;

        deleteResult = deleteTempFile(process.stderrTempFileInfo.name);

        if ((deleteResult.fServicePerformedRC != kSTAFOk) &&
            (result.fServicePerformedRC == kSTAFOk))
        {
            result = deleteResult;
        }
    }

    return result;
}
