/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFQueueService.h"
#include "STAFHandleQueue.h"
#include "STAFTrustManager.h"
#include "STAFHandleManager.h"
#include "STAFVariablePool.h"
#include "STAFUtil.h"
#include "STAFDiagManager.h"

// QUEUE  [HANDLE <Handle>] | [NAME <Name>] [PRIORITY <Priority>]
//        MESSAGE <Message>
// GET    [PRIORITY <Priority>] [MACHINE <Machine>] [NAME <Name>]
//        [HANDLE <Handle>] [CONTAINS <String>] [WAIT]
// PEEK   [PRIORITY <Priority>] [MACHINE <Machine>] [NAME <Name>]
//        [HANDLE <Handle>] [CONTAINS <String>] [WAIT]
// DELETE [PRIORITY <Priority>]... [MACHINE <Machine>]... [NAME <Name>]...
//        [HANDLE <Handle>] [CONTAINS <String>]...>
// LIST   [HANDLE <Handle>]
//
// Returned elements are in the form
//
// Priority:YYYYMMDD-HH:MM:SS:Machine:Process:Handle:Message

STAFQueueService::STAFQueueService() : STAFService("QUEUE")
{
    // queue options

    fQueueParser.addOption("QUEUE",    1, 
        STAFCommandParser::kValueNotAllowed);
    fQueueParser.addOption("HANDLE",   1, 
        STAFCommandParser::kValueRequired);
    fQueueParser.addOption("NAME",     1, 
        STAFCommandParser::kValueRequired);
    fQueueParser.addOption("PRIORITY", 1, 
        STAFCommandParser::kValueRequired);
    fQueueParser.addOption("MESSAGE",  1, 
        STAFCommandParser::kValueRequired);

    // queue option groups

    fQueueParser.addOptionGroup("HANDLE NAME", 0, 1);
    fQueueParser.addOptionGroup("MESSAGE",     1, 1);

    // get options

    fGetParser.addOption("GET",       1, 
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("PEEK",      1, 
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("PRIORITY",  0, 
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("MACHINE",   0, 
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("NAME",      0, 
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("HANDLE",    0, 
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("CONTAINS",  0, 
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("ICONTAINS", 0,
        STAFCommandParser::kValueRequired);
    fGetParser.addOption("WAIT",      1, 
        STAFCommandParser::kValueAllowed);

    // delete options

    fDeleteParser.addOption("DELETE",    1, 
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("PRIORITY",  0, 
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("MACHINE",   0, 
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("NAME",      0, 
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("HANDLE",    0, 
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("CONTAINS",  0, 
        STAFCommandParser::kValueRequired);
    fDeleteParser.addOption("ICONTAINS", 0,
        STAFCommandParser::kValueRequired);

    // list options

    fListParser.addOption("LIST",   1, 
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("HANDLE", 1, 
        STAFCommandParser::kValueRequired);
}


STAFQueueService::~STAFQueueService()
{
    /* Do Nothing */
}


STAFString STAFQueueService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFQueueService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if      (action == "queue")  return handleQueue(requestInfo);
    else if (action == "get")    return handleGetPeek(requestInfo, true);
    else if (action == "peek")   return handleGetPeek(requestInfo, false);
    else if (action == "delete") return handleDelete(requestInfo);
    else if (action == "list")   return handleList(requestInfo);
    else if (action == "help")   return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString);
}


STAFServiceResult STAFQueueService::handleQueue(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fQueueParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString errorBuffer;                  

    // HANDLE may only default if the request came from this machine

    if ((parsedResult->optionTimes("HANDLE") == 0) &&
        (parsedResult->optionTimes("NAME") == 0) &&
        !isLocalMachine(requestInfo.fMachine))
    {
        return STAFServiceResult(kSTAFInvalidRequestString);
    }

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    STAFString name;
    unsigned int theHandle = requestInfo.fHandle;
    unsigned int priority = 5;

    rc = resolveStringIfExists(parsedResult, "NAME", handlePool, name, 
                               errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool, theHandle,
                             errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    rc = resolveUIntIfExists(parsedResult, "PRIORITY", handlePool, priority,
                             errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Currently we won't resolve variables in MESSAGE

    STAFHandleQueuePtr handleQueue;
    STAFString machine = requestInfo.fMachine;
    STAFString result;

    if ((gUseLongNames == 0) && (machine.find(kUTF8_PERIOD) != 
                                 STAFString::kNPos))
    {
        machine = machine.subString(0, machine.find(kUTF8_PERIOD));
    }

    STAFHandleQueue::Message message(priority, machine, requestInfo.fHandleName,
                                     requestInfo.fHandle,
                                     parsedResult->optionValue("MESSAGE"));

    if (parsedResult->optionTimes("NAME") == 0)
    {
        rc = gHandleManagerPtr->handleQueue(theHandle, handleQueue);

        if (rc) return STAFServiceResult(rc);

        STAFMutexSemLock queueLock(handleQueue->fQueueSem);

        if (handleQueue->fQueue.size() >= gMaxQueueSize)
        {
            return STAFServiceResult(kSTAFQueueFull,
                                     STAFString(handleQueue->fQueue.size()));
        }

        handleQueue->fQueue.insert(STAFHandleQueue::HandleQueue::value_type(
            message.priority, message));
        handleQueue->fNotify->post();
    }
    else
    {
        std::vector<STAFHandle_t> handles =
                                  gHandleManagerPtr->handlesWithName(name);

        std::vector<STAFHandle_t>::iterator iter;

        unsigned int numberQueued = handles.size();

        for(iter = handles.begin(); iter != handles.end(); iter++)
        {
            if (gHandleManagerPtr->handleQueue(*iter, handleQueue))
                --numberQueued;
            else
            {
                STAFMutexSemLock queueLock(handleQueue->fQueueSem);

                if (handleQueue->fQueue.size() >= gMaxQueueSize)
                    --numberQueued;
                else
                {
                    handleQueue->fQueue.insert(
                    STAFHandleQueue::HandleQueue::value_type(
                        message.priority, message));
                    handleQueue->fNotify->post();
                }
            }
        }

        result = STAFString(numberQueued);
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFQueueService::handleGetPeek(
    const STAFServiceRequest &requestInfo, bool isGet)
{
    // This is redundant given that GET/PEEK are only valid locally
    //
    // if (gTrustManagerPtr->trustLevel(client) < 2)
    //     return kSTAFAccessDenied;

    if (!isLocalMachine(requestInfo.fMachine)) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fGetParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a GET/PEEK request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/QUEUE GET|PEEK",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFString errorBuffer;                  
    STAFHandleQueuePtr handleQueue;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = gHandleManagerPtr->handleQueue(requestInfo.fHandle,
                                                 handleQueue);
    if (rc) return STAFServiceResult(rc);

    std::vector<unsigned int> priorityList;
    std::vector<unsigned int> handleList;
    std::vector<STAFString> machineList;
    std::vector<STAFString> nameList;
    std::vector<STAFString> containsList;
    std::vector<STAFString> icontainsList;

    unsigned int index = 0;
    unsigned int optionCount = 0;
    unsigned int priority = 0;

    for(index = 1, optionCount = parsedResult->optionTimes("PRIORITY");
        (index <= optionCount) && !rc; ++index)
    {
        rc = resolveUInt(parsedResult->optionValue("PRIORITY", index), 
                         handlePool, priority, errorBuffer);
        
        if (!rc) priorityList.push_back(priority);
    }

    STAFHandle_t theHandle = 0;

    for(index = 1, optionCount = parsedResult->optionTimes("HANDLE");
        (index <= optionCount) && !rc; ++index)
    {
        rc = resolveUInt(parsedResult->optionValue("HANDLE", index), handlePool,
                         theHandle, errorBuffer);
        if (!rc) handleList.push_back(theHandle);
    }

    STAFString machine;

    for(index = 1, optionCount = parsedResult->optionTimes("MACHINE");
        (index <= optionCount) && !rc; ++index)
    {
        rc = resolveString(parsedResult->optionValue("MACHINE", index), 
                           handlePool, machine, errorBuffer);
        if (!rc) machineList.push_back(machine.toLowerCase());
    }

    STAFString name;

    for(index = 1, optionCount = parsedResult->optionTimes("NAME");
        (index <= optionCount) && !rc; ++index)
    {
        rc = resolveString(parsedResult->optionValue("NAME", index), 
                           handlePool, name, errorBuffer);
        if (!rc) nameList.push_back(name.toLowerCase());
    }

    STAFString contains;

    for(index = 1, optionCount = parsedResult->optionTimes("CONTAINS");
        (index <= optionCount) && !rc; ++index)
    {
        rc = resolveString(parsedResult->optionValue("CONTAINS", index), 
                           handlePool, contains, errorBuffer);
        if (!rc) containsList.push_back(contains);
    }

    STAFString icontains;

    for(index = 1, optionCount = parsedResult->optionTimes("ICONTAINS");
        (index <= optionCount) && !rc; ++index)
    {
        rc = resolveString(parsedResult->optionValue("ICONTAINS", index),
                           handlePool, icontains, errorBuffer);
        if (!rc) icontainsList.push_back(icontains.toLowerCase());
    }

    unsigned int timeout = 0;

    if (!rc) rc = resolveUIntIfExists(parsedResult, "WAIT", handlePool, 
                                      timeout, errorBuffer, (unsigned int)-1);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFHandleQueue::Message message;
    unsigned int doWait = 0;
    STAFEventSemPtr event;

    if (parsedResult->optionTimes("WAIT") != 0)
    {
        STAFMutexSemLock queueLock(handleQueue->fQueueSem);

        event = handleQueue->fNotify;
        doWait = 1;
    }


    // We enter into a loop here waiting for an appropriate queue element
    // to become available.  The event->wait() condition on the end of the
    // for loop is safe because we only do multiple loops if we are in
    // WAIT mode

    STAFTimestamp startTime = STAFTimestamp::now();
    unsigned int currTimeout = timeout;

    for(int foundIt = 0; !foundIt; foundIt ? 0 : event->wait(currTimeout))
    {
        STAFMutexSemLock queueLock(handleQueue->fQueueSem);

        STAFHandleQueue::HandleQueue::iterator iter;

        for(iter = handleQueue->fQueue.begin(); 
            (iter != handleQueue->fQueue.end()) && !foundIt;)
        {
            foundIt = 1;

            if (priorityList.size() != 0)
            {
                unsigned int checkPriority = iter->second.priority;

                for(foundIt = 0, index = 0;
                    index < priorityList.size(); ++index)
                {
                    if (priorityList[index] == checkPriority)
                        foundIt = 1;
                }
            }

            if (foundIt && (handleList.size() != 0))
            {
                STAFHandle_t checkHandle = iter->second.handle;

                for(foundIt = 0, index = 0;
                    index < handleList.size(); ++index)
                {
                    if (handleList[index] == checkHandle)
                        foundIt = 1;
                }
            }

            if (foundIt && (machineList.size() != 0))
            {
                STAFString lowerMachine = iter->second.machine.toLowerCase();

                for(foundIt = 0, index = 0;
                    index < machineList.size(); ++index)
                {
                    if (machineList[index] == lowerMachine)
                        foundIt = 1;
                }
            }

            if (foundIt && (nameList.size() != 0))
            {
                STAFString lowerName = iter->second.process.toLowerCase();

                for(foundIt = 0, index = 0;
                    index < nameList.size(); ++index)
                {
                    if (nameList[index] == lowerName)
                        foundIt = 1;
                }
            }

            if (foundIt && (containsList.size() != 0))
            {
                STAFString &checkMessage = iter->second.message;

                for(foundIt = 0, index = 0;
                    index < containsList.size(); ++index)
                {
                    if (checkMessage.find(containsList[index]) != 
                        STAFString::kNPos)
                    {
                        foundIt = 1;
                    }
                }
            }

            if (foundIt && (icontainsList.size() != 0))
            {
                STAFString checkMessage = iter->second.message.toLowerCase();

                for(foundIt = 0, index = 0;
                    index < icontainsList.size(); ++index)
                {
                    if (checkMessage.find(icontainsList[index]) != 
                        STAFString::kNPos)
                    {
                        foundIt = 1;
                    }
                }
            }

            if (!foundIt) ++iter;
        }

        if (!foundIt)
        {
            if (doWait)
            {
                unsigned int deltaTime = (STAFTimestamp::now() -
                                         startTime) * 1000;

                if (deltaTime >= timeout)
                {
                    return kSTAFTimeout;
                }
                else
                {
                    currTimeout = timeout - deltaTime;
                    event->reset();
                }
            }
            else return STAFServiceResult(kSTAFNoQueueElement);
        }
        else
        {
            message = iter->second;
            if (isGet) handleQueue->fQueue.erase(iter);
        }

    }   // found an appropriate entry

    STAFString result(STAFString(message.priority) + ";" +
                      message.timestamp.asString() + ";" +
                      message.machine + ";" + message.process +
                      ";" + STAFString(message.handle) + ";" +
                      message.message);

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFQueueService::handleDelete(
    const STAFServiceRequest &requestInfo)
{
    // This is redundant given that DELETE is only valid locally
    //
    // if (gTrustManagerPtr->trustLevel(client) < 2)
    //     return kSTAFAccessDenied;

    if (!isLocalMachine(requestInfo.fMachine)) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fDeleteParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString errorBuffer;                  

    unsigned int checkPriority  = parsedResult->optionTimes("PRIORITY");
    unsigned int checkMachine   = parsedResult->optionTimes("MACHINE");
    unsigned int checkName      = parsedResult->optionTimes("NAME");
    unsigned int checkHandle    = parsedResult->optionTimes("HANDLE");
    unsigned int checkContains  = parsedResult->optionTimes("CONTAINS");
    unsigned int checkIContains = parsedResult->optionTimes("ICONTAINS");
    STAFHandleQueuePtr handleQueue;

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = gHandleManagerPtr->handleQueue(requestInfo.fHandle,
                                                 handleQueue);
    if (rc) return STAFServiceResult(rc);

    unsigned int numDeleted = 0;
    STAFMutexSemLock queueLock(handleQueue->fQueueSem);
    STAFHandleQueue::HandleQueue::iterator iter;
    int foundIt = 0;

    for(iter = handleQueue->fQueue.begin(); iter != handleQueue->fQueue.end();)
    {
        foundIt = 1;

        if (checkPriority)
        {
            unsigned int priority = 0;
            foundIt = 0;

            for(int i = 1; (i <= parsedResult->optionTimes("PRIORITY")) && 
                !foundIt; ++i)
            {
                rc = resolveUInt(parsedResult->optionValue("PRIORITY", i), 
                                 handlePool, priority, errorBuffer);

                if (rc) return STAFServiceResult(rc, errorBuffer);

                if (priority == iter->second.priority) foundIt = 1;
            }
        }

        if (foundIt && checkMachine)
        {
            STAFString machine;
            foundIt = 0;

            for(int i = 1; (i <= parsedResult->optionTimes("MACHINE")) && 
                !foundIt; ++i)
            {
                rc = resolveString(parsedResult->optionValue("MACHINE", i),
                                   handlePool, machine, errorBuffer);
                
                if (rc) return STAFServiceResult(rc, errorBuffer);

                machine.lowerCase();

                if (machine == iter->second.machine.toLowerCase())
                    foundIt = 1;
            }
        }

        if (foundIt && checkName)
        {
            STAFString name;
            foundIt = 0;

            for(int i = 1; (i <= parsedResult->optionTimes("NAME")) && 
                !foundIt; ++i)
            {
                rc = resolveString(parsedResult->optionValue("NAME", i), 
                                   handlePool, name, errorBuffer);

                if (rc) return STAFServiceResult(rc, errorBuffer);

                name.lowerCase();

                if (name == iter->second.process.toLowerCase())
                    foundIt = 1;
            }
        }

        if (foundIt && checkHandle)
        {
            unsigned int theHandle = 0;
            foundIt = 0;

            for(int i = 1; (i <= parsedResult->optionTimes("HANDLE")) && 
                !foundIt; ++i)
            {
                rc = resolveUInt(parsedResult->optionValue("HANDLE", i), 
                                 handlePool, theHandle, errorBuffer);

                if (rc) return STAFServiceResult(rc, errorBuffer);

                if (theHandle == iter->second.handle) foundIt = 1;
            }
        }

        if (foundIt && checkContains)
        {
            STAFString contains;
            foundIt = 0;

            for(int i = 1; (i <= parsedResult->optionTimes("CONTAINS")) && 
                !foundIt; ++i)
            {
                rc = resolveString(parsedResult->optionValue("CONTAINS", i),
                                   handlePool, contains, errorBuffer);

                if (rc) return STAFServiceResult(rc, errorBuffer);

                if (iter->second.message.find(contains) != STAFString::kNPos)
                    foundIt = 1;
            }
        }

        if (foundIt && checkIContains)
        {
            STAFString checkMessage = iter->second.message.toLowerCase();
            STAFString icontains;
            foundIt = 0;

            for(int i = 1; (i <= parsedResult->optionTimes("ICONTAINS")) && 
                    !foundIt; ++i)
            {
                rc = resolveString(parsedResult->optionValue("ICONTAINS", i),
                                   handlePool, icontains, errorBuffer);

                if (rc) return STAFServiceResult(rc, errorBuffer);

                if (checkMessage.find(icontains.toLowerCase())
                    != STAFString::kNPos)
                {
                    foundIt = 1;
                }
            }
        }

        if (foundIt)
        {
            ++numDeleted;
            handleQueue->fQueue.erase(iter);
            iter = handleQueue->fQueue.begin();
        }
        else
        {
            ++iter;
        }
    }

    return STAFServiceResult(kSTAFOk, STAFString(numDeleted));
}


STAFServiceResult STAFQueueService::handleList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fListParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString errorBuffer;  

    // HANDLE may only default if the request came from this machine

    if ((parsedResult->optionTimes("HANDLE") == 0) &&
        !isLocalMachine(requestInfo.fMachine))
    {
        return STAFServiceResult(kSTAFInvalidRequestString);
    }

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a LIST request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/QUEUE LIST",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    unsigned int theHandle = requestInfo.fHandle;

    rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool, theHandle,
                             errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFHandleQueuePtr handleQueue;

    rc = gHandleManagerPtr->handleQueue(theHandle, handleQueue);

    if (rc) return STAFServiceResult(rc);

    STAFMutexSemLock queueLock(handleQueue->fQueueSem);

    STAFString result;
    STAFHandleQueue::HandleQueue::iterator iter;

    for(iter = handleQueue->fQueue.begin(); iter != handleQueue->fQueue.end();
        iter++)
    {
        STAFHandleQueue::Message message = iter->second;

        result += STAFString(message.priority) + ";" +
                  message.timestamp.asString() + ";" +
                  message.machine + ";" + message.process + ";" +
                  STAFString(message.handle) + ";" +
                  message.message + *gLineSeparatorPtr;
    }

    return STAFServiceResult(rc, result);
}


STAFServiceResult STAFQueueService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1)
        return kSTAFAccessDenied;

    STAFString result("QUEUE Service Help" + *gLineSeparatorPtr +
                      *gLineSeparatorPtr);

    result += "QUEUE  [HANDLE <Handle>] | [NAME <Name>] [PRIORITY "
              "<Priority>]" + *gLineSeparatorPtr + "       MESSAGE <Message>" +
              *gLineSeparatorPtr + *gLineSeparatorPtr;
    result += "GET    [PRIORITY <Priority>]... [MACHINE <Machine>]... "
              "[NAME <Name>]..." + *gLineSeparatorPtr +
              "       [HANDLE <Handle>]... [CONTAINS <String>]... " +
              "[ICONTAINS <String>]..." + *gLineSeparatorPtr +
              "       [WAIT [Timeout]]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "PEEK   [PRIORITY <Priority>]... [MACHINE <Machine>]... "
              "[NAME <Name>]..." + *gLineSeparatorPtr +
              "       [HANDLE <Handle>]... [CONTAINS <String>]... " +
              "[ICONTAINS <String>]..." + *gLineSeparatorPtr +
              "       [WAIT [Timeout]]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "DELETE [PRIORITY <Priority>]... [MACHINE <Machine>]... "
              "[NAME <Name>]..." + *gLineSeparatorPtr +
              "       [HANDLE <Handle>]... [CONTAINS <String>]... " +
              "[ICONTAINS <String>]..." + *gLineSeparatorPtr +
              "       [WAIT [Timeout]]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "LIST   [HANDLE <Handle>]" + *gLineSeparatorPtr +
              *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
