/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAFOSTypes.h"
#include "STAFProc.h"
#include "STAFRequestManager.h"
#include "STAFService.h"
#include "STAFString.h"
#include "STAFError.h"

static STAFRequestManager *pRequestManager = 0;


STAFRequestManager::STAFRequestManager() : fNextRequest(0)
{
    pRequestManager = this;
}


STAFRC_t STAFRequestManager::add(const RequestData &requestData, 
                                 STAFReqNum_t *requestNumber)
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    *requestNumber = ++fNextRequest;
    fRequestMap[*requestNumber] = requestData;

    return kSTAFOk;
}


STAFRC_t STAFRequestManager::requestCompleted(STAFReqNum_t request, 
                                              const STAFServiceResult &result) 
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return kSTAFRequestNumberNotFound;

    RequestData &theRequest = (*theIterator).second;

    theRequest.result = result;
    theRequest.processingState = kSTAFRequestComplete;

    return kSTAFOk;
}


STAFRC_t STAFRequestManager::getResult(STAFReqNum_t request, 
                                       STAFServiceResult *result)
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    return doGetResult(request, result);
}


STAFRC_t STAFRequestManager::freeRequest(STAFReqNum_t request, 
                                         STAFServiceResult *result)
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return kSTAFRequestNumberNotFound;

    STAFRC_t rc = doGetResult(request, result);
    
    if (rc == kSTAFOk) 
        fRequestMap.erase(request);

    return rc;
}


STAFRC_t STAFRequestManager::deleteRequest(STAFReqNum_t request)                     
{
    // This method will remove the request whether or not it is complete

    STAFMutexSemLock requestLock(fRequestMapSem);

    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return kSTAFRequestNumberNotFound;

    fRequestMap.erase(request);

    return kSTAFOk;
}


STAFRC_t STAFRequestManager::getRequestData(STAFReqNum_t request, 
                                            RequestData *reqData)
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return kSTAFRequestNumberNotFound;

    *reqData = (*theIterator).second;

    return kSTAFOk;
}


STAFRC_t STAFRequestManager::getStatus(STAFReqNum_t request, 
                                       STAFRequestState_t  *status)
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    return doGetStatus(request, status);
}


bool STAFRequestManager::requestExists(STAFReqNum_t request)
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return false;

    return true;
}


STAFRequestManager::RequestMap STAFRequestManager::getRequestMapCopy()
{
    STAFMutexSemLock requestLock(fRequestMapSem);

    return fRequestMap;
}


STAFRC_t STAFRequestManager::doGetStatus(STAFReqNum_t request, 
                                         STAFRequestState_t  *status)
{
    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return kSTAFRequestNumberNotFound;

    *status = (*theIterator).second.processingState;

    return kSTAFOk;
}


STAFRC_t STAFRequestManager::doGetResult(STAFReqNum_t request, 
                                         STAFServiceResult *result)
{
    RequestMap::iterator theIterator = fRequestMap.find(request);

    if (theIterator == fRequestMap.end())
        return kSTAFRequestNumberNotFound;

    STAFRequestState_t reqStatus;
    STAFRC_t rc = doGetStatus(request, &reqStatus);

    if (rc != kSTAFOk) return rc;

    if (reqStatus != kSTAFRequestComplete)
        return kSTAFRequestNotComplete;

    RequestData &theRequest = (*theIterator).second;

    *result = theRequest.result;

    return kSTAFOk;
}
