/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#ifndef STAF_RequestManager
#define STAF_RequestManager

#include <map>
#include <vector>
#include "STAFString.h"
#include "STAFRefPtr.h"
#include "STAFOSTypes.h"
#include "STAF.h"
#include "STAFProc.h"
#include "STAFError.h"
#include "STAFMutexSem.h"
#include "STAFTimestamp.h"
#include "STAFService.h"

// STAFRequestManager - This class manages all the requests used by
//                      STAFProc.    

class STAFRequestManager
{
public:

    typedef unsigned int STAFReqNum_t;    

    typedef enum STAFRequestState_e
    {
        kSTAFRequestPending = 0,
        kSTAFRequestComplete = 1,
        kSTAFRequestStateUnknown = 2
    } STAFRequestState_t;

    struct RequestData
    {
        RequestData() 
        { /* Do Nothing */ }    

        RequestData(const STAFString &aOriginatingMachine,
                    const STAFString &aOriginatingProcess,
                    STAFHandle_t aOriginatingHandle,
                    STAFTimestamp aStartStamp,
                    const STAFString &aTargetMachine,
                    const STAFString &aTargetService,
                    const STAFString &aTargetRequest)
            : originatingMachine(aOriginatingMachine),
              originatingProcess(aOriginatingProcess),
              originatingHandle(aOriginatingHandle),
              startStamp(aStartStamp),
              targetMachine(aTargetMachine), targetService(aTargetService),
              targetRequest(aTargetRequest), processingState(kSTAFRequestPending)
        { /* Do Nothing */ }
    
        STAFString originatingMachine;
        STAFString originatingProcess;
        STAFHandle_t originatingHandle;
        STAFTimestamp startStamp;
        STAFString targetMachine;
        STAFString targetService;
        STAFString targetRequest;
        STAFRequestState_t processingState;
        STAFServiceResult result;
    };    

    typedef std::map<STAFReqNum_t, RequestData> RequestMap;

    STAFRequestManager();

    STAFRC_t add(const RequestData &requestData, STAFReqNum_t *requestNumber);
    STAFRC_t requestCompleted(STAFReqNum_t request, 
                              const STAFServiceResult &result);
    STAFRC_t getResult(STAFReqNum_t request, STAFServiceResult *result);
    STAFRC_t freeRequest(STAFReqNum_t request, STAFServiceResult *result);
    STAFRC_t deleteRequest(STAFReqNum_t request);
    STAFRC_t getRequestData(STAFReqNum_t request, RequestData *reqData);
    STAFRC_t getStatus(STAFReqNum_t request, STAFRequestState_t *status);    
    bool requestExists(STAFReqNum_t request);
    RequestMap getRequestMapCopy();

private:

    // Don't allow copy construction or assignment
    STAFRequestManager(const STAFRequestManager &);
    STAFRequestManager &operator=(const STAFRequestManager &);

    STAFRC_t doGetResult(STAFReqNum_t request, STAFServiceResult *result);
    STAFRC_t doGetStatus(STAFReqNum_t request, STAFRequestState_t *status);

    STAFReqNum_t fNextRequest;
    STAFMutexSem fRequestMapSem;
    RequestMap fRequestMap;

};

#endif
