/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFSemService.h"
#include "STAFTrustManager.h"
#include "STAFCommandParser.h"
#include "STAFDiagManager.h"
#include "STAFUtil.h"

// MUTEX <Name> <REQUEST [Timeout] | RELEASE [FORCE] | DELETE | QUERY>
// EVENT <Name> <POST | RESET | WAIT [Timeout] | DELETE | QUERY>
// LIST <MUTEX | EVENT>
// HELP

STAFSemService::STAFSemService() : STAFService("SEM")
{
    // mutex options
 
    fMutexParser.addOption("MUTEX",   1, 
        STAFCommandParser::kValueRequired);
    fMutexParser.addOption("REQUEST", 1, 
        STAFCommandParser::kValueAllowed);
    fMutexParser.addOption("RELEASE", 1,
        STAFCommandParser::kValueNotAllowed);
    fMutexParser.addOption("FORCE",   1,
        STAFCommandParser::kValueNotAllowed);
    fMutexParser.addOption("DELETE",  1,
        STAFCommandParser::kValueNotAllowed);
    fMutexParser.addOption("QUERY",   1,
        STAFCommandParser::kValueNotAllowed);
 
    // mutex groups

    fMutexParser.addOptionGroup("MUTEX", 1, 1);
    fMutexParser.addOptionGroup("REQUEST RELEASE DELETE QUERY", 1, 1);
 
    // mutex needs
 
    fMutexParser.addOptionNeed("FORCE", "RELEASE");
 
    // event options
 
    fEventParser.addOption("EVENT",  1, 
        STAFCommandParser::kValueRequired);
    fEventParser.addOption("POST",   1, 
        STAFCommandParser::kValueNotAllowed);
    fEventParser.addOption("RESET",  1, 
        STAFCommandParser::kValueNotAllowed);
    fEventParser.addOption("PULSE",  1, 
        STAFCommandParser::kValueNotAllowed);
    fEventParser.addOption("WAIT",   1, 
        STAFCommandParser::kValueAllowed);
    fEventParser.addOption("DELETE", 1, 
        STAFCommandParser::kValueNotAllowed);
    fEventParser.addOption("QUERY",  1, 
        STAFCommandParser::kValueNotAllowed);          

    // event groups
 
    fEventParser.addOptionGroup("EVENT", 1, 1);
    fEventParser.addOptionGroup("POST RESET PULSE WAIT DELETE QUERY", 1, 1);
 
    // list options
 
    fListParser.addOption("LIST",  1, 
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("MUTEX", 1, 
        STAFCommandParser::kValueNotAllowed);
    fListParser.addOption("EVENT", 1, 
        STAFCommandParser::kValueNotAllowed);
 
    // list groups
 
    fListParser.addOptionGroup("MUTEX EVENT", 1, 1);
}


STAFSemService::~STAFSemService()
{
    /* Do Nothing */
}


STAFString STAFSemService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFSemService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if      (action == "mutex") return handleMutex(requestInfo);
    else if (action == "event") return handleEvent(requestInfo);
    else if (action == "list")  return handleList(requestInfo);
    else if (action == "help")  return handleHelp(requestInfo);
    else return kSTAFInvalidRequestString;
}


STAFServiceResult STAFSemService::handleMutex(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult =
                              fMutexParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
                                         
    STAFString errorBuffer;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    STAFString result;
    STAFString semName;

    rc = resolveStringIfExists(parsedResult, "MUTEX", handlePool, semName,
                               errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    if (parsedResult->optionTimes("REQUEST") != 0)
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;
        
        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for a MUTEX REQUEST 
            // request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM MUTEX REQUEST", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        unsigned int timeout = 0;

        rc = resolveUInt(parsedResult->optionValue("REQUEST"), handlePool, 
                         timeout, errorBuffer, (unsigned int)-1);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        fMutexSemListSem.request();

        if (fMutexSemList.find(uppercaseSemName) != fMutexSemList.end())
        {
            // The sempahore exists.  Get it and lock it.

            MutexSemPtr sem = fMutexSemList[uppercaseSemName];

            sem->lockPtr->request();

            if (sem->isOwned)
            {
                // It is safe to unlock the semaphore list here, as this
                // semaphore can't be deleted with requests pending

                fMutexSemListSem.release();

                // Add a request to the list, unlock the semaphore,
                // and wait to acquire ownership

                MutexRequest request(requestInfo.fMachine,
                                     requestInfo.fHandleName,
                                     requestInfo.fHandle);
                STAFEventSemPtr myTurn = request.avail;

                sem->requestList.push_back(request);
                sem->lockPtr->release();

                myTurn->wait(timeout);

                sem->lockPtr->request();

                if (myTurn->query() == kSTAFEventSemPosted)
                {
                    // We were posted.  So, we should have been given
                    // ownership.  Yeah.

                    sem->lockPtr->release();
                }
                else
                {
                    // We timed out waiting for the semaphore, so remove
                    // us from the list

                    int foundUs = 0;
                    MutexRequestList::iterator iter;

                    for(iter = sem->requestList.begin(); !foundUs && 
                        iter != sem->requestList.end(); iter++)
                    {
                        if ((iter->machine == requestInfo.fMachine) &&
                            (iter->handleName == requestInfo.fHandleName) &&
                            (iter->handle == requestInfo.fHandle))
                        {
                            foundUs = 1;
                            sem->requestList.erase(iter);
                            break;
                        }
                    }

                    // Now unlock the semaphore and return a timeout error

                    sem->lockPtr->release();

                    return kSTAFTimeout;
                }
            }
            else
            {
                // Take ownership of the semaphore

                sem->isOwned = 1;
                sem->owner = MutexRequest(requestInfo.fMachine,
                                          requestInfo.fHandleName,
                                          requestInfo.fHandle);
                sem->acquireTimeDate = STAFTimestamp::now();

                // Now release the locks on the semphore and the list

                fMutexSemListSem.release();
                sem->lockPtr->release();
            }
        }
        else
        {
            // The semaphore doesn't exist, so create it.

            fMutexSemList[uppercaseSemName] =
                MutexSemPtr(new MutexSem(semName, 1,
                                         MutexRequest(requestInfo.fMachine,
                                                      requestInfo.fHandleName,
                                                      requestInfo.fHandle)),
                            MutexSemPtr::INIT);

            // Now release the lock on the list

            fMutexSemListSem.release();
        }
    }
    else if (parsedResult->optionTimes("RELEASE") != 0)
    {
        if (requestInfo.fTrustLevel <
            ((parsedResult->optionTimes("FORCE") != 0) ? 4 : 3))
        {
            return kSTAFAccessDenied;
        }

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for a MUTEX RELEASE 
            // request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM MUTEX RELEASE", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        fMutexSemListSem.request();

        if (fMutexSemList.find(uppercaseSemName) == fMutexSemList.end())
        {
            fMutexSemListSem.release();
            return kSTAFSemaphoreDoesNotExist;
        }

        // Ok, the semaphore exists.  Get it and lock it, then release the
        // lock on the semaphore list.

        MutexSemPtr sem = fMutexSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        fMutexSemListSem.release();

        // Verify we are allowed to release the semaphore

        if (((requestInfo.fMachine != sem->owner.machine) ||
             (requestInfo.fHandle != sem->owner.handle)) &&
            (parsedResult->optionTimes("FORCE") == 0))
        {
            return kSTAFNotSemaphoreOwner;
        }

        if (sem->requestList.size() != 0)
        {
            // There is a pending request.  Remove it from the list and
            // notify the waiter.

            sem->acquireTimeDate = STAFTimestamp::now();

            sem->owner = sem->requestList.front();
            sem->requestList.pop_front();

            sem->owner.avail->post();
        }
        else sem->isOwned = 0;
    }
    else if (parsedResult->optionTimes("DELETE") != 0)
    {
        if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for a MUTEX DELETE 
            // request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM MUTEX DELETE", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        STAFMutexSemLock lock(fMutexSemListSem);

        if (fMutexSemList.find(uppercaseSemName) == fMutexSemList.end())
            return kSTAFSemaphoreDoesNotExist;

        // Ok, the semaphore exists.  Get it and lock it.

        MutexSemPtr sem = fMutexSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        if (!sem->isOwned)
            fMutexSemList.erase(uppercaseSemName);
        else
            return kSTAFSemaphoreHasPendingRequests;
    }
    else if (parsedResult->optionTimes("QUERY") != 0)
    {
        if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the output for a MUTEX QUERY request
            // will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM MUTEX QUERY", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        fMutexSemListSem.request();

        if (fMutexSemList.find(uppercaseSemName) == fMutexSemList.end())
        {
            fMutexSemListSem.release();
            return kSTAFSemaphoreDoesNotExist;
        }

        // Ok, the semaphore exists.  Get it and lock it, then release the
        // lock on the semaphore list.

        MutexSemPtr sem = fMutexSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        fMutexSemListSem.release();

        if (sem->isOwned)
        {
            result += "Owned by: " + effectiveMachineName(sem->owner.machine)
                      + ";" + sem->owner.handleName + ";" +
                      STAFString(sem->owner.handle) + ";" +
                      sem->owner.timeDate.asString() +
                      *gLineSeparatorPtr;
            result += "Acquired: " + sem->acquireTimeDate.asString() +
                      *gLineSeparatorPtr;
            result += "Pending requests: " +
                      STAFString(sem->requestList.size()) +
                      *gLineSeparatorPtr;

            if (!sem->requestList.empty())
            {
                result += *gLineSeparatorPtr;

                MutexRequestList::iterator iter;
                int i = 1;

                for(iter = sem->requestList.begin();
                    iter != sem->requestList.end();
                    ++i, ++iter)
                {
                    MutexRequest request = *iter;

                    result += "Request " + STAFString(i) + ": " +
                              effectiveMachineName(request.machine) +
                              ";" + request.handleName + ";" +
                              STAFString(request.handle) + ";" +
                              request.timeDate.asString() +
                              *gLineSeparatorPtr;
                }
            }
        }
        else result += "Unowned" + *gLineSeparatorPtr + *gLineSeparatorPtr;
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFSemService::handleEvent(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult =
                              fEventParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }                             

    STAFString errorBuffer;

    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    STAFString result;
    STAFString semName;

    rc = resolveStringIfExists(parsedResult, "EVENT", handlePool, semName,
                               errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    STAFString uppercaseSemName = semName.toUpperCase();

    if ((parsedResult->optionTimes("POST") != 0) ||
        (parsedResult->optionTimes("PULSE") != 0))
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for an EVENT
            // POST/PULSE request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM EVENT POST|PULSE", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        fEventSemListSem.request();

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
        {
            // Semaphore does not exist, so create it in a posted state
            // if on POST, or in a reset state if on PULSE

            EventSemPtr sem = EventSemPtr(new EventSem(semName), 
                              EventSemPtr::INIT);

            sem->lastPost = EventUnit(requestInfo.fMachine,
                                      requestInfo.fHandleName,
                                      requestInfo.fHandle);

            if (parsedResult->optionTimes("POST") != 0)
            {
                sem->eventSem->post();
            }
            else
            {
                sem->eventSem->reset();
                sem->lastReset = sem->lastPost;
            }

            fEventSemList[uppercaseSemName] = sem;

            fEventSemListSem.release();
        }
        else
        {
            // The semaphore does exist, so lock the sem, unlock the main
            // list, clear the waiter list, and post/pulse it.

            EventSemPtr sem = fEventSemList[uppercaseSemName];

            sem->lockPtr->request();
            fEventSemListSem.release();

            sem->waiterList.clear();
            sem->lastPost = EventUnit(requestInfo.fMachine,
                                      requestInfo.fHandleName,
                                      requestInfo.fHandle);
            sem->eventSem->post();

            if (parsedResult->optionTimes("PULSE") != 0)
            {
                sem->eventSem->reset();
                sem->lastReset = sem->lastPost;
            }

            sem->lockPtr->release();
        }
    }
    else if (parsedResult->optionTimes("RESET") != 0)
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for an EVENT
            // RESET request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM EVENT RESET", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        fEventSemListSem.request();

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
        {
            // Semaphore does not exist, so create it in a reset state

            EventSemPtr sem = EventSemPtr(new EventSem(semName), 
                              EventSemPtr::INIT);

            sem->lastReset = EventUnit(requestInfo.fMachine,
                                       requestInfo.fHandleName,
                                       requestInfo.fHandle);
            sem->eventSem->reset();

            fEventSemList[uppercaseSemName] = sem;

            fEventSemListSem.release();
        }
        else
        {
            // The semaphore does exist, so lock the sem, unlock the main
            // list, and reset it.

            EventSemPtr sem = fEventSemList[uppercaseSemName];

            sem->lockPtr->request();
            fEventSemListSem.release();

            sem->lastReset = EventUnit(requestInfo.fMachine,
                                       requestInfo.fHandleName,
                                       requestInfo.fHandle);
            sem->eventSem->reset();

            sem->lockPtr->release();
        }
    }
    else if (parsedResult->optionTimes("WAIT") != 0)
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for an EVENT
            // WAIT request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM EVENT WAIT", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        unsigned int timeout = 0;

        rc = resolveUInt(parsedResult->optionValue("WAIT"), handlePool, timeout,
                         errorBuffer, (unsigned int)-1);
        if (rc) return STAFServiceResult(rc, errorBuffer);

        EventUnit us = EventUnit(requestInfo.fMachine,
                                 requestInfo.fHandleName,
                                 requestInfo.fHandle);
        EventSemPtr sem;

        fEventSemListSem.request();

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
        {
            // Semaphore does not exist, so create it in a reset state

            sem = EventSemPtr(new EventSem(semName), EventSemPtr::INIT);

            sem->lastReset = us;
            sem->eventSem->reset();

            fEventSemList[uppercaseSemName] = sem;
        }
        else
        {
            // The semaphore does exist, so get it.

            sem = fEventSemList[uppercaseSemName];
        }

        // Lock the semaphore, and unlock the main list.

        sem->lockPtr->request();
        fEventSemListSem.release();

        // First check to make sure that the semaphore is not already posted.
        // If it is posted, simply unlock the semaphore and return.

        if (sem->eventSem->query() == kSTAFEventSemPosted)
        {
            sem->lockPtr->release();
            return kSTAFOk;
        }

        // Ok, the semaphore is reset, so add us to the waiting list, and
        // unlock the semaphore.

        sem->waiterList.push_front(us);
        sem->lockPtr->release();

        // Now wait on it

        unsigned int waitRC = sem->eventSem->wait(timeout);

        // Re-acquire our lock on the semaphore

        sem->lockPtr->request();

        if (waitRC == 0)
        {
            // The sem was posted while we were trying to get a lock.
            // Yeah!!  We were removed from the waiting list by the
            // posting thread.

            sem->lockPtr->release();
        }
        else
        {
            // We timed out waiting for the semaphore, so remove
            // us from the waiting list

            int foundUs = 0;
            EventWaiterList::iterator iter;

            for(iter = sem->waiterList.begin(); !foundUs && 
                iter != sem->waiterList.end(); iter++)
            {
                if ((iter->machine == requestInfo.fMachine) &&
                    (iter->handleName == requestInfo.fHandleName) &&
                    (iter->handle == requestInfo.fHandle))
                {
                    foundUs = 1;
                    sem->waiterList.erase(iter);
                    break;
                }
            }

            // Now unlock the semaphore and return a timeout error

            sem->lockPtr->release();

            return kSTAFTimeout;
        }
    }
    else if (parsedResult->optionTimes("DELETE") != 0)
    {
        if (requestInfo.fTrustLevel < 4) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request syntax for an EVENT
            // DELETE request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM EVENT DELETE", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        STAFMutexSemLock lock(fEventSemListSem);

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
            return kSTAFSemaphoreDoesNotExist;

        // Ok, the semaphore exists.  Get it and lock it.

        EventSemPtr sem = fEventSemList[uppercaseSemName];
        STAFMutexSemLock semLock(*sem->lockPtr);

        if (sem->waiterList.empty())
            fEventSemList.erase(uppercaseSemName);
        else
            return kSTAFSemaphoreHasPendingRequests;
    }
    else if (parsedResult->optionTimes("QUERY") != 0)
    {
        if (gTrustManagerPtr->trustLevel(requestInfo.fMachine) < 2)
            return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the request and output for an EVENT
            // QUERY request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/SEM EVENT QUERY", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        fEventSemListSem.request();

        if (fEventSemList.find(uppercaseSemName) == fEventSemList.end())
        {
            fEventSemListSem.release();
            return kSTAFSemaphoreDoesNotExist;
        }

        // Ok, the semaphore exists.  Get it and lock it, then release the
        // lock on the semaphore list.

        EventSemPtr sem = fEventSemList[uppercaseSemName];
        STAFMutexSemLock lock(*sem->lockPtr);

        fEventSemListSem.release();

        if (sem->eventSem->query() == kSTAFEventSemPosted)
            result += "State: Posted" + *gLineSeparatorPtr + *gLineSeparatorPtr;
        else
            result += "State: Reset" + *gLineSeparatorPtr + *gLineSeparatorPtr;


        result += "Last posted: " +
                  effectiveMachineName(sem->lastPost.machine) + ";" +
                  sem->lastPost.handleName + ";" + 
                  STAFString(sem->lastPost.handle) +
                  ";" + sem->lastPost.timeDate.asString() +
                  *gLineSeparatorPtr;

        result += "Last reset : " +
                  effectiveMachineName(sem->lastReset.machine) + ";" +
                  sem->lastReset.handleName + ";" +
                  STAFString(sem->lastReset.handle) +
                  ";" + sem->lastReset.timeDate.asString() +
                  *gLineSeparatorPtr;

        if (sem->eventSem->query() == kSTAFEventSemReset)
        {
            result += *gLineSeparatorPtr + "Number of waiters: " +
                      STAFString(sem->waiterList.size()) +
                      *gLineSeparatorPtr;

            if (sem->waiterList.size() != 0)
            {
                result += *gLineSeparatorPtr;

                EventWaiterList::iterator iter;
                int i = 1;

                for(iter = sem->waiterList.begin();
                    iter != sem->waiterList.end(); ++i, ++iter)
                {
                    EventUnit eventUnit = *iter;

                    result += STAFString("Waiter ") + STAFString(i) + ": " +
                              effectiveMachineName(eventUnit.machine) +
                              ";" + eventUnit.handleName + ";" +
                              STAFString(eventUnit.handle) + ";" +
                              eventUnit.timeDate.asString() +
                              *gLineSeparatorPtr;
                }
            }
        }
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFSemService::handleList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fListParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because output for a LIST request will
        // change in STAF V3.0

        gDiagManagerPtr->record("STAF/V3.0-Mig/SEM LIST",
                                requestInfo.fHandleName,
                                requestInfo.fMachine,
                                requestInfo.fHandle);
    }

    STAFString result;

    if (parsedResult->optionTimes("MUTEX") != 0)
    {
        STAFMutexSemLock lock(fMutexSemListSem);
        MutexSemList::iterator iter;

        for(iter = fMutexSemList.begin(); iter != fMutexSemList.end(); iter++)
        {
            MutexSemPtr sem = iter->second;
            STAFMutexSemLock semLock(*sem->lockPtr);

            result += sem->name + ": ";

            if (sem->isOwned)
            {
                result += "Owned - " +
                          STAFString(sem->requestList.size()) +
                          " pending request(s)" + *gLineSeparatorPtr;
            }
            else
            {
                result += "Unowned" + *gLineSeparatorPtr;
            }
        }
    }
    else
    {
        STAFMutexSemLock lock(fEventSemListSem);
        EventSemList::iterator iter;

        for(iter = fEventSemList.begin(); iter != fEventSemList.end(); iter++)
        {
            EventSemPtr sem = iter->second;
            STAFMutexSemLock semLock(*sem->lockPtr);

            result += sem->name + ": ";

            if (sem->eventSem->query() == kSTAFEventSemReset)
            {
                result += "Reset - " +
                          STAFString(sem->waiterList.size()) +
                          " waiter(s)" + *gLineSeparatorPtr;
            }
            else
            {
                result += "Posted" + *gLineSeparatorPtr;
            }
        }
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFSemService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString result("SEM (Semaphore) Service Help" + *gLineSeparatorPtr +
                       *gLineSeparatorPtr);

    result += "MUTEX <Name> <REQUEST [Timeout] | RELEASE [FORCE] | DELETE | "
              "QUERY>" + *gLineSeparatorPtr;
    result += "EVENT <Name> <POST | RESET | PULSE | WAIT [Timeout] | DELETE | "
              "QUERY>" + *gLineSeparatorPtr;
    result += "LIST  <MUTEX | EVENT>" + *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
