/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#ifndef STAF_Service
#define STAF_Service

#include <set>
#include "STAF.h"
#include "STAFString.h"
#include "STAFRefPtr.h"
#include "STAFUtil.h"
#include "STAFMutexSem.h"

// STAFServiceRequest - Encapsulates the data sent into a service

class STAFServiceRequest
{
public:

    STAFServiceRequest()
    { /* Do Nothing */ }

    STAFString   fMachine;
    STAFHandle_t fHandle;
    STAFString   fHandleName;
    STAFString   fRequest;
    unsigned int fTrustLevel;
    unsigned int fDiagEnabled;
};


// STAFServiceResult - Encapsulates the data returned from service request

class STAFServiceResult
{
public:

    // This might be a good place to define an enumeration or some
    // constants for the servicePerformedRC and doShutdown

    STAFServiceResult(STAFRC_t servicePerformedRC,
                      const STAFString &result = STAFString(),
                      unsigned int doShutdown = 0)
        : fServicePerformedRC(servicePerformedRC),
                              fResult(result), fDoShutdown(doShutdown)
    {
        /* Do Nothing */
    }

    STAFServiceResult() : fServicePerformedRC(kSTAFOk), fDoShutdown(0), fResult()
    {
        /* Do Nothing */
    }

    STAFRC_t fServicePerformedRC;
    unsigned int fDoShutdown;
    STAFString fResult;
};


// STAFService - An interface that encapsulates a service
//
// This class provides a generic front end to any service.  A particular
// service should be derived from this class.
//
// Facilities are provided to send a request to the service and to obtain the
// name of the service.
//
// Dervied classes should override the acceptRequest() method and should
// also initialize STAFService with the correct name of the service.

class STAFService;  // Forward declaration needed for typdef

// STAFServicePtr is also defined in STAFProc.h
typedef STAFRefPtr<STAFService> STAFServicePtr;

class STAFService
{
public:

    enum ServiceState { kNotReady = 0, kInitializing, kReady, kTerminating,
                        kTerminated };
    
    enum UnregisterableState { kUnregisterable = 1, kNotUnregisterable = 0 };

    STAFService() : fIsUnregisterable(0), fServiceState(kNotReady),
                    fName("UNKNOWN")
    { /* Do Nothing */ }

    STAFService(const STAFString &name,
                unsigned int isUnregisterable = kNotUnregisterable)
        : fIsUnregisterable(isUnregisterable), fServiceState(kNotReady),
          fName(name)
    {
        fName.upperCase();
    }

    STAFServiceResult submitRequest(const STAFServiceRequest &requestInfo);

    const STAFString &name() const { return fName; }
    virtual STAFString info(unsigned int raw = 0) const = 0;
    ServiceState state();

    void setName(const STAFString &name);

    virtual unsigned int isUnregisterable() const { return fIsUnregisterable; }

    STAFRC_t initialize();
    STAFRC_t terminate(unsigned int timeout = 50000);

    virtual ~STAFService() { /* Do Nothing */ }

    typedef std::set<STAFThreadID_t> ThreadList;

protected:

    ThreadList fThreadList;
    STAFMutexSem fRequestSem;

private:

    // Don't allow copy construction or assignment
    STAFService(const STAFService &);
    STAFService &operator=(const STAFService &);

    virtual STAFServiceResult acceptRequest(
                              const STAFServiceRequest &requestInfo) = 0;
    virtual STAFRC_t init();
    virtual STAFRC_t term();

    unsigned int fIsUnregisterable;
    ServiceState fServiceState;
    STAFString fName;
};


STAF_EXCEPTION_DEFINITION(STAFServiceException, STAFException);
STAF_EXCEPTION_DEFINITION(STAFServiceCreateException, STAFServiceException);
STAF_EXCEPTION_DEFINITION(STAFServiceInitException, STAFServiceException);

#endif
