/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFServiceManager.h"
#include "STAFProc.h"

STAFServiceManager::ServiceStringList STAFServiceManager::sServiceStringList;
STAFMutexSem STAFServiceManager::sServiceStringListSem;

STAFRC_t STAFServiceManager::add(STAFServicePtr theService)
{
    STAFMutexSemLock serviceLock(fServiceListSem);

    if (fOrderedServiceList.find(theService->name()) !=
        fOrderedServiceList.end())
    {
        return kSTAFAlreadyExists;
    }

    fOrderedServiceList[theService->name()] = theService;
    fServiceList.push_back(theService);

    return kSTAFOk;
}


STAFRC_t STAFServiceManager::addSLS(STAFServicePtr theService)
{
    STAFMutexSemLock serviceLock(fSLSListSem);

    fSLSList.push_back(theService);

    return kSTAFOk;
}


STAFRC_t STAFServiceManager::remove(STAFString name)
{
    name.upperCase();

    STAFMutexSemLock serviceLock(fServiceListSem);

    if (fOrderedServiceList.find(name) == fOrderedServiceList.end())
        return kSTAFDoesNotExist;

    ServiceList::iterator listIter;

    for(listIter = fServiceList.begin(); listIter != fServiceList.end();
        ++listIter)
    {
        if ((*listIter)->name() == name)
        {
            // Don't allow removing STAF internal services 
            if (! (*listIter)->isUnregisterable())
                return kSTAFServiceNotUnregisterable;
            else
                fOrderedServiceList.erase(name);

            fServiceList.erase(listIter);
            break;
        }
    }

    return kSTAFOk;
}

STAFRC_t STAFServiceManager::removeSLS(STAFServicePtr theService)
{
    STAFMutexSemLock serviceLock(fSLSListSem);

    // XXX: This should be done with some form of erase_if or remove_if

    ServiceList::iterator listIter;

    for(listIter = fSLSList.begin(); listIter != fSLSList.end();
        ++listIter)
    {
        if ((*listIter) == theService)
        {
            fSLSList.erase(listIter);
            break;
        }
    }

    return kSTAFOk;
}


STAFRC_t STAFServiceManager::replace(STAFServicePtr theService,
                                     STAFString newName,
                                     STAFString &errorBuffer)
{
    STAFServicePtr service;
    STAFString name = theService->name();
    STAFString upperName = name.upperCase();

    {
        STAFMutexSemLock serviceLock(fServiceListSem);

        // Make sure that service being replaced already exists
        if (fOrderedServiceList.find(upperName) == fOrderedServiceList.end())
        {
            errorBuffer = name;
            return kSTAFDoesNotExist;
        }

        // If the existing service is being renamed, make sure no service
        //  already exists with this new name.
        if (newName != STAFString())
        {
            if (fOrderedServiceList.find(newName.toUpperCase()) !=
                fOrderedServiceList.end())
            {
                errorBuffer = newName;
                return kSTAFAlreadyExists; 
            }
        }

        ServiceList::iterator listIter;

        for(listIter = fServiceList.begin(); listIter != fServiceList.end();
            ++listIter)
        {
            if ((*listIter)->name() == upperName)
            {
                // Don't allow removing STAF internal services 
                if (! (*listIter)->isUnregisterable())
                {
                    errorBuffer = name;
                    return kSTAFServiceNotUnregisterable;
                }

                service = *listIter;

                fOrderedServiceList.erase(upperName);

                if (newName == STAFString())
                    fServiceList.erase(listIter);

                break;
            }
        }

        // Add the new service 
        fOrderedServiceList[theService->name()] = theService;
        fServiceList.push_back(theService);

        if (newName != STAFString())
        {
            // Change the old service's name and add back to the service list
            service->setName(newName);
            fOrderedServiceList[newName.toUpperCase()] = service;
        }
    }

    if (newName == STAFString())
    {
        // Terminate the service that was replaced
        STAFRC_t rc = service->terminate();
    }

    return kSTAFOk;
}


STAFRC_t STAFServiceManager::rename(STAFString name, STAFString newName,
                                    STAFString &errorBuffer)
{
    STAFServicePtr service;
    STAFString upperName = name.toUpperCase();

    STAFMutexSemLock serviceLock(fServiceListSem);

    // Make sure that service being rename already exists
    if (fOrderedServiceList.find(upperName) == fOrderedServiceList.end())
    {
        errorBuffer = name;
        return kSTAFDoesNotExist;
    }

    // Make sure no service already exists with this new name.
    if (fOrderedServiceList.find(newName.toUpperCase()) !=
        fOrderedServiceList.end())
    {
        errorBuffer = newName;
        return kSTAFAlreadyExists; 
    }

    ServiceList::iterator listIter;

    for(listIter = fServiceList.begin(); listIter != fServiceList.end();
        ++listIter)
    {
        if ((*listIter)->name() == upperName)
        {
            // Don't allow renaming STAF internal services 
            if (! (*listIter)->isUnregisterable())
            {
                errorBuffer = name;
                return kSTAFServiceNotUnregisterable;
            }

            service = *listIter;

            fOrderedServiceList.erase(name);
            
            // Change the old service's name & add back to the service list
            service->setName(newName);
            fOrderedServiceList.erase(upperName);
            fOrderedServiceList[newName.toUpperCase()] = service;

            break;
        }
    }

    return kSTAFOk;
}


STAFRC_t STAFServiceManager::get(STAFString name, STAFServicePtr &theService,
    bool addingService)
{
    name.upperCase();

    bool foundService = false;

    {
        STAFMutexSemLock serviceLock(fServiceListSem);

        if (fOrderedServiceList.find(name) != fOrderedServiceList.end())
        {
            theService = fOrderedServiceList[name];
            return kSTAFOk;
        }
    }
    
    // service was not found

    // If we are adding a new service, do not call the SLSs
    if (!addingService)
    {
        STAFMutexSemLock serviceLock(fSLSListSem);

        // loop through all SLSs
        ServiceList::iterator listIter;

        for(listIter = fSLSList.begin(); listIter != fSLSList.end();
            ++listIter)
        {
            // send LOAD SERVICE request
            
            STAFServiceRequest requestInfo;

            requestInfo.fMachine = "local";
            requestInfo.fHandle = gSTAFProcHandle;
            requestInfo.fHandleName = "ServiceLoader";   // ???
            requestInfo.fRequest = "LOAD SERVICE " + name;

            STAFServiceResult serviceResult =
                              (*listIter)->submitRequest(requestInfo);

            {
                STAFMutexSemLock serviceLock(fServiceListSem);

                if (fOrderedServiceList.find(name) != 
                    fOrderedServiceList.end())
                {
                    theService = fOrderedServiceList[name];
                    return kSTAFOk;
                }
            }
        }
    }

    return kSTAFDoesNotExist;
}


STAFServiceManager::OrderedServiceList STAFServiceManager::getOrderedServiceListCopy()
{
    STAFMutexSemLock serviceLock(fServiceListSem);

    return fOrderedServiceList;
}


STAFServiceManager::ServiceList STAFServiceManager::getServiceListCopy()
{
    STAFMutexSemLock serviceLock(fServiceListSem);

    return fServiceList;
}

STAFServiceManager::ServiceList STAFServiceManager::getSLSListCopy()
{
    STAFMutexSemLock serviceLock(fSLSListSem);

    return fSLSList;
}


void STAFServiceManager::traceServices(const STAFString &services)
{
    ServiceStringList newList;

    for (int i = 0; i < services.numWords(); ++i)
    {
        newList.insert(services.subWord(i, 1).upperCase());
    }

    STAFMutexSemLock lock(sServiceStringListSem);

    sServiceStringList = newList;
}


void STAFServiceManager::traceAllServices()
{
    STAFMutexSemLock lock(sServiceStringListSem);

    sServiceStringList = ServiceStringList();
}


unsigned int STAFServiceManager::doTraceService(const STAFString &service)
{
    STAFMutexSemLock lock(sServiceStringListSem);

    if (sServiceStringList.size() == 0) return 1;

    if (sServiceStringList.find(service.toUpperCase()) ==
        sServiceStringList.end())
    {
        return 0;
    }

    return 1;
}


STAFServiceManager::ServiceStringList STAFServiceManager::getTraceServices()
{
    STAFMutexSemLock lock(sServiceStringListSem);

    return sServiceStringList;
}
