/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAFOSTypes.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFError.h"
#include "STAFServiceService.h"
#include "STAFCommandParser.h"
#include "STAFTrustManager.h"
#include "STAFServiceManager.h"
#include "STAFRequestManager.h"
#include "STAFExternalService.h"
#include "STAFDiagManager.h"

// XXX: This file needs some cleanup.
//      1) Some parse results are named after the command instead of just
//         a generic name
//      2) resolveXXXIfExists isn't leveraged, so there are extra checks
//      3) Lot's of redundant empty STAFString declarations

typedef STAFRefPtr<STAFCommandParser> STAFCommandParserPtr;

STAFCommandParserPtr fListParser;
STAFCommandParserPtr fFreeParser;
STAFCommandParserPtr fQueryParser;
STAFCommandParserPtr fAddParser;
STAFCommandParserPtr fReplaceParser;
STAFCommandParserPtr fRenameParser;
STAFCommandParserPtr fRemoveParser;
STAFCommandParserPtr fHelpParser;    

static STAFString sList("list");
static STAFString sQuery("query");
static STAFString sFree("free");
static STAFString sAdd("add");
static STAFString sReplace("replace");
static STAFString sRename("rename");
static STAFString sRemove("remove");
static STAFString sForce("force");
static STAFString sHelp("help");
static STAFString sServices("services");
static STAFString sServiceLoaders("serviceloaders");
static STAFString sRequests("requests");
static STAFString sPending("pending");
static STAFString sComplete("complete");
static STAFString sStatus("status");
static STAFString sResult("result");
static STAFString sRequest("request");
static STAFString sSemiColon(";");
static STAFString sColon(":");
static STAFString sRequestNumber("Request Number: ");
static STAFString sRequestStatus("Request Status: ");
static STAFString sSourceMachine("Source Machine: ");
static STAFString sSourceProcess("Source Process: ");
static STAFString sSourceHandle("Source Handle : ");
static STAFString sStartStamp("Started       : ");
static STAFString sTargetMachine("Target Machine: ");
static STAFString sService("Service       : ");
static STAFString sServiceRequest("Request       : ");
static STAFString sReturnCode("Return Code   : ");
static STAFString sRequestResult("Result        : ");

STAFServiceService::STAFServiceService() : STAFService("SERVICE")
{        

    //LIST Options
    fListParser = STAFCommandParserPtr(new STAFCommandParser,
                                       STAFCommandParserPtr::INIT);
    fListParser->addOption("LIST", 1, 
                           STAFCommandParser::kValueNotAllowed);  
    fListParser->addOption("SERVICES", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fListParser->addOption("SERVICELOADERS", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fListParser->addOption("REQUESTS", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fListParser->addOptionGroup("SERVICES REQUESTS", 0, 1);
    fListParser->addOption("PENDING", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fListParser->addOption("COMPLETE", 1, 
                           STAFCommandParser::kValueNotAllowed);   

    //FREE Options
    fFreeParser = STAFCommandParserPtr(new STAFCommandParser,
                                       STAFCommandParserPtr::INIT);
    fFreeParser->addOption("FREE", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fFreeParser->addOption("REQUEST", 1, 
                           STAFCommandParser::kValueRequired);
    fFreeParser->addOption("FORCE", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fFreeParser->addOptionNeed("FREE", "REQUEST");

    //QUERY Options
    fQueryParser = STAFCommandParserPtr(new STAFCommandParser,
                                        STAFCommandParserPtr::INIT);
    fQueryParser->addOption("QUERY", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fQueryParser->addOption("REQUEST", 1, 
                           STAFCommandParser::kValueRequired);
    fQueryParser->addOption("STATUS", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fQueryParser->addOption("RESULT", 1, 
                           STAFCommandParser::kValueNotAllowed);
    fQueryParser->addOptionNeed("QUERY", "REQUEST");
    fQueryParser->addOptionGroup("STATUS RESULT",
                                0, 1);

    // ADD Options
    fAddParser = STAFCommandParserPtr(new STAFCommandParser,
                                      STAFCommandParserPtr::INIT);
    fAddParser->addOption("ADD", 1, STAFCommandParser::kValueNotAllowed);
    fAddParser->addOption("SERVICE", 1, STAFCommandParser::kValueRequired);
    fAddParser->addOption("LIBRARY", 1, STAFCommandParser::kValueRequired);
    fAddParser->addOption("EXECUTE", 1, STAFCommandParser::kValueRequired);
    fAddParser->addOption("OPTION", 0, STAFCommandParser::kValueRequired);
    fAddParser->addOption("PARMS", 1, STAFCommandParser::kValueRequired);
    fAddParser->addOptionNeed("ADD", "SERVICE");
    fAddParser->addOptionNeed("ADD", "LIBRARY");
    
    // REPLACE Options
    fReplaceParser = STAFCommandParserPtr(new STAFCommandParser,
                                          STAFCommandParserPtr::INIT);
    fReplaceParser->addOption("REPLACE", 1,
                              STAFCommandParser::kValueNotAllowed);
    fReplaceParser->addOption("SERVICE", 1, STAFCommandParser::kValueRequired);
    fReplaceParser->addOption("LIBRARY", 1, STAFCommandParser::kValueRequired);
    fReplaceParser->addOption("EXECUTE", 1, STAFCommandParser::kValueRequired);
    fReplaceParser->addOption("OPTION", 0, STAFCommandParser::kValueRequired);
    fReplaceParser->addOption("PARMS", 1, STAFCommandParser::kValueRequired);
    fReplaceParser->addOption("RENAME", 1, STAFCommandParser::kValueRequired);
    fReplaceParser->addOptionNeed("REPLACE", "SERVICE");
    fReplaceParser->addOptionNeed("REPLACE", "LIBRARY");

    // RENAME Options
    fRenameParser = STAFCommandParserPtr(new STAFCommandParser,
                                      STAFCommandParserPtr::INIT);
    fRenameParser->addOption("RENAME", 1, STAFCommandParser::kValueNotAllowed);
    fRenameParser->addOption("SERVICE", 1, STAFCommandParser::kValueRequired);
    fRenameParser->addOption("TO", 1, STAFCommandParser::kValueRequired);

    // REMOVE Options
    fRemoveParser = STAFCommandParserPtr(new STAFCommandParser,
                                         STAFCommandParserPtr::INIT);
    fRemoveParser->addOption("REMOVE", 1, STAFCommandParser::kValueNotAllowed);
    fRemoveParser->addOption("SERVICE", 1, STAFCommandParser::kValueRequired);
    fRemoveParser->addOptionNeed("REMOVE", "SERVICE");

    // HELP Options
    fHelpParser = STAFCommandParserPtr(new STAFCommandParser,
                                       STAFCommandParserPtr::INIT);
    fHelpParser->addOption("HELP", 1, STAFCommandParser::kValueNotAllowed);
        
}


STAFServiceService::~STAFServiceService()
{
    /* Do Nothing */
}


STAFString STAFServiceService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFServiceService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).toLowerCase(); 

    if      (action == sList)    return handleList(requestInfo);
    else if (action == sHelp)    return handleHelp(requestInfo);
    else if (action == sQuery)   return handleQuery(requestInfo);
    else if (action == sFree)    return handleFree(requestInfo);
    else if (action == sAdd)     return handleAdd(requestInfo);
    // XXX: Commented out Replace and Rename options
    // else if (action == sReplace) return handleReplace(requestInfo);
    // else if (action == sRename)  return handleRename(requestInfo);
    else if (action == sRemove)  return handleRemove(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString);
}

STAFServiceResult STAFServiceService::handleList(
    const STAFServiceRequest &requestInfo)
{    
    STAFRC_t rc = kSTAFOk;
    STAFString result;

    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedListResult =
                              fListParser->parse(requestInfo.fRequest); 
    
    if (parsedListResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedListResult->errorBuffer, 0);
    }
     
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because the output for a LIST request
        // will change in STAF V3.0

        gDiagManagerPtr->record(
            "STAF/V3.0-Mig/SERVICE LIST", requestInfo.fHandleName,
            requestInfo.fMachine, requestInfo.fHandle);
    }

    if (parsedListResult->optionTimes(sServiceLoaders) > 0)
    {
        // LIST SERVICELOADERSS
        
        STAFServiceManager::ServiceList serviceLoaderList =
            gServiceManagerPtr->getSLSListCopy();
        
        STAFServiceManager::ServiceList::iterator iter;
        
        for(iter = serviceLoaderList.begin(); 
            iter != serviceLoaderList.end(); iter++)
        {            
            result += (*iter)->info() + *gLineSeparatorPtr;
        }
    } 
    else if (parsedListResult->optionTimes(sRequests) > 0)
    {
        //  LIST REQUESTS
        
        bool pending = true;  // PENDING is the default
        bool complete = false;        

        if (parsedListResult->optionTimes(sComplete) > 0) 
        {
            complete = true;

            if (parsedListResult->optionTimes(sPending) == 0)
                pending = false;
        }     

        STAFRequestManager::RequestMap requestMap = 
                                       gRequestManagerPtr->getRequestMapCopy();
        STAFRequestManager::RequestMap::iterator theIterator; 

        STAFRequestManager::RequestData requestData;

        for (theIterator = requestMap.begin(); theIterator != requestMap.end();
             ++theIterator)
        {            
            requestData = theIterator->second;        
            STAFRequestManager::STAFRequestState_t requestState = 
                requestData.processingState;

            if ((pending &&
                 (requestState == STAFRequestManager::kSTAFRequestPending)) ||
                (complete &&
                 (requestState == STAFRequestManager::kSTAFRequestComplete)))
            {
                result += theIterator->first; 
                result += sSemiColon;    

                if (requestState == STAFRequestManager::kSTAFRequestPending)
                {
                    result += sPending;
                }
                else if (requestState ==
                         STAFRequestManager::kSTAFRequestComplete)
                {
                    result += sComplete;
                }
                
                result += sSemiColon;
                result += requestData.originatingMachine;
                result += sSemiColon;                
                result += requestData.originatingProcess;
                result += sSemiColon;
                result += requestData.originatingHandle;
                result += sSemiColon;
                result += requestData.startStamp.asString();
                result += sSemiColon;
                result += requestData.targetMachine;
                result += sSemiColon;
                result += requestData.targetService;
                result += sSemiColon;
                result += requestData.targetRequest;                

                if (requestState == STAFRequestManager::kSTAFRequestComplete)
                {
                    result += sSemiColon;
                    result += requestData.result.fServicePerformedRC;
                    result += sSemiColon;
                    result += requestData.result.fResult;
                }

                result += *gLineSeparatorPtr;            
            }
        }    
    }
    else
    {
        // LIST SERVICES (services is the default)

        STAFServiceManager::OrderedServiceList serviceList =
            gServiceManagerPtr->getOrderedServiceListCopy();
        STAFServiceManager::OrderedServiceList::iterator iter;

        for(iter = serviceList.begin(); iter != serviceList.end(); iter++)
            result += iter->second->info() + *gLineSeparatorPtr;

    }
    
    return STAFServiceResult(rc, result);
}

STAFServiceResult STAFServiceService::handleQuery(
    const STAFServiceRequest &requestInfo)
{    
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedQueryResult = 
                              fQueryParser->parse(requestInfo.fRequest); 
    
    if (parsedQueryResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedQueryResult->errorBuffer, 0);
    }
    
    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because the output for a QUERY request
        // will change in STAF V3.0

        gDiagManagerPtr->record(
            "STAF/V3.0-Mig/SERVICE QUERY", requestInfo.fHandleName,
            requestInfo.fMachine, requestInfo.fHandle);
    }

    STAFRequestManager::STAFReqNum_t reqNum;
    STAFString reqNumString = parsedQueryResult->optionValue(sRequest);

    if (!reqNumString.isDigits())
        return STAFServiceResult(kSTAFInvalidValue);
    else
        reqNum = reqNumString.asUInt();

    if (parsedQueryResult->optionTimes(sStatus))
    {
        STAFRequestManager::STAFRequestState_t statusResult = 
            STAFRequestManager::kSTAFRequestStateUnknown;    
        STAFRC_t rc = gRequestManagerPtr->getStatus(reqNum, &statusResult);
        return STAFServiceResult(rc, statusResult);
    } 
    else if (parsedQueryResult->optionTimes(sResult))
    {
        STAFServiceResult requestResult;
        STAFRC_t rc = gRequestManagerPtr->getResult(reqNum, &requestResult);
        
        if (rc == kSTAFOk) return requestResult;
        else return STAFServiceResult(rc);
    }

    // Otherwise, return all the info about the request

    STAFRequestManager::RequestData requestData;       
    STAFRC_t rc = gRequestManagerPtr->getRequestData(reqNum, &requestData);

    if (rc != kSTAFOk) return STAFServiceResult(rc);   

    STAFRequestManager::STAFRequestState_t requestState = 
                requestData.processingState;

    STAFString result = sRequestNumber;

    result += reqNum;
    result += *gLineSeparatorPtr;       
    result += sRequestStatus;

    if (requestState == STAFRequestManager::kSTAFRequestPending)
            result += sPending;
    else 
        if (requestState == STAFRequestManager::kSTAFRequestComplete)
            result += sComplete;
    
    result += *gLineSeparatorPtr;       
    result += sSourceMachine;
    result += requestData.originatingMachine;
    result += *gLineSeparatorPtr;
    result += sSourceProcess;
    result += requestData.originatingProcess;
    result += *gLineSeparatorPtr;       
    result += sSourceHandle;
    result += requestData.originatingHandle;
    result += *gLineSeparatorPtr;       
    result += sStartStamp;
    result += requestData.startStamp.asString();
    result += *gLineSeparatorPtr;
    result += sTargetMachine;
    result += requestData.targetMachine;
    result += *gLineSeparatorPtr;       
    result += sService;
    result += requestData.targetService;
    result += *gLineSeparatorPtr;       
    result += sServiceRequest;
    result += requestData.targetRequest;
    result += *gLineSeparatorPtr;
    result += sReturnCode;

    if (requestState == STAFRequestManager::kSTAFRequestComplete)
        result += requestData.result.fServicePerformedRC;

     result += *gLineSeparatorPtr;       
    result += sRequestResult;

    if (requestState == STAFRequestManager::kSTAFRequestComplete)
        result += requestData.result.fResult;

    result += *gLineSeparatorPtr;       
    
    return STAFServiceResult(rc, result); 
}

STAFServiceResult STAFServiceService::handleFree(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedFreeResult =
                              fFreeParser->parse(requestInfo.fRequest);
    
    if (parsedFreeResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedFreeResult->errorBuffer, 0);
    }

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because the output for a FREE request
        // will change in STAF V3.0

        gDiagManagerPtr->record(
            "STAF/V3.0-Mig/SERVICE FREE", requestInfo.fHandleName,
            requestInfo.fMachine, requestInfo.fHandle);
    }

    STAFServiceResult result;
    STAFRequestManager::STAFReqNum_t requestNumber;
    STAFString reqNumString = parsedFreeResult->optionValue(sRequest);

    if (!reqNumString.isDigits())    
        return STAFServiceResult(kSTAFInvalidValue);
    else
       requestNumber = parsedFreeResult->optionValue(sRequest).asUInt();

    STAFRequestManager::RequestData requestData;
    STAFRC_t rc = gRequestManagerPtr->getRequestData(reqNumString.asUInt(),
                                                     &requestData);

    if (rc != kSTAFOk) return STAFServiceResult(rc);

    // check for FORCE if this is not the originating process

    if ((effectiveMachineName(requestInfo.fMachine) !=
         requestData.originatingMachine) ||
        (requestInfo.fHandle != requestData.originatingHandle))
    {
        if (parsedFreeResult->optionTimes(sForce) == 0)
            return STAFServiceResult(kSTAFAccessDenied);
    }

    rc = gRequestManagerPtr->freeRequest(requestNumber, &result);

    if (rc != kSTAFOk)
        return STAFServiceResult(rc);
 
    STAFString freeResult = result.fServicePerformedRC + sColon + 
                            result.fResult;

    return STAFServiceResult(kSTAFOk, freeResult);
}

STAFServiceResult STAFServiceService::handleAdd(
    const STAFServiceRequest &requestInfo)
{    
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fAddParser->parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
    
    STAFServiceResult result;
    STAFString errorBuffer;                  
    STAFString serviceName;
    STAFString libraryName;
    STAFString executable;
    STAFString parms;
    STAFString option;
    STAFExternalService::OptionList options;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);

    // Resolve SERVICE value
    STAFRC_t rc = resolveString(parsedResult->optionValue("SERVICE"), 
                                handlePool, serviceName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Resolve LIBRARY value
    rc = resolveString(parsedResult->optionValue("LIBRARY"), 
                       handlePool, libraryName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Resolve EXECUTE value (if specified)
    if (parsedResult->optionTimes("EXECUTE") != 0)
    {
        rc = resolveString(parsedResult->optionValue("EXECUTE"), 
                           handlePool, executable, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
    }

    // Resolve PARMS values (if any are specified)
    if (parsedResult->optionTimes("PARMS") != 0)
    {
        rc = resolveString(parsedResult->optionValue("PARMS"), 
                           handlePool, parms, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
    }
    
    // Resolve OPTION values (if any are specified)
    for (int i = 1; i <= parsedResult->optionTimes("OPTION"); ++i)
    {
        rc = resolveString(parsedResult->optionValue("OPTION", i),
                           handlePool, option, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
        options.push_back(option);
    }
    
    // Verify that the Service name specified does not already exist
    
    STAFServicePtr service;
   
    rc = gServiceManagerPtr->get(serviceName, service, true);
    
    if (!rc) return STAFServiceResult(kSTAFAlreadyExists, serviceName);

    // Create the service
    
    try
    {
        service = STAFServicePtr(new STAFExternalService(
                  serviceName, libraryName, executable, options, parms),
                  STAFServicePtr::INIT);
    }
    catch (STAFServiceCreateException &e)
    {
        return STAFServiceResult(kSTAFServiceConfigurationError,
                                 STAFString(e.getErrorCode()) + ":" +
                                 e.getText());
    }
    catch (STAFException &e)
    {
        return STAFServiceResult(kSTAFServiceConfigurationError,
                                 STAFString(e.getErrorCode()) + ":" +
                                 e.getText());
    }
    catch (...)
    {
        return STAFServiceResult(kSTAFServiceConfigurationError,
                                 STAFString(""));
    }
    
    // Initialize the service
    rc = service->initialize();
    if (rc) return STAFServiceResult(rc, STAFString(""));

    // Add the service
    rc = gServiceManagerPtr->add(service);

    if (rc)
    {
        service->terminate();
        return STAFServiceResult(rc, serviceName);
    }
    
    return STAFServiceResult(kSTAFOk, STAFString(""));
}

STAFServiceResult STAFServiceService::handleReplace(
    const STAFServiceRequest &requestInfo)
{    
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fReplaceParser->parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFServiceResult result;
    STAFString errorBuffer;                  
    STAFString serviceName;
    STAFString libraryName;
    STAFString executable;
    STAFString parms;
    STAFString option;
    STAFString newServiceName;
    STAFExternalService::OptionList options;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);

    // Resolve SERVICE value
    STAFRC_t rc = resolveString(parsedResult->optionValue("SERVICE"), 
                                handlePool, serviceName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);
    
    // Resolve LIBRARY value
    rc = resolveString(parsedResult->optionValue("LIBRARY"), 
                       handlePool, libraryName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Resolve EXECUTE value (if specified)
    if (parsedResult->optionTimes("EXECUTE") != 0)
    {
        rc = resolveString(parsedResult->optionValue("EXECUTE"), 
                           handlePool, executable, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
    }

    // Resolve PARMS value (if specified)
    if (parsedResult->optionTimes("PARMS") != 0)
    {
        rc = resolveString(parsedResult->optionValue("PARMS"), 
                           handlePool, parms, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
    }
    
    // Resolve OPTION values (if any are specified)
    for (int i = 1; i <= parsedResult->optionTimes("OPTION"); ++i)
    {
        rc = resolveString(parsedResult->optionValue("OPTION", i),
                           handlePool, option, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
        options.push_back(option);
    }

    // Resolve RENAME value (if specified)
    
    STAFServicePtr newService;

    if (parsedResult->optionTimes("RENAME") != 0)
    {
        rc = resolveString(parsedResult->optionValue("RENAME"), 
                           handlePool, newServiceName, errorBuffer);
        if (rc) return STAFServiceResult(rc, errorBuffer);
        if (newServiceName.toUpperCase() == serviceName.toUpperCase())
        {
            return STAFServiceResult(kSTAFAlreadyExists, newServiceName);
        }
    }

    // Create the new service
    try
    {
        newService = STAFServicePtr(new STAFExternalService(
                     serviceName, libraryName, executable, options, parms),
                     STAFServicePtr::INIT);
    }
    catch (STAFServiceCreateException &e)
    {
        return STAFServiceResult(kSTAFServiceConfigurationError,
                                 STAFString(e.getErrorCode()) + ":" +
                                 e.getText());
    }
    catch (STAFException &e)
    {
        return STAFServiceResult(kSTAFServiceConfigurationError,
                                 STAFString(e.getErrorCode()) + ":" +
                                 e.getText());
    }
    catch (...)
    {
        return STAFServiceResult(kSTAFServiceConfigurationError,
                                 STAFString(""));
    }

    // Initialize the new service
    rc = newService->initialize();
    if (rc) return STAFServiceResult(rc, STAFString(""));

    // Replace with the new service
    rc = gServiceManagerPtr->replace(newService, newServiceName, errorBuffer);
    if (rc)
    {
        newService->terminate();
        return STAFServiceResult(rc, STAFString(""));
    }
    
    return STAFServiceResult(kSTAFOk, STAFString(""));
}

STAFServiceResult STAFServiceService::handleRename(
    const STAFServiceRequest &requestInfo)
{    
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fRenameParser->parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFServiceResult result;
    STAFString errorBuffer;                  
    STAFString serviceName;
    STAFString newServiceName;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);

    // Resolve SERVICE value
    STAFRC_t rc = resolveString(parsedResult->optionValue("SERVICE"), 
                                handlePool, serviceName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);

    // Resolve TO value
        
    rc = resolveString(parsedResult->optionValue("TO"), 
                       handlePool, newServiceName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);
    
    if (newServiceName.toUpperCase() == serviceName.toUpperCase())
    {
        return STAFServiceResult(kSTAFAlreadyExists, newServiceName);
    }
 
    // Rename with the new service name
    rc = gServiceManagerPtr->rename(serviceName, newServiceName, errorBuffer);
    if (rc)
    {
        return STAFServiceResult(rc, errorBuffer);
    }
    
    return kSTAFOk;
}

STAFServiceResult STAFServiceService::handleRemove(
    const STAFServiceRequest &requestInfo)
{    
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fRemoveParser->parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    STAFServiceResult result;
    STAFString errorBuffer;                  
    STAFString serviceName;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);

    // Resolve SERVICE value
    STAFRC_t rc = resolveString(parsedResult->optionValue("SERVICE"), 
                                handlePool, serviceName, errorBuffer);
    if (rc) return STAFServiceResult(rc, errorBuffer);
    
    STAFServicePtr service;
    rc = gServiceManagerPtr->get(serviceName, service);
    if (rc) return STAFServiceResult(rc, serviceName);
    
    // Remove the service
    rc = gServiceManagerPtr->remove(serviceName);
    if (rc) return STAFServiceResult(rc, serviceName);
    
    // Terminate the service
    rc = service->terminate();
    if (rc) return STAFServiceResult(rc, STAFString(""));

    return kSTAFOk;
}

STAFServiceResult STAFServiceService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString result = "SERVICE Service Help" + *gLineSeparatorPtr +
                        *gLineSeparatorPtr;

    result += 
        "LIST    [SERVICES | SERVICELOADERS | REQUESTS [PENDING] [COMPLETE]]" 
        + *gLineSeparatorPtr;
    result += "FREE    REQUEST <Request Number> [FORCE]" + *gLineSeparatorPtr;
    result += "QUERY   REQUEST <Request Number>" + *gLineSeparatorPtr;
    result += "ADD     SERVICE <Service Name> LIBRARY <Library Name>" +
              *gLineSeparatorPtr +
              "        [EXECUTE <Executable>] [OPTION <Name[=Value]>]..." +
              *gLineSeparatorPtr +
              "        [PARMS <Parameters>]" + *gLineSeparatorPtr;
    // XXX: Commented out REPLACE and RENAME options
    //result += "REPLACE SERVICE <Service Name> LIBRARY <Library Name>" +
    //          *gLineSeparatorPtr +
    //          "        [EXECUTE <Executable>] [OPTION <Name[=Value]>]..." +
    //          *gLineSeparatorPtr +
    //          "        [PARMS <Parameters>] [RENAME <Service Name>]" +
    //          *gLineSeparatorPtr;
    //result += "RENAME  SERVICE <Service Name> TO <New Service Name>" +
    //          *gLineSeparatorPtr;
    result += "REMOVE  SERVICE <Service Name>" + *gLineSeparatorPtr;
    result += "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
