/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFShutdownService.h"
#include "STAFTrustManager.h"
#include "STAFNotificationList.h"
#include "STAFDiagManager.h"
#include "STAFUtil.h"

// SHUTDOWN
// NOTIFY REGISTER   [MACHINE <Machine>] [HANDLE <Handle> | NAME <Name>]
//                   [PRIORITY <Priority>]
// NOTIFY UNREGISTER [MACHINE <Machine>] [HANDLE <Handle> | NAME <Name>]
//                   [PRIORITY <Priority>]
// NOTIFY LIST

STAFShutdownService::STAFShutdownService() : STAFService("SHUTDOWN")
{
   fShutdownParser.addOption("NOTIFY",     1,
                             STAFCommandParser::kValueNotAllowed);
   fShutdownParser.addOption("REGISTER",   1,
                             STAFCommandParser::kValueNotAllowed);
   fShutdownParser.addOption("UNREGISTER", 1,
                             STAFCommandParser::kValueNotAllowed);
   fShutdownParser.addOption("MACHINE",    1,
                             STAFCommandParser::kValueRequired);
   fShutdownParser.addOption("HANDLE",     1,
                             STAFCommandParser::kValueRequired);
   fShutdownParser.addOption("NAME",       1,
                             STAFCommandParser::kValueRequired);
   fShutdownParser.addOption("PRIORITY",   1,
                             STAFCommandParser::kValueRequired);

   fShutdownParser.addOptionGroup("HANDLE NAME", 0, 1);
}


STAFString STAFShutdownService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFShutdownService::~STAFShutdownService()
{
    /* Do Nothing */
}


STAFServiceResult STAFShutdownService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if (action == "shutdown") return handleShutdown(requestInfo);
    else if (action == "notify")
    {
        STAFString subAction = requestInfo.fRequest.subWord(1, 1).lowerCase();

        if (subAction == "register")
            return handleNotifyRegistration(true, requestInfo);
        else if (subAction == "unregister")
            return handleNotifyRegistration(false, requestInfo);
        else if (subAction == "list")
            return handleNotifyList(requestInfo);
    }
    else if (action == "help") return handleHelp(requestInfo);
    
    return STAFServiceResult(kSTAFInvalidRequestString);
}


STAFServiceResult STAFShutdownService::handleShutdown(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    return STAFServiceResult(kSTAFOk, "", 1);
}


STAFServiceResult STAFShutdownService::handleNotifyRegistration(
    bool isRegister, const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;
 
    STAFCommandParseResultPtr parsedResult =
                              fShutdownParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString errorBuffer;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    unsigned int useName = parsedResult->optionTimes("NAME");
    unsigned int theHandle = requestInfo.fHandle;
    unsigned int priority = 5;
    STAFString machine = requestInfo.fMachine;
    STAFString name;

    STAFRC_t rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool,
                                           theHandle, errorBuffer);

    if (!rc) rc = resolveUIntIfExists(parsedResult, "PRIORITY", handlePool, 
                                      priority, errorBuffer);

    if (!rc) rc = resolveStringIfExists(parsedResult, "MACHINE", handlePool, 
                                        machine, errorBuffer);

    if (!rc) rc = resolveStringIfExists(parsedResult, "NAME", handlePool, name,
                                        errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (isRegister)
    {
        if (useName) rc = gNotifyOnShutdownPtr->reg(machine, name, priority);
        else rc = gNotifyOnShutdownPtr->reg(machine, theHandle, priority);
    }
    else
    {
        if (useName)
            rc = gNotifyOnShutdownPtr->unregister(machine, name, priority);
        else
            rc = gNotifyOnShutdownPtr->unregister(machine, theHandle, 
                                                  priority);
    }

    return rc;
}


STAFServiceResult STAFShutdownService::handleNotifyList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because the output for a NOTIFY LIST request
        // will change in STAF V3.0

        gDiagManagerPtr->record(
            "STAF/V3.0-Mig/SHUTDOWN NOTIFY LIST", requestInfo.fHandleName,
            requestInfo.fMachine, requestInfo.fHandle);
    }

    STAFNotificationList::NotifieeList notifieeList(
        gNotifyOnShutdownPtr->getNotifieeListCopy());

    STAFString result = "";
    STAFNotificationList::NotifieeList::iterator iter;

    for(iter = notifieeList.begin(); iter != notifieeList.end(); iter++)
    {
        STAFNotificationList::Notifiee notifiee = *iter;

        result += STAFString(notifiee.priority) + ":" + notifiee.machine + 
                  ":" + ((notifiee.nameOrHandle ==
                  STAFNotificationList::Notifiee::kName) ? ("NAME:" +
                  notifiee.process) : ("HANDLE:" + 
                  STAFString(notifiee.handle))) + *gLineSeparatorPtr;
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFShutdownService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

   STAFString result("SHUTDOWN Service Help" + *gLineSeparatorPtr +
                      *gLineSeparatorPtr);

   result += "SHUTDOWN" + *gLineSeparatorPtr;
   result += "NOTIFY REGISTER   [MACHINE <Machine>] [HANDLE <Handle> | "
             "NAME <Name>]" + *gLineSeparatorPtr;
   result += "                  [PRIORITY <Priority>]" + *gLineSeparatorPtr;
   result += "NOTIFY UNREGISTER [MACHINE <Machine>] [HANDLE <Handle> | "
             "NAME <Name>]" + *gLineSeparatorPtr;
   result += "                  [PRIORITY <Priority>]" + *gLineSeparatorPtr;
   result += "NOTIFY LIST" + *gLineSeparatorPtr;
   result += "HELP" + *gLineSeparatorPtr;

   return STAFServiceResult(kSTAFOk, result);
}
