/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFTrustService.h"
#include "STAFCommandParser.h"
#include "STAFTrustManager.h"
#include "STAFDiagManager.h"

// SET <MACHINE <Machine> | DEFAULT> LEVEL <Level>
// GET MACHINE <Machine>
// DELETE MACHINE <Machine>
// LIST

STAFTrustService::STAFTrustService() : STAFService("TRUST")
{
    // get options
 
    fGetParser.addOption("GET",     1,
        STAFCommandParser::kValueNotAllowed);
    fGetParser.addOption("MACHINE", 1,
        STAFCommandParser::kValueRequired);
 
    // get groups
 
    fGetParser.addOptionGroup("MACHINE", 1, 1);
 
    // set options

    fSetParser.addOption("SET",     1, 
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("MACHINE", 1, 
        STAFCommandParser::kValueRequired);
    fSetParser.addOption("DEFAULT", 1, 
        STAFCommandParser::kValueNotAllowed);
    fSetParser.addOption("LEVEL",   1, 
        STAFCommandParser::kValueRequired);
 
    // set groups
 
    fSetParser.addOptionGroup("MACHINE DEFAULT", 1, 1);
    fSetParser.addOptionGroup("LEVEL", 1, 1);
 
    // delete options
 
    fDeleteParser.addOption("DELETE",  1, 
        STAFCommandParser::kValueNotAllowed);
    fDeleteParser.addOption("MACHINE", 1, 
        STAFCommandParser::kValueRequired);
 
    // delete groups
 
    fDeleteParser.addOptionGroup("MACHINE", 1, 1);           
}


STAFTrustService::~STAFTrustService()
{
    /* Do Nothing */
}


STAFString STAFTrustService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFTrustService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFString action = requestInfo.fRequest.subWord(0, 1).lowerCase();
 
    if (action == "set")
        return handleSet(requestInfo);
    else if (action == "get")
        return handleGet(requestInfo);
    else if (action == "delete")
        return handleDelete(requestInfo);
    else if (action == "list")
        return handleList(requestInfo);
    else if (action == "help") 
        return handleHelp(requestInfo);
    else return STAFServiceResult(kSTAFInvalidRequestString, "");
}


STAFServiceResult STAFTrustService::handleSet(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fSetParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    STAFString levelString = parsedResult->optionValue("LEVEL");

    if (!levelString.isDigits())
        return STAFServiceResult(kSTAFInvalidRequestString, 
                                 levelString);

    if (parsedResult->optionTimes("MACHINE") != 0)
    {
        return STAFServiceResult(gTrustManagerPtr->setTrusteeLevel(
                                 parsedResult->optionValue("MACHINE"),
                                 levelString.asUInt()));
    }
    else
    {
        return STAFServiceResult(gTrustManagerPtr->setDefaultTrusteeLevel(
                                 levelString.asUInt()));
    }
}


STAFServiceResult STAFTrustService::handleGet(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    STAFCommandParseResultPtr parsedResult =
                              fGetParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }
 
    return STAFServiceResult(kSTAFOk,
                             STAFString(gTrustManagerPtr->trustLevel(
                             parsedResult->optionValue("MACHINE"))));
}


STAFServiceResult STAFTrustService::handleDelete(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 5) return kSTAFAccessDenied;
 
    STAFCommandParseResultPtr parsedResult =
                              fDeleteParser.parse(requestInfo.fRequest);

    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer, 0);
    }

    return STAFServiceResult(gTrustManagerPtr->deleteTrustee(
                             parsedResult->optionValue("MACHINE")));
}


STAFServiceResult STAFTrustService::handleList(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

    if (requestInfo.fDiagEnabled)
    {
        // Record diagnostics because the output for a LIST request
        // will change in STAF V3.0

        gDiagManagerPtr->record(
            "STAF/V3.0-Mig/TRUST LIST", requestInfo.fHandleName,
            requestInfo.fMachine, requestInfo.fHandle);
    }

    STAFTrustManager::TrustList trustList = 
        gTrustManagerPtr->getTrustListCopy();
    
    STAFString result;
    STAFTrustManager::TrustList::iterator iter;

    result += "Default: " +
              STAFString(gTrustManagerPtr->getDefaultTrusteeLevel()) +
              *gLineSeparatorPtr + *gLineSeparatorPtr;

    for(iter = trustList.begin(); iter != trustList.end(); iter++)
    {
        result += iter->second.machine + ": " +
                  STAFString(iter->second.trustLevel) + 
                  *gLineSeparatorPtr;
    }

    return STAFServiceResult(kSTAFOk, result);
}


STAFServiceResult STAFTrustService::handleHelp(
    const STAFServiceRequest &requestInfo)
{
    if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

    STAFString result("TRUST service help" + *gLineSeparatorPtr +
                      *gLineSeparatorPtr);

    result += "SET <MACHINE <Machine> | DEFAULT> LEVEL <Level>" +
              *gLineSeparatorPtr;
    result += "GET MACHINE <Machine>" + *gLineSeparatorPtr;
    result += "DELETE MACHINE <Machine>" + *gLineSeparatorPtr;
    result += "LIST" + *gLineSeparatorPtr + "HELP" + *gLineSeparatorPtr;

    return STAFServiceResult(kSTAFOk, result);
}
