/*****************************************************************************/
/* Software Testing Automation Framework (STAF)                              */
/* (C) Copyright IBM Corp. 2001                                              */
/*                                                                           */
/* This software is licensed under the Common Public License (CPL) V1.0.     */
/*****************************************************************************/

#include "STAF.h"
#include "STAFProc.h"
#include "STAFProcUtil.h"
#include "STAFUtil.h"
#include "STAFVariableService.h"
#include "STAFVariablePool.h"
#include "STAFTrustManager.h"
#include "STAFHandleManager.h"
#include "STAFDiagManager.h"


// XXX: Might want to think about being able to create other "global" variable
//      pools by name, and then adding a POOL <Name> option to the above.

STAFVariableService::STAFVariableService() : STAFService("VAR")
{
    // var options

    fVarParser.addOption("SET",		   0,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("GET",		   1,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("DELETE",	   0,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("LIST",	   1,
                         STAFCommandParser::kValueNotAllowed);
    fVarParser.addOption("RESOLVE",	   0,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("DELIMIT",	   1,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("GLOBAL",	   1,
                         STAFCommandParser::kValueNotAllowed);
    fVarParser.addOption("HANDLE",	   1,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("ONLYHANDLE",     1,
                         STAFCommandParser::kValueRequired);
    fVarParser.addOption("HELP",	   1,
                         STAFCommandParser::kValueNotAllowed);

    // var groups

    fVarParser.addOptionGroup("SET GET DELETE LIST RESOLVE HELP", 1, 1);
    fVarParser.addOptionGroup("GLOBAL HANDLE ONLYHANDLE", 0, 1);

    // var needs

    fVarParser.addOptionNeed("DELIMIT", "RESOLVE");
    fVarParser.addOptionNeed("ONLYHANDLE", "LIST");
}


STAFVariableService::~STAFVariableService()
{
    /* Do Nothing */
}


STAFString STAFVariableService::info(unsigned int) const
{
    return name() + ": Internal";
}


STAFServiceResult STAFVariableService::acceptRequest(
    const STAFServiceRequest &requestInfo)
{
    STAFCommandParseResultPtr parsedResult =
                              fVarParser.parse(requestInfo.fRequest);
 
    if (parsedResult->rc != kSTAFOk)
    {
        return STAFServiceResult(kSTAFInvalidRequestString,
                                 parsedResult->errorBuffer);
    }
 
    STAFString errorBuffer;                  
    STAFHandle_t theHandle = requestInfo.fHandle;
    STAFVariablePoolPtr handlePool = getClientVariablePool(requestInfo.fMachine,
                                                           requestInfo.fHandle);
    STAFRC_t rc = kSTAFOk;
    STAFVariablePoolPtr mainPool = *gGlobalVariablePoolPtr;
    STAFVariablePoolPtr secondaryPool;

    // Note: We may change the local value of the handle passed into us
    //       based on whether the user specifies HANDLE or ONLYHANDLE.

    rc = resolveUIntIfExists(parsedResult, "HANDLE", handlePool, theHandle,
                             errorBuffer);

    rc = resolveUIntIfExists(parsedResult, "ONLYHANDLE", handlePool, theHandle,
                             errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    // We need to make the handle pool the primary pool and the secondary pool
    // the global pool if:
    //   1) the user specified HANDLE, or
    //   2) we are handling a request for the local machine and the user did
    //      not specify GLOBAL
    //   3) and the user did not specify ONLYHANDLE

    if (((parsedResult->optionTimes("HANDLE") != 0) ||
        ((parsedResult->optionTimes("GLOBAL") == 0) && 
          isLocalMachine(requestInfo.fMachine))) &&
	    (parsedResult->optionTimes("ONLYHANDLE") == 0))
    {
        secondaryPool = mainPool;

        rc = gHandleManagerPtr->variablePool(theHandle, mainPool);

        if (rc != kSTAFOk) return STAFServiceResult(rc);
    }

    STAFString result;
    STAFString delim(kUTF8_NULL);

    rc = resolveStringIfExists(parsedResult, "DELIMIT", handlePool, delim,
                               errorBuffer);

    if (rc) return STAFServiceResult(rc, errorBuffer);

    if (parsedResult->optionTimes("SET") != 0)
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;
        
        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the syntax for a SET request
            // will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/VAR SET", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        for(int i = 1; i <= parsedResult->optionTimes("SET"); ++i)
        {
            STAFString nameAndValue = parsedResult->optionValue("SET", i);
            unsigned int equalPos = nameAndValue.find(kUTF8_EQUAL);

            if (equalPos == STAFString::kNPos)
            {
                return STAFServiceResult(kSTAFInvalidRequestString, "");
            }

            mainPool->set(nameAndValue.subString(0, equalPos),
                          nameAndValue.subString(equalPos +
                          nameAndValue.sizeOfChar(equalPos)));
        }
    }
    else if (parsedResult->optionTimes("GET") != 0)
    {
        if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the syntax for a GET request
            // will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/VAR GET", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        rc = mainPool->get(parsedResult->optionValue("GET"), result);
    }
    else if (parsedResult->optionTimes("DELETE") != 0)
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;
        
        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the syntax for a DELETE request
            // will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/VAR DELETE", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        for(int i = 1; (i <= parsedResult->optionTimes("DELETE")) &&
            (rc == kSTAFOk); ++i)
        {
            rc = mainPool->del(parsedResult->optionValue("DELETE", i));
        }
    }
    else if (parsedResult->optionTimes("LIST") != 0)
    {
        // We need to make the handle pool the primary pool if the user
        // specified ONLYHANDLE

        if ((parsedResult->optionTimes("ONLYHANDLE")) != 0)
        {
            rc = gHandleManagerPtr->variablePool(theHandle, mainPool);

            if (rc != kSTAFOk) return STAFServiceResult(rc);
        }

        if (requestInfo.fTrustLevel < 2) return kSTAFAccessDenied;

        if (requestInfo.fDiagEnabled)
        {
            // Record diagnostics because the syntax and output for a
            // LIST request will change in STAF V3.0

            gDiagManagerPtr->record(
                "STAF/V3.0-Mig/VAR LIST", requestInfo.fHandleName,
                requestInfo.fMachine, requestInfo.fHandle);
        }

        STAFVariablePool::VariableList mainVarList =
            mainPool->getVariableListCopy();

        STAFVariablePool::VariableList secondaryVarList;

        if (secondaryPool != 0)
            secondaryVarList = secondaryPool->getVariableListCopy();

        // First, find all variables in secondary list that are not in the
        // main list and add them to the result

        STAFVariablePool::VariableList::iterator secondaryIter;

        for(secondaryIter = secondaryVarList.begin();
            secondaryIter != secondaryVarList.end(); secondaryIter++)
        {
            const STAFVariablePool::Variable &var = secondaryIter->second;

            if (mainVarList.find(var.lowerCaseName) == mainVarList.end())
            {
                result += var.name + "=" + var.value +
                          *gLineSeparatorPtr;
            }
        }

        // Now, add all the variables in the main list to the result

        STAFVariablePool::VariableList::iterator mainIter;

        for(mainIter = mainVarList.begin(); mainIter != mainVarList.end();
            mainIter++)
        {
            const STAFVariablePool::Variable &var = mainIter->second;

            result += var.name + "=" + var.value + *gLineSeparatorPtr;
        }
    }
    else if (parsedResult->optionTimes("RESOLVE") != 0)
    {
        if (requestInfo.fTrustLevel < 3) return kSTAFAccessDenied;

        unsigned int numTimes = parsedResult->optionTimes("RESOLVE");

        if (numTimes == 1)
        {
            if (requestInfo.fDiagEnabled)
            {
                // Record diagnostics because the syntax for a RESOLVE
                // request will change in STAF V3.0

                gDiagManagerPtr->record(
                    "STAF/V3.0-Mig/VAR RESOLVE", requestInfo.fHandleName,
                    requestInfo.fMachine, requestInfo.fHandle);
            }

            rc = mainPool->resolve(parsedResult->optionValue("RESOLVE"), 
                                   secondaryPool, result);
        }
        else
        {
            if (requestInfo.fDiagEnabled)
            {
                // Record diagnostics because the syntax and output for a
                // request with more than one RESOLVE option will change
                // in STAF V3.0

                gDiagManagerPtr->record(
                    "STAF/V3.0-Mig/VAR RESOLVE (Multiple)", requestInfo.fHandleName,
                    requestInfo.fMachine, requestInfo.fHandle);
            }

            STAFString fullResult;

            for(int i = 1; i <= numTimes; ++i)
            {
                rc = mainPool->resolve(parsedResult->optionValue("RESOLVE", i),
                                       secondaryPool, result);
                fullResult += STAFString(rc) + ":" + result + delim;
            }

            rc = kSTAFOk;
            result = fullResult;
        }
    }
    else  // HELP
    {
        if (requestInfo.fTrustLevel < 1) return kSTAFAccessDenied;

        result = "VAR Service Help" + *gLineSeparatorPtr + *gLineSeparatorPtr;
        result += "[GLOBAL | HANDLE <Handle>] SET <Name=Value> "
                  "[SET <Name=Value>] ..." + *gLineSeparatorPtr;
        result += "[GLOBAL | HANDLE <Handle>] GET <Name>" + *gLineSeparatorPtr;
        result += "[GLOBAL | HANDLE <Handle>] DELETE <Name> [DELETE <Name>] "
                  "..." + *gLineSeparatorPtr;
        result += "[GLOBAL | HANDLE <Handle> | ONLYHANDLE <Handle>] LIST" +
                  *gLineSeparatorPtr;
        result += "[GLOBAL | HANDLE <Handle>] RESOLVE <String> "
                  "[RESOLVE <String>] ..." + *gLineSeparatorPtr +
                  "                           [DELIMIT <Delimiter>]" +
                  *gLineSeparatorPtr;
        result += "HELP" + *gLineSeparatorPtr;
    }

    return STAFServiceResult(rc, result);
}
